Package: ngme2
Type: Package
Title: Linear Latent Non-Gaussian Models with Flexible Distributions
Version: 0.9.8
Authors@R: c(
    person("David", "Bolin", email = "davidbolin@gmail.com", role = c("aut", "cph")),
    person("Xiaotian", "Jin", email = "xiaotian.jin@kaust.edu.sa", role =c("aut", "cre")),
    person("Alexandre", "Simas", email = "alexandre.impa@gmail.com", role = "aut"),
    person("Jonas", "Wallin", email = "jonas.wallin81@gmail.com", role = "aut"),
    person("Andrea V.", "Rocha", email = "dearocha@yahoo.com", role = "ctb", comment = "contributed to many vignettes and examples"),
    person("Timothy A.", "Davis", role = c("ctb", "cph"), comment = "SuiteSparse components"),
    person("Patrick R.", "Amestoy", role = c("ctb", "cph"), comment = "SuiteSparse AMD and CAMD components"),
    person("Iain S.", "Duff", role = c("ctb", "cph"), comment = "SuiteSparse AMD and CAMD components"),
    person("John K.", "Reid", role = c("ctb", "cph"), comment = "SuiteSparse AMD-derived code"),
    person("Yanqing", "Chen", role = c("ctb", "cph"), comment = "SuiteSparse CAMD components"),
    person("Sivasankaran", "Rajamanickam", role = c("ctb", "cph"), comment = "SuiteSparse CCOLAMD components"),
    person("Stefan", "Larimore", role = c("ctb", "cph"), comment = "SuiteSparse CCOLAMD and COLAMD components"),
    person("William W.", "Hager", role = c("ctb", "cph"), comment = "SuiteSparse CHOLMOD Modify components"),
    person("University of Florida", role = "cph", comment = "SuiteSparse CHOLMOD and CCOLAMD components"),
    person("Regents of the University of Minnesota", role = "cph", comment = "METIS/GKlib components bundled with SuiteSparse"),
    person("Free Software Foundation, Inc.", role = "cph", comment = "GKlib regex/getopt code bundled with SuiteSparse"),
    person("Makoto", "Matsumoto", role = c("ctb", "cph"), comment = "Mersenne Twister code bundled with SuiteSparse"),
    person("Takuji", "Nishimura", role = c("ctb", "cph"), comment = "Mersenne Twister code bundled with SuiteSparse"),
    person("Alexander", "Chemeris", role = c("ctb", "cph"), comment = "MS integer headers bundled with SuiteSparse"))
Maintainer: Xiaotian Jin <xiaotian.jin@kaust.edu.sa>
Description: Fits and analyzes linear latent non-Gaussian models for
    temporal, spatial, and space-time data. The package provides model
    components for autoregressive and Ornstein-Uhlenbeck processes, random
    walks, Matern fields based on stochastic partial differential equations,
    separable and non-separable space-time models, graph-based Matern models,
    bivariate type-G fields, and user-defined sparse operators. Latent fields
    and observation models can use Gaussian and non-Gaussian noise
    distributions, including normal inverse Gaussian, generalized asymmetric
    Laplace, and skew-t distributions. Functions are included for simulation,
    likelihood-based estimation, prediction, cross-validation, convergence
    diagnostics, stochastic gradient optimization, batch-means confidence
    intervals, and posterior-like sampling. The modeling framework is described
    in Bolin, Jin, Simas and Wallin (2026) "A Unified and Computationally
    Efficient Non-Gaussian Statistical Modeling Framework"
    <doi:10.48550/arXiv.2602.23987>.
License: GPL (>= 3)
Encoding: UTF-8
LazyData: true
Imports: Matrix, Rcpp, methods, rlang, ggplot2, graphics, fmesher,
        stats, gridExtra, withr
LinkingTo: Rcpp, RcppEigen,
Suggests: R.rsp, knitr, rmarkdown, testthat (>= 3.0.0), MASS, dplyr,
        fields, inlabru, MetricGraph, rSPDE, sf, zoo, mvtnorm
VignetteBuilder: knitr
RoxygenNote: 7.3.2
Config/testthat/edition: 3
URL: https://davidbolin.github.io/ngme2/,
        https://github.com/davidbolin/ngme2
Depends: R (>= 3.5.0)
BugReports: https://github.com/davidbolin/ngme2/issues
Copyright: The R package and code, and the main programs, were written
        by and are Copyright by David Bolin, Xiaotian Jin, Alexandre
        Simas, and Jonas Wallin, and are redistributable under the GNU
        Public License, version 3 or later. The package also includes
        bundled SuiteSparse components. For details see
        inst/COPYRIGHTS.
NeedsCompilation: yes
Packaged: 2026-05-19 19:01:57 UTC; jinx0a
Author: David Bolin [aut, cph],
  Xiaotian Jin [aut, cre],
  Alexandre Simas [aut],
  Jonas Wallin [aut],
  Andrea V. Rocha [ctb] (contributed to many vignettes and examples),
  Timothy A. Davis [ctb, cph] (SuiteSparse components),
  Patrick R. Amestoy [ctb, cph] (SuiteSparse AMD and CAMD components),
  Iain S. Duff [ctb, cph] (SuiteSparse AMD and CAMD components),
  John K. Reid [ctb, cph] (SuiteSparse AMD-derived code),
  Yanqing Chen [ctb, cph] (SuiteSparse CAMD components),
  Sivasankaran Rajamanickam [ctb, cph] (SuiteSparse CCOLAMD components),
  Stefan Larimore [ctb, cph] (SuiteSparse CCOLAMD and COLAMD components),
  William W. Hager [ctb, cph] (SuiteSparse CHOLMOD Modify components),
  University of Florida [cph] (SuiteSparse CHOLMOD and CCOLAMD
    components),
  Regents of the University of Minnesota [cph] (METIS/GKlib components
    bundled with SuiteSparse),
  Free Software Foundation, Inc. [cph] (GKlib regex/getopt code bundled
    with SuiteSparse),
  Makoto Matsumoto [ctb, cph] (Mersenne Twister code bundled with
    SuiteSparse),
  Takuji Nishimura [ctb, cph] (Mersenne Twister code bundled with
    SuiteSparse),
  Alexander Chemeris [ctb, cph] (MS integer headers bundled with
    SuiteSparse)
Repository: CRAN
Date/Publication: 2026-05-20 07:40:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-25 01:54:43 UTC; unix
Archs: ngme2.so.dSYM

Package: ngme2
Type: Package
Title: Linear Latent Non-Gaussian Models with Flexible Distributions
Version: 0.9.8
Authors@R: c(
    person("David", "Bolin", email = "davidbolin@gmail.com", role = c("aut", "cph")),
    person("Xiaotian", "Jin", email = "xiaotian.jin@kaust.edu.sa", role =c("aut", "cre")),
    person("Alexandre", "Simas", email = "alexandre.impa@gmail.com", role = "aut"),
    person("Jonas", "Wallin", email = "jonas.wallin81@gmail.com", role = "aut"),
    person("Andrea V.", "Rocha", email = "dearocha@yahoo.com", role = "ctb", comment = "contributed to many vignettes and examples"),
    person("Timothy A.", "Davis", role = c("ctb", "cph"), comment = "SuiteSparse components"),
    person("Patrick R.", "Amestoy", role = c("ctb", "cph"), comment = "SuiteSparse AMD and CAMD components"),
    person("Iain S.", "Duff", role = c("ctb", "cph"), comment = "SuiteSparse AMD and CAMD components"),
    person("John K.", "Reid", role = c("ctb", "cph"), comment = "SuiteSparse AMD-derived code"),
    person("Yanqing", "Chen", role = c("ctb", "cph"), comment = "SuiteSparse CAMD components"),
    person("Sivasankaran", "Rajamanickam", role = c("ctb", "cph"), comment = "SuiteSparse CCOLAMD components"),
    person("Stefan", "Larimore", role = c("ctb", "cph"), comment = "SuiteSparse CCOLAMD and COLAMD components"),
    person("William W.", "Hager", role = c("ctb", "cph"), comment = "SuiteSparse CHOLMOD Modify components"),
    person("University of Florida", role = "cph", comment = "SuiteSparse CHOLMOD and CCOLAMD components"),
    person("Regents of the University of Minnesota", role = "cph", comment = "METIS/GKlib components bundled with SuiteSparse"),
    person("Free Software Foundation, Inc.", role = "cph", comment = "GKlib regex/getopt code bundled with SuiteSparse"),
    person("Makoto", "Matsumoto", role = c("ctb", "cph"), comment = "Mersenne Twister code bundled with SuiteSparse"),
    person("Takuji", "Nishimura", role = c("ctb", "cph"), comment = "Mersenne Twister code bundled with SuiteSparse"),
    person("Alexander", "Chemeris", role = c("ctb", "cph"), comment = "MS integer headers bundled with SuiteSparse"))
Maintainer: Xiaotian Jin <xiaotian.jin@kaust.edu.sa>
Description: Fits and analyzes linear latent non-Gaussian models for
    temporal, spatial, and space-time data. The package provides model
    components for autoregressive and Ornstein-Uhlenbeck processes, random
    walks, Matern fields based on stochastic partial differential equations,
    separable and non-separable space-time models, graph-based Matern models,
    bivariate type-G fields, and user-defined sparse operators. Latent fields
    and observation models can use Gaussian and non-Gaussian noise
    distributions, including normal inverse Gaussian, generalized asymmetric
    Laplace, and skew-t distributions. Functions are included for simulation,
    likelihood-based estimation, prediction, cross-validation, convergence
    diagnostics, stochastic gradient optimization, batch-means confidence
    intervals, and posterior-like sampling. The modeling framework is described
    in Bolin, Jin, Simas and Wallin (2026) "A Unified and Computationally
    Efficient Non-Gaussian Statistical Modeling Framework"
    <doi:10.48550/arXiv.2602.23987>.
License: GPL (>= 3)
Encoding: UTF-8
LazyData: true
Imports: Matrix, Rcpp, methods, rlang, ggplot2, graphics, fmesher,
        stats, gridExtra, withr
LinkingTo: Rcpp, RcppEigen,
Suggests: R.rsp, knitr, rmarkdown, testthat (>= 3.0.0), MASS, dplyr,
        fields, inlabru, MetricGraph, rSPDE, sf, zoo, mvtnorm
VignetteBuilder: knitr
RoxygenNote: 7.3.2
Config/testthat/edition: 3
URL: https://davidbolin.github.io/ngme2/,
        https://github.com/davidbolin/ngme2
Depends: R (>= 3.5.0)
BugReports: https://github.com/davidbolin/ngme2/issues
Copyright: The R package and code, and the main programs, were written
        by and are Copyright by David Bolin, Xiaotian Jin, Alexandre
        Simas, and Jonas Wallin, and are redistributable under the GNU
        Public License, version 3 or later. The package also includes
        bundled SuiteSparse components. For details see
        inst/COPYRIGHTS.
NeedsCompilation: yes
Packaged: 2026-05-19 19:01:57 UTC; jinx0a
Author: David Bolin [aut, cph],
  Xiaotian Jin [aut, cre],
  Alexandre Simas [aut],
  Jonas Wallin [aut],
  Andrea V. Rocha [ctb] (contributed to many vignettes and examples),
  Timothy A. Davis [ctb, cph] (SuiteSparse components),
  Patrick R. Amestoy [ctb, cph] (SuiteSparse AMD and CAMD components),
  Iain S. Duff [ctb, cph] (SuiteSparse AMD and CAMD components),
  John K. Reid [ctb, cph] (SuiteSparse AMD-derived code),
  Yanqing Chen [ctb, cph] (SuiteSparse CAMD components),
  Sivasankaran Rajamanickam [ctb, cph] (SuiteSparse CCOLAMD components),
  Stefan Larimore [ctb, cph] (SuiteSparse CCOLAMD and COLAMD components),
  William W. Hager [ctb, cph] (SuiteSparse CHOLMOD Modify components),
  University of Florida [cph] (SuiteSparse CHOLMOD and CCOLAMD
    components),
  Regents of the University of Minnesota [cph] (METIS/GKlib components
    bundled with SuiteSparse),
  Free Software Foundation, Inc. [cph] (GKlib regex/getopt code bundled
    with SuiteSparse),
  Makoto Matsumoto [ctb, cph] (Mersenne Twister code bundled with
    SuiteSparse),
  Takuji Nishimura [ctb, cph] (Mersenne Twister code bundled with
    SuiteSparse),
  Alexander Chemeris [ctb, cph] (MS integer headers bundled with
    SuiteSparse)
Repository: CRAN
Date/Publication: 2026-05-20 07:40:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-21 23:33:30 UTC; unix
Archs: ngme2.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(noise_moments,ngme_noise)
S3method(plot,ngme_noise)
S3method(plot,ngme_sgld_ci)
S3method(plot,parameter_distance)
S3method(predict,ngme)
S3method(print,ngme)
S3method(print,ngme_model)
S3method(print,ngme_noise)
S3method(print,ngme_operator)
S3method(print,ngme_replicate)
S3method(print,ngme_trajectories)
S3method(print,noise_kld_comparison)
S3method(print,parameter_distance)
S3method(print,summary.ngme_batch_ci)
S3method(simulate,ngme)
S3method(simulate,ngme_model)
S3method(simulate,ngme_noise)
S3method(summary,ngme)
S3method(summary,ngme_batch_ci)
export(adagrad)
export(adam)
export(adamW)
export(adaptive_gd)
export(ar)
export(ar1)
export(arma)
export(arma11)
export(batch_decay)
export(batch_means_ci)
export(batch_means_estimator)
export(bfgs)
export(bv)
export(bv_matern)
export(calibrate_inv_exp_lambda)
export(calibrate_inv_exp_lambda_driven_nig)
export(compare_noise_kld)
export(compute_index_corr_from_map)
export(compute_log_like)
export(compute_ngme_CI)
export(compute_ngme_ci)
export(compute_ngme_sgld_samples)
export(control_ngme)
export(control_opt)
export(control_opt_batch_ci)
export(create_paired_cv_splits)
export(cross_validation)
export(dgal)
export(dgig)
export(dig)
export(digam)
export(dnig)
export(f)
export(generic)
export(generic_ns)
export(get_data_from_formula)
export(get_parameter_distance)
export(get_trace_trajectories)
export(get_trajectories)
export(iid)
export(make_time_series_cv_index)
export(matern)
export(mean_list)
export(merge_noise)
export(momentum)
export(name2fun)
export(ngme)
export(ngme_as_sparse)
export(ngme_batch_ci)
export(ngme_cov_matrix)
export(ngme_make_mesh_repls)
export(ngme_model_types)
export(ngme_noise_types)
export(ngme_optimizers)
export(ngme_post_samples)
export(ngme_prior_types)
export(ngme_result)
export(ngme_sgld_ci)
export(ngme_sgld_samples)
export(ngme_ts_make_A)
export(ngme_update)
export(noise_gal)
export(noise_gal_moments)
export(noise_moments)
export(noise_nig)
export(noise_nig_moments)
export(noise_normal)
export(noise_normal_nig)
export(noise_skew_t)
export(noise_t)
export(openmp_test)
export(ou)
export(pgal)
export(pgig)
export(pig)
export(pigam)
export(pnig)
export(poly_decay)
export(posterior_plot)
export(precision_matrix_multivariate)
export(precision_matrix_multivariate_spde)
export(precond_sgd)
export(prior_half_cauchy)
export(prior_inv_exp)
export(prior_inv_exponential)
export(prior_none)
export(prior_normal)
export(prior_pc_sd)
export(priors)
export(qgal)
export(qgig)
export(qig)
export(qigam)
export(qnig)
export(re)
export(rgal)
export(rgig)
export(rig)
export(rigam)
export(rmsprop)
export(rnig)
export(rw1)
export(rw2)
export(sgd)
export(sgld)
export(spacetime)
export(stepsize_control)
export(stepsize_decay)
export(stepsize_schedule)
export(tp)
export(traceplot)
export(var1)
import(Rcpp)
importFrom(Rcpp,evalCpp)
importFrom(fmesher,fm_basis)
importFrom(fmesher,fm_fem)
importFrom(fmesher,fm_mesh_1d)
importFrom(fmesher,fm_mesh_2d)
importFrom(ggplot2,aes)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,xlab)
importFrom(ggplot2,ylab)
importFrom(graphics,hist)
importFrom(methods,as)
importFrom(rlang,.data)
importFrom(stats,aggregate)
importFrom(stats,as.formula)
importFrom(stats,delete.response)
importFrom(stats,density)
importFrom(stats,dgamma)
importFrom(stats,dist)
importFrom(stats,dnorm)
importFrom(stats,formula)
importFrom(stats,median)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,qgamma)
importFrom(stats,qlogis)
importFrom(stats,quantile)
importFrom(stats,rgamma)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,simulate)
importFrom(stats,terms)
importFrom(stats,terms.formula)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(utils,str)
importFrom(utils,tail)
useDynLib(ngme2, .registration = TRUE)

# The `ngme2` Package  <img src="man/figures/logo.png" align="right" width="80" height="80" />

## Description
`ngme2` (https://davidbolin.github.io/ngme2/) is a unified, efficient, and flexible framework for fitting latent **non-Gaussian** models in R. It extends the SPDE-based Gaussian modeling toolkit to handle skewness, heavy tails, and non-smooth behavior while keeping familiar workflows for estimation, prediction, and model assessment.

## Features
- Temporal processes: AR(1), Ornstein–Uhlenbeck, random walks, and ARMA models.
- Spatial processes: Matérn fields (including graph-based meshes) and separable/non-separable space–time variants.
- Non-Gaussian driven noise: NIG, generalized asymmetric Laplace, skew-\(t\); non-Gaussian measurement noise is also supported.
- Joint and custom structures: bivariate type-G fields, longitudinal random-effects models, and user-defined operators via `generic` / `generic_ns`, flexible combinations of different models and driven noise.
- Practical tools: kriging-style prediction, cross-validation helpers, and diagnostics for convergence/fit quality.

## Quick start
```r
library(ngme2)
time_index <- seq(1, 1000, by = 1)
n <- length(time_index)

# Define the AR(1) model with NIG driven noise
ar1_nig <- f(time_index,
  model = ar1(rho = 0.7),
  noise = noise_nig(mu = 3, sigma = 2, nu = 0.5)
)

# Simulate the AR(1) process with NIG driven noise
nig_field <- simulate(ar1_nig, seed = 123, nsim = 1)[[1]]
Y <- nig_field + rnorm(n, mean = 0, sd = 1)
plot(time_index, nig_field, type = "l")

# Fit the model
fit <- ngme(
  formula = Y ~ 0 + f(time_index, model = ar1(), noise = noise_nig()),
  data    = data.frame(Y = Y, time_index = time_index),
  family  = "normal" # likelihood family
)

summary(fit)

```

Use `ngme_optimizers()` to see available optimizers and configure stochastic gradient settings via `control_opt`.

## Installation
The stable version can be installed with:
```r
install.packages("ngme2", repos = "https://davidbolin.github.io/ngme2/")
```
See the [Installation and Configuration](https://davidbolin.github.io/ngme2/articles/Installation.html) vignette if compilation tools are needed.

## Learn more
- [Get Started](https://davidbolin.github.io/ngme2/articles/ngme2.html): for a guided tour of the modeling framework.
- [Online vignettes](https://davidbolin.github.io/ngme2/articles/)
- [Online documentations](https://davidbolin.github.io/ngme2/reference/index.html)

Shared test fixtures live here.

- `data/` is for reusable tabular or serialized datasets used by multiple tests.
- `matrices/` is for reusable sparse or dense matrix fixtures.
- `models/` is for reusable model specs or saved fit objects when inline setup becomes noisy.

Keep fixtures small and deterministic. If a fixture is only used once, prefer building it inline in the test file.

This directory is for exploratory scripts and demos.

Code here is not part of the automated package test suite. When a demo becomes
important for regression coverage, promote it into `tests/testthat/`.

#' Generate control specifications for \code{ngme()} function.
#'
#' These are configurations for \code{ngme}
#' optimization process.
#'
#' @details
#' Convergence diagnostics (multi-chain):
#' * R-hat: per-parameter Gelman–Rubin statistic; passes if \code{R_hat <= max_R_hat}.
#' * Trend/Std: uses the last \code{n_slope_check} checkpoints after at least \code{n_min_batch} batches.
#'   Passes when both the relative std (\code{sqrt(var)/|mean| <= std_lim}) and linear trend
#'   of the means (\code{|slope| <= trend_lim}) satisfy their thresholds.
#' * Pflug: per-chain criterion \code{pflug_sum < pflug_alpha * max_pflug_sum} in the latest batch;
#'   if all chains satisfy it, overall convergence is declared.
#' Checks are evaluated every \code{iters_per_check = iterations / n_batch}. A parameter is marked
#' converged if any enabled parameter-level diagnostic (R-hat or Trend/Std) passes; the run stops
#' when all parameters converge or when the Pflug diagnostic triggers.
#' @param seed  set the seed for pesudo random number generator
#' @param burnin          interations for burn-in periods (before optimization)
#' @param iterations      optimization iterations
#' @param estimation      run the estimation process (call C++ in backend)
#' @param standardize_fixed  whether or not standardize the fixed effect. When
#'   \code{TRUE} (default), the design matrix is SVD-standardized internally and
#'   any user-supplied \code{control_ngme(beta_init = ...)} is automatically
#'   mapped from the original design scale to that standardized basis. Set to
#'   \code{FALSE} to keep both the design matrix and any provided
#'   \code{beta_init} on
#'   their original scale.
#'
#' @param n_parallel_chain number of parallel chains
#' @param n_batch     number of checkpoints; optimization is split into \code{n_batch} equal batches
#' @param iters_per_check run how many iterations between each check point (or specify \code{n_batch})
#' @param n_min_batch   minimum number of checkpoints before any convergence diagnostic is attempted
#' @param n_slope_check number of checkpoints used as the regression window for the trend test
#' @param std_lim         maximum allowed standard deviation
#' @param trend_lim       maximum allowed slope
#' @param print_check_info print the convergence information
#' @param start deprecated guard argument. Do not pass model starts through
#'   \code{control_opt()}; use \code{ngme(..., start = previous_fit)} instead.
#' @param start_sd        standard deviation of the initial parameter (1st chain fixed, other chains random), set 0 to be fixed for all chains
#' @param optimizer choose different sgd optimization method,
#' currently support "sgd", "sgld", "precond_sgd", "momentum", "adagrad", "rmsprop", "adam", "adamW"
#' see ?sgd, ?precond_sgd, ?momentum, ?adagrad, ?rmsprop, ?adam, ?adamW
#'
#' @param max_num_threads maximum number of threads used for parallel computing, by default will be set same as n_parallel_chain.
#' If it is more than n_parallel_chain, the rest will be used to parallel different replicates of the model.
#' @param max_relative_step   max relative step allowed in 1 iteration
#' @param max_absolute_step   max absolute step allowed in 1 iteration
#' @param trend_std_conv_check enable the trend/std diagnostic (uses \code{std_lim}, \code{trend_lim}, \code{n_slope_check})
#' @param solver_backend backend in ("eigen", "cholmod", "pardiso")
#' @param solver_type factorization type: "llt" or "ldlt"
#' @param rao_blackwellization  use rao_blackwellization
#' @param n_trace_iter  use how many iterations to approximate the trace (Hutchinson’s trick)
#'
#' @param verbose print estimation
#' @param store_traj store the optimizer trajectory for diagnostics (set FALSE to reduce memory)
#' @param sampling_strategy subsampling method of replicates of model, c("all", "ws")
#' "all" means using all replicates in each iteration,
#' "ws" means weighted sampling (each iteration use 1 replicate to compute the gradient, the sample probability is proption to its number of observations)
#' @param stepsize_control unified stepsize configuration created by
#'   \code{stepsize_control()}, \code{poly_decay()}, or \code{batch_decay()}.
#'   For polynomial schedule, \code{poly_decay(..., burnin_iter = B)} keeps
#'   schedule scale at 1 for the first \code{B} iterations, then starts decay
#'   with reset local time index.
#' @param robust use robust mode in the backend optimizer/model updates
#' @param R_hat_conv_check use the R-hat diagnostic for convergence checking
#' @param pflug_conv_check use Pflug diagnostic for convergence check
#' @param pflug_alpha scaling factor (0-1] for Pflug criterion: require \code{pflug_sum < pflug_alpha * max_pflug_sum}
#' @param max_R_hat maximum allowed R_hat
#' @return list of control variables
#' @export
control_opt <- function(
    seed = Sys.time(),
    burnin = 100,
    iterations = 500,
    estimation = TRUE,
    standardize_fixed = TRUE,
    n_batch = 10,
    iters_per_check = iterations / n_batch,
    optimizer = adam(),
    start = NULL,
    start_sd = 0.5,
    # parallel options
    n_parallel_chain = 4,
    max_num_threads = n_parallel_chain,
    print_check_info = FALSE,
    max_relative_step = 0.5,
    max_absolute_step = 0.5,
    rao_blackwellization = FALSE,
    n_trace_iter = 10,
    sampling_strategy = "all",
    solver_backend = "cholmod",
    solver_type = "llt",
    # opt print
    verbose = FALSE,
    store_traj = TRUE,
    robust = FALSE,
    stepsize_control = NULL,
    n_min_batch = min(n_batch, 3),
    n_slope_check = min(n_batch, 3),
    trend_std_conv_check = TRUE,
    std_lim = 0.01,
    trend_lim = 0.01,
    R_hat_conv_check = TRUE,
    max_R_hat = 1.1,
    pflug_conv_check = TRUE,
    pflug_alpha = 0.9) {
  strategy_list <- c("all", "ws")
  preconditioner_list <- c("none", "fast", "full")
  solver_backend_list <- c("eigen", "cholmod", "pardiso")
  solver_factor_list <- c("llt", "ldlt")
  stepsize_decay_list <- c("none", "grad_norm_plateau")
  stepsize_schedule_list <- c("constant", "poly")

  if (!is.null(start)) {
    stop(
      "`start` is not a `control_opt()` argument. ",
      "Pass previous fits via `ngme(..., start = previous_fit)` instead.",
      call. = FALSE
    )
  }

  if (is.null(stepsize_control)) {
    stepsize_control <- .default_stepsize_control()
  }
  stopifnot(inherits(stepsize_control, "ngme_stepsize_control"))
  stepsize_decay_method <- stepsize_control$decay$method
  stepsize_decay_patience <- stepsize_control$decay$patience
  stepsize_decay_gamma <- stepsize_control$decay$gamma
  stepsize_decay_min_delta <- stepsize_control$decay$min_delta
  stepsize_decay_warmup <- stepsize_control$decay$warmup
  stepsize_decay_min_stepsize <- stepsize_control$decay$min_stepsize
  stepsize_schedule_method <- stepsize_control$schedule$method
  stepsize_schedule_alpha <- stepsize_control$schedule$alpha
  stepsize_schedule_t0 <- stepsize_control$schedule$t0
  stepsize_schedule_burnin_iter <- stepsize_control$schedule$burnin_iter

  # read preconditioner from optimizer
  preconditioner <- "none"
  numerical_eps <- 1e-5
  precond_by_diff_chain <- FALSE
  compute_precond_each_iter <- FALSE
  if (optimizer$method == "precond_sgd") {
    preconditioner <- optimizer$preconditioner
    numerical_eps <- optimizer$numerical_eps
    precond_by_diff_chain <- optimizer$precond_by_diff_chain
    compute_precond_each_iter <- optimizer$compute_precond_each_iter
  }

  # if user inputs iters_per_check
  if (!missing(iters_per_check) && !missing(n_batch)) {
    stop("Specify only one of iters_per_check and n_batch")
  } else if (!missing(iters_per_check)) {
    stopifnot(
      "iterations should be multiple of iters_per_check" = iterations %% iters_per_check == 0
    )
    n_batch <- iterations / iters_per_check
  }

  # resolve solver backend + factorization; send both to C++ and let it map
  solver_backend <- match.arg(solver_backend, solver_backend_list)
  solver_factor <- match.arg(solver_type, solver_factor_list)
  stepsize_decay_method <- match.arg(stepsize_decay_method, stepsize_decay_list)
  stepsize_schedule_method <- match.arg(stepsize_schedule_method, stepsize_schedule_list)
  solver_backend_idx <- match(solver_backend, solver_backend_list) - 1L
  solver_factor_idx <- match(solver_factor, solver_factor_list) - 1L

  if (identical(optimizer$method, "sgld") &&
      identical(stepsize_schedule_method, "constant") &&
      identical(stepsize_decay_method, "none")) {
    warning(
      "optimizer = sgld() is using constant stepsize (no decay). ",
      "For asymptotically exact SGLD sampling, use diminishing stepsize ",
      "(e.g., stepsize_control = poly_decay(alpha = 0.6, t0 = 10)).",
      call. = FALSE
    )
  }

  # platform guard: pardiso disabled on builds without MKL
  is_mac <- Sys.info()["sysname"] == "Darwin"
  if (solver_backend == "pardiso" && !has_pardiso()) {
    stop(
      "solver_backend 'pardiso' requires MKL (compiled with USEMKL). ",
      "Reinstall ngme2 with MKLROOT set so MKL/Pardiso can be enabled."
    )
  }
  if (is_mac && solver_backend == "pardiso" && !has_pardiso()) {
    stop("solver_backend 'pardiso' is not available on this macOS build; reinstall with MKLROOT to enable Pardiso.")
  }
  stopifnot(
    sampling_strategy %in% strategy_list,
    preconditioner %in% preconditioner_list,
    "start_sd must be a numeric scalar" =
      is.numeric(start_sd) && length(start_sd) == 1 && is.finite(start_sd),
    is.numeric(max_num_threads) && length(max_num_threads) == 1,
    iterations > 0 && n_batch > 0,
    "iterations should be multiple of n_batch" = iterations %% n_batch == 0,
    "n_min_batch must be numeric" = is.numeric(n_min_batch),
    "n_min_batch must be a single value" = length(n_min_batch) == 1,
    "n_min_batch must be greater than 0" = n_min_batch > 0,
    "n_min_batch cannot be greater than n_batch" = n_min_batch <= n_batch,
    is.numeric(n_slope_check) && length(n_slope_check) == 1 &&
      n_slope_check > 0 && n_slope_check <= n_batch,
    inherits(optimizer, "ngme_optimizer"),
    "solver backend must map to 0:2" = solver_backend_idx %in% 0:2,
    "solver factor must be 0 (llt) or 1 (ldlt)" = solver_factor_idx %in% 0:1,
    is.numeric(pflug_alpha) && length(pflug_alpha) == 1 && pflug_alpha > 0 && pflug_alpha <= 1,
    "stepsize_decay must be one of 'none' or 'grad_norm_plateau'" =
      stepsize_decay_method %in% stepsize_decay_list,
    "stepsize_decay_patience must be >= 1" =
      stepsize_decay_method == "none" ||
        (is.numeric(stepsize_decay_patience) && length(stepsize_decay_patience) == 1 &&
          stepsize_decay_patience >= 1),
    "stepsize_decay_gamma must be in (0, 1)" =
      stepsize_decay_method == "none" ||
        (is.numeric(stepsize_decay_gamma) && length(stepsize_decay_gamma) == 1 &&
          stepsize_decay_gamma > 0 && stepsize_decay_gamma < 1),
    "stepsize_decay_min_delta must be >= 0" =
      stepsize_decay_method == "none" ||
        (is.numeric(stepsize_decay_min_delta) &&
          length(stepsize_decay_min_delta) == 1 && stepsize_decay_min_delta >= 0),
    "stepsize_decay_warmup must be >= 0" =
      stepsize_decay_method == "none" ||
        (is.numeric(stepsize_decay_warmup) && length(stepsize_decay_warmup) == 1 &&
          stepsize_decay_warmup >= 0),
    "stepsize_decay_min_stepsize must be >= 0" =
      stepsize_decay_method == "none" ||
        (is.numeric(stepsize_decay_min_stepsize) &&
          length(stepsize_decay_min_stepsize) == 1 && stepsize_decay_min_stepsize >= 0),
    "stepsize_schedule must be one of 'constant' or 'poly'" =
      stepsize_schedule_method %in% stepsize_schedule_list,
    "stepsize_schedule_alpha must be numeric scalar" =
      is.numeric(stepsize_schedule_alpha) && length(stepsize_schedule_alpha) == 1,
    "stepsize_schedule_t0 must be a non-negative numeric scalar" =
      is.numeric(stepsize_schedule_t0) && length(stepsize_schedule_t0) == 1 &&
        stepsize_schedule_t0 >= 0,
    "stepsize_schedule_burnin_iter must be a non-negative integer scalar" =
      is.numeric(stepsize_schedule_burnin_iter) &&
        length(stepsize_schedule_burnin_iter) == 1 &&
        is.finite(stepsize_schedule_burnin_iter) &&
        stepsize_schedule_burnin_iter >= 0 &&
        abs(stepsize_schedule_burnin_iter - round(stepsize_schedule_burnin_iter)) <
          sqrt(.Machine$double.eps),
    "for stepsize_schedule='poly', alpha must satisfy 1/2 < alpha < 1" =
      stepsize_schedule_method != "poly" ||
        (stepsize_schedule_alpha > 0.5 && stepsize_schedule_alpha < 1)
  )

  if (n_parallel_chain == 1) {
    precond_by_diff_chain <- FALSE
  }

  # variance reduction techniques (not used for now)
  {
    reduce_var <- FALSE
    reduce_power <- 0.75
    threshold <- 1e-5
    window_size <- 1
    stopifnot(
      "reduceVar should be in (0.5,1]" =
        (reduce_power > 0.5) && (reduce_power <= 1)
    )
  }

  control <- list(
    seed = seed,
    start_sd = start_sd,
    burnin = burnin,
    iterations = iterations,
    estimation = estimation,
    standardize_fixed = standardize_fixed,
    n_parallel_chain = n_parallel_chain,
    n_batch = n_batch,
    n_min_batch = n_min_batch, # minimum batches before checking
    n_slope_check = n_slope_check, # window for trend regression
    std_lim = std_lim,
    trend_lim = trend_lim,
    num_threads = c(
      max(n_parallel_chain, 1),
      max(floor(max_num_threads / n_parallel_chain), 1)
    ),
    rao_blackwellization = rao_blackwellization,
    n_trace_iter = n_trace_iter,
    print_check_info = print_check_info,
    verbose = verbose,
    store_traj = store_traj,
    sampling_strategy = which(strategy_list == sampling_strategy) - 1, # start from 0,

    max_relative_step = max_relative_step,
    max_absolute_step = max_absolute_step,
    trend_std_conv_check = trend_std_conv_check,

    # preconditioner related
    numerical_eps = numerical_eps,
    precond_by_diff_chain = precond_by_diff_chain,
    compute_precond_each_iter = compute_precond_each_iter,
    precond_strategy = which(preconditioner_list == preconditioner) - 1, # start from 0

    # optimization method related
    stepsize = optimizer$stepsize,
    sgd_method = optimizer$method,
    sgd_parameters = optimizer$sgd_parameters,
    line_search = optimizer$line_search,

    # solver related
    solver_backend = solver_backend_idx,
    solver_factor = solver_factor_idx,

    # stepsize decay
    stepsize_decay = stepsize_decay_method,
    stepsize_decay_patience = stepsize_decay_patience,
    stepsize_decay_gamma = stepsize_decay_gamma,
    stepsize_decay_min_delta = stepsize_decay_min_delta,
    stepsize_decay_warmup = stepsize_decay_warmup,
    stepsize_decay_min_stepsize = stepsize_decay_min_stepsize,
    stepsize_schedule = stepsize_schedule_method,
    stepsize_schedule_alpha = stepsize_schedule_alpha,
    stepsize_schedule_t0 = stepsize_schedule_t0,
    stepsize_schedule_burnin_iter = as.integer(round(stepsize_schedule_burnin_iter)),

    # variance reduction (not used for now)
    reduce_var = reduce_var,
    reduce_power = reduce_power,
    threshold = threshold,
    window_size = window_size,
    robust = robust,
    R_hat_conv_check = R_hat_conv_check,
    max_R_hat = max_R_hat,
    pflug_conv_check = pflug_conv_check,
    pflug_alpha = pflug_alpha
  )

  class(control) <- "control_opt"
  control
}

#' Generate CI-focused control settings for batch-means inference
#'
#' @description
#' Convenience wrapper for \code{control_opt()} with defaults tailored for
#' Xi-style batch-means confidence intervals.
#'
#' @details
#' This helper enforces trajectory-friendly defaults:
#' \itemize{
#'   \item \code{store_traj = TRUE}
#'   \item \code{trend_std_conv_check = FALSE}
#'   \item \code{R_hat_conv_check = FALSE}
#'   \item \code{pflug_conv_check = FALSE}
#'   \item \code{stepsize_control = poly_decay(alpha, t0, schedule_burnin_iter)}
#' }
#' Any of these can still be overridden through \code{...}.
#'
#' @param optimizer optimizer object, default \code{sgd(stepsize = 0.03)}.
#' @param burnin burn-in iterations before optimization.
#' @param iterations optimization iterations.
#' @param n_batch number of checkpoints.
#' @param n_parallel_chain number of parallel chains.
#' @param alpha polynomial stepsize exponent for \code{poly_decay(alpha, t0)}.
#' @param t0 non-negative schedule offset.
#' @param schedule_burnin_iter non-negative integer. Initial optimization
#'   iterations where polynomial schedule scaling is disabled.
#' @param ... additional arguments forwarded to \code{control_opt()} to override
#'   defaults.
#'
#' @return object of class \code{control_opt}.
#' @export
control_opt_batch_ci <- function(
    optimizer = sgd(stepsize = 0.03),
    burnin = 100,
    iterations = 2000,
    n_batch = 20,
    n_parallel_chain = 4,
    alpha = 0.501,
    t0 = 1,
    schedule_burnin_iter = 0,
    ...) {
  defaults <- list(
    optimizer = optimizer,
    burnin = burnin,
    iterations = iterations,
    n_batch = n_batch,
    n_parallel_chain = n_parallel_chain,
    store_traj = TRUE,
    trend_std_conv_check = FALSE,
    R_hat_conv_check = FALSE,
    pflug_conv_check = FALSE,
    stepsize_control = poly_decay(
      alpha = alpha,
      t0 = t0,
      burnin_iter = schedule_burnin_iter
    )
  )

  args <- utils::modifyList(defaults, list(...))
  do.call(control_opt, args)
}


#' Generate control specifications for the ngme model
#'
#' @param n_gibbs_samples    number of gibbs samples at each iteration
#' @param fix_beta       logical, fix fixed effect estimation
#' @param beta_init      fixed effect initial value on the original design
#'   scale. If \code{control_opt(standardize_fixed = TRUE)} (the default), the
#'   vector is internally rotated/scaled to match the SVD-standardized design
#'   matrix before being passed to the optimizer. Provide values on the raw
#'   design scale; no manual rescaling is required.
#' @param n_post_samples number of posterior samples, see ?ngme_post_samples()
#' @param debug          debug mode
#' @param ... additional arguments. Legacy aliases \code{feff} and
#'   \code{fix_feff} are still recognized and mapped to \code{beta_init} and
#'   \code{fix_beta}.
#' @return a list of control variables for ngme
#' @export
control_ngme <- function(
    n_gibbs_samples = 5,
    fix_beta = FALSE,
    n_post_samples = 100,
    beta_init = NULL,
    debug = FALSE,
    ...) {
  dots <- list(...)
  # backward compatibility for legacy arguments
  if (!is.null(dots$feff)) beta_init <- dots$feff
  if (!is.null(dots$fix_feff)) fix_beta <- dots$fix_feff

  control <- list(
    n_gibbs_samples = n_gibbs_samples,
    fix_beta = fix_beta,
    beta_init = beta_init,
    # legacy names preserved for downstream code until fully migrated
    fix_feff = fix_beta,
    feff = beta_init,
    n_post_samples = n_post_samples,
    debug = debug
  )

  class(control) <- "control_ngme"
  control
}

update_control_ngme <- function(control_ngme, control_opt) {
  control_ngme$rao_blackwellization <- control_opt$rao_blackwellization
  control_ngme$n_trace_iter <- control_opt$n_trace_iter
  control_ngme$stepsize <- control_opt$stepsize
  control_ngme$solver_backend <- control_opt$solver_backend
  control_ngme$solver_factor <- control_opt$solver_factor
  control_ngme$robust <- control_opt$robust

  control_ngme
}

test_that("test basic mn", {
  set.seed(100)
  n_obs <- 1000
  x <- rexp(n_obs)
  beta <- c(-1, 2)
  y <- beta[[1]] + x * beta[[2]] + rnorm(n_obs, sd = 0.5)

  # summary(lm(y~x))
  out <- ngme(
    y ~ x,
    data = data.frame(x = x, y = y),
    control_opt = control_opt(
      iterations = 100,
      # optimizer = precond_sgd(),
      print_check_info = FALSE,
      n_parallel_chain = 4
    )
  )
  expect_true(out$replicates[[1]]$noise$theta_sigma - log(1.5) < 1)
  out
  traceplot(out)

  # 2. nig case
  n_obs <- 1000
  x <- rnorm(n_obs)
  z <- rexp(n_obs)
  # beta <- c(-2, 3)
  mu <- 5
  sigma <- 0.8
  nu <- 0.1
  sim_noise <- rnig(n_obs, delta = -mu, mu = mu, nu = nu, sigma = sigma, seed = 10)
  mean(sim_noise)
  # y <- z*beta[[1]] + x * beta[[2]] + sim_noise - mean(sim_noise)
  y <- sim_noise - mean(sim_noise)

  out <- ngme(
    y ~ 0,
    data = data.frame(x = x, y = y), family = noise_nig(),
    control_opt = control_opt(
      iterations = 100,
      n_parallel_chain = 4,
      n_min_batch = 10,
      optimizer = precond_sgd(),
      print_check_info = FALSE
    )
  )
  out
  traceplot(out, hline = c(mu, sigma, nu, beta))
  expect_true(out$replicates[[1]]$noise$theta_sigma - log(2) < 1)
  expect_true(out$replicates[[1]]$noise$theta_mu - mu < 1)
  expect_true(out$replicates[[1]]$noise$theta_nu - log(nu) < 2)
  plot(noise_nig(mu = mu, nu = nu, sigma = sigma), out$replicates[[1]]$noise)
})

test_that("test ar1 (normal) + NIG noise", {
  set.seed(100)
  n_obs <- 1000
  rho <- 0.6
  sigma_ar <- 5
  model <- f(
    1:n_obs,
    model = ar1(rho = rho),
    noise = noise_normal(sigma = sigma_ar)
  )
  W <- simulate(model)[[1]]
  sd(W)

  x <- rexp(n_obs)
  mu <- 1
  sigma <- 1
  nu <- 0.5
  sim_noise <- rnig(n_obs, delta = -mu, mu = mu, nu = nu, sigma = sigma, seed = 10)
  sd(sim_noise)
  y <- W + sim_noise
  mean(sim_noise)

  out <- ngme(
    y ~ 0 + f(
      1:n_obs,
      model = ar1(),
      noise = noise_normal()
    ),
    data = data.frame(x = x, y = y),
    family = noise_nig(),
    control_opt = control_opt(
      iterations = 1000,
      n_parallel_chain = 4,
      # optimizer = precond_sgd(),
      n_min_batch = 10,
      # verbose=TRUE
    )
  )
  out
  traceplot(out, "field1", hline = c(rho, sigma_ar))
  traceplot(out, hline = c(mu, sigma, nu))

  expect_true(out$replicates[[1]]$noise$theta_sigma - log(sigma) < 2)
  expect_true(out$replicates[[1]]$noise$theta_mu - mu < 1)
  expect_true(out$replicates[[1]]$noise$theta_nu - log(nu) < 2)
})

library(ngme2)
library(testthat)

test_that("control_opt robust parameter works", {
  # Create a simple model
  n_obs <- 100
  sigma_eps <- 0.5
  x <- runif(n_obs)
  beta <- 2
  y <- beta * x + rnorm(n_obs, sd = sigma_eps)

  # Fit model with robust = TRUE
  control <- control_opt(
    iterations = 10,
    robust = TRUE,
    estimation = TRUE,
    verbose = TRUE
  )

  # Just checking if it runs without error
  expect_error(
    ngme(
      formula = y ~ x,
      data = data.frame(y = y, x = x),
      control_opt = control,
      family = "normal"
    ),
    NA
  )
})

test_that("test bv operator", {
  n <- 10

  op <- bv(
    mesh = 1:5,
    theta = 0.5,
    rho = 0.5,
    sub_models = list(
      A = ar1(rho = 0.5),
      B = ar1(rho = -0.5)
    )
  )
  op

  model <- f(
    c(1:(n / 2), 1:(n / 2)),
    model = bv(
      use_c_param = TRUE,
      mesh = 1:5,
      rho = 0.5,
      sub_models = list(
        A = ar1(rho = 0.5),
        B = ar1(rho = -0.5)
      )
    ),
    group = c(rep("A", n / 2), rep("B", n / 2)),
    noise = list(
      A = noise_normal(),
      B = noise_normal()
    )
  )
  model
})

test_that("test bv(ar1, ar1) with 2 noise", {
  op <- bv(
    mesh = 1:5,
    theta = 0.5, rho = 0.8,
    sub_models = list(
      A = ar1(rho = 0.5),
      B = ar1(rho = -0.5)
    )
  )

  n <- 10
  true_model <- f(
    c(1:(n / 2), 1:(n / 2)),
    model = bv(
      mesh = 1:5,
      rho = 0.8,
      sub_models = list(
        A = ar1(rho = 0.5),
        B = ar1(rho = -0.5)
      )
    ),
    group = c(rep("A", n / 2), rep("B", n / 2)),
    noise = list(
      A = noise_nig(mu = 3, sigma = 2, nu = 1),
      B = noise_nig(mu = -3, sigma = 2, nu = 1)
    )
  )
  true_model
  true_model$operator$K


  W <- simulate(true_model, seed = 1)[[1]]
  n_obs <- length(W)
  Y <- W + rnorm(n_obs, sd = 0.5)
  # Y = c(Y_A, Y_B)
  # label = c(A, B)

  out <- ngme(
    Y ~ 0 + f(
      c(1:(n / 2), 1:(n / 2)),
      model = bv(
        mesh = c(1:(n / 2)),
        theta = 0.45,
        sub_models = list(
          A = ar1(rho = 0.5),
          B = ar1(rho = -0.5)
        )
      ),
      noise = list(
        A = noise_nig(),
        B = noise_nig()
      )
    ),
    group = c(rep("A", n / 2), rep("B", n / 2)),
    data = data.frame(Y = Y),
    control_opt = control_opt(
      optimizer = precond_sgd(),
      seed = 3,
      iterations = 10,
      n_batch = 5,
      n_parallel_chain = 4,
      print_check_info = TRUE
    )
  )

  out
  traceplot(out, "field1")

  lp <- predict(out,
    map = list(field1 = c(1, 2, 3)),
    group = c("B", "B", "A")
  )
  str(lp)
})

test_that("test on bv(matern, matern)", {
  pl01 <- cbind(c(0, 1, 1, 0, 0) * 10, c(0, 0, 1, 1, 0) * 5)
  mesh <- fmesher::fm_mesh_2d(loc.domain = pl01, cutoff = 0.2, max.edge = c(1, 10))
  mesh$n

  n_obs1 <- 300
  loc1 <- cbind(runif(n_obs1, 0, 10), runif(n_obs1, 0, 5))
  n_obs2 <- 400
  loc2 <- cbind(runif(n_obs2, 0, 10), runif(n_obs2, 0, 5))

  true_model <- f(
    map = rbind(loc1, loc2),
    model = bv(
      mesh = mesh,
      sub_models = list(
        A = matern(theta_kappa = log(1)),
        B = matern(theta_kappa = log(3))
      )
    ),
    group = c(rep("A", 300), rep("B", 400)),
    noise = list(A = noise_nig(), B = noise_nig())
  )

  W <- simulate(true_model)[[1]]
  n_obs <- length(W)
  Y <- W + rnorm(n_obs, sd = 0.5)
  length(Y)

  out <- ngme(
    Y ~ 0 + f(
      map = rbind(loc1, loc2),
      model = bv(
        mesh = mesh,
        sub_models = list(
          A = matern(),
          B = matern()
        )
      ),
      noise = list(A = noise_nig(), B = noise_nig())
    ),
    group = c(rep("A", 300), rep("B", 400)),
    data = data.frame(
      Y = Y
    ),
    control_ngme = control_ngme(
      n_gibbs_samples = 5
    ),
    control_opt = control_opt(
      iterations = 10,
      n_parallel_chain = 4,
      estimation = T,
      verbose = T
    )
  )

  out
  traceplot(out, "field1")
  predict(out,
    map = list(field1 = cbind(c(1, 2, 3), c(2, 3, 4))),
    group = rep("A", 6)
  )
  # out$replicates[[1]]$models[[1]]$theta_K
  # out$replicates[[1]]$models[[1]]$operator$theta_K
})

test_that("test on bv matern NIG", {
  pl01 <- cbind(c(0, 1, 1, 0, 0) * 10, c(0, 0, 1, 1, 0) * 5)
  mesh <- fmesher::fm_mesh_2d(
    loc.domain = pl01,
    cutoff = 0.2,
    max.edge = c(1, 5)
  )
  mesh$n
  n_obs <- 500

  # generate random locations (same for two fields)
  long <- runif(n_obs / 2, 0, 10)
  lat <- runif(n_obs / 2, 0, 5)
  long <- c(long, long)
  lat <- c(lat, lat)
  group <- c(rep("W1", n_obs / 2), rep("W2", n_obs / 2))

  # plot the mesh
  # plot(mesh); points(long, lat)

  # parameters of the bivariate model
  theta <- pi / 4
  rho <- 0.2
  theta_K_1 <- log(2)
  theta_K_2 <- log(5)
  mu_1 <- -2
  sigma_1 <- 0.5
  nu_1 <- 1
  mu_2 <- 2
  sigma_2 <- 0.3
  nu_2 <- 0.5

  true_model <- f(
    ~ long + lat,
    model = bv_matern(
      mesh = mesh,
      theta = theta,
      rho = rho,
      sd1 = sigma_1,
      sd2 = sigma_2,
      sub_models = list(
        W1 = matern(theta_kappa = theta_K_1),
        W2 = matern(theta_kappa = theta_K_2)
      )
    ),
    group = group,
    noise = list(
      W1 = noise_nig(mu = mu_1, sigma = 1, nu = nu_1),
      W2 = noise_nig(mu = mu_2, sigma = 1, nu = nu_2)
    )
  )
  true_model

  sim_fields <- simulate(true_model)[[1]]

  sd_1 <- 0.6
  sd_2 <- 0.9
  rho_e <- 0.9
  Cov_same_idx <- matrix(c(sd_1^2, rho_e * sd_1 * sd_2, rho_e * sd_1 * sd_2, sd_2^2), nrow = 2)

  Cov_measurement <- Cov_same_idx %x% diag(n_obs / 2)

  # e ~ N(0, Cov_measurement)
  L <- t(chol(Cov_measurement))
  e <- L %*% rnorm(n_obs)

  # fixed effects
  x1 <- rexp(n_obs)
  x2 <- rnorm(n_obs)
  feff <- c(-3, 1.5)

  Y <- sim_fields + x1 * feff[1] + x2 * feff[2] + as.numeric(e)

  B_sigma <- matrix(0, nrow = n_obs, ncol = 2)
  B_sigma[group == "W1", 1] <- 1
  B_sigma[group == "W2", 2] <- 1

  #
  out_cor_gauss <- ngme(
    Y ~ 0 + x1 + x2 + f(
      ~ long + lat,
      name = "bv",
      model = bv_matern(
        mesh = mesh,
        sub_models = list(
          W1 = matern(),
          W2 = matern()
        )
      ),
      # debug=T,
      noise = list(
        W1 = noise_normal(),
        W2 = noise_normal()
      )
    ),
    group = group,
    family = noise_normal(
      corr_measurement = TRUE,
      index_corr = c(1:(n_obs / 2), 1:(n_obs / 2)),
      B_sigma = B_sigma,
      theta_sigma = c(0, 0)
    ),
    data = data.frame(Y, long, lat),
    control_opt = control_opt(
      iterations = 20,
      n_parallel_chain = 4,
      seed = 50
    ),
    debug = FALSE
  )
  out_cor_gauss
  traceplot(out_cor_gauss, "bv")
  traceplot(out_cor_gauss, hline = c(log(sd_1), log(sd_2), rho_e, feff))

  out_cor_nig <- ngme(
    Y ~ 0 + x1 + x2 + f(
      ~ long + lat,
      name = "bv",
      model = bv_matern(
        mesh = mesh,
        theta = pi / 4,
        fix_theta = TRUE,
        sub_models = list(
          W1 = matern(),
          W2 = matern()
        )
      ),
      # debug=T,
      noise = list(
        W1 = noise_nig(),
        W2 = noise_nig()
      )
    ),
    group = group,
    family = noise_normal(
      corr_measurement = TRUE,
      index_corr = c(1:(n_obs / 2), 1:(n_obs / 2)),
      B_sigma = B_sigma,
      theta_sigma = c(0, 0)
    ),
    data = data.frame(Y, long, lat),
    control_opt = control_opt(
      estimation = TRUE,
      iterations = 20,
      print_check_info = FALSE,
      seed = 50,
      n_parallel_chain = 1
    ),
    debug = FALSE,
    start = out_cor_gauss
  )
  out_cor_nig
  traceplot(out_cor_nig, "bv")
  traceplot(out_cor_nig, hline = c(log(sd_1), log(sd_2), rho_e, feff))
})

test_that("predict and cross_validation support predictive_average over chains", {
  withr::local_seed(1)

  n <- 40
  y <- as.numeric(arima.sim(list(ar = 0.6), n = n)) + rnorm(n, sd = 0.2)

  fit <- ngme(
    y ~ 0 + f(1:n, model = ar1(), name = "ar1_field"),
    data = data.frame(y = y),
    control_opt = control_opt(
      iterations = 30,
      n_batch = 3,
      n_parallel_chain = 3,
      burnin = 1,
      n_min_batch = 1,
      verbose = FALSE,
      print_check_info = FALSE,
      # store_traj = FALSE,
      seed = 1
    )
  )

  chain_outputs <- attr(fit, "chain_outputs")
  expect_true(is.list(chain_outputs))
  expect_equal(length(chain_outputs), 3)

  pred_param_mean <- predict(
    fit,
    map = list(ar1_field = seq_len(n)),
    estimator = "mean",
    sampling_size = 20,
    burnin_size = 5,
    seed = 1,
    chain_combine = "param_mean"
  )
  pred_predictive_avg <- predict(
    fit,
    map = list(ar1_field = seq_len(n)),
    estimator = "mean",
    sampling_size = 20,
    burnin_size = 5,
    seed = 1,
    chain_combine = "predictive_average"
  )

  expect_equal(length(pred_predictive_avg$mean), n)
  expect_true(all(is.finite(pred_predictive_avg$mean)))
  expect_true(ncol(attr(pred_predictive_avg, "samples")) >= ncol(attr(pred_param_mean, "samples")))

  cv_predictive_avg <- cross_validation(
    fit,
    type = "k-fold",
    k = 2,
    N_sim = 1,
    n_gibbs_samples = 5,
    n_burnin = 2,
    seed = 1,
    print = FALSE,
    parallel = FALSE,
    chain_combine = "predictive_average"
  )

  expect_true(all(is.finite(as.matrix(cv_predictive_avg$mean.scores))))
})

test_that("control_opt rejects model starts passed via partial matching", {
  expect_error(
    control_opt(start = list(dummy = TRUE)),
    "`start` is not a `control_opt\\(\\)` argument"
  )
})

test_that("control_opt validates start_sd eagerly", {
  expect_error(
    control_opt(start_sd = list(dummy = TRUE)),
    "start_sd must be a numeric scalar"
  )
})

test_that("cross_validation(data=...) rebuilds non-stationary model without formula globals", {
  skip_if_not_installed("fmesher")
  withr::local_seed(1)

  x <- seq(0, 1, length.out = 80)
  mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = 30))
  mesh_loc <- as.numeric(mesh$loc)
  B <- cbind(1, scale(mesh_loc, scale = FALSE))
  n_basis <- ncol(B)

  y <- sin(2 * pi * x) + rnorm(length(x), sd = 0.1)
  train_data <- data.frame(Y = y[1:60], x = x[1:60])
  test_data <- data.frame(Y = y[61:80], x = x[61:80])

  fit <- ngme(
    Y ~ 0 + f(
      x,
      model = matern(
        mesh = mesh,
        B_kappa = B,
        theta_kappa = rep(0, n_basis)
      ),
      noise = noise_normal(
        B_sigma = B,
        theta_sigma = rep(-2, n_basis)
      )
    ),
    data = train_data,
    control_opt = control_opt(estimation = FALSE)
  )

  rm(mesh, mesh_loc, B, n_basis)

  expect_no_error({
    cv <- cross_validation(
      fit,
      data = test_data,
      type = "k-fold",
      k = 2,
      N_sim = 1,
      n_gibbs_samples = 5,
      n_burnin = 2,
      print = FALSE,
      parallel = FALSE
    )
  })

  expect_equal(rownames(cv$mean.scores), "model_1")
  expect_true(all(is.finite(as.matrix(cv$mean.scores))))
})

test_that("predict handles normal-nig latent noise", {
  withr::local_seed(1)

  n <- 12
  data <- data.frame(
    y = as.numeric(arima.sim(list(ar = 0.4), n = n)),
    idx = seq_len(n)
  )

  fit <- ngme(
    y ~ 0 + f(idx, model = ar1(), noise = noise_normal_nig(), name = "latent"),
    data = data,
    control_opt = control_opt(
      iterations = 30,
      n_batch = 3,
      n_min_batch = 1,
      burnin = 1,
      verbose = FALSE,
      print_check_info = FALSE,
      store_traj = FALSE,
      seed = 1
    )
  )

  pred_latent <- predict(
    fit,
    map = list(latent = data$idx),
    type = "latent",
    estimator = c("mean", "sd", "0.025q", "0.975q"),
    sampling_size = 10,
    burnin_size = 2,
    seed = 1
  )
  pred_response <- predict(
    fit,
    map = list(latent = data$idx),
    type = "response",
    estimator = c("mean", "sd"),
    sampling_size = 10,
    burnin_size = 2,
    seed = 1
  )
  pred_latent_samples <- predict(
    fit,
    map = list(latent = data$idx),
    type = "latent",
    estimator = c("mean", "sd", "0.025q", "0.975q"),
    sampling_size = 10,
    burnin_size = 2,
    seed = 1,
    return_samples = TRUE
  )

  expect_length(pred_latent$mean, n)
  expect_length(pred_response$mean, n)
  expect_length(pred_response$sd, n)
  expect_true(all(is.finite(pred_latent$mean)))
  expect_true(all(is.finite(pred_response$mean)))
  expect_true(all(is.finite(pred_response$sd)))
  expect_equal(pred_latent$mean, pred_latent_samples$mean)
  expect_equal(pred_latent$sd, pred_latent_samples$sd)
  expect_equal(pred_latent$`0.025q`, pred_latent_samples$`0.025q`)
  expect_equal(pred_latent$`0.975q`, pred_latent_samples$`0.975q`)
})

library(ngme2)
library(testthat)

test_that("bv model with c1, c2 parameters works", {
  n_obs <- 1000
  n_each <- n_obs / 2
  group <- rep(c("W1", "W2"), n_each)
  reorder_loc <- 1:n_each

  # Simulate with c1=2, c2=0.5
  sim_fields <- simulate(
    f(
      c(reorder_loc, reorder_loc),
      model = bv(
        mesh = reorder_loc,
        theta = 0.5, rho = 0.5,
        c1 = 2, c2 = 0.5,
        use_c_param = TRUE,
        sub_models = list(W1 = ar1(rho = 0.5), W2 = ar1(rho = 0.5))
      ),
      group = group,
      noise = list(W1 = noise_normal(), W2 = noise_normal())
    )
  )[[1]]

  # Fit
  res <- ngme(
    Y ~ 0 + f(
      c(reorder_loc, reorder_loc),
      model = bv(
        mesh = reorder_loc,
        use_c_param = TRUE,
        sub_models = list(W1 = ar1(), W2 = ar1())
      ),
      group = group,
      noise = list(W1 = noise_normal(), W2 = noise_normal())
    ),
    data = data.frame(Y = sim_fields),
    control_opt = control_opt(
      estimation = TRUE,
      iterations = 10,
      print_check_info = FALSE
    )
  )

  expect_true(inherits(res, "ngme"))
  # Check parameter names
  print(res$replicates[[1]]$par_names)
  expect_true("c1" %in% res$replicates[[1]]$par_names)
  expect_true("c2" %in% res$replicates[[1]]$par_names)
})

test_that("test fractional model", {
  skip_if_not_installed("fmesher")
  skip_if_not_installed("rSPDE")
  skip_if_not_installed("inlabru")
  skip_if_not_installed("sf")
  library(fmesher)
  library(rSPDE)
  n_loc <- 1000
  loc_2d_mesh <- matrix(runif(n_loc * 2), n_loc, 2)
  mesh_2d <- fm_mesh_2d(
    loc = loc_2d_mesh,
    cutoff = 0.05,
    max.edge = c(0.1, 0.5)
  )
  plot(mesh_2d)
  mesh_2d$n

  sigma <- 2
  range <- 0.25
  nu <- 1.3
  kappa <- sqrt(8 * nu) / range
  op <- matern.operators(
    mesh = mesh_2d, nu = nu,
    range = range, sigma = sigma,
    m = 1,
    parameterization = "matern"
  )
  tau <- op$tau

  u <- simulate(op)
  A <- spde.make.A(
    mesh = mesh_2d,
    loc = loc_2d_mesh
  )
  sigma.e <- 0.1
  y <- A %*% u + rnorm(n_loc) * sigma.e

  ############ Fitting with ngme2 ############
  # library(ngme2)
  toy_df <- data.frame(
    coord1 = loc_2d_mesh[, 1],
    coord2 = loc_2d_mesh[, 2],
    y = as.vector(y)
  )
  ngme_fit <- ngme(
    y ~ -1 + f(
      ~ coord1 + coord2,
      model = matern(
        mesh = mesh_2d,
        kappa = 12,
        alpha = 2.3,
        rational_order = 1,
        fix_alpha = FALSE
      ),
      noise = noise_normal(sigma = 200, fix_theta_sigma = FALSE)
    ),
    data = toy_df,
    family = noise_normal(sigma = 0.1, fix_sigma = FALSE),
    # debug = TRUE,
    control_opt = control_opt(
      n_parallel_chain = 4,
      start_sd = 0.01,
      optimizer = precond_sgd(),
      iterations = 200,
      robust = TRUE,
      verbose = TRUE
    )
    # start = ngme_fit
  )

  summary(ngme_fit)
  traceplot(ngme_fit)
  traceplot(ngme_fit, "field1")

  ############ Fitting with INLAbru ############
  library(inlabru)
  rspde_model <- rspde.matern(
    mesh = mesh_2d,
    nu.upper.bound = 2,
    parameterization = "spde"
  )

  library(sf)
  toy_df <- data.frame(
    coord1 = loc_2d_mesh[, 1],
    coord2 = loc_2d_mesh[, 2],
    y = as.vector(y)
  )
  toy_df <- st_as_sf(toy_df, coords = c("coord1", "coord2"))

  cmp <-
    y ~ -1 + field(geometry,
      model = rspde_model
    )

  rspde_bru_fit <-
    bru(cmp,
      data = toy_df,
      options = list(
        family = "gaussian",
        num.threads = "1:1"
      )
    )

  summary(rspde_bru_fit)

  result_fit <- rspde.result(rspde_bru_fit, "field", rspde_model)
  summary(result_fit)
  tau <- op$tau
  result_df <- data.frame(
    parameter = c("tau", "kappa", "nu"),
    true = c(tau, kappa, nu), mean = c(
      result_fit$summary.tau$mean,
      result_fit$summary.kappa$mean,
      result_fit$summary.nu$mean
    ),
    mode = c(
      result_fit$summary.tau$mode,
      result_fit$summary.kappa$mode,
      result_fit$summary.nu$mode
    )
  )
  print(result_df)
})

#' Fit an additive linear mixed effect model over replicates
#'
#'  \code{ngme} function performs an analysis of non-gaussian additive models.
#'  It does the maximum likelihood estimation via stochastic gradient descent.
#'  The prediction of unknown location can be performed by leaving the response
#'  variable to be \code{NA}. The likelihood is specified by \code{family}.
#' The model estimation control can be setted in \code{control} using
#'  \code{control_opt()} function, see \code{?control_opt} for details.
#' See \code{ngme_model_types()} for available models.
#' @param formula formula
#' @param data    a dataframe or a list providing data
#'   (Only response variable can contain \code{NA} value,
#'    \code{NA} value in other columns will cause problem)
#' @param control_opt  control for optimizer. by default it is \code{control_opt()}. See \code{?control_opt} for details.
#' @param control_ngme control for ngme model. by default it is \code{control_ngme()}. See \code{?control_ngme} for details.
#' @param replicate factor, used for divide data into different replicates
#' @param group factor, used for bivariate model, indicating which group the observation belongs to
#' @param family likelihood type, same as measurement noise specification. It can be provided as:
#'   \enumerate{
#'     \item a string, e.g., \code{"normal"}, \code{"nig"}, \code{"t"}.
#'     \item an ngme noise object, e.g., \code{noise_normal()}, \code{noise_nig(mu = 0, sigma = 1, nu = 1)}, \code{noise_t(nu=5)}.
#'   }
#' @param prior_beta prior specification for fixed effects (`beta`), created by
#'   \code{prior_*()} or \code{priors(...)}.
#' @param start  starting ngme object (usually object from last fit)
#' @param moving_window number of iterations to average the estimation
#' @param debug  toggle debug mode
#'
#' @return random effects (for different replicate) + models(fixed effects, measuremnt noise, and latent process)
#' @export
#'
#' @examples
#' ngme(
#'   formula = Y ~ x1 + f(
#'     x2,
#'     model = ar1(rho = 0.5),
#'     noise = noise_nig()
#'   ) + f(x1,
#'     model = rw1(),
#'     noise = noise_normal(),
#'   ),
#'   family = noise_normal(sd = 0.5),
#'   data = data.frame(Y = 1:5, x1 = 2:6, x2 = 3:7),
#'   control_opt = control_opt(
#'     estimation = FALSE
#'   )
#' )
ngme <- function(
    formula,
    data,
    family = "normal",
    control_opt = NULL,
    control_ngme = NULL,
    group = NULL,
    replicate = NULL,
    start = NULL,
    moving_window = 1, # return the average estimation of last .. iterations
    prior_beta = NULL,
    debug = FALSE) {
  # -------------  CHECK INPUT ---------------
  if (is.null(data)) {
    stop("Missing data.frame/list `data'. Leaving `data' empty might lead to\n\t\tuncontrolled behaviour, therefore is it required.")
  }
  if (!is.data.frame(data)) {
    stop("\n\tArgument `data' must be a data.frame.")
  }

  # configure control parameters
  if (is.null(control_ngme)) control_ngme <- control_ngme()
  if (is.null(control_opt)) control_opt <- control_opt()
  stopifnot(
    inherits(control_ngme, "control_ngme"),
    inherits(control_opt, "control_opt"),
    "data provide should be of the same length" =
      all(diff(sapply(data, length)) == 0)
  )
  control_ngme <- update_control_ngme(control_ngme, control_opt)

  group_supplied <- !is.null(group)
  group <- validate_rep_or_group(group, data)
  replicate <- validate_rep_or_group(replicate, data)

  # model fit information
  fit <- list(
    formula = formula,
    data = data,
    family = family,
    replicate = replicate,
    group = group,
    n_data = nrow(data)
  )
  if (debug) control_ngme$debug <- TRUE

  noise <- if (is.character(family)) {
    switch(family,
      "normal" = noise_normal(),
      "gaussian" = noise_normal(),
      "nig" = noise_nig(),
      "t" = noise_t(),
      "skew_t" = noise_skew_t(),
      stop("Unknown family!")
    )
  } else {
    family
  } # ngme noise object

  stopifnot(class(noise) == "ngme_noise")

  # parse the formula get a list of ngme_replicate
  ngme_model <- ngme_parse_formula(
    formula, data, control_ngme, noise, group, replicate, prior_beta,
    control_opt$standardize_fixed # convert fixed effects
  )

  if (contain_bv_model(ngme_model) && group_supplied) {
    stopifnot(
      "Please supply `group` argument in ngme() function, not in f()." =
        length(levels(group)) == 2
    )
  }
  attr(ngme_model, "fit") <- fit
  attr(ngme_model, "estimation_enabled") <- control_opt$estimation

  # Check if using bfgs for non-Gaussian model
  if (control_opt$sgd_method == "bfgs") {
    stopifnot(
      "Please use other optimizer for non-Gaussian model, BFGS is used for optimizing Gaussian model likelihood." =
        ngme_model$replicates[[1]]$all_gaussian &&
          noise$noise_type == "normal"
    )
  }

  ####### Use Last_fit ngme object to update Rcpp_list
  if (!is.null(start) && !inherits(start, "ngme")) {
    stop("start should be an ngme object.")
  }

  # update with start (list of ngmes)
  if (inherits(start, "ngme")) {
    for (i in seq_along(ngme_model$replicates)) {
      ngme_model$replicates[[i]] <- within(ngme_model$replicates[[i]], {
        # check if feff is the same, then overwrite the feff
        same_feff <- ncol(X) == ncol(start$replicates[[i]]$X)
        if (!is.null(colnames(X)) && !is.null(colnames(start$replicates[[i]]$X))) {
          same_feff <- same_feff && identical(colnames(X), colnames(start$replicates[[i]]$X))
        }
        if (same_feff) {
          start_feff <- start$replicates[[i]]$feff
          if (!is.null(names(start_feff)) &&
            !is.null(colnames(X)) &&
            all(colnames(X) %in% names(start_feff))) {
            start_feff <- start_feff[colnames(X)]
          }

          if (!ngme_model$replicates[[i]]$standardize) {
            feff <- as.numeric(start_feff)
          } else {
            # start_feff is on the raw scale after restoration; map it to the
            # current centered + SVD-standardized parameterization.
            feff_center <- ngme_beta_raw_to_center(start_feff, X_center, X_center_target)
            svd_idx <- if (!is.null(svd) && !is.null(svd$idx)) svd$idx else integer(0)
            if (length(svd_idx) > 0) {
              beta_fixed <- feff_center[svd_idx]
              beta_fixed <- as.numeric(svd$d * drop(t(svd$v) %*% beta_fixed))
              feff_center[svd_idx] <- beta_fixed
              feff <- feff_center
            } else {
              feff <- feff_center
            }
          }
          names(feff) <- colnames(X)
        }

        noise <- update_noise(noise, new_noise = start$replicates[[i]]$noise)
        for (j in seq_along(start$replicates[[i]]$models)) {
          prev_model_type <-
            start$replicates[[i]]$models[[j]]$model
          prev_model_ope <-
            start$replicates[[i]]$models[[j]]$operator

          # update parameter of K
          if (models[[j]]$model == "bv_matern" && prev_model_type == "bv_matern") {
            # update theta_K
            if (!models[[j]]$operator$fix_theta) {
              models[[j]]$theta_K <- models[[j]]$operator$theta_K <-
                c(0, prev_model_ope$theta_K)
            } else {
              models[[j]]$theta_K <- models[[j]]$operator$theta_K <-
                prev_model_ope$theta_K
            }

            # for printing
            models[[j]]$operator$first <- prev_model_ope$first
            models[[j]]$operator$second <- prev_model_ope$second
          } else {
            stopifnot(
              "Please make sure model type are the same" =
                models[[j]]$model == prev_model_type
            )
            # default case: seed new operator with previous estimates without
            # overwriting the new model configuration (e.g., fix_alpha changes)
            op_new <- models[[j]]$operator
            op_prev <- prev_model_ope

            # try name-based matching first so we only overwrite shared params
            if (!is.null(op_new$param_name) && !is.null(op_prev$param_name)) {
              theta_new <- op_new$theta_K
              names(theta_new) <- op_new$param_name

              theta_prev <- op_prev$theta_K
              names(theta_prev) <- op_prev$param_name

              common <- intersect(names(theta_new), names(theta_prev))
              if (length(common) > 0) {
                theta_new[common] <- theta_prev[common]
              }

              op_new$theta_K <- theta_new
              models[[j]]$theta_K <- theta_new
            } else if (length(op_new$theta_K) == length(op_prev$theta_K)) {
              # fall back to position-wise copy when dimensions align
              op_new$theta_K <- op_prev$theta_K
              models[[j]]$theta_K <- op_prev$theta_K
            } else {
              # keep op_new defaults when shapes differ (e.g., freeing alpha)
              models[[j]]$theta_K <- op_new$theta_K
            }

            op_new$K <- op_new$update_K(op_new$theta_K)

            models[[j]]$operator <- op_new
          }

          stopifnot(
            "length of W should be the same" =
              models[[j]]$W_size == length(start$replicates[[i]]$models[[j]]$W)
          )
          stopifnot(
            "length of V should be the same" =
              models[[j]]$V_size == length(start$replicates[[i]]$models[[j]]$noise$V)
          )
          # update the rest
          models[[j]]$W <- start$replicates[[i]]$models[[j]]$W
          models[[j]]$noise$V <- start$replicates[[i]]$models[[j]]$noise$V
          models[[j]]$noise <- update_noise(
            models[[j]]$noise,
            new_noise = start$replicates[[i]]$models[[j]]$noise
          )
        }
      })
    }
  }
  if (debug) {
    message(paste(capture.output(str(ngme_model$replicates[[1]])), collapse = "\n"))
  }

  # configuration of controls

  # check all f has the same replicate
  ################# Run CPP ####################
  check_dim(ngme_model)
  if (control_opt$estimation) {
    if (isTRUE(control_opt$verbose)) {
      message("Starting estimation...")
    }
    outputs <- tryCatch(
      estimate_cpp(ngme_model, control_opt),
      error = function(err) {
        stop("C++ estimation failed: ", conditionMessage(err), call. = FALSE)
      }
    )

    ################# Update the estimates ####################
    est_output <- mean_list(outputs)
    for (i in seq_along(ngme_model$replicates)) {
      ngme_model$replicates[[i]] <- update_ngme_est(ngme_model$replicates[[i]], est_output[[i]])
    }

    # return posterior samples of W and V
    if (isTRUE(control_opt$verbose)) {
      message("Starting posterior sampling...")
    }
    for (i in seq_along(ngme_model$replicates)) {
      res <- tryCatch(
        sampling_cpp(
          ngme_model$replicates[[i]],
          n = control_ngme$n_post_samples,
          n_burnin = 1,
          posterior = TRUE,
          seed = control_opt$seed
        ),
        error = function(err) {
          stop(
            "C++ posterior sampling failed for replicate ", i, ": ",
            conditionMessage(err),
            call. = FALSE
          )
        }
      )

      df_V <- data.frame(res$V)
      colnames(df_V) <- paste0("sample_", 1:ncol(df_V))
      ngme_model$replicates[[i]]$post_V <- df_V

      df_W <- data.frame(res$W)
      colnames(df_W) <- paste0("sample_", 1:ncol(df_W))
      ngme_model$replicates[[i]]$post_W <- df_W

      sd_W <- mean(sapply(res$W, sd))
    }
    if (isTRUE(control_opt$verbose)) {
      message("Posterior sampling done.")
      message("Average standard deviation of the posterior W: ", sd_W)
    }

    # print R_hat
    # if (!is.null(attr(outputs, "R_hat"))) {
    #   cat("R_hat: \n")
    #   print(attr(outputs, "R_hat"))
    # }

    if (isTRUE(control_opt$verbose)) {
      message(
        "Use ngme_post_samples() to access posterior samples and ",
        "ngme_result() to access latent model parameters."
      )
    }

    # mn_nu <- ngme_model$replicates[[1]]$noise$nu
    # if (length(mn_nu) > 1 && mn_nu > 100)
    #   cat("The parameter nu for measurement noise is too big, consider using family=normal instead. \n")

    if (isTRUE(control_opt$store_traj)) {
      # Transform trajectory
      traj_df_chains <- transform_traj(attr(outputs, "opt_traj"))
      # dispatch trajs to each latent and block
      idx <- 0
      for (i in seq_along(ngme_model$replicates[[1]]$models)) {
        n_params <- ngme_model$replicates[[1]]$models[[i]]$n_params
        lat_traj_chains <- list()
        for (j in seq_along(traj_df_chains)) {
          lat_traj_chains[[j]] <- traj_df_chains[[j]][idx + 1:n_params, ]
        }

        attr(ngme_model$replicates[[1]]$models[[i]], "lat_traj") <- lat_traj_chains
        idx <- idx + n_params
      }

      # measurement noise and feff
      block_traj <- list()
      n_feff <- length(ngme_model$replicates[[1]]$feff)
      n_chains <- length(traj_df_chains)
      for (j in seq_len(n_chains)) {
        block_traj[[j]] <- traj_df_chains[[j]][(idx + 1):ngme_model$replicates[[1]]$n_params, ]
      }

      n_block_params <- nrow(block_traj[[1]])
      # Map fixed-effects trajectories back to raw parameterization so traceplot
      # is comparable with ngme_result()/printed feff.
      if (n_feff > 0) {
        rep1 <- ngme_model$replicates[[1]]
        feff_idx <- (n_block_params - n_feff + 1):n_block_params
        feff_names <- names(rep1$feff)

        for (i in seq_along(block_traj)) {
          betas <- as.matrix(block_traj[[i]][feff_idx, , drop = FALSE])

          # 1) Undo SVD standardization on the standardized subset.
          if (isTRUE(rep1$standardize) && !is.null(rep1$svd) && length(rep1$svd$d) > 0) {
            svd <- rep1$svd
            svd_idx <- if (!is.null(svd$idx)) svd$idx else seq_len(length(svd$d))
            betas[svd_idx, ] <- svd$v %*% ngme_diag_vec(1 / svd$d) %*% betas[svd_idx, , drop = FALSE]
          }

          # 2) Undo demeaning shift (centered beta -> raw beta), per iteration.
          if (!is.null(rep1$X_center) && length(rep1$X_center) > 0) {
            for (k in seq_len(ncol(betas))) {
              beta_k <- betas[, k]
              names(beta_k) <- feff_names
              beta_k <- ngme_beta_center_to_raw(beta_k, rep1$X_center, rep1$X_center_target)
              betas[, k] <- as.numeric(beta_k)
            }
          }

          block_traj[[i]][feff_idx, ] <- betas
        }
      }

      attr(ngme_model$replicates[[1]], "block_traj") <- block_traj
      attr(outputs, "opt_traj") <- NULL
    } else {
      attr(outputs, "opt_traj") <- NULL
    }

    if (is.list(outputs) && length(outputs) > 1) {
      attr(ngme_model, "chain_outputs") <- outputs
    } else {
      attr(ngme_model, "chain_outputs") <- NULL
    }
  } else {
    # Estimation skipped: map fixed effects and X back to the raw covariate
    # scale so outputs are comparable to pre-demean runs.
    for (i in seq_along(ngme_model$replicates)) {
      ngme_replicate <- ngme_model$replicates[[i]]
      ngme_model$replicates[[i]] <- ngme_restore_fixed_effect_scale(ngme_replicate)
    }
  }
  ngme_model
}

# helper function
# get trajs from a list of estimates
get_trajs <- function(outputs) {
  ret <- list()
  for (i in seq_along(outputs)) {
    ret[[i]] <- list()
    ret[[i]]$block_traj <- attr(outputs[[i]], "trajectory")
    for (j in seq_along(outputs[[i]]$models)) {
      ret[[i]]$models[[j]] <- list()
      ret[[i]]$models[[j]] <- attr(outputs[[i]]$models[[j]], "trajectory")
    }
  }
  ret
}

# helper function to tranform the trajectory
# input: a list (n_chain) of all parameters
# return a list (n_chain) of all parameters transposed
transform_traj <- function(traj) {
  n_chain <- length(traj)
  dfs <- list()
  for (i in 1:n_chain) {
    df <- as.data.frame(traj[[i]])
    names(df) <- NULL
    dfs[[i]] <- df
  }
  dfs
}

ngme_diag_vec <- function(x) {
  x <- as.numeric(x)
  if (length(x) == 0) {
    return(matrix(numeric(0), nrow = 0, ncol = 0))
  }
  diag(x, nrow = length(x), ncol = length(x))
}

# use estimate result to update ngme object
update_ngme_est <- function(
    ngme_replicate, est_output) {
  # Fixed effects
  names(est_output$feff) <- names(ngme_replicate$feff)
  ngme_replicate$feff <- est_output$feff
  ngme_replicate$log_likelihood <- est_output$log_likelihood

  ngme_replicate <- ngme_restore_fixed_effect_scale(ngme_replicate)

  ngme_replicate$noise <- update_noise(ngme_replicate$noise, new_noise = est_output$noise)
  for (i in seq_along(ngme_replicate$models)) {
    # update theta_K and K
    theta_K <- est_output$models[[i]]$theta_K
    if ("param_name" %in% names(ngme_replicate$models[[i]]$operator)) {
      names(theta_K) <- ngme_replicate$models[[i]]$operator$param_name
    }
    ngme_replicate$models[[i]]$operator$theta_K <-
      ngme_replicate$models[[i]]$theta_K <- theta_K

    # new_K <- ngme_replicate$models[[i]]$operator$update_K(theta_K)
    # ngme_replicate$models[[i]]$operator$K <- ngme_as_sparse(new_K)

    # update W and noise
    ngme_replicate$models[[i]]$W <- est_output$models[[i]]$W
    ngme_replicate$models[[i]]$noise <- update_noise(
      ngme_replicate$models[[i]]$noise,
      new_noise = est_output$models[[i]]
    )

    # tedious special case
    if (ngme_replicate$models[[i]]$model == "tp") {
      n1 <- ngme_replicate$models[[i]]$operator$first$n_theta_K
      n2 <- ngme_replicate$models[[i]]$operator$second$n_theta_K
      ngme_replicate$models[[i]]$operator$first$theta_K <- ngme_replicate$models[[i]]$theta_K[1:n1]
      ngme_replicate$models[[i]]$operator$second$theta_K <- ngme_replicate$models[[i]]$theta_K[(n1 + 1):(n1 + n2)]

      # update output for tp-bv model
      lat <- ngme_replicate$models[[i]]
      if (lat$operator$second$model %in% c("bv", "bv2", "bv_matern")) {
        bv <- lat$operator$second
        # theta_K = (theta_K_tp_first, theta_K_bv)
        # notice that theta_K_bv is already updated

        # number of bv-specific parameters before sub-model params
        n0 <- switch(bv$model,
          "bv" = (if (!isTRUE(bv$fix_theta)) 1 else 0) + 1 + (if (isTRUE(bv$use_c_param)) 2 else 0),
          "bv2" = 3,
          "bv_matern" = 4 - isTRUE(bv$fix_theta),
          0
        )
        n1 <- bv$first$n_theta_K
        n2 <- bv$second$n_theta_K
        bv$first$theta_K <- bv$theta_K[(n0 + 1):(n0 + n1)]
        bv$second$theta_K <- bv$theta_K[(n0 + n1 + 1):(n0 + n1 + n2)]

        ngme_replicate$models[[i]]$operator$second <- bv
      }
    }

    if (ngme_replicate$models[[i]]$model %in% c("bv", "bv_matern")) {
      n1 <- ngme_replicate$models[[i]]$operator$first$n_theta_K
      n2 <- ngme_replicate$models[[i]]$operator$second$n_theta_K

      n_param_bv <- if (ngme_replicate$models[[i]]$model == "bv") {
        (if (!isTRUE(ngme_replicate$models[[i]]$operator$fix_theta)) 1 else 0) + 1 + (if (isTRUE(ngme_replicate$models[[i]]$operator$use_c_param)) 2 else 0)
      } else {
        4 - isTRUE(ngme_replicate$models[[i]]$operator$fix_theta)
      }
      ngme_replicate$models[[i]]$operator$first$theta_K <- ngme_replicate$models[[i]]$theta_K[(n_param_bv + 1):(n1 + n_param_bv)]
      ngme_replicate$models[[i]]$operator$second$theta_K <- ngme_replicate$models[[i]]$theta_K[(n1 + n_param_bv + 1):(n1 + n2 + n_param_bv)]
    }
  }
  ngme_replicate
}

#' Print an ngme model
#'
#' @param x ngme model object
#' @param ... ...
#'
#' @return Invisibly returns \code{x}, an \code{ngme} object. The method is
#'   called for its side effect of printing a model summary.
#' @export
print.ngme <- function(x, ...) {
  print(x$replicates[[1]])
  if (x$n_repls > 1) {
    cat("Number of global replicates is", x$n_repls, "\n")
  }
  invisible(x)
}

######
check_dim <- function(ngme_model) {
  for (ngme in ngme_model$replicates) {
    if (ncol(ngme$X) != length(ngme$feff)) {
      stop("The number of columns of X is not equal to the length of feff")
    }
    for (latent in ngme$models) {
      # ncol(A) = W_size
      if (ncol(latent[["A"]]) != latent$W_size) {
        stop("The number of columns of A is not equal to the W_size of the latent model")
      }

      stopifnot(
        nrow(latent$noise$B_sigma) == latent$V_size,
        nrow(latent$noise$B_mu) == latent$V_size
      )
    }
  }
}

ngme_center_non_intercept_cols <- function(X, active_rows = NULL, center_target = NULL) {
  if (ncol(X) == 0) {
    return(list(X = X, center = numeric(0), center_target = character(0)))
  }

  if (is.null(colnames(X))) {
    colnames(X) <- as.character(seq_len(ncol(X)))
  }

  center <- setNames(numeric(ncol(X)), colnames(X))
  center_target_vec <- setNames(rep(NA_character_, ncol(X)), colnames(X))
  has_intercept <- any(startsWith(colnames(X), "(Intercept)"))
  if (!has_intercept) {
    # Keep no-intercept models on their original scale:
    # centering without an intercept changes the model semantics.
    return(list(X = X, center = center, center_target = center_target_vec))
  }

  if (is.null(active_rows)) {
    active_rows <- rep(TRUE, nrow(X))
  } else if (length(active_rows) != nrow(X)) {
    stop("active_rows must have the same length as nrow(X).")
  }

  out <- X
  non_intercept <- !startsWith(colnames(out), "(Intercept)")
  if (any(non_intercept)) {
    active_idx <- which(active_rows)
    if (length(active_idx) > 0) {
      center_vals <- colMeans(out[active_idx, non_intercept, drop = FALSE])
      center[non_intercept] <- center_vals
      out[active_idx, non_intercept] <- sweep(
        out[active_idx, non_intercept, drop = FALSE],
        2,
        center_vals,
        FUN = "-"
      )
    }

    default_target <- center_target
    if (is.null(default_target)) {
      default_target <- if ("(Intercept)" %in% colnames(out)) "(Intercept)" else NA_character_
    }
    center_target_vec[non_intercept] <- default_target
  }

  list(X = out, center = center, center_target = center_target_vec)
}

ngme_guess_intercept_targets <- function(coef_name, intercept_names) {
  if (length(intercept_names) == 0) {
    return(character(0))
  }

  grouped <- setdiff(intercept_names, "(Intercept)")
  if (length(grouped) > 0) {
    suffixes <- sub("^\\(Intercept\\)", "", grouped)
    keep <- nchar(suffixes) > 0
    if (any(keep)) {
      grouped <- grouped[keep]
      suffixes <- suffixes[keep]
      match_idx <- which(endsWith(coef_name, suffixes))
      if (length(match_idx) > 0) {
        # If multiple suffixes match, prefer the most specific (longest).
        best <- match_idx[which.max(nchar(suffixes[match_idx]))]
        return(grouped[best])
      }
    }
  }

  if ("(Intercept)" %in% intercept_names) {
    return("(Intercept)")
  }

  # Legacy fallback: if no unambiguous mapping exists, use all intercepts.
  intercept_names
}

ngme_collect_intercept_shifts <- function(beta, center, center_target = NULL) {
  if (length(beta) == 0 || is.null(names(beta))) {
    return(numeric(0))
  }

  center_beta <- center[names(beta)]
  center_beta[is.na(center_beta)] <- 0
  nonzero <- which(abs(center_beta) > 0)
  if (length(nonzero) == 0) {
    return(numeric(0))
  }

  intercept_names <- names(beta)[startsWith(names(beta), "(Intercept)")]
  if (length(intercept_names) == 0) {
    return(numeric(0))
  }

  shifts <- numeric(0)
  for (idx in nonzero) {
    coef_name <- names(beta)[idx]
    delta <- center_beta[idx] * beta[idx]
    if (abs(delta) < 1e-12) {
      next
    }

    targets <- character(0)
    if (!is.null(center_target)) {
      explicit <- center_target[coef_name]
      explicit <- explicit[!is.na(explicit) & nzchar(explicit)]
      if (length(explicit) > 0) {
        targets <- explicit[[1]]
      }
    }
    if (length(targets) == 0) {
      targets <- ngme_guess_intercept_targets(coef_name, intercept_names)
    }
    if (length(targets) == 0) {
      next
    }

    for (target in targets) {
      if (!(target %in% names(shifts))) {
        shifts[target] <- 0
      }
      shifts[target] <- shifts[target] + delta
    }
  }

  shifts
}

ngme_shift_intercepts <- function(beta, shifts, direction = c("to_raw", "to_center")) {
  direction <- match.arg(direction)
  if (length(beta) == 0 || length(shifts) == 0 || is.null(names(beta))) {
    return(beta)
  }

  sign <- if (direction == "to_raw") -1 else 1
  for (target in names(shifts)) {
    if (target %in% names(beta)) {
      beta[target] <- beta[target] + sign * shifts[target]
    }
  }

  beta
}

ngme_beta_center_to_raw <- function(beta_center, center, center_target = NULL) {
  if (length(beta_center) == 0) {
    return(beta_center)
  }

  shifts <- ngme_collect_intercept_shifts(beta_center, center, center_target)
  ngme_shift_intercepts(beta_center, shifts, direction = "to_raw")
}

ngme_beta_raw_to_center <- function(beta_raw, center, center_target = NULL) {
  if (length(beta_raw) == 0) {
    return(beta_raw)
  }

  shifts <- ngme_collect_intercept_shifts(beta_raw, center, center_target)
  ngme_shift_intercepts(beta_raw, shifts, direction = "to_center")
}

ngme_check_isotropic_normal_beta_prior <- function(prior_beta, idx_svd, tol = 1e-10) {
  if (length(idx_svd) == 0) {
    return(list(ok = TRUE, reason = NULL))
  }
  if (length(prior_beta) < max(idx_svd)) {
    return(list(ok = FALSE, reason = "invalid prior length"))
  }

  priors_fixed <- prior_beta[idx_svd]
  is_normal_coef <- vapply(priors_fixed, function(p) {
    identical(p$type, "normal") && identical(p$target, "coef")
  }, logical(1))
  if (!all(is_normal_coef)) {
    return(list(ok = FALSE, reason = "requires target='coef' normal priors"))
  }

  params <- lapply(priors_fixed, function(p) p$param)
  valid_param <- vapply(params, function(x) {
    is.numeric(x) && length(x) >= 2 &&
      all(is.finite(as.numeric(x[1:2])))
  }, logical(1))
  if (!all(valid_param)) {
    return(list(ok = FALSE, reason = "invalid normal prior parameters"))
  }

  precision <- as.numeric(vapply(params, function(x) x[[2]], numeric(1)))
  if (any(precision <= 0)) {
    return(list(ok = FALSE, reason = "normal prior precision must be positive"))
  }

  ref_precision <- precision[[1]]
  if (!all(abs(precision - ref_precision) <= tol * max(1, abs(ref_precision)))) {
    return(list(ok = FALSE, reason = "requires same variance across standardized fixed effects"))
  }

  list(ok = TRUE, reason = NULL)
}

# Map isotropic normal priors from raw fixed-effect coefficients (X basis)
# to the internal SVD basis (U basis): beta* = D V^T beta.
# Only columns indexed by idx_svd (the SVD-standardized block) are transformed.
ngme_map_isotropic_normal_priors_to_svd <- function(prior_beta, svd, idx_svd, tol = 1e-10) {
  if (is.null(svd) || length(idx_svd) == 0 || length(prior_beta) < max(idx_svd)) {
    return(prior_beta)
  }

  priors_fixed <- prior_beta[idx_svd]
  is_normal_coef <- vapply(priors_fixed, function(p) {
    identical(p$type, "normal") && identical(p$target, "coef")
  }, logical(1))
  if (!all(is_normal_coef)) {
    return(prior_beta)
  }

  params <- lapply(priors_fixed, function(p) p$param)
  valid_param <- vapply(params, function(x) {
    is.numeric(x) && length(x) >= 2 &&
      all(is.finite(as.numeric(x[1:2])))
  }, logical(1))
  if (!all(valid_param)) {
    return(prior_beta)
  }

  mean_raw <- as.numeric(vapply(params, function(x) x[[1]], numeric(1)))
  precision_raw <- as.numeric(vapply(params, function(x) x[[2]], numeric(1)))
  if (any(precision_raw <= 0)) {
    return(prior_beta)
  }

  ref_precision <- precision_raw[[1]]
  if (!all(abs(precision_raw - ref_precision) <= tol * max(1, abs(ref_precision)))) {
    # Only isotropic (same variance) normal priors are mapped.
    return(prior_beta)
  }

  d_fixed <- svd$d
  mean_svd <- as.numeric(d_fixed * drop(t(svd$v) %*% mean_raw))
  precision_svd <- ref_precision / (d_fixed^2)

  for (i in seq_along(idx_svd)) {
    prior_beta[[idx_svd[[i]]]]$param <- c(mean = mean_svd[[i]], precision = precision_svd[[i]])
  }

  prior_beta
}

ngme_restore_fixed_effect_scale <- function(ngme_replicate) {
  center <- ngme_replicate$X_center
  center_target <- ngme_replicate$X_center_target
  if (is.null(center) || length(center) == 0 || ncol(ngme_replicate$X) == 0) {
    return(ngme_replicate)
  }

  X_centered <- ngme_replicate$X
  feff_centered <- ngme_replicate$feff
  has_intercept <- any(startsWith(names(feff_centered), "(Intercept)"))

  if (isTRUE(ngme_replicate$standardize) &&
    !is.null(ngme_replicate$svd) &&
    length(ngme_replicate$svd$d) > 0) {
    svd_idx <- ngme_replicate$svd$idx
    if (is.null(svd_idx)) {
      svd_idx <- seq_len(length(ngme_replicate$svd$d))
    }
    stopifnot(length(svd_idx) == length(ngme_replicate$svd$d))

    feff_centered[svd_idx] <- as.numeric(
      ngme_replicate$svd$v %*%
        ((1 / ngme_replicate$svd$d) * feff_centered[svd_idx])
    )

    nrow_one_repl <- nrow(ngme_replicate$X)
    u_fixed <- ngme_replicate$svd$u
    if (nrow(u_fixed) != nrow_one_repl) {
      idx <- ngme_replicate$data_idx
      if (is.null(idx) || length(idx) != nrow_one_repl) {
        stop(
          "Unable to restore standardized fixed effects: replicate row count ",
          "does not match svd$u and data_idx is unavailable or invalid."
        )
      }
      if (any(idx < 1) || any(idx > nrow(u_fixed))) {
        stop(
          "Unable to restore standardized fixed effects: data_idx is out of ",
          "bounds for svd$u."
        )
      }
      u_fixed <- u_fixed[idx, , drop = FALSE]
    }

    X_fixed <- u_fixed %*% ngme_diag_vec(ngme_replicate$svd$d) %*% t(ngme_replicate$svd$v)
    X_centered[, svd_idx] <- X_fixed
  }

  X_raw <- X_centered
  feff_raw <- feff_centered
  if (has_intercept) {
    center_x <- center[colnames(X_centered)]
    center_x[is.na(center_x)] <- 0

    nonzero <- abs(center_x) > 0
    if (any(nonzero)) {
      intercept_cols <- colnames(X_centered)[startsWith(colnames(X_centered), "(Intercept)")]
      for (col_name in names(center_x)[nonzero]) {
        shift <- center_x[[col_name]]

        target <- NA_character_
        if (!is.null(center_target)) {
          explicit <- center_target[col_name]
          explicit <- explicit[!is.na(explicit) & nzchar(explicit)]
          if (length(explicit) > 0) {
            target <- explicit[[1]]
          }
        }
        if (is.na(target) || !nzchar(target) || !(target %in% colnames(X_centered))) {
          guessed <- ngme_guess_intercept_targets(col_name, intercept_cols)
          if (length(guessed) == 1 && guessed %in% colnames(X_centered)) {
            target <- guessed
          }
        }

        if (!is.na(target) && nzchar(target) && target %in% colnames(X_centered)) {
          X_raw[, col_name] <- X_raw[, col_name] + shift * X_centered[, target]
        } else {
          X_raw[, col_name] <- X_raw[, col_name] + shift
        }
      }
    }

    feff_raw <- ngme_beta_center_to_raw(feff_centered, center, center_target)
  }
  names(feff_raw) <- names(ngme_replicate$feff)

  ngme_replicate$feff <- feff_raw
  ngme_replicate$X <- X_raw
  ngme_replicate
}

#' Parse the formula for ngme function
#'
#' @param fm formula
#' @param data data.frame
#' @param control_ngme control_ngme
#' @param noise noise
#' @param group group factor
#' @param replicate replicate vector
#' @param prior_beta prior specification for fixed effects
#' @param standardize logical, whether fixed effects are standardized
#'
#' @return a list (replicate) of ngme_replicate models
ngme_parse_formula <- function(
    fm,
    data,
    control_ngme,
    noise,
    group,
    replicate,
    prior_beta,
    standardize) {
  enclos_env <- list2env(as.list(parent.frame()), parent = parent.frame(2))
  global_env_first <- list2env(as.list(parent.frame(2)), parent = parent.frame())

  tf <- terms.formula(fm, specials = c("f", "fe"))
  terms <- attr(tf, "term.labels")
  intercept <- attr(tf, "intercept")

  # order of f terms in labels
  f_order <- attr(tf, "specials")$f - 1
  fe_order <- attr(tf, "specials")$fe - 1
  # construct plain formula without f
  # watch out! terms[-double(0)] -> character(0)
  fixf <- if (length(f_order) == 0) terms else terms[-f_order]
  fixf <- if (length(fe_order) == 0) fixf else fixf[-fe_order]

  # construct fixed effect with resposne ~ intercept + fixf
  response <- deparse(attr(tf, "variables")[[2]])
  plain_fm_str <- paste(response, "~", intercept, paste(c("", fixf), collapse = " + "))
  plain_fm <- formula(plain_fm_str)
  data_eval <- as.data.frame(data)

  needed_vars <- all.vars(plain_fm)
  if (length(fe_order) > 0) {
    for (i in fe_order) {
      lang <- str2lang(terms[i])
      needed_vars <- c(needed_vars, all.vars(as.formula(lang[[2]])))
    }
  }
  needed_vars <- setdiff(unique(needed_vars), names(data_eval))
  if (length(needed_vars) > 0) {
    n_data <- nrow(data_eval)
    for (var_name in needed_vars) {
      if (!exists(var_name, envir = enclos_env, inherits = TRUE)) {
        next
      }
      value <- get(var_name, envir = enclos_env, inherits = TRUE)
      if (is.matrix(value) || is.data.frame(value)) {
        if (NROW(value) == n_data && NCOL(value) == 1) {
          value <- value[, 1, drop = TRUE]
        } else {
          next
        }
      }
      if (length(value) %in% c(1, n_data)) {
        data_eval[[var_name]] <- rep_len(value, n_data)
      }
    }
  }
  data_mask <- list2env(as.list(data_eval), parent = enclos_env)

  # eval the data
  ngme_response <- eval(stats::terms(fm)[[2]], envir = data_mask, enclos = enclos_env)
  stopifnot("Have NA in your response variable" = all(!is.na(ngme_response)))
  X_full <- model.matrix(delete.response(terms(plain_fm)), data = data_eval)
  centered <- ngme_center_non_intercept_cols(X_full)
  X_full <- centered$X
  X_center <- centered$center
  X_center_target <- centered$center_target
  if (ncol(X_full) == 0) {
    idx_svd <- integer(0)
  } else {
    if (is.null(colnames(X_full))) {
      colnames(X_full) <- as.character(seq_len(ncol(X_full)))
    }
    idx_svd <- which(!startsWith(colnames(X_full), "(Intercept)"))
  }

  # SVD-standardize only non-intercept fixed effects.
  svd <- NULL
  if (length(idx_svd) >= 1) {
    svd <- svd(X_full[, idx_svd, drop = FALSE])
    if (any((svd$d) < 1e-10)) stop("The design matrix is not full rank.")
  } else {
    standardize <- FALSE
  }

  # adding fixed effect (fe() syntax used for bivariate model..)
  for (i in fe_order) {
    lang <- str2lang(terms[i])
    stopifnot(
      "Please provide which_group=<group_name>" = !is.null(lang$which),
      "Please make sure <group_name> is in group" = lang$which %in% levels(group)
    )
    mask <- group == lang$which
    X_sub <- model.matrix(as.formula(lang[[2]]), data = data_eval)
    colnames(X_sub) <- paste0(colnames(X_sub), "_", lang$which)
    X_sub[!mask, ] <- 0
    intercept_sub <- colnames(X_sub)[startsWith(colnames(X_sub), "(Intercept)")]
    center_target_sub <- if (length(intercept_sub) == 1) intercept_sub else NA_character_
    centered_sub <- ngme_center_non_intercept_cols(
      X_sub,
      active_rows = mask,
      center_target = center_target_sub
    )
    X_sub <- centered_sub$X
    X_center <- c(X_center, centered_sub$center)
    X_center_target <- c(X_center_target, centered_sub$center_target)
    X_full <- cbind(X_full, X_sub)
  }

  # fe() terms are appended after the initial SVD-based check above.
  # Re-check rank here so aliased grouped columns fail early with a clear error.
  if (ncol(X_full) > 0) {
    qr_X <- qr(X_full)
    if (qr_X$rank < ncol(X_full)) {
      dep_cols <- colnames(X_full)[qr_X$pivot[(qr_X$rank + 1):ncol(X_full)]]
      stop(
        paste0(
          "The fixed-effects design matrix is not full rank after fe() expansion. ",
          "Please remove redundant terms. Non-identifiable columns: ",
          paste(dep_cols, collapse = ", ")
        )
      )
    }
  }

  prior_beta_compiled <- compile_beta_priors(prior_beta, colnames(X_full))
  if (standardize && !is.null(prior_beta)) {
    prior_check <- ngme_check_isotropic_normal_beta_prior(
      prior_beta = prior_beta_compiled,
      idx_svd = idx_svd
    )
    if (!isTRUE(prior_check$ok)) {
      warning(
        paste0(
          "control_opt(standardize_fixed = TRUE) was disabled because custom ",
          "prior_beta on standardized fixed effects is not isotropic normal (",
          prior_check$reason,
          ")."
        ),
        call. = FALSE
      )
      standardize <- FALSE
    }
  }

  if (standardize) {
    colnames(svd$u) <- colnames(X_full)[idx_svd]
    X_full[, idx_svd] <- svd$u
    prior_beta_compiled <- ngme_map_isotropic_normal_priors_to_svd(
      prior_beta = prior_beta_compiled,
      svd = svd,
      idx_svd = idx_svd
    )
    svd$idx <- idx_svd
  }

  # If the user supplies beta_init/feff on the original design scale, map it to
  # the standardized basis used internally (svd$u) while leaving any additional
  # fe() columns untouched.
  user_beta <- if (!is.null(control_ngme$beta_init)) control_ngme$beta_init else control_ngme$feff
  if (!is.null(user_beta)) {
    stopifnot(
      "The length of beta_init must equal the number of fixed-effect columns" =
        length(user_beta) == ncol(X_full)
    )

    # Internal estimation uses centered covariates; map user-provided beta from
    # raw scale to centered scale first.
    user_beta <- ngme_beta_raw_to_center(user_beta, X_center, X_center_target)
  }

  if (standardize && !is.null(user_beta)) {
    beta_fixed <- user_beta[idx_svd]
    beta_fixed <- as.numeric(svd$d * drop(t(svd$v) %*% beta_fixed))
    user_beta[idx_svd] <- beta_fixed
  }

  if (!is.null(user_beta)) {
    control_ngme$beta_init <- user_beta
    control_ngme$feff <- user_beta # legacy name retained for downstream code
  }

  ########## parse models terms
  pre_model <- list()
  all_gaussian <- noise$noise_type == "normal"
  idx_effect <- 1
  idx_field <- 1 # for setting names
  for (i in f_order) {
    str <- gsub("^f\\(", "ngme2::f(", terms[i])
    lang <- str2lang(str)
    if (is.null(lang$model)) stop("Please provide model=<model_name>.")
    # pass extra argument into f
    if (is.null(lang$data)) lang$data <- data
    if (is.null(lang$group)) lang$group <- group
    if (is.null(lang$name) && lang$model == "re") {
      lang$name <- paste0("effect", idx_effect)
      idx_effect <- idx_effect + 1
    }
    if (is.null(lang$name)) {
      lang$name <- paste0("field", idx_field)
      idx_field <- idx_field + 1
    }

    pre_model[[lang$name]] <- lang
  }

  levels <- levels(replicate)
  blocks_rep <- list() # of length n_repl

  # Validate mesh lists if any f() calls use them
  for (tmp in pre_model) {
    if (!is.null(tmp$mesh) && is.list(tmp$mesh) && !inherits(tmp$mesh, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))) {
      n_meshes <- length(tmp$mesh)
      n_repls <- length(levels)
      if (n_meshes < n_repls) {
        stop(paste("Insufficient meshes provided for field '", tmp$name, "'. ",
          "Expected ", n_repls, " meshes for ", n_repls, " replicates, ",
          "but only ", n_meshes, " meshes were provided.",
          sep = ""
        ))
      }
    }
  }

  noise_new <- update_noise(noise, n = length(ngme_response))
  if (noise_new$corr_measurement) {
    stopifnot(
      "Please make sure the len(index_corr) == observations" =
        length(ngme_response) == length(noise_new$index_corr)
    )
  }
  for (level in levels) {
    idx <- replicate == level
    data_idx <- which(idx) # record the original index

    Y <- ngme_response[idx]
    X <- X_full[idx, , drop = FALSE]
    svd_rep <- NULL
    if (standardize) {
      svd_rep <- list(
        u = svd$u[idx, , drop = FALSE],
        d = svd$d,
        v = svd$v,
        idx = svd$idx
      )
    }

    # re-evaluate each f model using idx
    models_rep <- list()
    for (tmp in pre_model) {
      tmp$subset <- idx

      # Evaluate mesh parameter to get actual value
      actual_mesh <- if (is.null(tmp$mesh)) NULL else eval(tmp$mesh, envir = data, enclos = global_env_first)

      # Handle mesh selection for different replicates
      if (
        tmp$model != "spacetime" &&
          !is.null(actual_mesh) &&
          is.list(actual_mesh) &&
          !inherits(actual_mesh, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))
      ) {
        # mesh is a list of meshes for different replicates
        mesh_list <- actual_mesh

        # Convert level to numeric index if needed
        replicate_idx <- which(levels == level)

        # Check if we have enough meshes for this replicate
        if (replicate_idx <= length(mesh_list)) {
          selected_mesh <- mesh_list[[replicate_idx]]
        } else {
          stop(paste("Not enough meshes provided for replicate", level, ". Expected at least", replicate_idx, "meshes, but only", length(mesh_list), "provided."))
        }

        # Replace the mesh parameter in the call with the selected mesh
        tmp$mesh <- selected_mesh

        # Force A matrix to be NULL so it gets rebuilt with the correct mesh
        tmp$A <- NULL
      }

      model_eval <- eval(tmp, envir = data, enclos = global_env_first)

      models_rep[[model_eval$name]] <- model_eval
      if (all(model_eval$noise$noise_type != "normal")) all_gaussian <- FALSE
    }
    # give initial value (whole dataset)
    lm.model <- stats::lm.fit(X_full, ngme_response)
    if (is.null(control_ngme$beta_init) && is.null(control_ngme$feff)) {
      control_ngme$beta_init <- control_ngme$feff <- lm.model$coeff
    }
    if (is.null(noise$theta_sigma)) noise$theta_sigma <- log(sd(lm.model$residuals))
    noise_rep <- subset_noise(noise_new, sub_idx = idx, compute_corr = FALSE)
    group_rep <- group[idx]

    # Re-order according to index_corr!
    # s.t. noise$index_corr=1,1,2,2,3,4,4,....

    # p_oder is the order after permutation
    # original is just 1 2 3, ...
    p_order <- seq_along(Y)
    if (noise$corr_measurement) {
      stopifnot(
        "The length of noise$index_corr should be the same as the number of observations" = length(noise_rep$index_corr) == sum(idx),
        "Now more than 2 locations are correlated in 1 replicate is not allowed" = !any(table(noise_rep$index_corr) > 2)
      )
      p_order <- order(noise_rep$index_corr)
      data_idx <- data_idx[p_order]
      X <- X[p_order, , drop = FALSE]
      Y <- Y[p_order]
      if (standardize) svd_rep$u <- svd_rep$u[p_order, , drop = FALSE]

      group_rep <- group_rep[p_order]
      for (j in seq_along(models_rep)) {
        models_rep[[j]]$A <- models_rep[[j]]$A[p_order, , drop = FALSE]
      }

      # update noise, consider index_corr
      noise_rep <- subset_noise(noise_rep, sub_idx = p_order, compute_corr = TRUE)
    }

    blocks_rep[[level]] <- ngme_replicate(
      data_idx = data_idx,
      Y = Y,
      X = X,
      prior_beta = prior_beta_compiled,
      group = group_rep,
      noise = noise_rep,
      models = models_rep,
      control_ngme = control_ngme,
      n_repl = length(levels),
      all_gaussian = all_gaussian,
      standardize = standardize,
      X_center = X_center,
      X_center_target = X_center_target,
      svd = svd_rep
    )
  }

  n_repls <- length(blocks_rep)
  n_params <- blocks_rep[[1]]$n_params

  structure(
    list(
      replicates   = blocks_rep,
      n_repls      = n_repls,
      n_params     = n_params,
      repls_ngme   = replicate,
      control_ngme = control_ngme
    ),
    class = "ngme"
  )
}

#' Helper function to compute the index_corr vector
#'
#' @param map used as location to compute distance, can be 1d (numerical) or 2d (data.frame)
#' @param eps threshold to determine if two points are close (if close, we consider them as the same point)
#'
#' @return the index_corr vector for ngme correlated measurement noise
#' @examples
#' x_coord <- c(1.11, 1.12, 2, 1.3, 1.3)
#' y_coord <- c(2.11, 2.11, 2, 3.3, 3.3)
#' coord <- data.frame(x_coord, y_coord)
#' compute_index_corr_from_map(map = coord, 0.1)
#' @export
compute_index_corr_from_map <- function(map, eps = 0.1) {
  if (is.null(map)) {
    return(NULL)
  }

  index_corr <- 1:length_map(map)
  if (length(index_corr) == 1) {
    return(index_corr)
  }
  for (i in 2:length_map(map)) {
    for (j in 1:(i - 1)) {
      # compute dist of i and j entry
      d <- dist(sub_map(map, c(i, j)))
      if (d < eps) {
        index_corr[j] <- index_corr[i]
      }
    }
  }

  as.integer(as.factor(index_corr))
}

# idx: integer vector, indicating which observations are correlated
compute_corr_index <- function(idx) {
  stopifnot(sum(round(idx - as.integer(idx))) < 1e8)
  n <- length(idx)
  rows <- cols <- 1:n

  has_correlation <- rep(FALSE, n)
  unique_idx <- unique(idx)
  count <- 1
  for (i in seq_along(unique_idx)) {
    idx_i <- which(idx == unique_idx[i])
    if (length(idx_i) == 1) next
    stopifnot(
      "Now we don't accept measurement noise over 2 places are correlated" = length(idx_i) == 2
    )
    rows[n + count] <- max(idx_i)
    cols[n + count] <- min(idx_i)
    count <- count + 1
    has_correlation[idx_i] <- TRUE
  }

  sort_idx <- order(cols, rows) # col_major order
  list(
    cor_rows = rows[sort_idx] - 1,
    cor_cols = cols[sort_idx] - 1,
    has_correlation = has_correlation,
    n_corr_pairs = sum(has_correlation) / 2
  )
}

#' Summary of ngme fit result
#' @param object an object of class \code{ngme}
#' @param name name of the latent model to be summarized (if NULL, will print all)
#' @param ... other arguments
#'
#' @return a list of summary
#' @export
summary.ngme <- function(
    object,
    name = NULL,
    ...) {
  stopifnot(inherits(object, "ngme"))

  result <- object

  if (!is.null(name)) {
    ngme_rep <- result$replicates[[1]]
    stopifnot(
      "Please provide the correct name of the model" =
        name %in% names(ngme_rep$models)
    )
    result <- ngme_rep$models[[name]]
  }

  result
}

#' Access the result of a ngme fitted model
#' @param ngme_object a ngme fitted model object
#' @param model latent model name to filter by (e.g., "my_ar1", "my_matern", "my_rw1", etc.), "data" for fixed effects and measurement noise, if NULL, return all models
#' @param transformed logical, if TRUE (default) return transformed parameters, if FALSE return raw parameters
#'
#' @return a list of parameters for the specified model, or all models if model is NULL
#' @export
#'
#' @examples
#' # Fit a simple AR(1) model
#' set.seed(1)
#' Y <- 1:10
#' n_obs <- length(Y)
#' x1 <- rnorm(n_obs)
#' x2 <- runif(n_obs)
#'
#' ngme_out <- ngme(
#'   Y ~ x1 + x2 + f(
#'     1:n_obs,
#'     name = "my_ar",
#'     model = ar1(rho = 0.5),
#'     noise = noise_nig(mu = 2, sigma = 3, nu = 1)
#'   ),
#'   data = data.frame(x1 = x1, x2 = x2, Y = Y),
#'   control_opt = control_opt(estimation = FALSE)
#' )
#'
#' # Get all model parameters (transformed)
#' all_params <- ngme_result(ngme_out)
#' # Returns: list(my_ar = list(rho = 0.5, mu = 2, sigma = 3, nu = 1),
#' #               data = list(fixed_effects = c(...), sigma = 0.5))
#'
#' # Get parameters for specific latent model
#' ar_params <- ngme_result(ngme_out, model = "my_ar")
#' # Returns: list(rho = 0.5, mu = 2, sigma = 3, nu = 1)
#'
#' # Get raw (untransformed) parameters
#' ar_raw <- ngme_result(ngme_out, model = "my_ar", transformed = FALSE)
#' # Returns: list(theta_rho = 1.099, mu = 2, sigma = 3, nu = 1)
#'
#' # Get fixed effects and measurement noise
#' data_params <- ngme_result(ngme_out, model = "data")
#' # Returns: list(fixed_effects = c(...), sigma = 0.5, ...)
#'
#' # For models with multiple latent processes
#' ngme_out2 <- ngme(
#'   Y ~ x1 + x2 + f(
#'     1:n_obs,
#'     name = "my_ar",
#'     model = ar1(rho = 0.5),
#'     noise = noise_nig(mu = 2, sigma = 3, nu = 1)
#'   ) + f(
#'     1:n_obs,
#'     name = "my_ou",
#'     model = ou(),
#'     noise = noise_normal(sigma = 1)
#'   ),
#'   data = data.frame(x1 = x1, x2 = x2, Y = Y),
#'   control_opt = control_opt(estimation = FALSE)
#' )
#'
#' # Get all models
#' all_models <- ngme_result(ngme_out2)
#' # Returns: list(my_ar = list(...), my_ou = list(...), data = list(...))
#'
#' # Get specific model
#' ou_params <- ngme_result(ngme_out2, model = "my_ou")
#' # Returns: list(theta_K1 = 0.5, sigma = 1)
ngme_result <- function(
    ngme_object,
    model = NULL,
    transformed = TRUE) {
  stopifnot(inherits(ngme_object, "ngme"))

  # Extract all parameters using extract_parameters
  all_params <- extract_parameters(ngme_object)

  # Choose transformed or raw parameters
  params <- if (transformed) all_params$transformed else all_params$raw

  # Return specific model or all models
  if (is.null(model)) {
    return(params)
  } else {
    stopifnot(
      "Please provide the correct model name" = model %in% names(params)
    )
    return(params[[model]])
  }
}

test_that("test sum of ar1 and matern", {
  skip_on_cran()
  withr::local_seed(123)
  library(ngme2)
  library(fmesher)

  # 1. Simulation
  n_obs <- 1000
  time_points <- 1:n_obs

  # AR1 process
  rho_true <- 0.7
  ar1_model <- f(time_points, model = ar1(rho = rho_true), noise = noise_normal(sigma = 1.5))
  W_ar1 <- simulate(ar1_model)[[1]]
  sd(W_ar1)

  # Matern process
  # Create a 1D mesh
  mesh <- fm_mesh_1d(seq(1, n_obs, length.out = 50))
  kappa_true <- 0.3
  matern_model <- f(time_points, model = matern(mesh, kappa = kappa_true), noise = noise_normal(sigma = 4.0))
  W_matern <- simulate(matern_model)[[1]]
  sd(W_matern)

  # Measurement noise
  sigma_eps_true <- 0.5
  Y <- W_ar1 + W_matern + rnorm(n_obs, sd = sigma_eps_true)

  # 2. Estimation
  # We need to specify the model structure for estimation
  # Note: For Matern 1D, we need to provide the mesh

  out <- ngme(
    Y ~ 0 + f(time_points, model = ar1(), name = "ar1_comp") +
      f(time_points, model = matern(mesh), name = "matern_comp"),
    data = data.frame(Y = Y),
    control_opt = control_opt(
      seed = 123,
      iterations = 1000,
      optimizer = adamW(),
      n_parallel_chain = 4,
      print_check_info = FALSE,
      verbose = FALSE
    )
  )
  traceplot(out, name = "ar1_comp", hline = c(rho_true, 1.5))
  traceplot(out, name = "matern_comp", hline = c(kappa_true, 4))

  # 3. Comparison
  res_ar1 <- ngme_result(out, "ar1_comp")
  res_matern <- ngme_result(out, "matern_comp")
  res_noise <- ngme_result(out, "data") # Measurement noise is usually under "data" or similar if not specified in f()

  # Parallel stochastic fitting is somewhat noisy here, so keep the assertions
  # focused on recovering a sensible decomposition rather than exact targets.
  expect_true(is.finite(res_ar1$rho))
  expect_true(res_ar1$rho > 0.4 && res_ar1$rho < 1.05)
  expect_true(is.finite(res_matern$kappa))
  expect_true(res_matern$kappa > 0)
  expect_true(is.finite(res_noise$sigma))
  expect_true(res_noise$sigma > 0.2 && res_noise$sigma < 0.85)

  # Print results for manual inspection
  print(paste("True AR1 rho:", rho_true, "Est:", res_ar1$rho))
  print(paste("True Matern kappa:", kappa_true, "Est:", res_matern$kappa)) # Assuming kappa is directly available or we need to extract it
  print(paste("True Sigma eps:", sigma_eps_true, "Est:", res_noise$sigma))
})

test_that("grouped fe() terms fail early when design matrix is rank deficient", {
  df <- data.frame(
    lon = c(11.2, 11.2, 13.1, 13.1, 15, 15),
    lat = c(46.7, 46.7, 46.7, 46.7, 46.7, 46.7),
    direction = c("u_wind", "v_wind", "u_wind", "v_wind", "u_wind", "v_wind"),
    wind = c(0.1, -0.2, 0.3, 0.1, 0.8, 0.3)
  )

  expect_error(
    ngme(
      formula = wind ~
        0 +
        fe(~ 1 + lon, which = "u_wind") +
        fe(~ 1 + lat, which = "v_wind"),
      data = df,
      group = df$direction,
      family = noise_normal(),
      control_opt = control_opt(estimation = FALSE)
    ),
    "not full rank.*lat_v_wind"
  )
})

test_that("grouped fe() regression matches lm.fit and is invariant to SVD standardization", {
  skip_on_cran()
  withr::local_seed(20260225)

  n_per_group <- 200
  n <- 2 * n_per_group

  direction <- rep(c("u_wind", "v_wind"), each = n_per_group)
  lon <- runif(n, min = 8, max = 16)
  lat <- runif(n, min = 42, max = 50)

  true_coef <- c(
    "(Intercept)_u_wind" = 1.5,
    "lon_u_wind" = 0.8,
    "(Intercept)_v_wind" = -2.0,
    "lat_v_wind" = 0.35
  )

  wind <- ifelse(
    direction == "u_wind",
    true_coef["(Intercept)_u_wind"] + true_coef["lon_u_wind"] * lon,
    true_coef["(Intercept)_v_wind"] + true_coef["lat_v_wind"] * lat
  )

  df <- data.frame(
    lon = lon,
    lat = lat,
    direction = direction,
    wind = wind
  )

  fit_std <- ngme(
    formula = wind ~
      0 +
      fe(~ 1 + lon, which = "u_wind") +
      fe(~ 1 + lat, which = "v_wind"),
    data = df,
    group = df$direction,
    family = noise_normal(),
    control_opt = control_opt(
      estimation = FALSE,
      standardize_fixed = TRUE
    )
  )

  fit_nosvd <- ngme(
    formula = wind ~
      0 +
      fe(~ 1 + lon, which = "u_wind") +
      fe(~ 1 + lat, which = "v_wind"),
    data = df,
    group = df$direction,
    family = noise_normal(),
    control_opt = control_opt(
      estimation = FALSE,
      standardize_fixed = FALSE
    )
  )

  ngme_rep_std <- fit_std$replicates[[1]]
  ngme_rep_nosvd <- fit_nosvd$replicates[[1]]

  coef_ngme_std <- ngme_rep_std$feff
  coef_ngme_nosvd <- ngme_rep_nosvd$feff
  coef_lm <- stats::lm.fit(ngme_rep_std$X, ngme_rep_std$Y)$coefficients

  pred_ngme_std <- as.numeric(ngme_rep_std$X %*% coef_ngme_std)
  pred_ngme_nosvd <- as.numeric(ngme_rep_nosvd$X %*% coef_ngme_nosvd)
  pred_lm <- as.numeric(ngme_rep_std$X %*% coef_lm)

  # Structural zeros from fe() should remain zeros outside each group.
  expect_equal(
    unname(ngme_rep_std$X[direction != "u_wind", "lon_u_wind"]),
    rep(0, sum(direction != "u_wind")),
    tolerance = 1e-10
  )
  expect_equal(
    unname(ngme_rep_std$X[direction != "v_wind", "lat_v_wind"]),
    rep(0, sum(direction != "v_wind")),
    tolerance = 1e-10
  )

  expect_equal(max(abs(coef_ngme_std - coef_lm)), 0, tolerance = 1e-10)
  expect_equal(max(abs(pred_ngme_std - pred_lm)), 0, tolerance = 1e-10)
  expect_equal(max(abs(pred_ngme_std - wind)), 0, tolerance = 1e-10)

  # Compare with no-SVD fixed-effect standardization.
  expect_equal(coef_ngme_std, coef_ngme_nosvd, tolerance = 1e-10)
  expect_equal(pred_ngme_std, pred_ngme_nosvd, tolerance = 1e-10)

  # Slopes should be recovered exactly in this noiseless simulation.
  expect_equal(coef_ngme_std["lon_u_wind"], true_coef["lon_u_wind"], tolerance = 1e-10)
  expect_equal(coef_ngme_std["lat_v_wind"], true_coef["lat_v_wind"], tolerance = 1e-10)
})

test_that("demeaned fixed effects are mapped back to raw parameterization", {
  withr::local_seed(123)

  df <- data.frame(
    x1 = rnorm(80),
    x2 = rnorm(80)
  )
  df$y <- 2 - 0.5 * df$x1 + 1.3 * df$x2 + rnorm(80, sd = 0.2)

  fit <- ngme(
    y ~ x1 + x2,
    data = df,
    control_opt = control_opt(
      estimation = FALSE,
      standardize_fixed = TRUE
    )
  )

  coef_ngme <- fit$replicates[[1]]$feff
  coef_lm <- stats::lm.fit(model.matrix(~ x1 + x2, data = df), df$y)$coefficients
  expect_equal(coef_ngme, coef_lm, tolerance = 1e-10)

  beta_init <- c("(Intercept)" = 2, "x1" = -0.5, "x2" = 1.3)
  fit_init <- ngme(
    y ~ x1 + x2,
    data = df,
    control_ngme = control_ngme(beta = beta_init),
    control_opt = control_opt(
      estimation = FALSE,
      standardize_fixed = TRUE
    )
  )
  expect_equal(fit_init$replicates[[1]]$feff, beta_init, tolerance = 1e-10)
})

test_that("no-intercept fixed effects are not centered and keep original semantics", {
  withr::local_seed(456)

  n <- 100
  df <- data.frame(
    x1 = runif(n),
    x2 = rnorm(n)
  )
  beta_true <- c("x1" = 1.8, "x2" = -0.7)
  X_raw <- model.matrix(~ 0 + x1 + x2, data = df)
  df$y <- as.numeric(X_raw %*% beta_true) + rnorm(n, sd = 0.1)

  fit <- ngme(
    y ~ 0 + x1 + x2,
    data = df,
    control_opt = control_opt(
      estimation = FALSE,
      standardize_fixed = TRUE
    )
  )

  rep1 <- fit$replicates[[1]]
  coef_lm <- stats::lm.fit(X_raw, df$y)$coefficients

  expect_equal(as.numeric(rep1$X), as.numeric(X_raw), tolerance = 1e-10)
  expect_equal(dim(rep1$X), dim(X_raw))
  expect_equal(colnames(rep1$X), colnames(X_raw))
  expect_equal(rep1$feff, coef_lm, tolerance = 1e-10)
})

test_that("fixed-effect standardization excludes intercept columns", {
  withr::local_seed(789)

  n <- 120
  df <- data.frame(
    x1 = rnorm(n),
    x2 = runif(n)
  )
  df$y <- 1.5 + 0.7 * df$x1 - 0.3 * df$x2 + rnorm(n, sd = 0.2)

  fit <- ngme(
    y ~ x1 + x2,
    data = df,
    control_opt = control_opt(
      estimation = FALSE,
      standardize_fixed = TRUE
    )
  )

  rep1 <- fit$replicates[[1]]
  X_raw <- model.matrix(~ x1 + x2, data = df)
  expect_true(rep1$standardize)
  expect_equal(rep1$svd$idx, which(!startsWith(colnames(X_raw), "(Intercept)")))
  expect_equal(colnames(rep1$svd$u), colnames(X_raw)[rep1$svd$idx])
  expect_equal(as.numeric(rep1$X), as.numeric(X_raw), tolerance = 1e-10)
})

test_that("restore uses replicate indices when svd$u is global", {
  x <- c(0, 1, 2, 10, 11, 12)
  X_raw <- cbind("(Intercept)" = 1, "x" = x)
  centered <- ngme2:::ngme_center_non_intercept_cols(X_raw)
  X_centered <- centered$X

  svd_full <- svd(X_centered)
  colnames(svd_full$u) <- colnames(X_raw)

  beta_raw <- c("(Intercept)" = 3, "x" = 0.5)
  beta_centered <- ngme2:::ngme_beta_raw_to_center(beta_raw, centered$center)
  beta_svd <- as.numeric(svd_full$d * drop(t(svd_full$v) %*% beta_centered))
  names(beta_svd) <- names(beta_raw)

  idx_rep2 <- 4:6
  ngme_rep <- list(
    X = svd_full$u[idx_rep2, , drop = FALSE],
    feff = beta_svd,
    X_center = centered$center,
    X_center_target = centered$center_target,
    standardize = TRUE,
    svd = svd_full,
    data_idx = idx_rep2
  )

  restored <- ngme2:::ngme_restore_fixed_effect_scale(ngme_rep)

  expect_equal(restored$X, X_raw[idx_rep2, , drop = FALSE], tolerance = 1e-10)
  expect_equal(restored$feff, beta_raw, tolerance = 1e-10)
})

test_that("stored fixed-effect trajectories are on raw scale (intercept included)", {
  withr::local_seed(20260226)

  n <- 80
  df <- data.frame(
    x1 = runif(n),
    x2 = rexp(n)
  )
  beta_true <- c("(Intercept)" = 3.0, "x1" = -1.2, "x2" = 1.8)
  X <- model.matrix(~ x1 + x2, data = df)
  df$y <- as.numeric(X %*% beta_true) + rnorm(n, sd = 0.2)

  fit <- ngme(
    y ~ x1 + x2,
    data = df,
    family = noise_normal(),
    control_opt = control_opt(
      optimizer = sgd(),
      iterations = 60,
      burnin = 10,
      n_parallel_chain = 1,
      standardize_fixed = FALSE,
      store_traj = TRUE,
      verbose = FALSE,
      print_check_info = FALSE,
      seed = 20260226
    )
  )

  rep1 <- fit$replicates[[1]]
  block_traj <- attr(rep1, "block_traj")
  expect_true(length(block_traj) == 1)

  n_feff <- length(rep1$feff)
  feff_idx <- (nrow(block_traj[[1]]) - n_feff + 1):nrow(block_traj[[1]])
  last_feff <- unlist(block_traj[[1]][feff_idx, ncol(block_traj[[1]]), drop = FALSE], use.names = FALSE)
  names(last_feff) <- names(rep1$feff)

  expect_equal(last_feff, rep1$feff, tolerance = 1e-3)
})

test_that("trace trajectories use fixed-effect variable names when available", {
  withr::local_seed(20260226)

  n <- 60
  df <- data.frame(
    x1 = runif(n),
    x2 = rnorm(n)
  )
  df$y <- 2.5 - 0.8 * df$x1 + 1.1 * df$x2 + rnorm(n, sd = 0.2)

  fit <- ngme(
    y ~ x1 + x2,
    data = df,
    family = noise_normal(),
    control_opt = control_opt(
      optimizer = sgd(),
      iterations = 40,
      burnin = 10,
      n_parallel_chain = 1,
      standardize_fixed = FALSE,
      store_traj = TRUE,
      verbose = FALSE,
      print_check_info = FALSE,
      seed = 20260226
    )
  )

  traj <- get_trace_trajectories(fit, name = "data", apply_transform = FALSE)
  fe_names <- names(fit$replicates[[1]]$feff)
  expect_equal(utils::tail(traj$parameter_names, length(fe_names)), fe_names)
})

test_that("trace trajectories disambiguate duplicated latent parameter names", {
  fake_latent <- structure(
    list(
      operator = list(
        param_name = c("rho", "rho"),
        param_trans = list(identity, identity)
      ),
      noise = noise_normal(fix_theta_sigma = TRUE)
    ),
    class = "ngme_model"
  )
  attr(fake_latent, "lat_traj") <- list(matrix(
    c(0.10, 0.20, 0.30,
      0.60, 0.70, 0.80),
    nrow = 2,
    byrow = TRUE
  ))

  fake_ngme <- structure(
    list(
      replicates = list(
        list(
          models = list(field1 = fake_latent),
          feff = numeric(0)
        )
      )
    ),
    class = "ngme"
  )

  traj <- get_trace_trajectories(fake_ngme, name = "field1", apply_transform = FALSE)
  expect_equal(length(traj$parameter_names), 2)
  expect_equal(length(unique(traj$parameter_names)), 2)
  expect_equal(sort(names(traj$trajectories)), sort(traj$parameter_names))
})

test_that("traceplot supports configurable facet columns via ncol", {
  fake_latent <- structure(
    list(
      operator = list(
        param_name = c("rho", "rho"),
        param_trans = list(identity, identity)
      ),
      noise = noise_normal(fix_theta_sigma = TRUE)
    ),
    class = "ngme_model"
  )
  attr(fake_latent, "lat_traj") <- list(matrix(
    c(0.10, 0.20, 0.30,
      0.60, 0.70, 0.80),
    nrow = 2,
    byrow = TRUE
  ))

  fake_ngme <- structure(
    list(
      replicates = list(
        list(
          models = list(field1 = fake_latent),
          feff = numeric(0)
        )
      )
    ),
    class = "ngme"
  )

  p <- traceplot(fake_ngme, name = "field1", combine = TRUE, ncol = 1)
  expect_true(inherits(p, "ggplot"))
  expect_equal(p$facet$params$ncol, 1)

  expect_error(
    traceplot(fake_ngme, name = "field1", combine = TRUE, ncol = 0),
    "`ncol` must be a positive integer or NULL."
  )
})

test_that("isotropic normal beta prior is invariant to fixed-effect standardization", {
  withr::local_seed(20260226)

  n <- 80
  df <- data.frame(
    x1 = rnorm(n),
    x2 = rnorm(n)
  )
  df$y <- 1.2 + 0.8 * df$x1 - 1.5 * df$x2 + rnorm(n, sd = 0.4)

  beta_prior <- priors(beta = prior_normal(mean = 0, sd = 0.3))

  fit_std <- ngme(
    y ~ x1 + x2,
    data = df,
    prior_beta = beta_prior,
    control_opt = control_opt(
      optimizer = adam(),
      burnin = 100,
      iterations = 400,
      n_parallel_chain = 1,
      seed = 101,
      standardize_fixed = TRUE,
      verbose = FALSE,
      print_check_info = FALSE
    )
  )

  fit_raw <- ngme(
    y ~ x1 + x2,
    data = df,
    prior_beta = beta_prior,
    control_opt = control_opt(
      optimizer = adam(),
      burnin = 100,
      iterations = 400,
      n_parallel_chain = 1,
      seed = 101,
      standardize_fixed = FALSE,
      verbose = FALSE,
      print_check_info = FALSE
    )
  )

  coef_std <- fit_std$replicates[[1]]$feff
  coef_raw <- fit_raw$replicates[[1]]$feff

  expect_equal(coef_std, coef_raw, tolerance = 1e-5)
})

test_that("incompatible custom beta prior disables fixed-effect standardization", {
  withr::local_seed(20260226)

  n <- 40
  df <- data.frame(
    x1 = rnorm(n),
    x2 = rnorm(n)
  )
  df$y <- 0.5 + 1.1 * df$x1 - 0.7 * df$x2 + rnorm(n, sd = 0.2)

  # Different variances across fixed effects => not isotropic normal.
  prior_bad <- priors(
    "(Intercept)" = prior_normal(mean = 0, sd = 1.0),
    x1 = prior_normal(mean = 0, sd = 0.3),
    x2 = prior_normal(mean = 0, sd = 0.8)
  )

  expect_warning(
    fit <- ngme(
      y ~ x1 + x2,
      data = df,
      prior_beta = prior_bad,
      control_opt = control_opt(
        estimation = FALSE,
        standardize_fixed = TRUE
      )
    ),
    "standardize_fixed = TRUE\\) was disabled"
  )

  rep1 <- fit$replicates[[1]]
  X_raw <- model.matrix(~ x1 + x2, data = df)

  expect_false(rep1$standardize)
  expect_equal(as.numeric(rep1$X), as.numeric(X_raw), tolerance = 1e-10)
  expect_equal(dim(rep1$X), dim(X_raw))
  expect_equal(colnames(rep1$X), colnames(X_raw))
})

test_that("start works when previous fit disabled fixed-effect standardization", {
  withr::local_seed(20260226)

  n <- 40
  df <- data.frame(
    x1 = rnorm(n),
    x2 = rnorm(n)
  )
  df$y <- 0.8 + 1.2 * df$x1 - 0.4 * df$x2 + rnorm(n, sd = 0.2)

  prior_bad <- priors(
    "(Intercept)" = prior_normal(mean = 0, sd = 1.0),
    x1 = prior_normal(mean = 0, sd = 0.3),
    x2 = prior_normal(mean = 0, sd = 0.8)
  )

  expect_warning(
    fit_prev <- ngme(
      y ~ x1 + x2,
      data = df,
      prior_beta = prior_bad,
      control_opt = control_opt(
        estimation = FALSE,
        standardize_fixed = TRUE
      )
    ),
    "standardize_fixed = TRUE\\) was disabled"
  )
  expect_false(fit_prev$replicates[[1]]$standardize)

  expect_error(
    fit_next <- ngme(
      y ~ x1 + x2,
      data = df,
      start = fit_prev,
      control_opt = control_opt(
        estimation = FALSE,
        standardize_fixed = TRUE
      )
    ),
    NA
  )

  expect_true(fit_next$replicates[[1]]$standardize)
})

test_that("generic_ns basic functionality works", {
  # Create test matrices
  n <- 5
  A <- matrix(1, n, n)
  B <- matrix(2, n, n)
  alpha <- 0.4

  # Create a simple generic_ns model
  model <- generic_ns(
    theta_K = list(alpha = c(alpha)),
    matrices = list(A, B),
    h = rep(1, n),
    position = list(c(2, 1), c(3)) # Diagonal(alpha) * A + B
  )

  # First position c(2, 1) means:
  # - Matrix at index 2 (A) multiplied by
  # - Matrix at index 1 (Diagonal with alpha)
  # Second position c(3) means matrix at index 3 (B)
  # So expected_K = A * Diagonal(alpha) + B
  expected_K <- A %*% Matrix::Diagonal(n, alpha) + B

  expect_equal(as.matrix(model$K), as.matrix(expected_K))
  expect_equal(length(model$matrices), 2) # Only original matrices are stored
  expect_equal(model$h, rep(1, n))
})

test_that("generic_ns with parameter transformations works", {
  n <- 5
  A <- matrix(1, n, n)
  B <- matrix(2, n, n)
  alpha <- 0.6

  # Create model with parameter transformations
  model <- generic_ns(
    theta_K = list(alpha = c(log(alpha))),
    matrices = list(A, B),
    h = rep(1, n),
    trans = list(alpha = "exp"),
    position = list(c(1, 2), c(3)) # Diagonal(exp(log(alpha))) * A + B = Diagonal(alpha) * A + B
  )

  # With exp transformation, expected_K = A * Diagonal(alpha) + B
  expected_K <- A %*% Matrix::Diagonal(n, alpha) + B

  expect_equal(as.matrix(model$K), as.matrix(expected_K))

  # Check flat parameters
  expect_equal(model$theta_K, c(alpha = log(alpha)))
})

test_that("generic_ns with custom basis expansion works", {
  n <- 10
  A <- matrix(1, n, n)
  B <- matrix(2, n, n)
  C <- matrix(3, n, n)

  # Create basis matrices (simple example for testing)
  basis_alpha <- matrix(rnorm(n), n, 1) # One coefficient per location
  alpha <- 0.5

  # Create model with basis expansion
  model <- generic_ns(
    theta_K = list(alpha = c(alpha)),
    matrices = list(A, B, C),
    h = rep(1, n),
    B_theta_K = list(alpha = basis_alpha),
    trans = list(alpha = "identity"),
    position = list(c(1, 2), c(3, 2, 1), c(4)) # Diagonal(basis_alpha * alpha) * A + B
  )

  # Expected: K = D_alpha * A + B
  alpha_vec <- as.vector(basis_alpha %*% alpha)
  D_alpha <- Matrix::Diagonal(x = alpha_vec)
  expected_K <- D_alpha %*% A + B %*% A %*% D_alpha + C

  expect_equal(as.matrix(model$K), as.matrix(expected_K))

  expect_false(identical(as.matrix(A %*% D_alpha), as.matrix(D_alpha %*% A)))
})

test_that("generic_ns with B_theta_K defaults to matrices of 1s", {
  n <- 5
  A <- matrix(1, n, n)
  B <- matrix(2, n, n)
  alpha <- 0.7

  # Create model without explicit B_theta_K
  model <- generic_ns(
    theta_K = list(alpha = c(alpha)),
    matrices = list(A, B),
    h = rep(1, n),
    position = list(c(1, 2), c(3)) # Diagonal(alpha*ones(n)) * A + B = Diagonal(alpha) * A + B
  )

  # Default B_theta_K is a matrix of 1s, so expected_K = A * Diagonal(alpha) + B
  expected_K <- A %*% Matrix::Diagonal(n, alpha) + B

  expect_equal(as.matrix(model$K), as.matrix(expected_K))
})

test_that("generic_ns handles complex matrix combinations", {
  n <- 5
  A <- matrix(1, n, n)
  B <- matrix(2, n, n)
  C <- matrix(3, n, n)
  D <- matrix(4, n, n)

  alpha <- 0.8
  beta <- 1.2

  # Create a complex model with multiple operations
  model <- generic_ns(
    theta_K = list(alpha = c(alpha), beta = c(beta)),
    matrices = list(A, B, C, D),
    h = rep(1, n),
    trans = list(alpha = "identity", beta = "identity"),
    position = list(c(1, 3), c(2, 4), c(1, 5), c(2, 6))
    # This represents: Diagonal(alpha)*A + Diagonal(beta)*B + Diagonal(alpha)*C + Diagonal(beta)*D
  )

  # Calculate expected result
  D_alpha <- Matrix::Diagonal(n, alpha)
  D_beta <- Matrix::Diagonal(n, beta)
  expected_K <- (D_alpha %*% A) + (D_beta %*% B) + (D_alpha %*% C) + (D_beta %*% D)

  expect_equal(as.matrix(model$K), as.matrix(expected_K))
})

test_that("generic_ns with spatially-varying kappa works", {
  # Create a simple 1D mesh
  n <- 10
  mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = n))

  # Create basis for spatially-varying kappa
  B_kappa <- matrix(0, n, 2)
  B_kappa[1:(n / 2), 1] <- 1 # First half of the domain
  B_kappa[(n / 2 + 1):n, 2] <- 1 # Second half of the domain

  # Create standard Matern components
  matern_model <- matern(mesh)
  C <- matern_model$C
  G <- matern_model$G

  # Create model with space-varying kappa
  ns_model <- generic_ns(
    theta_K = list(kappa = c(log(1), log(2))), # log(kappa) values
    matrices = list(C, G), # Include both matrices C and G
    B_theta_K = list(kappa = B_kappa),
    trans = list(kappa = "exp2"), # kappa^2 transformation for C
    h = matern_model$h,
    position = list(c(1, 2), c(3)), # D_kappa * C + G
    mesh = mesh
  )

  # Create the expected K matrix manually
  kappa_vec <- exp(2 * (B_kappa %*% c(log(1), log(2))))
  D_kappa <- Matrix::Diagonal(x = kappa_vec)
  expected_K <- D_kappa %*% C + G

  expect_equal(as.matrix(ns_model$K), as.matrix(expected_K))
})

test_that("generic_ns handles edge cases correctly", {
  n <- 5
  A <- matrix(1, n, n)
  B <- matrix(2, n, n)

  # Single matrix, no parameters
  model1 <- generic_ns(
    theta_K = list(),
    matrices = list(A),
    position = list(c(1)),
    h = rep(1, n)
  )
  expect_equal(as.matrix(model1$K), A)

  # Multiple parameters with same transformation
  model3 <- generic_ns(
    theta_K = list(alpha = c(1), beta = c(2)),
    matrices = list(A, B),
    h = rep(1, n),
    trans = list(alpha = "exp", beta = "exp"),
    position = list(c(1, 2), c(3, 2, 1), c(4))
  )
  D_alpha <- Matrix::Diagonal(n, exp(1))
  D_beta <- Matrix::Diagonal(n, exp(2))
  expected_K <- D_alpha %*% D_beta + A %*% D_beta %*% D_alpha + B
  expect_equal(as.matrix(model3$K), as.matrix(expected_K))
})

test_that("generic_ns with complex matrix combinations works", {
  n <- 5
  A <- matrix(rnorm(n^2), n, n)
  B <- matrix(rnorm(n^2), n, n)
  C <- matrix(rnorm(n^2), n, n)
  D <- matrix(rnorm(n^2), n, n)

  # Create a complex model with nested matrix operations
  model <- generic_ns(
    theta_K = list(alpha = c(1), beta = c(2)),
    matrices = list(A, B, C, D),
    h = rep(1, n),
    trans = list(alpha = "identity", beta = "exp"),
    position = list(c(1, 2), c(3, 4), c(1, 5), c(2, 6))

    # This represents: D_alpha * D_beta + A * B + D_alpha * C + D_beta * D
  )
  D_alpha <- Matrix::Diagonal(n, 1)
  D_beta <- Matrix::Diagonal(n, exp(2))
  expected_K <- D_alpha %*% D_beta + A %*% B + D_alpha %*% C + D_beta %*% D

  expect_equal(as.matrix(model$K), as.matrix(expected_K))
})


test_that("generic_ns model == AR1 model", {
  # Test AR1 representation
  seed <- 10
  n_obs <- 5
  Y <- rnorm(n_obs)
  ar1 <- ar1(1:n_obs, rho = 0.5)
  g <- name2fun("tanh", inv = TRUE)
  ar1$param_name
  ar1$param_trans

  generic_ar1 <- generic_ns(
    theta_K = list(x = c(g(0.5))), # trans(X) = rho
    trans = list(x = "tanh"),
    matrices = list(ar1$C, ar1$G),
    position = list(c(1, 2), c(3)),
    h = ar1$h,
    mesh = 1:n_obs
  )
  generic_ar1
  generic_ar1$param_name
  generic_ar1$param_trans

  expect_equal(ar1$K, ar1$C * 0.5 + ar1$G)
  expect_equal(generic_ar1$K, ar1$K)
  expect_equal(generic_ar1$matrices, list(ar1$C, ar1$G))

  control <- control_opt(
    seed = seed,
    iterations = 100,
    n_parallel_chain = 4,
    n_batch = 1
  )

  fit_ar1 <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      name = "my_ar",
      model = ar1(rho = 0.5)
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_ar1

  est_rho_ar1 <- ar1_th2a(ngme_result(fit_ar1, "my_ar")$rho)
  print(est_rho_ar1)

  print(generic_ar1$K)
  generic_ar1$param_map
  generic_ar1$position
  fit_generic <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      name = "generic",
      model = generic_ns(
        theta_K = list(rho = g(0.5)),
        trans = list(rho = "tanh"),
        matrices = list(ar1$C, ar1$G),
        h = ar1$h,
        position = list(c(1, 2), c(3))
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_generic

  est_rho_generic <- ar1_th2a(ngme_result(fit_generic, "generic")$rho)
  print(est_rho_generic)

  expect_equal(est_rho_generic[[1]][1], est_rho_ar1[[1]][1])
})


test_that("generic model == Matern model (alpha == 2 or 4)", {
  x <- seq(0, 1, length.out = 10)
  y <- seq(0, 1, length.out = 10)
  mesh <- fmesher::fm_mesh_2d(cbind(x, y))
  mesh$n

  B_kappa <- matrix(rnorm(mesh$n * 2), mesh$n, 2)
  theta_kappa <- c(0.5, 0.3)
  matern <- matern(
    mesh,
    alpha = 4,
    theta_kappa = theta_kappa,
    B_kappa = B_kappa
  )
  kappas <- as.numeric(exp(B_kappa %*% theta_kappa))
  D_kappa <- Matrix::Diagonal(x = kappas)
  C <- matern$C
  G <- matern$G
  Cinv <- C
  diag(Cinv) <- 1 / Matrix::diag(C)

  # Complicated K matrix
  #  D_\kappa C D_\kappa C^{-1} D_\kappa C D_\kappa + D_\kappa C D_\kappa C^{-1}  G + G C^{-1}  D_\kappa C D_\kappa  + G C^{-1} G
  full_K <- D_kappa %*% C %*% D_kappa %*% Cinv %*% D_kappa %*% C %*% D_kappa +
    D_kappa %*% C %*% D_kappa %*% Cinv %*% G +
    G %*% Cinv %*% D_kappa %*% C %*% D_kappa +
    G %*% Cinv %*% G
  expect_equal(matern$K, full_K)

  # matrices list : D_kappa, C, G, Cinv
  generic_model_matern <- generic_ns(
    theta_K = list(theta = theta_kappa),
    trans = list(theta = c("exp")),
    B_theta_K = list(theta = B_kappa),
    matrices = list(C, G, Cinv),
    position = list(
      c(1, 2, 1, 4, 1, 2, 1),
      c(1, 2, 1, 4, 3),
      c(3, 4, 1, 2, 1),
      c(3, 4, 3)
    ),
    h = matern$h
  )
  expect_equal(generic_model_matern$symmetric, matern$symmetric)
  expect_equal(generic_model_matern$zero_trace, matern$zero_trace)
  expect_equal(generic_model_matern$h, matern$h)
  expect_equal(generic_model_matern$K, matern$K, tolerance = 1e-4)

  # Fitting the matern model
  control <- control_opt(
    seed = 10,
    iterations = 50,
    n_parallel_chain = 4,
    n_batch = 1,
  )

  Y <- rnorm(10)
  fit_matern_2 <- ngme(
    Y ~ 0 + f(
      cbind(x, y),
      model = matern(
        mesh = mesh,
        theta_kappa = theta_kappa,
        B_kappa = B_kappa,
        alpha = 2
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_matern_2
  est_theta_matern_2 <- ngme_result(fit_matern_2, "field1")$kappa
  est_theta_matern_2[[1]]

  fit_generic_2 <- ngme(
    Y ~ 0 + f(
      cbind(x, y),
      name = "generic",
      model = generic_ns(
        mesh = mesh,
        theta_K = list(theta = theta_kappa),
        trans = list(theta = c("exp")),
        B_theta_K = list(theta = B_kappa),
        matrices = list(C, G),
        position = list(
          c(1, 2, 1),
          c(3)
        ),
        h = matern$h
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_generic_2
  est_theta_generic_2 <- ngme_result(fit_generic_2, "generic")$theta
  expect_equal(est_theta_generic_2[[1]], est_theta_matern_2[[1]], tolerance = 1e-4)

  fit_matern_4 <- ngme(
    Y ~ 0 + f(
      cbind(x, y),
      model = matern(
        mesh = mesh,
        theta_kappa = theta_kappa,
        B_kappa = B_kappa,
        alpha = 4
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_matern_4
  est_theta_matern_4 <- ngme_result(fit_matern_4, "field1")$operator$theta_K
  est_theta_matern_4[[1]]

  fit_generic_alpha_4 <- ngme(
    Y ~ 0 + f(
      cbind(x, y),
      name = "generic",
      model = generic_ns(
        mesh = mesh,
        theta_K = list(theta = theta_kappa),
        trans = list(theta = c("exp")),
        B_theta_K = list(theta = B_kappa),
        matrices = list(C, G, Cinv),
        position = list(
          c(1, 2, 1, 4, 1, 2, 1),
          c(1, 2, 1, 4, 3),
          c(3, 4, 1, 2, 1),
          c(3, 4, 3)
        ),
        h = matern$h
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_generic_alpha_4
  est_theta_generic_alpha_4 <- ngme_result(fit_generic_alpha_4, "generic")$operator$theta_K
  expect_equal(est_theta_generic_alpha_4[[1]], est_theta_matern_4[[1]], tolerance = 1e-4)
})

test_that("ou (generic) equals rho*C + G on uniform mesh", {
  mesh <- 1:8
  rho <- 0.3
  theta <- -log(rho)

  op <- ou(mesh = mesh, theta = theta)

  C <- Matrix::sparseMatrix(j = 1:(length(mesh) - 1), i = 2:length(mesh), x = -1, dims = c(length(mesh), length(mesh)))
  G <- Matrix::Diagonal(length(mesh))
  G[1, 1] <- sqrt(1 - rho^2)
  expected <- rho * C + G

  expect_equal(as.matrix(op$K), as.matrix(expected))
})

# #' Build random effect structure
# #' @param formula formula
# #' @param data data
# #' @param family distribution of random effect
# #' @return a list of random effect structure
# #' @export
# randeff <- function(
#   map,
#   data,
#   effect_type,
#   name,
#   # replicate,
#   theta_K=NULL,
#   n_repl=NULL, # number of replicates
#   ...
# ) {
#   stopifnot("Please specify random effect type, using effect_type=...,
#     check ngme_randeff_types()" = inherits(effect_type, "ngme_noise") || effect_type %in% ngme_randeff_types())

#   # Design Matrix and Covariance matrix
#   B_reff <- map
#   n_reff <- ncol(B_reff)
#   n_cov_params <- sum(1:n_reff)
#   if (is.null(theta_K)) theta_K <- rep(0, n_cov_params)
#   stopifnot(length(theta_K) == n_cov_params)

#   if (is.character(effect_type)) {
#     mix_var <- if (effect_type=="normal") noise_normal() else noise_nig()
#   } else {
#     mix_var <- effect_type
#   }
#   mix_var <- update_noise(mix_var, n=n_reff)
#   mix_var$single_V = TRUE
#   mix_var$fix_theta_sigma = TRUE
#   mix_var$hessian = FALSE
#   # mix_var$B_mu <- matrix(rep(1, length(B_reff), ncol=n_reff))
#   # mix_var$theta_mu <- rep(0, n_reff)

#   args <- within(list(...), {
#     model        = "re"
#     map          = B_reff
#     A            = ngme_as_sparse(B_reff)
#     n_map        = length_map(map)
#     noise        = mix_var
#     W_size       = n_reff
#     V_size       = n_reff
#     theta_K      = theta_K
#     n_cov_params = n_cov_params
#     n_repl       = n_repl
#     # n_params     = n_cov_params + 1 + 1 #n_reff + 1 #sigma mu nu
#     name         = name
#   })
#   do.call(ngme_model, args)
# }

# #' Print ngme random effect
# #'
# #' @param x random effect object
# #' @param padding number of white space padding in front
# #' @param ... ...
# #'
# #' @return radom effect model
# #' @export
# print.randeff <- function(x, padding = 0, ...) {
#   reff <- x
#   pad_space <- paste(rep(" ", padding), collapse = "")
#   pad_add4_space <- paste(rep(" ", padding + 4), collapse = "")

#   if (is.null(reff)) {
#     cat(pad_space); cat("Effect type - "); cat("NULL"); cat("\n")
#   } else {
#     cat(pad_space); cat("Effect type - "); cat(reff$effect_type); cat("\n")

#     cat(pad_space); cat("Effect formula: ");
#     cat(paste(as.character(reff$formula))); cat("\n")

#     cat(pad_space); cat("Effect covariance: \n")
#     cat(pad_space); print(reff$Sigma)
#   }
#   cat("\n")
#   invisible(reff)
# }

#' Create Time Series Cross-Validation Indices
#'
#' @description Creates indices for time series cross-validation with options for expanding window
#' or sliding window approaches. Supports both single-step and multi-step forecasting, and can
#' handle replicated observations.
#'
#' @param time_idx A numeric vector of time indices in ascending order
#' @param train_length An integer specifying the fixed length of training sets.
#'        If NULL (default), an expanding window approach is used.
#' @param test_length An integer specifying the number of observations to include in each test set.
#'        Default is 1 (single-step forecasting).
#' @param replicate An optional vector of the same length as time_idx, indicating which observations
#'        belong to the same replicate group. When provided, ensures that all observations with the
#'        same replicate value are either entirely in the training set or entirely in the test set.
#' @param gap An integer specifying the gap between the training set and test set.
#'        Default is 0 (no gap, test set starts immediately after training set).
#'        For example, gap=1 means skip one time point between training and test (useful for 2-step ahead forecasting).
#'
#' @return A list with two components:
#' \describe{
#'   \item{train}{A list of numeric vectors, where each vector contains the time indices for training in that fold}
#'   \item{test}{A list of numeric vectors, where each vector contains the time indices for testing in that fold}
#' }
#'
#' @details
#' Time series cross-validation requires respecting the temporal order of observations.
#' This function implements two common approaches:
#'
#' 1. Expanding window (when train_length = NULL): The training set grows with each fold,
#'    starting with a minimal set and expanding to include all but the test data.
#'
#' 2. Sliding window (when train_length is specified): Uses a fixed-length window
#'    that slides through the time series, maintaining the same training size across folds.
#'
#' The test_length parameter allows for multi-step forecasting evaluation.
#'
#' When replicate is provided, the function ensures that all observations with the same
#' replicate value are kept together, either all in the training set or all in the test set.
#' This is useful for scenarios where multiple observations at the same time point should
#' be treated as a group.
#'
#' The gap parameter creates a separation between training and test sets, which is useful
#' for multi-step ahead forecasting validation.
#'
#' @examples
#' # Expanding window approach with single-step forecasting
#' cv_expanding <- make_time_series_cv_index(1:10)
#'
#' # Sliding window approach with window size 3 and single-step forecasting
#' cv_sliding <- make_time_series_cv_index(1:10, train_length = 3)
#'
#' # Sliding window with multi-step forecasting (predict 2 steps ahead)
#' cv_multistep <- make_time_series_cv_index(1:10, train_length = 3, test_length = 2)
#'
#' # Working with replicates
#' time_idx <- c(1, 1, 1, 2, 2, 3, 3)
#' replicates <- c(1, 1, 1, 2, 2, 3, 3)
#' cv_with_replicates <- make_time_series_cv_index(time_idx, replicate = replicates)
#'
#' # 2-step ahead forecasting with a gap
#' cv_with_gap <- make_time_series_cv_index(1:10, gap = 1)
#'
#' @export
make_time_series_cv_index <- function(
    time_idx,
    train_length = NULL,
    test_length = 1,
    replicate = time_idx,
    gap = 0) {
  # Validate inputs
  if (!is.numeric(time_idx)) {
    stop("time_idx must be numeric")
  }

  n <- length(time_idx)
  if (n < 2) {
    stop("time_idx must have at least 2 elements")
  }

  # Validate test_length
  if (test_length < 1) {
    stop("test_length must be at least 1")
  }

  # Validate gap
  if (gap < 0) {
    stop("gap must be non-negative")
  }

  # Initialize lists for train and test indices
  train_indices <- list()
  test_indices <- list()

  # Validate and process replicate information if provided
  if (!is.null(replicate)) {
    if (length(replicate) != length(time_idx)) {
      stop("replicate must have the same length as time_idx")
    }
    # Create a data frame to keep track of time points and their replicates
    data <- data.frame(
      idx = 1:length(time_idx),
      time = time_idx,
      replicate = replicate
    )

    # Sort by time and then by replicate to ensure proper ordering
    data <- data[order(data$time, data$replicate), ]

    # Get unique replicates in order of their first appearance
    unique_replicates <- unique(data$replicate)
    replicate_first_times <- tapply(data$time, data$replicate, min)
    replicate_order <- order(replicate_first_times)
    sorted_unique_replicates <- unique_replicates[replicate_order]

    # We'll work with replicate groups instead of individual time points
    time_groups <- sort(unique(replicate_first_times))
    n_groups <- length(time_groups)

    # Time series CV with replicates - we'll work with groups
    if (is.null(train_length)) {
      # Expanding window mode with replicates
      # We'll create folds by using all earlier replicates for training and later ones for testing

      # Calculate the maximum number of folds considering the gap
      max_folds <- n_groups - gap - test_length

      # Ensure we have at least one fold
      if (max_folds < 1) {
        stop("Not enough replicate groups for the specified gap and test_length")
      }

      # For each possible train/test split of replicate groups
      for (i in 1:max_folds) {
        # Identify which replicates are in training set
        train_replicates <- sorted_unique_replicates[1:i]

        # Get all rows where the replicate is in the training set
        train_idx <- data$idx[data$replicate %in% train_replicates]
        train_indices[[i]] <- time_idx[train_idx]

        # Get the test replicates after the gap
        test_replicates <- sorted_unique_replicates[
          (i + 1 + gap):min(i + gap + test_length, n_groups)
        ]

        # Get all rows where the replicate is in the test set
        test_idx <- data$idx[data$replicate %in% test_replicates]
        test_indices[[i]] <- time_idx[test_idx]
      }
    } else {
      # Fixed-length sliding window with replicates

      # Validate train_length - now in terms of number of replicate groups
      if (train_length <= 0 || train_length >= n_groups) {
        stop("train_length must be between 1 and the number of replicate groups - 1")
      }

      # The maximum number of folds we can create, accounting for the gap
      max_folds <- n_groups - train_length - gap - test_length + 1

      # Ensure we have at least one fold
      if (max_folds < 1) {
        stop("Not enough replicate groups for the specified train_length, gap, and test_length")
      }

      # Generate sliding window CV folds based on replicate groups
      for (i in 1:max_folds) {
        # Training set: replicates from position i to i+train_length-1
        train_replicates <- sorted_unique_replicates[i:(i + train_length - 1)]

        # Get all rows where the replicate is in the training set
        train_idx <- data$idx[data$replicate %in% train_replicates]
        train_indices[[i]] <- time_idx[train_idx]

        # Test set: the next test_length replicates after the gap
        test_replicates <- sorted_unique_replicates[
          (i + train_length + gap):(min(i + train_length + gap + test_length - 1, n_groups))
        ]

        # Get all rows where the replicate is in the test set
        test_idx <- data$idx[data$replicate %in% test_replicates]
        test_indices[[i]] <- time_idx[test_idx]
      }
    }
  } else {
    # Standard time series CV without replicates
    # Sort time indices to ensure chronological order if they aren't already
    if (!all(diff(time_idx) >= 0)) {
      time_idx <- sort(unique(time_idx))
      n <- length(time_idx)
    }

    if (is.null(train_length)) {
      # Expanding window mode (default behavior)
      # We'll create as many folds as possible while ensuring enough test points
      # and accounting for the gap
      max_folds <- n - gap - test_length

      if (max_folds < 1) {
        stop("Not enough data points for the specified gap and test_length")
      }

      for (i in 1:max_folds) {
        # Training set: indices from 1 to i
        train_indices[[i]] <- time_idx[1:i]

        # Test set: observations after training and gap
        test_start <- i + 1 + gap
        test_end <- min(test_start + test_length - 1, n)
        test_indices[[i]] <- time_idx[test_start:test_end]
      }
    } else {
      # Fixed-length sliding window mode

      # Validate train_length
      if (train_length <= 0 || train_length >= n) {
        stop("train_length must be between 1 and length(time_idx)-1")
      }

      # The maximum number of folds we can create, accounting for the gap
      max_folds <- n - train_length - gap - test_length + 1

      # Ensure we have at least one fold
      if (max_folds < 1) {
        stop("Not enough data points for the specified train_length, gap, and test_length")
      }

      # Generate sliding window CV folds
      for (i in 1:max_folds) {
        # For sliding window: start with indices i:(i+train_length-1)
        start_idx <- i
        end_idx <- i + train_length - 1

        # Training set: window of length train_length
        train_indices[[i]] <- time_idx[start_idx:end_idx]

        # Test set: observations after training and gap
        test_start <- end_idx + 1 + gap
        test_end <- min(test_start + test_length - 1, n)
        test_indices[[i]] <- time_idx[test_start:test_end]
      }
    }
  }

  # Return the result as a list containing train and test indices
  return(list(
    train = train_indices,
    test = test_indices
  ))
}



#' Convert sparse matrix into sparse dgCMatrix
#' @param G matrix
#'
#' @return sparse dgCMatrix
#' @export
ngme_as_sparse <- function(G) {
  tryCatch(
    expr = {
      G <- as(as(G, "CsparseMatrix"), "generalMatrix")
      # G <- as(as(G, "dMatrix"), "generalMatrix")
    },
    error = function(e) {
      G <- as(G, "dgTMatrix")
      idx <- which(G@i <= G@j)
      G <- Matrix::sparseMatrix(
        i = G@i[idx], j = G@j[idx], x = G@x[idx],
        symmetric = FALSE, index1 = FALSE
      )
      G <- as(as(G, "CsparseMatrix"), "generalMatrix")
      # G <- as(as(G, "dMatrix"), "generalMatrix")
    },
    finally = {
      G
    }
  )
}


#' @name get_inla_mesh_dimension
#' @title Get the dimension of an INLA mesh
#' @description Get the dimension of an INLA mesh
#' @param inla_mesh An INLA mesh
#' @return The dimension of an INLA mesh.
#' @noRd
#'
get_inla_mesh_dimension <- function(inla_mesh) {
  cond1 <- inherits(inla_mesh, "inla.mesh.1d")
  cond2 <- inherits(inla_mesh, "inla.mesh")
  stopifnot(cond1 || cond2)
  if (inla_mesh$manifold == "R1") {
    d <- 1
  } else if (inla_mesh$manifold %in% c("R2", "S2")) {
    d <- 2
  } else {
    stop("The mesh should be from a flat manifold.")
  }
  return(d)
}

# format output
ngme_format <- function(param, val, model = NULL, ...) {
  stationary <- (length(val) == 1)
  dne <- (length(val) == 0)
  dots <- list(...)
  nu_lower_bound <- if (!is.null(dots$nu_lower_bound)) dots$nu_lower_bound else 0

  if (is.null(model)) { # noise
    if (stationary) {
      val <- if (grepl("sigma", param, fixed = TRUE)) {
        format(exp(val), digits = 3)
      } else if (grepl("nu", param, fixed = TRUE)) {
        format(nu_lower_bound + exp(val), digits = 3)
      } else {
        format(val, digits = 3)
      }
    } else {
      val <- paste0(format(val, digits = 3), collapse = ", ")
    }

    switch(param,
      "sigma" = if (stationary) {
        paste0("sigma = ", val)
      } else {
        paste0("theta_sigma = ", val)
      },
      "sigma_nig" = if (stationary) {
        paste0("sigma_nig = ", val)
      } else {
        paste0("theta_sigma_nig = ", val)
      },
      "sigma_normal" = if (stationary) {
        paste0("sigma_normal = ", val)
      } else {
        paste0("theta_sigma_normal = ", val)
      },
      "mu" = if (stationary) {
        paste0("mu = ", val)
      } else {
        paste0("theta_mu = ", val)
      },
      "nu" = if (stationary) {
        paste0("nu = ", val)
      } else {
        paste0("theta_nu = ", val)
      },
      "feff" = if (dne) "No fixed effects" else paste0("feff = ", val)
    )
  } else { # model
    switch(model,
      "ar1" = paste0("alpha = ", format(ar1_th2a(val), digits = 3)),
      "matern" = paste0("theta_kappa = ", paste0(format(val, digits = 3), collapse = ", ")),
      "ou" = paste0("theta_K = ", paste0(format(val, digits = 3), collapse = ", ")),
      "re" = {
        paste(capture.output(vecK_to_Sigma(val, list(...)[[1]])), collapse = "\n")
      }
    )
  }
}

#' taking mean over a list of nested lists
#'
#' @param lls a list
#' @param weights weights of each list
#'
#' @return a list of nested lists
#' @export
#'
#' @examples
#' ls <- list(
#'   list(a = 1, b = 2, t = "nig", ll = list(a = 1, b = 2, w = "ab")),
#'   list(a = 3, b = 5, t = "nig", ll = list(a = 1, b = 6, w = "ab")),
#'   list(a = 5, b = 5, t = "nig", ll = list(a = 4, b = 2, w = "ab"))
#' )
#' mean_list(ls)
mean_list <- function(lls, weights = NULL) {
  n <- length(lls)
  weights <- if (is.null(weights)) {
    rep(1 / n, n)
  } else {
    weights / sum(weights)
  }

  # helpers
  nest_list_add <- function(l1, l2) {
    for (i in seq_along(l2)) {
      if (is.numeric(l2[[i]])) {
        l1[[i]] <- l1[[i]] + l2[[i]]
      }
      if (is.list(l2[[i]])) {
        l1[[i]] <- nest_list_add(l1[[i]], l2[[i]])
      }
    }
    l1
  }
  nest_list_mult <- function(l, n) {
    for (i in seq_along(l)) {
      if (is.numeric(l[[i]])) l[[i]] <- l[[i]] * n
      if (is.list(l[[i]])) l[[i]] <- nest_list_mult(l[[i]], n)
    }
    l
  }

  ret <- nest_list_mult(lls[[1]], 0)
  for (i in seq_along(lls)) {
    tmp <- nest_list_mult(lls[[i]], weights[[i]])
    ret <- nest_list_add(ret, tmp)
  }
  ret
}

# helper functions
# ar1 alpha (-1~1) to theta_K
ar1_a2th <- function(a) {
  log((-1 - a) / (-1 + a))
}

# theta_K to ar1 alpha (0~1)
ar1_th2a <- function(th) {
  -1 + (2 * exp(th)) / (1 + exp(th))
}


# #' Make index for the matern model
# #'
# #' @param name
# #' @param n.spde
# #' @param n.repl
# #' @param mesh
# #' @param dim
# #'
# #' @return
# #' @export
# #'
# #' @examples
# ngme.matern.make.index <- function(
#   n.spde=NULL,
#   n.repl = 1,
#   mesh = NULL,
#   dim = NULL
# ){
#   if(is.null(n.spde)&&is.null(mesh)){
#     stop("You should provide either n.spde or mesh!")
#   }

#   if(!is.null(mesh)){
#     n_mesh = mesh$n

#     if(mesh$manifold == "R1"){
#       dim = 1
#     } else if(mesh$manifold == "R2"){
#       dim = 2
#     } else{
#       stop("The domain must be flat manifolds of dimension 1 or 2, that is,
#          the domain must be a line or a plane.")
#     }

#   } else{
#     n_mesh <- n.spde
#     if(is.null(dim)) {
#       stop("You should provide the dimension d!")
#     }
#   }

#   out <- list()
#   out$index <- rep(1:n_mesh, times = n.repl)

#   out$replicate <- rep(1:n.repl, each = n_mesh)
#   return(out)
# }

#' Make observation matrix for time series
#'
#' @param loc   integers (after sorting, no gaps > 1)
#' @param replicate indicating replicate measure at same location
#' @param range range for the mesh
#'  by default range=(min(loc), max(loc))
#'
#' @return A matrix (length(loc) * length(unique(loc)))
#' @export
#'
#' @examples
#' ngme_ts_make_A(c(1, 2, 2), replicate = c(1, 1, 2))
#' ngme_ts_make_A(c(1, 2, 2), range = c(1, 5))
ngme_ts_make_A <- function(
    loc,
    replicate = NULL,
    range = c(min(loc), max(loc))) {
  if (is.null(loc) || length(loc) == 0) {
    return(NULL)
  }

  n_loc <- length(loc)
  nrep <- 1

  start <- range[1]
  end <- range[2]
  n_range <- end - start + 1

  if (is.null(replicate)) {
    replicate <- rep(1, n_range)
  }

  unique_rep <- unique(replicate)
  nrep <- length(unique_rep)

  A <- matrix(0, nrow = n_loc, ncol = n_range * nrep)

  for (i in 1:n_loc) {
    ncol_rep <- which(unique_rep == replicate[i])
    A[i, (ncol_rep - 1) * n_range + loc[i] - start + 1] <- 1
  }
  # as(A, "dgCMatrix")
  as(as(A, "dMatrix"), "generalMatrix")
}

# compute the mode of data
emprical_mode <- function(x, breaks = max(20, length(x) / 20)) {
  h <- hist(x, breaks = breaks, plot = FALSE)
  idx <- which.max(h$counts)
  h$mids[idx]
}

# build <- function(A1, A2) {
# }

# t, (x, y)

# f(mesh = list(mesh1, mesh2), replicate=112233) + f(ar1 )


# given a list of replicate
# merge_repls <- function(repls) {
#   # merge the list of data frames
#   # input: list of numerics
#   # output: merged list
#   # assert of equal length
#   stopifnot(length(unique(as.numeric(lapply(repls, length)))) == 1)
#   # helper function of merge 2 repls
#   merge_repl <- function(repl, group) {
#     halas <- FALSE
#     while (!halas) {
#       unique_group <- unique(group)
#       halas <- TRUE
#       if (length(unique_group) == 1) return (group)
#       for (i in 1:(length(unique_group)-1)) {
#         for (j in (i+1):length(unique_group)) {
#           if (length(intersect(repl[group == unique_group[i]], repl[group == unique_group[j]])) > 0) {
#             group[group == unique_group[j]] <- unique_group[i]
#             halas <- FALSE
#           }
#           if (!halas) break
#         }
#         if (!halas) break
#       }
#     }
#     group
#   }

#   Reduce(function(x, y) merge_repl(x, y), repls)
# }

split_matrix <- function(mat, repl) {
  split_mat <- lapply(split(mat, repl, drop = FALSE),
    matrix,
    ncol = ncol(mat)
  )
  split_mat
}

# helper function to unfiy way of accessing 1d and 2d index
as_map <- function(locs) {
  if (inherits(locs, c("data.frame", "matrix"))) {
    as.matrix(locs)
  } else {
    as.numeric(locs)
  }
}

sub_map <- function(locs, idx) {
  if (inherits(locs, c("data.frame", "matrix"))) {
    locs[idx, , drop = FALSE]
  } else {
    locs[idx]
  }
}

dim_map <- function(map) {
  if (inherits(map, c("data.frame", "matrix"))) ncol(map) else 1
}

rep_map <- function(map, times) {
  if (inherits(map, c("data.frame", "matrix"))) {
    do.call(rbind, replicate(times, map, simplify = FALSE))
  } else {
    rep(map, times = times)
  }
}

length_map <- function(map) {
  if (inherits(map, c("data.frame", "matrix"))) {
    nrow(map)
  } else {
    length(map)
  }
}

# vectorize a matrix
veci <- function(v, n, m) {
  if (length(v) != n * m) {
    stop(
      "Wrong dimensions in reshape: ", length(v),
      " (", n, ", ", m, ")",
      call. = FALSE
    )
  }
  M <- matrix(0, nrow = n, ncol = m)
  count <- 1
  for (i in 1:m) {
    M[, i] <- v[count:(count + n - 1)]
    count <- count + n
  }
  return(M)
}

# M <- matrix(c(1,2,3,2,4,5,3,5,6), 3, 3); M
vech <- function(M) {
  stopifnot(nrow(M) == ncol(M))
  n <- nrow(M)
  V <- double(n * (n - 1) / 2 + n)
  V[1:n] <- diag(M)

  k <- n + 1
  for (i in seq_len(n - 1)) {
    V[k:(k + n - i - 1)] <- tail(M[, i], n - i)
    k <- k + n - i
  }
  return(V)
}

vech_to_mat <- function(vech, n) {
  stopifnot(n * (n - 1) / 2 + n == length(vech))
  mat <- matrix(0, nrow = n, ncol = n)
  mat[lower.tri(mat)] <- tail(vech, length(vech) - n)
  mat <- mat + t(mat)
  diag(mat) <- vech[1:n]
  mat
}

vecK_to_Sigma <- function(theta_K, n) {
  tmp <- theta_K
  tmp[1:n] <- exp(theta_K[1:n])
  K <- vech_to_mat(tmp, n)
  Kinv <- solve(K)
  Sigma <- Kinv %*% t(Kinv)
  Sigma
}

build_D <- function(theta, rho) {
  d11 <- cos(theta) + rho * sin(theta)
  d12 <- -sin(theta) * (sqrt(1 + rho^2))
  d21 <- sin(theta) - rho * cos(theta)
  d22 <- cos(theta) * (sqrt(1 + rho^2))
  matrix(c(d11, d21, d12, d22), nrow = 2, ncol = 2)
}

build_effect_K <- function(n_reff, theta_K) {
  n_theta_K <- sum(1:n_reff)
  K <- diag(n_reff)
  diag(K) <- exp(theta_K[1:n_reff])
  if (n_reff > 1) {
    K[lower.tri(K)] <- theta_K[(n_reff + 1):n_theta_K]
  }
  K
}


# given a basis matrix, find if it is stationary
is_stationary <- function(B) {
  stopifnot(is.matrix(B))
  ncol(B) == 1 && all(B == 1)
}


#' @title ngme make mesh for different replicates
#' @description
#' Make different mesh for different replicates
#'
#' @param data provide the data.frame
#' @param map provide the map to make mesh, i.g. ~x+y, x and y will be extracted from data to make a 2d mesh
#' @param replicate provide the replicate information, i.g. ~id
#' @param mesh_type type of mesh, "regular" means use all the point from same replicate to make a mesh
#'
#' @return a list of mesh of length of different replicates
#' @export
ngme_make_mesh_repls <- function(
    data,
    map,
    replicate,
    mesh_type = "regular") {
  if (inherits(map, "formula")) {
    map <- model.matrix(map, data)[, -1]
  }

  if (inherits(replicate, "formula")) {
    replicate <- model.matrix(replicate, data)[, -1]
  }

  stopifnot(length_map(map) == length(replicate))

  mesh_repls <- NULL
  for (repl in replicate) {
    map_repl <- subset(map, replicate == repl)
    if (dim_map(map) == 1) {
      mesh_repls[[as.character(repl)]] <- tryCatch(
        fmesher::fm_mesh_1d(map_repl),
        error = function(e) {
          stop("The nodes for making mesh is not valid for replicate id=", repl)
        }
      )
    } else if (dim_map(map) == 2) {
      stop("Not implemented yet.")
      mesh_repls[[repl]] <- fmesher::fm_mesh_2d(map_repl)
    } else {
      stop("The dimension of the mesh should be 1 or 2.")
    }
  }

  mesh_repls
}


# check if group is valid
# return as integer
validate_rep_or_group <- function(replicate, data) {
  if (is.null(data)) {
    return(as.factor(replicate))
  }

  if (is.null(replicate)) {
    replicate <- rep(1, nrow(data))
  }

  if (inherits(replicate, "formula")) {
    # input as: replicate = ~id
    stopifnot(
      "Allow 1 variable (column in data) as replicate. i.g. replicate=~id" = length(replicate) == 2 && length(replicate[[2]]) == 1
    )

    replicate <- eval(replicate[[2]], envir = data, enclos = parent.frame())
  }

  if (inherits(replicate, "character") && (length(replicate) == 1)) {
    # input as: replicate = "id"
    replicate <- data[[replicate]]
  }

  stopifnot(
    "Please make sure the length of replicate/group is equal to the number of rows of data" = nrow(data) == length(replicate)
  )

  return(as.factor(replicate))
}

# #' @title Use more interpretable parameterization of the matern model
# #' @description
# #' From SPDE parameter (kappa, sigma, alpha) to (theta_kappa, theta_sigma, theta_alpha)
# #'
# #' @param ope provide the operator
# #'
# #' @return a list of mesh of length of different replicates
# #' @export
# matern_result <- function(ope) {
#   stop("to-do")
# }


# return the relative idx among all latent models
# using W_size assume W_size == V_size
idx_range <- function(ngme_rep, name_or_idx) {
  if (is.character(name_or_idx)) {
    idx <- which(sapply(ngme_rep$models, function(model) model$name) == name_or_idx)
  } else {
    idx <- name_or_idx
  }

  models <- ngme_rep$models
  tmp <- sapply(models, function(model) model$W_size)
  sizes <- Reduce(`+`, tmp, accumulate = TRUE)

  if (idx == 1) {
    return(1:sizes[1])
  } else {
    return((sizes[idx - 1] + 1):sizes[idx])
  }
}

#' @title posterior samples of different latent models
#' @description
#' Extract the posterior samples of different latent models
#'
#' @param ngme_object ngme object
#' @param model_name name of the model, or index of the model
#' @param type type of samples, "W" or "V"
#' @param replicate which replicate
#'
#' @return a data.frame of posterior samples (mesh_size * n_post_samples)
#' @export
ngme_post_samples <- function(
    ngme_object,
    model_name = 1,
    type = "W",
    replicate = 1) {
  ngme_rep <- ngme_object$replicates[[replicate]]
  stopifnot(
    inherits(ngme_rep, "ngme_replicate"),
    type %in% c("W", "V")
  )
  idx <- idx_range(ngme_rep, model_name)

  if (type == "W") {
    return(ngme_rep$post_W[idx, , ])
  } else {
    return(ngme_rep$post_V[idx, , ])
  }
}

#' @title variance of the data or the latent field
#' @description
#' Compute the variance of the data or the latent field
#'
#' @param ngme_object ngme_model
#' @param model_name
#'   if model_name = "data", then return the covariance matrix of the data (without measurement noise)
#'   if the model_name is the name or index of the latent, then return the covariance matrix of the latent field
#' @param replicate which replicate (default = 1)
#'
#' @return a data.frame of posterior samples (mesh_size * n_post_samples)
#' @export
ngme_cov_matrix <- function(
    ngme_object,
    model_name = "data",
    replicate = 1) {
  ngme_rep <- ngme_object$replicates[[replicate]]

  stopifnot(
    "Please provide the correct ngme object." = inherits(ngme_object, "ngme"),
    "Please provide the correct model name or index." =
      model_name %in% c(
        "data",
        sapply(ngme_rep$models, function(model) model$name),
        1:length(ngme_rep$models)
      )
  )

  if (model_name == "data") {
    V_mean <- apply(ngme_rep$post_V, 1, mean)

    diag_K <- Matrix::bdiag(
      sapply(ngme_rep$models, function(model) model$operator$K)
    )

    Q <- diag_K %*% diag(1 / V_mean) %*% Matrix::t(diag_K)
    block_A <- Reduce(cbind, sapply(ngme_rep$models, function(model) model$A))

    # return A Q^(-1) A^t,
    return(block_A %*% Matrix::solve(Q, Matrix::t(block_A)))
  } else {
    K <- ngme_rep$models[[model_name]]$operator$K

    idx <- idx_range(ngme_rep, model_name)
    V_mean <- apply(ngme_rep$post_V[idx, , ], 1, mean)

    Q <- K %*% diag(1 / V_mean) %*% Matrix::t(K)
    return(solve(Q))
  }
}


ngme_build_A <- function(model, mesh, map, operator, group, group_levels = NULL) {
  group <- validate_rep_or_group(group, NULL)

  if (is.null(group_levels)) group_levels <- levels(group)

  if (inherits(mesh, "metric_graph")) {
    A <- if (is.null(map)) NULL else mesh$fem_basis(map)
    return(A)
  }

  if (model %in% c("tp", "spacetime")) {
    mesh_t <- if (model == "tp") operator$first$mesh else operator$mesh[[1]]
    mesh_s <- if (model == "tp") operator$second$mesh else operator$mesh[[2]]
    stopifnot(
      "Now only support first to be 1d model" = inherits(mesh_t, "inla.mesh.1d")
    )

    # watch-out! Should use as.factor to start at 1, not min(map[[1]])
    # blk_group <- as.integer(as.factor(map[[1]]))

    blk_group <- as.integer(map[[1]])
    # important to start with 1
    min_loc <- mesh_t$interval[[1]]
    blk_group <- blk_group - min_loc + 1
    stopifnot("The map[[1]] of the tp model is not valid, out of range." = all(blk_group <= (mesh_t$range[2] + 1)))

    blk <- fmesher::fm_block(blk_group, n_block = mesh_t$n)
    basis <- fmesher::fm_basis(mesh_s, loc = map[[2]])
    A0 <- fmesher::fm_row_kron(Matrix::t(blk), basis)

    if (model == "tp" && operator$second$model %in% c("bv", "bv_matern", "bv2")) {
      # for tp-bv model
      # check if group is valid
      if (length(group) == 0) stop("Please provide the `group` argument.")
      all(group %in% group_levels) || stop("The group is not valid.")
      length(group) == length_map(map[[1]]) || stop("The length of group should be equal to the length of map.")
      # tp bv model
      # 1. expand A <- cbind(A, 0), double the column
      # 2. move 2nd field to the 2nd half
      # 3. re-order the 1st and 2nd
      # e.g., (each field is of size 2)
      # 1 2 3 4 5 6 7 8 to
      # 1 2 5 6 3 4 7 8
      A_expand <- cbind(
        A0,
        matrix(0, nrow = nrow(A0), ncol = ncol(A0))
      )

      # select row (of the 2nd field)
      row_2nd_field <- group == group_levels[[2]]
      half_1st <- 1:ncol(A0)
      half_2nd <- 1:ncol(A0) + ncol(A0)

      # Move the 2nd field to the right
      A_expand[row_2nd_field, half_2nd] <-
        A_expand[row_2nd_field, half_1st]

      if (any(row_2nd_field)) {
        A_expand[row_2nd_field, half_1st] <- 0
      }

      # Re-order to match K = K_t %x% K_bv (time-major, fields within time)
      n_t <- operator$first$mesh$n
      n_s <- operator$second$mesh$n
      n_ts <- n_t * n_s
      field1 <- matrix(seq_len(n_ts), nrow = n_s, ncol = n_t)
      field2 <- field1 + n_ts
      reorder_idx <- as.vector(rbind(field1, field2))

      # return after re-order
      return(ngme_as_sparse(A_expand[, reorder_idx]))
    } else {
      return(ngme_as_sparse(A0))
    }
  }

  # bivariate model (including VAR(1))
  if (model %in% c("bv", "bv2", "bv_matern", "var1")) {
    # check if group is valid
    if (length(group) == 0) stop("Please provide the `group` argument.")
    all(group %in% group_levels) || stop("The group is not valid.")
    length(group) == length_map(map) || stop("The length of group should be equal to the length of map.")

    blk_group <- as.integer(as.factor(group))
    blk <- fmesher::fm_block(blk_group, n_block = length(group_levels))
    basis <- fmesher::fm_basis(mesh, loc = map)
    A <- fmesher::fm_row_kron(Matrix::t(blk), basis)
    return(ngme_as_sparse(A))
  }

  if (model == "re") {
    return(ngme_as_sparse(operator$B_theta_K))
  }

  if (is.factor(map)) map <- as.integer(map) # for iid model
  return(fmesher::fm_basis(mesh, loc = map))
}




contain_bv_model <- function(ngme) {
  any(sapply(ngme$replicates[[1]]$models, function(model) {
    model$model %in% c("bv", "bv2", "bv_matern")
  })) ||
    any(sapply(ngme$replicates[[1]]$models, function(model) {
      model$model %in% c("tp") && model$operator$second$model %in% c("bv", "bv2", "bv_matern")
    }))
}


#' @title get the trajectories of parameters of the model
#' @description
#' Get the trajectories of the parameters of the model
#'
#' @param ngme_object ngme object
#' @param model_name name of the model, if NULL, return the trajectories of the parameters of the measurement noise and fixed effects
#' @return a list of trajectories, each element is a matrix of trajectories (n_samples * n_trajectories)
#' @export
get_trajectories <- function(
    ngme_object,
    model_name) {
  stopifnot(inherits(ngme_object, "ngme"))

  ngme <- ngme_object$replicates[[1]]
  if (!is.null(model_name) && model_name %in% names(ngme$models)) {
    traj <- attr(ngme$models[[model_name]], "lat_traj")
    ts <- get_latent_info(ngme$models[[model_name]])
  } else {
    traj <- attr(ngme, "block_traj")
    ts <- get_noise_info(ngme$noise)
  }
  #  str(traj)
  #  List of 4
  #  $ :'data.frame':       4 obs. of  1000 variables:

  convert_to_3d_array <- function(traj) {
    # Extract the dimensions
    num_replicates <- length(traj) # Number of replicates
    num_variables <- nrow(traj[[1]]) # Number of variables (4)
    num_iterations <- ncol(traj[[1]]) # Number of iterations (1000)

    # Initialize an empty 3D array
    array_3d <- array(NA, dim = c(num_variables, num_iterations, num_replicates))

    # Fill the array
    for (replicate_index in seq_len(num_replicates)) {
      array_3d[, , replicate_index] <- as.matrix(traj[[replicate_index]])
    }

    return(array_3d)
  }
  traj_3d <- convert_to_3d_array(traj)

  ret <- list()
  for (i in seq_along(ts$name)) {
    name <- ts$name[i]
    data <- traj_3d[i, , ]
    ret[[name]] <- ts$trans[[i]](data)
  }

  return(ret)
}




#' @title get the parameters of the noise
#' @description
#' Get the parameters of the noise
#'
#' @param noise noise object
#' @param prefix prefix of the parameter
#' @return a list with parameter name as key and parameter value as value
#' @keywords internal
get_noise_param <- function(noise, prefix) {
  B <- noise[[paste0("B_", prefix)]]
  theta <- noise[[paste0("theta_", prefix)]]
  if (length(theta) == 0) {
    return(NULL)
  }

  if (is_stationary(B)) {
    value <- if (prefix == "mu") theta else exp(theta)
  } else {
    value <- theta
  }

  name <- if (is_stationary(B)) prefix else paste0("theta_", prefix)
  result <- list()
  result[[name]] <- value
  result
}



# extract parameters from all models in ngme object
extract_parameters <- function(ngme_object) {
  stopifnot(inherits(ngme_object, "ngme"))
  ngme_1rep <- ngme_object$replicates[[1]]
  result_transformed <- list() # transformed parameters (user friendly)
  result_raw <- list() # raw parameters (for optimizer)

  # LATENT MODELS
  models <- ngme_1rep$models
  model_names <- names(models)

  for (j in seq_along(models)) {
    names <- models[[j]]$operator$param_name
    trans <- models[[j]]$operator$param_trans
    result_model_transformed <- list()
    result_model_raw <- list()
    for (i in seq_along(names)) {
      result_model_transformed[[names[i]]] <- trans[[i]](models[[j]]$operator$theta_K[[i]])
      if (is.function(trans[[i]]) && trans[[i]](3) == 3) {
        raw_name <- names[i]
      } else {
        raw_name <- paste0("theta_", names[i])
      }
      result_model_raw[[raw_name]] <- models[[j]]$operator$theta_K[[i]]
    }

    # Use get_noise_param for noise parameters
    noise <- models[[j]]$noise
    noise_params_transformed <- list()
    noise_params_raw <- list()
    for (p in c("mu", "sigma", "nu")) {
      noise_param <- get_noise_param(noise, p)

      # special case for rw1 and rw2
      if (models[[j]]$model %in% c("rw1", "rw2") && p == "sigma") {
        noise_param$theta_sigma <- noise_param$theta_sigma[-1]
        if (length(noise_param$theta_sigma) == 1) {
          noise_param <- list(sigma = exp(noise_param$theta_sigma))
        }
      }

      if (!is.null(noise_param)) {
        noise_params_transformed <- c(noise_params_transformed, noise_param)
        noise_params_raw <- c(noise_params_raw, noise_param)
      }
    }

    result_model_transformed <- c(result_model_transformed, noise_params_transformed)
    result_model_raw <- c(result_model_raw, noise_params_raw)

    result_transformed[[model_names[j]]] <- result_model_transformed
    result_raw[[model_names[j]]] <- result_model_raw
  }

  # FIXED EFFECTS and MEASUREMENT NOISE
  beta <- ngme_1rep$feff
  result_transformed$data$beta <- beta
  result_raw$data$beta <- beta

  m_noise <- ngme_1rep$noise
  m_noise_params_transformed <- list()
  m_noise_params_raw <- list()
  for (p in c("mu", "sigma", "nu")) {
    noise_param <- get_noise_param(m_noise, p)
    if (!is.null(noise_param)) {
      m_noise_params_transformed <- c(m_noise_params_transformed, noise_param)
      m_noise_params_raw <- c(m_noise_params_raw, noise_param)
    }
  }

  result_transformed$data <- c(result_transformed$data, m_noise_params_transformed)
  result_raw$data <- c(result_raw$data, m_noise_params_raw)

  list(transformed = result_transformed, raw = result_raw)
}




#' Extracts design matrix from a formula and data.
#'
#' This function takes a formula and a data frame, and returns a design matrix
#' based on the right-hand side of the formula. If the formula is `~.`, it treats
#' all columns in the data as the design matrix. Otherwise, it constructs a
#' design matrix without an intercept based on the specified formula.
#'
#' @param form A formula object, e.g., `~ x1 + x2` or `~.`.
#' @param data A data frame or matrix from which to extract the design matrix.
#' @return A numeric matrix representing the design matrix.
#' @export
#' @examples
#' data(mtcars)
#' # Extract a design matrix for 'mpg' and 'cyl'
#' X1 <- get_data_from_formula(~ mpg + cyl, mtcars)
#' # Extract a design matrix for all columns
#' X2 <- get_data_from_formula(~., mtcars)
#' # Extract a design matrix for 'hp'
#' X3 <- get_data_from_formula(~hp, mtcars)
get_data_from_formula <- function(form, data) {
  form <- as.formula(form)

  if (identical(form, ~.)) {
    return(as.matrix(data))
  }

  rhs <- deparse(form[[2]])
  form_no_intercept <- as.formula(paste("~ 0 +", rhs))

  mf <- model.frame(form_no_intercept, data = data)

  X <- model.matrix(form_no_intercept, data = mf)

  return(X)
}

#' Check whether a newer stable version of ngme2 is available
#'
#' This function checks the package repository for the latest available ngme2
#' version. It does not install or update packages.
#'
#' @return Invisibly returns a list with the local version, remote version,
#'   repository URL, and a logical \code{update_available} flag.
#' @export
ngme_update <- function() {
  local_version <- utils::packageVersion("ngme2")
  repos <- "https://davidbolin.github.io/ngme2/"

  # Check available version in the repository
  available <- tryCatch(
    utils::available.packages(repos = repos),
    error = function(e) NULL
  )

  if (is.null(available) || !"ngme2" %in% rownames(available)) {
    message("Could not check for updates. Please check your internet connection or the repository URL.")
    return(invisible(list(
      local_version = as.character(local_version),
      remote_version = NA_character_,
      repository = repos,
      update_available = NA
    )))
  }

  remote_version <- available["ngme2", "Version"]
  update_available <- utils::compareVersion(as.character(remote_version), as.character(local_version)) > 0

  if (update_available) {
    message(paste0("New stable version available: ", remote_version, " (local: ", local_version, ")"))
    message("Install it manually from repository: ", repos)
  } else {
    message(paste0("ngme2 is up to date (version ", local_version, ")"))
  }

  invisible(list(
    local_version = as.character(local_version),
    remote_version = as.character(remote_version),
    repository = repos,
    update_available = update_available
  ))
}


#' Test OpenMP availability and report the number of threads.
#'
#' This function checks if OpenMP is available in the current R environment
#' and, if so, reports the number of OpenMP threads detected.
#' If OpenMP is not available, it reports a corresponding message.
#' It relies on an internal or external function `get_openmp_threads()`
#' which is assumed to return the number of threads or 0 if unavailable.
#'
#' @return Invisibly returns the detected number of OpenMP threads. A value of
#'   zero means OpenMP is not available.
#' @export
openmp_test <- function() {
  num_threads <- get_openmp_threads()
  if (num_threads == 0) {
    message("OpenMP not available.")
  } else {
    message("OpenMP is available, the default thread number is ", num_threads, ".")
  }
  invisible(num_threads)
}

# -------- ngme operators --------

#' ngme iid model specification
#'
#' @param mesh integer or factor, index to build the mesh
#'
#' @return ngme_operator object
#' @export
#'
iid <- function(mesh = NULL) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, "inla.mesh.1d"))) {
    return(structure(list(model = "iid", args = as.list(environment())), class = "ngme_operator_def"))
  }

  mesh <- ngme_build_mesh(mesh, "iid")
  n <- mesh$n
  K <- ngme_as_sparse(Matrix::Diagonal(n))

  ngme_operator(
    matrices = list(K),
    mesh = mesh,
    model = "iid",
    theta_K = double(0),
    update_K = function(theta_K) {
      K
    },
    K = K,
    h = rep(1, n),
    A = K,
    symmetric = TRUE,
    zero_trace = FALSE,
    param_name = character(0),
    generic_type = "generic"
  )
}

#' ngme AR(1) model specification
#'
#' @param mesh integer vector or inla.mesh.1d object, index to build the mesh
#' @param rho the correlation parameter (between -1 and 1)
#'
#' @return ngme_operator object
#' @export
#'
ar1 <- function(mesh = NULL, rho = 0) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, "inla.mesh.1d"))) {
    return(structure(list(model = "ar1", args = as.list(environment())), class = "ngme_operator_def"))
  }
  stopifnot("rho should be between -1 and 1" = rho >= -1 && rho <= 1)

  mesh <- ngme_build_mesh(mesh)
  n <- mesh$n
  h <- c(diff(mesh$loc), 1)
  G <- Matrix::Diagonal(n)
  G[1, 1] <- sqrt(1 - rho**2)
  C <- Matrix::sparseMatrix(j = 1:(n - 1), i = 2:n, x = -1, dims = c(n, n))
  G <- ngme_as_sparse(G)
  C <- ngme_as_sparse(C)
  stopifnot("The mesh should be 1d and has gap 1." = all(h == 1))

  theta_K <- ar1_a2th(rho)
  stopifnot("The length of rho(theta_K) should be 1." = length(theta_K) == 1)

  update_K <- function(theta_K) {
    ar1_th2a(theta_K) * C + G
  }

  # Create a generic model internally
  g <- name2fun("tanh", inv = TRUE)

  ngme_operator(
    mesh = mesh,
    model = "ar1",
    theta_K = c(rho = g(rho)),
    trans = c(rho = "tanh"),
    matrices = list(C, G),
    h = h,
    C = C,
    G = G,
    update_K = update_K,
    K = ngme_as_sparse(update_K(theta_K)),
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name = "rho",
    param_trans = list(ar1_th2a),
    # using the generic structure to build the operator
    generic_type = "generic"
  )
}

#' ngme ARMA(p, q) model specification
#'
#' General ARMA(p, q) latent operator under AZW: K W = Z eps,
#' where K encodes the AR part and Z encodes the MA part.
#'
#' @param mesh Integer vector or `inla.mesh.1d` object.
#' @param ar_order Integer p (0, 1, or 2).
#' @param ma_order Integer q (0, 1, or 2).
#' @param ar Numeric length-p vector of AR coefficients in (-1, 1). Defaults to zeros.
#' @param ma Numeric length-q vector of MA coefficients in (-1, 1). Defaults to zeros.
#' @param fix_ar Logical or length-p logical vector to fix AR coefficients.
#' @param fix_ma Logical or length-q logical vector to fix MA coefficients.
#'
#' @details
#' - Supports p, q <= 2. The user provides standard AR/MA coefficients (ar, ma).
#'   For estimation, these are transformed in R into an unbounded parameter vector
#'   `theta_K` via PACF-style mappings so optimizers (e.g., Adam/SGD) operate on
#'   unconstrained variables while K and Z remain valid:
#'   - AR(1): phi1 = tanh(raw1/2). AR(2): pi1=tanh(raw1/2), pi2=tanh(raw2/2),
#'     phi2 = pi2, phi1 = pi1 * (1 - pi2).
#'   - MA(1): theta1 = tanh(raw1/2). MA(2): psi1=tanh(raw1/2), psi2=tanh(raw2/2),
#'     theta2 = psi2, theta1 = psi1 * (1 - psi2).
#' - The returned operator includes both K (AR part) and Z (MA part). During
#'   estimation the backend updates Z synchronously every time `theta_K` changes.
#'
#' @return An `ngme_operator` describing the ARMA(p, q) latent model.
#'
#' @examples
#' mesh <- 1:10
#' op <- arma(mesh, ar_order = 2, ma_order = 2, ar = c(0.5, 0.3), ma = c(0.6, 0.4))
#'
#' @export
arma <- function(
    mesh = NULL,
    ar_order = 1,
    ma_order = 1,
    ar = NULL, # AR coefficients (friendly scale)
    ma = NULL, # MA coefficients (friendly scale)
    fix_ar = FALSE,
    fix_ma = FALSE) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, "inla.mesh.1d"))) {
    return(structure(list(model = "arma", args = as.list(environment())), class = "ngme_operator_def"))
  }
  p <- as.integer(ar_order)
  q <- as.integer(ma_order)
  stopifnot("ar_order >= 0" = p >= 0, "ma_order >= 0" = q >= 0)

  if (is.null(ar)) ar <- rep(0, p) else stopifnot(length(ar) == p)
  if (is.null(ma)) ma <- rep(0, q) else stopifnot(length(ma) == q)

  mesh <- ngme_build_mesh(mesh)
  n <- mesh$n
  h <- c(diff(mesh$loc), 1)
  stopifnot("The mesh should be 1d and has gap 1." = all(h[1:(n - 1)] == 1))

  # Build AR bases for placeholder K printing (Cj and G)
  Cs <- vector("list", max(1, p))
  if (p > 0) {
    for (j in 1:p) {
      Cs[[j]] <- Matrix::sparseMatrix(i = (p + 1):n, j = (p + 1 - j):(n - j), x = -1, dims = c(n, n))
    }
  } else {
    Cs[[1]] <- Matrix::sparseMatrix(i = 1, j = 1, x = 0, dims = c(n, n))
  }
  G <- Matrix::Diagonal(n)
  Cs <- lapply(Cs, ngme_as_sparse)
  G <- ngme_as_sparse(G)

  # Helper mappings (friendly <-> raw via AR1 transforms)
  clamp_unit <- function(x) pmax(-1 + 1e-8, pmin(1 - 1e-8, x))
  to_raw_vec <- function(coef, order) {
    if (order == 0) {
      return(numeric(0))
    }
    coef <- clamp_unit(coef)
    if (order == 1) {
      return(ar1_a2th(coef[1]))
    } else if (order == 2) {
      u2 <- coef[2]
      u1 <- coef[1] / (1 - u2)
      # clamp derived PACF before transform
      u1 <- clamp_unit(u1)
      u2 <- clamp_unit(u2)
      return(c(ar1_a2th(u1), ar1_a2th(u2)))
    } else {
      stop("Only order <= 2 supported")
    }
  }
  from_raw_vec <- function(raw, order) {
    if (order == 0) {
      return(numeric(0))
    }
    if (order == 1) {
      v1 <- ar1_th2a(raw[1])
      return(v1)
    } else if (order == 2) {
      t1 <- ar1_th2a(raw[1])
      t2 <- ar1_th2a(raw[2])
      v2 <- t2
      v1 <- t1 * (1 - t2)
      return(c(v1, v2))
    } else {
      stop("Only order <= 2 supported")
    }
  }

  # Build raw theta_K from friendly ar/ma
  theta_ar_raw <- to_raw_vec(ar, p)
  theta_ma_raw <- to_raw_vec(ma, q)
  theta_K <- c(theta_ar_raw, theta_ma_raw)
  names(theta_K) <- c(if (p > 0) paste0("ar", 1:p), if (q > 0) paste0("ma", 1:q))

  # placeholder K using friendly AR coefficients (printing)
  K0 <- G
  if (p > 0) {
    for (j in 1:p) K0 <- K0 + ar[j] * Cs[[j]]
  }
  K0 <- ngme_as_sparse(K0)

  # Attach spec for MA part (names for MA params now 'ma1..ma_q')
  Z_spec <- list(type = "ma", order = q, param = paste0("ma", 1:q))

  # Build MA shift matrices L^k = -Cs[[k]] (since Cs has -1 on sub-diagonal positions)
  Ls <- vector("list", max(1, q))
  if (q > 0) {
    for (j in 1:q) Ls[[j]] <- -Cs[[j]]
  } else {
    Ls[[1]] <- Matrix::sparseMatrix(i = 1, j = 1, x = 0, dims = c(n, n))
  }

  # placeholder Z using friendly MA coefficients: Z = I + sum theta_j L^j
  Z0 <- G
  if (q > 0) {
    for (j in 1:q) Z0 <- Z0 + ma[j] * Ls[[j]]
  }
  Z0 <- ngme_as_sparse(Z0)

  # Build per-parameter transforms mapping raw -> user-friendly proxy.
  # Note: for order 2, transforms map to PACF components (tanh(raw/2)),
  # because ar1 depends on both raw1 and raw2 and cannot be reconstructed
  # element-wise. The print method shows true AR/MA; traceplot will display PACF
  # unless we add grouped transforms (can be added on request).
  make_trans_list <- function(p, q) {
    trans <- list()
    if (p > 0) trans <- c(trans, rep(list(ar1_th2a), p))
    if (q > 0) trans <- c(trans, rep(list(ar1_th2a), q))
    trans
  }

  ngme_operator(
    mesh = mesh,
    model = "arma",
    theta_K = theta_K,
    # param_space omitted: R passes raw theta_K now
    matrices = c(Cs, list(G)),
    h = h,
    K = K0,
    Z = Z0,
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name = names(theta_K),
    param_trans = make_trans_list(p, q),
    Z_spec = Z_spec,
    fix_rho = rep(as.logical(fix_ar), p),
    fix_phi = rep(as.logical(fix_ma), q),
    p = p, q = q,
    generic_type = "none"
  )
}

#' Convenience wrapper for ARMA(1,1)
#'
#' @inheritParams arma
#' @return An \code{ngme_operator} object describing an ARMA(1,1) latent
#'   operator. The object contains the AR precision component \code{K}, the MA
#'   filter component \code{Z}, integration weights \code{h}, and the
#'   unconstrained AR and MA parameters used by the optimizer. If \code{mesh} is
#'   \code{NULL}, returns an \code{ngme_operator_def} specification for delayed
#'   construction inside \code{f()}.
#' @export
arma11 <- function(
    mesh = NULL,
    ar = 0,
    ma = 0,
    fix_ar = FALSE,
    fix_ma = FALSE) {
  arma(mesh,
    ar_order = 1,
    ma_order = 1,
    ar = ar,
    ma = ma,
    fix_ar = fix_ar,
    fix_ma = fix_ma
  )
}

#' Helper function to compute AR(p) autocovariance
#'
#' @param rho vector of AR coefficients
#' @param p AR order
#' @return vector of autocovariances from lag 0 to lag p-1
#' @keywords internal
ARcov <- function(rho, p) {
  # Solve Yule-Walker equations for autocovariances
  # gamma(0), gamma(1), ..., gamma(p-1)

  # Create the coefficient matrix A
  A <- matrix(0, p, p)
  for (i in 1:p) {
    for (j in 1:p) {
      lag <- abs(i - j)
      if (lag == 0) {
        A[i, j] <- 1
      } else {
        A[i, j] <- -rho[lag]
      }
    }
  }

  # Right hand side vector b
  b <- c(1, rep(0, p - 1))

  # Solve A * gamma = b
  gamma <- solve(A, b)

  return(gamma)
}

#' ngme AR(p) model specification
#'
#' @param mesh integer vector or inla.mesh.1d object, index to build the mesh
#' @param rho vector of AR coefficients of length p (should satisfy stationarity conditions)
#' @param order integer, the AR order p. If provided and rho is not specified, rho will be initialized as rep(0, p)
#'
#' @return ngme_operator object
#' @export
#'
#' @examples
#' # AR(2) model with specified coefficients
#' ar(c(1:10), rho = c(0.5, -0.3))
#' # AR(3) model with specified coefficients
#' ar(c(1:10), rho = c(0.4, 0.2, -0.1))
#' # AR(2) model with default coefficients (zeros)
#' ar(c(1:10), order = 2)
#' # AR(3) model with specified order and coefficients
#' ar(c(1:10), order = 3, rho = c(0.4, 0.2, -0.1))
ar <- function(
    mesh = NULL, rho = NULL, order = NULL) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, "inla.mesh.1d"))) {
    return(structure(list(model = "ar", args = as.list(environment())), class = "ngme_operator_def"))
  }
  # Determine p and rho based on input parameters
  if (is.null(order) && is.null(rho)) {
    stop("Either 'rho' or 'order' must be provided")
  } else if (is.null(order) && !is.null(rho)) {
    # Infer order from rho length
    p <- length(rho)
  } else if (!is.null(order) && is.null(rho)) {
    # Use specified order and initialize rho as zeros
    p <- order
    rho <- rep(0, p)
  } else {
    # Both order and rho provided - check consistency
    p <- order
    stopifnot("Length of rho must match the specified order" = length(rho) == p)
  }

  stopifnot("p should be >= 1" = p >= 1)

  # Check stationarity conditions for AR(p)
  # For now, we'll do a simple check for AR(2)
  # if (p == 2) {
  #   stopifnot("AR(2) stationarity conditions violated" =
  #             rho[1] + rho[2] < 1 &&
  #             rho[2] - rho[1] < 1 &&
  #             abs(rho[2]) < 1)
  # } else if (p > 2) {
  #   # For higher order AR, we check if all roots of characteristic polynomial
  #   # are outside unit circle (simplified check)
  #   char_poly <- c(1, -rho)
  #   roots <- polyroot(char_poly)
  #   if (any(abs(roots) <= 1.001)) {  # small tolerance for numerical errors
  #     warning("AR(p) model may not be stationary")
  #   }
  # }

  mesh <- ngme_build_mesh(mesh)
  n <- mesh$n
  h <- c(diff(mesh$loc), 1)

  stopifnot("The mesh should be 1d and has gap 1." = all(h[1:(n - 1)] == 1))
  stopifnot("n should be >= p" = n >= p)

  # Construct C1, ..., Cp and G matrices
  Cs <- vector("list", p)
  for (j in 1:p) {
    # Cj: for lag j, put -1 at (t, t-j) for t = (j+1):n
    # The first p rows should be all 0, so we only fill from row (p+1) to n
    Cs[[j]] <- Matrix::sparseMatrix(
      i = (p + 1):n,
      j = (p + 1 - j):(n - j),
      x = rep(-1, n - p),
      dims = c(n, n)
    )
  }
  # G: identity matrix
  G <- Matrix::Diagonal(n)
  G <- ngme_as_sparse(G)

  # Create transformation functions
  rho_trans <- rep(list(identity), p)
  names(rho_trans) <- paste0("rho", 1:p)
  trans <- setNames(rep("tanh", p), paste0("rho", 1:p))

  # Initial theta_K values (no transformation for now)
  theta_K <- rho
  names(theta_K) <- paste0("rho", 1:p)

  # update_K: sum_j rho[j] * Cj + G
  update_K <- function(theta_K) {
    K <- G
    for (j in 1:p) {
      K <- K + theta_K[j] * Cs[[j]]
    }
    K
  }

  K <- ngme_as_sparse(update_K(theta_K))

  ngme_operator(
    mesh = mesh,
    model = paste0("ar", p),
    theta_K = theta_K,
    trans = trans,
    matrices = c(Cs, list(G)),
    h = h,
    update_K = update_K,
    K = K,
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name = paste0("rho", 1:p),
    param_trans = rho_trans,
    generic_type = "generic"
  )
}

#' Random Walk Model of Order 1 (RW1)
#'
#' Constructs a first-order random walk model for spatial or temporal processes.
#' The RW1 model assumes that first-order differences \eqn{\Delta W_i = W_i - W_{i-1}}
#' are independent and identically distributed Gaussian variables.
#'
#' @details
#' The RW1 model is defined by the precision matrix \eqn{K} that penalizes
#' first-order differences. The model has different structures depending on the
#' constraints:
#'
#' **Non-cyclic, constrained (default)**: The first row of \eqn{K} enforces the
#' constraint \eqn{\sum_{i=1}^n h_i W_i = 0}, where \eqn{h_i} are the mesh weights.
#'
#' **Non-cyclic, unconstrained**: The first element is fixed at 0 (\eqn{W_1 = 0}).
#'
#' **Cyclic**: Treats the domain as circular, connecting the first and last locations
#' as neighbors. The constraint \eqn{\sum_{i=1}^n h_i W_i = 0} is always enforced.
#' The precision matrix is expanded to size \eqn{(n+1) \times (n+1)} to handle
#' the constraint properly.
#'
#' @param mesh numerical vector or inla.mesh.1d object, locations to build the mesh.
#'   For numerical vectors, assumes equally spaced locations.
#' @param cyclic logical, whether the mesh is circular. If TRUE, the first and last
#'   locations are treated as neighbors. Cannot be FALSE when constr = TRUE.
#' @param constr logical, whether to enforce the sum-to-zero constraint
#'   \eqn{\sum_{i=1}^n h_i W_i = 0}. If FALSE, fixes the first element \eqn{W_1 = 0}.
#'   Must be TRUE for cyclic models.
#'
#' @return An `ngme_operator` object containing the precision matrix and related
#'   components for the RW1 model.
#' @export
#'
#' @examples
#' # Non-cyclic constrained RW1 (default)
#' rw1_default <- rw1(1:5)
#' print(rw1_default$K)
#'
#' # Non-cyclic unconstrained RW1 (fixes first element to 0)
#' rw1_fixed <- rw1(1:5, constr = FALSE)
#' print(rw1_fixed$K)
#'
#' # Cyclic RW1 (connects first and last locations)
#' rw1_cyclic <- rw1(1:5, cyclic = TRUE)
#' print(rw1_cyclic$K)
#'
#' # Using with unequally spaced locations
#' locations <- c(0, 1, 3, 6, 10)
#' rw1_unequal <- rw1(locations)
#' print(rw1_unequal$K)
rw1 <- function(
    mesh = NULL,
    cyclic = FALSE,
    constr = TRUE) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, "inla.mesh.1d"))) {
    return(structure(list(model = "rw1", args = as.list(environment())), class = "ngme_operator_def"))
  }
  mesh <- ngme_build_mesh(mesh)

  n <- mesh$n
  h_left <- c(0, diff(mesh$loc))
  h_right <- c(diff(mesh$loc), 0)
  h <- 0.5 * (h_left + h_right)

  if (!cyclic) {
    C <- Matrix::sparseMatrix(i = 1:n, j = 1:n, x = 1, dims = c(n, n))
    G <- Matrix::sparseMatrix(i = 2:n, j = 1:(n - 1), x = -1, dims = c(n, n))
    K_mat <- C + G
    if (constr) K_mat[1, ] <- h else K_mat[1, 1] <- 1
    K <- ngme_as_sparse(K_mat)
  } else {
    stopifnot("constraint cannot be false for cyclic case" = constr)
    stopifnot("Too less data point" = length(h) >= 3)
    C <- Matrix::Diagonal(n)
    G <- Matrix::sparseMatrix(i = 1:n, j = c(2:n, 1), x = -1, dims = c(n, n))
    K_mat <- C + G
    # expand K to be (0 1 1 K)
    K_mat <- rbind(0, K_mat)
    K_mat <- cbind(1, K_mat)
    K_mat[1, -1] <- h
    K_mat[1, 1] <- 0
    K <- ngme_as_sparse(K_mat)
    h <- c(1, h)
  }

  ngme_operator(
    mesh = mesh,
    cyclic = cyclic,
    matrices = list(K),
    model = "rw1",
    generic_type = "generic",
    theta_K = double(0),
    update_K = function(theta_K) {
      K
    },
    K = K,
    h = h,
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name = character(0)
  )
}

#' Random Walk Model of Order 2 (RW2)
#'
#' Constructs a second-order random walk model for spatial or temporal processes.
#' The RW2 model assumes that second-order differences
#' \eqn{\Delta^2 W_i = W_i - 2W_{i-1} + W_{i-2}} are independent and identically
#' distributed Gaussian variables.
#'
#' @details
#' The RW2 model is defined by the precision matrix \eqn{K} that penalizes
#' second-order differences, making it smoother than RW1. The model enforces
#' constraints to ensure identifiability:
#'
#' **Non-cyclic (default)**: The first two rows of \eqn{K} enforce constraints:
#' the first row implements \eqn{\sum_{i=1}^n h_i W_i = 0} (sum-to-zero),
#' and the second row implements \eqn{\sum_{i=1}^n h_i \cdot i \cdot W_i = 0}
#' (removes linear trend). The remaining rows penalize second-order differences.
#'
#' **Cyclic**: Treats the domain as circular, connecting the first and last locations
#' as neighbors. The constraint \eqn{\sum_{i=1}^n h_i W_i = 0} is enforced.
#' The precision matrix is expanded to size \eqn{(n+1) \times (n+1)} to handle
#' the constraint and circular structure properly.
#'
#' @param mesh numerical vector or inla.mesh.1d object, locations to build the mesh.
#'   For numerical vectors, assumes equally spaced locations. Must have at least 3 locations.
#' @param constr logical, whether to enforce the sum-to-zero constraint \eqn{\sum_{i=1}^n h_i W_i = 0}
#'   and the linear trend constraint \eqn{\sum_{i=1}^n (\sum_{j=1}^i h_j - h_1) \cdot W_i = 0}.
#'   If FALSE, fixes the first and second elements \eqn{W_1 = 0} and \eqn{W_2 = 0}.
#' @param cyclic logical, whether the mesh is circular. If TRUE, the first and last
#'   locations are treated as neighbors with second-order differences computed
#'   across the boundary.
#'
#' @return An `ngme_operator` object containing the precision matrix and related
#'   components for the RW2 model.
#' @export
#'
#' @examples
#' # Non-cyclic RW2 with constraints (default)
#' rw2_default <- rw2(1:6)
#' print(rw2_default$K)
#'
#' # Cyclic RW2 (connects first and last locations)
#' rw2_cyclic <- rw2(1:6, cyclic = TRUE)
#' print(rw2_cyclic$K)
#'
#' # Using with unequally spaced locations
#' locations <- c(0, 1, 3, 6, 10, 15)
#' rw2_unequal <- rw2(locations)
#' print(rw2_unequal$K)
rw2 <- function(
    mesh = NULL,
    cyclic = FALSE,
    constr = TRUE) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, "inla.mesh.1d"))) {
    return(structure(list(model = "rw2", args = as.list(environment())), class = "ngme_operator_def"))
  }
  mesh <- ngme_build_mesh(mesh)

  stopifnot("Mesh should be inla.mesh.1d." = inherits(mesh, c("inla.mesh.1d")))
  n <- mesh$n
  h_left <- c(0, diff(mesh$loc))
  h_right <- c(diff(mesh$loc), 0)
  h <- 0.5 * (h_left + h_right)
  stopifnot("mesh too small" = n >= 3)

  if (!cyclic) {
    C <- Matrix::sparseMatrix(i = 3:n, j = 2:(n - 1), x = -2, dims = c(n, n))
    G <- Matrix::sparseMatrix(i = c(1:n, 3:n), j = c(1:n, 1:(n - 2)), x = 1, dims = c(n, n))
    K <- as.matrix(C + G)
    if (constr) {
      K[1, ] <- h
      K[2, ] <- cumsum(h) - h[1]
    } else {
      K[1, 1] <- 1
      K[2, 2] <- 1
    }
    K <- ngme_as_sparse(K)
  } else {
    C <- Matrix::sparseMatrix(i = 1:n, j = c(2:n, 1), x = -2, dims = c(n, n))
    G <- Matrix::sparseMatrix(i = rep(1:n, 2), j = c(1:n, 3:n, 1, 2), x = 1, dims = c(n, n))
    K <- as.matrix(C + G)
    K <- rbind(0, K)
    K <- cbind(1, K)
    K[1, -1] <- h
    K[1, 1] <- 0
    K <- ngme_as_sparse(K)
    h <- c(1, h)
    K <- ngme_as_sparse(K)
  }

  ngme_operator(
    mesh = mesh,
    model = "rw2",
    generic_type = "generic",
    cyclic = cyclic,
    matrices = list(K),
    theta_K = double(0),
    update_K = function(theta_K) {
      K
    },
    K = K,
    h = h,
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name = character(0)
  )
}

#' Ornstein-Uhlenbeck Process Model
#'
#' Implements the exact discrete-time representation of the continuous
#' Ornstein-Uhlenbeck (OU) process using a matrix formulation
#' \eqn{\mathbf{K}\mathbf{X} = \boldsymbol{\delta}}.
#'
#' @details
#'
#' The OU process is defined by the stochastic differential equation (SDE):
#'
#' \deqn{dX_t = -\theta X_t dt + \sigma dW_t}
#'
#' where \eqn{\theta > 0} is the mean-reversion rate (stiffness) and
#' \eqn{\sigma} is the volatility (diffusion coefficient).
#'
#' ## Exact Discretization
#'
#' Unlike Euler-Maruyama approximations, the OU process has an exact solution
#' between time points \eqn{t_i} and \eqn{t_{i+1}} with step size \eqn{\Delta t_i}:
#'
#' \deqn{X_{i+1} = X_i e^{-\theta \Delta t_i} + \epsilon_i}
#'
#' This is an AR(1) form with autoregressive coefficient:
#'
#' \deqn{\rho_i = e^{-\theta \Delta t_i}}
#'
#' and Gaussian noise:
#'
#' \deqn{\epsilon_i \sim \mathcal{N}\left(0, \frac{\sigma^2}{2\theta}(1 - \rho_i^2)\right)}
#'
#' ## Matrix Representation
#'
#' The precision matrix \eqn{\mathbf{K}} for the linear system
#' \eqn{\mathbf{K}\mathbf{X} = \boldsymbol{\delta}} is constructed as:
#'
#' \deqn{\mathbf{K} = \begin{bmatrix}
#' \sqrt{1-\rho_1^2} & 0 & 0 & \cdots & 0 \\
#' -\rho_1 & 1 & 0 & \cdots & 0 \\
#' 0 & -\rho_2 & 1 & \cdots & 0 \\
#' \vdots & \vdots & \ddots & \ddots & \vdots \\
#' 0 & 0 & \cdots & -\rho_{n-1} & 1
#' \end{bmatrix}}
#'
#' where \eqn{\rho_i = \exp(-\theta \Delta t_i)} with \eqn{\Delta t_i = t_{i+1} - t_i}.
#'
#' **Why the \eqn{\sqrt{1-\rho_1^2}} term?** The first row ensures that \eqn{X_1}
#' is drawn from the stationary distribution \eqn{\mathcal{N}(0, \sigma^2/(2\theta))}.
#' This scaling factor matches the variance of the first component of
#' \eqn{\boldsymbol{\delta}} with the transition noise variance in subsequent steps.
#'
#' ## Non-uniform Mesh
#'
#' For non-uniform meshes, each \eqn{\rho_i} is computed locally based on the
#' varying time step \eqn{\Delta t_i}, allowing the model to naturally handle
#' irregularly spaced observations.
#'
#' ## The \eqn{h} vector
#'
#' The \eqn{h} vector represents integration weights for the mass matrix.
#' For the OU model, it is set as:
#'
#' \deqn{h = [\Delta t_1, \Delta t_2, \ldots, \Delta t_{n-1}, \Delta t_{n-1}]}
#'
#' where the last element is duplicated. This ensures proper weighting in the
#' finite element formulation.
#'
#' ## The \eqn{\boldsymbol{\delta}} vector
#'
#' In the equation \eqn{\mathbf{K}\mathbf{X} = \boldsymbol{\delta}}, the vector
#' \eqn{\boldsymbol{\delta}} contains independent identically distributed (i.i.d.)
#' random variables with unit variance. The precision matrix \eqn{\mathbf{K}}
#' is constructed such that the resulting process \eqn{\mathbf{X}} has the
#' correct OU covariance structure.
#'
#' @param mesh numerical vector or `inla.mesh.1d` object giving the ordered
#'   time locations. Must be strictly increasing.
#' @param theta positive mean-reversion rate (stiffness parameter). Internally
#'   stored on the log scale so optimization remains unconstrained.
#'
#' @return An `ngme_operator` object containing:
#' \itemize{
#'   \item `K`: the precision matrix
#'   \item `h`: integration weights
#'   \item `theta_K`: log-transformed theta for unconstrained optimization
#'   \item `mesh`: the mesh object
#' }
#'
#' @references
#' Uhlenbeck, G. E., & Ornstein, L. S. (1930). On the theory of the Brownian motion.
#' Physical review, 36(5), 823.
#'
#' @export
#' @examples
#' # Uniform mesh
#' mesh_uniform <- seq(0, 10, by = 0.5)
#' ou_uniform <- ou(mesh = mesh_uniform, theta = 0.5)
#' print(ou_uniform)
#'
#' # Non-uniform mesh
#' mesh_nonuniform <- c(0, 1, 2.5, 3.5, 5, 8, 11)
#' ou_nonunif <- ou(mesh = mesh_nonuniform, theta = 0.8)
#' print(ou_nonunif)
ou <- function(
    mesh = NULL,
    theta = 1) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, "inla.mesh.1d"))) {
    return(structure(list(model = "ou", args = as.list(environment())), class = "ngme_operator_def"))
  }

  mesh <- ngme_build_mesh(mesh)
  stopifnot("Mesh should be inla.mesh.1d." = inherits(mesh, c("inla.mesh.1d")))

  n <- mesh$n
  stopifnot("mesh should have at least two locations" = n >= 2)

  dt <- diff(mesh$loc)
  stopifnot("mesh locations must be strictly increasing" = all(dt > 0))

  theta_raw <- log(theta)
  stopifnot("theta must be a single positive value" = length(theta) == 1 && is.finite(theta) && theta > 0)

  build_K <- function(theta_val) {
    rho <- exp(-theta_val * dt)
    stopifnot("All rho values must lie in (0, 1)" = all(rho > 0 & rho < 1))

    diag_main <- c(sqrt(pmax(0, 1 - rho[1]^2)), rep(1, n - 1))
    sub_diag <- -rho

    Matrix::bandSparse(
      n = n,
      k = c(0, -1),
      diagonals = list(diag_main, sub_diag)
    )
  }

  update_K <- function(theta_K) {
    ngme_as_sparse(build_K(exp(theta_K)))
  }

  K <- update_K(theta_raw)
  h <- c(dt, utils::tail(dt, 1))

  ngme_operator(
    mesh = mesh,
    model = "ou",
    theta_K = c(theta = theta_raw),
    trans = c(theta = "exp"),
    update_K = update_K,
    K = K,
    h = h,
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name = "theta",
    param_trans = list(exp),
    generic_type = "none"
  )
}

#' ngme Matern SPDE model specification
#'
#' @param mesh an fmesher::fm_mesh_2d object, mesh for build the SPDE model
#' @param alpha SPDE smoothness parameter (alpha = 2beta) 2 or 4 for integer case, otherwise for fractional case
#' @param kappa the range parameter, for the stationary model, it is the range parameter
#' @param theta_kappa kappa = exp(B_kappa * theta_kappa), for the non-stationary model, it is the range parameter
#' @param fix_alpha if FALSE, enable fractional modeling
#' @param rational_order order of rational approximation for fractional operators (default: 2)
#' @param B_kappa bases for theta_kappa, ignore if use the stationary model
#'
#' @return ngme_operator object
#' @export
matern <- function(
    mesh = NULL,
    alpha = 2,
    fix_alpha = TRUE,
    kappa = NULL,
    theta_kappa = NULL,
    rational_order = 2,
    B_kappa = NULL) {
  if (is.null(mesh) || (is.list(mesh) && !inherits(mesh, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph")))) {
    return(structure(list(model = "matern", args = as.list(environment())), class = "ngme_operator_def"))
  }

  mesh <- ngme_build_mesh(mesh)
  if (fix_alpha && alpha != 2 && alpha != 4) {
    if (!requireNamespace("rSPDE", quietly = TRUE)) {
      stop("For fixed alpha values not equal to 2 or 4, the 'rSPDE' package is required. Please install it before using this option.")
    }
  }

  # Get mesh dimensions first (needed for B_kappa initialization)
  if (inherits(mesh, "metric_graph")) {
    if (is.null(mesh$mesh$C)) mesh$compute_fem()
    C <- mesh$mesh$C
    G <- mesh$mesh$G
    h <- mesh$mesh$weights
    d <- 0
    # C is not mass lumped..
    C <- Matrix::Diagonal(nrow(C), Matrix::rowSums(C))
  } else {
    d <- get_inla_mesh_dimension(mesh)
    fem <- fmesher::fm_fem(mesh, order = alpha)
    C <- fem$c0
    G <- fem$g1
    h <- Matrix::diag(fem$c0)
  }
  Ci <- Matrix::Diagonal(x = 1 / Matrix::diag(C))

  mesh_n <- length(h)

  # Optimized initialization logic for kappa, theta_kappa, and B_kappa
  if (is.null(kappa)) {
    if (is.null(theta_kappa)) {
      # Case 1: kappa = theta_kappa = NULL
      if (is.null(B_kappa)) {
        # Stationary case: set theta_kappa=0, B_kappa as n_mesh*1 matrix of ones
        theta_kappa <- 0
        B_kappa <- matrix(1, nrow = mesh_n, ncol = 1)
      } else {
        # Non-stationary case: B_kappa is given, auto-set theta_kappa
        theta_kappa <- rep(0, ncol(B_kappa))
      }
    } else {
      # Case 2: theta_kappa is provided, kappa is NULL
      if (is.null(B_kappa)) {
        if (length(theta_kappa) == 1) {
          # Stationary case
          B_kappa <- matrix(1, nrow = mesh_n, ncol = 1)
        } else {
          stop("Please provide B_kappa for non-stationary case.")
        }
      }
    }
  } else {
    # Case 3: kappa is provided
    stopifnot("kappa should be a single value" = length(kappa) == 1)
    theta_kappa <- log(kappa)
    if (is.null(B_kappa)) {
      B_kappa <- matrix(1, nrow = mesh_n, ncol = 1)
    }
  }

  stationary <- is_stationary(B_kappa)
  # Build parameter vector theta_K
  # theta_K = c(alpha, theta_kappa) if alpha is not fixed, otherwise theta_K = theta_kappa
  L <- d / 2
  alpha_clamped <- max(L + 1e-6, min(4 - 1e-6, alpha))
  if (fix_alpha) {
    # Do not include alpha in theta when fixed; only pass kappa params
    theta_K <- if (stationary) c(theta_kappa) else c(as.numeric(theta_kappa))
  } else {
    eta_alpha <- qlogis((alpha_clamped - L) / (4 - L))
    theta_K <- if (stationary) c(eta_alpha, theta_kappa) else c(eta_alpha, as.numeric(theta_kappa))
  }

  C <- ngme_as_sparse(C)
  G <- ngme_as_sparse(G)
  Ci <- ngme_as_sparse(Ci)
  B_K <- if (stationary) matrix(1, nrow = mesh_n, ncol = 1) else B_kappa

  # Placeholder K for printing; real K is built in C++
  update_K <- function(theta_K) {
    if (stationary && fix_alpha && alpha == 4) {
      kappa_sq <- exp(2 * theta_K[1])
      K <- kappa_sq^2 * C + 2 * kappa_sq * G + G %*% Ci %*% G
    } else if (stationary) {
      K <- exp(if (fix_alpha) theta_K[1] else theta_K[2])^2 * C + G
    } else if (fix_alpha && alpha == 4) {
      kappas <- exp(as.numeric(B_K %*% theta_K))
      D_kappa <- Matrix::Diagonal(x = kappas)
      K <- D_kappa %*% C %*% D_kappa %*% Ci %*% D_kappa %*% C %*% D_kappa +
        D_kappa %*% C %*% D_kappa %*% Ci %*% G +
        G %*% Ci %*% D_kappa %*% C %*% D_kappa +
        G %*% Ci %*% G
    } else if (fix_alpha && alpha == 2) {
      kappas <- exp(as.numeric(B_K %*% theta_K))
      D_kappa <- Matrix::Diagonal(x = kappas)
      K <- D_kappa %*% C %*% D_kappa + G
    } else {
      kappas <- exp(as.numeric(B_K %*% (if (fix_alpha) theta_K else theta_K[-1])))
      K <- Matrix::Diagonal(x = kappas^2) %*% C + G
    }
    ngme_as_sparse(K)
  }

  # Build operator args and conditionally include roots and param meta
  op_args <- list(
    mesh = mesh,
    model = "matern",
    update_K = update_K,
    K = update_K(theta_K),
    h = h,
    theta_K = theta_K,
    generic_type = "none",
    C = C,
    Ci = Ci,
    G = G,
    B_K = B_K,
    alpha = alpha,
    fix_alpha = fix_alpha,
    spatial_dim = d,
    symmetric = TRUE,
    stationary = stationary,
    rational_order = rational_order,
    param_name = NULL,
    param_trans = NULL
  )
  # Parameter names/transforms for traceplot
  if (fix_alpha) {
    if (stationary) {
      op_args$param_name <- c("kappa")
      op_args$param_trans <- list(exp)
    } else {
      p <- ncol(B_kappa)
      op_args$param_name <- make.unique(rep("kappa", p))
      op_args$param_trans <- rep(list(identity), p)
    }
  } else {
    L <- d / 2
    alpha_trans <- function(eta) L + (4 - L) * (1 / (1 + exp(-eta)))
    if (stationary) {
      op_args$param_name <- c("alpha", "kappa")
      op_args$param_trans <- list(alpha_trans, exp)
    } else {
      p <- ncol(B_kappa)
      op_args$param_name <- c("alpha", make.unique(rep("kappa", p)))
      op_args$param_trans <- c(list(alpha_trans), rep(list(identity), p))
    }
  }
  do.call(ngme_operator, op_args)
}

#' ngme random effect model
#'
#' @param map numerical vector, covariates to build index for the process (can be formula, provided data)
#' @param theta_K initial value for theta_K (build covariance matrix)
#' @param ... currently ignored
#'
#' @return ngme_operator object
#' @export
re <- function(
    map,
    theta_K = NULL,
    ...) {
  if (missing(map) || is.null(map)) {
    return(structure(list(model = "re", args = list(theta_K = theta_K, ...)), class = "ngme_operator_def"))
  }
  if (inherits(map, "formula")) {
    map <- model.matrix(map)
  } else {
    map <- as.matrix(map)
  }

  B_theta_K <- map
  n_reff <- ncol(B_theta_K) # number of random effects
  n_theta_K <- sum(1:n_reff) # number of theta_K
  h <- rep(1, n_reff)

  # provide initial value for theta_K
  if (!is.null(theta_K)) {
    stopifnot(length(theta_K) == n_theta_K)
  } else {
    theta_K <- rep(0, n_theta_K)
  }

  # build K
  update_K <- function(theta_K) {
    K <- diag(n_reff)
    diag(K) <- exp(theta_K[1:n_reff])
    if (n_reff > 1) {
      K[lower.tri(K)] <- theta_K[(n_reff + 1):n_theta_K]
    }
    Matrix::Matrix(K)
  }
  K <- update_K(theta_K)

  ngme_operator(
    mesh = NULL,
    model = "re",
    theta_K = theta_K,
    K = ngme_as_sparse(K),
    update_K = update_K,
    h = h,
    B_theta_K = B_theta_K,
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name = paste("re_coeff", seq_len(length(theta_K)), sep = ""),
    param_trans = rep(list(identity), length(theta_K))
  )
}

# ----  For computing precision matrix of multivariate model

# p: dimension
# rho: a list of rho parameter, see paper D_l
D_l <- function(p, rho) {
  stopifnot(
    "rho should be of length p(p-1)/2" =
      length(rho) == p * (p - 1) / 2
  )
  D_l <- diag(p)
  D_l[lower.tri(D_l)] <- rho
  # compute k(j)
  k <- double(p)
  k[1] <- 1
  for (j in 2:p) {
    # print(D_l[j, 1:(j-1)])
    k[j] <- sqrt(1 + sum(D_l[j, 1:(j - 1)]^2))
  }
  D_l <- solve(D_l, diag(k))
  D_l
}

# return Q * D_l
dependence_matrix <- function(p, rho, theta = NULL, Q = NULL) {
  Q_2d <- function(theta) {
    Q <- matrix(0, nrow = 2, ncol = 2)
    Q[1, 1] <- cos(theta)
    Q[2, 2] <- cos(theta)
    Q[1, 2] <- -sin(theta)
    Q[2, 1] <- sin(theta)
    Q
  }
  stopifnot(
    p - round(p) == 0, p > 1,
    "Please provide theta (p <= 3) or Q matrix, see ?precision_matrix_multivariate" = !is.null(theta) | !is.null(Q)
  )

  # compute D_l
  D_l <- D_l(p, rho)
  # compute Q
  if (p == 2) {
    stopifnot(
      "Length of theta should be 1 for p=2 case" = length(theta) == 1
    )
    Q <- Q_2d(theta)
  } else if (p == 3) {
    stopifnot(
      "Length of theta should be 3 for p=3 case" = length(theta) == 3
    )

    Q_3x <- Matrix::bdiag(Q_2d(theta[1]), 1)
    Q_3z <- Matrix::bdiag(1, Q_2d(theta[3]))
    Q_3y <- diag(3)
    Q_3y[c(1, 3, 7, 9)] <- Q_2d(theta[2])
    Q <- Q_3x %*% Q_3y %*% Q_3z
  } else {
    if (is.null(Q)) stop("Please provide Q (p*p) for p > 3 case.")
  }

  Q %*% D_l
}

#' Compute the precision matrix for multivariate model
#'
#' @param p dimension, should be integer and greater than 1
#' @param operator_list a list of ngme_operator object (length should be p)
#' @param rho vector with the p(p-1)/2 correlation parameters rho_11, rho_21, rho_22, ... rho_p1, rho_p2, ... rho_p(p-1)
#' @param theta parameter for Q matrix (length of 1 when p=2, length of 3 when p=3)
#' @param Q orthogonal matrix of dim p*p (provide when p > 3)
#' @param scale A vector of length p with constants to multiply each operator matrix with
#'
#' @return the precision matrix of the multivariate model
#' @details The general model is defined as $D diag(L_1, ..., L_p) x = M$. D is the dependence matrix, it is paramterized by $D = Q(theta) * D_l(cor_mat)$, where $Q$ is the orthogonal matrix, and $D_l$ is matrix controls the cross-correlation.
#' See the section 2.2 of Bolin and Wallin (2020) for exact parameterization of Dependence matrix.
#' @references
#' Bolin, D. and Wallin, J. (2020), Multivariate type G Matérn stochastic partial differential equation random fields. J. R. Stat. Soc. B, 82: 215-239. https://doi.org/10.1111/rssb.12351
#' @export
#' @examples
#' rho <- c(-0.5, 0.5, -0.25) # correlation parameters
#' operator_list <- list(ar1(1:5, rho = 0.4), ar1(1:5, rho = 0.5), ar1(1:5, rho = 0.6))
#' precision_matrix_multivariate(3, operator_list, rho, theta = c(1, 2, 3))
precision_matrix_multivariate <- function(p,
                                          operator_list,
                                          rho,
                                          theta = NULL,
                                          Q = NULL,
                                          scale = NULL) {
  stopifnot("Please provide a list of p models" = length(operator_list) == p)
  stopifnot(
    "rho should be of length p(p-1)/2" =
      length(rho) == p * (p - 1) / 2
  )

  D <- dependence_matrix(p, rho, theta, Q)
  bigD <- kronecker(D, Matrix::Diagonal(length(operator_list[[1]]$h)))

  if (is.null(scale)) {
    scale <- rep(1, p)
  }
  K_list <- list()
  for (i in 1:p) {
    K_list[[i]] <- scale[i] * operator_list[[i]]$K
  }
  K <- bigD %*% Matrix::bdiag(K_list)

  # mass lumping version
  Cinv <- rep(1 / operator_list[[1]]$h, p)
  Matrix::t(K) %*% Matrix::Diagonal(x = Cinv) %*% K
}


#' Compute the precision matrix for multivariate spde Matern model
#'
#' @param p dimension, should be integer and greater than 1
#' @param mesh an fmesher::fm_mesh_2d object, mesh for build the SPDE model
#' @param alpha_list a list of SPDE smoothness parameter
#' @param theta_K_list a list (length is p) of theta_K
#' @param B_K_list a list (length is p) of B_K (non-stationary case)
#' @param variance_list If provided, it should be a vector of length p, where the
#' kth element corresponds to a desired variance of the kth field. The kth operator
#' is then scaled by a constant c so that this variance is achieved in the stationary case
#' (default no scaling)
#' @param rho vector with the p(p-1)/2 correlation parameters rho_11, rho_21, rho_22, ... rho_p1, rho_p2, ... rho_p(p-1)
#' @param theta parameter for Q matrix (length of 1 when p=2, length of 3 when p=3)
#' @param Q orthogonal matrix of dim p*p (provide when p > 3)
#'
#' @return the precision matrix of the multivariate model
#' @details The general model is defined as $D diag(L_1, ..., L_p) x = M$. D is the dependence matrix, it is paramterized by $D = Q(theta) * D_l(cor_mat)$, where $Q$ is the orthogonal matrix, and $D_l$ is matrix controls the cross-correlation.
#' See the section 2.2 of Bolin and Wallin (2020) for exact parameterization of Dependence matrix.
#' @references
#' Bolin, D. and Wallin, J. (2020), Multivariate type G Matérn stochastic partial differential equation random fields. J. R. Stat. Soc. B, 82: 215-239. https://doi.org/10.1111/rssb.12351
#' @export
#' @examples
#' library(fmesher)
#' library(fields)
#' # Define mesh
#' x <- seq(from = 0, to = 1, length.out = 40)
#' mesh <- fm_rcdt_2d_inla(lattice = fm_lattice_2d(x, x), extend = FALSE)
#' # Set parameters
#' p <- 3 # number of fields
#' rho <- c(-0.5, 0.5, -0.25) # correlation parameters
#' log_kappa <- list(2, 2, 2) # log(kappa)
#' variances <- list(1, 1, 1) # set marginal variances to 1
#' alpha <- list(2, 2, 2) # smoothness parameters
#' # Compute precision
#' Q <- precision_matrix_multivariate_spde(p,
#'   mesh = mesh, rho = rho,
#'   alpha = alpha, theta_K_list = log_kappa,
#'   variance_list = variances
#' )
#' # Plot the cross covariances
#' A <- as.vector(fm_basis(mesh, loc = matrix(c(0.5, 0.5), 1, 2)))
#' Sigma <- as.vector(solve(Q, c(A, rep(0, 2 * mesh$n))))
#' r11 <- Sigma[1:mesh$n]
#' r12 <- Sigma[(mesh$n + 1):(2 * mesh$n)]
#' r13 <- Sigma[(2 * mesh$n + 1):(3 * mesh$n)]
#' Sigma <- as.vector(solve(Q, c(rep(0, mesh$n), A, rep(0, mesh$n))))
#' r22 <- Sigma[(mesh$n + 1):(2 * mesh$n)]
#' r23 <- Sigma[(2 * mesh$n + 1):(3 * mesh$n)]
#' Sigma <- as.vector(solve(Q, v <- c(rep(0, 2 * mesh$n), A)))
#' r33 <- Sigma[(2 * mesh$n + 1):(3 * mesh$n)]
#'
#' proj <- fm_evaluator(mesh)
#'
#' oldpar <- par(no.readonly = TRUE)
#' par(mfrow = c(3, 3))
#' image.plot(fm_evaluate(proj, r11), main = "Cov(X_1(s0),X_1(s)")
#' plot.new()
#' plot.new()
#' image.plot(fm_evaluate(proj, r12), main = "Cov(X_1(s0),X_2(s)")
#' image.plot(fm_evaluate(proj, r22), main = "Cov(X_2(s0),X_2(s)")
#' plot.new()
#' image.plot(fm_evaluate(proj, r13), main = "Cov(X_1(s0),X_3(s)")
#' image.plot(fm_evaluate(proj, r23), main = "Cov(X_2(s0),X_3(s)")
#' image.plot(fm_evaluate(proj, r33), main = "Cov(X_3(s0),X_3(s)")
#' par(oldpar)
precision_matrix_multivariate_spde <- function(
    p,
    mesh,
    rho,
    alpha_list = NULL,
    theta_K_list = NULL,
    variance_list = NULL,
    B_K_list = NULL,
    theta = NULL,
    Q = NULL) {
  if (is.null(B_K_list)) {
    B_K_list <- lapply(1:p, function(i) NULL)
  } else {
    stopifnot(
      "Please provide a list of B_K matrices for each matern model" = length(B_K_list) == p
    )
  }

  if (is.null(theta_K_list)) {
    theta_K_list <- lapply(1:p, function(i) 0)
  } else {
    stopifnot(
      "Please provide a list of p parameters" = length(theta_K_list) == p
    )
  }

  if (is.null(alpha_list)) {
    alpha_list <- lapply(1:p, function(i) 2)
  } else {
    stopifnot(
      "Please provide a list of p alpha parameters" = length(alpha_list) == p
    )
  }

  if (is.null(variance_list)) {
    variance_list <- lapply(1:p, function(i) NULL)
    c <- NULL
  } else {
    stopifnot(
      "Please provide a list of p variance parameters" = length(variance_list) == p
    )
    if (mesh$manifold %in% c("R2", "S2")) {
      d <- 2
    } else if (mesh$manifold == "R1") {
      d <- 1
    }
    scale <- rep(0, p)
    for (i in 1:p) {
      scale[i] <- sqrt(gamma(alpha_list[[i]] - d / 2) / (variance_list[[i]] * (4 * pi)^(d / 2) * exp(theta_K_list[[i]][1])^(2 * (alpha_list[[i]] - d / 2)) * gamma(alpha_list[[i]])))
    }
  }

  operator_list <- NULL
  for (i in 1:p) {
    operator_list[[i]] <- matern(
      mesh = mesh,
      alpha = alpha_list[[i]],
      theta_kappa = theta_K_list[[i]],
      B_kappa = B_K_list[[i]]
    )
  }
  if (is.null(theta)) {
    theta <- rep(0, (p - 1) * p / 2)
  }
  precision_matrix_multivariate(p, operator_list, rho, theta, Q, scale = scale)
}

test_that("predict supports response draws and output-level samples", {
  withr::local_seed(42)

  n <- 30
  w <- as.numeric(arima.sim(list(ar = 0.5), n = n))
  y <- w + rnorm(n, sd = 0.3)

  fit <- ngme(
    y ~ 0 + f(1:n, model = ar1(), name = "ar1_field"),
    data = data.frame(y = y),
    family = noise_normal(),
    control_opt = control_opt(
      iterations = 30,
      burnin = 1,
      n_batch = 3,
      n_parallel_chain = 2,
      n_min_batch = 1,
      verbose = FALSE,
      print_check_info = FALSE,
      store_traj = FALSE,
      seed = 42
    )
  )

  pred <- predict(
    fit,
    map = list(ar1_field = seq_len(n)),
    type = "response",
    estimator = c("mean", "0.25q", "0.75q"),
    sampling_size = 25,
    burnin_size = 5,
    seed = 7,
    return_samples = TRUE
  )

  response_samples <- attr(pred, "samples")
  latent_samples <- attr(pred, "latent_samples")

  expect_true(is.matrix(response_samples))
  expect_equal(dim(response_samples), c(n, 25))
  expect_true(is.matrix(latent_samples))
  expect_equal(ncol(latent_samples), 25)

  expect_equal(pred$mean, rowMeans(response_samples), tolerance = 1e-8)
  expect_equal(
    pred$`0.25q`,
    apply(response_samples, 1, quantile, 0.25),
    tolerance = 1e-8
  )
  expect_equal(
    pred$`0.75q`,
    apply(response_samples, 1, quantile, 0.75),
    tolerance = 1e-8
  )
})


test_that("predict only requires maps for requested latent fields", {
  withr::local_seed(123)

  n <- 20
  w1 <- as.numeric(arima.sim(list(ar = 0.4), n = n))
  w2 <- cumsum(rnorm(n, sd = 0.15))
  y <- w1 + w2 + rnorm(n, sd = 0.2)

  fit <- ngme(
    y ~ 0 +
      f(1:n, model = ar1(), name = "ar1_field") +
      f(1:n, model = rw1(), name = "rw1_field"),
    data = data.frame(y = y),
    family = noise_normal(),
    control_opt = control_opt(
      iterations = 20,
      burnin = 1,
      n_batch = 2,
      n_parallel_chain = 2,
      n_min_batch = 1,
      verbose = FALSE,
      print_check_info = FALSE,
      store_traj = FALSE,
      seed = 123
    )
  )

  pred_mean <- predict(
    fit,
    map = list(ar1_field = seq_len(n)),
    type = "ar1_field",
    estimator = "mean",
    sampling_size = 12,
    burnin_size = 3,
    seed = 11
  )

  pred_samples <- predict(
    fit,
    map = list(ar1_field = seq_len(n)),
    type = "ar1_field",
    estimator = "mean",
    sampling_size = 12,
    burnin_size = 3,
    seed = 11,
    return_samples = TRUE
  )

  expect_length(pred_mean$mean, n)
  expect_length(pred_samples$mean, n)
  expect_equal(dim(attr(pred_samples, "samples")), c(n, 12))

  expect_error(
    predict(
      fit,
      map = list(ar1_field = seq_len(n)),
      type = "lp",
      estimator = "mean",
      sampling_size = 12,
      burnin_size = 3,
      seed = 11
    ),
    "rw1_field"
  )
})

test_that("tp-bv-matern operator structure and simulation", {
  withr::local_seed(42)

  time_len <- 30
  half_bv_len <- 10
  rho <- 0.6

  # specify indices for 2 parts
  time_idx <- rep(1:time_len, each = half_bv_len * 2)
  bv_idx <- rep(c(1:half_bv_len, 1:half_bv_len), time_len)

  # group to indicate two different fields for bivariate model
  group <- rep(rep(c("f1", "f2"), each = half_bv_len), time_len)

  mesh_t <- fmesher::fm_mesh_1d(1:time_len)
  mesh_s <- fmesher::fm_mesh_1d(1:half_bv_len)

  f_model <- f(
    map = list(time_idx, bv_idx),
    model = tp(
      first = ar1(
        mesh = mesh_t,
        rho = rho
      ),
      second = bv_matern(
        mesh = mesh_s,
        group = group,
        sub_models = list(
          f1 = matern(),
          f2 = matern()
        )
      )
    ),
    group = group,
    noise = noise_nig(
      mu = -2,
      sigma = 1,
      nu = 2
    )
  )

  op <- f_model$operator
  expect_equal(op$model, "tp")
  expect_equal(op$first$model, "ar1")
  expect_equal(op$second$model, "bv_matern")
  expect_equal(op$second$model_names, c("f1", "f2"))

  rho_from_theta <- ar1_th2a(op$first$theta_K)
  expect_equal(as.numeric(rho_from_theta), rho, tolerance = 1e-10)

  K_ar1 <- op$first$K
  expect_equal(as.numeric(K_ar1[1, 1]), sqrt(1 - rho^2), tolerance = 1e-10)
  expect_equal(as.numeric(K_ar1[2, 1]), -rho, tolerance = 1e-10)
  expect_equal(as.numeric(K_ar1[3, 2]), -rho, tolerance = 1e-10)
  expect_equal(as.numeric(K_ar1[2, 2]), 1, tolerance = 1e-10)

  expected_tp_K <- op$first$K %x% op$second$K
  expect_equal(as.matrix(op$K), as.matrix(expected_tp_K))

  expected_bv_size <- 2 * mesh_s$n
  expect_equal(nrow(op$second$K), expected_bv_size)
  expect_equal(ncol(op$second$K), expected_bv_size)
  expect_equal(nrow(op$K), time_len * expected_bv_size)

  sim <- simulate(f_model)
  expect_length(sim, 1)

  W <- sim[[1]]
  expect_length(W, length(time_idx))
  expect_true(is.numeric(W))
  expect_false(anyNA(W))
  expect_true(all(is.finite(W)))

  # Estimate rho from latent W (not A %*% W)
  W_latent <- as.numeric(attr(sim, "W_sim")[[1]])
  expect_length(W_latent, time_len * 2 * mesh_s$n)

  w_mat <- matrix(W_latent, nrow = time_len, byrow = TRUE)
  w_centered <- sweep(w_mat, 2, colMeans(w_mat), FUN = "-")
  x <- w_centered[1:(time_len - 1), , drop = FALSE]
  y <- w_centered[2:time_len, , drop = FALSE]

  rho_by_col <- colSums(x * y) / colSums(x * x)
  rho_by_col <- rho_by_col[is.finite(rho_by_col)]
  rho_hat <- median(rho_by_col, na.rm = TRUE)
  expect_true(is.finite(rho_hat))
  expect_true(rho_hat > 0.35 && rho_hat < 0.9)
})

test_that("f() accepts mesh in model", {
  mesh <- fmesher::fm_mesh_1d(1:10)
  # f_args$mesh <- temp_mesh # mesh is not in f_args
  #
  # print(paste("DEBUG: model class:", paste(class(model), collapse = ", ")))
  out <- f(map = 1:10, model = ar1(mesh = mesh), noise = noise_normal())
  expect_true(inherits(out, "ngme_model"))
  expect_equal(out$operator$model, "ar1")
})

test_that("f() accepts list of meshes in model for replicates", {
  mesh1 <- fmesher::fm_mesh_1d(1:10)
  mesh2 <- fmesher::fm_mesh_1d(1:10)
  mesh_list <- list(mesh1, mesh2)

  # Replicate with different meshes
  out <- f(
    map = 1:20, replicate = rep(1:2, each = 10),
    model = ar1(mesh = mesh_list),
    noise = noise_normal()
  )

  expect_true(inherits(out, "ngme_model"))
  # Check if operator list is created internally (not directly exposed in ngme_model usually, but we can check structure)
  # Actually f() returns ngme_model which has operator.
  # If replicates, operator might be block diagonalized?
  # Yes, f() block diagonalizes.
  expect_equal(length(out$operator$h), mesh1$n + mesh2$n)
})

test_that("f() builds mesh from map if not provided", {
  out <- f(map = 1:10, model = ar1(), noise = noise_normal())
  expect_true(inherits(out, "ngme_model"))
  # 1:10 creates mesh of size 10 (min to max)
  expect_equal(out$mesh$n, 10)
})

test_that("ou discretisation matches exact AR(1) form", {
  loc <- c(0, 1, 2.5, 3.5)
  theta <- 0.5

  op <- ou(mesh = loc, theta = theta)

  dt <- diff(loc)
  rho <- exp(-theta * dt)
  diag_main <- c(sqrt(1 - rho[1]^2), rep(1, length(loc) - 1))
  sub_diag <- -rho
  expected <- Matrix::bandSparse(
    n = length(loc),
    k = c(0, -1),
    diagonals = list(diag_main, sub_diag)
  )

  expect_equal(as.matrix(op$K), as.matrix(expected))
})

test_that("test tp", {
  n <- 10
  time <- sample(2001:2004, n, replace = TRUE)
  loc <- cbind(runif(n), runif(n)) * 10

  # Display the space-time observation structure
  data.frame(
    observation = 1:n,
    time = time,
    x = round(loc[, 1], 2),
    y = round(loc[, 2], 2)
  )

  # Create spatial mesh
  mesh_spatial <- fmesher::fm_mesh_2d(
    loc.domain = cbind(
      c(0, 10, 10, 0, 0), # x range: 0 to 10
      c(0, 0, 10, 10, 0) # y range: 0 to 10 (to cover all points)
    ),
    max.edge = c(1, 10),
    cutoff = 0.1,
    offset = c(0.5, 2) # Offset to ensure boundary coverage
  )

  tp_model <- f(
    map = list(time, loc),
    model = tp(
      first = ar1(mesh = time, rho = 0.5),
      second = matern(mesh = mesh_spatial, kappa = 2)
    )
  )
  tp_model
})

test_that("test tp-bv", {
  time_len <- 20
  half_bv_len <- 50

  # specify indices for 2 parts
  time_idx <- rep(1:time_len, each = half_bv_len * 2)
  bv_idx <- rep(c(1:half_bv_len, 1:half_bv_len), time_len)

  # group to indicate two different fields for bivariate model
  group <- rep(rep(c("f1", "f2"), each = half_bv_len), time_len)

  tp(
    first = ar1(mesh = time_idx),
    second = bv(
      mesh = bv_idx,
      sub_models = list(
        f1 = ar1(),
        f2 = ar1()
      )
    )
  )
})

test_that("test bivariate (bv)", {
  temp <- c(32, 33, 35.5, 36)
  year_temp <- c(2001, 2002, 2003, 2004)
  precip <- c(0.1, 0.2, 0.5, 1, 0.2)
  year_pre <- c(2001, 2002, 2003, 2004, 2005)

  # bind 2 fields in one vector, and make labels for them
  y <- c(temp, precip)
  year <- c(year_temp, year_pre)
  labels <- c(rep("temp", 4), rep("precip", 5)) # group is label for 2 fields

  x1 <- 1:9
  data <- data.frame(y, year, x1, labels)

  bv_model <- bv(
    year,
    theta = pi / 8,
    rho = 0.5,
    sub_models = list(A = ar1(), B = rw1())
  )

  bv_model
})

#' Show ngme noise types
#'
#' @return available types for noise
#' @export
ngme_noise_types <- function() {
    c("normal", "nig", "gal", "t", "skew_t", "normal_nig")
}

#' Show ngme model types
#'
#' @return available types for models
#' @export
ngme_model_types <- function() {
    c(
        "iid", "re",
        "ar1", "ar", "arma",
        "rw1", "rw2",
        "ou", "matern",
        "tp",
        "bv", "bv2", "bv_matern",
        "spacetime",
        "generic", "generic_ns"
    )
}

#' Show ngme priors
#'
#' @return available types of priors
#' @export
ngme_prior_types <- function() {
    c("none", "normal", "pc.sd", "half.cauchy", "inv.exponential")
}

# This script is used for fitting different matern models using Parana data
############################### Create INLA mesh
{
  library(INLA)
  library(splancs)
  library(lattice)
  library(ggplot2)
  library(grid)
  library(gridExtra)
  library(viridis)

  # read data
  data(PRprec)
  data(PRborder)

  str(PRprec)
  str(PRborder)

  # maps
  plot(PRborder[, 1], PRborder[, 2], type = "l")
  # plot the seashore
  lines(PRborder[1034:1078, 1], PRborder[1034:1078, 2], col = 2, lwd = 2)

  # create mesh
  coords <- as.matrix(PRprec[, 1:2])
  prdomain <- inla.nonconvex.hull(coords, -0.03, -0.05, resolution = c(100, 100))
  prmesh <- fmesher::fm_mesh_2d(boundary = prdomain, max.edge = c(0.45, 1), cutoff = 0.2)
  prmesh2 <- fmesher::fm_mesh_2d(coords, max.edge = c(0.45, 1), cutoff = 0.2)
  plot(prmesh)

  # monthly mean at each location
  colnames(PRprec)
  Y <- rowMeans(PRprec[, 3 + 1:31]) # January
  any(is.na(Y))

  ind <- !is.na(Y) # non-NA index
  sum(ind)
  Y <- Y[ind]

  coords <- as.matrix(PRprec[ind, 1:2])
  seaDist <- apply(spDists(coords, PRborder[1034:1078, ],
    longlat = TRUE
  ), 1, min)

  ggplot() +
    geom_point(aes(
      x = coords[, 1], y = coords[, 2],
      colour = Y
    ), size = 2, alpha = 1) +
    scale_color_gradientn(colours = viridis(100)) +
    geom_path(aes(x = PRborder[, 1], y = PRborder[, 2])) +
    geom_path(aes(x = PRborder[1034:1078, 1], y = PRborder[
      1034:1078,
      2
    ]), colour = "red")
}
############################### Matern model and predict location
{
  # matern model
  library(ngme2)
  matern_spde <- model_matern(
    loc = coords,
    mesh = prmesh
  )

  # make prediction at locations
  projgrid <- inla.mesh.projector(
    prmesh,
    xlim = range(PRborder[, 1]),
    ylim = range(PRborder[, 2]),
    dims = c(300, 200)
  )
  # get the points inside the boundary idx
  xy.in <- inout(projgrid$lattice$loc, cbind(PRborder[, 1], PRborder[, 2]))
  coord.prd <- projgrid$lattice$loc[xy.in, ]

  plot(coord.prd, type = "p", cex = 0.1)
  lines(PRborder)
  points(coords[, 1], coords[, 2], pch = 19, cex = 0.5, col = "red")

  dim(coord.prd)
  coord.prd.df <- data.frame(
    x1 = coord.prd[, 1],
    x2 = coord.prd[, 2]
  )
  coordinates(coord.prd.df) <- c("x1", "x2")

  # compute the distance to the shore
  seaDist.prd <- apply(spDists(coord.prd,
    PRborder[1034:1078, ],
    longlat = TRUE
  ), 1, min)
  coord.prd.df$seaDist <- seaDist.prd
  str(coord.prd.df)
  head(coord.prd.df)
}


################ case 1. spde_nig + normal measurement nosie
{
  out_nig_norm <- ngme(
    formula = Y ~ 1 +
      f(inla.group(seaDist), model = "rw1", noise = noise_normal(), name = "rw") +
      f(model = matern_spde, noise = noise_nig(), name = "spde1"),
    data = list(
      Y = Y
    ),
    family = noise_normal(),
    control = control_opt(
      estimation = T,
      iterations = 1000,
      n_min_batch = 4,
      n_batch = 10,
      std_lim = 0.1,
      n_parallel_chain = 4,
      print_check_info = TRUE
    ),
    # debug = TRUE,
    seed = 16
  )
  out_nig_norm
  traceplot(out_nig_norm, name = "rw")
  traceplot(out_nig_norm, name = "spde1")

  { # make prediction
    head(coord.prd.df@coords)
    preds <- predict(out_nig_norm, loc = list(
      rw = seaDist.prd,
      spde1 = coord.prd.df@coords
    ))

    pred_df <- data.frame(
      x1 = coord.prd[, 1],
      x2 = coord.prd[, 2],
      preds = preds
    )
    ggplot(pred_df, aes(x = x1, y = x2, fill = preds)) +
      geom_raster() +
      scale_fill_viridis()
  }
}

################ case 2. spde_nig + spde_normal + normal measurement nosie
{
  out_nig_norm_norm <- ngme(
    formula = Y ~ 1 +
      f(inla.group(seaDist), model = "rw1", noise = noise_normal(), name = "rw") +
      f(model = matern_spde, noise = noise_nig(), name = "spde1") +
      f(model = matern_spde, noise = noise_normal(), name = "spde2"),
    data = list(
      Y = Y
    ),
    family = noise_normal(),
    control = control_opt(
      estimation = T,
      iterations = 1000,
      n_min_batch = 4,
      n_batch = 10,
      std_lim = 0.1,
      n_parallel_chain = 4,
      print_check_info = TRUE
    ),
    # debug = TRUE,
    seed = 16
  )
  traceplot(out_nig_norm_norm_2, name = "rw")
  traceplot(out_nig_norm_norm_2, name = "spde1")

  { # make prediction
    str(coord.prd.df@coords)
    preds <- predict(out_nig_norm_norm_2, loc = list(
      rw = seaDist.prd,
      spde1 = coord.prd.df@coords,
      spde2 = coord.prd.df@coords
    ))

    pred_df <- data.frame(
      x1 = coord.prd[, 1],
      x2 = coord.prd[, 2],
      preds = preds
    )
    ggplot(pred_df, aes(x = x1, y = x2, fill = preds)) +
      geom_raster() +
      scale_fill_viridis()
  }
}

################ case 3. spde_nig + spde_normal + normal measurement nosie
{
  out_normnig_norm <- ngme(
    formula = Y ~ 1 +
      f(inla.group(seaDist), model = "rw1", noise = noise_normal(), name = "rw") +
      f(model = matern_spde, noise = noise_normal_nig(), name = "spde1"),
    data = list(
      Y = Y
    ),
    family = noise_normal(),
    control = control_opt(
      estimation = T,
      iterations = 1000,
      n_min_batch = 4,
      n_batch = 10,
      std_lim = 0.1,
      n_parallel_chain = 4,
      print_check_info = TRUE
    ),
    # debug = TRUE,
    seed = 16
  )
  traceplot(out_normnig_norm, name = "rw")
  traceplot(out_normnig_norm, name = "spde1")

  { # make prediction
    str(coord.prd.df@coords)
    preds <- predict(out_normnig_norm, loc = list(
      rw = seaDist.prd,
      spde1 = coord.prd.df@coords
    ))

    pred_df <- data.frame(
      x1 = coord.prd[, 1],
      x2 = coord.prd[, 2],
      preds = preds
    )
    ggplot(pred_df, aes(x = x1, y = x2, fill = preds)) +
      geom_raster() +
      scale_fill_viridis()
  }
}

# cross validation
cross_validation(out_nig_norm, k = 5, N = 500, print = TRUE)
cross_validation(out_nig_norm_norm, k = 5, N = 500, print = TRUE)
cross_validation(out_normnig_norm, k = 5, N = 500, print = TRUE)

library(fmesher)
library(rSPDE)
library(ngme2)

m <- 200
loc_2d_mesh <- matrix(runif(m * 2), m, 2)
mesh_2d <- fm_mesh_2d(
  loc = loc_2d_mesh,
  cutoff = 0.05,
  offset = c(0.1, 0.4),
  max.edge = c(0.05, 0.5)
)

plot(mesh_2d, main = "")
points(loc_2d_mesh[, 1], loc_2d_mesh[, 2])
nu <- 0.8
sigma <- 1.3
range <- 0.15
op <- matern.operators(
  range = range, sigma = sigma,
  nu = nu, m = 2, mesh = mesh_2d,
  parameterization = "matern"
)

n.rep <- 1
u <- simulate(op, nsim = n.rep)
A <- fm_basis(
  x = mesh_2d,
  loc = loc_2d_mesh
)
sigma.e <- 0.1
Y <- A %*% u + matrix(rnorm(m * n.rep), ncol = n.rep) * sigma.e

library(viridis)
library(ggplot2)
proj <- fm_evaluator(mesh_2d, dims = c(70, 70))

df_field <- data.frame(
  x = proj$lattice$loc[, 1],
  y = proj$lattice$loc[, 2],
  field = as.vector(fm_evaluate(proj,
    field = as.vector(u[, 1])
  )),
  type = "field"
)

df_loc <- data.frame(
  x = loc_2d_mesh[, 1],
  y = loc_2d_mesh[, 2],
  field = as.vector(Y[, 1]),
  type = "locations"
)
df_plot <- rbind(df_field, df_loc)

ggplot(df_plot) +
  aes(x = x, y = y, fill = field) +
  facet_wrap(~type) +
  xlim(0, 1) +
  ylim(0, 1) +
  geom_raster(data = df_field) +
  geom_point(
    data = df_loc, aes(colour = field),
    show.legend = FALSE
  ) +
  scale_fill_viridis() +
  scale_colour_viridis()

op_obj <- matern.operators(
  m = 1,
  type = "operator", mesh = mesh_2d
)

y_vec <- as.vector(Y)
repl <- rep(1:n.rep, each = m)
df_data_2d <- data.frame(
  y = y_vec, x_coord = loc_2d_mesh[, 1],
  y_coord = loc_2d_mesh[, 2]
)

# fit_2d <- rspde_lme(y ~ -1, model = op_obj,
#           data = df_data_2d, repl = repl,
#           loc = c("x_coord", "y_coord"),
#           parallel = TRUE)
# glance(fit_2d)
# print(data.frame(
#   sigma = c(sigma, fit_2d$alt_par_coeff$coeff["sigma"]),
#   range = c(range, fit_2d$alt_par_coeff$coeff["range"]),
#   nu = c(nu, fit_2d$alt_par_coeff$coeff["nu"]),
#   row.names = c("Truth", "Estimates")
# ))

# # Total time
# print(fit_2d$fitting_time)

fit <- ngme(
  y ~ 0 + f(
    ~ x_coord + y_coord,
    model = "matern",
    name = "spde",
    kappa = 20,
    mesh = mesh_2d,
    fix_alpha = FALSE,
    rational_order = 2,
    noise = noise_normal(sigma = 5)
  ),
  data = df_data_2d,
  replicate = repl,
  control_opt = control_opt(
    estimation = TRUE,
    iterations = 300,
    optimizer = adam(),
    n_batch = 10,
    rao_blackwellization = F,
    n_parallel_chain = 4,
    print_check_info = F,
    verbose = T,
    std_lim = 0.01,
    start_sd = 0.01
  )
)
fit
traceplot(fit, "spde")
traceplot(fit)

#############################################################################################
#' @name gig
#' @title  The Generalised Inverse-Gaussian (GIG) Distribution
#' @aliases dgig rgig pgig qgig
#' @description Density, distribution function, quantile function and
#' random generation for the generalised inverse-Gaussian distribution
#'  with parameters \code{p}, \code{a} and \code{b}.
#' @param x,q vector of quantiles.
#' @param prob vector of probabilities.
#' @param n, number of observations.
#' @param a,b parameters \code{a} and \code{b}. Must be positive.
#' @param p parameter \code{p}.
#' @param log,log.p logical; if \code{TRUE}, probabilities/densities \eqn{p} are
#' returned as \eqn{log(p)}.
#' @param lower.tail logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
#' otherwise, \eqn{P[X>x]}.
#' @param seed Seed for the random generation.
#' @return
#' dgig gives the density, pgig gives the distribution function,
#' qgig gives the quantile function, and rgig generates random deviates.
#'
#' Invalid arguments will result in return value NaN, with a warning.
#'
#' The length of the result is determined by \code{n} for rgig.
#' @details The generalised inverse-Gaussian distribution has density given
#' by
#'  \deqn{f(x; p, a, b) = ((a/b)^{p/2})/(2K_p(\sqrt{ab})) x^{p-1} \exp\{-(a/2)x - (b/2)/x\},}
#'  where \eqn{K_p} is modified Bessel function of the second kind of order \eqn{p},
#'  \eqn{x>0}, \eqn{a,b>0} and \eqn{p\in\mathbb{R}}.
#'  See Jørgensen (1982) for further details.
#'
#' @references
#' Jørgensen, Bent (1982). Statistical Properties of the Generalized
#' Inverse Gaussian Distribution. Lecture Notes in Statistics. 9.
#' New York–Berlin: Springer-Verlag. \doi{10.1007/978-1-4612-5698-4}
#'
#' @seealso
#' \code{\link{dnig}}, \code{\link{dig}}, \code{\link{digam}}
#' @examples
#' rgig(20, p = 1, a = 1, b = 1)
#' pgig(0.4, p = 1, a = 1, b = 1)
#' qgig(0.8, p = 1, a = 1, b = 1)
#' plot(function(x){dgig(x, p = 1, a = 1, b = 1)}, main =
#' "Generalised inverse-Gaussian density", ylab = "Probability density",
#' xlim = c(0,10))
#' @rdname gig
#' @export
dgig <- function(x, p, a ,b, log=FALSE){
  if(missing(p)){
    stop('argument "p" missing, with no default')
  }
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(p)){
    if(length(a) == 1){
      a <- rep(a, length(p))
    } else if(length(p)==1){
      p <- rep(p, length(a))
    } else {
      stop("a and p are vectors of different lengths")
    }
  }
  if (length(a) != length(b)){
    if(length(a) == 1){
      a <- rep(a, length(b))
      if(length(p)==1){
        p <- rep(p, length(b))
      }
    } else if(length(b) == 1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  n = length(x)
  if(n < length(p)){
    p_new = p[1:n]
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(p)){
    quot <- n%/%length(p)
    rem <- n%%length(p)
    p_new <- rep(p,quot)
    if(rem > 0){
      p_new  <- c(p_new, p[1:rem])
    }
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    p_new <- p
    a_new <- a
    b_new <- b
  }

  densgig <- sapply(1:n, function(i){
    if (b_new[i] == 0) {
      return (dgamma(x[i], shape = p_new[i], rate = a_new[i]/2, log = log))
    }
    if (a_new[i] == 0) {
      return (digam(x[i], a =
      -p_new[i], b = b_new[i]/2, log = log))
    }

    if(x[i] <= 0){
      dens <- ifelse(log, -Inf,0)
      return(dens)
    } else{
      z <- sqrt(a_new[i] * b_new[i])
      logcoef <- (p_new[i]/2) * (log(a_new[i]) - log(b_new[i])) - log(2)
      bk_scaled <- besselK(z, nu = p_new[i], expon.scaled = TRUE)
      if (any(is.na(bk_scaled))) stop("besselK returned NA; parameters out of range for this routine.")
      logK <- ifelse(bk_scaled > 0, log(bk_scaled) - z, -Inf)
      lx <- (p_new[i] - 1) * log(x[i]) - (a_new[i] * x[i] + b_new[i] / x[i]) / 2
      logf <- logcoef - logK + lx
      dens <- ifelse(log, logf, exp(logf))
      return(dens)
    }
  })
  return(densgig)
}

#' @rdname gig
#' @export
rgig <- function (n, p, a, b, seed = 0)
{
  if(missing(p)){
    stop('argument "p" missing, with no default')
  }
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(p)){
    if(length(a) == 1){
      a <- rep(a, length(p))
    } else if(length(p)==1){
      p <- rep(p, length(a))
    } else {
      stop("a and p are vectors of different lengths")
    }
  }
  if (length(a) != length(b)){
    if(length(a) == 1){
      a <- rep(a, length(b))
      if(length(p)==1){
        p <- rep(p, length(b))
      }
    } else if(length(b) == 1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  if(n < length(p)){
    p_new = p[1:n]
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(p)){
    quot <- n%/%length(p)
    rem <- n%%length(p)
    p_new <- rep(p,quot)
    if(rem > 0){
      p_new  <- c(p_new, p[1:rem])
    }
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    p_new <- p
    a_new <- a
    b_new <- b
  }

  if (all(a==0)) {
    stopifnot("p should be negative when a is zero" =
                all(p_new < 0))
    return (rigam(n, a = -p_new, b = b_new/2))
  }
  
  if (all(b==0)) {
    stopifnot("p should be positive when b is zero" =
                all(p_new > 0))
    return (rgamma(n, shape = p_new, rate = a_new/2))
  }

  return(rGIG_cpp(p_new, a_new, b_new, seed))
}

#' @rdname gig
#' @export
pgig <- function(q, p, a, b, lower.tail = TRUE, log.p = FALSE){
  if(missing(p)){
    stop('argument "p" missing, with no default')
  }
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(p)){
    if(length(a) == 1){
      a <- rep(a, length(p))
    } else if(length(p)==1){
      p <- rep(p, length(a))
    } else {
      stop("a and p are vectors of different lengths")
    }
  }
  if (length(a) != length(b)){
    if(length(a) == 1){
      a <- rep(a, length(b))
      if(length(p)==1){
        p <- rep(p, length(b))
      }
    } else if(length(b) == 1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  n = length(q)
  if(n < length(p)){
    p_new = p[1:n]
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(p)){
    quot <- n%/%length(p)
    rem <- n%%length(p)
    p_new <- rep(p,quot)
    if(rem > 0){
      p_new  <- c(p_new, p[1:rem])
    }
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    p_new <- p
    a_new <- a
    b_new <- b
  }
  prob_gig <- sapply(1:n, function(i){
    if(q[i]<=0){
      p_gig <- ifelse(lower.tail, 0, 1)
      p_gig <- ifelse(log.p, log(p_gig), p_gig)
      return(p_gig)
    } else{
      p_gig <- stats::integrate(dgig,lower = 0, upper = q[i],
                                p=p_new[i],a=a_new[i],b=b_new[i])$value
      if(p_gig < 10^{-5} & q[i] > sqrt(b_new[i]/a_new[i])*
         besselK(sqrt(a_new[i]*b_new[i]),p_new[i]+1)/
         besselK(sqrt(a_new[i]*b_new[i]),p_new[i])){
        p_gig = 1
      }
      p_gig <- ifelse(lower.tail, p_gig, 1-p_gig)
      p_gig <- ifelse(log.p, log(p_gig), p_gig)
      return(p_gig)
    }
  })
  return(prob_gig)
}

#' @rdname gig
#' @export
qgig <- function(prob, p, a, b, lower.tail = TRUE, log.p = FALSE){
  if(missing(p)){
    stop('argument "p" missing, with no default')
  }
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(p)){
    if(length(a) == 1){
      a <- rep(a, length(p))
    } else if(length(p)==1){
      p <- rep(p, length(a))
    } else {
      stop("a and p are vectors of different lengths")
    }
  }
  if (length(a) != length(b)){
    if(length(a) == 1){
      a <- rep(a, length(b))
      if(length(p)==1){
        p <- rep(p, length(b))
      }
    } else if(length(b) == 1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  n = length(prob)
  if(n < length(p)){
    p_new = p[1:n]
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(p)){
    quot <- n%/%length(p)
    rem <- n%%length(p)
    p_new <- rep(p,quot)
    if(rem > 0){
      p_new  <- c(p_new, p[1:rem])
    }
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    p_new <- p
    a_new <- a
    b_new <- b
  }
  quant_gig <- sapply(1:n, function(i){
    if (b_new[i] == 0){
      return (qgamma(prob[i], shape = p_new[i], rate = a_new[i]/2, lower.tail = lower.tail, log.p = log.p))
    }
    if (a_new[i] == 0){
      return (qigam(prob[i], a = -p_new[i], b = b_new[i]/2, lower.tail = lower.tail, log.p = log.p))
    }

    if(a_new[i]<1 & b_new[i]<1){
      up_bd <- max(1/a_new[i] * 1/b_new[i], 100)
    } else if (a[i]>=1){
      up_bd <- max(a_new[i]/b_new[i], 100)
    } else {
      up_bd <- 100
    }
    if(prob[i] < 0 | prob[i] > 1){
      warn_qgig <- TRUE
      return(NaN)} else{
        while(pgig(up_bd, p=p_new[i],a=a_new[i],
                b=b_new[i],lower.tail=lower.tail) < prob[i]){
          up_bd = up_bd + 10
        }

        q_gig <- stats::uniroot(function(y) {
          pgig(y, p=p_new[i],a=a_new[i],b=b_new[i],lower.tail=lower.tail) - prob[i]
        },lower = 0, upper = up_bd)$root
        q_gig <- ifelse(log.p, log(q_gig), q_gig)
        return(q_gig)
      }
  }
  )
  if(any(is.nan(quant_gig))){
    warning("NaNs produced", call. = TRUE, domain = "R")
  }
  return(quant_gig)
}

#############################################################################################
#' @name ig
#' @title  The Inverse-Gaussian (IG) Distribution
#' @aliases dig rig pig qig
#' @description Density, distribution function, quantile function and
#' random generation for the inverse-Gaussian distribution
#'  with parameters \code{a} and \code{b}.
#' @param x,q vector of quantiles.
#' @param p vector of probabilities.
#' @param n, number of observations.
#' @param a,b parameters \code{a} and \code{b}. Must be positive.
#' @param log,log.p logical; if \code{TRUE}, probabilities/densities \eqn{p} are
#' returned as \eqn{log(p)}.
#' @param lower.tail logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
#' otherwise, \eqn{P[X>x]}.
#' @param seed Seed for the random generation.
#' @return
#' dig gives the density, pig gives the distribution function,
#' qig gives the quantile function, and rig generates random deviates.
#'
#' Invalid arguments will result in return value NaN, with a warning.
#'
#' The length of the result is determined by \code{n} for rig.
#' @details The inverse-Gaussian distribution has density given
#' by
#'  \deqn{f(x; a, b) = \frac{\sqrt{b}}{\sqrt{2\pi x^3}}\exp(
#'  -\frac{a}{2}x -\frac{b}{2x} + \sqrt{ab}),}
#'  where \eqn{x>0} and \eqn{a,b>0}. In this parameterization,
#'  \eqn{E(X) = \sqrt{b}/\sqrt{a}}. See Tweedie (1957a, 1957b) for
#'  further details.
#'
#' @references
#' Tweedie, M. C. K. (1957a). "Statistical Properties of Inverse Gaussian Distributions I". Annals of Mathematical Statistics. 28 (2): 362–377. \doi{10.1214/aoms/1177706964}
#'
#' Tweedie, M. C. K. (1957b). "Statistical Properties of Inverse Gaussian Distributions II". Annals of Mathematical Statistics. 28 (3): 696–705. \doi{10.1214/aoms/1177706881}
#'
#' @seealso
#' \code{\link{dnig}}, \code{\link{dgig}}, \code{\link{digam}}
#' @examples
#' rig(100, a = 1, b = 1)
#' pig(0.4, a = 1, b = 1)
#' qig(0.8, a = 1, b = 1)
#' plot(function(x){dig(x, a = 1, b = 1)}, main =
#' "Inverse-Gaussian density", ylab = "Probability density",
#' xlim = c(0,10))
#' @rdname ig
#' @export
dig <- function(x, a ,b, log=FALSE){
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(b)){
    if(length(a)==1){
      a <- rep(a, length(b))
    } else if(length(b)==1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  n = length(x)
  if(n < length(a)){
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(a)){
    quot <- n%/%length(a)
    rem <- n%%length(a)
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    a_new <- a
    b_new <- b
  }

  densig <- sapply(1:n, function(i){
    if(x[i] <= 0){
      dens <- ifelse(log, -Inf,0)
      return(dens)
    } else{
      l = (log(b_new[i]) - log(2*pi) - 3 *log(x[i]) )
      l = l - a_new[i]*x[i] - b_new[i]/x[i] + 2*sqrt(a_new[i]*b_new[i])
      l = 0.5 * l
      dens <- ifelse(log, l, exp(l))
      return(dens)
    }
  })
  return(densig)
}

#' @rdname ig
#' @export
rig <- function (n, a, b, seed = 0)
{
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(b)){
    if(length(a)==1){
      a <- rep(a, length(b))
    } else if(length(b)==1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  if(n < length(a)){
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(a)){
    quot <- n%/%length(a)
    rem <- n%%length(a)
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    a_new <- a
    b_new <- b
  }
  p_new <- rep(-0.5, length(a_new))
  return(rGIG_cpp(p_new, a_new, b_new, seed))
}

#' @rdname ig
#' @export
pig <- function(q, a, b, lower.tail = TRUE, log.p = FALSE){
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
      mu <- sqrt(b/a)
      p_ig <- stats::pnorm(sqrt(b/q)*(q/mu-1)) + exp(2*b/mu)*stats::pnorm(-sqrt(b/q)*(q/mu+1))
      if(!lower.tail){
        p_ig <- 1 - p_ig
      }

      if(log.p){
          p_ig <- log(p_ig)
      }
      return(p_ig)
}

#' @rdname ig
#' @export
qig <- function(p, a, b, lower.tail = TRUE, log.p = FALSE){
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(b)){
    if(length(a)==1){
      a <- rep(a, length(b))
    } else if(length(b)==1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  n = length(p)
  if(n < length(a)){
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(a)){
    quot <- n%/%length(a)
    rem <- n%%length(a)
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    a_new <- a
    b_new <- b
  }
  quant_ig <- sapply(1:n, function(i){
    if(a_new[i]<1 & b_new[i]<1){
      up_bd <- max(1/a_new[i] * 1/b_new[i], 100)
    } else if (a[i]>=1){
      up_bd <- max(a_new[i]/b_new[i], 100)
    } else {
      up_bd <- 100
    }
    if(p[i] < 0 | p[i] > 1){
      warn_qig <- TRUE
      return(NaN)} else{
        q_ig <- stats::uniroot(function(y) {
          pig(y, a=a_new[i],b=b_new[i],lower.tail=lower.tail) - p[i]
        },lower = 0, upper = up_bd)$root
        q_ig <- ifelse(log.p, log(q_ig), q_ig)
        return(q_ig)
      }
  }
  )
  if(any(is.nan(quant_ig))){
    warning("NaNs produced", call. = TRUE, domain = "R")
  }
  return(quant_ig)
}

#############################################################################################
#' @name igam
#' @title  The Inverse-Gamma (IGam) Distribution
#' @aliases digam rigam pigam qigam
#' @description Density, distribution function, quantile function and
#' random generation for the inverse-Gamma distribution
#'  with parameters \code{a} and \code{b}.
#' @param x,q vector of quantiles.
#' @param p vector of probabilities.
#' @param n, number of observations.
#' @param a,b parameters \code{a} and \code{b}. Must be positive.
#' @param log,log.p logical; if \code{TRUE}, probabilities/densities \eqn{p} are
#' returned as \eqn{log(p)}.
#' @param lower.tail logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
#' otherwise, \eqn{P[X>x]}.
#' @return
#' digam gives the density, pigam gives the distribution function,
#' qigam gives the quantile function, and rigam generates random deviates.
#'
#' Invalid arguments will result in return value NaN, with a warning.
#'
#' The length of the result is determined by \code{n} for rig.
#' @details The inverse-Gamma distribution has density given
#' by
#'  \deqn{f(x; a, b) = \frac{b^a}{\Gamma(a)}x^{-a-1}\exp(
#'  -\frac{b}{x}),}
#'  where \eqn{x>0} and \eqn{a,b>0}.
#'
#' @seealso
#' \code{\link{dnig}}, \code{\link{dgig}}
#' @examples
#' rigam(100, a = 1, b = 1)
#' pigam(0.4, a = 1, b = 1)
#' qigam(0.8, a = 1, b = 1)
#' plot(function(x){digam(x, a = 1, b = 1)}, main =
#' "Inverse-Gamma density", ylab = "Probability density",
#' xlim = c(0,10))
#' @rdname igam
#' @export
digam <- function(x, a ,b, log=FALSE){
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  if (length(a) != length(b)){
    if(length(a)==1){
      a <- rep(a, length(b))
    } else if(length(b)==1){
      b <- rep(b, length(a))
    } else{
      stop("a and b are vectors of different lengths")
    }
  }
  if (min(a) < 0)
    stop("vector a must be  positive")
  if (min(b) < 0)
    stop("vector b must be  positive")
  n = length(x)
  if(n < length(a)){
    a_new = a[1:n]
    b_new = b[1:n]
  } else if(n>length(a)){
    quot <- n%/%length(a)
    rem <- n%%length(a)
    a_new <- rep(a,quot)
    if(rem > 0){
      a_new <- c(a_new,a[1:rem])
    }
    b_new <- rep(b,quot)
    if(rem > 0){
      b_new <- c(b_new, b[1:rem])
    }
  } else{
    a_new <- a
    b_new <- b
  }

  densigam <- sapply(1:n, function(i){
    if(x[i] <= 0){
      dens <- ifelse(log, -Inf,0)
      return(dens)
    } else{
      l = a_new[i]*log(b_new[i]) - lgamma(a_new[i]) -
        (a_new[i]+1)*log(x[i]) - b_new[i]/x[i]
      dens <- ifelse(log, l, exp(l))
      return(dens)
    }
  })
  return(densigam)
}

#' @rdname igam
#' @export
rigam <- function (n, a, b)
{
  1/stats::rgamma(n,shape=a, rate = b)
}

#' @rdname igam
#' @export
pigam <- function(q, a, b, lower.tail = TRUE, log.p = FALSE){
  if(missing(a)){
    stop('argument "a" missing, with no default')
  }
  if(missing(b)){
    stop('argument "b" missing, with no default')
  }
  p_igam <- 1 - stats::pgamma(1/q, shape = a, rate = b)
  return(p_igam)
}

#' @rdname igam
#' @export
qigam <- function(p, a, b, lower.tail = TRUE, log.p = FALSE){
 1/stats::qgamma(1-p,shape = a, rate = b, lower.tail = lower.tail, log.p = log.p)
}

#############################################################################################
#' @name nig
#' @title  The Normal Inverse-Gaussian (NIG) Distribution
#' @aliases dnig rnig pnig qnig
#' @description Density, distribution function, quantile function and
#' random generation for the normal inverse-Gaussian distribution
#'  with parameters \code{p}, \code{a} and \code{b}.
#' @param x,q vector of quantiles.
#' @param p vector of probabilities.
#' @param n, number of observations.
#' @param delta A numeric value for the location parameter.
#' @param mu    A numeric value for the shift parameter.
#' @param nu    A numeric value for the shape parameter.
#' @param sigma A numeric value for the scaling parameter.
#' @param h     A numeric value for the additional parameter, see details.
#' @param log,log.p logical; if \code{TRUE}, probabilities/densities \eqn{p} are
#' returned as \eqn{log(p)}.
#' @param lower.tail logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
#' otherwise, \eqn{P[X>x]}.
#' @param seed Seed for the random generation.
#' @return
#' dnig gives the density, pnig gives the distribution function,
#' qnig gives the quantile function, and rnig generates random deviates.
#'
#' Invalid arguments will result in return value NaN, with a warning.
#'
#' The length of the result is determined by \code{n} for rnig.
#' @details The normal inverse-Gaussian distribution has density given
#' by
#'  \deqn{f(x; \delta, \mu, \sigma, \nu) =
#'  \frac{e^{\nu+\mu(x-\delta)/\sigma^2}\sqrt{\nu\mu^2/\sigma^2+\nu^2}}{\pi\sqrt{\nu\sigma^2+(x-\delta)^2}}
#'  K_1(\sqrt{(\nu\sigma^2+(x-\delta)^2)(\mu^2/\sigma^4+\nu/\sigma^2)}),}
#'  where \eqn{K_p} is modified Bessel function of the second kind of order \eqn{p},
#'  \eqn{x>0}, \eqn{\nu>0} and \eqn{\mu,\delta, \sigma\in\mathbb{R}}.
#'  See Barndorff-Nielsen (1977, 1978 and 1997) for further details.
#'
#' The additional parameter h is used when \deqn{V\sim IG(\nu,\nu h^{2})}.
#' By the infinite divisibility, \deqn{\frac{1}{h} V \sim IG(\nu h, \nu h)}.
#' Then \deqn{\delta+\mu V + \sigma \sqrt{V} Z} has the distribution of \deqn{NIG(\delta=-\mu h,\mu= \mu h, \sigma=\sigma \sqrt{h}, \nu=\nu h).}
#'
#' @references
#'  Barndorff-Nielsen, O. (1977) Exponentially decreasing distributions for the logarithm of particle size. Proceedings of the Royal Society of London.
#'
#'  Series A, Mathematical and Physical Sciences. The Royal Society. 353, 401–409. \doi{10.1098/rspa.1977.0041}
#'
#'  Barndorff-Nielsen, O. (1978) Hyperbolic Distributions and Distributions on Hyperbolae, Scandinavian Journal of Statistics. 5, 151–157.
#'
#'  Barndorff-Nielsen, O. (1997) Normal Inverse Gaussian Distributions and Stochastic Volatility Modelling, Scandinavian Journal of Statistics. 24, 1-13. \doi{10.1111/1467-9469.00045}
#'
#' @seealso
#' \code{\link{dgig}}, \code{\link{dig}}, \code{\link{digam}}
#' @examples
#' rnig(100, delta = 0, mu = 5, sigma = 1, nu = 1)
#' pnig(0.4, delta = 0, mu = 5, sigma = 1, nu = 1)
#' qnig(0.8, delta = 0, mu = 5, sigma = 1, nu = 1)
#' plot(function(x){dnig(x, delta = 0, mu = 5, sigma = 1, nu = 1)}, main =
#' "Normal inverse-Gaussian density", ylab = "Probability density",
#' xlim = c(0,10))
#' @rdname nig
#' @export
dnig <- function(x, delta, mu, nu, sigma, h=NULL, log=FALSE){
  if (is.numeric(h)) {
    delta = h * delta
    mu = mu * h
    sigma = sigma * sqrt(h)
    nu = nu * h
  }
  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lenghts")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
      } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
      } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  n = length(x)
  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }

  densnig <- sapply(1:n, function(i){
    c0 <- sqrt(nu_new[i]) * sqrt( mu_new[i]^2/sigma_new[i]^2 + nu_new[i]) / pi
    l <- nu_new[i] + mu_new[i] * (x[i] - delta_new[i]) /sigma_new[i]^2
    coeff <- sqrt( nu_new[i] * sigma_new[i]^2 + (x[i] - delta_new[i])^2)
    l <- l + log(c0) -  log(coeff)
    l <- l + log(besselK(coeff  * sqrt(mu_new[i]^2/sigma_new[i]^4
                                        + nu_new[i]/sigma_new[i]^2),
                          -1,TRUE)) -coeff  * sqrt(mu_new[i]^2/sigma_new[i]^4 +
                                                    nu_new[i]/sigma_new[i]^2)
    dens <- ifelse(log, l, exp(l))
    return(dens)
  })
  return(densnig)
}

#' @rdname nig
#' @export
rnig <- function (n, delta, mu, nu, sigma, h=NULL, seed = 0)
{
  if (is.numeric(h)) {
    delta = h * delta
    mu = mu * h
    sigma = sigma * sqrt(h)
    nu = nu * h
  }
  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lenghts")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
    } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
    } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }

  V = as.vector(rig(n, nu, nu, seed = seed))
  res <- delta_new + mu_new*V + sigma_new*sqrt(V)*stats::rnorm(n)
  attr(res, "V") <- V
  return(res)
}

#' @rdname nig
#' @export
pnig <- function(q, delta, mu, nu, sigma, h=NULL, lower.tail = TRUE, log.p = FALSE){
  if (is.numeric(h)) {
    delta = h * delta
    mu = mu * h
    sigma = sigma * sqrt(h)
    nu = nu * h
  }

  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lenghts")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
    } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
    } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  n = length(q)
  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }

  prob_nig <- sapply(1:n, function(i){
    if(q[i]<=0){
      p_nig <- ifelse(lower.tail, 0, 1)
      p_nig <- ifelse(log.p, log(p_nig), p_nig)
      return(p_nig)
    } else{
      p_nig <- stats::integrate(dnig,lower = 0, upper = q[i],
                                delta=delta_new[i],mu=mu_new[i],
                                nu=nu_new[i],sigma=sigma_new[i])$value
      if(p_nig < 10^{-5} & q[i] > 200){
        p_nig = 1
      }
      p_nig <- ifelse(lower.tail, p_nig, 1-p_nig)
      p_nig <- ifelse(log.p, log(p_nig), p_nig)
      return(p_nig)
    }
  })
  return(prob_nig)
}

#' @rdname nig
#' @export
qnig <- function(p, delta, mu, nu, sigma, h=NULL, lower.tail = TRUE, log.p = FALSE){
  if (is.numeric(h)) {
    delta = h * delta
    mu = mu * h
    sigma = sigma * sqrt(h)
    nu = nu * h
  }
  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lenghts")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
    } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
    } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  n = length(p)
  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }
  quant_nig <- sapply(1:n, function(i){
    if(nu_new[i]<1){
      up_bd <- max(1/nu_new[i]^2, 1000)
    } else {
      up_bd <- 1000
    }
    if(p[i] < 0 | p[i] > 1){
      warn_qnig <- TRUE
      return(NaN)} else{
        while(pnig(up_bd, delta=delta_new[i],mu=mu_new[i],
                   nu=nu_new[i],sigma=sigma_new[i],
                   lower.tail=lower.tail) < p[i]){
          up_bd = up_bd + 1000
        }
        q_nig <- stats::uniroot(function(y) {
          pnig(y, delta=delta_new[i],mu=mu_new[i],
               nu=nu_new[i],sigma=sigma_new[i],
               lower.tail=lower.tail) - p[i]
        },lower = 0, upper = up_bd)$root
        q_nig <- ifelse(log.p, log(q_nig), q_nig)
        return(q_nig)
      }
  }
  )
  if(any(is.nan(quant_nig))){
    warning("NaNs produced", call. = TRUE, domain = "R")
  }
  return(quant_nig)
}

#!/usr/bin/env Rscript

# Demo: verify analytical Hessian of log f(W|V) w.r.t. theta_K
# Model: K(a,b) = exp(a) M_a + 3 b^2 M_b
#
# Notation (matching C++ implementation and derivation):
# - A = diag(1 / (sigma^2 * V))
# - m = mu * (V - h)
# - e = W - K^{-1} m
# - KWm = K W - m
#
# Hessian formula for entries H_ij with i,j in {a,b}:
#   H_ij = - tr(K^{-1} K_j K^{-1} K_i) + tr(K^{-1} K_ij)
#          - [ (K_j W)^T A K_i e + (K W - m)^T A K_ij e
#              + (K W - m)^T A K_i K^{-1} K_j K^{-1} m ]
#
suppressPackageStartupMessages({
  # No extra packages strictly needed; keep it base R compatible
})

make_spd <- function(n, jitter = 1e-1) {
  M <- matrix(rnorm(n * n), n, n)
  S <- crossprod(M)
  S <- 0.5 * (S + t(S)) + diag(jitter, n)
  S
}

ell_logf <- function(par, env) {
  a <- par[1]; b <- par[2]
  with(env, {
    K <- exp(a) * Ma + 3 * b^2 * Mb
    # logdet(K)
    # Use base::determinant for robust log |K|
    ld <- as.numeric(determinant(K, logarithm = TRUE)$modulus)
    KWm <- K %*% W - m
    qf <- sum((KWm * A_diag) * KWm) # (KWm)^T A (KWm), A is diagonal
    ld - 0.5 * qf
  })
}

analytic_hessian <- function(par, env) {
  a <- par[1]; b <- par[2]
  with(env, {
    K   <- exp(a) * Ma + 3 * b^2 * Mb
    Ka  <- exp(a) * Ma
    Kb  <- 6 * b * Mb
    Kaa <- exp(a) * Ma
    Kab <- 0 * Ma
    Kbb <- 6 * Mb

    # solvers
    Kinvm <- solve(K, m)
    e     <- W - Kinvm
    KWm   <- K %*% W - m

    # Helper trace(K^{-1} X)
    tr_Kinv_X <- function(X) {
      # tr(K^{-1} X) = sum_i e_i^T K^{-1} X e_i ~ exact via solve of each column (small n)
      # For small n this is fine; for large n use randomized trace.
      sum(diag(solve(K, X)))
    }

    # - tr(K^{-1} K_j K^{-1} K_i)
    tr_Kinv_Kj_Kinv_Ki <- function(Ki, Kj) {
      # compute tr(K^{-1} Kj K^{-1} Ki) exactly (small n)
      X <- solve(K, Kj)
      Y <- solve(K, Ki)
      sum(diag(X %*% Y))
    }

    # Common pieces
    A_dot <- A_diag

    # Terms builder
    Hij_fun <- function(Ki, Kj, Kij) {
      term_tr  <- - tr_Kinv_Kj_Kinv_Ki(Ki, Kj) + tr_Kinv_X(Kij)
      term1    <- sum((Kj %*% W) * A_dot * (Ki %*% e))
      term2    <- sum(KWm * A_dot * (Kij %*% e))
      yj1      <- Kj %*% Kinvm
      yj2      <- solve(K, yj1)
      yi3      <- Ki %*% yj2
      term3    <- sum(KWm * A_dot * yi3)
      term_tr - (term1 + term2 + term3)
    }

    H <- matrix(0, 2, 2)
    # (a,a)
    H[1,1] <- Hij_fun(Ka, Ka, Kaa)
    # (a,b) and (b,a)
    H[1,2] <- Hij_fun(Ka, Kb, Kab)
    H[2,1] <- H[1,2]
    # (b,b)
    H[2,2] <- Hij_fun(Kb, Kb, Kbb)
    H
  })
}

numeric_hessian <- function(par, fn, env, eps = 1e-5) {
  d <- length(par)
  H <- matrix(0, d, d)
  f00 <- fn(par, env)
  # central difference second derivatives
  for (i in 1:d) {
    for (j in i:d) {
      e_i <- rep(0, d); e_i[i] <- eps
      e_j <- rep(0, d); e_j[j] <- eps
      fpp <- fn(par + e_i + e_j, env)
      fpm <- fn(par + e_i - e_j, env)
      fmp <- fn(par - e_i + e_j, env)
      fmm <- fn(par - e_i - e_j, env)
      H[i, j] <- (fpp - fpm - fmp + fmm) / (4 * eps * eps)
      if (j != i) H[j, i] <- H[i, j]
    }
  }
  # symmetrize for numerical noise
  0.5 * (H + t(H))
}

run_once <- function(n = 10, seed = 1L, a = 0.1, b = -0.2) {
  set.seed(seed)
  Ma <- make_spd(n)
  Mb <- make_spd(n)
  sigma <- exp(rnorm(n, sd = 0.3))
  V <- rexp(n, rate = 1) + 0.5
  A_diag <- 1 / (sigma^2 * V)
  mu <- rnorm(n)
  h <- rep(1, n)
  m <- mu * (V - h)
  W <- rnorm(n)

  env <- list(Ma = Ma, Mb = Mb, A_diag = A_diag, mu = mu, V = V, h = h, m = m, W = W)
  par <- c(a, b)

  H_ana <- analytic_hessian(par, env)
  H_num <- numeric_hessian(par, ell_logf, env, eps = 1e-5)
  list(analytic = H_ana, numeric = H_num, diff = H_ana - H_num,
       max_abs_err = max(abs(H_ana - H_num)))
}

main <- function() {
  cat("Hessian check for K(a,b) = exp(a) M_a + 3 b^2 M_b\n")
  res <- run_once(n = 12, seed = 42, a = 0.3, b = -0.4)
  cat("Analytical Hessian:\n"); print(signif(res$analytic, 5))
  cat("Numeric Hessian:\n"); print(signif(res$numeric, 5))
  cat("Difference (ana - num):\n"); print(signif(res$diff, 5))
  cat(sprintf("Max abs error: %.3e\n", res$max_abs_err))

  # Run a few random trials
  errs <- numeric(5)
  for (k in 1:5) {
    a <- rnorm(1)
    b <- rnorm(1)
    tmp <- run_once(n = 10, seed = 100 + k, a = a, b = b)
    errs[k] <- tmp$max_abs_err
  }
  cat("Random trials max abs errors:\n")
  print(signif(errs, 5))
  cat(sprintf("Median max abs error across trials: %.3e\n", median(errs)))
}

if (identical(environment(), globalenv())) {
  main()
}

#' Specifying a latent process model (wrapper function for each model)
#'
#' Function used for defining of smooth and spatial terms
#' within ngme model formulae.
#' The function is a wrapper function for specific submodels.
#' (see ngme_models_types() for available models).
#'
#' @param map  symbol or numerical value: index or covariates to build index
#' @param model  an \code{ngme_operator} or \code{ngme_operator_def} created by
#'   model constructors such as \code{ar1()}, \code{matern()}, \code{bv()}, etc.
#' @param noise  ngme_noise object, noise_nig() or noise_gal()
#' @param name   name of the field, for later use, if not provided, will be "field1" etc.
#' @param data      specifed or inherit from ngme() function
#' @param group   group factor indicate resposne variable, can be inherited from ngme() function, (used for bivariate model)
#' @param which_group  belong to which group
#' @param replicate factor indicating replicate structure for INLA-style replicates.
#'   When provided, operators and A matrices will be block-diagonalized across replicates.
#' @param A  observation matrix, automatically computed given map and model
#' @param W      starting value of the process
#' @param fix_K  fix the estimation of parameter of K
#' @param prior prior specification created by \code{prior_*()} or
#'   \code{priors(...)} for operator parameters (\code{theta_K}).
#' @param fix_W  stop sampling for W
#' @param debug     debug mode
#' @param subset    subset of the model
#'
#' @details When using different meshes for different replicates, provide the mesh parameter
#' as a list of mesh objects. The number of meshes should match the number of replicates.
#' For example: \code{mesh = list(mesh1, mesh2, mesh3)} for 3 replicates.
#'
#' When using the `replicate` argument, the operator matrices will be block-diagonalized.
#' For a generic operator K = param1 * Matrix1 + param2 * Matrix2, with 2 replicates,
#' the resulting operator becomes:
#' K = param1 * bdiag(Matrix1_rep1, Matrix1_rep2) + param2 * bdiag(Matrix2_rep1, Matrix2_rep2)
#' Similarly, the A matrix will be block-diagonalized as bdiag(A_rep1, A_rep2, ...).
#'
#' @return a list for constructing latent model, e.g. A, h, C, G,
#' which also has
#' 1. Information about K matrix
#' 2. Information about noise
#' 3. Control variables
#'
#' @export
f <- function(
    map,
    model,
    noise = noise_normal(),
    name = "field",
    data = NULL,
    group = NULL,
    which_group = NULL,
    replicate = NULL,
    A = NULL,
    W = NULL,
    fix_W = FALSE,
    fix_K = FALSE,
    prior = NULL,
    subset = rep(TRUE, length_map(map)),
    debug = FALSE) {

  # examine the noise
  if (is.list(noise) && !inherits(noise, "ngme_noise")) {
    noise <- convert_noise_list_to_normal_nig(noise)
  }

  # helper: check whether all provided noises are normal
  is_noise_all_normal <- function(noise_obj) {
    if (inherits(noise_obj, "ngme_noise")) {
      return(noise_obj$noise_type == "normal")
    }
    if (is.list(noise_obj) && length(noise_obj) > 0) {
      return(all(vapply(noise_obj, function(x) inherits(x, "ngme_noise") && x$noise_type == "normal", logical(1))))
    }
    FALSE
  }

  maybe_apply_default_nu_prior <- function(noise_obj, h_vec) {
    if (isTRUE(noise_obj$fix_theta_nu) || noise_obj$n_theta_nu == 0) {
      return(noise_obj)
    }
    if (isTRUE(noise_obj$prior_nu_user)) {
      return(noise_obj)
    }
    # Keep the legacy weakly-informative default for non-stationary nu.
    if (length(noise_obj$theta_nu) > 1 || ncol(noise_obj$B_nu) > 1) {
      return(noise_obj)
    }
    if (!any(noise_obj$noise_type %in% c("nig", "normal_nig"))) {
      return(noise_obj)
    }

    h_star <- stats::median(h_vec, na.rm = TRUE)
    if (!is.finite(h_star) || h_star <= 0) {
      return(noise_obj)
    }

    lambda <- log(2) / h_star
    if (!is.finite(lambda) || lambda <= 0) {
      return(noise_obj)
    }

    lower <- if (!is.null(noise_obj$nu_lower_bound)) noise_obj$nu_lower_bound else 0
    noise_obj$prior_nu <- as_internal_prior(
      prior_inv_exp(lambda = lambda, lower = lower, target = "coef")
    )
    noise_obj
  }

  noise_all_normal <- is_noise_all_normal(noise)

  # If the user builds a bv/bv2/bv_matern model inline and all noises are normal,
  # ensure fix_theta = TRUE in that model call.
  model_expr <- substitute(model)
  # Check if model is a call to bv/bv2/bv_matern
  if (noise_all_normal && is.call(model_expr)) {
    op_name <- tryCatch(as.character(model_expr[[1]]), error = function(e) "")
    if (op_name %in% c("bv", "bv2", "bv_matern")) {
      model_list <- as.list(model_expr)
      name_vec <- names(model_list)
      idx <- which(name_vec %in% "fix_theta")

      if (length(idx) > 0 && isFALSE(model_list[[idx]])) {
        stop("For bv/bv2/bv_matern models with all normal noises, fix_theta must be TRUE. Please set fix_theta = TRUE.")
      } else if (length(idx) == 0) {
        model_list <- c(model_list, list(fix_theta = TRUE))
        model_expr <- as.call(model_list)
      } else { # fix_theta is present and not FALSE
        model_list[[idx]] <- TRUE
        model_expr <- as.call(model_list)
      }
    }
  }

  # Evaluate model expression (promises are forced here so we can inspect it consistently)
  model <- eval(model_expr, envir = parent.frame())
  stopifnot(
    "`model` must evaluate to an ngme_operator or ngme_operator_def. String model names are not supported." =
      inherits(model, c("ngme_operator", "ngme_operator_def"))
  )
  model_type <- model$model
  map <- eval(substitute(map), envir = data, enclos = parent.frame())

  # Evaluate map expression (promises are forced here so we can inspect it consistently)
  if (inherits(map, "formula")) {
    map <- get_data_from_formula(map, if (is.null(data)) parent.frame() else data)
  } else if (is.list(map)) {
    map <- lapply(map, function(m) {
      if (inherits(m, "formula")) {
        get_data_from_formula(m, if (is.null(data)) parent.frame() else data)
      } else {
        m
      }
    })
  }

  if (!is.null(data) && is.null(A)) {
    if (!model_type %in% c("tp", "spacetime")) {
      stopifnot(
        "Please make sure length of map is same as nrow(data) in f()" =
          length_map(map) == nrow(data)
      )
    } else {
      stopifnot(
        "Please make sure length of map is same as nrow(data) in f()" =
          length_map(map[[1]]) == nrow(data) &&
            length_map(map[[2]]) == nrow(data)
      )
    }
  }

  group <- validate_rep_or_group(group, data)

  # Validate and process replicate argument
  if (!is.null(replicate)) {
    replicate <- validate_rep_or_group(replicate, data)
    if (!is.null(data)) {
      stopifnot(
        "Length of replicate must match number of observations" =
          length(replicate) == nrow(data)
      )
    } else {
      stopifnot(
        "Length of replicate must match length of map" =
          length(replicate) == length_map(map)
      )
    }
  }

  # set the subset if provide group and which_group
  if (!is.null(which_group)) {
    stopifnot(
      "Please provide group factor" = !is.null(group),
      "Please check if which_group is in group" = which_group %in% levels(group)
    )
    subset <- group %in% which_group
  }

  # 1. Extract or build mesh
  mesh <- NULL
  mesh_list <- NULL

  if (inherits(model, "ngme_operator")) {
    mesh <- model$mesh
  } else if (inherits(model, "ngme_operator_def")) {
    if (model$model == "re") {
      model$args$map <- map
      mesh <- NULL
    } else if (!is.null(model$args$mesh)) {
      mesh <- model$args$mesh
    } else {
      # Build mesh from map if not provided
      mesh <- ngme_build_mesh(sub_map(map, subset), model$model)
      model$args$mesh <- mesh
    }
  } else {
    stop("Model must be ngme_operator or ngme_operator_def")
  }

  # 2. Check if mesh is a list (for replicates)
  is_replicate_mesh <- is.list(mesh) && !inherits(mesh, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))

  if (is_replicate_mesh && model_type %in% c("tp", "spacetime")) {
    # Check if it is a list of meshes for the model itself (not replicates)
    if (inherits(mesh[[1]], c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))) {
      is_replicate_mesh <- FALSE
    }
  }

  if (is_replicate_mesh) {
    mesh_list <- mesh
    # Use the first mesh as a template for building the operator
    mesh <- mesh_list[[1]]

    # Update model definition to use the template mesh
    if (inherits(model, "ngme_operator_def")) {
      model$args$mesh <- mesh
    }
  }

  # 3. Convert ngme_operator_def to ngme_operator
  if (inherits(model, "ngme_operator_def")) {
    operator <- do.call(model$model, model$args)
  } else {
    operator <- model
  }
  model_name <- operator$model

  if (is.null(mesh) && model_type != "re") {
    mesh <- ngme_build_mesh(sub_map(map, subset), model_name)
    operator$mesh <- mesh
  }

  # At this point, 'operator' is an ngme_operator object (possibly a template)
  # and 'model' is the model name string.

  # Build A matrix
  A <- if (is.null(A)) {
    # If mesh_list is present, we defer A matrix construction for replicates
    if (!is.null(mesh_list)) {
      NULL
    } else {
      ngme_build_A(model_name, mesh, map, operator, group)
    }
  } else {
    ngme_as_sparse(A)
  }

  # subset the A matrix only if it was built
  if (!is.null(A) && !all(subset)) {
    A <- A[subset, , drop = FALSE]
  }

  # Sanity warning for ARMA with fixed W and free MA/AR params
  if (model_name == "arma" && isTRUE(fix_W)) {
    ar_mask <- if (!is.null(operator$fix_rho)) operator$fix_rho else logical(0)
    ma_mask <- if (!is.null(operator$fix_phi)) operator$fix_phi else logical(0)
    if ((length(ar_mask) == 0 || !all(ar_mask)) || (length(ma_mask) == 0 || !all(ma_mask))) {
      warning("For ARMA with fix_W=TRUE, ar/ma determine Z and W=P^{-1}x; fixing W while estimating ar/ma leads to mismatch. Either (i) set fix_ar/fix_ma=TRUE and pass W=P^{-1}x, or (ii) allow W to be estimated.")
    }
  }

  # Handle replicates: block-diagonalize operator matrices and A matrix
  if (!is.null(replicate)) {
    n_repl <- length(levels(replicate))

    # Build separate operators for each replicate using their respective meshes
    if (!is.null(mesh_list)) {
      operator_list <- list()
      A_list <- list()
      h_total <- NULL

      for (i in 1:n_repl) {
        # Get indices for this replicate
        repl_idx <- which(replicate == levels(replicate)[i])

        # Get map for this replicate
        if (model_name %in% c("tp", "spacetime")) {
          map_repl <- list(map[[1]][repl_idx], map[[2]][repl_idx])
        } else {
          map_repl <- sub_map(map, repl_idx)
        }

        # Get mesh for this replicate
        mesh_repl <- if (length(mesh_list) >= i) mesh_list[[i]] else mesh_list[[1]]

        # Build operator for this replicate with its specific mesh
        # If model is ngme_operator, we might need to clone it and update mesh?
        # Or if it's ngme_operator_def, we build it.
        # The original code called build_operator.

        if (inherits(model, "ngme_operator")) {
          # If it's already an operator, we can't easily change the mesh if it's baked in?
          # Usually ngme_operator has a mesh.
          # If the user passed a single operator but multiple meshes in f(),
          # we need to re-instantiate the operator with the new mesh.
          # But we don't have the constructor arguments easily if it's already an operator.
          # This suggests that for replicates with different meshes, passing ngme_operator_def is better.

          # If it is an operator, we can try to update the mesh if the operator supports it.
          operator_repl <- model
          operator_repl$mesh <- mesh_repl
          # We might need to recompute K or other things?
          # This is tricky. If the user provides a fully built operator, they might expect it to be used as is.
          # But if mesh changes, the operator matrices change.

          warning("Using a pre-built ngme_operator with replicate and different meshes. The mesh in the operator will be replaced, but internal matrices might not update if they depend on the mesh during construction. Consider using ngme_operator_def (e.g. ar1(..., mesh=NULL)) instead.")
        } else if (inherits(model, "ngme_operator_def")) {
          args <- model$args
          args$mesh <- NULL
          operator_repl <- do.call(model$model, c(list(mesh = mesh_repl), args))
        } else {
          stop("Model must be ngme_operator or ngme_operator_def.")
        }

        operator_list[[i]] <- operator_repl

        # Build A matrix for this replicate
        if (is.null(A)) {
          A_repl <- ngme_build_A(
            model_name, mesh_repl, map_repl, operator_repl,
            if (!is.null(group)) group[repl_idx] else NULL
          )
          A_list[[i]] <- A_repl
        }

        # Accumulate h vectors
        h_total <- c(h_total, operator_repl$h)
      }

      # Create block-diagonal operator matrices
      if (operator_list[[1]]$generic_type %in% c("generic", "generic_ns")) {
        # For generic operators, block-diagonalize each matrix type
        n_matrices <- length(operator_list[[1]]$matrices)
        combined_matrices <- list()

        for (j in 1:n_matrices) {
          matrix_list <- lapply(operator_list, function(op) op$matrices[[j]])
          combined_matrices[[j]] <- do.call(Matrix::bdiag, matrix_list)
        }

        operator$matrices <- combined_matrices

        # Update K matrix using the new block-diagonal matrices
        if (!is.null(operator$update_K)) {
          operator$K <- operator$update_K(operator$theta_K)
        } else {
          # Block-diagonalize K matrices
          K_list <- lapply(operator_list, function(op) op$K)
          operator$K <- do.call(Matrix::bdiag, K_list)
        }
      } else {
        # For non-generic operators, block-diagonalize K directly
        K_list <- lapply(operator_list, function(op) op$K)
        operator$K <- do.call(Matrix::bdiag, K_list)

        # If the operator has component matrices, block-diagonalize them too
        if (!is.null(operator_list[[1]]$C)) {
          operator$C <- do.call(Matrix::bdiag, lapply(operator_list, function(op) op$C))
        }
        if (!is.null(operator_list[[1]]$G)) {
          operator$G <- do.call(Matrix::bdiag, lapply(operator_list, function(op) op$G))
        }
        if (!is.null(operator_list[[1]]$Ci)) {
          operator$Ci <- do.call(Matrix::bdiag, lapply(operator_list, function(op) op$Ci))
        }
        if (!is.null(operator_list[[1]]$B_K)) {
          operator$B_K <- do.call(rbind, lapply(operator_list, function(op) op$B_K))
        }
        if (!is.null(operator_list[[1]]$Z)) {
          operator$Z <- do.call(Matrix::bdiag, lapply(operator_list, function(op) op$Z))
        }

        # Update update_K to return block-diagonal K for shared parameters
        if (!is.null(operator_list[[1]]$update_K)) {
          operator$update_K <- function(theta_K) {
            K_list <- lapply(operator_list, function(op) op$update_K(theta_K))
            do.call(Matrix::bdiag, K_list)
          }
        }
      }

      # Update h vector
      operator$h <- h_total

      # Block-diagonalize A matrices if they were built
      if (!is.null(A_list) && length(A_list) > 0) {
        A <- do.call(Matrix::bdiag, A_list)
      }
    } else {
      # Original logic for same mesh across replicates
      # Block-diagonalize operator matrices if it's a generic operator
      if (operator$generic_type %in% c("generic", "generic_ns")) {
        # Create block-diagonal versions of all matrices
        operator$matrices <- lapply(operator$matrices, function(mat) {
          # Create list of matrices for each replicate
          mat_list <- replicate(n_repl, mat, simplify = FALSE)
          # Block diagonalize
          do.call(Matrix::bdiag, mat_list)
        })

        # Upda simple cases, just block-diagonalize K directly
        K_list <- replicate(n_repl, operator$K, simplify = FALSE)
        operator$K <- do.call(Matrix::bdiag, K_list)
        # Update h vector (concatenate for each replicate)
        operator$h <- rep(operator$h, n_repl)
      } else {
        stop("Not implemented replicate for non-generic operators yet.")
        # For non-generic operators, block-diagonalize K directly
        K_list <- replicate(n_repl, operator$K, simplify = FALSE)
        operator$K <- do.call(Matrix::bdiag, K_list)
        operator$h <- rep(operator$h, n_repl)
      }

      # Block-diagonalize A matrix if it exists
      if (!is.null(A)) {
        # Build A matrix for each replicate
        A_list <- list()
        for (i in 1:n_repl) {
          # Get indices for this replicate
          repl_idx <- which(replicate == levels(replicate)[i])

          # Get map for this replicate
          if (model_name %in% c("tp", "spacetime")) {
            map_repl <- list(map[[1]][repl_idx], map[[2]][repl_idx])
          } else {
            map_repl <- sub_map(map, repl_idx)
          }

          # Build A matrix for this replicate using the same mesh
          A_repl <- ngme_build_A(
            model_name, mesh, map_repl, operator,
            if (!is.null(group)) group[repl_idx] else NULL
          )
          A_list[[i]] <- A_repl
        }

        # Block-diagonalize A matrices
        A <- do.call(Matrix::bdiag, A_list)
      }
    }

    # Apply subset if needed
    if (!is.null(A) && !all(subset)) {
      A <- A[subset, , drop = FALSE]
    }

    # Update W if provided (replicate for each replicate)
    if (!is.null(W)) {
      W <- rep(W, n_repl)
    }
  }

  # 2. build noise given operator
  # bivariate noise
  # 2. build noise given operator
  # bivariate noise
  if (model_name %in% c("bv", "bv_matern", "bv_matern_normal", "bv2", "bv_matern_nig")) {
    stopifnot(
      "Please specify noise for each field" = length(noise) >= 2,
      "Input: noise=list(a=<noise>,b=<noise>)" = inherits(noise[[1]], "ngme_noise"),
      "Input: noise=list(a=<noise>,b=<noise>)" = inherits(noise[[2]], "ngme_noise"),
      "Please specify noise with same name as in the sub_models argument!" = all(names(noise[1:2]) %in% operator$model_names),
      "Keep the noise same if you want to specify single V for each noise" = noise[[1]]$single_V == noise[[2]]$single_V,
      "Two noise should be the same type" = noise[[1]]$noise_type == noise[[2]]$noise_type
    )

    noise1 <- update_noise(noise[[operator$model_names[[1]]]],
      n = length(operator$h) / 2
    )
    noise2 <- update_noise(noise[[operator$model_names[[2]]]],
      n = length(operator$h) / 2
    )
    bv_noises <- list(noise1, noise2)
    names(bv_noises) <- operator$model_names
    share_V <- !is.null(noise$share_V) && noise$share_V
    if (share_V) {
      stopifnot(
        "share_V option is only supported for 2 NIG noise." = noise1$noise_type == noise2$noise_type,
        "share_V option requires nu from both noise are same." = noise1$theta_nu == noise2$theta_nu
      )
    }

    noise <- ngme_noise(
      noise_type = c(noise1$noise_type, noise2$noise_type),
      B_mu = as.matrix(Matrix::bdiag(noise1$B_mu, noise2$B_mu)),
      B_sigma = as.matrix(Matrix::bdiag(noise1$B_sigma, noise2$B_sigma)),
      B_nu = as.matrix(Matrix::bdiag(noise1$B_nu, noise2$B_nu)),
      theta_mu = c(noise1$theta_mu, noise2$theta_mu),
      theta_sigma = c(noise1$theta_sigma, noise2$theta_sigma),
      theta_nu = c(noise1$theta_nu, noise2$theta_nu),
      fix_theta_nu = noise1$fix_theta_nu && noise2$fix_theta_nu,
      fix_theta_mu = noise1$fix_theta_mu && noise2$fix_theta_mu,
      fix_theta_sigma = c(noise1$fix_theta_sigma, noise2$fix_theta_sigma),
      share_V = !is.null(noise$share_V) && noise$share_V,
      single_V = noise1$single_V,
      bv_noises = bv_noises,
      fix_V = !is.null(noise$fix_V) && noise$fix_V,
      V = if (!is.null(noise$V)) noise$V else NULL
    )
  } else {
    n <- length(operator$h)
    if (model_name %in% c("rw1", "rw2") && operator$cyclic) {
      # Compensate the constraint for cyclic rw1/rw2 (we introduce lagrange in definition)
      if (nrow(noise$B_sigma) > 1) { # user-defined B_sigma
        noise$B_sigma <- rbind(0, noise$B_sigma)
      }
      if (nrow(noise$B_mu) > 1) { # user-defined B_mu
        noise$B_mu <- rbind(0, noise$B_mu)
      }
      if (nrow(noise$B_nu) > 1) { # user-defined B_nu
        noise$B_nu <- rbind(0, noise$B_nu)
      }
    }
    noise <- update_noise(noise, n = n)
  }

  if (noise$share_V &&
    !(model_name %in% c("bv", "bv2", "bv_matern", "bv_matern_normal", "bv_matern_nig"))) {
    stop("Not allow for share_V for univariate model")
  }

  if (model_name %in% c("rw1", "rw2")) {
    # force the constrain (e.g. fixed the 1st position to be 0)
    theta_sigma <- c(-10, noise$theta_sigma)
    fix_theta_sigma <- c(TRUE, noise$fix_theta_sigma)
    B_sigma <- cbind(0, noise$B_sigma)
    B_sigma[1, 1] <- 1
    noise$B_sigma[1, -1] <- 0
    if (any(abs(eigen(t(B_sigma) %*% B_sigma)$values) < 1e-10)) {
      # cannot apply transformation
      stop("Do not fix the 1st position of sigma for non-cyclic rw1")
    } else {
      noise$B_sigma <- B_sigma
      noise$theta_sigma <- theta_sigma
      noise$fix_theta_sigma <- fix_theta_sigma
    }

    if (model_name == "rw2" && !operator$cyclic) {
      noise$B_sigma[2, 1] <- 1
      noise$B_sigma[2, -1] <- 0
      noise$theta_sigma[1] <- -6 # set larger for numerical stability
    }

    if (operator$cyclic) A <- cbind(0, A) # ignore the 1st auxiliary element
  }

  # Reshape the noise structure for normal_nig
  if (all(noise$noise_type %in% c(
    "normal_nig", "normal_gal", "nig_gal"
  ))) {
    # update operator matrices M to be diag(M, M)
    if (operator$generic_type %in% c("generic", "generic_ns")) {
      operator$matrices <- lapply(operator$matrices, function(x) {
        Matrix::bdiag(x, x)
      })
      operator$K <- operator$K <- Matrix::bdiag(operator$K, operator$K)
      operator$h <- operator$h <- c(operator$h, operator$h)
    }
    # A <- [A A]
    A <- cbind(A, A)

    W <- c(W, W)
  }

  noise <- maybe_apply_default_nu_prior(noise, operator$h)

  operator_prior_names <- operator$param_name
  if (is.null(operator_prior_names) ||
      length(operator_prior_names) != length(operator$theta_K)) {
    operator_prior_names <- paste0("theta", seq_along(operator$theta_K))
  }
  prior_theta_K <- compile_operator_priors(prior, operator_prior_names)

  ngme_model(
    model = model_name,
    operator = operator,
    noise = noise,
    W_size = ncol(operator$K),
    V_size = nrow(operator$K),
    A = A,
    map = map,
    mesh = mesh,
    mesh_list = mesh_list,
    n_map = length_map(map),
    W = W,
    fix_W = fix_W,
    name = name,
    fix_K = fix_K,
    prior_theta_K = prior_theta_K,
    debug = debug,
    replicate = replicate
  )
}

# build operator


# Convert a list containing both nig and normal noise to a normal_nig noise object
convert_noise_list_to_normal_nig <- function(noise_list) {
  # Check if we have a list with both nig and normal noise types
  if (!(length(noise_list) == 2 &&
    ((inherits(noise_list[[1]], "ngme_noise") && noise_list[[1]]$noise_type == "nig" &&
      inherits(noise_list[[2]], "ngme_noise") && noise_list[[2]]$noise_type == "normal") ||
      (inherits(noise_list[[1]], "ngme_noise") && noise_list[[1]]$noise_type == "normal" &&
        inherits(noise_list[[2]], "ngme_noise") && noise_list[[2]]$noise_type == "nig")))) {
    return(noise_list)
  }

  # Extract the nig and normal components
  nig_noise <- if (noise_list[[1]]$noise_type == "nig") noise_list[[1]] else noise_list[[2]]
  normal_noise <- if (noise_list[[1]]$noise_type == "normal") noise_list[[1]] else noise_list[[2]]

  # Create a normal_nig noise object
  noise_normal_nig(
    mu = nig_noise$theta_mu,
    sigma_nig = exp(nig_noise$theta_sigma),
    sigma_normal = exp(normal_noise$theta_sigma),
    nu = nig_noise$nu_lower_bound + exp(nig_noise$theta_nu),
    nu_lower_bound = nig_noise$nu_lower_bound,
    B_mu = nig_noise$B_mu,
    B_sigma_nig = nig_noise$B_sigma,
    B_sigma_normal = normal_noise$B_sigma,
    B_nu = nig_noise$B_nu,
    fix_theta_mu = nig_noise$fix_theta_mu,
    fix_theta_sigma = c(
      nig_noise$fix_theta_sigma,
      normal_noise$fix_theta_sigma
    ),
    fix_theta_nu = nig_noise$fix_theta_nu,
    corr_measurement = nig_noise$corr_measurement || normal_noise$corr_measurement,
    V = nig_noise$V,
    fix_V = nig_noise$fix_V
  )
}

# help to build a list of mesh for different replicates
ngme_build_mesh <- function(
    loc,
    model = NULL,
    ...) {
  if (inherits(loc, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))) {
    return(loc)
  }

  # Check if loc is a list of meshes
  if (is.list(loc) && length(loc) > 0 && all(vapply(loc, function(x) inherits(x, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph")), logical(1)))) {
    return(loc)
  }

  if (!is.null(model)) {
    model_name <- if (inherits(model, "ngme_operator")) model$model else model
    if (model_name %in% c("re", "tp")) {
      return(NULL)
    }
    if (model_name == "iid") loc <- as.integer(as.factor(loc))
    if (model_name %in% c("ar", "ar1", "arma")) {
      stopifnot(
        "The map should be integers." = is.numeric(loc) && all(loc == round(loc))
      )
      return(fmesher::fm_mesh_1d(loc = min(loc):max(loc)))
      # return (fmesher::fm_mesh_1d(as.integer(as.factor(loc))))
    }
  }

  if (is.matrix(loc) && ncol(loc) == 2) {
    stop("Please build and provide the mesh for spatial data using fmesher::fm_mesh_2d()")
  } else if (is.numeric(loc)) {
    mesh <- fmesher::fm_mesh_1d(loc = loc)
  } else {
    stop("The mesh provided is invalid.")
  }

  mesh
}

# create the general replicate model
ngme_replicate <- function(
    Y = NULL,
    X = NULL,
    prior_beta = NULL,
    noise = noise_normal(),
    models = list(),
    control_ngme = control_ngme(),
    standardize = TRUE,
    log_likelihood = NULL,
    ...) {
  # compute W_sizes and V_sizes
  W_sizes <- sum(unlist(lapply(models, function(x) x[["W_size"]]))) # W_sizes = sum(ncol_K)
  V_sizes <- sum(unlist(lapply(models, function(x) x[["V_size"]]))) # W_sizes = sum(nrow_K)

  n_la_params <- sum(unlist(lapply(models, function(x) x["n_params"])))
  n_feff <- ncol(X)
  feff <- if (!is.null(control_ngme$beta_init)) control_ngme$beta_init else control_ngme$feff
  names(feff) <- colnames(X)
  if (is.null(prior_beta)) {
    prior_beta <- compile_beta_priors(NULL, colnames(X))
  }

  # Collect parameter names as a vector (not concatenated string)
  par_names <- character(0)

  # Add latent model parameter names
  for (latent in models) {
    if (!is.null(latent$par_names)) {
      par_names <- c(par_names, latent$par_names)
    }
  }

  # Add fixed effect names
  if (ncol(X) > 0) {
    par_names <- c(par_names, paste0("feff_", seq_along(feff)))
  }

  # Add measurement error parameter names (free parameters only)
  merr_names <- {
    mu_params <- if (length(noise$theta_mu) == 0 || isTRUE(noise$fix_theta_mu)) {
      character(0)
    } else {
      paste0("mu_", seq_along(noise$theta_mu))
    }

    sigma_idx <- if (length(noise$theta_sigma) == 0) integer(0) else which(!noise$fix_theta_sigma)
    sigma_params <- if (length(sigma_idx) == 0) character(0) else
      paste0("sigma_", sigma_idx)

    nu_params <- if (length(noise$theta_nu) == 0 || isTRUE(noise$fix_theta_nu)) {
      character(0)
    } else {
      paste0("nu_", seq_along(noise$theta_nu))
    }

    rho_params <- if (length(noise$rho) == 0 || isTRUE(noise$fix_rho)) {
      character(0)
    } else {
      paste0("rho_", seq_along(noise$rho))
    }

    c(mu_params, sigma_params, nu_params, rho_params)
  }
  par_names <- c(par_names, merr_names)

  n_params <- n_feff + n_la_params + noise$n_params
  structure(
    list(
      Y                 = Y,
      X                 = X,
      feff              = feff,
      prior_beta        = prior_beta,
      models            = models,
      noise             = noise,
      control_ngme      = control_ngme,
      par_names         = par_names,
      W_sizes           = W_sizes,
      V_sizes           = V_sizes,
      n_merr            = noise$n_params,
      n_params          = n_params,
      n_la_params       = n_la_params,
      standardize       = standardize,
      log_likelihood    = log_likelihood,
      ...
    ),
    class = c("ngme_replicate", "list")
  )
}

#' Print ngme object
#'
#' @param x ngme object
#' @param ... ignored
#'
#' @return Invisibly returns \code{x}, an \code{ngme_replicate} object. The
#'   method is called for its side effect of printing a replicate summary.
#' @export
print.ngme_replicate <- function(x, ...) {
  ngme_rep <- x
  cat("*** Ngme object ***\n\n")

  cat("Fixed effects: \n")
  if (length(ngme_rep$feff) > 0) {
    print(ngme_rep$feff, digits = 3)
  } else {
    cat("  None\n")
  }
  cat("\n")
  # cat(paste("  ", ngme_rep_format("feff", ngme_rep$feff)));

  cat("Models: \n")
  for (i in seq_along(ngme_rep$models)) {
    # cat("[["); cat(i); cat("]]")
    # cat("\""); cat(names(ngme_rep$models)[[i]]); cat("\"\n")
    cat("$")
    cat(names(ngme_rep$models)[[i]])
    cat("\n")
    print(ngme_rep$models[[i]], padding = 2)
    cat("\n")
  }

  cat("Measurement noise: \n")
  print(ngme_rep$noise, padding = 2)
  cat("\n\n")

  if (ngme_rep$all_gaussian && !is.null(ngme_rep$log_likelihood)) {
    cat("Log likelihood: ")
    cat(ngme_rep$log_likelihood)
    cat("\n\n")
  }
  invisible(x)
}

library(rSPDE)

s <- seq(from = 0, to = 1, length.out = 1001)
mesh <- fmesher::fm_mesh_1d(s)
kappa <- 20
sigma <- 10
tau <- 1 / sigma
nu <- 0.8
alpha <- nu + 1 / 2
r <- sqrt(8 * nu) / kappa

# integer case
kappa <- 2
sigma <- 5
tau <- 1 / sigma
nu <- 1.5
alpha <- nu + 1 / 2
r <- sqrt(8 * nu) / kappa

# alpha=nu + d/2
C <- matern(mesh, kappa = kappa)$C
G <- matern(mesh, kappa = kappa)$G

# op <- matern.operators(
#   C = C,
#   G = G,
#   sigma = sigma,
#   range = r,
#   # nu = nu,
#   nu = 2 - 1/2,
#   # loc_mesh = s,
#   d = 1,
#   m = 1,
#   type = "operator",
#   parameterization = "matern"
# )
alpha
op <- matern.operators(
  C = C,
  G = G,
  tau = tau,
  kappa = kappa,
  alpha = alpha,
  d = 1,
  m = 2,
  type = "operator",
  parameterization = "spde"
)
# op$beta
# op$Pl
# op$Pr

w <- simulate(op, seed = 10)[, 1]
var(w)
y <- w + rnorm(length(w), sd = 0.1)
dat <- data.frame(y = y, s = s)

# 2) 用 rspde_lme 拟合（rspde_order 与 m 一致）
rspde_fit <- rspde_lme(
  formula = y ~ 0,
  loc = "s",
  data = dat,
  model = matern.operators(
    mesh = mesh,
    kappa = kappa, tau = tau, alpha = alpha,
    d = 1, m = 2,
    type = "operator",
    parameterization = "spde"
  ),
  rspde_order = 2,
  model_options = list(
    start_sigma_e = 0.1,
    fix_alpha = alpha
    # 可选：fix_alpha = alpha, fix_kappa = kappa, fix_tau = tau
  )
)
summary(rspde_fit)

# options(error = recover)
fit <- ngme(
  y ~ 0 + f(
    1:length(y),
    model = "matern",
    name = "spde",
    kappa = 20,
    mesh = mesh,
    alpha = alpha,
    fix_alpha = TRUE,
    rational_order = 2,
    noise = noise_normal(sigma = 5)
  ),
  data = data.frame(y = y),
  control_opt = control_opt(
    estimation = TRUE,
    iterations = 1000,
    optimizer = adam(),
    n_batch = 10,
    rao_blackwellization = F,
    n_parallel_chain = 4,
    print_check_info = F,
    verbose = T,
    std_lim = 0.01,
    start_sd = 0.01
  )
)
fit
traceplot(fit, "spde")
traceplot(fit, hline = 0.1)

#' Batch-Means Covariance Estimator for SGD Trajectories
#'
#' Estimate the asymptotic covariance from a single SGD trajectory using the
#' increasing-batch construction from Xi et al. (2020).
#'
#' @param trajectory numeric matrix with rows = iterations and columns = parameters.
#' @param alpha stepsize decay exponent in
#'   \eqn{\eta_i = \eta i^{-\alpha}}. Must satisfy \eqn{1/2 < \alpha < 1}.
#' @param M number of retained batches (excluding burn-in batch 0). If `NULL`,
#'   use \eqn{\lfloor n^{(1-\alpha)/2} \rfloor}.
#' @param N decorrelation constant in
#'   \eqn{e_k = \lfloor ((k+1)N)^{1/(1-\alpha)} \rfloor}. If `NULL`, use
#'   \eqn{N = n^{1-\alpha}/(M+1)}.
#' @param drop_burnin logical; if `TRUE`, discard batch 0.
#' @param burnin_iter non-negative integer. Explicitly discard the first
#'   `burnin_iter` iterations before building Xi-style batches. After trimming,
#'   batch boundaries are rebuilt from iteration 1 of the retained trajectory.
#'
#' @return A list containing the covariance estimate, pooled mean, batch sizes,
#' batch boundaries, and chain-level metadata.
#' @export
batch_means_estimator <- function(
    trajectory,
    alpha = 0.501,
    M = NULL,
    N = NULL,
    drop_burnin = TRUE,
    burnin_iter = 0) {
  stopifnot(
    is.matrix(trajectory) || is.data.frame(trajectory),
    is.numeric(alpha),
    length(alpha) == 1,
    alpha > 0.5,
    alpha < 1,
    is.logical(drop_burnin),
    length(drop_burnin) == 1,
    is.numeric(burnin_iter),
    length(burnin_iter) == 1,
    is.finite(burnin_iter),
    burnin_iter >= 0,
    abs(burnin_iter - round(burnin_iter)) < sqrt(.Machine$double.eps)
  )

  trajectory <- as.matrix(trajectory)
  storage.mode(trajectory) <- "double"

  burnin_iter <- as.integer(round(burnin_iter))
  n_iter_total <- nrow(trajectory)
  n_param <- ncol(trajectory)
  stopifnot(
    n_iter_total >= 2,
    n_param >= 1
  )

  if (burnin_iter > 0) {
    stopifnot(
      "burnin_iter is too large; at least 2 iterations must remain after trimming." =
        n_iter_total - burnin_iter >= 2
    )
    trajectory <- trajectory[seq.int(burnin_iter + 1L, n_iter_total), , drop = FALSE]
  }
  n_iter <- nrow(trajectory)

  boundary <- .batch_means_boundaries(
    n_iter = n_iter,
    alpha = alpha,
    M = M,
    N = N
  )

  keep_idx <- if (drop_burnin) {
    seq.int(2L, length(boundary$end))
  } else {
    seq_along(boundary$end)
  }

  stopifnot(
    "No valid batch remains after burn-in removal; increase iterations or reduce M." =
      length(keep_idx) > 0
  )

  if (length(keep_idx) < 2) {
    warning(
      "Only one batch is retained; covariance estimate may be unstable.",
      call. = FALSE
    )
  }

  batch_means <- t(vapply(keep_idx, function(k) {
    idx <- boundary$start[k]:boundary$end[k]
    colMeans(trajectory[idx, , drop = FALSE])
  }, numeric(n_param)))

  batch_sizes <- boundary$size[keep_idx]
  global_mean <- colSums(batch_means * batch_sizes) / sum(batch_sizes)

  centered <- sweep(batch_means, 2, global_mean, "-")
  covariance <- matrix(0, n_param, n_param)
  for (i in seq_len(nrow(centered))) {
    covariance <- covariance + batch_sizes[i] * tcrossprod(centered[i, ])
  }
  covariance <- covariance / length(keep_idx)

  param_names <- colnames(trajectory)
  if (is.null(param_names)) {
    param_names <- paste0("param_", seq_len(n_param))
  }
  names(global_mean) <- param_names
  dimnames(covariance) <- list(param_names, param_names)

  ret <- list(
    mean = global_mean,
    covariance = covariance,
    n_eff = sum(batch_sizes),
    batch_sizes = batch_sizes,
    batch_start = boundary$start[keep_idx],
    batch_end = boundary$end[keep_idx],
    batch_means = batch_means,
    alpha = alpha,
    M = length(keep_idx),
    N = boundary$N,
    dropped_burnin = drop_burnin,
    burnin_iter = burnin_iter,
    n_iterations = n_iter,
    n_iterations_total = n_iter_total
  )

  class(ret) <- "ngme_batch_means"
  ret
}

#' Pooled Batch-Means Confidence Intervals from Multiple Chains
#'
#' Compute pooled point estimates, covariance, and Wald confidence intervals from
#' multiple chain trajectories.
#'
#' @param chain_trajectories list of numeric matrices, each with rows = iterations
#'   and columns = parameters.
#' @param level confidence level.
#' @param alpha stepsize decay exponent in \eqn{(1/2, 1)}.
#' @param M number of retained batches per chain (excluding burn-in batch 0).
#' @param N decorrelation constant used in the batch boundary formula.
#' @param drop_burnin logical; if `TRUE`, discard batch 0.
#' @param burnin_iter non-negative integer. Explicitly discard the first
#'   `burnin_iter` iterations of each chain before Xi-style batching.
#'
#' @return A list with pooled estimates, covariance, standard errors, and
#' confidence intervals.
#' @export
batch_means_ci <- function(
    chain_trajectories,
    level = 0.95,
    alpha = 0.501,
    M = NULL,
    N = NULL,
    drop_burnin = TRUE,
    burnin_iter = 0) {
  if (is.matrix(chain_trajectories) || is.data.frame(chain_trajectories)) {
    chain_trajectories <- list(as.matrix(chain_trajectories))
  }

  stopifnot(
    is.list(chain_trajectories),
    length(chain_trajectories) >= 1,
    is.numeric(level),
    length(level) == 1,
    level > 0,
    level < 1
  )

  chain_trajectories <- lapply(chain_trajectories, function(x) {
    stopifnot(is.matrix(x) || is.data.frame(x))
    x <- as.matrix(x)
    storage.mode(x) <- "double"
    x
  })

  n_param <- unique(vapply(chain_trajectories, ncol, integer(1)))
  stopifnot(
    "All chains must have the same number of parameters." = length(n_param) == 1
  )

  n_iter_vec <- vapply(chain_trajectories, nrow, integer(1))
  if (length(unique(n_iter_vec)) > 1) {
    min_iter <- min(n_iter_vec)
    warning(
      "Chains have different lengths; truncating all chains to the minimum length.",
      call. = FALSE
    )
    chain_trajectories <- lapply(chain_trajectories, function(x) {
      x[seq_len(min_iter), , drop = FALSE]
    })
  }

  param_names <- colnames(chain_trajectories[[1]])
  if (is.null(param_names)) {
    param_names <- paste0("param_", seq_len(n_param))
  }

  chain_results <- lapply(chain_trajectories, function(x) {
    if (is.null(colnames(x))) {
      colnames(x) <- param_names
    }
    batch_means_estimator(
      trajectory = x,
      alpha = alpha,
      M = M,
      N = N,
      drop_burnin = drop_burnin,
      burnin_iter = burnin_iter
    )
  })

  n_chain <- length(chain_results)
  chain_means <- do.call(rbind, lapply(chain_results, `[[`, "mean"))
  pooled_mean <- colMeans(chain_means)

  pooled_cov <- Reduce(`+`, lapply(chain_results, `[[`, "covariance")) / n_chain
  dimnames(pooled_cov) <- list(param_names, param_names)

  n_eff_vec <- vapply(chain_results, `[[`, numeric(1), "n_eff")
  n_eff <- min(n_eff_vec)
  if (length(unique(n_eff_vec)) > 1) {
    warning(
      "Effective sample sizes differ across chains; using the minimum for CI width.",
      call. = FALSE
    )
  }

  z_value <- stats::qnorm(1 - (1 - level) / 2)
  std_error <- sqrt(pmax(diag(pooled_cov), 0) / (n_eff * n_chain))
  conf_int <- cbind(
    lower = pooled_mean - z_value * std_error,
    upper = pooled_mean + z_value * std_error
  )

  pooled_mean <- stats::setNames(pooled_mean, param_names)
  std_error <- stats::setNames(std_error, param_names)
  rownames(conf_int) <- param_names

  ret <- list(
    estimates = pooled_mean,
    std_error = std_error,
    conf_int = conf_int,
    covariance = pooled_cov,
    level = level,
    z_value = z_value,
    alpha = alpha,
    M = chain_results[[1]]$M,
    N = chain_results[[1]]$N,
    burnin_iter = burnin_iter,
    n_eff = n_eff,
    n_chains = n_chain,
    n_iterations = chain_results[[1]]$n_iterations,
    n_iterations_total = chain_results[[1]]$n_iterations_total,
    chain_results = chain_results
  )

  class(ret) <- "ngme_batch_ci"
  ret
}

#' Batch-Means Confidence Intervals from an ngme Fit
#'
#' Compute Xi-style batch-means confidence intervals directly from trajectories
#' stored in an `ngme` object.
#'
#' @param ngme fitted `ngme` object with `store_traj = TRUE` during optimization.
#' @param name either `"all"` (default) to use all parameters jointly, a latent
#'   model name, or `"general"` for measurement-noise/fixed-effect block.
#' @param level confidence level.
#' @param alpha stepsize decay exponent in \eqn{(1/2, 1)}.
#' @param M number of retained batches (excluding burn-in batch 0).
#' @param N decorrelation constant in the batch boundary formula.
#' @param apply_transform logical; if `TRUE`, inference is first computed in the
#'   raw optimization space and then mapped to user scale via post-hoc Delta
#'   method (Post-delta).
#' @param drop_burnin logical; if `TRUE`, discard batch 0.
#' @param burnin_iter non-negative integer. Explicitly discard the first
#'   `burnin_iter` optimization iterations before Xi-style batching.
#'
#' @return Same structure as [batch_means_ci()] with extra metadata fields
#' `name` and `apply_transform`.
#' @export
ngme_batch_ci <- function(
    ngme,
    name = "all",
    level = 0.95,
    alpha = 0.501,
    M = NULL,
    N = NULL,
    apply_transform = TRUE,
    drop_burnin = TRUE,
    burnin_iter = 0) {
  stopifnot(
    inherits(ngme, "ngme"),
    is.character(name),
    length(name) == 1,
    is.logical(apply_transform),
    length(apply_transform) == 1
  )

  chain_data <- .ngme_chain_trajectories(
    ngme = ngme,
    name = name,
    apply_transform = FALSE
  )

  raw_ret <- batch_means_ci(
    chain_trajectories = chain_data$chains,
    level = level,
    alpha = alpha,
    M = M,
    N = N,
    drop_burnin = drop_burnin,
    burnin_iter = burnin_iter
  )

  ret <- if (apply_transform) {
    .delta_transform_batch_ci(raw_ret, chain_data$transforms)
  } else {
    raw_ret
  }

  ret$name <- name
  ret$apply_transform <- apply_transform
  if (apply_transform) {
    ret$raw_estimates <- raw_ret$estimates
    ret$raw_covariance <- raw_ret$covariance
    ret$raw_std_error <- raw_ret$std_error
    ret$raw_conf_int <- raw_ret$conf_int
  }
  ret
}

#' Refit an Existing ngme Object with SGD and Compute Batch-Means CI
#'
#' Run one additional SGD stage (warm-started from an existing `ngme` fit),
#' then compute Xi-style batch-means confidence intervals from the newly stored
#' trajectories.
#'
#' @param fit existing fitted `ngme` object (can be obtained with any optimizer).
#' @param iterations optimization iterations for the SGD CI stage.
#' @param alpha Xi-style exponent used for both polynomial schedule and
#'   batch-means inference.
#' @param optimizer optimizer for the CI stage; must be `sgd(...)` for Xi theory.
#' @param burnin burn-in iterations before SGD optimization.
#' @param n_batch number of optimization checkpoints.
#' @param n_parallel_chain number of parallel chains for SGD.
#' @param t0 non-negative schedule offset in
#'   \eqn{\eta_t = \eta_0 (t + t0)^{-\alpha}}.
#' @param start_sd standard deviation for randomized chain initialization.
#' @param seed random seed for CI-stage optimization.
#' @param verbose logical; print optimization progress.
#' @param level confidence level for CI.
#' @param name parameter block for CI (`"all"`, latent name, or `"general"`).
#' @param M number of retained Xi batches (excluding batch 0).
#' @param N decorrelation constant in Xi batch boundaries.
#' @param apply_transform logical; if `TRUE`, apply Post-delta transform after
#'   raw-space inference.
#' @param drop_burnin logical; if `TRUE`, discard Xi batch 0.
#' @param burnin_iter non-negative integer used both as optimizer schedule
#'   warmup (`stepsize_schedule_burnin_iter`) and as explicit CI trajectory
#'   trimming before Xi-style batching.
#' @param control_opt optional pre-built `control_opt` object for the CI stage.
#'   If supplied, it is used directly (with `store_traj` forced to `TRUE`).
#' @param ... additional arguments forwarded to [control_opt_batch_ci()] when
#'   `control_opt` is `NULL`.
#'
#' @return An object of class `ngme_batch_ci` with extra fields:
#'   `refit` (the SGD-refit model), `refit_control_opt`, and `refit_source`.
#' @export
compute_ngme_ci <- function(
    fit,
    iterations = 4000,
    alpha = 0.501,
    optimizer = sgd(stepsize = 0.03),
    burnin = 100,
    n_batch = 20,
    n_parallel_chain = 4,
    t0 = 1,
    start_sd = 0.2,
    seed = Sys.time(),
    verbose = FALSE,
    level = 0.95,
    name = "all",
    M = NULL,
    N = NULL,
    apply_transform = TRUE,
    drop_burnin = TRUE,
    burnin_iter = 0,
    control_opt = NULL,
    ...) {
  stopifnot(inherits(fit, "ngme"))

  fit_meta <- attr(fit, "fit")
  stopifnot(
    "Missing fit metadata in `fit`; please pass a fitted ngme object from ngme()." =
      !is.null(fit_meta),
    "Missing formula in fit metadata." = !is.null(fit_meta$formula),
    "Missing data in fit metadata." = !is.null(fit_meta$data)
  )

  dots <- list(...)
  if ("schedule_burnin_iter" %in% names(dots)) {
    stop(
      "In compute_ngme_ci(), `schedule_burnin_iter` is unified with `burnin_iter`. ",
      "Please set only `burnin_iter`.",
      call. = FALSE
    )
  }

  if (is.null(control_opt)) {
    control_opt <- do.call(
      control_opt_batch_ci,
      utils::modifyList(
        list(
          optimizer = optimizer,
          burnin = burnin,
          iterations = iterations,
          n_batch = n_batch,
          n_parallel_chain = n_parallel_chain,
          alpha = alpha,
          t0 = t0,
          schedule_burnin_iter = burnin_iter,
          start_sd = start_sd,
          seed = seed,
          verbose = verbose
        ),
        dots
      )
    )
  } else {
    stopifnot(
      inherits(control_opt, "control_opt"),
      "When `control_opt` is supplied, extra `...` arguments are not used." =
        length(dots) == 0
    )
  }

  stopifnot(
    "compute_ngme_ci currently requires optimizer = sgd(...) for Xi-style guarantees." =
      identical(control_opt$sgd_method, "sgd")
  )

  if (isTRUE(control_opt$trend_std_conv_check) ||
    isTRUE(control_opt$R_hat_conv_check) ||
    isTRUE(control_opt$pflug_conv_check)) {
    warning(
      "Convergence diagnostics are enabled in `control_opt`; early stopping may reduce fixed-iteration Xi-style validity.",
      call. = FALSE
    )
  }

  control_opt$stepsize_schedule_burnin_iter <- as.integer(round(burnin_iter))
  control_opt$store_traj <- TRUE

  fit_env <- build_cv_rebuild_env(
    formula = fit_meta$formula,
    ngme_obj = fit,
    data = fit_meta$data
  )

  control_ngme_refit <- fit$replicates[[1]]$control_ngme
  if (is.null(control_ngme_refit)) {
    control_ngme_refit <- control_ngme()
  }

  refit <- do.call(
    "ngme",
    list(
      formula = fit_meta$formula,
      data = fit_meta$data,
      family = fit_meta$family,
      control_opt = control_opt,
      control_ngme = control_ngme_refit,
      group = fit_meta$group,
      replicate = fit_meta$replicate,
      start = fit
    ),
    envir = fit_env
  )

  ci <- ngme_batch_ci(
    ngme = refit,
    name = name,
    level = level,
    alpha = alpha,
    M = M,
    N = N,
    apply_transform = apply_transform,
    drop_burnin = drop_burnin,
    burnin_iter = burnin_iter
  )

  ci$refit <- refit
  ci$refit_control_opt <- control_opt
  ci$refit_source <- fit
  ci
}

#' @rdname compute_ngme_ci
#' @export
compute_ngme_CI <- compute_ngme_ci

#' Refit an Existing ngme Object with SGLD and Extract Samples
#'
#' Run one additional SGLD stage (warm-started from an existing `ngme` fit),
#' then extract posterior-like samples from stored trajectories.
#'
#' @param fit existing fitted `ngme` object (can be obtained with any optimizer).
#' @param iterations optimization iterations for the SGLD stage.
#' @param optimizer optimizer for the sampling stage; must be `sgld(...)`.
#' @param burnin burn-in iterations before optimization.
#' @param n_batch number of optimization checkpoints.
#' @param n_parallel_chain number of parallel chains.
#' @param alpha polynomial schedule exponent used by `poly_decay(alpha, t0)`.
#' @param t0 non-negative schedule offset.
#' @param start_sd standard deviation for randomized chain initialization.
#' @param seed random seed for SGLD stage.
#' @param verbose logical; print optimization progress.
#' @param name parameter block to extract: `"all"` (default), latent model name,
#'   or `"general"`.
#' @param burnin_iter non-negative integer used both as optimizer schedule
#'   warmup (`stepsize_schedule_burnin_iter`) and as explicit trajectory
#'   trimming before sampling.
#' @param thinning positive integer thinning interval.
#' @param n_gibbs_samples optional positive integer. If supplied, override
#'   `control_ngme$n_gibbs_samples` for the SGLD refit stage only; otherwise
#'   inherit the value from `fit`.
#' @param apply_transform logical; apply parameter transforms to user scale.
#' @param combine_chains logical; if `TRUE`, return one combined data.frame,
#'   otherwise return one data.frame per chain.
#' @param control_opt optional pre-built `control_opt` object for the SGLD
#'   stage. If supplied, it is used directly (with `store_traj` forced to
#'   `TRUE`).
#' @param ... additional arguments forwarded to [control_opt()] when
#'   `control_opt` is `NULL`.
#'
#' @return A data.frame (or list of data.frames) from [ngme_sgld_samples()]
#' with extra attributes `refit`, `refit_control_opt`, and `refit_source`.
#' @export
compute_ngme_sgld_samples <- function(
    fit,
    iterations = 4000,
    optimizer = sgld(stepsize = 0.01, temperature = 1),
    burnin = 100,
    n_batch = 20,
    n_parallel_chain = 4,
    alpha = 0.6,
    t0 = 10,
    start_sd = 0.2,
    seed = Sys.time(),
    verbose = FALSE,
    name = "all",
    burnin_iter = 0,
    thinning = 1,
    n_gibbs_samples = NULL,
    apply_transform = TRUE,
    combine_chains = TRUE,
    control_opt = NULL,
    ...) {
  stopifnot(inherits(fit, "ngme"))
  stopifnot(
    is.null(n_gibbs_samples) || (
      is.numeric(n_gibbs_samples) &&
        length(n_gibbs_samples) == 1 &&
        is.finite(n_gibbs_samples) &&
        n_gibbs_samples > 0 &&
        abs(n_gibbs_samples - round(n_gibbs_samples)) < sqrt(.Machine$double.eps)
    )
  )

  fit_meta <- attr(fit, "fit")
  stopifnot(
    "Missing fit metadata in `fit`; please pass a fitted ngme object from ngme()." =
      !is.null(fit_meta),
    "Missing formula in fit metadata." = !is.null(fit_meta$formula),
    "Missing data in fit metadata." = !is.null(fit_meta$data)
  )

  dots <- list(...)
  if ("schedule_burnin_iter" %in% names(dots)) {
    stop(
      "In compute_ngme_sgld_samples(), `schedule_burnin_iter` is unified with `burnin_iter`. ",
      "Please set only `burnin_iter`.",
      call. = FALSE
    )
  }

  if (is.null(control_opt)) {
    control_opt <- do.call(
      "control_opt",
      utils::modifyList(
        list(
          optimizer = optimizer,
          burnin = burnin,
          iterations = iterations,
          n_batch = n_batch,
          n_parallel_chain = n_parallel_chain,
          store_traj = TRUE,
          trend_std_conv_check = FALSE,
          R_hat_conv_check = FALSE,
          pflug_conv_check = FALSE,
          stepsize_control = poly_decay(
            alpha = alpha,
            t0 = t0,
            burnin_iter = burnin_iter
          ),
          start_sd = start_sd,
          seed = seed,
          verbose = verbose
        ),
        dots
      )
    )
  } else {
    stopifnot(
      inherits(control_opt, "control_opt"),
      "When `control_opt` is supplied, extra `...` arguments are not used." =
        length(dots) == 0
    )
  }

  stopifnot(
    "compute_ngme_sgld_samples currently requires optimizer = sgld(...)." =
      identical(control_opt$sgd_method, "sgld")
  )

  if (isTRUE(control_opt$trend_std_conv_check) ||
    isTRUE(control_opt$R_hat_conv_check) ||
    isTRUE(control_opt$pflug_conv_check)) {
    warning(
      "Convergence diagnostics are enabled in `control_opt`; early stopping may reduce fixed-iteration sampling consistency.",
      call. = FALSE
    )
  }

  control_opt$stepsize_schedule_burnin_iter <- as.integer(round(burnin_iter))
  control_opt$store_traj <- TRUE

  fit_env <- build_cv_rebuild_env(
    formula = fit_meta$formula,
    ngme_obj = fit,
    data = fit_meta$data
  )

  control_ngme_refit <- fit$replicates[[1]]$control_ngme
  if (is.null(control_ngme_refit)) {
    control_ngme_refit <- control_ngme()
  }
  if (!is.null(n_gibbs_samples)) {
    control_ngme_refit$n_gibbs_samples <- as.integer(round(n_gibbs_samples))
  }

  refit <- do.call(
    "ngme",
    list(
      formula = fit_meta$formula,
      data = fit_meta$data,
      family = fit_meta$family,
      control_opt = control_opt,
      control_ngme = control_ngme_refit,
      group = fit_meta$group,
      replicate = fit_meta$replicate,
      start = fit
    ),
    envir = fit_env
  )

  ret <- ngme_sgld_samples(
    ngme = refit,
    name = name,
    burnin_iter = burnin_iter,
    thinning = thinning,
    apply_transform = apply_transform,
    combine_chains = combine_chains
  )

  attr(ret, "refit") <- refit
  attr(ret, "refit_control_opt") <- control_opt
  attr(ret, "refit_source") <- fit
  ret
}

#' Extract Posterior-like Samples from Stored SGLD Trajectories
#'
#' Build posterior-like samples from optimizer trajectories by dropping an
#' initial burn-in segment and applying thinning.
#'
#' @param ngme fitted `ngme` object with `store_traj = TRUE`.
#' @param name parameter block to extract: `"all"` (default), latent model name,
#'   or `"general"`.
#' @param burnin_iter non-negative integer. Number of initial iterations to
#'   discard before sampling.
#' @param thinning positive integer thinning interval.
#' @param apply_transform logical; apply parameter transforms to user scale.
#' @param combine_chains logical; if `TRUE`, return one combined data.frame,
#'   otherwise return one data.frame per chain.
#'
#' @return A data.frame (or list of data.frames when `combine_chains = FALSE`)
#' with columns `.chain`, `.draw`, `.iter`, and one column per parameter.
#' @export
ngme_sgld_samples <- function(
    ngme,
    name = "all",
    burnin_iter = 0,
    thinning = 1,
    apply_transform = TRUE,
    combine_chains = TRUE) {
  stopifnot(
    inherits(ngme, "ngme"),
    is.character(name),
    length(name) == 1,
    is.numeric(burnin_iter),
    length(burnin_iter) == 1,
    is.finite(burnin_iter),
    burnin_iter >= 0,
    abs(burnin_iter - round(burnin_iter)) < sqrt(.Machine$double.eps),
    is.numeric(thinning),
    length(thinning) == 1,
    is.finite(thinning),
    thinning >= 1,
    abs(thinning - round(thinning)) < sqrt(.Machine$double.eps),
    is.logical(apply_transform),
    length(apply_transform) == 1,
    is.logical(combine_chains),
    length(combine_chains) == 1
  )

  chain_data <- .ngme_chain_trajectories(
    ngme = ngme,
    name = name,
    apply_transform = apply_transform
  )

  burnin_iter <- as.integer(round(burnin_iter))
  thinning <- as.integer(round(thinning))

  n_iter_vec <- vapply(chain_data$chains, nrow, integer(1))
  n_iter <- min(n_iter_vec)
  if (length(unique(n_iter_vec)) > 1) {
    warning(
      "Chains have different lengths; truncating to the minimum length.",
      call. = FALSE
    )
  }

  stopifnot(
    "burnin_iter must be smaller than the available trajectory length." =
      burnin_iter < n_iter
  )

  keep_iter <- seq.int(burnin_iter + 1L, n_iter, by = thinning)
  stopifnot(length(keep_iter) >= 1)

  per_chain <- lapply(seq_along(chain_data$chains), function(chain_idx) {
    mat <- chain_data$chains[[chain_idx]][seq_len(n_iter), , drop = FALSE]
    mat <- mat[keep_iter, , drop = FALSE]
    df <- as.data.frame(mat, check.names = FALSE)
    df$.chain <- chain_idx
    df$.draw <- seq_len(nrow(df))
    df$.iter <- keep_iter
    df <- df[, c(".chain", ".draw", ".iter", colnames(mat)), drop = FALSE]
    rownames(df) <- NULL
    df
  })

  if (isTRUE(combine_chains)) {
    ret <- do.call(rbind, per_chain)
    rownames(ret) <- NULL
  } else {
    ret <- per_chain
  }

  attr(ret, "name") <- name
  attr(ret, "burnin_iter") <- burnin_iter
  attr(ret, "thinning") <- thinning
  attr(ret, "apply_transform") <- apply_transform
  attr(ret, "n_chains") <- length(per_chain)
  attr(ret, "n_samples_per_chain") <- length(keep_iter)
  ret
}

#' Quantile Confidence Intervals from SGLD Samples
#'
#' Compute parameter-wise quantile confidence intervals from posterior-like
#' SGLD samples returned by [ngme_sgld_samples()].
#'
#' @param samples data.frame (or list of data.frames) returned by
#'   [ngme_sgld_samples()].
#' @param lower lower quantile probability (e.g. 0.025).
#' @param upper upper quantile probability (e.g. 0.975).
#'
#' @return A list with posterior-like point estimates (`estimates`), quantile
#' intervals (`ci`), and sample covariance matrix (`covariance`).
#' Class: `ngme_sgld_ci`.
#' @export
ngme_sgld_ci <- function(samples, lower = 0.025, upper = 0.975) {
  stopifnot(
    is.data.frame(samples) ||
      (is.list(samples) &&
        length(samples) >= 1 &&
        all(vapply(samples, is.data.frame, logical(1)))),
    is.numeric(lower),
    length(lower) == 1,
    is.finite(lower),
    lower > 0,
    lower < 1,
    is.numeric(upper),
    length(upper) == 1,
    is.finite(upper),
    upper > 0,
    upper < 1,
    lower < upper
  )

  if (is.list(samples) && !is.data.frame(samples)) {
    samples <- do.call(rbind, samples)
    rownames(samples) <- NULL
  }

  burnin_iter <- attr(samples, "burnin_iter")
  thinning <- attr(samples, "thinning")
  apply_transform <- attr(samples, "apply_transform")
  name <- attr(samples, "name")

  param_cols <- setdiff(colnames(samples), c(".chain", ".draw", ".iter"))
  stopifnot(length(param_cols) >= 1)
  stopifnot(
    "All parameter columns in `samples` must be numeric." =
      all(vapply(samples[, param_cols, drop = FALSE], is.numeric, logical(1)))
  )

  sample_mat <- as.matrix(samples[, param_cols, drop = FALSE])
  storage.mode(sample_mat) <- "double"
  colnames(sample_mat) <- param_cols

  estimates <- colMeans(sample_mat)
  std_error <- if (nrow(sample_mat) > 1) {
    apply(sample_mat, 2, stats::sd) / sqrt(nrow(sample_mat))
  } else {
    setNames(rep(0, ncol(sample_mat)), colnames(sample_mat))
  }

  qmat <- t(vapply(seq_len(ncol(sample_mat)), function(j) {
    stats::quantile(
      sample_mat[, j],
      probs = c(lower, upper),
      names = FALSE,
      na.rm = TRUE
    )
  }, numeric(2)))
  colnames(qmat) <- c("lower", "upper")
  rownames(qmat) <- colnames(sample_mat)

  cov_mat <- if (nrow(sample_mat) > 1) {
    stats::cov(sample_mat)
  } else {
    matrix(0, ncol(sample_mat), ncol(sample_mat))
  }
  dimnames(cov_mat) <- list(colnames(sample_mat), colnames(sample_mat))

  n_chains <- attr(samples, "n_chains")
  n_samples_per_chain <- attr(samples, "n_samples_per_chain")
  if (is.null(n_chains)) {
    n_chains <- if (".chain" %in% colnames(samples)) {
      length(unique(samples$.chain))
    } else {
      1L
    }
  }
  if (is.null(n_samples_per_chain)) {
    n_samples_per_chain <- if (".chain" %in% colnames(samples)) {
      as.integer(min(table(samples$.chain)))
    } else {
      nrow(samples)
    }
  }

  ret <- list(
    estimates = stats::setNames(estimates, colnames(sample_mat)),
    std_error = stats::setNames(as.numeric(std_error), colnames(sample_mat)),
    ci = qmat,
    covariance = cov_mat,
    lower = lower,
    upper = upper,
    burnin_iter = burnin_iter,
    thinning = thinning,
    n_chains = n_chains,
    n_samples_per_chain = n_samples_per_chain,
    n_samples = nrow(sample_mat),
    name = name,
    apply_transform = apply_transform,
    samples = samples
  )

  class(ret) <- "ngme_sgld_ci"
  ret
}

#' Summary for Batch-Means CI Results
#'
#' Summarize an object of class `ngme_batch_ci`, including parameter means,
#' standard errors, confidence intervals, and covariance matrix.
#'
#' @param object an object of class `ngme_batch_ci`.
#' @param digits number of digits used for printing.
#' @param ... currently unused.
#'
#' @return A list with summary table and covariance matrix.
#' @export
summary.ngme_batch_ci <- function(object, digits = 4, ...) {
  stopifnot(inherits(object, "ngme_batch_ci"))

  conf_int <- as.matrix(object$conf_int)
  est <- as.numeric(object$estimates)
  se <- as.numeric(object$std_error)
  pnames <- names(object$estimates)
  if (is.null(pnames)) {
    pnames <- rownames(conf_int)
  }
  if (is.null(pnames)) {
    pnames <- paste0("param_", seq_along(est))
  }

  table <- data.frame(
    parameter = pnames,
    mean = est,
    std_error = se,
    lower = conf_int[, "lower"],
    upper = conf_int[, "upper"],
    check.names = FALSE,
    row.names = NULL
  )

  ret <- list(
    table = table,
    covariance = object$covariance,
    level = object$level,
    alpha = object$alpha,
    M = object$M,
    N = object$N,
    burnin_iter = object$burnin_iter,
    n_eff = object$n_eff,
    n_chains = object$n_chains,
    n_iterations = object$n_iterations,
    n_iterations_total = object$n_iterations_total,
    name = object$name,
    apply_transform = object$apply_transform,
    digits = digits
  )

  if (!is.null(object$raw_estimates)) {
    raw_conf <- as.matrix(object$raw_conf_int)
    raw_names <- names(object$raw_estimates)
    if (is.null(raw_names)) {
      raw_names <- rownames(raw_conf)
    }
    if (is.null(raw_names)) {
      raw_names <- paste0("param_", seq_along(object$raw_estimates))
    }

    ret$raw_table <- data.frame(
      parameter = raw_names,
      mean = as.numeric(object$raw_estimates),
      std_error = as.numeric(object$raw_std_error),
      lower = raw_conf[, "lower"],
      upper = raw_conf[, "upper"],
      check.names = FALSE,
      row.names = NULL
    )
    ret$raw_covariance <- object$raw_covariance
  }

  class(ret) <- "summary.ngme_batch_ci"
  ret
}

#' @export
print.summary.ngme_batch_ci <- function(x, ...) {
  .format_table <- function(tbl, digits) {
    out <- tbl
    is_num <- vapply(out, is.numeric, logical(1))
    out[is_num] <- lapply(out[is_num], signif, digits = digits)
    out
  }

  cat("Batch-means CI summary\n")
  cat(
    "  level =", x$level,
    "| alpha =", x$alpha,
    "| M =", x$M,
    "| n_eff =", x$n_eff,
    "| n_chains =", x$n_chains, "\n"
  )
  if (!is.null(x$name)) {
    cat("  block =", x$name, "\n")
  }
  if (!is.null(x$apply_transform)) {
    cat("  apply_transform =", x$apply_transform, "\n")
  }

  cat("\nParameter summary:\n")
  print(.format_table(x$table, x$digits), row.names = FALSE)

  cat("\nCovariance matrix:\n")
  print(signif(x$covariance, x$digits))

  if (!is.null(x$raw_table)) {
    cat("\nRaw-space parameter summary (pre-transform):\n")
    print(.format_table(x$raw_table, x$digits), row.names = FALSE)
    cat("\nRaw-space covariance matrix:\n")
    print(signif(x$raw_covariance, x$digits))
  }

  invisible(x)
}

.batch_means_boundaries <- function(n_iter, alpha, M = NULL, N = NULL) {
  stopifnot(
    is.numeric(n_iter),
    length(n_iter) == 1,
    n_iter >= 2,
    is.numeric(alpha),
    length(alpha) == 1,
    alpha > 0.5,
    alpha < 1
  )

  if (is.null(M)) {
    M <- floor(n_iter^((1 - alpha) / 2))
  }
  M <- as.integer(M)

  stopifnot(
    is.numeric(M),
    length(M) == 1,
    M >= 1
  )

  if (is.null(N)) {
    N <- (n_iter^(1 - alpha)) / (M + 1)
  }

  stopifnot(
    is.numeric(N),
    length(N) == 1,
    N > 0
  )

  end <- floor(((seq_len(M + 1L)) * N)^(1 / (1 - alpha)))
  end <- pmin(pmax(as.integer(end), 1L), as.integer(n_iter))
  end <- cummax(end)
  end[length(end)] <- as.integer(n_iter)

  # Remove zero-size batches induced by integer rounding
  keep <- c(TRUE, diff(end) > 0)
  end <- end[keep]

  start <- c(1L, head(end, -1) + 1L)
  size <- end - start + 1L

  list(
    start = start,
    end = end,
    size = size,
    N = N
  )
}

.ngme_chain_trajectories <- function(ngme, name = "all", apply_transform = TRUE) {
  ngme_rep <- ngme$replicates[[1]]

  if (!identical(name, "all")) {
    tr <- get_trace_trajectories(
      ngme = ngme,
      name = name,
      apply_transform = apply_transform
    )

    tr_list <- unname(tr$trajectories)
    parameter_names <- make.unique(tr$parameter_names)

    chains <- lapply(seq_len(tr$n_chains), function(chain_idx) {
      mat <- do.call(cbind, lapply(tr_list, function(x) {
        x[, chain_idx, drop = FALSE]
      }))
      mat <- as.matrix(mat)
      colnames(mat) <- parameter_names
      mat
    })

    transforms <- tr$transformations
    stopifnot(length(transforms) == length(parameter_names))
    return(list(chains = chains, parameter_names = parameter_names, transforms = transforms))
  }

  latent_names <- names(ngme_rep$models)
  parts <- lapply(latent_names, function(lat_name) {
    latent <- ngme_rep$models[[lat_name]]
    traj <- attr(latent, "lat_traj")
    stopifnot(
      "Please run ngme() with store_traj=TRUE before using ngme_batch_ci()." =
        !is.null(traj)
    )

    ts <- get_latent_info(latent)
    ts$name <- paste0(ts$name, " (", lat_name, ")")

    list(traj = traj, ts = ts)
  })

  block_traj <- attr(ngme_rep, "block_traj")
  stopifnot(
    "Please run ngme() with store_traj=TRUE before using ngme_batch_ci()." =
      !is.null(block_traj)
  )

  ts_block <- get_noise_info(ngme_rep$noise)
  fe_names <- if (length(ngme_rep$feff) == 0) {
    NULL
  } else if (!is.null(names(ngme_rep$feff)) &&
    all(!is.na(names(ngme_rep$feff))) &&
    all(nzchar(names(ngme_rep$feff)))) {
    names(ngme_rep$feff)
  } else {
    paste("fixed effect", seq_len(length(ngme_rep$feff)))
  }
  fe_trans <- rep(list(identity), length(ngme_rep$feff))
  ts_block$name <- c(ts_block$name, fe_names)
  ts_block$trans <- c(ts_block$trans, fe_trans)

  parts <- c(parts, list(list(traj = block_traj, ts = ts_block)))

  n_chain <- length(parts[[1]]$traj)
  stopifnot(
    "All trajectory components must have the same number of chains." =
      all(vapply(parts, function(x) length(x$traj) == n_chain, logical(1)))
  )

  parameter_names <- make.unique(unlist(lapply(parts, function(x) x$ts$name), use.names = FALSE))
  parameter_transforms <- unlist(lapply(parts, function(x) x$ts$trans), recursive = FALSE, use.names = FALSE)
  stopifnot(length(parameter_transforms) == length(parameter_names))

  chains <- lapply(seq_len(n_chain), function(chain_idx) {
    mats <- lapply(parts, function(part) {
      mat <- as.matrix(part$traj[[chain_idx]])
      if (apply_transform) {
        mat <- .apply_parameter_transforms(mat, part$ts$trans)
      }
      mat
    })

    min_iter <- min(vapply(mats, ncol, integer(1)))
    mats <- lapply(mats, function(x) x[, seq_len(min_iter), drop = FALSE])

    chain_mat <- t(do.call(rbind, mats))
    colnames(chain_mat) <- parameter_names
    chain_mat
  })

  list(chains = chains, parameter_names = parameter_names, transforms = parameter_transforms)
}

.apply_parameter_transforms <- function(mat, transforms) {
  stopifnot(length(transforms) == nrow(mat))

  for (i in seq_len(nrow(mat))) {
    ff <- transforms[[i]]
    if (is.null(ff)) {
      next
    }

    vals <- ff(as.numeric(mat[i, ]))
    stopifnot(length(vals) == ncol(mat))
    mat[i, ] <- vals
  }

  mat
}

.delta_transform_batch_ci <- function(ci, transforms) {
  stopifnot(
    is.list(ci),
    is.numeric(ci$estimates),
    is.matrix(ci$covariance),
    is.list(transforms),
    length(transforms) == length(ci$estimates)
  )

  param_names <- names(ci$estimates)
  mean_raw <- as.numeric(ci$estimates)

  mean_trans <- numeric(length(mean_raw))
  deriv_trans <- numeric(length(mean_raw))

  for (i in seq_along(mean_raw)) {
    ff <- transforms[[i]]
    if (is.null(ff)) {
      ff <- identity
    }
    mean_trans[i] <- ff(mean_raw[i])
    deriv_trans[i] <- .delta_derivative(ff, mean_raw[i])
  }

  jacobian <- diag(deriv_trans, nrow = length(deriv_trans), ncol = length(deriv_trans))
  cov_trans <- jacobian %*% ci$covariance %*% jacobian

  std_error <- sqrt(pmax(diag(cov_trans), 0) / (ci$n_eff * ci$n_chains))
  conf_int <- cbind(
    lower = mean_trans - ci$z_value * std_error,
    upper = mean_trans + ci$z_value * std_error
  )

  names(mean_trans) <- param_names
  names(std_error) <- param_names
  dimnames(cov_trans) <- list(param_names, param_names)
  rownames(conf_int) <- param_names

  ci$estimates <- mean_trans
  ci$covariance <- cov_trans
  ci$std_error <- std_error
  ci$conf_int <- conf_int

  if (!is.null(ci$chain_results)) {
    ci$chain_results <- lapply(ci$chain_results, function(cr) {
      cr_mean <- as.numeric(cr$mean)
      cr_mean_trans <- numeric(length(cr_mean))
      cr_deriv <- numeric(length(cr_mean))
      for (j in seq_along(cr_mean)) {
        ff <- transforms[[j]]
        if (is.null(ff)) {
          ff <- identity
        }
        cr_mean_trans[j] <- ff(cr_mean[j])
        cr_deriv[j] <- .delta_derivative(ff, cr_mean[j])
      }
      cr_jacobian <- diag(cr_deriv, nrow = length(cr_deriv), ncol = length(cr_deriv))
      cr$mean <- stats::setNames(cr_mean_trans, param_names)
      cr$covariance <- cr_jacobian %*% cr$covariance %*% cr_jacobian
      dimnames(cr$covariance) <- list(param_names, param_names)
      if (!is.null(cr$batch_means) && is.matrix(cr$batch_means)) {
        for (j in seq_along(transforms)) {
          ff <- transforms[[j]]
          if (is.null(ff)) {
            ff <- identity
          }
          cr$batch_means[, j] <- ff(as.numeric(cr$batch_means[, j]))
        }
      }
      cr
    })
  }

  ci
}

.delta_derivative <- function(ff, x) {
  h <- 1e-6 * max(1, abs(x))
  if (!is.finite(h) || h <= 0) {
    h <- 1e-6
  }

  x_plus <- x + h
  x_minus <- x - h
  y_plus <- ff(x_plus)
  y_minus <- ff(x_minus)

  if (is.finite(y_plus) && is.finite(y_minus)) {
    return((y_plus - y_minus) / (2 * h))
  }

  # fallback to one-sided differences for domain-constrained transforms
  y0 <- ff(x)
  if (is.finite(y_plus) && is.finite(y0)) {
    return((y_plus - y0) / h)
  }
  if (is.finite(y0) && is.finite(y_minus)) {
    return((y0 - y_minus) / h)
  }

  stop("Failed to compute Delta-method derivative for parameter transform.")
}

#' Predict function of ngme2
#' predict using ngme after estimation
#'
#' @param object a ngme object
#' @param map a named list (or dataframe) of the locations to make the prediction
#' @param data a data.frame or matrix of covariates (used for fixed effects)
#'  names(loc) corresponding to the name each latent model
#'  vector or matrix (n * 2) for spatial coords
#' @param type what type of prediction, c("fe", "lp", <model_name>)
#' "fe" is fixed effect prediction
#' <model_name> is prediction of a specific model
#' "lp" is linear predictor (including fixed effect and all sub-models)
#' "response" is the linear predictor plus a fresh measurement-noise draw
#' @param group which filed to predict
#'   (used for bivariate model, should be of same length as map)
#' @param estimator what type of estimator. Options include:
#'   - "mean", "median", "mode", "sd": standard estimators
#'   - "0.XXXq": any quantile specified as probability (e.g., "0.025q", "0.5q", "0.975q")
#' @param sampling_size size of posterior sampling
#' @param burnin_size size of posterior burnin
#' @param seed random seed
#' @param train_idx optional vector of training indices to use for posterior sampling.
#'   If provided, only these indices from the original data will be used for training,
#'   similar to cross-validation. If NULL, uses all original training data.
#' @param chain_combine how to combine multiple optimization chains:
#'   \itemize{
#'     \item \code{"param_mean"}: default behavior using the fitted object parameters.
#'     \item \code{"predictive_average"}: run prediction for each optimization chain and
#'       average predictions across chains.
#'   }
#' @param return_samples logical; when `TRUE`, attach sample draws for the
#'   requested output in `attr(ret, "samples")`. For `type = "response"`, the
#'   attached samples are response predictive draws.
#' @param ... additional arguments (currently unused)
#'
#' @return a list of outputs contains estimation of operator paramters, noise parameters
#' @export
predict.ngme <- function(
    object,
    map,
    data = NULL,
    type = "lp",
    group = NULL,
    estimator = c("mean", "sd", "0.05q", "0.95q", "median", "mode"),
    sampling_size = 500,
    burnin_size = 100,
    seed = Sys.time(),
    train_idx = NULL,
    chain_combine = c("param_mean", "predictive_average"),
    return_samples = FALSE,
    ...) {
  chain_combine <- match.arg(chain_combine)
  stopifnot(
    sampling_size > 0,
    "Make sure the object is of class 'ngme'." = inherits(object, "ngme")
  )

  if (chain_combine == "predictive_average") {
    chain_fits <- resolve_ngme_chain_fits(object)
    if (length(chain_fits) > 1) {
      chain_preds <- lapply(seq_along(chain_fits), function(i) {
        predict_ngme_param_mean(
          object = chain_fits[[i]],
          map = map,
          data = data,
          type = type,
          group = group,
          estimator = estimator,
          sampling_size = sampling_size,
          burnin_size = burnin_size,
          seed = seed + i - 1L,
          train_idx = train_idx,
          return_samples = return_samples
        )
      })

      ret <- list()
      for (est in names(chain_preds[[1]])) {
        ret[[est]] <- rowMeans(do.call(cbind, lapply(chain_preds, `[[`, est)))
      }

      samples_list <- lapply(chain_preds, function(x) attr(x, "samples"))
      if (all(vapply(samples_list, is.matrix, logical(1)))) {
        attr(ret, "samples") <- do.call(cbind, samples_list)
      }

      return(ret)
    }
  }

  predict_ngme_param_mean(
    object = object,
    map = map,
    data = data,
    type = type,
    group = group,
    estimator = estimator,
    sampling_size = sampling_size,
    burnin_size = burnin_size,
    seed = seed,
    train_idx = train_idx,
    return_samples = return_samples
  )
}


predict_ngme_param_mean <- function(
    object,
    map,
    data = NULL,
    type = "lp",
    group = NULL,
    estimator = c("mean", "sd", "0.05q", "0.95q", "median", "mode"),
    sampling_size = 500,
    burnin_size = 100,
    seed = Sys.time(),
    train_idx = NULL,
    return_samples = FALSE) {
  fm <- attr(object, "fit")$formula
  ngme <- object$replicates[[1]]
  seed_int <- normalize_prediction_seed(seed)
  type_names <- prediction_type_names(type, ngme)
  requested_model_names <- prediction_requested_model_names(type_names, ngme)

  # If train_idx is provided, subset the data for posterior sampling
  if (!is.null(train_idx)) {
    # Validate train_idx
    n_data <- attr(object, "fit")$n_data
    if (any(train_idx > n_data) || any(train_idx < 1)) {
      stop("train_idx out of bounds. Data has ", n_data, " observations.")
    }

    # Subset training data (similar to cross-validation logic)
    ngme$X <- ngme$X[train_idx, , drop = FALSE]
    ngme$Y <- ngme$Y[train_idx]
    ngme$noise <- subset_noise(
      ngme$noise,
      sub_idx = train_idx, compute_corr = TRUE
    )

    # Subset A matrices for training data
    for (i in seq_along(ngme$models)) {
      ngme$models[[i]]$A <- ngme$models[[i]]$A[train_idx, , drop = FALSE]
    }
  }

  samples_W <- sampling_cpp(ngme,
    n = sampling_size,
    n_burnin = burnin_size,
    posterior = TRUE,
    seed = seed_int
  )[["W"]]

  # Convert samples_W from list to matrix
  if (is.list(samples_W)) {
    samples_W <- do.call(cbind, samples_W)
  }

  use_output_samples <- identical(type, "response") ||
    isTRUE(return_samples) ||
    prediction_needs_output_samples_for_composite_noise(
      type_names = type_names,
      estimator = estimator,
      ngme = ngme
    )

  if (use_output_samples) {
    output_samples <- build_prediction_output_samples(
      ngme = ngme,
      fm = fm,
      map = map,
      data = data,
      type = type,
      group = group,
      samples_W = samples_W
    )

    if (identical(type, "response")) {
      if (isTRUE(ngme$noise$corr_measurement)) {
        stop("type = 'response' does not currently support correlated measurement noise.")
      }

      noise_draws <- simulate_predictive_noise_draws(
        noise = ngme$noise,
        n_pred = nrow(output_samples),
        n_samp = ncol(output_samples),
        seed = seed_int + 100000L
      )
      output_samples <- ensure_prediction_sample_matrix(output_samples + noise_draws)
    }

    ret <- summarize_prediction_samples(output_samples, estimator)
    attr(ret, "samples") <- output_samples
    attr(ret, "latent_samples") <- samples_W
    return(ret)
  }

  ret <- NULL
  for (estimator in estimator) {
    # Check if estimator matches pattern like "0.025q", "0.975q", etc.
    if (grepl("^0\\.[0-9]*[1-9][0-9]*q$", estimator)) {
      # Extract the probability from the estimator string
      quantile_prob <- as.numeric(gsub("q$", "", estimator))
      # Validate the quantile probability (should be between 0 and 1, exclusive)
      if (quantile_prob <= 0 || quantile_prob >= 1) {
        stop("Quantile probability should be between 0 and 1 (exclusive). Got: ", quantile_prob)
      }
      post_W <- apply(samples_W, 1, function(x) {
        quantile(x, quantile_prob)
      })
    } else {
      post_W <- switch(estimator,
        "mean"      = apply(samples_W, 1, mean),
        "median"    = apply(samples_W, 1, median),
        "sd"        = apply(samples_W, 1, sd),
        "mode"      = apply(samples_W, 1, emprical_mode),
        stop("No such estimator available: ", estimator)
      )
    }

    # update post W (notice here W is concated)
    j <- 1
    for (i in seq_along(ngme$models)) {
      sz <- ngme$models[[i]]$W_size
      ngme$models[[i]][[estimator]] <- post_W[j:(j + sz - 1)]
      j <- j + sz
    }

    AW <- list()
    if (length(requested_model_names) > 0) {
      validate_prediction_map(map, requested_model_names)

      for (model_name in requested_model_names) {
        model <- ngme$models[[model_name]]
        loc <- map[[model_name]]
        group_use <- group
        if (inherits(model$operator$mesh, "metric_graph")) {
          loc <- as.data.frame(loc)
        } else if (model$model != "tp") {
          loc <- as.matrix(loc)
        } else {
          stopifnot(
            length(loc) == 2, # map 1 and map 2
            length_map(loc[[1]]) == length_map(loc[[2]])
          )
        }

        if (model$model %in% c("bv", "bv2", "bv_matern") && !is.null(group_use)) {
          group_use <- rep(group_use, length.out = length_map(loc))
        }

        A <- ngme_build_A(
          model$model,
          model$operator$mesh,
          loc,
          model$operator,
          group_use,
          group_levels = levels(ngme$group)
        )
        A <- expand_prediction_A_for_composite_noise(A, model)

        AW[[model_name]] <- as.numeric(A %*% model[[estimator]])
      }
    }

    preds <- 0
    for (i in seq_along(type_names)) {
      name <- type_names[[i]]
      if (name == "fe" && length(ngme$feff) > 0) {
        X_pred <- if (is.null(data) && attr(terms(fm), "intercept")) {
          matrix(1, nrow = length(AW[[1]]), ncol = 1)
        } else {
          stopifnot("Please provide covariates for predictions" = !is.null(data))
          # build plain_fm
          tf <- terms.formula(fm, specials = c("f"))
          terms <- attr(tf, "term.labels")
          intercept <- attr(tf, "intercept")
          spec_order <- attr(tf, "specials")$f - 1
          fixf <- if (length(spec_order) == 0) terms else terms[-spec_order]
          plain_fm_str <- paste("~", intercept, paste(c("", fixf), collapse = " + "))
          plain_fm <- formula(plain_fm_str)
          model.matrix(plain_fm, data = data)
        }
        if (!is.null(data) && nrow(X_pred) < nrow(data)) {
          stop("The data has NA values in the covariates.")
        }
        preds <- preds + as.numeric(X_pred %*% ngme$feff)
      } else if (name %in% names(ngme$models)) {
        preds <- preds + AW[[name]]
      }
    }
    ret[[estimator]] <- preds
  }
  attr(ret, "samples") <- samples_W
  ret
}

normalize_prediction_seed <- function(seed) {
  if (inherits(seed, "POSIXt")) {
    seed <- as.numeric(seed)
  }

  seed <- as.numeric(seed)[1]
  if (!is.finite(seed)) {
    stop("seed must be coercible to a finite numeric value.")
  }

  as.integer(abs(seed) %% 2147483647)
}


build_prediction_output_samples <- function(
    ngme,
    fm,
    map,
    data,
    type,
    group,
    samples_W) {
  n_samp <- ncol(samples_W)

  type_names <- prediction_type_names(type, ngme)
  requested_model_names <- prediction_requested_model_names(type_names, ngme)
  needs_map <- length(requested_model_names) > 0

  if (needs_map) {
    validate_prediction_map(map, requested_model_names)
  }

  fixed_effect <- NULL
  n_pred <- NULL
  if ("fe" %in% type_names && length(ngme$feff) > 0) {
    X_pred <- build_predict_model_matrix(
      fm = fm,
      ngme = ngme,
      data = data,
      n_pred = if (!is.null(data)) {
        nrow(data)
      } else {
        prediction_map_size(map, requested_model_names)
      }
    )
    fixed_effect <- as.numeric(X_pred %*% ngme$feff)
    n_pred <- length(fixed_effect)
  }

  model_draws <- list()
  if (needs_map) {
    j <- 1
    for (i in seq_along(ngme$models)) {
      model <- ngme$models[[i]]
      sz <- model$W_size
      if (model$name %in% requested_model_names) {
        loc <- map[[model$name]]
        if (inherits(model$operator$mesh, "metric_graph")) {
          loc <- as.data.frame(loc)
        } else if (model$model != "tp") {
          loc <- as.matrix(loc)
        } else {
          stopifnot(
            length(loc) == 2,
            length_map(loc[[1]]) == length_map(loc[[2]])
          )
        }

        A <- ngme_build_A(
          model$model,
          model$operator$mesh,
          loc,
          model$operator,
          group,
          group_levels = levels(ngme$group)
        )
        A <- expand_prediction_A_for_composite_noise(A, model)

        model_draws[[model$name]] <- A %*% samples_W[j:(j + sz - 1), , drop = FALSE]
        if (is.null(n_pred)) n_pred <- nrow(model_draws[[model$name]])
      }
      j <- j + sz
    }
  }

  if (is.null(n_pred)) {
    stop("Could not determine prediction size. Please provide either `map` or `data`.")
  }

  output_samples <- matrix(0, nrow = n_pred, ncol = n_samp)

  for (name in type_names) {
    if (name == "fe" && !is.null(fixed_effect)) {
      output_samples <- output_samples +
        matrix(rep(fixed_effect, n_samp), nrow = n_pred, ncol = n_samp)
    } else if (name %in% names(model_draws)) {
      output_samples <- output_samples + model_draws[[name]]
    }
  }

  ensure_prediction_sample_matrix(output_samples)
}


expand_prediction_A_for_composite_noise <- function(A, model) {
  composite_noise <- all(model$noise$noise_type %in% c(
    "normal_nig", "normal_gal", "nig_gal"
  ))

  if (composite_noise && ncol(A) * 2 == model$W_size) {
    return(cbind(A, A))
  }

  A
}


prediction_needs_output_samples_for_composite_noise <- function(
    type_names,
    estimator,
    ngme) {
  if (all(estimator == "mean")) {
    return(FALSE)
  }

  requested_model_names <- prediction_requested_model_names(type_names, ngme)
  any(vapply(requested_model_names, function(model_name) {
    model <- ngme$models[[model_name]]
    all(model$noise$noise_type %in% c("normal_nig", "normal_gal", "nig_gal"))
  }, logical(1)))
}


build_predict_model_matrix <- function(fm, ngme, data, n_pred = NULL) {
  if (is.null(data) && attr(terms(fm), "intercept")) {
    if (is.null(n_pred)) stop("Please provide covariates for predictions")
    return(matrix(1, nrow = n_pred, ncol = 1))
  }

  stopifnot("Please provide covariates for predictions" = !is.null(data))
  tf <- terms.formula(fm, specials = c("f"))
  terms <- attr(tf, "term.labels")
  intercept <- attr(tf, "intercept")
  spec_order <- attr(tf, "specials")$f - 1
  fixf <- if (length(spec_order) == 0) terms else terms[-spec_order]
  plain_fm_str <- paste("~", intercept, paste(c("", fixf), collapse = " + "))
  plain_fm <- formula(plain_fm_str)
  X_pred <- model.matrix(plain_fm, data = data)
  if (!is.null(data) && nrow(X_pred) < nrow(data)) {
    stop("The data has NA values in the covariates.")
  }
  X_pred
}


summarize_prediction_samples <- function(samples, estimator) {
  samples <- ensure_prediction_sample_matrix(samples)
  ret <- list()
  for (est in estimator) {
    if (grepl("^0\\.[0-9]*[1-9][0-9]*q$", est)) {
      quantile_prob <- as.numeric(gsub("q$", "", est))
      if (quantile_prob <= 0 || quantile_prob >= 1) {
        stop("Quantile probability should be between 0 and 1 (exclusive). Got: ", quantile_prob)
      }
      ret[[est]] <- apply(samples, 1, quantile, quantile_prob)
    } else {
      ret[[est]] <- switch(est,
        "mean" = rowMeans(samples),
        "median" = apply(samples, 1, median),
        "sd" = apply(samples, 1, sd),
        "mode" = apply(samples, 1, emprical_mode),
        stop("No such estimator available: ", est)
      )
    }
  }
  ret
}


ensure_prediction_sample_matrix <- function(samples) {
  if (is.matrix(samples)) {
    return(samples)
  }

  if (inherits(samples, "Matrix")) {
    return(as.matrix(samples))
  }

  if (is.null(dim(samples))) {
    return(matrix(samples, ncol = 1))
  }

  as.matrix(samples)
}


prediction_type_names <- function(type, ngme) {
  if (type %in% c("lp", "response")) c("fe", names(ngme$models)) else type
}


prediction_requested_model_names <- function(type_names, ngme) {
  intersect(type_names, names(ngme$models))
}


validate_prediction_map <- function(map, required_model_names) {
  if (length(required_model_names) == 0) {
    return(invisible(NULL))
  }

  stopifnot(
    "map should be a named list (name for each model)" = is.list(map) && !is.null(names(map))
  )

  missing_names <- setdiff(required_model_names, names(map))
  if (length(missing_names) > 0) {
    stop(
      "The loction for model ",
      paste(missing_names, collapse = ", "),
      " is not provided"
    )
  }

  invisible(NULL)
}


prediction_map_size <- function(map, model_names = names(map)) {
  if (is.null(map) || length(model_names) == 0) {
    return(NULL)
  }

  first_name <- model_names[[1]]
  loc <- map[[first_name]]
  if (is.null(loc)) {
    return(NULL)
  }

  if (is.list(loc) && !inherits(loc, c("data.frame", "matrix"))) {
    stopifnot(
      length(loc) == 2,
      length_map(loc[[1]]) == length_map(loc[[2]])
    )
    return(length_map(loc[[1]]))
  }

  length_map(loc)
}


simulate_predictive_noise_draws <- function(noise, n_pred, n_samp, seed) {
  if (length(noise$noise_type) == 2) {
    stop("type = 'response' does not currently support bivariate measurement noise.")
  }

  h_vec <- rep(1, n_pred)
  mu_vec <- recycle_predictive_noise_param(
    as.numeric(noise$B_mu %*% noise$theta_mu),
    n_pred = n_pred,
    param_name = "mu"
  )
  sigma_vec <- recycle_predictive_noise_param(
    as.numeric(exp(noise$B_sigma %*% noise$theta_sigma)),
    n_pred = n_pred,
    param_name = "sigma"
  )
  nu_vec <- recycle_predictive_noise_param(
    as.numeric(noise$nu_lower_bound + exp(noise$B_nu %*% noise$theta_nu)),
    n_pred = n_pred,
    param_name = "nu"
  )

  noise_draws <- vapply(seq_len(n_samp), function(i) {
    simulate_noise(
      noise_type = noise$noise_type,
      h_vec = h_vec,
      mu_vec = mu_vec,
      sigma_vec = sigma_vec,
      nu_vec = nu_vec,
      seed = seed + i,
      single_V = noise$single_V
    )
  }, numeric(n_pred))

  matrix(noise_draws, nrow = n_pred, ncol = n_samp)
}


recycle_predictive_noise_param <- function(param_vec, n_pred, param_name) {
  if (length(param_vec) == n_pred) {
    return(param_vec)
  }

  if (length(param_vec) == 1) {
    return(rep(param_vec, n_pred))
  }

  if (all(isTRUE(all.equal(param_vec, rep(param_vec[[1]], length(param_vec)))))) {
    return(rep(param_vec[[1]], n_pred))
  }

  stop(
    "type = 'response' requires measurement-noise ", param_name,
    " to be scalar or already match prediction length. Got length ",
    length(param_vec), " for ", n_pred, " prediction locations."
  )
}

# spatial data
library(sp)
data(meuse)
head(meuse)
str(meuse)
coordinates(meuse) <- ~ x + y
class(meuse)
summary(meuse)

plot(meuse)
title("points")

#  --- grid ---

data(meuse.grid)
head(meuse.grid)
str(meuse.grid)

coordinates(meuse.grid) <- ~ x + y
meuse.grid <- as(meuse.grid, "SpatialPixels")
image(meuse.grid, col = "grey")
title("grid")

# --- river ---

data(meuse.riv)
meuse.lst <- list(Polygons(list(Polygon(meuse.riv)), "meuse.riv"))
meuse.pol <- SpatialPolygons(meuse.lst)
plot(meuse.pol, col = "grey")


image(meuse.grid, col = "lightgrey")
plot(meuse.grid, col = "lightgrey", add = TRUE)
plot(meuse.pol, col = "grey", add = TRUE)
plot(meuse, add = TRUE)

##############################
# create mesh
library(INLA)
mesh <- fmesher::fm_mesh_2d(loc = meuse@coords, max.n = 200, max.edge = c(1, 2))
plot(mesh)
mesh$n

str(mesh1)
points(meuse, col = "red")

head(meuse)

prdomain <- inla.nonconvex.hull(meuse@coords, -0.03, -0.05, resolution = c(100, 100))
prmesh <- fmesher::fm_mesh_2d(meuse@coords, boundary = prdomain, max.n = 400, cutoff = 50)
prmesh$n
plot(prmesh)

##############################
# fit using ngme
bubble(meuse, "copper")

spde <- model_matern(loc = meuse@coords, mesh = mesh)

out1 <- ngme(
  zinc ~ 1 # + f(dist.m, model="rw1", noise=noise_normal(), name="rw")
    + f(model = spde, name = "spde", noise = noise_nig()),
  data = meuse@data,
  control = control_opt(
    iterations = 100,
    n_parallel_chain = 4,
    print_check_info = TRUE
  ),
  debug = TRUE
)

out1$models[[1]]$h

traceplot(out1, "rw")
traceplot(out1, "rw")
traceplot(out1, "spde")

class(meuse)

library(INLA)

n_meuse <- nrow(meuse)
spde_inla <- inla.spde2.matern(mesh = prmesh, loc = meuse@coords)
res_inla <- inla(
  formula = zinc ~ 0 +
    f(dist.m, model = "rw1") +
    f(i, model = spde_inla),
  data = list(dist.m = meuse$dist.m, i = 1:spde_inla$n.spde, zinc = meuse$zinc),
  family = "gaussian",
  control.predictor = list(1, A = inla.spde.make.A(mesh = prmesh, loc = meuse@coords))
)
list(dist.m = meuse$dist.m, i = 1:spde_inla$n.spde, zinc = meuse$zinc)
summary(res_inla)

meuse@data$copper

#' Inference and prediction for mixed effects models with flexible non-Gaussian and Gaussian distributions.
#'
#' @name ngme2
#' @keywords internal 
"_PACKAGE"
#' @import Rcpp
#' @importFrom Rcpp evalCpp
#' @importFrom stats aggregate as.formula delete.response density dgamma dnorm formula median model.frame model.matrix qgamma qlogis quantile rnorm rgamma sd setNames simulate terms terms.formula
#' @importFrom methods as
#' @importFrom utils capture.output str head tail modifyList
#' @importFrom ggplot2 ggplot aes geom_line xlab ylab guides labs geom_hline
#' @importFrom rlang .data
#' @importFrom fmesher fm_mesh_1d fm_mesh_2d fm_basis fm_fem
#' @importFrom graphics hist
#' @importFrom stats dist
#' @useDynLib ngme2, .registration = TRUE
#' @name ngme2
NULL

utils::globalVariables(c("density", "mean_moving_avg"))

.onAttach <- function(libname, pkgname) {
  version <- utils::packageVersion("ngme2")
   packageStartupMessage(
    "This is ngme2 of version ", version, "\n",
    "- See our homepage: https://davidbolin.github.io/ngme2 for more details."
  )
}

#' @title Test ngme function
#'
#' @description
#' Test ngme function for different models
#' @param model model name
#' @param n_obs_per_rep number of observation per replicate
#' @param n_replicate number of replicate
#' @param numer_grad numerical gradient
#' @param max.n maximum number for building mesh
#' @param debug debug mode
#' @param debug_f debug mode for latent process
#' @param f_noise noise function
#' @param n_gibbs_samples number of gibbs samples
#' @param family family of noise
#' @param seed seed
#' @param start start value for optimization
#' @param fix_theta_K fix theta_K
#' @param fix_theta_mu fix theta_mu
#' @param fix_theta_sigma fix theta_sigma
#' @param fix_theta_nu fix nu
#' @param control_opt control options for optimization, see \code{\link{control_opt}}
#' @keywords internal
test_ngme <- function(
    model,
    n_obs_per_rep,
    n_replicate = 1,
    control_opt = NULL,
    numer_grad = TRUE,
    f_noise = noise_nig(mu = -2, sigma = 1.5, nu = 0.5),
    n_gibbs_samples = 5,
    family = "normal",
    max.n = 1000,
    debug = FALSE,
    debug_f = FALSE,
    start = NULL,
    fix_theta_K = FALSE,
    fix_theta_mu = FALSE,
    fix_theta_sigma = FALSE,
    fix_theta_nu = FALSE,
    seed = Sys.time()) {
  if (is.null(control_opt)) control_opt <- control_opt()

  set.seed(seed)
  # create 2d mesh
  if (model %in% c("matern", "bvmatern", "ar1+matern")) {
    pl01 <- cbind(c(0, 1, 1, 0, 0) * 10, c(0, 0, 1, 1, 0) * 5)
    mesh <- fmesher::fm_mesh_2d(
      loc.domain = pl01, cutoff = 0.2,
      max.edge = c(0.4, 10),
      max.n = max.n
    )
    if (isTRUE(debug)) {
      message("nodes of mesh = ", mesh$n)
    }
  }

  stopifnot(length(f_noise$noise_type) == 1)
  f_fm_noise <- f_noise
  f_fm_noise$fix_theta_mu <- fix_theta_mu
  f_fm_noise$fix_theta_sigma <- fix_theta_sigma
  f_fm_noise$fix_theta_nu <- fix_theta_nu
  f_fm_noise$theta_mu <- rep(0, ncol(f_fm_noise$B_mu))
  f_fm_noise$theta_sigma <- rep(0, ncol(f_fm_noise$B_sigma))
  f_fm_noise$theta_nu <- rep(0, ncol(f_fm_noise$B_nu))

  if (family == "nig") {
    mn_noise <- rnig(n_obs_per_rep, delta = 2, mu = -2, nu = 1, sigma = 0.5, seed = seed)
    real_mn_noise <- noise_nig(mu = -2, sigma = 0.5, nu = 1)
    fm_mn_noise <- noise_nig(
      # nu=2, fix_theta_nu=TRUE
      # V = attr(mn_noise, "V"), fix_V = TRUE
    )
  } else if (family == "normal") {
    mn_noise <- rnorm(n_obs_per_rep, sd = 0.1)
    real_mn_noise <- noise_normal(sigma = 0.1)
    fm_mn_noise <- noise_normal()
  } else if (family == "cor_normal") {
    stopifnot(n_obs_per_rep %% 2 == 0)
    rho <- -0.5
    Cov_kron <- matrix(c(.5, rho * .5, rho * .5, .5), nrow = 2) %x% diag(n_obs_per_rep / 2)
    L <- t(chol(Cov_kron))
    mn_noise <- as.numeric(L %*% rnorm(n_obs_per_rep))
    real_mn_noise <- noise_normal(
      corr_measurement = TRUE,
      index_corr = rep(1:(n_obs_per_rep / 2), 2),
      rho = -0.5
    )
    fm_mn_noise <- noise_normal(
      corr_measurement = TRUE,
      index_corr = rep(1:(n_obs_per_rep / 2), 2)
    )
  } else if (family == "cor_nig") {
    # simulation of cor_nig?
    stopifnot(n_obs_per_rep %% 2 == 0)
    rho <- -0.5
    Cov_kron <- matrix(c(.5, rho * .5, rho * .5, .5), nrow = 2) %x% diag(n_obs_per_rep / 2)
    L <- t(chol(Cov_kron))
    mn_noise <- as.numeric(L %*% rnorm(n_obs_per_rep))
    real_mn_noise <- noise_nig(
      corr_measurement = TRUE,
      index_corr = rep(1:(n_obs_per_rep / 2), 2),
      rho = -0.5
    )
    fm_mn_noise <- noise_nig(
      corr_measurement = TRUE,
      index_corr = rep(1:(n_obs_per_rep / 2), 2)
    )
  }

  # ------- Simulate data for each model --------
  sim_data <- switch(model,
    "none" = {
      list(Y = mn_noise, group = rep(1, n_obs_per_rep))
    },
    "iid" = {
      idx <- 1:n_obs_per_rep
      iid_model <- f(idx,
        model = iid(),
        noise = f_noise
      )
      W <- simulate(iid_model, seed = seed)[[1]]
      Y <- W + mn_noise
      list(Y = Y, idx = idx, group = rep(1, n_obs_per_rep))
    },
    "ar1" = {
      idx <- 1:n_obs_per_rep
      ar1_model <- f(idx,
        model = ar1(rho = 0.5),
        noise = f_noise
      )
      W <- simulate(ar1_model, seed = seed)[[1]]
      Y <- W + mn_noise
      list(Y = Y, idx = idx, group = rep(1, n_obs_per_rep))
    },
    "matern" = {
      loc <- cbind(stats::runif(n_obs_per_rep, 0, 10), stats::runif(n_obs_per_rep, 0, 5))
      matern_model <- f(
        map = loc,
        model = matern(mesh = mesh, theta_kappa = log(4)),
        noise = f_noise
      )
      W <- simulate(matern_model, seed = seed)[[1]]
      Y <- W + mn_noise
      list(Y = Y, idx = loc, group = rep(1, n_obs_per_rep))
    },
    "ar1+ar1" = {
      idx <- 1:n_obs_per_rep
      ar1_model_1 <- f(idx, model = ar1(rho = 0.5), noise = f_noise)
      ar1_model_2 <- f(idx, model = ar1(rho = -0.6), noise = f_noise)
      W1 <- simulate(ar1_model_1, seed = seed)[[1]]
      W2 <- simulate(ar1_model_2, seed = seed)[[1]]
      Y <- W1 + W2 + mn_noise
      list(Y = Y, idx = idx, group = rep(1, n_obs_per_rep))
    },
    "ar1+matern" = {
      idx <- 1:n_obs_per_rep
      ar1_model <- f(idx, model = ar1(rho = 0.5), noise = f_noise)
      loc <- cbind(stats::runif(n_obs_per_rep, 0, 10), stats::runif(n_obs_per_rep, 0, 5))
      matern_model <- f(
        map = loc,
        model = matern(mesh = mesh, theta_kappa = log(4)),
        noise = f_noise
      )
      W1 <- simulate(ar1_model, seed = seed)[[1]]
      W2 <- simulate(matern_model, seed = seed)[[1]]
      Y <- W1 + W2 + mn_noise
      list(Y = Y, idx = list(idx, loc), group = rep(1, n_obs_per_rep))
    },
    "bvar1" = {
      group_per_rep <- c(rep("first", n_obs_per_rep / 2), rep("second", n_obs_per_rep / 2))
      idx_per_rep <- c(1:(n_obs_per_rep / 2), 1:(n_obs_per_rep / 2))
      true_model <- f(
        idx_per_rep,
        model = bv(
          theta = 0.5, rho = 0.8,
          sub_models = list(
            first = ar1(rho = 0.5),
            second = ar1(rho = -0.5)
          )
        ),
        group = group_per_rep,
        noise = list(
          first = f_noise,
          second = f_noise
        )
      )
      W <- simulate(true_model, seed = seed)[[1]]
      Y <- W + mn_noise
      list(Y = Y, idx = idx_per_rep, group = group_per_rep)
    },
    "bvmatern" = {
      group_per_rep <- c(rep("first", n_obs_per_rep / 2), rep("second", n_obs_per_rep / 2))
      idx_per_rep <- c(1:(n_obs_per_rep / 2), 1:(n_obs_per_rep / 2))
      loc <- cbind(stats::runif(n_obs_per_rep / 2, 0, 10), stats::runif(n_obs_per_rep / 2, 0, 5))
      loc <- rbind(loc, loc)
      true_model <- f(
        loc,
        model = bv(
          mesh = mesh,
          theta = pi / 4, rho = 2,
          sub_models = list(
            first = matern(kappa = 2),
            second = matern(kappa = 5)
          )
        ),
        group = group_per_rep,
        noise = list(
          first = noise_nig(mu = -2, sigma = 0.5, nu = 1),
          second = noise_nig(mu = 2, sigma = 0.3, nu = 0.5)
        )
      )
      W <- simulate(true_model, seed = seed)[[1]]
      Y <- W + mn_noise
      list(Y = Y, idx = loc, group = group_per_rep)
    },
    "graph" = {
      # library(MetricGraph)
      edge1 <- rbind(c(0, 0), c(1, 0))
      edge2 <- rbind(c(0, 0), c(0, 1))
      edge3 <- rbind(c(0, 1), c(-1, 1))
      theta <- seq(from = pi, to = 3 * pi / 2, length.out = 20)
      edge4 <- cbind(sin(theta), 1 + cos(theta))
      graph <- MetricGraph::metric_graph$new(edges = list(edge1, edge2, edge3, edge4))
      # graph <- MetricGraph::metric_graph$new(edges = list(edge1))
      graph$build_mesh(h = 0.02)

      matern_graph <- f(
        model = matern(mesh = graph, theta_kappa = log(8)),
        noise = f_noise
      )

      W <- simulate(matern_graph, seed = seed)[[1]]

      # build observation and A matrices
      obs.per.edge <- n_obs_per_rep / graph$nE
      obs.loc <- NULL
      for (i in 1:graph$nE) {
        obs.loc <- rbind(
          obs.loc,
          cbind(rep(i, obs.per.edge), stats::runif(obs.per.edge))
        )
      }
      A <- graph$fem_basis(obs.loc)
      Y <- as.numeric(A %*% W) + mn_noise

      df_data <- data.frame(
        Y = Y, edge_number = obs.loc[, 1],
        distance_on_edge = obs.loc[, 2]
      )

      graph$clear_observations()
      graph$add_observations(data = df_data, normalized = TRUE)
      # browser()

      list(Y = graph$get_data()$Y, idx = NULL, group = rep(1, n_obs_per_rep))
    },
    stop("Unknown test model")
  )

  # ------- Specify formula for each model -------
  formula <- switch(model,
    "none" = Y ~ 0,
    "iid" = Y ~ 0 + f(idx,
      fix_theta_K = fix_theta_K,
      model = iid(),
      noise = f_fm_noise,
      debug = debug_f,
      control = control_f(numer_grad = numer_grad)
    ),
    "ar1" = Y ~ 0 + f(idx,
      fix_theta_K = fix_theta_K,
      model = ar1(),
      noise = f_fm_noise,
      debug = debug_f,
      control = control_f(numer_grad = numer_grad)
    ),
    "matern" = Y ~ 0 + f(idx,
      fix_theta_K = fix_theta_K,
      model = matern(mesh = mesh),
      noise = f_fm_noise,
      debug = debug_f,
      control = control_f(numer_grad = numer_grad)
    ),
    "ar1+ar1" = Y ~ 0 + f(idx,
      fix_theta_K = fix_theta_K,
      model = ar1(),
      noise = f_fm_noise,
      debug = debug_f,
      name = "field1",
      control = control_f(numer_grad = numer_grad)
    ) + f(idx,
      fix_theta_K = fix_theta_K,
      model = ar1(),
      noise = f_fm_noise,
      debug = debug_f,
      name = "field2",
      control = control_f(numer_grad = numer_grad)
    ),
    "ar1+matern" = Y ~ 0 + f(idx[[1]],
      fix_theta_K = fix_theta_K,
      model = ar1(),
      noise = f_fm_noise,
      debug = debug_f,
      control = control_f(numer_grad = numer_grad)
    ) + f(idx[[2]],
      fix_theta_K = fix_theta_K,
      model = matern(mesh = mesh),
      noise = f_fm_noise,
      debug = debug_f,
      control = control_f(numer_grad = numer_grad)
    ),
    "bvar1" = Y ~ f(idx,
      model = bv(
        sub_models = list(first = ar1(), second = ar1())
      ),
      fix_theta_K = fix_theta_K,
      control = control_f(numer_grad = numer_grad),
      debug = debug_f,
      noise = list(first = f_fm_noise, second = f_fm_noise)
    ),
    "bvmatern" = Y ~ f(idx,
      model = bv(
        mesh = mesh,
        sub_models = list(first = matern(), second = matern())
      ),
      fix_theta_K = fix_theta_K,
      debug = debug_f,
      control = control_f(numer_grad = numer_grad),
      noise = list(first = f_fm_noise, second = f_fm_noise)
    ),
    "graph" = Y ~ 0 + f(idx,
      fix_theta_K = fix_theta_K,
      model = matern(mesh = graph),
      noise = f_fm_noise,
      debug = debug_f,
      control = control_f(numer_grad = numer_grad)
    )
  )
  # make replicate
  idx <- rep_map(sim_data$idx, n_replicate)
  group <- rep(sim_data$group, n_replicate)
  Y <- rep(sim_data$Y, n_replicate)
  repl <- rep(1:n_replicate, each = n_obs_per_rep)

  # fit
  start_time <- proc.time()
  out <- ngme(
    formula,
    replicate = repl,
    group = group,
    data = data.frame(Y = Y),
    control_ngme = control_ngme(
      n_gibbs_samples = n_gibbs_samples
    ),
    control_opt = control_opt,
    start = start,
    family = fm_mn_noise,
    debug = debug
  )

  if (isTRUE(debug)) {
    message("elapsed time: ", paste(proc.time() - start_time, collapse = " "))
  }
  list(
    out = out,
    time = proc.time() - start_time,
    f_noise = f_noise,
    m_noise = real_mn_noise
  )
}

# ngme2 demo

# doing replicate using INLA
library(INLA)
n = 500
z1 = arima.sim(n, model = list(ar = 0.5), sd = 0.5) # independent replication
z2 = arima.sim(n, model = list(ar = 0.5), sd = 0.5) # from AR(1) process
y = c(z1, z2)

idx <- c(1:n, 1:n)
repl <- rep(1:2, each=n)

# use INLA
formula <- y ~ -1 + f(idx, model="ar1", replicate = repl)
res_inla <- inla(
  formula,
  family="gaussian",
  data = list(y=c(z1, z2), idx=idx, repl=repl)
)
summary(res_inla)

# use ngme2
formula <- y ~ -1 + f(idx, model="ar1", replicate = repl)
res_ngme <- ngme(
  formula,
  family="gaussian",
  data = list(y=c(z1, z2), idx=idx, repl=repl)
)
res_ngme

# Doing simulation using ngme2
myar <- f(1:n, model="ar1", alpha = 0.75,
  noise=noise_nig(
  mu = -3,
  sigma = 2,
  nu = 2
))

W1 <- simulate(myar)[[1]]
W2 <- simulate(myar)[[1]]
Y = c(W1, W2) + rnorm(2*n, sd=0.9)

res_ngme2 <- ngme(
  y ~ -1 + f(idx, model="ar1", replicate = repl, noise=noise_nig()),
  family="gaussian",
  data = list(y=Y, idx=idx, repl=repl),
  control = control_opt(
    iterations = 1000
  )
)
res_ngme2

plot(myar$noise,
  res_ngme2$models[["field1"]]$noise)

# This file contains ngme noise specifications

#' @title ngme noise specification
#' @aliases noise_nig noise_normal noise_t noise_skew_t
#' @description Function for specifying ngme noise.
#' Please use \code{noise_nig} and \code{noise_normal} for simpler usage.
#' Use \code{ngme_noise_types()} to check all the available types.
#'
#' @details The parameterization is given in \code{?nig} and \code{?gal}. Moreover,
#' for specifying non-stationary mu and sigma, nu
#' \deqn{\mu = B_{\mu} \theta_{\mu},} and
#' \deqn{\sigma = \exp (B_{\sigma} \theta_{\sigma}),}
#' \deqn{\nu = \nu_{\mathrm{lower}} + \exp (B_{\nu} \theta_{\nu}).}
#'
#' @param noise_type    type of noise, "normal", "nig", "gal", "t", "skew_t"
#' @param mu          specify the NIG noise parameter mu, see \code{?nig}
#' @param sigma       specify the noise parameter sigma, see \code{?nig}
#' @param nu          specify the noise parameter nu, see \code{?nig} and \code{?gal}
#' @param V             start value for V
#' @param theta_mu      specify a non-stationary noise using theta_mu
#' @param B_mu          Basis matrix for mu (if non-stationary)
#' @param theta_sigma   specify a non-stationary noise using theta_sigma
#' @param B_sigma       Basis matrix for sigma (if non-stationary)
#' @param theta_nu      specify a non-stationary noise using theta_nu
#' @param nu_lower_bound specify the lower bound of parameter nu; effective
#'   parametrization is \eqn{nu = nu_lower_bound + \exp(B_\nu \theta_\nu)} so
#'   \code{theta_nu} remains unconstrained. Default 0.
#' @param B_nu          Basis matrix for nu (if non-stationary)
#' @param fix_theta_mu     fix the parameter of theta_mu
#' @param fix_theta_sigma  fix the parameter of theta_sigma, can be a single
#'   logical value or a vector of logical values with length equal to
#'   length(theta_sigma)
#' @param fix_theta_nu     fix the parameter of nu
#' @param fix_theta_sigma_normal  fix the parameter of sigma_normal, used in noise_normal_nig()
#' @param fix_rho    fix the parameter of rho
#' @param fix_V         fix the sampling of V
#'  gives Y|W ~ N(mean * sigma, sigma^2)
#' @param theta_sigma_normal for normal nosie with nig noise sharing same parameter
#' @param B_sigma_normal    for normal nosie with nig noise sharing same parameter
#' @param sigma_normal  for normal nosie with nig noise sharing same parameter
#' @param theta_sigma_nig similar to theta_sigma_normal
#' @param B_sigma_nig     similar to B_sigma_nig
#' @param sigma_nig     similar to sigma_normal
#' @param single_V  TRUE if V is a single number
#' @param share_V  used only for bivariate model
#' @param corr_measurement TRUE if we use correlated measurement noise
#' @param index_corr used when corr_measurement=TRUE, indicate which observation has correlation
#' @param map_corr 1d, 2d, or formula, used when corr_measurement=TRUE, specify use which covariate to infer the index_corr.
#' @param rho used when corr_measurement=TRUE, starting point for correlation
#' @param prior prior specification created by \code{priors(...)}. Supported
#'   keys are \code{mu}, \code{sigma}, and \code{nu}.
#' @param ...       additional arguments
#'
#' @return a list of specification of noise
ngme_noise <- function(
    noise_type,
    mu = 0,
    sigma = 1,
    nu = 1,
    B_mu = NULL,
    theta_mu = NULL,
    B_sigma = NULL,
    theta_sigma = NULL,
    B_nu = NULL,
    theta_nu = NULL,
    theta_sigma_normal = NULL,
    B_sigma_normal = NULL,
    fix_theta_mu = FALSE,
    fix_theta_sigma = FALSE,
    fix_rho = FALSE,
    fix_theta_sigma_normal = FALSE,
    fix_theta_nu = FALSE,
    V = NULL,
    fix_V = FALSE,
    single_V = FALSE,
    share_V = FALSE,
    corr_measurement = FALSE,
    index_corr = NULL,
    map_corr = NULL,
    nu_lower_bound = 0,
    rho = double(0),
    prior = NULL,
    ...) {
  theta_mu_supplied <- !is.null(theta_mu)
  theta_sigma_supplied <- !is.null(theta_sigma)
  theta_nu_supplied <- !is.null(theta_nu)

  if (is.null(theta_mu)) theta_mu <- mu
  if (is.null(theta_sigma)) {
    theta_sigma <- if (sigma > 0) log(sigma) else stop("ngme_noise: sigma should be positive.")
  }
  if (is.null(theta_nu)) {
    theta_nu <- if (nu > nu_lower_bound) log(nu - nu_lower_bound) else stop("ngme_noise: nu must exceed nu_lower_bound.")
  }

  stopifnot(
    "Unkown noise type. Please check ngme_noise_types()" =
      noise_type %in% ngme_noise_types()
  )

  if (is.null(B_mu)) B_mu <- as.matrix(1)
  if (is.null(B_sigma)) B_sigma <- as.matrix(1)
  if (is.null(B_nu)) B_nu <- as.matrix(1)

  # If user supplies basis but omits coefficients, create zero starts that match
  if (!theta_mu_supplied) theta_mu <- rep(0, ncol(B_mu))
  if (!theta_sigma_supplied) theta_sigma <- rep(0, ncol(B_sigma))
  if (!theta_nu_supplied) theta_nu <- rep(0, ncol(B_nu))

  # Validate fix_theta_sigma parameter
  if (length(fix_theta_sigma) == 1) {
    # If single logical value, replicate for all theta_sigma parameters
    fix_theta_sigma <- rep(fix_theta_sigma, length(theta_sigma))
  } else if (length(fix_theta_sigma) != length(theta_sigma)) {
    stop("fix_theta_sigma must be either a single logical value or a vector of length equal to length(theta_sigma)")
  }

  stopifnot(
    "nu_lower_bound must be non-negative" = nu_lower_bound >= 0,
    "Please input B_mu as a matrix." = is.matrix(B_mu),
    "Please input B_sigma as a matrix." = is.matrix(B_sigma),
    "Please make sure ncol(B_mu) == length(theta_mu)." = ncol(B_mu) == length(theta_mu),
    "Please make sure ncol(B_sigma) == length(theta_sigma)." = ncol(B_sigma) == length(theta_sigma),
    "Please make sure ncol(B_nu) == length(theta_nu)." = ncol(B_nu) == length(theta_nu),
    "fix_theta_sigma must be logical" = is.logical(fix_theta_sigma)
  )

  prior_nu_user <- FALSE
  if (!is.null(prior)) {
    if (is_prior_spec(prior)) {
      prior_nu_user <- TRUE
    } else if (is_prior_collection(prior)) {
      prior_nu_user <- "nu" %in% names(prior)
    }
  }

  compiled_prior <- compile_noise_priors(prior)

  if (all(noise_type == "normal")) {
    theta_mu <- double(0)
    B_mu <- matrix(ncol = 0, nrow = nrow(B_mu))
    theta_nu <- double(0)
    B_nu <- matrix(ncol = 0, nrow = nrow(B_nu))
  }

  if (all(noise_type == "t")) {
    theta_mu <- double(0)
    B_mu <- matrix(ncol = 0, nrow = nrow(B_mu))
  }

  if (all(noise_type != "normal_nig")) {
    theta_sigma_normal <- double(0)
    B_sigma_normal <- matrix(ncol = 0, nrow = nrow(B_mu))
  }

  # init rho
  if (corr_measurement && length(rho) == 0) {
    rho <- 0
  }

  n_theta_mu <- if (fix_theta_mu) 0 else length(theta_mu)
  n_theta_sigma <- sum(!fix_theta_sigma) # Count parameters that are NOT fixed
  n_theta_nu <- if (fix_theta_nu) 0 else length(theta_nu)
  n_rho <- if (fix_rho) 0 else length(rho)
  n_theta_sigma_normal <- if (fix_theta_sigma_normal) 0 else length(theta_sigma_normal)

  structure(
    list(
      noise_type = noise_type,
      V = V,
      theta_mu = theta_mu,
      theta_sigma = theta_sigma,
      theta_sigma_normal = theta_sigma_normal,
      theta_nu = theta_nu,
      B_mu = B_mu,
      B_sigma = B_sigma,
      B_sigma_normal = B_sigma_normal,
      B_nu = B_nu,
      n_theta_mu = n_theta_mu,
      n_theta_sigma = n_theta_sigma,
      n_theta_nu = n_theta_nu,
      n_rho = n_rho,
      n_theta_sigma_normal = n_theta_sigma_normal,
      fix_theta_mu = fix_theta_mu,
      fix_theta_sigma = fix_theta_sigma,
      fix_theta_nu = fix_theta_nu,
      nu_lower_bound = nu_lower_bound,
      fix_V = fix_V,
      fix_rho = fix_rho,
      fix_theta_sigma_normal = fix_theta_sigma_normal,
      n_params = n_theta_mu + n_theta_sigma + n_theta_nu + n_rho,
      single_V = single_V,
      share_V = share_V,
      corr_measurement = corr_measurement,
      index_corr = index_corr,
      map_corr = map_corr,
      rho = rho,
      prior_nu_user = prior_nu_user,
      prior_mu = compiled_prior$mu,
      prior_sigma = compiled_prior$sigma,
      prior_nu = compiled_prior$nu,
      ...
    ),
    class = "ngme_noise"
  )
}

#' @rdname ngme_noise
#' @export
#' @examples
#' noise_normal(sigma = 2)
noise_normal <- normal <- function(
    sigma = NULL,
    theta_sigma = NULL,
    B_sigma = matrix(1),
    corr_measurement = FALSE,
    index_corr = NULL,
    ...) {
  sd <- sigma

  if (!is.null(sd) && !is.null(theta_sigma)) {
    stop("Please only use sigma or theta_sigma as input")
  }

  # both are null, use default value
  if (is.null(sd) && is.null(theta_sigma)) {
    theta_sigma <- rep(0, ncol(B_sigma))
  }

  if (!is.null(sd)) {
    stopifnot(
      "sd is a double" = is.double(sd),
      "sd should be positive" = sd > 0
    )

    theta_sigma <- log(sd)
  }

  if (!is.null(theta_sigma)) {
    stopifnot(
      "Make sure ncol of B_sigma = length of theta_sigma" = ncol(B_sigma) == length(theta_sigma)
    )
  }

  ngme_noise(
    noise_type = "normal",
    theta_sigma = theta_sigma,
    B_sigma = B_sigma,
    corr_measurement = corr_measurement,
    index_corr = index_corr,
    ...
  )
}

#' @rdname ngme_noise
#' @export
#' @examples
#' noise_nig(mu = 1, sigma = 2, nu = 1)
noise_nig <- nig <- function(
    mu = NULL,
    sigma = NULL,
    nu = NULL,
    V = NULL,
    theta_mu = NULL,
    theta_sigma = NULL,
    theta_nu = NULL,
    nu_lower_bound = 0,
    B_mu = matrix(1),
    B_sigma = matrix(1),
    B_nu = matrix(1),
    corr_measurement = FALSE,
    index_corr = NULL,
    ...) {
  # if nothing, then fill with default
  stopifnot("Please use theta_mu for non-stationary mu." = length(mu) < 2)
  if (is.null(mu) && is.null(theta_mu)) theta_mu <- rep(0, ncol(B_mu))
  if (is.null(sigma) && is.null(theta_sigma)) theta_sigma <- rep(0, ncol(B_sigma))
  if (is.null(nu) && is.null(theta_nu)) theta_nu <- rep(0, ncol(B_nu))

  if (!is.null(nu) && nu <= nu_lower_bound) {
    stop("ngme_noise: nu must exceed nu_lower_bound.")
  }
  if (!is.null(sigma) && sigma <= 0) stop("ngme_nosie: sigma should be positive.")

  if (!is.null(mu)) theta_mu <- mu
  if (!is.null(sigma)) theta_sigma <- log(sigma)
  if (!is.null(nu)) theta_nu <- log(nu - nu_lower_bound)

  ngme_noise(
    noise_type = "nig",
    theta_mu = theta_mu,
    theta_sigma = theta_sigma,
    theta_nu = theta_nu,
    nu_lower_bound = nu_lower_bound,
    V = V,
    B_mu = B_mu,
    B_sigma = B_sigma,
    B_nu = B_nu,
    corr_measurement = corr_measurement,
    index_corr = index_corr,
    ...
  )
}

#' @rdname ngme_noise
#' @export
#' @examples
#' noise_gal(mu = 1, sigma = 2, nu = 1)
noise_gal <- gal <- function(
    mu = NULL,
    sigma = NULL,
    nu = NULL,
    V = NULL,
    theta_mu = NULL,
    theta_sigma = NULL,
    theta_nu = NULL,
    nu_lower_bound = 0.01,
    B_mu = matrix(1),
    B_sigma = matrix(1),
    B_nu = matrix(1),
    corr_measurement = FALSE,
    index_corr = NULL,
    ...) {
  # if nothing, then fill with default
  stopifnot("Please use theta_mu for non-stationary mu." = length(mu) < 2)
  if (is.null(mu) && is.null(theta_mu)) theta_mu <- rep(0, ncol(B_mu))
  if (is.null(sigma) && is.null(theta_sigma)) theta_sigma <- rep(0, ncol(B_sigma))
  if (is.null(nu) && is.null(theta_nu)) theta_nu <- rep(0, ncol(B_nu))

  if (!is.null(nu) && nu <= nu_lower_bound) {
    stop("ngme_noise: nu must exceed nu_lower_bound.")
  }
  if (!is.null(sigma) && sigma <= 0) stop("ngme_nosie: sigma should be positive.")

  if (!is.null(mu)) theta_mu <- mu
  if (!is.null(sigma)) theta_sigma <- log(sigma)
  if (!is.null(nu)) theta_nu <- log(nu - nu_lower_bound)

  ngme_noise(
    noise_type = "gal",
    theta_mu = theta_mu,
    theta_sigma = theta_sigma,
    theta_nu = theta_nu,
    nu_lower_bound = nu_lower_bound,
    V = V,
    B_mu = B_mu,
    B_sigma = B_sigma,
    B_nu = B_nu,
    corr_measurement = corr_measurement,
    index_corr = index_corr,
    ...
  )
}

#' Moments of stationary NIG and GAL noise
#'
#' @description
#' Compute the mean, variance, skewness, kurtosis, and excess kurtosis for
#' stationary NIG and GAL noise. The distribution is evaluated through the
#' normal mean-variance mixture
#' \deqn{X = \delta + \mu V + \sigma \sqrt{V} Z,}
#' where \eqn{Z \sim N(0, 1)}. For the package's stationary noise
#' parameterization, \code{delta = -mu}, so the mean is zero.
#'
#' @param noise An \code{ngme_noise} object created by \code{noise_nig()} or
#'   \code{noise_gal()}.
#' @param delta Location parameter. Defaults to \code{-mu}, matching the
#'   stationary ngme noise convention.
#' @param mu Shift/skewness parameter in the mixture mean.
#' @param sigma Positive Gaussian scale parameter.
#' @param nu Positive mixing-distribution shape parameter.
#' @param ... Currently unused.
#'
#' @return A named numeric vector with entries \code{skewness},
#'   \code{kurtosis}, \code{excess_kurtosis}, \code{variance}, \code{sd},
#'   \code{central_moment3}, and \code{central_moment4}.
#'
#' @examples
#' noise_nig_moments(mu = 1, sigma = 2, nu = 3)
#' noise_gal_moments(mu = 1, sigma = 2, nu = 3)
#' noise_moments(noise_nig(mu = 1, sigma = 2, nu = 3))
#'
#' @export
noise_moments <- function(noise, ...) {
  UseMethod("noise_moments")
}

#' @rdname noise_moments
#' @export
noise_moments.ngme_noise <- function(noise, ...) {
  if (length(noise$noise_type) != 1) {
    stop("noise_moments() only supports univariate stationary noise.")
  }

  mu <- .stationary_noise_parameter(noise, "mu")
  sigma <- .stationary_noise_parameter(noise, "sigma")
  nu <- .stationary_noise_parameter(noise, "nu")

  switch(noise$noise_type,
    "nig" = noise_nig_moments(mu = mu, sigma = sigma, nu = nu),
    "gal" = noise_gal_moments(mu = mu, sigma = sigma, nu = nu),
    stop("noise_moments() currently supports only NIG and GAL noise.")
  )
}

#' @rdname noise_moments
#' @export
noise_nig_moments <- function(delta = -mu, mu = 0, sigma = 1, nu = 1) {
  .check_stationary_moment_params(delta, mu, sigma, nu)
  mixing_moments <- .gig_moments(p = -0.5, a = nu, b = nu, k = 1:4)
  .normal_mean_variance_mixture_moments(delta, mu, sigma, mixing_moments)
}

#' @rdname noise_moments
#' @export
noise_gal_moments <- function(delta = -mu, mu = 0, sigma = 1, nu = 1) {
  .check_stationary_moment_params(delta, mu, sigma, nu)
  mixing_moments <- vapply(1:4, function(k) {
    prod(nu + seq_len(k) - 1) / nu^k
  }, numeric(1))
  .normal_mean_variance_mixture_moments(delta, mu, sigma, mixing_moments)
}

.stationary_noise_parameter <- function(noise, parameter) {
  if (parameter == "mu") {
    basis <- noise$B_mu
    theta <- noise$theta_mu
    transform <- identity
  } else if (parameter == "sigma") {
    basis <- noise$B_sigma
    theta <- noise$theta_sigma
    transform <- exp
  } else if (parameter == "nu") {
    basis <- noise$B_nu
    theta <- noise$theta_nu
    transform <- function(x) noise$nu_lower_bound + exp(x)
  } else {
    stop("Unknown noise parameter.")
  }

  value <- as.numeric(transform(basis %*% theta))
  if (length(value) < 1) {
    stop("noise_moments() requires a non-empty stationary noise parameter.")
  }
  if (any(!is.finite(value)) || any(abs(value - value[[1]]) > sqrt(.Machine$double.eps))) {
    stop("noise_moments() only supports stationary noise parameters.")
  }
  value[[1]]
}

.check_stationary_moment_params <- function(delta, mu, sigma, nu) {
  stopifnot(
    "delta must be a finite scalar." = length(delta) == 1 && is.finite(delta),
    "mu must be a finite scalar." = length(mu) == 1 && is.finite(mu),
    "sigma must be a positive finite scalar." = length(sigma) == 1 && is.finite(sigma) && sigma > 0,
    "nu must be a positive finite scalar." = length(nu) == 1 && is.finite(nu) && nu > 0
  )
}

.gig_moments <- function(p, a, b, k) {
  stopifnot(
    "GIG parameter a must be positive." = length(a) == 1 && is.finite(a) && a > 0,
    "GIG parameter b must be positive." = length(b) == 1 && is.finite(b) && b > 0
  )

  z <- sqrt(a * b)
  log_kp <- log(besselK(z, nu = p, expon.scaled = TRUE)) - z

  vapply(k, function(ki) {
    log_kp_ki <- log(besselK(z, nu = p + ki, expon.scaled = TRUE)) - z
    (b / a)^(ki / 2) * exp(log_kp_ki - log_kp)
  }, numeric(1))
}

.normal_mean_variance_mixture_moments <- function(delta, beta, sigma, mixing_moments) {
  m1 <- mixing_moments[[1]]
  m2 <- mixing_moments[[2]]
  m3 <- mixing_moments[[3]]
  m4 <- mixing_moments[[4]]

  variance_v <- m2 - m1^2
  central_v3 <- m3 - 3 * m1 * m2 + 2 * m1^3
  central_v4 <- m4 - 4 * m1 * m3 + 6 * m1^2 * m2 - 3 * m1^4
  v_weighted_central_v2 <- m3 - 2 * m1 * m2 + m1^3

  variance <- sigma^2 * m1 + beta^2 * variance_v
  central_moment3 <- beta^3 * central_v3 + 3 * beta * sigma^2 * variance_v
  central_moment4 <- beta^4 * central_v4 +
    6 * beta^2 * sigma^2 * v_weighted_central_v2 +
    3 * sigma^4 * m2

  if (!is.finite(variance) || variance <= 0) {
    stop("Computed variance is not positive and finite.")
  }

  c(
    skewness = central_moment3 / variance^(3 / 2),
    kurtosis = central_moment4 / variance^2,
    excess_kurtosis = central_moment4 / variance^2 - 3,
    variance = variance,
    sd = sqrt(variance),
    central_moment3 = central_moment3,
    central_moment4 = central_moment4
  )
}

#' @rdname ngme_noise
#' @export
#' @examples
#' noise_skew_t(mu = 0, sigma = 1, nu = 5)
noise_skew_t <- skew_t_noise <- function(
    mu = NULL,
    sigma = NULL,
    nu = NULL,
    theta_mu = NULL,
    theta_sigma = NULL,
    theta_nu = NULL,
    nu_lower_bound = 0.01,
    B_mu = matrix(1),
    B_sigma = matrix(1),
    B_nu = matrix(1),
    corr_measurement = FALSE,
    index_corr = NULL,
    ...) {
  # if nothing, then fill with default
  stopifnot("Please use theta_mu for non-stationary mu." = length(mu) < 2)
  if (is.null(mu) && is.null(theta_mu)) theta_mu <- rep(0, ncol(B_mu))
  if (is.null(sigma) && is.null(theta_sigma)) theta_sigma <- rep(0, ncol(B_sigma))
  if (is.null(nu) && is.null(theta_nu)) {
    stopifnot("Default nu=5 must exceed nu_lower_bound." = 5 > nu_lower_bound)
    theta_nu <- rep(log(5 - nu_lower_bound), ncol(B_nu)) # default to 5 degrees of freedom
  }

  if (!is.null(nu) && nu <= nu_lower_bound) {
    stop("ngme_noise: nu (degrees of freedom) must exceed nu_lower_bound.")
  }
  if (!is.null(sigma) && sigma <= 0) stop("ngme_noise: sigma should be positive.")

  if (!is.null(mu)) theta_mu <- mu
  if (!is.null(sigma)) theta_sigma <- log(sigma)
  if (!is.null(nu)) theta_nu <- log(nu - nu_lower_bound)

  ngme_noise(
    noise_type = "skew_t",
    theta_mu = theta_mu,
    theta_sigma = theta_sigma,
    theta_nu = theta_nu,
    nu_lower_bound = nu_lower_bound,
    B_mu = B_mu,
    B_sigma = B_sigma,
    B_nu = B_nu,
    corr_measurement = corr_measurement,
    index_corr = index_corr,
    ...
  )
}


#' @rdname ngme_noise
#' @export
#' @examples
#' noise_t(nu = 5)
noise_t <- t_noise <- function(
    nu = NULL,
    theta_nu = NULL,
    nu_lower_bound = 0,
    B_nu = matrix(1),
    corr_measurement = FALSE,
    index_corr = NULL,
    ...) {
  # if nothing, then fill with default
  if (is.null(nu) && is.null(theta_nu)) {
    stopifnot("Default nu=5 must exceed nu_lower_bound." = 5 > nu_lower_bound)
    theta_nu <- rep(log(5 - nu_lower_bound), ncol(B_nu)) # default to 5 degrees of freedom
  }

  if (!is.null(nu) && nu <= nu_lower_bound) {
    stop("ngme_noise: nu (degrees of freedom) must exceed nu_lower_bound.")
  }

  if (!is.null(nu)) theta_nu <- log(nu - nu_lower_bound)

  ngme_noise(
    noise_type = "t",
    theta_sigma = 0,
    theta_nu = theta_nu,
    nu_lower_bound = nu_lower_bound,
    B_sigma = matrix(1),
    B_nu = B_nu,
    fix_theta_sigma = TRUE,
    corr_measurement = corr_measurement,
    index_corr = index_corr,
    ...
  )
}


# update noise
update_noise <- function(noise, n = NULL, new_noise = NULL) {
  # update with length n
  if (!is.null(n)) {
    stopifnot("n should be integer" = is.numeric(n))
    B_mu <- noise$B_mu
    stopifnot("n / nrow(B_mu) not integer" = abs(n / nrow(B_mu) - round(n / nrow(B_mu))) < 1e-4)
    noise$B_mu <- matrix(data = rep(B_mu, n / nrow(B_mu)), nrow = n)

    B_nu <- noise$B_nu
    stopifnot("n / nrow(B_nu) not integer" = abs(n / nrow(B_nu) - round(n / nrow(B_nu))) < 1e-4)
    noise$B_nu <- matrix(data = rep(B_nu, n / nrow(B_nu)), nrow = n)

    # Reshape Basis Matrix for Normal-NIG noise
    if (all(noise$noise_type == "normal_nig")) {
      # Merge B_sigma_normal and B_sigma_nig
      B_sigma_normal <- noise$B_sigma_normal
      noise$B_sigma_normal <- matrix(data = rep(B_sigma_normal, n / nrow(B_sigma_normal)), nrow = n)
      B_sigma_nig <- noise$B_sigma_nig
      noise$B_sigma_nig <- matrix(data = rep(B_sigma_nig, n / nrow(B_sigma_nig)), nrow = n)

      noise$B_sigma <- as.matrix(Matrix::bdiag(
        noise$B_sigma_nig, # place holder
        noise$B_sigma_normal
      ))
      noise$B_mu <- rbind(noise$B_mu, matrix(0, nrow(noise$B_mu), ncol(noise$B_sigma_normal)))
      noise$B_nu <- rbind(noise$B_nu, noise$B_nu)
    } else {
      B_sigma <- noise$B_sigma
      stopifnot("n / nrow(B_sigma) not integer" = abs(n / nrow(B_sigma) - round(n / nrow(B_sigma))) < 1e-4)
      noise$B_sigma <- matrix(data = rep(B_sigma, n / nrow(B_sigma)), nrow = n)
    }
  } else if (!is.null(new_noise)) {
    # update noise after estimation
    if (all(new_noise$noise_type != "normal")) {
      noise$theta_mu <- new_noise$theta_mu
      noise$theta_nu <- new_noise$theta_nu
    }
    noise$theta_sigma <- new_noise$theta_sigma
    noise$rho <- new_noise$rho
    if (!is.null(new_noise$V)) noise$V <- new_noise$V

    # bv noise
    if (length(noise$noise_type) == 2) {
      # pass mu, sigma, nu to sub_models
      n_theta_mu1 <- noise$bv_noises[[1]]$n_theta_mu
      n_theta_mu2 <- noise$bv_noises[[2]]$n_theta_mu
      n_theta_sigma1 <- noise$bv_noises[[1]]$n_theta_sigma
      n_theta_sigma2 <- noise$bv_noises[[2]]$n_theta_sigma
      n_theta_nu1 <- noise$bv_noises[[1]]$n_theta_nu
      n_theta_nu2 <- noise$bv_noises[[2]]$n_theta_nu
      if (!noise$fix_theta_mu) {
        noise$bv_noises[[1]]$theta_mu <- head(noise$theta_mu, n_theta_mu1)
        noise$bv_noises[[2]]$theta_mu <- tail(noise$theta_mu, n_theta_mu2)
      }
      if (!all(noise$fix_theta_sigma)) {
        noise$bv_noises[[1]]$theta_sigma <- head(noise$theta_sigma, n_theta_sigma1)
        noise$bv_noises[[2]]$theta_sigma <- tail(noise$theta_sigma, n_theta_sigma2)
      }
      if (!noise$fix_theta_nu) {
        noise$bv_noises[[1]]$theta_nu <- head(noise$theta_nu, n_theta_nu1)
        noise$bv_noises[[2]]$theta_nu <- tail(noise$theta_nu, n_theta_nu2)
      }
    }

    if (all(noise$noise_type == "normal_nig")) {
      n_theta_sigma_nig <- length(noise$theta_sigma_nig)
      n_theta_sigma_normal <- length(noise$theta_sigma_normal)
      noise$theta_sigma_nig <- noise$theta_sigma[1:n_theta_sigma_nig]
      noise$theta_sigma_normal <- noise$theta_sigma[(n_theta_sigma_nig + 1):(n_theta_sigma_nig + n_theta_sigma_normal)]
    }
  }
  noise
}

#' @rdname ngme_noise
#' @export
noise_normal_nig <- normal_nig <- function(
    sigma_normal = NULL,
    mu = NULL,
    sigma_nig = NULL,
    nu = NULL,
    V = NULL,
    theta_mu = NULL,
    theta_sigma_nig = NULL,
    theta_sigma_normal = NULL,
    theta_nu = NULL,
    B_mu = matrix(1),
    B_sigma_nig = matrix(1),
    B_sigma_normal = matrix(1),
    B_nu = matrix(1),
    corr_measurement = FALSE,
    index_corr = NULL,
    ...) {
  dots <- list(...)
  nu_lower_bound <- if (!is.null(dots$nu_lower_bound)) dots$nu_lower_bound else 0

  # if nothing, then fill with default
  stopifnot("Please use theta_mu for non-stationary mu." = length(mu) < 2)
  if (is.null(mu) && is.null(theta_mu)) theta_mu <- rep(0, ncol(B_mu))
  if (is.null(sigma_nig) && is.null(theta_sigma_nig)) theta_sigma_nig <- rep(0, ncol(B_sigma_nig))
  if (is.null(nu) && is.null(theta_nu)) theta_nu <- rep(0, ncol(B_nu))
  if (is.null(sigma_normal) && is.null(theta_sigma_normal)) theta_sigma_normal <- rep(0, ncol(B_sigma_normal))

  if (!is.null(nu) && nu <= nu_lower_bound) {
    stop("ngme_noise: nu must exceed nu_lower_bound.")
  }
  if (!is.null(sigma_nig) && sigma_nig <= 0) stop("ngme_nosie: sigma_nig should be positive.")
  if (!is.null(sigma_normal) && sigma_normal <= 0) stop("ngme_nosie: sigma_nig should be positive.")

  if (!is.null(mu)) theta_mu <- mu
  if (!is.null(sigma_nig)) theta_sigma_nig <- log(sigma_nig)
  if (!is.null(sigma_normal)) theta_sigma_normal <- log(sigma_normal)
  if (!is.null(nu)) theta_nu <- log(nu - nu_lower_bound)

  ngme_noise(
    noise_type = "normal_nig",
    theta_mu = theta_mu,
    theta_sigma_nig = theta_sigma_nig,
    theta_sigma_normal = theta_sigma_normal,
    theta_sigma = c(theta_sigma_nig, theta_sigma_normal),
    theta_nu = theta_nu,
    V = V,
    B_mu = B_mu,
    B_sigma = cbind(B_sigma_nig, B_sigma_normal),
    B_nu = B_nu,
    B_sigma_nig = B_sigma_nig,
    B_sigma_normal = B_sigma_normal,
    corr_measurement = corr_measurement,
    index_corr = index_corr,
    ...
  )
}

#' Print ngme noise
#'
#' @param x noise object
#' @param padding number of white space padding in front
#' @param prefix prefix
#' @param model_type model type
#' @param ... ...
#'
#' @return a list (noise specifications)
#' @export
print.ngme_noise <- function(
    x,
    padding = 0,
    prefix = "Noise type",
    model_type = NULL,
    ...) {
  noise <- x
  pad_space <- paste(rep(" ", padding), collapse = "")
  pad_add4_space <- paste(rep(" ", padding + 4), collapse = "")

  if (is.null(noise)) {
    cat(pad_space)
    cat(prefix)
    cat(": ")
    cat("NULL")
    cat("\n")
  } else {
    if (length(noise$noise_type) == 2) {
      # bivariate noise
      cat(pad_space)
      cat(" ")
      if (noise$single_V && noise$share_V) {
        cat("Bivariate type-G1 noise (single_V && share_V):")
      } else if (noise$single_V && !noise$share_V) {
        cat("Bivariate type-G2 noise (single_V):")
      } else if (!noise$single_V && noise$share_V) {
        cat("Bivariate type-G3 noise (share_V):")
      } else {
        cat("Bivariate type-G4 noise:")
      }
      cat("\n")
      names <- names(noise$bv_noises)
      print(noise$bv_noises[[1]], padding = padding + 4, prefix = names[[1]])
      print(noise$bv_noises[[2]], padding = padding + 4, prefix = names[[2]])
    } else {
      # single noise
      cat(pad_space)
      cat(prefix)
      cat(": ")
      cat(toupper(noise$noise_type))
      cat("\n")

      known_type <- !is.null(model_type)
      if ((known_type && model_type == "re") && noise$noise_type == "normal") {
        # skip
      } else if (known_type && model_type == "re") {
        # only print mu and nu
        cat(paste0(
          pad_add4_space, ngme_format("mu", noise$theta_mu), "\n",
          pad_add4_space, ngme_format("nu", noise$theta_nu, nu_lower_bound = noise$nu_lower_bound)
        ))
      } else {
        cat(pad_space)
        cat("Noise parameters: \n")
        params <- with(noise, {
          if (known_type && model_type %in% c("rw1", "rw2")) {
            theta_sigma <- theta_sigma[-1] # suppress the first fixed parameter
          }
          switch(noise_type,
            "normal" = paste0(pad_add4_space, ngme_format("sigma", theta_sigma)),
            "nig" = paste0(
              pad_add4_space, ngme_format("mu", theta_mu),
              "\n", pad_add4_space, ngme_format("sigma", theta_sigma),
              "\n", pad_add4_space, ngme_format("nu", theta_nu, nu_lower_bound = nu_lower_bound)
            ),
            "gal" = paste0(
              pad_add4_space, ngme_format("mu", theta_mu),
              "\n", pad_add4_space, ngme_format("sigma", theta_sigma),
              "\n", pad_add4_space, ngme_format("nu", theta_nu, nu_lower_bound = nu_lower_bound)
            ),
            "t" = paste0(pad_add4_space, ngme_format("nu", theta_nu, nu_lower_bound = nu_lower_bound)),
            "skew_t" = paste0(
              pad_add4_space, ngme_format("mu", theta_mu),
              "\n", pad_add4_space, ngme_format("sigma", theta_sigma),
              "\n", pad_add4_space, ngme_format("nu", theta_nu, nu_lower_bound = nu_lower_bound)
            ),
            "normal_nig" = paste0(
              pad_add4_space, ngme_format("mu", theta_mu),
              "\n", pad_add4_space, ngme_format("sigma_nig", theta_sigma_nig),
              "\n", pad_add4_space, ngme_format("nu", theta_nu, nu_lower_bound = nu_lower_bound),
              "\n", pad_add4_space, ngme_format("sigma_normal", theta_sigma_normal)
            ),
            NULL
          )
        })
        if (noise$nu_lower_bound > 0 && length(noise$theta_nu) > 0) {
          lb_txt <- paste0(" (lower bound ", format(noise$nu_lower_bound, digits = 3), ")")
          params <- sub("(nu = [^\\n]+)", paste0("\\1", lb_txt), params, perl = TRUE)
          params <- sub("(theta_nu = [^\\n]+)", paste0("\\1", lb_txt), params, perl = TRUE)
          params <- sub("\\(lower bound ([0-9.]+)\\)\\s*\\(lower bound [0-9.]+\\)", "(lower bound \\1)", params, perl = TRUE)
        }
        cat(params)
      }
    }
  }
  cat("\n")
  if (noise$corr_measurement) {
    cat(pad_add4_space)
    cat("correlation(rho) = ")
    cat(format(noise$rho, digits = 3))
    cat("\n")
  }
  if (all(noise$B_nu %*% noise$theta_nu > 1000)) {
    cat("(Notice: Parameter nu seems too big, consider use Gaussian noise.)\n")
  }

  invisible(noise)
}

subset_noise <- function(noise, sub_idx, compute_corr) {
  noise$B_mu <- noise$B_mu[sub_idx, , drop = FALSE]
  noise$B_sigma <- noise$B_sigma[sub_idx, , drop = FALSE]
  noise$B_nu <- noise$B_nu[sub_idx, , drop = FALSE]
  noise$V <- noise$V[sub_idx]

  if (!is.null(noise$index_corr)) noise$index_corr <- noise$index_corr[sub_idx]

  if (!is.null(noise$corr_measurement) &&
    noise$corr_measurement && compute_corr
  ) {
    p_order <- order(noise$index_corr)
    cov_rc <- compute_corr_index(noise$index_corr[p_order])

    # update noise with extra terms about correlation
    noise$cor_rows <- cov_rc$cor_rows
    noise$cor_cols <- cov_rc$cor_cols
    noise$has_correlation <- cov_rc$has_correlation
    noise$n_corr_pairs <- cov_rc$n_corr_pairs
    noise$index_corr <- noise$index_corr[p_order]
  }
  noise
}


#' Merge 2 noise into 1 noise
#'
#' @param noise1 noise 1
#' @param noise2 noise 2
#'
#' @return merged noise
#' @export
merge_noise <- function(noise1, noise2) {
  noise1$B_mu <- rbind(noise1$B_mu, noise2$B_mu)
  noise1$B_sigma <- rbind(noise1$B_sigma, noise2$B_sigma)
  noise1$B_nu <- rbind(noise1$B_nu, noise2$B_nu)
  noise1$V <- c(noise1$V, noise2$V)
  noise1

  if (!is.null(noise1$index_corr) && !is.null(noise2$index_corr)) {
    max_idx <- max(noise1$index_corr, noise2$index_corr)
    # avoid overlap index
    noise1$index_corr <- c(noise1$index_corr, noise2$index_corr + max_idx)
  }

  # Update correlation
  if (!is.null(noise1$corr_measurement) &&
    noise1$corr_measurement
  ) {
    p_order <- order(noise1$index_corr)
    cov_rc <- compute_corr_index(noise1$index_corr[p_order])

    # update noise with extra terms about correlation
    noise1$cor_rows <- cov_rc$cor_rows
    noise1$cor_cols <- cov_rc$cor_cols
    noise1$has_correlation <- cov_rc$has_correlation
    noise1$n_corr_pairs <- cov_rc$n_corr_pairs
    noise1$index_corr <- noise1$index_corr[p_order]
  }

  noise1
}

test_that("stationary GAL moments match gamma mixing formulas", {
  moments <- noise_gal_moments(mu = 1, sigma = 2, nu = 3)

  expect_equal(names(moments), c(
    "skewness", "kurtosis", "excess_kurtosis",
    "variance", "sd", "central_moment3", "central_moment4"
  ))
  expect_false("mean" %in% names(moments))
  expect_equal(moments[["variance"]], 13 / 3)
  expect_equal(moments[["central_moment3"]], 38 / 9)
  expect_equal(moments[["central_moment4"]], 701 / 9)
  expect_equal(moments[["skewness"]], (38 / 9) / (13 / 3)^(3 / 2))
  expect_equal(moments[["kurtosis"]], (701 / 9) / (13 / 3)^2)
  expect_equal(moments[["excess_kurtosis"]], (701 / 9) / (13 / 3)^2 - 3)
})

test_that("stationary NIG moments match inverse-Gaussian mixing formulas", {
  moments <- noise_nig_moments(mu = 1, sigma = 2, nu = 3)

  expect_equal(names(moments), c(
    "skewness", "kurtosis", "excess_kurtosis",
    "variance", "sd", "central_moment3", "central_moment4"
  ))
  expect_false("mean" %in% names(moments))
  expect_equal(moments[["variance"]], 13 / 3, tolerance = 1e-12)
  expect_equal(moments[["central_moment3"]], 13 / 3, tolerance = 1e-12)
  expect_equal(moments[["central_moment4"]], 728 / 9, tolerance = 1e-12)
  expect_equal(moments[["skewness"]], (13 / 3) / (13 / 3)^(3 / 2), tolerance = 1e-12)
  expect_equal(moments[["kurtosis"]], (728 / 9) / (13 / 3)^2, tolerance = 1e-12)
  expect_equal(moments[["excess_kurtosis"]], (728 / 9) / (13 / 3)^2 - 3, tolerance = 1e-12)
})

test_that("stationary NIG moments match standard NIG closed-form formulas", {
  mu <- 1
  sigma <- 2
  nu <- 3

  moments <- noise_nig_moments(mu = mu, sigma = sigma, nu = nu)

  alpha <- sqrt(mu^2 / sigma^4 + nu / sigma^2)
  beta <- mu / sigma^2
  delta <- sqrt(nu) * sigma
  gamma <- sqrt(alpha^2 - beta^2)

  expect_equal(moments[["variance"]], delta * alpha^2 / gamma^3, tolerance = 1e-12)
  expect_equal(moments[["skewness"]], 3 * beta / sqrt(alpha^2 * delta * gamma), tolerance = 1e-12)
  expect_equal(
    moments[["excess_kurtosis"]],
    3 * (1 + 4 * beta^2 / alpha^2) / (delta * gamma),
    tolerance = 1e-12
  )
  expect_equal(
    moments[["kurtosis"]],
    3 + 3 * (1 + 4 * beta^2 / alpha^2) / (delta * gamma),
    tolerance = 1e-12
  )
})

test_that("stationary NIG and GAL object methods match direct moment functions", {
  nig_noise <- noise_nig(mu = 1, sigma = 2, nu = 3)
  gal_noise <- noise_gal(mu = 1, sigma = 2, nu = 3)

  expect_equal(noise_moments(nig_noise), noise_nig_moments(mu = 1, sigma = 2, nu = 3))
  expect_equal(noise_moments(gal_noise), noise_gal_moments(mu = 1, sigma = 2, nu = 3))
})

test_that("stationary moment object method accepts repeated stationary bases", {
  nig_noise <- update_noise(noise_nig(mu = 1, sigma = 2, nu = 3), n = 4)
  gal_noise <- update_noise(noise_gal(mu = 1, sigma = 2, nu = 3), n = 4)

  expect_equal(noise_moments(nig_noise), noise_nig_moments(mu = 1, sigma = 2, nu = 3))
  expect_equal(noise_moments(gal_noise), noise_gal_moments(mu = 1, sigma = 2, nu = 3))
})

test_that("stationary moment object method rejects non-stationary bases", {
  noise <- noise_nig(
    theta_mu = c(0, 1),
    B_mu = diag(2),
    sigma = 2,
    nu = 3
  )

  expect_error(noise_moments(noise), "stationary")
})

ngme_operator <- function(
    mesh,
    model,
    K,
    h,
    Z = NULL,
    theta_K = NULL,
    zero_trace = FALSE,
    symmetric = FALSE,
    generic_type = "none", # "none", "generic", "generic_ns"
    ...) {
  if (is.null(K)) stop("K is NULL.")
  if (nrow(K) != ncol(K)) stop("K is not a square matrix.")
  if (is.null(Z)) Z <- diag(nrow(K))

  structure(
    list(
      mesh = mesh,
      model = model,
      K = K,
      h = h,
      Z = Z,
      theta_K = theta_K,
      n_theta_K = length(theta_K),
      zero_trace = zero_trace,
      symmetric = symmetric,
      generic_type = generic_type,
      ...
    ),
    class = "ngme_operator"
  )
}


#' Print ngme operator
#'
#' @param x ngme operator object
#' @param padding number of white space padding in front
#' @param prefix prefix string
#' @param ... ...
#'
#' @return a list (operator specifications)
#' @export
print.ngme_operator <- function(x, padding = 0, prefix = "Model type", ...) {
  operator <- x
  pad_space <- paste(rep(" ", padding), collapse = "")
  pad_add4_space <- paste(rep(" ", padding + 4), collapse = "")
  pad_add8_space <- paste(rep(" ", padding + 8), collapse = "")

  limited_format <- function(values, digits = 3, max_items = 6) {
    if (length(values) == 0) {
      return("<none>")
    }
    formatted <- format(values, digits = digits)
    if (length(values) > max_items) {
      formatted <- c(formatted[seq_len(max_items)], "...")
    }
    paste(formatted, collapse = ", ")
  }

  model_name <- switch(operator$model,
    ar1 = "AR(1)",
    arma = "ARMA",
    matern = "Matern",
    tp = "Tensor product",
    bv = "Bivariate model",
    bv_matern = "Bivariate Matern model",
    bv2 = "Bivariate model (v2)",
    iid = "IID model",
    rw1 = "Random walk (order 1)",
    rw2 = "Random walk (order 2)",
    ou = "Ornstein-Uhlenbeck",
    re = "Random effect",
    generic = "Generic",
    generic_ns = "Generic (non-stationary)",
    var1 = "VAR(1) bivariate model",
    spacetime = if (operator$method == "galerkin") "Space-time (Galerkin)" else "Space-time (Implicit Euler)",
    {
      # Handle general AR models (ar2, ar3, etc.)
      if (grepl("^ar[0-9]+$", operator$model)) {
        p <- as.numeric(gsub("^ar([0-9]+)$", "\\1", operator$model))
        paste0("AR(", p, ")")
      } else {
        "Unknown"
      }
    }
  )

  model_name <- paste(prefix, ": ", model_name, "\n", sep = "")
  cat(pad_space, model_name, sep = "")

  parameter <- with(operator, switch(model,
    arma = {
      # Compute friendly AR/MA from raw theta_K for p,q<=2
      get_ar_from_raw <- function(raw, p) {
        if (p <= 0) {
          return(numeric(0))
        }
        if (p == 1) {
          return(ar1_th2a(raw[1]))
        }
        if (p == 2) {
          t1 <- ar1_th2a(raw[1])
          t2 <- ar1_th2a(raw[2])
          return(c(t1 * (1 - t2), t2))
        }
        numeric(0)
      }
      get_ma_from_raw <- function(raw, q) {
        if (q <= 0) {
          return(numeric(0))
        }
        if (q == 1) {
          return(ar1_th2a(raw[1]))
        }
        if (q == 2) {
          t1 <- ar1_th2a(raw[1])
          t2 <- ar1_th2a(raw[2])
          return(c(t1 * (1 - t2), t2))
        }
        numeric(0)
      }
      pp <- if (!is.null(p)) p else sum(startsWith(names(theta_K), "ar"))
      qq <- if (!is.null(q)) q else sum(startsWith(names(theta_K), "ma"))
      raw_ar <- if (pp > 0) theta_K[seq_len(pp)] else numeric(0)
      raw_ma <- if (qq > 0) theta_K[pp + seq_len(qq)] else numeric(0)
      ar_vals <- get_ar_from_raw(raw_ar, pp)
      ma_vals <- get_ma_from_raw(raw_ma, qq)
      if (length(ar_vals) > 0) {
        cat(pad_add4_space, "ar = ", limited_format(ar_vals, digits = 3), "\n", sep = "")
      } else {
        cat(pad_add4_space, "ar = <none>\n", sep = "")
      }
      if (length(ma_vals) > 0) {
        cat(pad_add4_space, "ma = ", limited_format(ma_vals, digits = 3), "\n", sep = "")
      } else {
        cat(pad_add4_space, "ma = <none>\n", sep = "")
      }
    },
    ar1 = cat(pad_add4_space, "rho = ", format(ar1_th2a(theta_K), digits = 3), "\n", sep = ""),
    matern = {
      # Print alpha from operator$alpha when available; fixed/free from operator$fix_alpha
      d <- if (!is.null(operator$spatial_dim)) operator$spatial_dim else 2
      L <- d / 2
      alpha_fixed <- isTRUE(operator$fix_alpha)
      if (!is.null(operator$alpha)) {
        alpha_val <- operator$alpha
      } else if (length(theta_K) >= 1 && !alpha_fixed) {
        # Fall back: derive from eta_alpha in theta_K[1]
        eta_alpha <- theta_K[1]
        alpha_val <- L + (4 - L) * (1 / (1 + exp(-eta_alpha)))
      } else {
        alpha_val <- NA
      }
      alpha_status <- if (alpha_fixed) "(fixed)" else "(free)"
      cat(pad_add4_space, "alpha = ", format(alpha_val, digits = 3), " ", alpha_status, "\n", sep = "")

      stationary <- isTRUE(operator$stationary)
      # theta_K layout:
      #   if fix_alpha:      [theta_kappa...]
      #   else (not fixed):  [eta_alpha, theta_kappa...]
      offset <- if (alpha_fixed) 0 else 1
      if (stationary) {
        if (length(theta_K) >= 1) {
          idx <- 1 + offset # stationary has single kappa
          kappa <- exp(theta_K[idx])
          cat(pad_add4_space, "kappa = ", format(kappa, digits = 3), "\n", sep = "")
        }
      } else {
        if (length(theta_K) > offset) {
          theta_kappa <- theta_K[(1 + offset):length(theta_K)]
          cat(pad_add4_space, "theta_kappa = ", limited_format(theta_kappa, digits = 3), "\n", sep = " ")
        }
      }
    },
    tp = {
      print(operator$first, padding = padding + 4, prefix = "first")
      print(operator$second, padding = padding + 4, prefix = "second")
    },
    bv = {
      model_names <- operator$model_names
      fix_theta <- isTRUE(operator$fix_theta)
      use_c_param <- isTRUE(operator$use_c_param)

      idx <- 1
      if (fix_theta) {
        theta <- operator$bv_theta
      } else {
        theta <- operator$param_trans[[idx]](theta_K[idx])
        idx <- idx + 1
      }
      rho <- operator$param_trans[[idx]](theta_K[idx])
      idx <- idx + 1

      cat(pad_add4_space, "theta = ", format(theta, digits = 3),
        if (fix_theta) " (fixed)" else "", "\n",
        sep = ""
      )
      cat(pad_add4_space, "rho = ", format(rho, digits = 3), "\n", sep = "")

      if (use_c_param) {
        c1 <- operator$param_trans[[idx]](theta_K[idx])
        idx <- idx + 1
        c2 <- operator$param_trans[[idx]](theta_K[idx])
        cat(pad_add4_space, "c1 = ", format(c1, digits = 3), "\n", sep = "")
        cat(pad_add4_space, "c2 = ", format(c2, digits = 3), "\n", sep = "")
      }

      print(operator$first, padding = padding + 4, prefix = model_names[[1]])
      print(operator$second, padding = padding + 4, prefix = model_names[[2]])
    },
    bv2 = {
      theta <- 0
      rho <- operator$param_trans[[1]](theta_K[1])
      c1 <- operator$param_trans[[2]](theta_K[2])
      c2 <- operator$param_trans[[3]](theta_K[3])
      cat(pad_add4_space, "theta = ", format(theta, digits = 3), "(fixed) \n", sep = "")
      cat(pad_add4_space, "rho = ", format(rho, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "c1 = ", format(c1, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "c2 = ", format(c2, digits = 3), "\n", sep = "")
      print(operator$first, padding = padding + 4, prefix = model_names[[1]])
      print(operator$second, padding = padding + 4, prefix = model_names[[2]])
    },
    bv_matern = {
      model_names <- operator$model_names
      fix_theta <- isTRUE(operator$fix_theta)

      idx <- 1
      if (fix_theta) {
        theta <- operator$bv_theta
      } else {
        theta <- operator$param_trans[[idx]](theta_K[idx])
        idx <- idx + 1
      }
      rho <- operator$param_trans[[idx]](theta_K[idx])
      idx <- idx + 1
      sd1 <- operator$param_trans[[idx]](theta_K[idx])
      idx <- idx + 1
      sd2 <- operator$param_trans[[idx]](theta_K[idx])

      cat(pad_add4_space, "theta = ", format(theta, digits = 3),
        if (fix_theta) " (fixed)" else "", "\n",
        sep = ""
      )
      cat(pad_add4_space, "rho = ", format(rho, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "sd1 = ", format(sd1, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "sd2 = ", format(sd2, digits = 3), "\n", sep = "")

      print(operator$first, padding = padding + 4, prefix = model_names[[1]])
      print(operator$second, padding = padding + 4, prefix = model_names[[2]])
    },
    bv_matern_normal = {
      theta <- 0
      rho <- operator$param_trans[[1]](theta_K[1])
      sd1 <- operator$param_trans[[2]](theta_K[2])
      sd2 <- operator$param_trans[[3]](theta_K[3])
      cat(pad_add4_space, "theta = ", format(theta, digits = 3), "(fixed) \n", sep = "")
      cat(pad_add4_space, "rho = ", format(rho, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "sd1 = ", format(sd1, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "sd2 = ", format(sd2, digits = 3), "\n", sep = "")
      print(operator$first, padding = padding + 4, prefix = model_names[[1]])
      print(operator$second, padding = padding + 4, prefix = model_names[[2]])
    },
    bv_matern_nig = {
      fix_theta <- operator$fix_theta
      theta <- if (fix_theta) operator$bv_theta else operator$param_trans[[1]](theta_K[1])
      rho <- operator$param_trans[[2 - fix_theta]](theta_K[2 - fix_theta])
      sd1 <- operator$param_trans[[3 - fix_theta]](theta_K[3 - fix_theta])
      sd2 <- operator$param_trans[[4 - fix_theta]](theta_K[4 - fix_theta])
      cat(pad_add4_space, "theta = ", format(theta, digits = 3), if (fix_theta) "(fixed)" else "", "\n", sep = "")
      cat(pad_add4_space, "rho = ", format(rho, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "sd1 = ", format(sd1, digits = 3), "\n", sep = "")
      cat(pad_add4_space, "sd2 = ", format(sd2, digits = 3), "\n", sep = "")
      print(operator$first, padding = padding + 4, prefix = model_names[[1]])
      print(operator$second, padding = padding + 4, prefix = model_names[[2]])
    },
    ou = {
      theta <- exp(theta_K)
      cat(pad_add4_space, "theta = ", format(theta, digits = 3), "\n", sep = "")
    },
    var1 = {
      A <- cayley_to_A(theta_K)
      sr <- max(abs(eigen(A, only.values = TRUE)$values))
      cat(pad_add4_space, "A (VAR coefficient matrix, Cayley reparameterization):\n", sep = "")
      cat(pad_add4_space, sprintf("  [[ %6.3f,  %6.3f ],\n", A[1, 1], A[1, 2]), sep = "")
      cat(pad_add4_space, sprintf("   [  %6.3f,  %6.3f ]]\n", A[2, 1], A[2, 2]), sep = "")
      cat(pad_add4_space, "spectral radius = ", format(sr, digits = 4), "\n", sep = "")
      cat(pad_add4_space, "unconstrained (p1, p2, p3, p4) = ",
        limited_format(theta_K, digits = 4), "\n",
        sep = ""
      )
    },
    re = {
      cat(pad_add4_space, "Covariance matrix (Sigma): \n")
      K <- build_effect_K(nrow(operator$K), operator$theta_K)
      print(solve(t(K) %*% K))
    },
    spacetime = {
      cat(pad_add4_space)
      cat("alpha =", format(alpha, digits = 2), "(fixed)", "\n", sep = " ")
      if (!fix_gamma) {
        if (shared_theta_gamma) {
          # When shared_theta_gamma = TRUE, only extract theta_gamma_x and use it for both
          theta_gamma_shared <- theta_K[3:(2 + n_theta_gamma_x)]
          cat(pad_add4_space)
          cat("theta_gamma (x and y) =", format(theta_gamma_shared, digits = 2), "\n", sep = " ")
        } else {
          # When shared_theta_gamma = FALSE, extract both separately
          theta_gamma_x <- theta_K[3:(2 + n_theta_gamma_x)]
          theta_gamma_y <- theta_K[(3 + n_theta_gamma_x):(2 + n_theta_gamma_x + n_theta_gamma_y)]
          cat(pad_add4_space)
          cat("theta_gamma_x =", format(theta_gamma_x, digits = 2), "\n", sep = " ")
          cat(pad_add4_space)
          cat("theta_gamma_y =", format(theta_gamma_y, digits = 2), "\n", sep = " ")
        }
      } else {
        cat(pad_add4_space)
        cat("theta_gamma_x =", format(theta_gamma_x, digits = 2), "(fixed)", "\n", sep = " ")
        cat(pad_add4_space)
        cat("theta_gamma_y =", format(theta_gamma_y, digits = 2), "(fixed)", "\n", sep = " ")
      }

      cat(pad_add4_space)
      cat("cc =", format(exp(theta_K[1]), digits = 2), "\n", sep = " ")

      cat(pad_add4_space)
      cat("kappa =", format(exp(theta_K[2]), digits = 2), "\n", sep = " ")
    },
    generic = {
      theta_vals <- operator$theta_K
      theta_names <- names(theta_vals)
      param_names <- operator$param_name
      trans_list <- operator$trans
      trans_names <- names(trans_list)

      get_label <- function(idx) {
        if (!is.null(param_names) && length(param_names) >= idx &&
          !is.na(param_names[idx]) && nzchar(param_names[idx])) {
          return(param_names[idx])
        }
        if (!is.null(theta_names) && length(theta_names) >= idx &&
          !is.na(theta_names[idx]) && nzchar(theta_names[idx])) {
          return(theta_names[idx])
        }
        paste0("theta_", idx)
      }

      if (length(theta_vals) > 0) {
        cat(pad_add4_space, "Parameters:\n", sep = "")
        for (i in seq_along(theta_vals)) {
          label <- get_label(i)
          cat(pad_add8_space, label, " raw: ", limited_format(theta_vals[i]), "\n", sep = "")

          trans_key <- NULL
          if (!is.null(theta_names) && length(theta_names) >= i &&
            !is.na(theta_names[i]) && nzchar(theta_names[i]) &&
            !is.null(trans_names) && theta_names[i] %in% trans_names) {
            trans_key <- theta_names[i]
          } else if (!is.null(param_names) && length(param_names) >= i &&
            !is.na(param_names[i]) && nzchar(param_names[i]) &&
            !is.null(trans_names) && param_names[i] %in% trans_names) {
            trans_key <- param_names[i]
          }

          if (!is.null(trans_key)) {
            mapping <- trans_list[[trans_key]]
            active_idx <- which(mapping != "null")
            if (length(active_idx) > 0) {
              combo <- paste0("#", active_idx, ": ", mapping[active_idx])
              cat(pad_add8_space, "matrix transforms: ", paste(combo, collapse = ", "), "\n", sep = "")
            }
          }
        }
      } else {
        cat(pad_add4_space, "Parameters: none\n", sep = "")
      }

      if (!is.null(operator$matrices)) {
        dims <- tryCatch(dim(operator$K), error = function(...) NULL)
        size_info <- if (!is.null(dims)) paste0(dims[1], " x ", dims[2]) else "unknown size"
        cat(pad_add4_space, "Stored matrices: ", length(operator$matrices), " (", size_info, ")\n", sep = "")
      }
    },
    generic_ns = {
      param_map <- operator$param_map
      param_names <- names(param_map)

      if (length(param_map) > 0) {
        cat(pad_add4_space, "Parameters:\n", sep = "")
        for (name in param_names) {
          indices <- param_map[[name]]
          coeff <- operator$theta_K[indices]
          cat(pad_add8_space, name, " coefficients: ", limited_format(coeff), "\n", sep = "")

          if (!is.null(operator$trans) && !is.null(operator$trans[[name]])) {
            cat(pad_add8_space, "transform: ", operator$trans[[name]], "\n", sep = "")
          }

          if (!is.null(operator$B_theta_K) && !is.null(operator$B_theta_K[[name]])) {
            basis <- operator$B_theta_K[[name]]
            cat(pad_add8_space, "basis dim: ", nrow(basis), " x ", ncol(basis), "\n", sep = "")
          }
        }
      } else {
        cat(pad_add4_space, "Parameters: none\n", sep = "")
      }

      if (!is.null(operator$position) && length(operator$position) > 0) {
        cat(pad_add4_space, "Matrix combinations (position indices):\n", sep = "")
        for (i in seq_along(operator$position)) {
          combo <- operator$position[[i]]
          cat(pad_add8_space, "#", i, ": ", paste(combo, collapse = " -> "), "\n", sep = "")
        }
      }

      if (!is.null(operator$matrices)) {
        dims <- tryCatch(dim(operator$K), error = function(...) NULL)
        size_info <- if (!is.null(dims)) paste0(dims[1], " x ", dims[2]) else "unknown size"
        cat(pad_add4_space, "Stored matrices: ", length(operator$matrices), " (", size_info, ")\n", sep = "")
      }
    },
    # Handle general AR models (ar2, ar3, etc.) or default case
    {
      if (grepl("^ar[0-9]+$", operator$model)) {
        # Print all rho parameters for general AR models
        for (i in seq_along(theta_K)) {
          cat(pad_add4_space, "rho", i, " = ", format(theta_K[i], digits = 3), "\n", sep = "")
        }
      } else {
        cat(pad_add4_space, "No parameter.", "\n", sep = "")
      }
    }
  ))

  invisible(operator)
}


# function for specify ngme.model basic structure
# keep same with cpp latent structure
ngme_model <- function(
    model,
    operator,
    noise = noise_normal(),
    theta_K = NULL,
    fix_K = FALSE,
    W_size = NULL,
    W = NULL,
    fix_W = FALSE,
    A = NULL,
    V_size = NULL,
    debug = FALSE,
    n_params = NULL,
    name = "field",
    mesh = NULL,
    par_string = NULL,
    map = NULL, # map is the covariates
    group = NULL,
    ...) {
  stopifnot(is.character(model))

  stopifnot(inherits(operator, "ngme_operator"))
  stopifnot(inherits(noise, "ngme_noise"))

  ope_params <- operator$param_name

  mu_params <- if (length(noise$theta_mu) == 0 || isTRUE(noise$fix_theta_mu)) {
    character(0)
  } else {
    paste0("mu_", seq_along(noise$theta_mu))
  }

  sigma_params <- {
    sigma_idx <- if (length(noise$theta_sigma) == 0) integer(0) else which(!noise$fix_theta_sigma)
    if (length(sigma_idx) == 0) character(0) else paste0("sigma_", sigma_idx)
  }

  nu_params <- if (length(noise$theta_nu) == 0 || isTRUE(noise$fix_theta_nu)) {
    character(0)
  } else {
    paste0("nu_", seq_along(noise$theta_nu))
  }

  rho_params <- if (length(noise$rho) == 0 || isTRUE(noise$fix_rho)) {
    character(0)
  } else {
    paste0("rho_", seq_along(noise$rho))
  }

  # Order matches the underlying parameter vector: operator -> mu -> sigma -> nu -> rho
  par_names <- c(ope_params, mu_params, sigma_params, nu_params, rho_params)

  if (is.null(n_params)) n_params <- length(operator$theta_K) + with(noise, n_params)

  structure(
    list(
      model = model,
      operator = operator,
      theta_K = operator$theta_K,
      noise_type = noise$noise_type,
      W_size = W_size,
      A = A,
      noise = noise,
      W = W,
      fix_theta_K = fix_K,
      fix_W = fix_W,
      V_size = V_size,
      n_params = n_params,
      debug = debug,
      par_names = par_names,
      name = name,
      mesh = mesh,
      map = map,
      n_map = length_map(map),
      group = group,
      ...
    ),
    class = "ngme_model"
  )
}

#' Print ngme model
#'
#' @param x ngme model object
#' @param padding number of white space padding in front
#' @param ... ...
#'
#' @return a list (model specifications)
#' @export
print.ngme_model <- function(x, padding = 0, ...) {
  model <- x
  pad_space <- paste(rep(" ", padding), collapse = "")

  # Print operator information using the fitted parameters stored on the object.
  op <- model$operator
  print.ngme_operator(op, padding = padding)

  # Print replicate information if replicates are used
  if (!is.null(model$replicate)) {
    n_replicates <- length(levels(model$replicate))
    cat(pad_space, "Number of replicates: ", n_replicates, "\n\n", sep = "")
  }

  # Print noise information
  model_type <- model$model
  print.ngme_noise(
    model$noise,
    padding = padding,
    model_type = model_type
  )

  invisible(model)
}

test_that("ngme_model par_names drop fixed sigma entries", {
  noise <- noise_normal(
    theta_sigma = c(0, 1),
    B_sigma = diag(2),
    fix_theta_sigma = c(TRUE, FALSE)
  )

  op <- ngme2:::ngme_operator(mesh = NULL, model = "iid", K = diag(1), h = 1)

  model <- ngme2:::ngme_model(
    model = "iid",
    operator = op,
    noise = noise,
    theta_K = op$theta_K
  )

  expect_equal(model$par_names, "sigma_2")
  expect_equal(model$n_params, length(model$par_names))
})

test_that("ngme_replicate par_names stay consistent with fixed sigma", {
  noise <- noise_normal(
    theta_sigma = c(0, 1),
    B_sigma = diag(2),
    fix_theta_sigma = c(TRUE, FALSE)
  )
  ctrl <- control_ngme(beta_init = numeric(0))

  ng <- ngme_replicate(
    Y = NULL,
    X = matrix(numeric(0), nrow = 0, ncol = 0),
    noise = noise,
    models = list(),
    control_ngme = ctrl,
    standardize = FALSE
  )

  expect_equal(ng$par_names, "sigma_2")
  expect_equal(ng$n_params, length(ng$par_names))
})

get_noise_info <- function(noise) {
  if (length(noise$noise_type) == 1) {
    # mu
    if (noise$n_theta_mu == 0) {
      name_mu <- NULL
    } else if (noise$n_theta_mu == 1 && is_stationary(noise$B_mu)) {
      name_mu <- "mu"
    } else {
      name_mu <- paste("theta_mu", seq_len(noise$n_theta_mu))
    }
    trans_mu <- rep(list(identity), noise$n_theta_mu)

    # sigma
    if (noise$noise == "normal_nig") {
      if (is_stationary(noise$B_sigma_nig)) {
        name_sigma_nig <- "sigma_nig"
        trans_sigma_nig <- list(exp)
      } else {
        name_sigma_nig <- paste("theta_sigma_nig", seq_len(noise$n_theta_sigma_nig))
        trans_sigma_nig <- rep(list(identity), noise$n_theta_sigma_nig)
      }
      if (is_stationary(noise$B_sigma_normal)) {
        name_sigma_normal <- "sigma_normal"
        trans_sigma_normal <- list(exp)
      } else {
        name_sigma_normal <- paste("theta_sigma_normal", seq_len(noise$n_theta_sigma_normal))
        trans_sigma_normal <- rep(list(identity), noise$n_theta_sigma_normal)
      }
      name_sigma <- c(name_sigma_nig, name_sigma_normal)
      trans_sigma <- c(trans_sigma_nig, trans_sigma_normal)
    } else {
      if (is_stationary(noise$B_sigma) ||
        (ncol(noise$B_sigma) == 2 && noise$B_sigma[1, 1] == 1 && all(noise$B_sigma[-1, 1] == 0)) # RW1 pattern
      ) {
        name_sigma <- "sigma"
        trans_sigma <- list(exp)
      } else {
        name_sigma <- paste("theta_sigma", seq_len(noise$n_theta_sigma))
        trans_sigma <- rep(list(identity), noise$n_theta_sigma)
      }
    }

    # nu
    if (noise$n_theta_nu == 0) {
      name_nu <- trans_nu <- NULL
    } else if (is_stationary(noise$B_nu)) {
      nu_lower_bound <- if (is.null(noise$nu_lower_bound)) 0 else noise$nu_lower_bound
      nu_trans <- function(theta_nu) {
        nu_lower_bound + exp(theta_nu)
      }
      name_nu <- "nu"
      trans_nu <- list(nu_trans)
    } else {
      name_nu <- paste("theta_nu", seq_len(noise$n_theta_nu))
      trans_nu <- rep(list(identity), noise$n_theta_nu)
    }

    if (noise$fix_theta_mu) name_mu <- trans_mu <- NULL
    if (all(noise$fix_theta_sigma)) {
      name_sigma <- trans_sigma <- NULL
    }
    if (noise$fix_theta_nu) name_nu <- trans_nu <- NULL

    ts <- list(
      # for bv noise
      all = list(
        name_mu = name_mu,
        name_sigma = name_sigma,
        name_nu = name_nu,
        trans_mu = trans_mu,
        trans_sigma = trans_sigma,
        trans_nu = trans_nu
      ),
      # for plot
      name = c(name_mu, name_sigma, name_nu),
      trans = c(trans_mu, trans_sigma, trans_nu)
    )
  } else {
    # bivariate noise
    n1 <- get_noise_info(noise$bv_noises[[1]])
    n2 <- get_noise_info(noise$bv_noises[[2]])
    n1 <- lapply(n1$all, function(x) if (is.character(x)) paste(x, "(1st)") else x)
    n2 <- lapply(n2$all, function(x) if (is.character(x)) paste(x, "(2nd)") else x)

    # re-arrange
    ts <- list(
      name = c(
        n1$name_mu, n2$name_mu,
        n1$name_sigma, n2$name_sigma,
        n1$name_nu, n2$name_nu
      ),
      trans = c(
        n1$trans_mu, n2$trans_mu,
        n1$trans_sigma, n2$trans_sigma,
        n1$trans_nu, n2$trans_nu
      )
    )
  }
  if (noise$corr_measurement) {
    ts$name <- c(ts$name, "rho(measurement)")
    ts$trans <- c(ts$trans, list(ar1_th2a))
  }
  ts
}

get_latent_info <- function(latent) {
  ts <- get_noise_info(latent$noise)
  ts$name <- c(latent$operator$param_name, ts$name)
  ts$trans <- c(latent$operator$param_trans, ts$trans)
  ts
}

get_feff_plot_names <- function(feff) {
  n_feff <- length(feff)
  if (n_feff == 0) {
    return(NULL)
  }

  feff_names <- names(feff)
  if (is.null(feff_names) || anyNA(feff_names) || any(!nzchar(feff_names))) {
    return(paste("fixed effect", seq_len(n_feff)))
  }

  feff_names
}

disambiguate_trace_param_names <- function(param_names) {
  if (length(param_names) == 0) {
    return(param_names)
  }
  make.unique(as.character(param_names))
}

#' Get trace trajectories from ngme fitting
#'
#' Extract numerical trajectories of parameters during ngme optimization
#' without creating plots. This function provides the underlying data
#' used by traceplot().
#'
#' @param ngme ngme object
#' @param name name of latent models, otherwise get fixed effects and measurement noise
#' should be in names(ngme$models) or other
#' @param apply_transform logical, whether to apply parameter transformations (default: TRUE)
#'
#' @return list containing:
#' \describe{
#'   \item{trajectories}{named list of trajectory matrices for each parameter}
#'   \item{parameter_names}{character vector of parameter names}
#'   \item{transformations}{list of transformation functions used}
#'   \item{n_chains}{number of chains}
#'   \item{n_iterations}{number of iterations}
#' }
#' @export
#'
get_trace_trajectories <- function(
    ngme,
    name = "general",
    apply_transform = TRUE) {
  stopifnot(inherits(ngme, "ngme"))
  stopifnot(!is.null(name))

  ngme <- ngme$replicates[[1]]

  if (name %in% names(ngme$models)) {
    # Get trajectory of parameters of the model
    traj <- attr(ngme$models[[name]], "lat_traj")
    stopifnot(
      "Please run ngme() to estimate the model before using get_trace_trajectories()" = !is.null(traj)
    )
    ts <- get_latent_info(ngme$models[[name]])
  } else {
    # Get trajectory of parameters of the noise
    traj <- attr(ngme, "block_traj")
    stopifnot(
      "Please run ngme() to estimate the model before using get_trace_trajectories()" = !is.null(traj)
    )
    # get titles
    ts <- get_noise_info(ngme$noise)
    name_feff <- get_feff_plot_names(ngme$feff)
    trans_feff <- rep(list(identity), length(ngme$feff))
    ts$name <- c(ts$name, name_feff)
    ts$trans <- c(ts$trans, trans_feff)
  }
  ts$name <- disambiguate_trace_param_names(ts$name)

  n_parameters <- nrow(traj[[1]])
  n_chains <- length(traj)

  # Get lengths and use minimum if chains have different lengths
  lengths <- sapply(traj, function(x) ncol(x))
  if (length(unique(lengths)) > 1) {
    min_length <- min(lengths)
    traj <- lapply(traj, function(x) x[, seq_len(min_length), drop = FALSE])
    warning("Some chains are of different lengths. Only the minimum length is used.")
  }
  n_iterations <- lengths[1]

  # Extract trajectories for each parameter
  trajectories <- list()

  for (idx in seq_len(n_parameters)) {
    # Extract the idx-th parameter from each chain
    param_data <- lapply(traj, function(x) x[idx, , drop = FALSE])
    param_data <- lapply(param_data, as.numeric)

    # Combine into matrix: rows = iterations, columns = chains
    param_matrix <- do.call(cbind, param_data)
    colnames(param_matrix) <- paste0("chain_", seq_len(n_chains))

    # Apply transformation if requested
    if (apply_transform && !is.null(ts$trans[[idx]])) {
      ff <- ts$trans[[idx]]
      param_matrix <- apply(param_matrix, 2, ff)
      if (!is.matrix(param_matrix)) {
        param_matrix <- matrix(param_matrix, nrow = n_iterations, ncol = n_chains)
        colnames(param_matrix) <- paste0("chain_", seq_len(n_chains))
      }
    }

    trajectories[[ts$name[idx]]] <- param_matrix
  }

  result <- list(
    trajectories = trajectories,
    parameter_names = ts$name,
    transformations = ts$trans,
    n_chains = n_chains,
    n_iterations = n_iterations
  )

  class(result) <- "ngme_trajectories"
  return(result)
}

#' Print method for ngme_trajectories
#' @param x ngme_trajectories object
#' @param ... additional arguments
#' @return Invisibly returns \code{x}, an \code{ngme_trajectories} object. The
#'   method is called for its side effect of printing a summary of stored
#'   parameter trajectory matrices, including parameter names, chain count, and
#'   iteration count.
#' @export
print.ngme_trajectories <- function(x, ...) {
  cat("NGME Parameter Trajectories\n")
  cat("==========================\n")
  cat("Parameters:", length(x$parameter_names), "\n")
  cat("Chains:", x$n_chains, "\n")
  cat("Iterations:", x$n_iterations, "\n\n")

  cat("Parameter names:\n")
  for (i in seq_along(x$parameter_names)) {
    cat(sprintf("  [%d] %s\n", i, x$parameter_names[i]))
  }

  cat("\nAccess trajectories via $trajectories$parameter_name\n")
  cat("Each trajectory is a matrix: rows = iterations, columns = chains\n")
  invisible(x)
}

#' Calculate parameter distance from true values
#'
#' Compute ||theta - theta_hat|| distance for all parameters across iterations.
#' This provides a single measure of convergence combining all parameters.
#'
#' @param trajectories ngme_trajectories object from get_trace_trajectories()
#' @param true_values named list or vector of true parameter values
#' @param norm_type character, type of norm to use: "euclidean" (default), "manhattan", "max"
#' @param chain_summary character, how to summarize across chains: "mean" (default), "median", "chain1"
#'
#' @return numeric vector of distances across iterations
#' @export
#'
get_parameter_distance <- function(
    trajectories,
    true_values,
    norm_type = "euclidean",
    chain_summary = "mean") {
  stopifnot(inherits(trajectories, "ngme_trajectories"))
  stopifnot(is.list(true_values) || is.numeric(true_values))
  stopifnot(norm_type %in% c("euclidean", "manhattan", "max"))
  stopifnot(chain_summary %in% c("mean", "median", "chain1"))

  # Match parameter names
  if (is.numeric(true_values) && is.null(names(true_values))) {
    if (length(true_values) != length(trajectories$parameter_names)) {
      stop("Length of true_values must match number of parameters")
    }
    names(true_values) <- trajectories$parameter_names
  }

  # Check all required parameters are provided
  missing_params <- setdiff(trajectories$parameter_names, names(true_values))
  if (length(missing_params) > 0) {
    stop("Missing true values for parameters: ", paste(missing_params, collapse = ", "))
  }

  n_iterations <- trajectories$n_iterations
  distances <- numeric(n_iterations)

  for (iter in seq_len(n_iterations)) {
    # Extract parameter values for this iteration
    param_values <- numeric(length(trajectories$parameter_names))

    for (i in seq_along(trajectories$parameter_names)) {
      param_name <- trajectories$parameter_names[i]
      traj_matrix <- trajectories$trajectories[[param_name]]

      # Summarize across chains for this iteration
      iter_values <- traj_matrix[iter, , drop = TRUE]
      param_values[i] <- switch(chain_summary,
        "mean" = mean(iter_values, na.rm = TRUE),
        "median" = median(iter_values, na.rm = TRUE),
        "chain1" = iter_values[1]
      )
    }

    # Calculate distance to true values
    true_vec <- unname(true_values[trajectories$parameter_names])
    diff_vec <- param_values - true_vec

    distances[iter] <- switch(norm_type,
      "euclidean" = sqrt(sum(diff_vec^2)),
      "manhattan" = sum(abs(diff_vec)),
      "max" = max(abs(diff_vec))
    )
  }

  structure(distances,
    norm_type = norm_type,
    chain_summary = chain_summary,
    parameter_names = trajectories$parameter_names,
    true_values = true_values[trajectories$parameter_names],
    class = c("parameter_distance", "numeric")
  )
}

#' Print method for parameter_distance
#' @param x parameter_distance object
#' @param ... additional arguments
#' @return Invisibly returns \code{x}, a numeric vector with class
#'   \code{parameter_distance}. The method is called for its side effect of
#'   printing the norm type, chain summary, number of iterations, final distance,
#'   and true parameter values.
#' @export
print.parameter_distance <- function(x, ...) {
  cat("Parameter Distance Trajectory\n")
  cat("============================\n")
  cat("Norm type:", attr(x, "norm_type"), "\n")
  cat("Chain summary:", attr(x, "chain_summary"), "\n")
  cat("Iterations:", length(x), "\n")
  cat("Final distance:", sprintf("%.6f", x[length(x)]), "\n\n")

  cat("True parameter values:\n")
  true_vals <- attr(x, "true_values")
  for (name in names(true_vals)) {
    cat(sprintf("  %s: %.4f\n", name, true_vals[[name]]))
  }

  cat("\nUse plot() to visualize convergence\n")
  invisible(x)
}

#' Plot method for parameter_distance
#' @param x parameter_distance object
#' @param ... additional arguments for ggplot
#' @return A \code{ggplot} object showing the parameter-distance trajectory over
#'   optimization iterations. The y-axis is the stored distance
#'   \eqn{||theta - \hat{theta}||}; plot annotations report the norm and
#'   across-chain summary used when \code{x} was computed.
#' @export
plot.parameter_distance <- function(x, ...) {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("ggplot2 package is required for plotting")
  }

  iteration <- distance <- NULL # avoid R CMD check warnings

  df <- data.frame(
    iteration = seq_along(x),
    distance = as.numeric(x)
  )

  ggplot2::ggplot(df, ggplot2::aes(x = iteration, y = distance)) +
    ggplot2::geom_line() +
    ggplot2::labs(
      title = paste(
        "Parameter Distance Trajectory (",
        attr(x, "norm_type"), "norm )"
      ),
      subtitle = paste("Chain summary:", attr(x, "chain_summary")),
      x = "Iteration",
      y = expression(paste("||", theta, " - ", hat(theta), "||"))
    ) +
    ggplot2::theme_minimal()
}

#' Trace plot of ngme fitting
#'
#' @param ngme ngme object
#' @param name name of latent models, otherwise plot fixed effects and measurement noise.
#'   Use \code{"all"} to plot all latent models plus data parameters in one figure.
#' should be in names(ngme$models) or other
#' @param moving_window moving window for the traceplot
#' @param hline vector, add hline to each plot
#' @param combine bool, if TRUE, return a single faceted ggplot; otherwise return
#'   a list of ggplot objects and print them one by one.
#' @param ncol number of facet columns when \code{combine = TRUE}. If \code{NULL}
#'   (default), use 2 columns for multi-parameter plots and 1 for single-parameter plots.
#'
#' @return A ggplot object when \code{combine = TRUE}; otherwise a list of ggplot
#'   objects. The mean parameter trajectories are stored in the \code{avg_lines}
#'   attribute of the returned object.
#' @export
#'
traceplot <- function(
    ngme,
    name = "all",
    moving_window = 1,
    hline = NULL,
    combine = TRUE,
    ncol = NULL) {
  stopifnot(inherits(ngme, "ngme"))
  stopifnot(!is.null(name))
  ngme <- ngme$replicates[[1]]
  ps <- list()

  get_latent_traj_and_ts <- function(latent_obj, latent_name = NULL) {
    traj <- attr(latent_obj, "lat_traj")
    stopifnot(
      "Please run ngme() to estimate the model before using traceplot()" = !is.null(traj)
    )
    ts <- get_latent_info(latent_obj)

    # Special handling for ARMA(p,q<=2): convert raw (PACF) to AR/MA coefficients
    if (!is.null(latent_obj$operator) && identical(latent_obj$operator$model, "arma")) {
      p <- latent_obj$operator$p %||% 0L
      q <- latent_obj$operator$q %||% 0L
      if (p <= 2 && q <= 2) {
        # Helper: map raw -> coeff per time column
        map_ar_coeff <- function(raw_ar) {
          if (length(raw_ar) == 0) {
            return(numeric(0))
          }
          if (length(raw_ar) == 1) {
            t1 <- ar1_th2a(raw_ar[1])
            return(c(t1))
          }
          t1 <- ar1_th2a(raw_ar[1])
          t2 <- ar1_th2a(raw_ar[2])
          c(t1 * (1 - t2), t2)
        }
        map_ma_coeff <- function(raw_ma) {
          if (length(raw_ma) == 0) {
            return(numeric(0))
          }
          if (length(raw_ma) == 1) {
            s1 <- ar1_th2a(raw_ma[1])
            return(c(s1))
          }
          s1 <- ar1_th2a(raw_ma[1])
          s2 <- ar1_th2a(raw_ma[2])
          c(s1 * (1 - s2), s2)
        }
        # Rewrite trajectories per chain
        traj <- lapply(traj, function(M) {
          M <- as.matrix(M)
          if ((p + q) > 0 && nrow(M) >= (p + q)) {
            ar_idx <- if (p > 0) seq_len(p) else integer(0)
            ma_idx <- if (q > 0) (p + seq_len(q)) else integer(0)
            for (col in seq_len(ncol(M))) {
              if (p > 0) {
                phi <- map_ar_coeff(M[ar_idx, col])
                M[seq_len(length(phi)), col] <- phi
              }
              if (q > 0) {
                theta <- map_ma_coeff(M[ma_idx, col])
                M[p + seq_len(length(theta)), col] <- theta
              }
            }
            M
          } else {
            M
          }
        })
        # Replace transforms for AR/MA with identity since we already mapped
        if ((p + q) > 0 && length(ts$trans) >= (p + q)) {
          ts$trans[seq_len(p + q)] <- replicate(p + q, identity, simplify = FALSE)
        }
      }
    }

    if (!is.null(latent_name)) {
      ts$name <- paste0(ts$name, " (", latent_name, ")")
    }

    list(traj = traj, ts = ts)
  }

  if (identical(name, "all")) {
    latent_names <- names(ngme$models)
    latent_parts <- lapply(latent_names, function(lat_name) {
      get_latent_traj_and_ts(ngme$models[[lat_name]], latent_name = lat_name)
    })

    # Data part (measurement noise + fixed effects)
    traj_noise <- attr(ngme, "block_traj")
    stopifnot(
      "Please run ngme() to estimate the model before using traceplot()" = !is.null(traj_noise)
    )
    ts_noise <- get_noise_info(ngme$noise)
    name_feff <- get_feff_plot_names(ngme$feff)
    trans_feff <- rep(list(identity), length(ngme$feff))
    ts_noise$name <- c(ts_noise$name, name_feff)
    ts_noise$trans <- c(ts_noise$trans, trans_feff)

    parts <- c(latent_parts, list(list(traj = traj_noise, ts = ts_noise)))

    n_chains <- length(parts[[1]]$traj)
    if (any(vapply(parts, function(p) length(p$traj) != n_chains, logical(1)))) {
      stop("All components must have the same number of chains for name = 'all'.")
    }

    traj <- lapply(seq_len(n_chains), function(chain_idx) {
      mats <- lapply(parts, function(p) p$traj[[chain_idx]])
      mats <- lapply(mats, function(m) {
        if (is.null(dim(m))) {
          matrix(m, nrow = 1)
        } else {
          as.matrix(m)
        }
      })

      ncol_vec <- vapply(mats, ncol, integer(1))
      nrow_vec <- vapply(mats, nrow, integer(1))
      expected_iter <- max(ncol_vec)
      if (max(nrow_vec) > expected_iter) {
        expected_iter <- max(nrow_vec)
      }

      mats <- Map(function(m, nc, nr) {
        if (nc != expected_iter && nr == expected_iter) {
          m <- t(m)
        }
        m
      }, mats, ncol_vec, nrow_vec)

      lengths <- vapply(mats, ncol, integer(1))
      if (length(unique(lengths)) > 1) {
        warning("Some components are of different lengths. Only the minimum length is used.")
      }
      min_length <- min(lengths)
      mats <- lapply(mats, function(m) m[, seq_len(min_length), drop = FALSE])
      do.call(rbind, mats)
    })

    ts <- list(
      name = unlist(lapply(parts, function(p) p$ts$name), use.names = FALSE),
      trans = do.call(c, lapply(parts, function(p) p$ts$trans))
    )
  } else if (name %in% names(ngme$models)) {
    # Plot trajectory of parameters of the model
    lt <- get_latent_traj_and_ts(ngme$models[[name]])
    traj <- lt$traj
    ts <- lt$ts
  } else {
    # Plot trajectory of parameters of the noise
    traj <- attr(ngme, "block_traj")
    stopifnot(
      "Please run ngme() to estimate the model before using traceplot()" = !is.null(traj)
    )
    # get titles
    ts <- get_noise_info(ngme$noise)
    name_feff <- get_feff_plot_names(ngme$feff)
    trans_feff <- rep(list(identity), length(ngme$feff))
    ts$name <- c(ts$name, name_feff)
    ts$trans <- c(ts$trans, trans_feff)
  }
  ts$name <- disambiguate_trace_param_names(ts$name)

  # Record mean trajectories for comparison later.
  avg_lines <- NULL
  plot_data_all <- list()
  mean_data_all <- list()
  hline_data_all <- list()

  n_parameters <- nrow(traj[[1]]) # number of parameters to draw
  for (idx in seq_len(n_parameters)) {
    # extract the idx-th parameter from each chain
    df <- lapply(traj, function(x) x[idx, , drop = F])
    df <- lapply(df, as.numeric)

    # if not all chains are of the same length, use the minimum length
    lengths <- sapply(df, length)
    if (length(unique(lengths)) > 1) {
      min_length <- min(lengths)
      df <- lapply(df, function(x) x[seq_len(min_length)])
      warning("Some chains are of different lengths. Only the minimum length is used.")
    }

    df <- as.data.frame(df)
    df$x <- seq_len(nrow(df))
    x <- NULL # get around check note
    keys <- setdiff(names(df), "x")
    df_long <- data.frame(
      x = rep(df$x, times = length(keys)),
      key = rep(keys, each = nrow(df)),
      value = as.numeric(unlist(df[keys], use.names = FALSE)),
      stringsAsFactors = FALSE
    )
    ff <- ts$trans[[idx]]
    df_long$value <- ff(df_long$value)

    # update df using moving window
    if (moving_window > 1) {
      if (!requireNamespace("zoo", quietly = TRUE)) {
        message("Package 'zoo' is not installed. Using original data without moving average.")
        df_long$moving_avg <- df_long$value
      } else {
        df_long$moving_avg <- df_long$value
        for (k in keys) {
          idx_k <- df_long$key == k
          df_long$moving_avg[idx_k] <- zoo::rollapply(
            df_long$value[idx_k],
            width = moving_window,
            FUN = mean,
            align = "right",
            fill = NA
          )
        }
      }
    } else {
      df_long$moving_avg <- df_long$value
    }
    df_long <- df_long[!is.na(df_long$moving_avg), , drop = FALSE]

    df_mean <- aggregate(moving_avg ~ x, data = df_long, FUN = mean)
    names(df_mean)[names(df_mean) == "moving_avg"] <- "mean_moving_avg"
    parameter_name <- ts$name[[idx]]
    df_long$parameter <- parameter_name
    df_mean$parameter <- parameter_name

    plot_data_all[[idx]] <- df_long
    mean_data_all[[idx]] <- df_mean

    hline_idx <- if (length(hline) >= idx) hline[[idx]] else NULL
    if (!is.null(hline_idx) && length(hline_idx) > 0) {
      hline_data_all[[idx]] <- data.frame(
        parameter = rep(parameter_name, length(hline_idx)),
        yintercept = as.numeric(hline_idx),
        stringsAsFactors = FALSE
      )
    }

    if (!combine) {
      p <- ggplot2::ggplot() +
        ggplot2::geom_line(
          data = df_long,
          mapping = ggplot2::aes(
            x = .data[["x"]],
            y = .data[["moving_avg"]],
            group = .data[["key"]]
          )
        ) +
        ggplot2::geom_line(
          data = df_mean,
          ggplot2::aes(x = .data[["x"]], y = .data[["mean_moving_avg"]]),
          col = "red"
        ) +
        ggplot2::labs(title = parameter_name) +
        ggplot2::xlab(NULL) +
        ggplot2::ylab(NULL) +
        ggplot2::theme(
          plot.title = ggplot2::element_text(size = 14, face = "bold")
        )

      if (!is.null(hline_idx) && length(hline_idx) > 0) {
        p <- p + ggplot2::geom_hline(
          yintercept = as.numeric(hline_idx),
          color = "blue"
        )
      }
      ps[[idx]] <- p
    }

    avg_lines[[parameter_name]] <- df_mean$mean_moving_avg
  }

  # Display results based on combine parameter
  result <- NULL
  if (combine) {
    if (length(plot_data_all) == 0) {
      stop("No parameters available to plot.")
    }
    facet_ncol <- if (is.null(ncol)) {
      if (length(plot_data_all) > 1) 2 else 1
    } else {
      if (!is.numeric(ncol) || length(ncol) != 1 || is.na(ncol) || ncol < 1 || abs(ncol - round(ncol)) > 0) {
        stop("`ncol` must be a positive integer or NULL.")
      }
      as.integer(ncol)
    }
    plot_data <- do.call(rbind, plot_data_all)
    mean_data <- do.call(rbind, mean_data_all)
    plot_levels <- unique(vapply(plot_data_all, function(x) x$parameter[[1]], character(1)))
    plot_data$parameter <- factor(plot_data$parameter, levels = plot_levels)
    mean_data$parameter <- factor(mean_data$parameter, levels = plot_levels)
    if (length(hline_data_all) > 0) {
      hline_data <- do.call(rbind, hline_data_all)
      hline_data$parameter <- factor(hline_data$parameter, levels = plot_levels)
    } else {
      hline_data <- NULL
    }

    result <- ggplot2::ggplot() +
      ggplot2::geom_line(
        data = plot_data,
        mapping = ggplot2::aes(
          x = .data[["x"]],
          y = .data[["moving_avg"]],
          group = .data[["key"]]
        )
      ) +
      ggplot2::geom_line(
        data = mean_data,
        mapping = ggplot2::aes(x = .data[["x"]], y = .data[["mean_moving_avg"]]),
        col = "red"
      ) +
      ggplot2::facet_wrap(~parameter, scales = "free_y", ncol = facet_ncol) +
      ggplot2::xlab(NULL) +
      ggplot2::ylab(NULL) +
      ggplot2::theme(
        strip.text = ggplot2::element_text(size = 14, face = "bold")
      )

    if (!is.null(hline_data)) {
      result <- result + ggplot2::geom_hline(
        data = hline_data,
        mapping = ggplot2::aes(yintercept = .data[["yintercept"]]),
        color = "blue",
        inherit.aes = FALSE
      )
    }

    # Keep the historical behavior of drawing a plot when traceplot() is called.
    print(result)
  } else {
    # Print plots one by one
    for (p in ps) {
      print(p)
    }
    result <- ps
  }

  attr(result, "avg_lines") <- avg_lines

  invisible(result)
}

#' Plot Posterior Distributions from SGLD Samples
#'
#' Visualize marginal posterior distributions of parameters extracted from
#' [ngme_sgld_samples()] or summarized by [ngme_sgld_ci()].
#'
#' @param x a data.frame (or list of data.frames) returned by
#'   [ngme_sgld_samples()], or an object of class `ngme_sgld_ci`.
#' @param parameters optional character vector of parameter names to plot.
#'   Defaults to all parameter columns.
#' @param type character; `"density"` (default) or `"histogram"`.
#' @param by_chain logical; if `TRUE`, color distributions by chain when a
#'   `.chain` column is available.
#' @param show_estimate logical; add a vertical line at the posterior mean.
#' @param show_interval logical; add vertical lines for the equal-tail interval.
#' @param lower lower quantile probability used for intervals. If `x` is an
#'   `ngme_sgld_ci` object, the stored value is used by default.
#' @param upper upper quantile probability used for intervals. If `x` is an
#'   `ngme_sgld_ci` object, the stored value is used by default.
#' @param bins number of bins for histograms.
#' @param ncol number of facet columns. If `NULL`, use 2 columns for multiple
#'   parameters and 1 for a single parameter.
#' @param ... unused.
#'
#' @return A faceted ggplot object.
#' @export
posterior_plot <- function(
    x,
    parameters = NULL,
    type = c("density", "histogram"),
    by_chain = FALSE,
    show_estimate = TRUE,
    show_interval = TRUE,
    lower = NULL,
    upper = NULL,
    bins = 30,
    ncol = NULL,
    ...) {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("ggplot2 package is required for plotting")
  }

  type <- match.arg(type)

  if (inherits(x, "ngme_sgld_ci")) {
    samples <- x$samples
    if (is.null(lower)) lower <- x$lower
    if (is.null(upper)) upper <- x$upper
    estimate_map <- x$estimates
    interval_map <- x$ci
  } else if (is.data.frame(x)) {
    samples <- x
    estimate_map <- NULL
    interval_map <- NULL
  } else if (is.list(x) &&
    length(x) >= 1 &&
    all(vapply(x, is.data.frame, logical(1)))) {
    samples <- do.call(rbind, x)
    rownames(samples) <- NULL
    estimate_map <- NULL
    interval_map <- NULL
  } else {
    stop(
      "`x` must be an `ngme_sgld_ci` object or a data.frame/list returned by `ngme_sgld_samples()`."
    )
  }

  if (is.null(lower)) lower <- 0.025
  if (is.null(upper)) upper <- 0.975

  stopifnot(
    is.numeric(lower),
    length(lower) == 1,
    is.finite(lower),
    lower > 0,
    lower < 1,
    is.numeric(upper),
    length(upper) == 1,
    is.finite(upper),
    upper > 0,
    upper < 1,
    lower < upper,
    is.logical(by_chain),
    length(by_chain) == 1,
    is.logical(show_estimate),
    length(show_estimate) == 1,
    is.logical(show_interval),
    length(show_interval) == 1,
    is.numeric(bins),
    length(bins) == 1,
    is.finite(bins),
    bins >= 1,
    abs(bins - round(bins)) < sqrt(.Machine$double.eps)
  )

  param_cols <- setdiff(colnames(samples), c(".chain", ".draw", ".iter"))
  stopifnot(length(param_cols) >= 1)
  stopifnot(
    "All parameter columns in `x` must be numeric." =
      all(vapply(samples[, param_cols, drop = FALSE], is.numeric, logical(1)))
  )

  if (!is.null(parameters)) {
    stopifnot(is.character(parameters))
    missing_params <- setdiff(parameters, param_cols)
    if (length(missing_params) > 0) {
      stop(
        "Unknown parameter(s): ",
        paste(missing_params, collapse = ", ")
      )
    }
    param_cols <- parameters
  }

  facet_ncol <- if (is.null(ncol)) {
    if (length(param_cols) > 1) 2L else 1L
  } else {
    if (!is.numeric(ncol) || length(ncol) != 1 || is.na(ncol) ||
      ncol < 1 || abs(ncol - round(ncol)) > 0) {
      stop("`ncol` must be a positive integer or NULL.")
    }
    as.integer(ncol)
  }

  long_df <- data.frame(
    parameter = rep(param_cols, each = nrow(samples)),
    value = as.numeric(unlist(samples[, param_cols, drop = FALSE], use.names = FALSE)),
    stringsAsFactors = FALSE
  )
  if (".chain" %in% colnames(samples)) {
    long_df$.chain <- rep(samples$.chain, times = length(param_cols))
  }

  summary_df <- data.frame(
    parameter = param_cols,
    estimate = numeric(length(param_cols)),
    lower = numeric(length(param_cols)),
    upper = numeric(length(param_cols)),
    stringsAsFactors = FALSE
  )

  for (idx in seq_along(param_cols)) {
    param_name <- param_cols[[idx]]
    vals <- samples[[param_name]]

    if (!is.null(estimate_map) && !is.null(names(estimate_map)) &&
      param_name %in% names(estimate_map)) {
      summary_df$estimate[idx] <- estimate_map[[param_name]]
    } else {
      summary_df$estimate[idx] <- mean(vals, na.rm = TRUE)
    }

    if (!is.null(interval_map) && is.matrix(interval_map) &&
      param_name %in% rownames(interval_map)) {
      summary_df$lower[idx] <- interval_map[param_name, "lower"]
      summary_df$upper[idx] <- interval_map[param_name, "upper"]
    } else {
      qq <- stats::quantile(vals, probs = c(lower, upper), na.rm = TRUE, names = FALSE)
      summary_df$lower[idx] <- qq[[1]]
      summary_df$upper[idx] <- qq[[2]]
    }
  }

  long_df$parameter <- factor(long_df$parameter, levels = param_cols)
  summary_df$parameter <- factor(summary_df$parameter, levels = param_cols)

  p <- ggplot2::ggplot(long_df, ggplot2::aes(x = .data[["value"]]))

  if (identical(type, "density")) {
    if (isTRUE(by_chain) && ".chain" %in% colnames(long_df)) {
      p <- p + ggplot2::geom_density(
        ggplot2::aes(
          color = factor(.data[[".chain"]]),
          fill = factor(.data[[".chain"]])
        ),
        alpha = 0.2
      ) +
        ggplot2::labs(color = "Chain", fill = "Chain")
    } else {
      p <- p + ggplot2::geom_density(fill = "grey70", alpha = 0.5)
    }
  } else {
    if (isTRUE(by_chain) && ".chain" %in% colnames(long_df)) {
      p <- p + ggplot2::geom_histogram(
        ggplot2::aes(
          y = ggplot2::after_stat(density),
          fill = factor(.data[[".chain"]])
        ),
        bins = as.integer(bins),
        alpha = 0.4,
        position = "identity"
      ) +
        ggplot2::labs(fill = "Chain")
    } else {
      p <- p + ggplot2::geom_histogram(
        ggplot2::aes(y = ggplot2::after_stat(density)),
        bins = as.integer(bins),
        fill = "grey70",
        color = "white"
      )
    }
  }

  if (isTRUE(show_interval)) {
    p <- p +
      ggplot2::geom_vline(
        data = summary_df,
        ggplot2::aes(xintercept = .data[["lower"]]),
        color = "firebrick",
        linetype = "dashed",
        inherit.aes = FALSE
      ) +
      ggplot2::geom_vline(
        data = summary_df,
        ggplot2::aes(xintercept = .data[["upper"]]),
        color = "firebrick",
        linetype = "dashed",
        inherit.aes = FALSE
      )
  }

  if (isTRUE(show_estimate)) {
    p <- p + ggplot2::geom_vline(
      data = summary_df,
      ggplot2::aes(xintercept = .data[["estimate"]]),
      color = "royalblue",
      inherit.aes = FALSE
    )
  }

  plot_name <- attr(samples, "name")
  plot_title <- "Posterior parameter distributions"
  if (!is.null(plot_name) && nzchar(plot_name)) {
    plot_title <- paste0(plot_title, " (", plot_name, ")")
  }

  p +
    ggplot2::facet_wrap(~parameter, scales = "free", ncol = facet_ncol) +
    ggplot2::labs(
      title = plot_title,
      subtitle = paste0(
        "Intervals: ",
        formatC(lower, format = "f", digits = 3),
        " to ",
        formatC(upper, format = "f", digits = 3)
      ),
      x = "Value",
      y = "Density"
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      strip.text = ggplot2::element_text(size = 12, face = "bold"),
      plot.title = ggplot2::element_text(size = 14, face = "bold")
    )
}

#' @rdname posterior_plot
#' @export
plot.ngme_sgld_ci <- function(x, ...) {
  posterior_plot(x, ...)
}



#' Plot the density of one or more stationary noise objects
#'
#' This function plots the probability density function for one or more stationary noise objects
#' (e.g., NIG, GAL, or normal noise). Multiple noise objects can be compared on the same plot.
#'
#' @param x An ngme_noise object (required).
#' @param ... Additional ngme_noise objects to plot, or plotting parameters such as \code{xlim}.
#'   Named arguments will be used as legend labels.
#'
#' @return A ggplot object showing the density curves for the provided noise objects.
#' @export
#'
#' @examples
#' plot(noise_nig(mu = 1, sigma = 2, nu = 1))
#' plot(n1 = noise_nig(mu = 0, sigma = 1, nu = 1), n2 = noise_nig(mu = 1, sigma = 1.5, nu = 0.5))
plot.ngme_noise <- function(x = NULL, ...) {
  # Get all arguments including the first one
  call_args <- as.list(match.call())[-1] # Remove function name
  all_args <- c(if (!is.null(x)) list(x), list(...))

  # Get names from the call
  call_names <- names(call_args)
  if (is.null(call_names)) {
    call_names <- rep("", length(call_args))
  }

  # Helper function to check if an object is a noise object
  is_noise_object <- function(obj) {
    is.list(obj) &&
      !is.null(obj$theta_mu) &&
      !is.null(obj$theta_sigma) &&
      !is.null(obj$noise_type)
  }

  # Initialize noise objects list
  noise_objects <- list()

  # Extract plotting parameters and noise objects
  plot_params <- c("xlim", "ylim", "main", "xlab", "ylab")
  xlim <- NULL
  unnamed_count <- 1

  for (i in seq_along(all_args)) {
    arg <- all_args[[i]]
    arg_name <- call_names[i]

    if (arg_name %in% plot_params) {
      # This is a plotting parameter
      if (arg_name == "xlim") xlim <- arg
    } else if (is_noise_object(arg)) {
      # This is a noise object
      if (is.null(arg_name) || arg_name == "") {
        # Unnamed argument
        noise_name <- paste0("noise_", unnamed_count)
        unnamed_count <- unnamed_count + 1
      } else {
        # Named argument
        noise_name <- arg_name
      }
      noise_objects[[noise_name]] <- arg
    }
  }

  # Check that we have at least one noise object
  if (length(noise_objects) == 0) {
    stop("No noise objects provided")
  }

  # Set default xlim if not provided
  if (is.null(xlim)) xlim <- c(-10, 10)

  xx <- seq(xlim[[1]], xlim[[2]], length = 400)

  # Define colors for multiple lines
  colors <- c("black", "red", "blue", "green", "purple", "orange", "brown", "pink", "gray", "cyan")

  # Initialize plot data
  plot_data <- data.frame()

  # Process each noise object
  for (i in seq_along(noise_objects)) {
    noise <- noise_objects[[i]]
    mu <- noise$theta_mu
    sigma <- exp(noise$theta_sigma)
    nu_lower_bound <- if (is.null(noise$nu_lower_bound)) 0 else noise$nu_lower_bound
    nu <- nu_lower_bound + exp(noise$theta_nu)

    stopifnot(
      "only implemented for stationary mu" =
        length(mu) == 1 || noise$noise_type == "normal"
    )
    stopifnot("only implemented for stationary sigma" = length(sigma) == 1)
    stopifnot(
      "only implemented for stationary nu" =
        length(nu) == 1 || noise$noise_type == "normal"
    )

    switch(noise$noise_type,
      "nig"     = dd <- dnig(xx, -mu, mu, nu, sigma),
      "gal"     = dd <- dgal(xx, -mu, mu, nu, sigma),
      "normal"  = dd <- dnorm(xx, sd = sigma),
      stop("Plot for this type is not implemented")
    )

    # Create data frame for this noise object
    noise_name <- names(noise_objects)[i]
    if (is.null(noise_name) || noise_name == "") {
      noise_name <- paste("noise", i)
    }

    temp_data <- data.frame(
      x = xx,
      y = dd,
      noise = noise_name,
      stringsAsFactors = FALSE
    )

    plot_data <- rbind(plot_data, temp_data)
  }


  # Create the plot
  gg <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(
      x = .data[["x"]],
      y = .data[["y"]],
      color = .data[["noise"]]
    )
  ) +
    ggplot2::geom_line() +
    ggplot2::labs(title = "Noise Density Plot") +
    ggplot2::theme_minimal()

  # Add color scale and legend handling
  if (length(noise_objects) > 1) {
    gg <- gg +
      ggplot2::scale_color_manual(values = colors[1:length(noise_objects)]) +
      ggplot2::labs(color = "Noise Objects")
  } else {
    gg <- gg + ggplot2::theme(legend.position = "none")
  }

  gg
}


compare_traceplot <- function(l1, l2) {
  l1 <- as.data.frame(l1)
  l2 <- as.data.frame(l2)

  ps <- list()
  n_plots <- length(l1)
  n_iter <- length(l1[[1]])

  for (i in seq_len(n_plots)) {
    c1 <- c2 <- NULL
    df <- data.frame(c1 = l1[[i]], c2 = l2[[i]], title = names(l1)[[i]])
    ps[[i]] <- ggplot(data = df) +
      geom_line(aes(x = 1:n_iter, y = c1), col = "1") +
      geom_line(aes(x = 1:n_iter, y = c2), col = "2") +
      labs(title = df$title) +
      xlab(NULL) +
      ylab(NULL)
  }

  do.call(gridExtra::grid.arrange, ps)
}


#' Compare noise objects using Kullback-Leibler divergence
#'
#' This function compares multiple noise objects by calculating the KLD
#' of each noise against the first noise object (reference).
#'
#' @param x first noise object (reference)
#' @param ... additional noise objects to compare, and optional parameters
#' @param xlim x-axis range for evaluation (default: c(-10, 10))
#' @param n_points number of evaluation points (default: 1000)
#'
#' @return named vector of KLD values
#' @export
#'
#' @examples
#' n1 <- noise_nig(mu = 0, sigma = 1, nu = 1)
#' n2 <- noise_nig(mu = 0.5, sigma = 1.2, nu = 0.8)
#' compare_noise_kld(n1, method2 = n2)
compare_noise_kld <- function(x = NULL, ..., xlim = c(-10, 10), n_points = 1000) {
  # Get all arguments including the first one
  call_args <- as.list(match.call())[-1] # Remove function name
  all_args <- c(if (!is.null(x)) list(x), list(...))

  # Get names from the call
  call_names <- names(call_args)
  if (is.null(call_names)) {
    call_names <- rep("", length(call_args))
  }

  # Helper function to check if an object is a noise object
  is_noise_object <- function(obj) {
    is.list(obj) &&
      !is.null(obj$theta_mu) &&
      !is.null(obj$theta_sigma) &&
      !is.null(obj$noise_type)
  }

  # Initialize noise objects list
  noise_objects <- list()

  # Extract parameters and noise objects
  param_names <- c("xlim", "n_points")
  unnamed_count <- 1

  for (i in seq_along(all_args)) {
    arg <- all_args[[i]]
    arg_name <- call_names[i]

    if (arg_name %in% param_names) {
      # This is a function parameter - already handled by function signature
      next
    } else if (is_noise_object(arg)) {
      # This is a noise object
      if (is.null(arg_name) || arg_name == "") {
        # Unnamed argument
        noise_name <- paste0("noise_", unnamed_count)
        unnamed_count <- unnamed_count + 1
      } else {
        # Named argument
        noise_name <- arg_name
      }
      noise_objects[[noise_name]] <- arg
    }
  }

  # Check that we have at least two noise objects
  if (length(noise_objects) < 2) {
    stop("Need at least two noise objects to compare")
  }

  # Generate evaluation points
  xx <- seq(xlim[[1]], xlim[[2]], length = n_points)

  # Calculate densities for all noise objects
  densities <- list()
  for (i in seq_along(noise_objects)) {
    noise <- noise_objects[[i]]
    mu <- noise$theta_mu
    sigma <- exp(noise$theta_sigma)
    nu_lower_bound <- if (is.null(noise$nu_lower_bound)) 0 else noise$nu_lower_bound
    nu <- nu_lower_bound + exp(noise$theta_nu)

    stopifnot(
      "only implemented for stationary mu" =
        length(mu) == 1 || noise$noise_type == "normal"
    )
    stopifnot("only implemented for stationary sigma" = length(sigma) == 1)
    stopifnot(
      "only implemented for stationary nu" =
        length(nu) == 1 || noise$noise_type == "normal"
    )

    switch(noise$noise_type,
      "nig"     = dd <- dnig(xx, -mu, mu, nu, sigma),
      "gal"     = dd <- dgal(xx, -mu, mu, nu, sigma),
      "normal"  = dd <- dnorm(xx, sd = sigma),
      stop("KLD comparison for this noise type is not implemented")
    )

    densities[[i]] <- dd
  }

  # Calculate KLD of each noise against the first one (reference)
  reference_density <- densities[[1]]
  kld_values <- numeric(length(noise_objects) - 1)
  names(kld_values) <- names(noise_objects)[-1]

  for (i in 2:length(noise_objects)) {
    comparison_density <- densities[[i]]

    # Add small epsilon to avoid log(0) and division by 0
    epsilon <- 1e-10
    p <- pmax(reference_density, epsilon)
    q <- pmax(comparison_density, epsilon)

    # Calculate KLD: sum(p * log(p/q)) * dx
    # Since we're using discrete points, we need to multiply by dx
    dx <- (xlim[2] - xlim[1]) / (n_points - 1)
    kld_values[i - 1] <- sum(p * log(p / q)) * dx
  }

  # Create results
  result <- list(
    kld_values = kld_values,
    reference = names(noise_objects)[1],
    n_comparisons = length(kld_values),
    closest = names(kld_values)[which.min(kld_values)]
  )

  class(result) <- "noise_kld_comparison"
  return(result)
}


#' Print method for noise_kld_comparison
#' @param x noise_kld_comparison object
#' @param ... additional arguments
#' @return Invisibly returns \code{x}, a \code{noise_kld_comparison} object. The
#'   method is called for its side effect of printing the reference noise object,
#'   the Kullback-Leibler divergence values for each comparison noise, and the
#'   closest comparison.
#' @export
print.noise_kld_comparison <- function(x, ...) {
  cat("Noise KLD Comparison\n")
  cat("===================\n")
  cat("Reference:", x$reference, "\n\n")
  cat("KLD values (lower is closer to reference):\n")

  # Sort by KLD value for better display
  sorted_kld <- sort(x$kld_values)
  for (i in seq_along(sorted_kld)) {
    cat(sprintf("  %s: %.6f", names(sorted_kld)[i], sorted_kld[i]))
    if (names(sorted_kld)[i] == x$closest) {
      cat(" <- CLOSEST")
    }
    cat("\n")
  }

  cat("\nClosest to reference:", x$closest, "\n")
  invisible(x)
}

test_that("trace transformations for stationary nu include nu_lower_bound", {
  noise <- noise_nig(
    theta_nu = 0,
    B_nu = matrix(1, 1, 1),
    nu_lower_bound = 0.7,
    fix_theta_mu = TRUE,
    fix_theta_sigma = TRUE
  )

  rep1 <- list(noise = noise, feff = numeric(0))
  attr(rep1, "block_traj") <- list(matrix(0, nrow = 1, ncol = 3))
  fake_ngme <- structure(list(replicates = list(rep1)), class = "ngme")

  tr <- get_trace_trajectories(fake_ngme, name = "data", apply_transform = TRUE)

  expect_equal(tr$parameter_names, "nu")
  expect_equal(as.numeric(tr$trajectories$nu[, 1]), rep(1.7, 3))
})

test_that("posterior_plot returns ggplot for SGLD sample data", {
  withr::local_seed(11)

  samples <- data.frame(
    .chain = rep(1:2, each = 40),
    .draw = rep(1:40, times = 2),
    .iter = rep(seq_len(40), times = 2),
    alpha = rnorm(80, mean = 0.5, sd = 0.2),
    beta = rnorm(80, mean = -1, sd = 0.4)
  )
  attr(samples, "name") <- "general"

  p <- posterior_plot(
    samples,
    parameters = c("alpha", "beta"),
    by_chain = TRUE
  )

  expect_s3_class(p, "ggplot")
  expect_true(inherits(plot.ngme_sgld_ci(ngme_sgld_ci(samples)), "ggplot"))
})

test_that("posterior_plot validates parameter selection and layout inputs", {
  samples <- data.frame(
    .chain = 1,
    .draw = 1:5,
    .iter = 1:5,
    alpha = rnorm(5)
  )

  expect_error(
    posterior_plot(samples, parameters = "missing"),
    "Unknown parameter"
  )

  expect_error(
    posterior_plot(samples, ncol = 0),
    "`ncol` must be a positive integer"
  )
})

#############################################################################################
#' @name gal
#' @title  The Generalized Asymmetric Laplace (GAL) Distribution
#' @aliases dgal rgal pgal qgal
#' @description Density, distribution function, quantile function and
#' random generation for the generalized asymmetric Laplace distribution
#'  with parameters \code{mu}, \code{sigma} and \code{nu}, \code{delta}.
#' @param x,q vector of quantiles.
#' @param p vector of probabilities.
#' @param n, number of observations.
#' @param delta A numeric value for the location parameter.
#' @param mu    A numeric value for the shift parameter.
#' @param nu    A numeric value for the shape parameter.
#' @param sigma A numeric value for the scaling parameter.
#' @param log,log.p logical; if \code{TRUE}, probabilities/densities \eqn{p} are
#' returned as \eqn{log(p)}.
#' @param lower.tail logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
#' otherwise, \eqn{P[X>x]}.
#' @param seed Seed for the random generation.
#' @return
#' dgal gives the density, pgal gives the distribution function,
#' qgal gives the quantile function, and rgal generates random deviates.
#'
#' Invalid arguments will result in return value NaN, with a warning.
#'
#' The length of the result is determined by \code{n} for rgal.
#' @details
#' The generalized asymmetric Laplace distribution has density given
#' by
#'  \deqn{f(x; p, a, b) =
#'  \frac{e^{\nu+\mu(x-\delta)/\sigma^2}\sqrt{\nu\mu^2/\sigma^2+\nu^2}}{\pi\sqrt{\nu\sigma^2+(x-\delta)^2}}
#'  K_1(\sqrt{(\nu\sigma^2+(x-\delta)^2)(\mu^2/\sigma^4+\nu/\sigma^2)}),}
#'  where \eqn{K_p} is modified Bessel function of the second kind of order \eqn{p},
#'  \eqn{x>0}, \eqn{\nu>0} and \eqn{\mu,\delta, \sigma\in\mathbb{R}}.
#'  See Barndorff-Nielsen (1977, 1978 and 1997) for further details.
#'
#' If the mixing variable \eqn{V} follows a Gamma distribution (same parameterization in R):
#' \deqn{V \sim \Gamma(h \nu, \nu),} then the poserior follows the GAL distribution (a special case of GIG distribution):
#' \deqn{
#' -\mu +\mu V + \sigma \sqrt{V} Z \sim GIG(h \nu - 0.5, 2 \nu + (\frac{\mu}{\sigma})^{2}, 0)
#' }
#' @references
#'  Barndorff-Nielsen, O. (1977) Exponentially decreasing distributions for the logarithm of particle size. Proceedings of the Royal Society of London.
#'
#'  Series A, Mathematical and Physical Sciences. The Royal Society. 353, 401–409. \doi{10.1098/rspa.1977.0041}
#'
#'  Barndorff-Nielsen, O. (1978) Hyperbolic Distributions and Distributions on Hyperbolae, Scandinavian Journal of Statistics. 5, 151–157.
#'
#' @seealso
#' \code{\link{dgig}}, \code{\link{dig}}, \code{\link{digam}}
#' @examples
#' rgal(100, delta = 0, mu = 5, sigma = 1, nu = 1)
#' pgal(0.4, delta = 0, mu = 5, sigma = 1, nu = 1)
#' qgal(0.8, delta = 0, mu = 5, sigma = 1, nu = 1)
#' plot(function(x){dgal(x, delta = 0, mu = 5, sigma = 1, nu = 1)}, main =
#' "generalized asymmetric Laplace density", ylab = "Probability density",
#' xlim = c(0,10))
#' @rdname gal
#' @export
dgal <- function(x, delta, mu, nu, sigma, log=FALSE){
  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lengths")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
      } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
      } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  n = length(x)

  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }

  # p_vec = nu_new - 0.5
  # a_vec = 2 * nu_new + (mu_new / sigma_new)^2
  # dgig(x, p_vec, a_vec, b = 1e-14, log)

  densgal <- sapply(1:n, function(i){
    l <- exp(nu_new[i] + mu_new[i] * (x[i] - delta_new[i]) /sigma_new[i]^2) *
      sqrt(nu_new[i] * mu_new[i]^2/sigma_new[i]^2 + nu_new[i]^2) /
      (pi * sqrt(nu_new[i] * sigma_new[i]^2 + (x[i] - delta_new[i])^2)) *
      besselK(sqrt((nu_new[i] * sigma_new[i]^2 + (x[i] - delta_new[i])^2) *
                     (mu_new[i]^2/sigma_new[i]^4 + nu_new[i]/sigma_new[i]^2)),
              1,FALSE)

    dens <- ifelse(log, log(l), l)
    return(dens)
  })

  return(densgal)
}

#' @rdname gal
#' @importFrom stats rgamma
#' @export
rgal <- function (n, delta, mu, nu, sigma, seed = 0)
{
  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lenghts")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
    } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
    } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }

  V = as.vector(stats::rgamma(n, nu_new, nu_new))
  return(delta_new + mu_new*V + sigma_new^2*sqrt(V)*stats::rnorm(n))
}

#' @rdname gal
#' @export
pgal <- function(q, delta, mu, nu, sigma, lower.tail = TRUE, log.p = FALSE){
  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lenghts")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
    } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
    } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  n = length(q)
  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }

  # prob_gal <- sapply(1:n, function(i){
  #   if(q[i]<=0){
  #     p_gal <- ifelse(lower.tail, 0, 1)
  #     p_gal <- ifelse(log.p, log(p_gal), p_gal)
  #     return(p_gal)
  #   } else{
  #     p_gal <- stats::integrate(dgal,lower = 0, upper = q[i],
  #                               delta=delta_new[i],mu=mu_new[i],
  #                               nu=nu_new[i],sigma=sigma_new[i])$value
  #     if(p_gal < 10^{-5} & q[i] > 200){
  #       p_gal = 1
  #     }
  #     p_gal <- ifelse(lower.tail, p_gal, 1-p_gal)
  #     p_gal <- ifelse(log.p, log(p_gal), p_gal)
  #     return(p_gal)
  #   }
  # })
  # return(prob_gal)
  p_vec = nu_new - 0.5
  a_vec = 2 * nu_new + (mu_new / sigma_new)^2

  pgig(q, p_vec, a_vec, b = 1e-14, lower.tail, log.p)
}

#' @rdname gal
#' @export
qgal <- function(p, delta, mu, nu, sigma, lower.tail = TRUE, log.p = FALSE){
  if(missing(delta)){
    stop('argument "delta" missing, with no default')
  }
  if(missing(mu)){
    stop('argument "mu" missing, with no default')
  }
  if(missing(nu)){
    stop('argument "nu" missing, with no default')
  }
  if(missing(sigma)){
    stop('argument "sigma" missing, with no default')
  }
  if (length(delta) != length(mu)){
    if(length(delta)==1){
      delta <- rep(delta, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(delta))
    } else{
      stop("delta and mu are vectors of different lenghts")
    }
  }
  if (length(mu) != length(nu)){
    if(length(nu)==1){
      nu <- rep(nu, length(mu))
    } else if(length(mu)==1){
      mu <- rep(mu, length(nu))
      if(length(delta)==1){
        delta <- rep(delta, length(nu))
      }
    } else{
      stop("mu and nu are vectors of different lengths")
    }
  }
  if (length(nu) != length(sigma)){
    if(length(sigma)==1){
      sigma <- rep(sigma, length(nu))
    } else if(length(nu)==1){
      nu <- rep(nu, length(sigma))
      if(length(mu)==1){
        mu <- rep(mu,length(sigma))
      }
      if(length(delta)==1){
        delta <- rep(delta, length(sigma))
      }
    } else{
      stop("nu and sigma are vectors of different lengths")
    }
  }
  if (min(nu) < 0)
    stop("vector nu must be  positive")
  n = length(p)
  if(n < length(delta)){
    delta_new = delta[1:n]
    mu_new = mu[1:n]
    nu_new = nu[1:n]
    sigma_new = sigma[1:n]
  } else if(n>length(delta)){
    quot <- n%/%length(delta)
    rem <- n%%length(delta)
    delta_new <- rep(delta,quot)
    if(rem > 0){
      delta_new <- c(delta_new, delta[1:rem])
    }
    mu_new <- rep(mu,quot)
    if(rem > 0){
      mu_new <- c(mu_new, mu[1:rem])
    }
    nu_new <- rep(nu,quot)
    if(rem > 0){
      nu_new  <- c(nu_new, nu[1:rem])
    }
    sigma_new <- rep(sigma,quot)
    if(rem > 0){
      sigma_new <- c(sigma_new, sigma[1:rem])
    }
  } else{
    delta_new = delta
    mu_new = mu
    nu_new = nu
    sigma_new = sigma
  }

  # quant_gal <- sapply(1:n, function(i){
  #   if(nu_new[i]<1){
  #     up_bd <- max(1/nu_new[i]^2, 1000)
  #   } else {
  #     up_bd <- 1000
  #   }
  #   if(p[i] < 0 | p[i] > 1){
  #     warn_qgal <- TRUE
  #     return(NaN)} else{
  #       while(pgal(up_bd, delta=delta_new[i],mu=mu_new[i],
  #                  nu=nu_new[i],sigma=sigma_new[i],
  #                  lower.tail=lower.tail) < p[i]){
  #         up_bd = up_bd + 1000
  #       }
  #       q_gal <- stats::uniroot(function(y) {
  #         pgal(y, delta=delta_new[i],mu=mu_new[i],
  #              nu=nu_new[i],sigma=sigma_new[i],
  #              lower.tail=lower.tail) - p[i]
  #       },lower = 0, upper = up_bd)$root
  #       q_gal <- ifelse(log.p, log(q_gal), q_gal)
  #       return(q_gal)
  #     }
  # }
  # )
  # if(any(is.nan(quant_gal))){
  #   warning("NaNs produced", call. = TRUE, domain = "R")
  # }
  # return(quant_gal)

  p_vec = nu_new - 0.5
  a_vec = 2 * nu_new + (mu_new / sigma_new)^2

  qgig(p, p_vec, a_vec, b = 1e-14, lower.tail, log.p)
}

#' Simulate latent process with noise
#'
#' @param object ngme model specified by f() function
#' @param nsim number of simulations
#' @param seed seed
#' @param ... ignored
#'
#' @return a realization of latent model
#' @export
#'
#' @examples
#' simulate(f(1:10, model = ar1(rho = 0.4), noise = noise_nig()))
#' simulate(f(rnorm(10), model = rw1(), noise = noise_normal()))
#' simulate(f(1:10, model = ar1(rho = 0.4), noise = noise_t(nu = 5)))
simulate.ngme_model <- function(
    object,
    nsim = 1,
    seed = NULL,
    ...) {
  if (is.null(seed)) seed <- Sys.time()
  model <- object
  noise <- model$noise

  sims <- list()
  V_sim <- list()
  W_sim <- list()
  for (nn in 1:nsim) {
    seed <- seed + nn
    # simulate noise
    h <- model$operator$h
    mu <- as.numeric(noise$B_mu %*% noise$theta_mu)
    sigma <- as.numeric(exp(noise$B_sigma %*% noise$theta_sigma))
    nu <- as.numeric(noise$nu_lower_bound + exp(noise$B_nu %*% noise$theta_nu))
    n <- length(mu)

    if (length(noise$noise_type) == 2) {
      # bivariate noise
      e1 <- simulate_noise(
        noise$noise_type[[1]],
        head(h, n / 2), head(mu, n / 2), head(sigma, n / 2), head(nu, n / 2),
        seed + 1000, noise$single_V
      )
      if (noise$share_V) {
        e2 <- e1
      } else {
        e2 <- simulate_noise(
          noise$noise_type[[2]],
          tail(h, n / 2), tail(mu, n / 2), tail(sigma, n / 2), tail(nu, n / 2),
          seed + 2000, noise$single_V
        )
      }
      e <- c(e1, e2)
      attr(e, "V") <- c(attr(e1, "V"), attr(e2, "V"))
    } else {
      e <- simulate_noise(noise$noise_type, h, mu, sigma, nu, seed + 3, noise$single_V)
    }

    W <- as.numeric(Matrix::solve(model$operator$K, model$operator$Z %*% e))

    # Need to map W and V to data by A matrix!!!
    if (is.null(model$A)) {
      # model does not contain A matrix, e.g., Matern graph model!
      sims[[paste0("sim_", nn)]] <- W
      V_sim[[paste0("sim_", nn)]] <- attr(e, "V")
      W_sim[[paste0("sim_", nn)]] <- W
    } else {
      sims[[paste0("sim_", nn)]] <- as.numeric(model$A %*% W)
      V_sim[[paste0("sim_", nn)]] <- attr(e, "V")
      W_sim[[paste0("sim_", nn)]] <- W
    }
  }

  structure(
    as.data.frame(sims),
    W_sim = as.data.frame(W_sim),
    V_sim = as.data.frame(V_sim),
    noise = noise
  )
}

simulate_noise <- function(
    noise_type, h_vec, mu_vec, sigma_vec, nu_vec, seed, single_V) {
  set.seed(seed)
  stopifnot(
    length(mu_vec) == length(sigma_vec),
    length(mu_vec) == length(nu_vec),
    length(mu_vec) == length(h_vec)
  )
  n <- length(mu_vec)

  if (noise_type == "normal") {
    V <- h_vec
  } else if (noise_type == "nig" || noise_type == "normal_nig") {
    V <- if (single_V) {
      h_vec * ngme2::rig(1, a = nu_vec[1], b = nu_vec[1], seed = seed)
    } else {
      ngme2::rig(n, a = nu_vec, b = nu_vec * (h_vec)^2, seed = seed)
    }
  } else if (noise_type == "gal") {
    V <- if (single_V) {
      h_vec * rgamma(1, nu_vec[1], nu_vec[1])
    } else {
      rgamma(n, shape = h_vec * nu_vec, rate = nu_vec)
    }
  } else if (noise_type == "t" || noise_type == "skew_t") {
    # For t-distribution, auxiliary variable V ~ InverseGamma(nu/2, nu/2)
    V <- if (single_V) {
      h_vec * ngme2::rigam(1, a = nu_vec[1] / 2, b = nu_vec[1] / 2)
    } else {
      ngme2::rigam(n, a = nu_vec / 2, b = nu_vec / 2)
    }
  } else {
    stop("unknown noise to simulate")
  }

  e <- mu_vec * (V - h_vec) + sigma_vec * sqrt(V) * rnorm(n)
  attr(e, "V") <- V
  e
}

#' Simulate ngme noise object
#'
#' @param object  ngme noise object
#' @param h should be of same length as nsim
#' @param seed seed
#' @param nsim number of simulations
#' @param ... ignored
#'
#' @return data.frame (each col is a realization)
#' @export
simulate.ngme_noise <- function(
    object,
    nsim = 1,
    seed = NULL,
    h = NULL,
    ...) {
  n_noise <- max(nrow(object$B_mu), nrow(object$B_sigma), nrow(object$B_nu))
  if (is.null(seed)) seed <- Sys.time()
  if (is.null(h)) h <- rep(1, n_noise)
  if (length(h) > n_noise) n_noise <- length(h)
  stopifnot(length(h) == n_noise)

  sims <- list()
  for (nn in 1:nsim) {
    seed <- seed + nn
    sims[[paste0("sim_", nn)]] <- with(object, {
      res <- numeric(n_noise)
      mu_vec <- as.numeric(B_mu %*% theta_mu)
      sigma_vec <- as.numeric(exp(B_sigma %*% theta_sigma))
      nu_vec <- as.numeric(nu_lower_bound + exp(B_nu %*% theta_nu))
      if (length(mu_vec) == 1) mu_vec <- rep(mu_vec, n_noise)
      if (length(sigma_vec) == 1) sigma_vec <- rep(sigma_vec, n_noise)
      if (length(nu_vec) == 1) nu_vec <- rep(nu_vec, n_noise)

      if (!corr_measurement) {
        res <- simulate_noise(noise_type, h, mu_vec, sigma_vec, nu_vec, seed, single_V)
      } else {
        # simulate correlated noise
        stopifnot(
          is.numeric(object$index_corr),
          "index_corr must be non-decreasing" = all(diff(object$index_corr) >= 0)
        )
        i <- 1
        while (i <= n_noise) {
          if (i == n_noise || index_corr[[i]] != index_corr[[i + 1]]) {
            res[i] <- simulate_noise(noise_type, h[i], mu_vec[i], sigma_vec[i], nu_vec[i], seed + i, single_V)
            i <- i + 1
          } else {
            # simulate a pair correlated noise
            if (noise_type == "normal") {
              set.seed(seed + i)
              cov_mat <- matrix(
                c(
                  sigma_vec[i]^2, sigma_vec[i] * sigma_vec[i + 1] * rho,
                  sigma_vec[i] * sigma_vec[i + 1] * rho, sigma_vec[i + 1]^2
                ),
                nrow = 2
              )
              res[i:(i + 1)] <- mvtnorm::rmvnorm(1, rep(0, 2), cov_mat)
              i <- i + 2
            } else {
              stop("Simulation of correlated NIG and GAL is not implemented yet, return only 0 for now.")
            }
          }
        }
      }
      res
    })
  }
  structure(
    as.data.frame(sims),
    V_sim = as.data.frame(lapply(sims, function(x) attr(x, "V")))
  )
}


#' Simulate from a ngme object (possibly with replicates)
#'
#' @param object  ngme object
#' @param nsim number of simulations
#' @param seed seed
#' @param ... optional arguments. Supported names are \code{posterior}
#'   (whether to simulate from posterior sampling of latent fields) and
#'   \code{m_noise} (whether to add the measurement noise).
#'
#' @return a realization of ngme object
#' @export
simulate.ngme <- function(
    object,
    nsim = 1,
    seed = NULL,
    ...) {
  dots <- list(...)
  posterior <- if (!is.null(dots$posterior)) dots$posterior else TRUE
  m_noise <- if (!is.null(dots$m_noise)) dots$m_noise else TRUE

  if (is.null(seed)) seed <- Sys.time()
  attr <- attributes(object)
  sims <- list()

  for (nn in 1:nsim) {
    seed <- seed + nn
    Y <- numeric(attr$fit$n_data)
    # simulate from different replicates
    replicate <- attr$fit$replicate
    for (repl in levels(replicate)) {
      repl_idx <- replicate == repl
      this_repl <- object$replicates[[repl]]
      Y[repl_idx] <- simulate_1rep(this_repl, posterior, seed)

      sim_noise <- simulate(this_repl$noise, nsim = 1, seed = seed)[[1]]
      # add measurement noise
      if (m_noise) Y[repl_idx] <- Y[repl_idx] + sim_noise
    }

    # If Y is correlated, we need to re-order the simulated Y to match the original Y
    corr <- object$replicates[[1]]$noise$corr_measurement
    if (corr) {
      Y <- Y[order(object$repls_ngme)]
      data_idx <- c()
      for (repl in levels(replicate)) {
        data_idx <- c(data_idx, object$replicates[[repl]]$data_idx)
      }
      Y <- Y[order(data_idx)]
    }

    sims[[paste0("sim_", nn)]] <- Y
  }

  as.data.frame(sims)
}

# simulate from one replicate
simulate_1rep <- function(ngme_1rep, posterior = TRUE, seed = NULL) {
  # extract A and cbind!
  As <- list()
  for (i in seq_along(ngme_1rep$models)) {
    As[[i]] <- ngme_1rep$models[[i]]$A
  }
  A_block <- Reduce(cbind, x = As)

  if (is.null(seed)) seed <- Sys.time()
  Ws <- sampling_cpp(ngme_1rep, n = 1, n_burnin = 1, posterior = posterior, seed = seed)[["W"]][[1]]

  # return A W + X beta
  if (!is.null(A_block)) {
    as.numeric(A_block %*% Ws + ngme_1rep$X %*% ngme_1rep$feff)
  } else {
    # No latent fields
    as.numeric(ngme_1rep$X %*% ngme_1rep$feff)
  }
}

coerce_sub_model <- function(sub_model, mesh) {
  if (inherits(sub_model, "ngme_operator")) {
    return(sub_model)
  }

  if (inherits(sub_model, "ngme_operator_def")) {
    model_name <- sub_model$model
    args <- sub_model$args
    args$mesh <- mesh
    return(do.call(model_name, args))
  }

  stop("Each sub-model must be an ngme_operator or ngme_operator_def.")
}

#' Ngme bivariate model specification
#'
#' Giving 2 sub_models, build a correlated bivariate operator based on K = D(theta, rho) %*% diag(c1*K_1, c2*K_2)
#' \deqn{D(\theta, \rho) = \begin{pmatrix}
#'   \cos(\theta) + \rho \sin(\theta) & -\sin(\theta) \sqrt{1+\rho^2} \\
#'   \sin(\theta) - \rho \cos(\theta) & \cos(\theta) \sqrt{1+\rho^2}
#' \end{pmatrix}}
#'
#' @param mesh the mesh where model is defined
#' @param sub_models a list of sub_models (total 2 sub_models)
#' @param theta the rotation parameter (starting point should be from -pi/4 to pi/4)
#' @param rho the parameter related to correlation
#' @param c1 scaling factor for the first sub-model (default: 1)
#' @param c2 scaling factor for the second sub-model (default: 1)
#' @param share_param TRUE if share the same parameter for 2 sub_models (of same type)
#' @param fix_theta TRUE if fix the theta of bv model
#' @param use_c_param TRUE to estimate c1 and c2 as parameters, FALSE to fix them at specified values (default: FALSE)
#'
#' @details
#' The bivariate model combines two sub-models with a rotation matrix D and optional scaling factors c1 and c2.
#' When \code{use_c_param = TRUE}, c1 and c2 are estimated as parameters (on log scale).
#' When \code{use_c_param = FALSE} (default), c1 and c2 are fixed at their specified values (both default to 1).
#'
#' @return a ngme_operator of bivariate model
#' @export
bv <- function(
    mesh,
    sub_models,
    theta = 0,
    rho = 0,
    c1 = 1,
    c2 = 1,
    share_param = FALSE,
    fix_theta = FALSE,
    use_c_param = FALSE) {
  if (is.list(mesh) && !inherits(mesh, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))) {
    return(structure(list(model = "bv", args = as.list(environment())), class = "ngme_operator_def"))
  }
  mesh <- ngme_build_mesh(mesh)
  model_names <- sort(names(sub_models))

  stopifnot(
    "Please provide names for sub_models" = !is.null(model_names),
    "Length of sub_models should be 2" = length(sub_models) == 2
  )

  # Inject mesh if sub_models are def
  first <- coerce_sub_model(sub_models[[model_names[1]]], mesh)
  second <- coerce_sub_model(sub_models[[model_names[2]]], mesh)

  stopifnot(
    "the theta should be in (-pi/2, pi/2)" = theta >= -pi / 2 & theta <= pi / 2
  )
  eta <- tan(theta)

  theta_K <- c()
  if (!fix_theta) theta_K <- c(theta_K, eta)
  theta_K <- c(theta_K, rho)
  if (use_c_param) theta_K <- c(theta_K, log(c1), log(c2))
  theta_K <- c(theta_K, first$theta_K, second$theta_K)

  # pass the theta_K to first and second
  n_bv <- (if (!fix_theta) 1 else 0) + 1 + (if (use_c_param) 2 else 0)
  first$theta_K <- theta_K[(n_bv + 1):(n_bv + first$n_theta_K)]
  second$theta_K <- theta_K[(n_bv + first$n_theta_K + 1):length(theta_K)]

  D <- build_D(theta, rho)
  bigD <- kronecker(D, Matrix::Diagonal(nrow(first$K)))

  # the range of theta is (-pi/4, pi/4), we expand it to (-pi/2, pi/2)
  # so that the maximum can be reached in either direction
  # in the end, we will NOT transform it back to (-pi/4, pi/4)
  eta_to_theta <- function(eta) atan(eta)

  update_K <- function(theta_K) {
    idx <- 1
    if (!fix_theta) {
      eta <- theta_K[idx]
      theta <- atan(eta)
      idx <- idx + 1
    } else {
      theta <- 0 # fixed theta
    }
    rho <- theta_K[idx]
    idx <- idx + 1

    c1 <- 1
    c2 <- 1
    if (use_c_param) {
      c1 <- exp(theta_K[idx])
      c2 <- exp(theta_K[idx + 1])
      idx <- idx + 2
    }

    D <- build_D(theta, rho)
    bigD <- kronecker(D, Matrix::Diagonal(nrow(first$K)))

    theta_K1 <- theta_K[idx:(idx + first$n_theta_K - 1)]
    idx <- idx + first$n_theta_K
    theta_K2 <- theta_K[idx:length(theta_K)]

    first$K <- first$update_K(theta_K1)
    second$K <- second$update_K(theta_K2)
    bigD %*% Matrix::bdiag(c1 * first$K, c2 * second$K)
  }

  ngme_operator(
    mesh = mesh,
    model = "bv",
    first = first,
    second = second,
    theta_K = theta_K,
    update_K = update_K,
    K = update_K(theta_K),
    h = c(first$h, second$h),
    symmetric = FALSE,
    zero_trace = FALSE,
    model_names = model_names,
    share_param = share_param,
    fix_theta = fix_theta,
    bv_theta = theta,
    use_c_param = use_c_param,
    param_name = c(
      if (!fix_theta) "theta" else NULL,
      "rho",
      if (use_c_param) c("c1", "c2") else NULL,
      if (!is.null(first$param_name)) paste(first$param_name, "(1st)") else NULL,
      if (!is.null(second$param_name)) paste(second$param_name, "(2nd)") else NULL
    ),
    param_trans = c(
      if (!fix_theta) list(eta_to_theta) else NULL,
      list(identity),
      if (use_c_param) list(exp, exp) else NULL,
      first$param_trans,
      second$param_trans
    )
  )
}


#' ngme tensor-product model specification
#'
#' Given 2 operator (first and second), build a tensor-product operator based on K = K_first x K_second (here x is Kronecker product)
#'
#' @param first ngme_operator, left side of kronecker model (usually a temporal or iid model)
#' @param second ngme_operator, right side of kronecker model (ususally a temporal or spatial model)
#'
#' @return a list of specification of model
#' @export
tp <- function(
    first,
    second) {
  stopifnot(
    "Please provide ngme_operator for first" = inherits(first, "ngme_operator"),
    "Please provide ngme_operator for second" = inherits(second, "ngme_operator")
  )

  theta_K <- c(first$theta_K, second$theta_K)
  stopifnot(length(theta_K) == first$n_theta_K + second$n_theta_K)

  update_K <- function(theta_K) {
    first$K %x% second$K
  }
  ngme_operator(
    mesh = list(first$mesh, second$mesh),
    model = "tp",
    first = first,
    second = second,
    theta_K = theta_K,
    update_K = update_K,
    K = first$K %x% second$K,
    h = first$h %x% second$h,
    symmetric = first$symmetric & second$symmetric,
    zero_trace = first$zero_trace & second$zero_trace,
    param_name = c(first$param_name, second$param_name),
    param_trans = c(first$param_trans, second$param_trans)
  )
}


#' Ngme bivariate model with Matern sub_models
#'
#' Giving 2 sub_models, build a correlated bivaraite operator based on K = D(theta, eta) %*% diag(K_1, K_2)
#' \deqn{D(\theta, \rho) = \begin{pmatrix}
#'   \cos(\theta) + \rho \sin(\theta) & -\sin(\theta) \sqrt{1+\rho^2} \\
#'   \sin(\theta) - \rho \cos(\theta) & \cos(\theta) \sqrt{1+\rho^2}
#' \end{pmatrix}}
#'
#' @param mesh the mesh where model is defined
#' @param sub_models a list of sub_models (should be two matern models)
#' @param theta the parameter related to rotation
#' @param rho the parameter related to correlation
#' @param sd1 scaling parameter for the first sub-model
#' @param sd2 scaling parameter for the second sub-model
#' @param group group vector, can be inherited from ngme() function
#' @param share_param TRUE if share the same parameter for 2 sub_models (of same type)
#' @param fix_theta TRUE if the rotation parameter is fixed
#' @param ... ignore
#'
#' @return a list of specification of model
#' @export
bv_matern <- function(
    mesh,
    sub_models,
    theta = 0,
    rho = 0,
    sd1 = 1,
    sd2 = 1,
    group = NULL,
    share_param = FALSE,
    fix_theta = FALSE,
    ...) {
  if (is.list(mesh) && !inherits(mesh, c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))) {
    return(structure(list(model = "bv_matern", args = as.list(environment())), class = "ngme_operator_def"))
  }
  mesh <- ngme_build_mesh(mesh)
  d <- switch(mesh$manifold,
    "R1" = 1,
    "R2" = 2,
    stop("manifold not supported")
  )

  # sort the sub_models
  model_names <- sort(names(sub_models))

  # read group argument from parent frame
  if (is.null(group) &&
    exists("group", parent.frame())) {
    group <- get("group", parent.frame())
  }

  stopifnot(
    "theta should be within (-pi/2, pi/2)" = theta > -pi / 2 & theta < pi / 2,
    "Please provide names for sub_models" = !is.null(model_names),
    "Please provide group argument to indicate different fields" = !is.null(group),
    "Length of sub_models should be 2" = length(sub_models) == 2,
    "Name of sub_models should be in group" = all(model_names %in% levels(as.factor(group))),
    "sd1 and sd2 should be positive" = sd1 > 0 & sd2 > 0
  )
  group <- factor(group, levels = model_names)

  # Inject mesh if sub_models are def
  first <- coerce_sub_model(sub_models[[model_names[1]]], mesh)
  second <- coerce_sub_model(sub_models[[model_names[2]]], mesh)

  eta <- tan(theta)
  if (!fix_theta) {
    theta_K <- c(
      eta, rho, log(sd1), log(sd2),
      first$theta_K, second$theta_K
    )
  } else {
    theta_K <- c(
      rho, log(sd1), log(sd2),
      first$theta_K, second$theta_K
    )
  }

  update_K <- function(theta_K) {
    if (!fix_theta) {
      theta <- atan(theta_K[1])
      rho <- theta_K[2]
      sd1 <- exp(theta_K[3])
      sd2 <- exp(theta_K[4])
      theta_K1 <- theta_K[5:(4 + first$n_theta_K)]
      theta_K2 <- theta_K[(5 + first$n_theta_K):length(theta_K)]
    } else {
      rho <- theta_K[1]
      sd1 <- exp(theta_K[2])
      sd2 <- exp(theta_K[3])
      theta_K1 <- theta_K[4:(3 + first$n_theta_K)]
      theta_K2 <- theta_K[(4 + first$n_theta_K):length(theta_K)]
    }

    alpha1 <- first$alpha
    alpha2 <- second$alpha
    kappa1 <- exp(first$theta_K)
    kappa2 <- exp(second$theta_K)

    nu1 <- alpha1 - d / 2
    nu2 <- alpha2 - d / 2
    c1 <- sqrt(gamma(nu1) / (gamma(alpha1))) / (kappa1^(nu1) * (4 * pi)^(d / 4) * sd1)
    c2 <- sqrt(gamma(nu2) / (gamma(alpha2))) / (kappa2^(nu2) * (4 * pi)^(d / 4) * sd2)

    D <- build_D(theta, rho)
    bigD <- kronecker(D, Matrix::Diagonal(nrow(first$K)))
    first$K <- first$update_K(theta_K1)
    second$K <- second$update_K(theta_K2)
    bigD %*% Matrix::bdiag(c1 * first$K, c2 * second$K)
  }

  eta_to_theta <- function(eta) atan(eta)
  ngme_operator(
    mesh = mesh,
    model = "bv_matern",
    first = first,
    second = second,
    theta_K = theta_K,
    update_K = update_K,
    K = update_K(theta_K),
    h = c(first$h, second$h),
    dim = d,
    symmetric = FALSE,
    zero_trace = FALSE,
    model_names = model_names,
    share_param = share_param,
    fix_theta = fix_theta,
    bv_theta = theta,
    param_name = c(
      if (!fix_theta) "theta" else NULL, "rho", "sd1", "sd2",
      if (!is.null(first$param_name)) paste(first$param_name, "(1st)") else NULL,
      if (!is.null(second$param_name)) paste(second$param_name, "(2nd)") else NULL
    ),
    param_trans = c(if (!fix_theta) eta_to_theta else NULL, identity, exp, exp, first$param_trans, second$param_trans)
  )
}





#' Ngme space-time non-separable model specification
#'
#' Implements a non-separable space-time model based on the advection-diffusion SPDE with first-order derivative in time.
#' The model combines temporal and spatial components through a finite difference method (implicit Euler) for temporal discretization
#' and finite element method (continuous Galerkin) for spatial discretization. When advection dominates diffusion,
#' the "Streamline Diffusion" stabilization technique is applied.
#'
#' The model is particularly useful for large space-time datasets in environmental science, offering computationally
#' efficient methods for parameter estimation, kriging prediction, and conditional simulations.
#'
#' For details, see \doi{10.1016/j.spasta.2024.100847}
#'
#' @param mesh A list of two objects:
#'   \itemize{
#'     \item mesh_t - The temporal mesh
#'     \item mesh_s - The spatial mesh
#'   }
#' @param alpha 2 or 4, SPDE smoothness parameter.
#' @param cc Parameter c in the SPDE.
#' @param kappa Kappa parameter from Matern SPDE.
#' @param lambda The spatial damping parameter.
#' @param fix_gamma TRUE if fix gamma (advection term), FALSE if estimate gamma.
#' @param theta_gamma_x The x component of the advection term: \code{gamma_x = B_gamma_x \%*\% theta_gamma_x}.
#' @param theta_gamma_y The y component of the advection term: \code{gamma_y = B_gamma_y \%*\% theta_gamma_y}.
#' @param shared_theta_gamma TRUE if share the same theta_gamma for all time nodes. (theta_gamma_x and theta_gamma_y will be the same)
#' @param B_gamma_x The design matrix for the x component of the advection term.
#' @param B_gamma_y The design matrix for the y component of the advection term.
#' @param B_gamma_x_list A list of design matrices for the x component of the advection term on every time node, length(B_gamma_x_list) == nt-1.
#' @param B_gamma_y_list A list of design matrices for the y component of the advection term on every time node, length(B_gamma_y_list) == nt-1.
#' @param stabilization TRUE if using a stabilization term (for implicit Euler).
#'
#' @return ngme_operator object.
#' @export
spacetime <- function(
    mesh,
    lambda = 1, # fixed
    alpha = 2, # alpha = 2, 4, fixed
    cc = 1,
    kappa = 1,
    # advection term
    fix_gamma = FALSE,
    theta_gamma_x = 0,
    theta_gamma_y = 0,
    shared_theta_gamma = FALSE,
    # theta_gamma_t = 0,
    B_gamma_x = matrix(1, nrow = mesh[[2]]$n, ncol = 1),
    B_gamma_y = matrix(1, nrow = mesh[[2]]$n, ncol = 1),
    # or, a list of B_gamma_x, B_gamma_y on every time node.
    B_gamma_x_list = NULL,
    B_gamma_y_list = NULL,
    # B_gamma_t = matrix(1, nrow = mesh[[1]]$n, ncol = 1),
    stabilization = TRUE) {
  if (is.list(mesh) && length(mesh) > 0 && is.list(mesh[[1]]) && !inherits(mesh[[1]], c("inla.mesh.1d", "inla.mesh", "fm_mesh_1d", "fm_mesh_2d", "metric_graph"))) {
    return(structure(list(model = "spacetime", args = as.list(environment())), class = "ngme_operator_def"))
  }
  method <- "euler" # for now only support implicit euler
  if (theta_gamma_x == 0 && theta_gamma_y == 0 && fix_gamma) {
    stabilization <- FALSE
  }

  mesh_t <- mesh[[1]]
  nt <- mesh_t$n
  mesh_s <- mesh[[2]]
  ns <- mesh_s$n
  stopifnot(
    "Please provide mesh as a list of length 2" = length(mesh) == 2,
    "alpha should be 2 or 4" = alpha == 2 || alpha == 4,
    "method should be galerkin or euler" = method %in% c("galerkin", "euler"),
    "First mesh should be 1d" = fmesher::fm_manifold_dim(mesh_t) == 1,
    "Second mesh should be 2d" =
      fmesher::fm_manifold_dim(mesh_s) == 2,
    "require package rSPDE" = requireNamespace("rSPDE", quietly = TRUE)
  )

  # Additional validation for shared_theta_gamma
  if (shared_theta_gamma && !fix_gamma) {
    if (!all(theta_gamma_x == theta_gamma_y)) {
      warning("When shared_theta_gamma = TRUE, theta_gamma_x and theta_gamma_y should be the same. Using theta_gamma_x.")
    }
    # Check that B_gamma_x and B_gamma_y have the same number of columns when shared
    if (ncol(B_gamma_x) != ncol(B_gamma_y)) {
      stop("When shared_theta_gamma = TRUE, B_gamma_x and B_gamma_y must have the same number of columns.")
    }
  }

  if (!is.null(B_gamma_x_list) && !is.null(B_gamma_y_list)) {
    stopifnot(
      "B_gamma_x_list and B_gamma_y_list should have the same length" =
        length(B_gamma_x_list) == length(B_gamma_y_list),
      "The number of time nodes should be the same as the length of B_gamma_x_list and B_gamma_y_list" =
        nt - 1 == length(B_gamma_x_list),
      "The ncol of items in B_gamma_x_list should be the same as the length of theta_gamma_x" =
        all(sapply(B_gamma_x_list, ncol) == length(theta_gamma_x)),
      "The ncol of items in B_gamma_y_list should be the same as the length of theta_gamma_y" =
        all(sapply(B_gamma_y_list, ncol) == length(theta_gamma_y))
    )

    # Additional validation for shared_theta_gamma with lists
    if (shared_theta_gamma && !fix_gamma) {
      # Check that all matrices in B_gamma_x_list and B_gamma_y_list have the same number of columns
      if (!all(sapply(B_gamma_x_list, ncol) == sapply(B_gamma_y_list, ncol))) {
        stop("When shared_theta_gamma = TRUE, all matrices in B_gamma_x_list and B_gamma_y_list must have the same number of columns.")
      }
    }
  } else {
    # Use B_gamma_x and B_gamma_y
    stopifnot(
      "B_gamma_x and B_gamma_y should have the same number of rows as the spatial mesh" =
        nrow(B_gamma_x) == ns && nrow(B_gamma_y) == ns,
      "theta_gamma_x should be of length ncol(B_gamma_x)" =
        length(theta_gamma_x) == ncol(B_gamma_x),
      "theta_gamma_y should be of length ncol(B_gamma_y)" =
        length(theta_gamma_y) == ncol(B_gamma_y)
    )

    # turn it into a list
    B_gamma_x_list <- lapply(1:(nt - 1), function(i) B_gamma_x)
    B_gamma_y_list <- lapply(1:(nt - 1), function(i) B_gamma_y)
  }

  ##### temporal FEM #####
  fem_t <- fmesher::fm_fem(mesh_t, order = 2)
  Ct <- fem_t$c0
  Gt <- fem_t$g1

  # Build Bt
  Bt <- Matrix::bandSparse(
    n = nt, m = nt, k = c(-1, 0, 1),
    diagonals = cbind(rep(0.5, nt), rep(0, nt), rep(-0.5, nt))
  )
  Bt[1, 1] <- -0.5
  Bt[nt, nt] <- 0.5

  # Bt
  # [1,] -0.5 -0.5  .    .    .
  # [2,]  0.5  .   -0.5  .    .
  # [3,]  .    0.5  .   -0.5  .
  # [4,]  .    .    0.5  .   -0.5
  # [5,]  .    .    .    0.5  0.5


  ##### space FEM #####
  fem_s <- fmesher::fm_fem(mesh_s, order = alpha)
  Cs <- fem_s$c0
  Gs <- fem_s$g1
  cs_diag <- Matrix::diag(Cs)

  # compute h
  if (method == "galerkin") {
    h <- Matrix::diag(Ct) %x% Matrix::diag(Cs)
  } else {
    # implicit euler
    dt <- c(1, diff(mesh_t$loc))
    h <- dt %x% Matrix::diag(Cs) / cc
  }

  FV <- mesh_s$graph$tv
  P <- sf::st_coordinates(fmesher::fm_vertices(mesh_s))[, 1:2]
  fem2d <- rSPDE::rSPDE.fem2d(FV = FV, P = P)
  Bx <- fem2d$Bx
  By <- fem2d$By
  Hxx <- fem2d$Hxx
  Hyy <- fem2d$Hyy
  Hxy <- fem2d$Hxy
  Hyx <- fem2d$Hyx

  theta_K <- if (fix_gamma) {
    c(log(cc), log(kappa))
  } else if (shared_theta_gamma) {
    c(log(cc), log(kappa), theta_gamma_x)
  } else {
    c(log(cc), log(kappa), theta_gamma_x, theta_gamma_y)
  }
  n_theta_gamma_x <- if (fix_gamma) 0 else ncol(B_gamma_x)
  n_theta_gamma_y <- if (fix_gamma) 0 else if (shared_theta_gamma) 0 else ncol(B_gamma_y)

  # gamma_x <- B_gamma_x %*% theta_gamma_x
  # gamma_y <- B_gamma_y %*% theta_gamma_y

  if (shared_theta_gamma) {
    # Use theta_gamma_x for both x and y components
    gamma_x_list <- lapply(1:(nt - 1), function(i) B_gamma_x_list[[i]] %*% theta_gamma_x)
    gamma_y_list <- lapply(1:(nt - 1), function(i) B_gamma_y_list[[i]] %*% theta_gamma_x)
  } else {
    gamma_x_list <- lapply(1:(nt - 1), function(i) B_gamma_x_list[[i]] %*% theta_gamma_x)
    gamma_y_list <- lapply(1:(nt - 1), function(i) B_gamma_y_list[[i]] %*% theta_gamma_y)
  }
  # gamma_t <- B_gamma_t %*% theta_gamma_t
  build_Bs <- function(gamma_x, gamma_y) {
    Dx <- Matrix::Diagonal(x = gamma_x)
    Dy <- Matrix::Diagonal(x = gamma_y)
    Dx %*% Bx %*% Dx + Dy %*% By %*% Dy
  }

  build_Bs_list <- function(gamma_x_list, gamma_y_list) {
    lapply(1:(nt - 1), function(i) build_Bs(gamma_x_list[[i]], gamma_y_list[[i]]))
  }

  build_S <- function(gamma_x, gamma_y) {
    # if gamma_x and gamma_y are 0, then S = 0
    if (sum(gamma_x^2 + gamma_y^2) < 1e-8) {
      return(Matrix::Diagonal(n = ns, x = 0))
    }
    gamma_xx <- gamma_x^2
    gamma_yy <- gamma_y^2
    gamma_xy <- gamma_x * gamma_y
    Dxx <- Matrix::Diagonal(x = gamma_xx)
    Dyy <- Matrix::Diagonal(x = gamma_yy)
    Dxy <- Matrix::Diagonal(x = gamma_xy)
    tmp <- Dxx %*% Hxx %*% Dxx + Dyy %*% Hyy %*% Dyy + Dxy %*% (Hxy + Hyx) %*% Dxy
    gamma_norm <- sqrt(sum(gamma_x^2 + gamma_y^2))
    Matrix::Diagonal(x = cs_diag) %*% tmp / gamma_norm
  }

  build_S_list <- function(gamma_x_list, gamma_y_list) {
    lapply(1:(nt - 1), function(i) build_S(gamma_x_list[[i]], gamma_y_list[[i]]))
  }

  # compute L_s
  update_K <- function(theta_K) {
    cc <- exp(theta_K[1])
    kappa <- exp(theta_K[2])

    if (shared_theta_gamma) {
      theta_gamma_x <- if (length(theta_K) > 2) theta_K[3:(2 + n_theta_gamma_x)] else double(0)
      theta_gamma_y <- theta_gamma_x # Use same parameter for both x and y
    } else {
      theta_gamma_x <- if (length(theta_K) > 2) theta_K[3:(2 + n_theta_gamma_x)] else double(0)
      theta_gamma_y <- if (length(theta_K) > 2) theta_K[(3 + n_theta_gamma_x):length(theta_K)] else double(0)
    }

    # gamma_x <- as.vector(B_gamma_x %*% theta_gamma_x)
    # gamma_y <- as.vector(B_gamma_y %*% theta_gamma_y)
    gamma_x_list <- lapply(1:(nt - 1), function(i) as.vector(B_gamma_x_list[[i]] %*% theta_gamma_x))
    gamma_y_list <- lapply(1:(nt - 1), function(i) as.vector(B_gamma_y_list[[i]] %*% theta_gamma_y))

    if (!fix_gamma) {
      # Bs = build_Bs(gamma_x, gamma_y)
      # S = build_S(gamma_x, gamma_y)
      Bs_list <- build_Bs_list(gamma_x_list, gamma_y_list)
      S_list <- build_S_list(gamma_x_list, gamma_y_list)
    }

    # compute L_s
    # Turn L into a list
    # L = kappa^2 * Cs + lambda * Gs + Bs
    L_list <- lapply(1:(nt - 1), function(i) kappa^2 * Cs + lambda * Gs + Bs_list[[i]])

    Cs_inv <- Matrix::Diagonal(n = ns, x = 1 / Matrix::diag(Cs))

    # watch-out! make sure which t
    if (alpha == 4) L <- L %*% Cs_inv %*% Matrix::t(L)

    if (method == "galerkin") {
      # not used
      K <- Bt %x% Cs + 1 / cc * Ct %x% L
    } else if (method == "euler") {
      # implicit euler
      null_matrix <- Matrix::Diagonal(n = ns, x = 0)

      if (stabilization) {
        # L = L + S
        L_list <- lapply(1:(nt - 1), function(i) L_list[[i]] + S_list[[i]])
      }

      # diagonalize the list of L
      # diag_L <- Matrix::bdiag(
      #   lapply(1:(nt-1), function(i) L)
      # )

      diag_L_list <- Matrix::bdiag(
        lapply(1:(nt - 1), function(i) L_list[[i]])
      )

      # K <- rw1(1:nt)$K %x% Cs + 1/cc *
      #   Matrix::bdiag(null_matrix, diag_L)

      K <- rw1(1:nt)$K %x% Cs + 1 / cc *
        Matrix::bdiag(null_matrix, diag_L_list)

      K <- sqrt(cc) * K
    }
    return(K)
  }

  Bs_list <- build_Bs_list(gamma_x_list, gamma_y_list)
  S_list <- build_S_list(gamma_x_list, gamma_y_list)
  K <- update_K(theta_K)

  BtCs <- if (method == "galerkin") {
    ngme_as_sparse(Bt %x% Cs)
  } else {
    ngme_as_sparse(rw1(1:nt)$K %x% Cs)
  }

  ngme_operator(
    model = "spacetime",
    mesh = mesh,
    alpha = alpha,
    theta_gamma_x = theta_gamma_x,
    theta_gamma_y = theta_gamma_y,
    fix_gamma = fix_gamma,
    shared_theta_gamma = shared_theta_gamma,
    lambda = lambda,
    nt = nt,
    # B_gamma_x = B_gamma_x,
    # B_gamma_y = B_gamma_y,
    B_gamma_x_list = B_gamma_x_list,
    B_gamma_y_list = B_gamma_y_list,
    n_theta_gamma_x = n_theta_gamma_x,
    n_theta_gamma_y = n_theta_gamma_y,
    method = method,
    theta_K = theta_K,
    K = ngme_as_sparse(K),
    update_K = update_K,
    h = h,
    Ct_diag = Matrix::diag(Ct),
    Cs_diag = Matrix::diag(Cs),
    BtCs = BtCs,
    Ct = ngme_as_sparse(Ct),
    Cs = ngme_as_sparse(Cs),
    Gs = ngme_as_sparse(Gs),
    Bx = ngme_as_sparse(Bx),
    By = ngme_as_sparse(By),
    Bs = ngme_as_sparse(Bs_list[[1]]), # only for fixed gamma case
    S = ngme_as_sparse(S_list[[1]]), # only for fixed gamma case
    Hxx = ngme_as_sparse(Hxx),
    Hyy = ngme_as_sparse(Hyy),
    Hxy = ngme_as_sparse(Hxy),
    Hyx = ngme_as_sparse(Hyx),
    stabilization = stabilization,
    symmetric = FALSE,
    zero_trace = FALSE,
    param_name =
      c("cc", "kappa", if (!fix_gamma) {
        if (shared_theta_gamma) "theta_gamma" else c("theta_gamma_x", "theta_gamma_y")
      } else {
        NULL
      }),
    param_trans =
      c(exp, exp, if (!fix_gamma) {
        if (shared_theta_gamma) identity else c(identity, identity)
      } else {
        NULL
      })
  )
}


#' ngme VAR(1) bivariate model specification (Cayley re-parameterization)
#'
#' Builds a Vector Autoregressive order-1 (VAR(1)) operator for bivariate
#' time-series data.  The model follows:
#' \deqn{Y_t = A\,Y_{t-1} + \varepsilon_t, \quad
#'   A = \begin{pmatrix} a_{11} & a_{12} \\ a_{21} & a_{22} \end{pmatrix}}
#'
#' @section Cayley re-parameterization:
#' To guarantee stationarity (\eqn{\rho(A) < 1}) at every SGD step, the
#' \eqn{2 \times 2} coefficient matrix \eqn{A} is obtained via the
#' \emph{Cayley transform}:
#' \deqn{A = (I + S)(I - S)^{-1}}
#' where \eqn{S} is constructed from four unconstrained real parameters
#' \eqn{(p_1, p_2, p_3, p_4) \in \mathbb{R}^4} as \eqn{S = J - R}:
#' \describe{
#'   \item{Skew-symmetric part (rotation):}{
#'     \eqn{J = \begin{pmatrix} 0 & p_1 \\ -p_1 & 0 \end{pmatrix}}}
#'   \item{Positive-definite part (contraction):}{
#'     \eqn{R = L L^{\top} + \varepsilon I}, where
#'     \eqn{L = \begin{pmatrix} p_2 & 0 \\ p_3 & p_4 \end{pmatrix}}
#'     and \eqn{\varepsilon = 10^{-5}}.}
#' }
#' Because \eqn{S + S^{\top} = -2R} is negative definite, all eigenvalues of
#' \eqn{S} have strictly negative real parts, and the Cayley transform then
#' guarantees \eqn{|\lambda_i(A)| < 1} for every \eqn{i}.
#'
#' The default values \eqn{(p_1, p_2, p_3, p_4) = (0, 1, 0, 1)} give
#' \eqn{A \approx 0} (no auto-regression at initialization).
#'
#' @section Precision operator:
#' The 2T x 2T precision operator is:
#' \deqn{K = M_0 + a_{11}\,M_{11} + a_{22}\,M_{22}
#'          + a_{12}\,M_{12} + a_{21}\,M_{21}}
#' where the \eqn{M_{ij}} are fixed sparse block matrices constructed from the
#' \eqn{T \times T} first-order difference matrix \eqn{C_T}.
#'
#' @param mesh Integer vector or \code{inla.mesh.1d} object for the time mesh
#'   (length \eqn{T}).  Pass \code{NULL} for deferred construction.
#' @param p1 Unconstrained skew-symmetric parameter (\eqn{\in \mathbb{R}});
#'   controls the rotation part of the Cayley mapping.  Default \code{0}.
#' @param p2,p3,p4 Unconstrained lower-triangular Cholesky parameters
#'   (\eqn{\in \mathbb{R}}); control the contraction (positive-definite) part.
#'   Defaults \code{p2 = 1}, \code{p3 = 0}, \code{p4 = 1}.
#'
#' @return An \code{ngme_operator} (or \code{ngme_operator_def} when
#'   \code{mesh = NULL}) suitable for use inside \code{f()}.
#'
#' @examples
#' set.seed(1)
#' n_obs <- 10
#' dat <- data.frame(
#'   y     = rnorm(2 * n_obs),
#'   idx   = rep(seq_len(n_obs), 2),
#'   group = factor(rep(c("y1", "y2"), each = n_obs))
#' )
#'
#' fit <- ngme(
#'   y ~ 0 + f(idx, model = var1(mesh = 1:n_obs), group = group,
#'             noise = noise_nig()),
#'   data    = dat,
#'   family  = noise_normal(sigma = 0.01, fix_sigma = TRUE),
#'   control_opt = control_opt(estimation = FALSE)
#' )
#' print(fit)
#'
#' @export
var1 <- function(mesh = NULL, p1 = 0, p2 = 1, p3 = 0, p4 = 1) {
  # Deferred construction (mesh not yet available)
  if (is.null(mesh) ||
    (is.list(mesh) && !inherits(mesh, c("inla.mesh.1d", "inla.mesh",
                                        "fm_mesh_1d", "fm_mesh_2d")))) {
    return(structure(list(model = "var1", args = as.list(environment())),
                     class = "ngme_operator_def"))
  }

  mesh <- ngme_build_mesh(mesh)
  n    <- mesh$n       # number of time points T per series
  eps  <- 1e-5         # small regulariser for positive definiteness

  # ---- Cayley transform: (p1, p2, p3, p4) -> 2x2 stationary matrix A ----
  #
  # Constructs S = J - R where:
  #   J = [[0, p1], [-p1, 0]]              (skew-symmetric, purely imaginary ev)
  #   R = L L^T + eps I,  L = [[p2, 0], [p3, p4]]   (positive definite)
  #
  # S + S^T = -2R  =>  all eigenvalues of S have Re < 0
  # Cayley:  A = (I + S)(I - S)^{-1}  =>  |eigenvalue(A)| < 1  (stationarity)
  #
  # Note: (I+S) and (I-S) commute  =>  solve(I-S, I+S) = (I+S)(I-S)^{-1}
  cayley_to_A <- function(p) {
    p1 <- p[1]; p2 <- p[2]; p3 <- p[3]; p4 <- p[4]
    I2 <- diag(2)

    # Skew-symmetric rotation part
    J  <- matrix(c(0, -p1, p1, 0), nrow = 2)      # J[1,2]=p1, J[2,1]=-p1

    # Positive-definite contraction part  (lower triangular Cholesky)
    L  <- matrix(c(p2, p3, 0, p4), nrow = 2)       # L[1,1]=p2, L[2,1]=p3, L[2,2]=p4
    R  <- L %*% t(L) + eps * I2

    # Stable matrix: all eigenvalues have negative real parts
    S  <- J - R

    # Cayley transform: spectral radius of A is strictly < 1
    solve(I2 - S, I2 + S)
  }

  # ---- Build T x T base matrices ----
  G      <- Matrix::Diagonal(n)                                  # identity
  C      <- Matrix::sparseMatrix(                                # C[t, t-1] = -1
    i = 2:n, j = 1:(n - 1), x = -1, dims = c(n, n)
  )
  zero_n <- Matrix::Diagonal(n, x = 0)

  # ---- Build 2T x 2T block matrices ----
  M0  <- Matrix::bdiag(G, G)                                    # base
  M11 <- Matrix::bdiag(C, zero_n)                               # own-lag y1
  M22 <- Matrix::bdiag(zero_n, C)                               # own-lag y2
  M12 <- rbind(cbind(zero_n, C),      cbind(zero_n, zero_n))   # cross-lag y1 <- y2
  M21 <- rbind(cbind(zero_n, zero_n), cbind(C,      zero_n))   # cross-lag y2 <- y1

  # ---- update_K: called by the optimizer at every gradient step ----
  update_K <- function(theta_K) {
    A   <- cayley_to_A(theta_K)
    a11 <- A[1, 1]; a12 <- A[1, 2]
    a21 <- A[2, 1]; a22 <- A[2, 2]
    ngme_as_sparse(M0 + a11 * M11 + a22 * M22 + a12 * M12 + a21 * M21)
  }

  theta_K <- c(p1 = p1, p2 = p2, p3 = p3, p4 = p4)

  ngme_operator(
    mesh         = mesh,
    model        = "var1",
    theta_K      = theta_K,
    K            = ngme_as_sparse(update_K(theta_K)),
    h            = rep(1, 2 * n),
    update_K     = update_K,
    cayley_to_A  = cayley_to_A,           # stored for print / tests
    matrices     = list(
      ngme_as_sparse(M0),  ngme_as_sparse(M11), ngme_as_sparse(M22),
      ngme_as_sparse(M12), ngme_as_sparse(M21)
    ),
    symmetric    = FALSE,
    zero_trace   = FALSE,
    generic_type = "none",   # Cayley is nonlinear; numerical gradient used
    param_name   = c("p1", "p2", "p3", "p4"),
    param_trans  = list(identity, identity, identity, identity)
  )
}

## Unit tests for the VAR(1) bivariate operator (var1()) -- Cayley reparameterization
##
## The operator takes four UNCONSTRAINED parameters (p1, p2, p3, p4) and maps
## them to a stationary 2x2 VAR coefficient matrix A via the Cayley transform:
##
##   J = [[0, p1], [-p1, 0]]              (skew-symmetric)
##   L = [[p2, 0], [p3, p4]]              (lower-triangular)
##   R = L L^T + eps*I                    (positive definite, eps = 1e-5)
##   S = J - R                            (stable: S + S^T = -2R < 0)
##   A = (I + S)(I - S)^{-1}             (Cayley: spectral radius < 1 guaranteed)
##
## Precision matrix:  K = M0 + a11*M11 + a22*M22 + a12*M12 + a21*M21
## Sign convention:   K[t, t-1] = -a  (because C[t, t-1] = -1)
##
## Tests cover:
##   1.  Deferred construction (mesh = NULL)
##   2.  Operator meta-fields: model name, param_name, n_theta_K
##   3.  K has dimension 2T x 2T for various T
##   4.  h vector is rep(1, 2*T)
##   5.  Diagonal entries of K are always 1
##   6.  K sub-diagonals match -A entries (generic params)
##   7.  Off-block entries are zero when cross-lags are zero (p1=0, p3=0)
##   8.  Stationarity guarantee: spectral radius < 1 for random unconstrained params
##   9.  Default params give A close to -eps/(2+eps)*I (near-zero A)
##  10.  update_K() reproduces the expected K after a parameter change
##  11.  cayley_to_A() is stored in operator and is callable
##  12.  f() with group builds a 2T x 2T identity A matrix (integer map)
##  13.  f() model can be simulated without error

## ---- helper ----------------------------------------------------------------
make_op <- function(T, p1 = 0, p2 = 1, p3 = 0, p4 = 1) {
  var1(mesh = 1:T, p1 = p1, p2 = p2, p3 = p3, p4 = p4)
}
## ---------------------------------------------------------------------------

test_that("var1: deferred construction returns ngme_operator_def", {
  op_def <- var1()
  expect_s3_class(op_def, "ngme_operator_def")
  expect_equal(op_def$model, "var1")

  # args are stored correctly
  op_def2 <- var1(mesh = NULL, p1 = 0.7, p2 = 1.2)
  expect_s3_class(op_def2, "ngme_operator_def")
  expect_equal(op_def2$args$p1, 0.7)
  expect_equal(op_def2$args$p2, 1.2)
})

test_that("var1: operator meta-fields are correct", {
  T  <- 8
  op <- var1(mesh = 1:T, p1 = 0.5, p2 = 1.0, p3 = 0.2, p4 = 0.8)

  expect_s3_class(op, "ngme_operator")
  expect_equal(op$model, "var1")
  expect_equal(op$n_theta_K, 4L)
  expect_equal(op$param_name, c("p1", "p2", "p3", "p4"))
  # theta_K stores the unconstrained params
  expect_equal(unname(op$theta_K), c(0.5, 1.0, 0.2, 0.8))
})

test_that("var1: h is rep(1, 2*T)", {
  for (T in c(5, 10, 20)) {
    op <- var1(mesh = 1:T)
    expect_equal(op$h, rep(1, 2 * T))
  }
})

test_that("var1: K has dimension 2T x 2T", {
  for (T in c(5, 10, 20)) {
    op <- var1(mesh = 1:T)
    expect_equal(nrow(op$K), 2L * T)
    expect_equal(ncol(op$K), 2L * T)
  }
})

test_that("var1: diagonal entries of K are always 1", {
  # The diagonal is 1 regardless of A, because K = I + A*C and C has zeros on diagonal
  param_sets <- list(
    c(0,   1, 0, 1),       # default  (A near 0)
    c(1.5, 0.5, 0.3, 0.8), # moderate
    c(-2,  2,  -1, 3),     # large
    c(0,   0.1, 0, 0.1)    # small L  (A near -1)
  )
  for (params in param_sets) {
    T  <- 7
    op <- make_op(T, params[1], params[2], params[3], params[4])
    K  <- as.matrix(op$K)
    expect_equal(diag(K), rep(1, 2 * T), tolerance = 1e-12,
                 label = paste("p =", paste(round(params, 3), collapse = ",")))
  }
})

## Sign convention note:
## The VAR(1) model is  y_t = A * y_{t-1} + eps_t.
## Rearranged: K * y = eps  where K[t, t-1] = -a  (NOT +a).
## This is because the sub-diagonal matrix C has C[t, t-1] = -1, so
##   K = I + a * C  =>  K[t, t-1] = a * (-1) = -a.
## All sub-diagonal expectations below use -a accordingly.

test_that("var1: K sub-diagonals match -A entries from cayley_to_A", {
  T  <- 6
  # Use params that give non-trivial A with all four entries non-zero
  op <- make_op(T, p1 = 1.2, p2 = 0.8, p3 = 0.3, p4 = 1.5)
  A  <- op$cayley_to_A(op$theta_K)
  a11 <- A[1, 1]; a12 <- A[1, 2]
  a21 <- A[2, 1]; a22 <- A[2, 2]
  K   <- as.matrix(op$K)

  for (t in 2:T) {
    expect_equal(K[t,         t - 1],     -a11, tolerance = 1e-12,
                 label = sprintf("K11[%d,%d]", t, t - 1))
    expect_equal(K[T + t, T + t - 1],     -a22, tolerance = 1e-12,
                 label = sprintf("K22[%d,%d]", t, t - 1))
    expect_equal(K[t,     T + t - 1],     -a12, tolerance = 1e-12,
                 label = sprintf("K12[%d,%d]", t, t - 1))
    expect_equal(K[T + t,     t - 1],     -a21, tolerance = 1e-12,
                 label = sprintf("K21[%d,%d]", t, t - 1))
  }

  # Also verify upper-triangular entries in each T x T block are zero
  K11 <- K[1:T, 1:T]
  K22 <- K[(T + 1):(2 * T), (T + 1):(2 * T)]
  expect_equal(sum(abs(K11[upper.tri(K11)])), 0, tolerance = 1e-12)
  expect_equal(sum(abs(K22[upper.tri(K22)])), 0, tolerance = 1e-12)
})

test_that("var1: off-block entries are zero when p1 = 0 and p3 = 0 (diagonal A)", {
  ## With J = 0 (p1 = 0) and L diagonal (p3 = 0):
  ##   R = L L^T + eps*I is diagonal
  ##   S = J - R = -R is diagonal
  ##   A = (I+S)(I-S)^{-1} is diagonal  => a12 = a21 = 0
  T  <- 6
  op <- make_op(T, p1 = 0, p2 = 1.5, p3 = 0, p4 = 2.0)
  A  <- op$cayley_to_A(op$theta_K)
  K  <- as.matrix(op$K)

  expect_equal(A[1, 2], 0, tolerance = 1e-12)
  expect_equal(A[2, 1], 0, tolerance = 1e-12)

  # K12 and K21 blocks should be all zero
  expect_equal(sum(abs(K[1:T, (T + 1):(2 * T)])), 0, tolerance = 1e-12)
  expect_equal(sum(abs(K[(T + 1):(2 * T), 1:T])), 0, tolerance = 1e-12)
})

test_that("var1: stationarity guaranteed for arbitrary unconstrained params", {
  set.seed(2025)
  for (i in 1:30) {
    p <- rnorm(4, sd = 3)    # deliberately large / varied values
    # Keep p2 and p4 positive so that R = LL^T + eps*I stays positive definite
    p[2] <- abs(p[2]) + 0.01
    p[4] <- abs(p[4]) + 0.01

    op <- make_op(5, p[1], p[2], p[3], p[4])
    A  <- op$cayley_to_A(op$theta_K)
    sr <- max(abs(eigen(A, only.values = TRUE)$values))

    expect_lt(sr, 1,
              label = sprintf("spectral radius (%.3f) for p = (%.2f,%.2f,%.2f,%.2f)",
                              sr, p[1], p[2], p[3], p[4]))
  }
})

test_that("var1: default params give A close to -eps/(2+eps)*I", {
  ## With (p1=0, p2=1, p3=0, p4=1):
  ##   J = 0, L = I, R = (1+eps)*I
  ##   S = -(1+eps)*I
  ##   I+S = -eps*I,  I-S = (2+eps)*I
  ##   A = solve((2+eps)*I, -eps*I) = -eps/(2+eps) * I
  eps      <- 1e-5
  A_diag   <- -eps / (2 + eps)   # expected diagonal value (≈ -5e-6)
  op <- var1(mesh = 1:5)
  A  <- op$cayley_to_A(op$theta_K)

  expect_equal(A[1, 1], A_diag, tolerance = 1e-10)
  expect_equal(A[2, 2], A_diag, tolerance = 1e-10)
  expect_equal(A[1, 2], 0,      tolerance = 1e-12)
  expect_equal(A[2, 1], 0,      tolerance = 1e-12)
})

test_that("var1: update_K rebuilds K consistently when theta_K changes", {
  T  <- 6
  op <- make_op(T, p1 = 0.3, p2 = 1.0, p3 = 0.2, p4 = 0.9)

  # New unconstrained params (arbitrary)
  new_theta <- c(p1 = -0.8, p2 = 1.5, p3 = -0.4, p4 = 1.2)
  K_new <- as.matrix(op$update_K(new_theta))

  # Diagonal entries still 1
  expect_equal(diag(K_new), rep(1, 2 * T), tolerance = 1e-12)

  # Sub-diagonals must match -A entries for the new params
  A_new <- op$cayley_to_A(new_theta)
  for (t in 2:T) {
    expect_equal(K_new[t,         t - 1], -A_new[1, 1], tolerance = 1e-12)
    expect_equal(K_new[T + t, T + t - 1], -A_new[2, 2], tolerance = 1e-12)
    expect_equal(K_new[t,     T + t - 1], -A_new[1, 2], tolerance = 1e-12)
    expect_equal(K_new[T + t,     t - 1], -A_new[2, 1], tolerance = 1e-12)
  }

  # Must match a fresh operator with the same params
  op2 <- make_op(T, p1 = new_theta["p1"], p2 = new_theta["p2"],
                    p3 = new_theta["p3"], p4 = new_theta["p4"])
  expect_equal(K_new, as.matrix(op2$K), tolerance = 1e-12)
})

test_that("var1: cayley_to_A is stored in operator and is callable", {
  op <- var1(mesh = 1:5, p1 = 1.0, p2 = 0.5, p3 = 0.0, p4 = 0.5)

  expect_true(is.function(op$cayley_to_A))

  A <- op$cayley_to_A(op$theta_K)
  expect_equal(dim(A), c(2L, 2L))

  # stationarity from the stored function
  sr <- max(abs(eigen(A, only.values = TRUE)$values))
  expect_lt(sr, 1)
})

test_that("var1: f() builds a 2T x 2T identity A matrix for integer map + group", {
  T     <- 8
  n_obs <- 2 * T
  group <- factor(c(rep("y1", T), rep("y2", T)), levels = c("y1", "y2"))
  map   <- c(1:T, 1:T)

  f_model <- f(
    map,
    model = var1(mesh = 1:T),
    group = group,
    noise = noise_normal()
  )

  A <- as.matrix(f_model$A)
  expect_equal(dim(A), c(n_obs, n_obs))
  expect_equal(A, diag(n_obs), tolerance = 1e-12)
})

test_that("var1: f() model can be simulated without error", {
  set.seed(42)
  T     <- 20
  group <- factor(c(rep("s1", T), rep("s2", T)), levels = c("s1", "s2"))

  f_model <- f(
    c(1:T, 1:T),
    model = var1(mesh = 1:T, p1 = 0.5, p2 = 1.0, p3 = 0.3, p4 = 0.8),
    group = group,
    noise = noise_nig(mu = 0, sigma = 1, nu = 1)
  )

  sim <- simulate(f_model, seed = 1)
  W   <- sim[[1]]

  expect_length(W, 2 * T)
  expect_true(all(is.finite(W)))
  expect_false(anyNA(W))
})

# different methods for SGD
# sgd (vanilla)
# preconditioner_sgd
# momentum
# adagrad
# rmsprop
# adam
# adamW

#' Preconditioner SGD optimization
#'
#' @details
#' Preconditioner SGD is using fisher information matrix as preconditioner (natural gradient descent).
#'
#' @param stepsize stepsize for SGD
#' @param numerical_eps   numerical, the gap used for estimate preconditioner, default is 1e-5
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
precond_sgd <- function(
    stepsize = 1,
    numerical_eps = 1e-5
    # precond_by_diff_chain = FALSE
    ) {
  ret <- list(
    # sgd related
    method = "precond_sgd",
    stepsize = stepsize,
    sgd_parameters = NULL,
    numerical_eps = numerical_eps
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' Vanilla SGD optimization
#'
#' Simple stochastic gradient descent without momentum or preconditioning.
#'
#' @param stepsize stepsize for SGD
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
sgd <- function(
    stepsize = 0.001) {
  ret <- list(
    method         = "sgd",
    stepsize       = stepsize,
    sgd_parameters = double(0)
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' Stochastic Gradient Langevin Dynamics (SGLD) optimization
#'
#' @details
#' SGLD adds Gaussian noise to vanilla SGD updates:
#' \deqn{x_{t+1} = x_t - \eta_t \nabla U(x_t) + \sqrt{2 T \eta_t}\,\xi_t, \quad \xi_t \sim \mathcal{N}(0, I)}
#' where \eqn{T} is \code{temperature}. The implementation applies this
#' component-wise using the current effective stepsize.
#'
#' @param stepsize base stepsize
#' @param temperature non-negative Langevin temperature
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
sgld <- function(
    stepsize = 0.001,
    temperature = 1) {
  stopifnot(
    is.numeric(stepsize),
    length(stepsize) == 1,
    is.finite(stepsize),
    stepsize > 0,
    is.numeric(temperature),
    length(temperature) == 1,
    is.finite(temperature),
    temperature >= 0
  )

  ret <- list(
    method         = "sgld",
    stepsize       = stepsize,
    sgd_parameters = temperature
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' Momentum SGD optimization
#'
#' @details
#' The update rule for momentum is:
#' \deqn{v_t = \beta_1 v_{t-1} + \beta_2 g_t}
#' \deqn{x_{t+1} = x_t - \text{stepsize} * v_t}
#'
#' @param stepsize stepsize for SGD
#' @param beta1 beta1 for momentum
#' @param beta2 beta2 for momentum
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
momentum <- function(
    stepsize = 0.05,
    beta1 = 0.9,
    beta2 = 1 - beta1) {
  ret <- list(
    method         = "momentum",
    stepsize       = stepsize,
    sgd_parameters = c(beta1, beta2)
  )
  class(ret) <- "ngme_optimizer"
  ret
}


#' AdaGrad SGD optimization
#'
#' @details
#' The update rule for AdaGrad is:
#' \deqn{v_t = v_{t-1} + g_t^2}
#' \deqn{x_{t+1} = x_t - \text{stepsize} * \frac{g_t}{\sqrt{v_t} + \epsilon}}
#'
#' @param stepsize stepsize for SGD
#' @param epsilon epsilon for numerical stability
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
adagrad <- function(
    stepsize = 0.05,
    epsilon = 1e-8) {
  ret <- list(
    method         = "adagrad",
    stepsize       = stepsize,
    sgd_parameters = epsilon
  )
  class(ret) <- "ngme_optimizer"
  ret
}


#' Root Mean Square Propagation (RMSProp) SGD optimization
#'
#' @details
#' The update rule for RMSProp is:
#' \deqn{v_t = \beta_1 v_{t-1} + (1 - \beta_1) g_t^2}
#' \deqn{x_{t+1} = x_t - \text{stepsize} * \frac{g_t}{\sqrt{v_t} + \epsilon}}
#'
#' @param stepsize stepsize for SGD
#' @param beta1 beta1 for momentum
#' @param epsilon epsilon for numerical stability
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
rmsprop <- function(
    stepsize = 0.05,
    beta1 = 0.9,
    epsilon = 1e-8) {
  ret <- list(
    method         = "rmsprop",
    stepsize       = stepsize,
    sgd_parameters = c(beta1, epsilon)
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' Adam SGD optimization
#'
#' @details
#' The update rule for Adam is:
#' \deqn{m_t = \beta_1 m_{t-1} + (1 - \beta_1) g_t}
#' \deqn{v_t = \beta_2 v_{t-1} + (1 - \beta_2) g_t^2}
#' \deqn{\hat{m_t} = m_t / (1 - \beta_1^t)}
#' \deqn{\hat{v_t} = v_t / (1 - \beta_2^t)}
#' \deqn{x_{t+1} = x_t - \text{stepsize} * \frac{\hat{m_t}}{\sqrt{\hat{v_t}} + \epsilon}}
#'
#' @param stepsize stepsize for SGD
#' @param beta1 beta1 for Adam
#' @param beta2 beta2 for Adam
#' @param epsilon epsilon for numerical stability
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
adam <- function(
    stepsize = 0.05,
    beta1 = 0.9,
    beta2 = 0.999,
    epsilon = 1e-8) {
  ret <- list(
    method         = "adam",
    stepsize       = stepsize,
    sgd_parameters = c(beta1, beta2, epsilon)
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' AdamW SGD optimization
#'
#' @details
#' The update rule for AdamW is:
#' \deqn{m_t = \beta_1 m_{t-1} + (1 - \beta_1) g_t}
#' \deqn{v_t = \beta_2 v_{t-1} + (1 - \beta_2) g_t^2}
#' \deqn{\hat{m_t} = m_t / (1 - \beta_1^t)}
#' \deqn{\hat{v_t} = v_t / (1 - \beta_2^t)}
#' \deqn{x_{t+1} = x_t - \text{stepsize} * \left( \lambda x_t + \frac{\hat{m_t}}{\sqrt{\hat{v_t}} + \epsilon} \right)}
#'
#' @param stepsize stepsize for SGD
#' @param beta1 beta1 for AdamW
#' @param beta2 beta2 for AdamW
#' @param lambda lambda (weight decay) for AdamW
#' @param epsilon epsilon for numerical stability
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
adamW <- function(
    stepsize = 0.05,
    beta1 = 0.9,
    beta2 = 0.999,
    lambda = 0.01,
    epsilon = 1e-8) {
  ret <- list(
    method         = "adamW",
    stepsize       = stepsize,
    sgd_parameters = c(beta1, beta2, lambda, epsilon)
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' Stepsize decay schedule
#'
#' @details
#' Define a stepsize decay strategy for use in \code{control_opt()}.
#' Currently supports the plateau-based schedule using grad.norm().
#'
#' @param method decay strategy. "none" disables decay; "grad_norm_plateau"
#'   decays when mean grad.norm() across chains fails to decrease for a number
#'   of epochs.
#' @param patience number of consecutive epochs without grad.norm() improvement
#'   before decaying stepsize.
#' @param gamma decay factor applied when triggered (0 < gamma < 1).
#' @param min_delta minimum required decrease in grad.norm() to be counted as improvement.
#' @param warmup number of initial epochs to skip decay checks.
#' @param min_stepsize lower bound for stepsize after decay (absolute value).
#'
#' @return a list of control variables for stepsize decay (used in \code{control_opt}).
#' @export
stepsize_decay <- function(
    method = c("none", "grad_norm_plateau"),
    patience = 3,
    gamma = 0.5,
    min_delta = 0,
    warmup = 0,
    min_stepsize = 0) {
  method <- match.arg(method)
  ret <- list(
    method = method,
    patience = patience,
    gamma = gamma,
    min_delta = min_delta,
    warmup = warmup,
    min_stepsize = min_stepsize
  )
  class(ret) <- "ngme_stepsize_decay"
  ret
}

#' Stepsize schedule
#'
#' @details
#' Define an iteration-dependent stepsize schedule for use in \code{control_opt()}.
#' This schedule is independent of \code{stepsize_decay()} and applies a direct
#' multiplicative scaling by iteration index:
#' \deqn{\eta_t = \eta_0 (t + t_0)^{-\alpha}}.
#'
#' @param method schedule type. "constant" keeps \eqn{\eta_t = \eta_0} and
#'   "poly" uses polynomial decay.
#' @param alpha polynomial exponent used when \code{method = "poly"}.
#'   Must satisfy \eqn{1/2 < \alpha < 1}.
#' @param t0 non-negative offset in iteration index.
#' @param burnin_iter non-negative integer. During these initial iterations,
#'   schedule scaling is fixed to 1. Afterward, polynomial decay is applied
#'   with reset local time index.
#'
#' @return a list of control variables for stepsize schedule
#' (used in \code{control_opt}).
#' @export
stepsize_schedule <- function(
    method = c("constant", "poly"),
    alpha = 0.501,
    t0 = 1,
    burnin_iter = 0) {
  method <- match.arg(method)

  stopifnot(
    is.numeric(alpha),
    length(alpha) == 1,
    is.numeric(t0),
    length(t0) == 1,
    t0 >= 0,
    is.numeric(burnin_iter),
    length(burnin_iter) == 1,
    is.finite(burnin_iter),
    burnin_iter >= 0,
    abs(burnin_iter - round(burnin_iter)) < sqrt(.Machine$double.eps)
  )
  if (method == "poly") {
    stopifnot(alpha > 0.5, alpha < 1)
  }

  ret <- list(
    method = method,
    alpha = alpha,
    t0 = t0,
    burnin_iter = as.integer(round(burnin_iter))
  )
  class(ret) <- "ngme_stepsize_schedule"
  ret
}

.default_stepsize_control <- function() {
  ret <- list(
    schedule = stepsize_schedule(method = "constant"),
    decay = stepsize_decay(method = "none")
  )
  class(ret) <- "ngme_stepsize_control"
  ret
}

#' Unified stepsize control
#'
#' @details
#' Bundle an iteration schedule (\code{stepsize_schedule}) and an optional
#' checkpoint-based decay rule (\code{stepsize_decay}) into one object for
#' \code{control_opt()}.
#'
#' @param schedule schedule component. Either an object from
#'   \code{stepsize_schedule()} or a method string accepted by
#'   \code{stepsize_schedule()}.
#' @param decay decay component. Either an object from
#'   \code{stepsize_decay()} or a method string accepted by
#'   \code{stepsize_decay()}.
#'
#' @return a bundled stepsize-control object for \code{control_opt()}.
#' @export
stepsize_control <- function(
    schedule = stepsize_schedule(method = "constant"),
    decay = stepsize_decay(method = "none")) {
  if (is.character(schedule)) {
    stopifnot(length(schedule) == 1)
    schedule <- stepsize_schedule(method = schedule)
  }
  if (is.character(decay)) {
    stopifnot(length(decay) == 1)
    decay <- stepsize_decay(method = decay)
  }

  stopifnot(
    inherits(schedule, "ngme_stepsize_schedule"),
    inherits(decay, "ngme_stepsize_decay")
  )

  ret <- list(
    schedule = schedule,
    decay = decay
  )
  class(ret) <- "ngme_stepsize_control"
  ret
}

#' Polynomial schedule helper
#'
#' @details
#' Convenience helper that enables polynomial schedule and disables
#' checkpoint-based decay.
#'
#' @param alpha polynomial exponent in \eqn{(1/2,1)}.
#' @param t0 non-negative schedule offset.
#' @param burnin_iter non-negative integer. Initial iterations without
#'   polynomial schedule scaling.
#'
#' @return a \code{stepsize_control()} object.
#' @export
poly_decay <- function(alpha = 0.501, t0 = 1, burnin_iter = 0) {
  stepsize_control(
    schedule = stepsize_schedule(
      method = "poly",
      alpha = alpha,
      t0 = t0,
      burnin_iter = burnin_iter
    ),
    decay = stepsize_decay(method = "none")
  )
}

#' Batch/checkpoint decay helper
#'
#' @details
#' Convenience helper that enables grad-norm plateau decay and keeps
#' iteration schedule constant.
#'
#' @param patience number of consecutive checkpoints without improvement before decay.
#' @param gamma decay factor in \eqn{(0,1)}.
#' @param min_delta minimum required decrease to count as improvement.
#' @param warmup number of initial checkpoints to skip.
#' @param min_stepsize lower bound for absolute stepsize.
#'
#' @return a \code{stepsize_control()} object.
#' @export
batch_decay <- function(
    patience = 3,
    gamma = 0.5,
    min_delta = 0,
    warmup = 0,
    min_stepsize = 0) {
  stepsize_control(
    schedule = stepsize_schedule(method = "constant"),
    decay = stepsize_decay(
      method = "grad_norm_plateau",
      patience = patience,
      gamma = gamma,
      min_delta = min_delta,
      warmup = warmup,
      min_stepsize = min_stepsize
    )
  )
}


#' BFGS optimization
#'
#' @param line_search line search method, can be c("backtracking", "wofle")
#' "bfgs" means use BFGS preconditioner
# ' The update rule for BFGS is:
# ' \deqn{H_{t+1} = H_t + \frac{y_t y_t^T}{y_t^T s_t} - \frac{H_t s_t s_t^T H_t}{s_t^T H_t s_t}}
# ' \deqn{x_{t+1} = x_t - H_{t+1} g_t}
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
bfgs <- function(
    line_search = "wolfe") {
  stopifnot(line_search %in% c("backtracking", "wolfe"))

  ret <- list(
    method         = "bfgs",
    stepsize       = 1,
    line_search    = line_search,
    sgd_parameters = NULL
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' Adaptive gradient descent
#'
#' @description
#' Adaptive gradient descent optimizer.
#'
#' @details
#' Based on the method described in
#' \url{https://arxiv.org/pdf/1910.09529}.
#' The update rule for adaptive gradient descent is:
#' \deqn{\lambda_k = \min(\sqrt{1 + \theta_{k-1}} \lambda_{k-1}, \frac{||x_k - x_{k-1}||}{2 ||\nabla f(x_k) - \nabla f(x_{k-1})||} )}
#' \deqn{x_{k+1} = x_k - \lambda_k \nabla f(x_k)}
#' \deqn{\theta_k = \lambda_k / \lambda_{k-1}}
#'
#' @param stepsize initial stepsize for SGD
#'
#' @return a list of control variables for optimization
#' (used in \code{control_opt} function)
#' @export
adaptive_gd <- function(
    stepsize = 0.01) {
  ret <- list(
    method         = "adaptive_gd",
    stepsize       = stepsize,
    sgd_parameters = double(0)
  )
  class(ret) <- "ngme_optimizer"
  ret
}

#' List supported optimizers
#'
#' @description
#' This function returns a list of supported optimizers in the \code{ngme} package.
#' The optimizers are categorized into three groups:
#' \itemize{
#' \item \strong{Gradient descent}: \code{"sgd"}, \code{"sgld"}, \code{"momentum"}, \code{"adaptive_gd"}
#' \item \strong{Adaptive learning rate}: \code{"adagrad"}, \code{"rmsprop"}, \code{"adam"}, \code{"adamW"}
#' \item \strong{Preconditioner}: \code{"precond_sgd"}, \code{"bfgs"}
#' }
#'
#' @return a character vector of supported optimizers
#' @export
ngme_optimizers <- function() {
  c("sgd", "sgld", "precond_sgd", "momentum", "adagrad", "rmsprop", "adam", "adamW", "bfgs", "adaptive_gd")
}

test_that("control_opt_batch_ci sets recommended CI defaults", {
  ctl <- control_opt_batch_ci(
    iterations = 400,
    n_batch = 10,
    n_parallel_chain = 3,
    alpha = 0.6,
    t0 = 5,
    schedule_burnin_iter = 12
  )

  expect_s3_class(ctl, "control_opt")
  expect_equal(ctl$iterations, 400)
  expect_equal(ctl$n_batch, 10)
  expect_equal(ctl$n_parallel_chain, 3)

  expect_true(isTRUE(ctl$store_traj))
  expect_false(isTRUE(ctl$trend_std_conv_check))
  expect_false(isTRUE(ctl$R_hat_conv_check))
  expect_false(isTRUE(ctl$pflug_conv_check))
  expect_equal(ctl$stepsize_decay, "none")

  expect_equal(ctl$stepsize_schedule, "poly")
  expect_equal(ctl$stepsize_schedule_alpha, 0.6)
  expect_equal(ctl$stepsize_schedule_t0, 5)
  expect_equal(ctl$stepsize_schedule_burnin_iter, 12)
})


test_that("control_opt_batch_ci allows overriding defaults via dots", {
  ctl <- control_opt_batch_ci(
    store_traj = FALSE,
    R_hat_conv_check = TRUE,
    stepsize_control = batch_decay(),
    optimizer = adam(stepsize = 0.01)
  )

  expect_false(isTRUE(ctl$store_traj))
  expect_true(isTRUE(ctl$R_hat_conv_check))
  expect_equal(ctl$stepsize_schedule, "constant")
  expect_equal(ctl$stepsize_decay, "grad_norm_plateau")
  expect_equal(ctl$sgd_method, "adam")
})

# this script is try to fit the Argo data at North Atlantic Ocean
# large-scale mapping using global sinusoidal or Mollweide projection.
load("tests/data/argo_2018_10.Rdata")
data(argo_float)
str(argo_float)

argo_al$JulDay <- NULL
nrow(argo_df)
argo_df$Long <- (argo_df$Long + 180) %% 360 - 180
names(argo_df)

library(maps)
library(ggplot2)
library(splancs)
library(viridis)

# North Atlantic Ocean
world_map <- ggplot2::map_data("world")

sel_lon = c(-80, -6, -6, -80)
sel_lat = c(50, 50, 0, 0)
sel_poly <- cbind(sel_lon, sel_lat)
argo_idx_sel <- inout(with(argo_df, cbind(Long, Lat)), sel_poly)
idx <- inout(with(world_map, cbind(long, lat)), sel_poly)

# Plot
gg <- ggplot(world_map[idx, ], aes(x = long, y = lat)) +
  geom_polygon(aes(group=group), fill = "grey70", color = "white") +
  geom_point(
    data = as.data.frame(argo_df[argo_idx_sel,]),
    aes(x = Long, y = Lat, col = temp_10), size = 0.7
  ) +
  scale_color_gradientn(colours = viridis(100)) +
  coord_fixed(1.3) +
  theme_void()
print(gg)

# Project longlat -> x and y
library(sp)
coordinates(argo_df) <- ~ Long + Lat
proj4string(argo_df) <- CRS("+proj=longlat +datum=WGS84")
moll_crs <- "+proj=moll +lon_0=0 +x_0=0 +y_0=0 +datum=WGS84 +units=km +no_defs"
argo_df_xy <- spTransform(argo_df, CRS(moll_crs))
head(argo_df_xy)

# INLA build mesh
library(INLA)
argo_al <- argo_df_xy[argo_idx_sel, ]
mesh <- fmesher::fm_mesh_2d(coordinates(argo_al),
  max.n = 1000, cutoff=100, max.edge = c(200, 1000))
mesh$n
plot(mesh)
points(coordinates(argo_al))
head(argo_al)
str(argo_al)
# plot salinity + temperature
gtemp <- ggplot(data = as.data.frame(argo_al)) +
  geom_point(aes(x=Long, y=Lat, col=temp_10)) +
  scale_color_gradientn(colours = viridis(100))

gpsal <- ggplot(data = as.data.frame(argo_al)) +
  geom_point(aes(x=Long, y=Lat, col=psal_10)) +
  scale_color_gradientn(colours = viridis(100))

library(gridExtra)
grid.arrange(gtemp, gpsal, ncol = 2)

##### ngme fit
library(ngme2)
myspde <- model_matern(
  mesh=mesh,
  loc=coordinates(argo_al),
  name="spde"
)

############################## Model 1
m0 <- y ~ 1 + lat +
# f(lat, model="rw1", noise=noise_normal()) +
  f(model=myspde, noise=noise_nig())

res0 <- ngme(
  formula = m0,
  data = list(y = argo_al$temp_10,
    lat = scale(argo_al@coords[, 2])),
  control = control_opt(
    iterations = 1000,
    stepsize = 0.5,
    n_parallel_chain = 4
  ),
  family = "normal"
)

res0
traceplot(res0, "mn")
traceplot(res0, "beta")
traceplot(res0, "spde")

w <- res0$models[["field1"]]$W
length(w)
plot(w)


############################## Model 2
# 1. interpolate salinity first
mod_sal <- ngme(
  formula = sal ~ 1 + f(model=myspde, noise=noise_nig()),
  data = list(
    sal = argo_al$psal_10
  ),
  control = control_opt(
    iterations = 1000,
    n_parallel_chain = 4
  )
)

mod_sal
sal_intp <- predict(mod_sal, data=rep(1, mesh$n), loc=list(myspde=mesh$loc[, 1:2]))

# 2. fit the non-stationary model with intepolated salinity

# non-stationary mu
m1 <- y ~ 1 + f(model=myspde, noise=noise_nig(
  theta_mu = c(0, 0),
  # B_mu = cbind(1, scale(mesh$loc[, 2]))
  B_mu = cbind(1, sal_intp)
))

mod_temp <- ngme(
  formula = m1,
  data = list(y = argo_al$temp_10),
  control = control_opt(
    iterations = 1000,
    n_parallel_chain = 4
  ),
  family = "normal"
)

mod_temp
traceplot(mod_temp, "mn")
traceplot(mod_temp, "spde")

traceplot2(mod_temp, "spde", "mu", 2)

p1 <- traceplot2(mod_temp, "spde", "sigma")
p1 + ylim(0, 0.1)
#
m1 <- y ~ 1 + f(model=myspde, noise=noise_nig(
  theta_mu = c(0, 0),
  # B_mu = cbind(1, scale(mesh$loc[, 2]))
  B_mu = cbind(1, sal_intp-mean(sal_intp))
))

mod_temp <- ngme(
  formula = m1,
  data = list(y = argo_al$temp_10),
  control = control_opt(
    iterations = 1000,
    n_parallel_chain = 4,
    max_relative_step = 20,
    max_absolute_step = 20
  ),
  family = "normal"
)

mod_temp
traceplot(mod_temp, "mn")
traceplot(mod_temp, "spde")

# Doing comparison with INLA and ngme2
# using dataset mcycle
# in this script, mainly show how to fit the model in INLA, and ngme2
set.seed(17)
library(MASS)
library(INLA)

data(mcycle)
str(mcycle)
with(mcycle, {plot(times, accel)})
mesh <- inla.mesh.1d(loc=mcycle$times, max.edge=c(1, 10))
mesh$n
points(mesh$loc, rep(0, mesh$n), col=2)

############################## 1. model estimation inla / ngme2
# fit use INLA
spde <- inla.spde2.matern(mesh, alpha=2)
A <- inla.spde.make.A(mesh, loc=mcycle$times)
ind <- inla.spde.make.index("time", spde$n.spde)
data <- list(accel = mcycle$accel, time = ind$time)

result_inla <- inla(
  accel ~ -1 + f(time, model=spde),
  data = data,
  control.predictor = list(A = A),
  control.compute = list(config=TRUE)
)
summary(result_inla)
spde_res <- inla.spde.result(result_inla, "time", spde)
summary(spde_res)
sqrt(1 / 0.002)
# estimation results
summary(result_inla)
plot(spde_res$marginals.kappa$kappa.1,
  type="l", main="kappa")
mean(spde_res$marginals.kappa$kappa.1[, 1])

plot(spde_res$marginals.tau$tau.1, type="l", main="tau")
mean(spde_res$marginals.tau$tau.1[, 1])
names(spde_res)
exp(spde_res$summary.log.kappa)
exp(spde_res$summary.log.tau)
with(mcycle, {plot(times, accel)})
lines(mesh$loc, result_inla$summary.random$time[, "mean"], col=2)
############################## using ngme2
library(ngme2)
spde_ngme <- model_matern(alpha=2, mesh = mesh, loc = mcycle$times)
result_ngme <- ngme(
  accel ~ -1 + f(model = spde_ngme, name="myspde"),
  data = mcycle,
  family = "normal",
  control = control_opt(
    iterations = 1000
  )
)
result_ngme
traceplot(result_ngme, "myspde")
traceplot(result_ngme, "mn")

with(mcycle, {plot(times, accel)})
lines(mesh$loc, result_ngme$models[["myspde"]]$W, lwd=2)
lines(mesh$loc, result_inla$summary.random$time[, "mean"], col=2, lwd=2)
pred_W <- predict(result_ngme, loc=list(myspde = mesh$loc))
# by default we compute, a bunch of statistics for the given location
str(pred_W)

with(mcycle, {plot(times, accel)})
lines(mesh$loc, pred_W[["median"]], col=3, lwd=2)
lines(mesh$loc, pred_W[["5quantile"]], col=5, lwd=2)
lines(mesh$loc, pred_W[["95quantile"]], col=4, lwd=2)

# refit the model using nig noise
result_ngme2 <- ngme(
  accel ~ 0 + f(model = spde_ngme, name="myspde", noise=noise_nig()),
  data = mcycle,
  family = "normal",
  control = control_opt(
    n_parallel_chain = 4,
    iterations = 2000
  ),
  start = result_ngme
)
result_ngme2
traceplot(result_ngme2, "myspde")
plot(result_ngme2$models[["myspde"]]$noise)

############################# model prediction using inla / ngme2
rg <- range(mcycle$times)
rg
locs <- seq(from=rg[1], to=rg[2], length = 100)
locs

# predict with INLA
stk.dat <- inla.stack(
  data = list(accel = mcycle$accel),
  A = list(A),
  effects = list(c(ind)),
  tag = "est"
)
A.prd <- inla.spde.make.A(mesh, loc = locs)
stk.prd <- inla.stack(
  data = list(accel = NA),
  A = list(A.prd),
  effects = list(c(ind)),
  tag = "pred"
)
stk <- inla.stack(stk.dat, stk.prd)
# str(stk)

# doing prediction
result_inla_prd <- inla(
  accel ~ -1 + f(time, model=spde),
  data = inla.stack.data(stk),
  control.predictor = list(A = inla.stack.A(stk)),
  control.compute = list(config=TRUE)
)
# summary(result_inla_prd)

prd_idx <- inla.stack.index(stk, "pred")$data
prd_inla <- result_inla_prd$summary.fitted.values[prd_idx, 1] # mean

# predict with ngme
prd_ngme <- predict(result_ngme2, loc = list(myspde=locs))[["mean"]]

with(mcycle, {plot(times, accel)})
lines(locs, prd_inla)
lines(locs, prd_ngme, col=2)

#' @title Prior Normal
#' @param mean prior mean
#' @param sd prior standard deviation
#' @param target apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)
#' @return prior specification
#' @export
prior_normal <- function(mean = 0, sd = 1, target = "coef") {
  target <- .validate_prior_target(target)
  stopifnot(
    is.numeric(mean), length(mean) == 1, is.finite(mean),
    is.numeric(sd), length(sd) == 1, is.finite(sd), sd > 0
  )
  structure(
    list(dist = "normal", hyper = c(mean = mean, sd = sd), target = target),
    class = "ngme_prior_spec"
  )
}

#' @title Prior PC-SD
#' @param u tail event threshold
#' @param alpha tail probability
#' @param target apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)
#' @return prior specification
#' @export
prior_pc_sd <- function(u, alpha, target = "coef") {
  target <- .validate_prior_target(target)
  stopifnot(
    is.numeric(u), length(u) == 1, is.finite(u), u > 0,
    is.numeric(alpha), length(alpha) == 1, is.finite(alpha),
    alpha > 0, alpha < 1
  )
  structure(
    list(dist = "pc.sd", hyper = c(u = u, alpha = alpha), target = target),
    class = "ngme_prior_spec"
  )
}

#' @title Prior Half-Cauchy
#' @param scale half-Cauchy scale
#' @param target apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)
#' @return prior specification
#' @export
prior_half_cauchy <- function(scale = 1, target = "coef") {
  target <- .validate_prior_target(target)
  stopifnot(
    is.numeric(scale), length(scale) == 1, is.finite(scale), scale > 0
  )
  structure(
    list(dist = "half.cauchy", hyper = c(scale = scale), target = target),
    class = "ngme_prior_spec"
  )
}

#' @title Prior Inverse-Exponential
#' @description Prior induced by \eqn{\kappa = 1 / \nu \sim \mathrm{Exp}(\lambda)},
#' giving \eqn{p(\nu) = \lambda \exp(-\lambda / \nu)\nu^{-2}} for \eqn{\nu > 0}.
#' Internally this prior is applied to \eqn{\nu = \mathrm{lower} + \exp(\theta)}.
#' @param lambda exponential rate on \eqn{\kappa = 1/\nu}
#' @param lower lower shift used in \eqn{\nu = \mathrm{lower} + \exp(\theta)}
#' @param target apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)
#' @return prior specification
#' @export
prior_inv_exponential <- function(lambda = 1, lower = 0, target = "coef") {
  target <- .validate_prior_target(target)
  stopifnot(
    is.numeric(lambda), length(lambda) == 1, is.finite(lambda), lambda > 0,
    is.numeric(lower), length(lower) == 1, is.finite(lower), lower >= 0
  )
  structure(
    list(
      dist = "inv.exponential",
      hyper = c(lambda = lambda, lower = lower),
      target = target
    ),
    class = "ngme_prior_spec"
  )
}

#' @rdname prior_inv_exponential
#' @export
prior_inv_exp <- function(lambda = 1, lower = 0, target = "coef") {
  prior_inv_exponential(lambda = lambda, lower = lower, target = target)
}

#' @title Prior None
#' @param target apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)
#' @return prior specification
#' @export
prior_none <- function(target = "coef") {
  target <- .validate_prior_target(target)
  structure(
    list(dist = "none", hyper = double(0), target = target),
    class = "ngme_prior_spec"
  )
}

#' @title Prior Container
#' @param ... named prior specifications
#' @return named prior container
#' @export
priors <- function(...) {
  ps <- list(...)
  if (length(ps) == 0) {
    return(structure(ps, class = "ngme_priors"))
  }
  nms <- names(ps)
  stopifnot(
    "priors(...) requires named entries." = !is.null(nms) &&
      all(nzchar(nms)) &&
      length(unique(nms)) == length(nms),
    "All entries in priors(...) must be prior_*() objects." =
      all(vapply(ps, inherits, logical(1), what = "ngme_prior_spec"))
  )
  structure(ps, class = "ngme_priors")
}

.validate_prior_target <- function(target) {
  stopifnot(
    is.character(target),
    length(target) == 1,
    target %in% c("coef", "field")
  )
  target
}

is_prior_spec <- function(x) inherits(x, "ngme_prior_spec")
is_prior_collection <- function(x) inherits(x, "ngme_priors")

as_internal_prior <- function(prior) {
  stopifnot("prior must be a prior_*() object." = is_prior_spec(prior))

  target <- .validate_prior_target(prior$target)
  dist <- prior$dist
  hyper <- prior$hyper
  internal <- switch(dist,
    "none" = list(type = "none", param = double(0), target = target),
    "normal" = {
      mean <- as.numeric(hyper["mean"])
      sd <- as.numeric(hyper["sd"])
      list(type = "normal", param = c(mean, 1 / (sd * sd)), target = target)
    },
    "pc.sd" = {
      u <- as.numeric(hyper["u"])
      alpha <- as.numeric(hyper["alpha"])
      lambda <- -log(alpha) / u
      list(type = "pc.sd", param = c(lambda), target = target)
    },
    "half.cauchy" = {
      list(type = "half.cauchy", param = c(as.numeric(hyper["scale"])), target = target)
    },
    "inv.exponential" = {
      lambda <- as.numeric(hyper["lambda"])
      lower <- as.numeric(hyper["lower"])
      list(type = "inv.exponential", param = c(lambda, lower), target = target)
    },
    stop("Unknown prior dist: ", dist)
  )
  internal
}

default_noise_priors <- function() {
  priors(
    mu = prior_normal(0, 10, target = "coef"),
    sigma = prior_normal(0, 10, target = "coef"),
    nu = prior_normal(0, 10, target = "coef")
  )
}

default_operator_prior <- function() {
  prior_normal(0, 10, target = "coef")
}

default_beta_prior <- function() {
  prior_normal(0, 10, target = "coef")
}

default_beta_prior_for_param <- function(param_name) {
  if (!is.null(param_name) && startsWith(param_name, "(Intercept)")) {
    return(prior_none(target = "coef"))
  }
  default_beta_prior()
}

default_beta_priors <- function(param_names) {
  lapply(param_names, default_beta_prior_for_param)
}

compile_noise_priors <- function(prior) {
  if (is.null(prior)) prior <- default_noise_priors()
  if (is_prior_spec(prior)) {
    prior <- priors(mu = prior, sigma = prior, nu = prior)
  }
  stopifnot("noise prior must be priors(...)." = is_prior_collection(prior))

  allowed <- c("mu", "sigma", "nu")
  bad <- setdiff(names(prior), allowed)
  if (length(bad) > 0) {
    stop("Unknown noise prior names: ", paste(bad, collapse = ", "))
  }

  defaults <- default_noise_priors()
  for (nm in names(prior)) defaults[[nm]] <- prior[[nm]]

  list(
    mu = as_internal_prior(defaults$mu),
    sigma = as_internal_prior(defaults$sigma),
    nu = as_internal_prior(defaults$nu)
  )
}

compile_operator_priors <- function(prior, param_names) {
  if (is.null(param_names) || length(param_names) == 0) {
    return(list())
  }
  if (is.null(prior)) {
    return(rep(list(as_internal_prior(default_operator_prior())), length(param_names)))
  }

  if (is_prior_spec(prior)) {
    prior <- priors(theta = prior)
  }
  stopifnot("operator prior must be prior_*() or priors(...)." = is_prior_collection(prior))

  unknown <- setdiff(names(prior), c("theta", param_names))
  if (length(unknown) > 0) {
    stop("Unknown operator prior names: ", paste(unknown, collapse = ", "))
  }

  per_param <- rep(list(default_operator_prior()), length(param_names))
  if ("theta" %in% names(prior)) {
    per_param <- rep(list(prior$theta), length(param_names))
  }
  for (i in seq_along(param_names)) {
    nm <- param_names[[i]]
    if (nm %in% names(prior)) per_param[[i]] <- prior[[nm]]
  }

  lapply(per_param, as_internal_prior)
}

compile_beta_priors <- function(prior, param_names) {
  if (is.null(param_names) || length(param_names) == 0) {
    return(list())
  }
  if (is.null(prior)) {
    return(lapply(default_beta_priors(param_names), as_internal_prior))
  }

  if (is_prior_spec(prior)) {
    prior <- priors(beta = prior)
  }
  stopifnot("beta prior must be prior_*() or priors(...)." = is_prior_collection(prior))

  index_names <- paste0("feff_", seq_along(param_names))
  unknown <- setdiff(names(prior), c("beta", param_names, index_names))
  if (length(unknown) > 0) {
    stop("Unknown beta prior names: ", paste(unknown, collapse = ", "))
  }

  per_param <- default_beta_priors(param_names)
  if ("beta" %in% names(prior)) {
    per_param <- rep(list(prior$beta), length(param_names))
  }
  for (i in seq_along(param_names)) {
    nm <- param_names[[i]]
    idx_nm <- index_names[[i]]
    if (idx_nm %in% names(prior)) per_param[[i]] <- prior[[idx_nm]]
    if (nm %in% names(prior)) per_param[[i]] <- prior[[nm]]
  }

  internal <- lapply(per_param, as_internal_prior)
  bad_target <- which(vapply(internal, function(x) x$target != "coef", logical(1)))
  if (length(bad_target) > 0) {
    stop("beta prior currently supports target='coef' only.")
  }
  internal
}

test_that("prior constructors validate and carry target", {
  p <- prior_normal(mean = 1, sd = 2, target = "field")
  expect_s3_class(p, "ngme_prior_spec")
  expect_equal(p$dist, "normal")
  expect_equal(unname(p$hyper), c(1, 2))
  expect_equal(p$target, "field")

  expect_error(prior_normal(sd = -1))
  expect_error(prior_pc_sd(u = 0, alpha = 0.1))
  expect_error(prior_pc_sd(u = 1, alpha = 1.2))
  expect_error(prior_half_cauchy(scale = 0))
  expect_error(prior_inv_exponential(lambda = 0))
  expect_error(prior_inv_exponential(lower = -1))
})

test_that("prior_inv_exp is an alias of prior_inv_exponential", {
  p1 <- prior_inv_exponential(lambda = 1.2, lower = 0.3, target = "coef")
  p2 <- prior_inv_exp(lambda = 1.2, lower = 0.3, target = "coef")
  expect_equal(p2, p1)
})

test_that("ngme_noise compiles named priors to internal format", {
  p <- priors(
    mu = prior_normal(0, 2, target = "coef"),
    sigma = prior_pc_sd(1, 0.1, target = "field"),
    nu = prior_inv_exponential(2, lower = 0.5, target = "coef")
  )

  noise <- ngme_noise(
    noise_type = "nig",
    theta_mu = 0,
    theta_sigma = 0,
    theta_nu = 0,
    B_mu = matrix(1, 1, 1),
    B_sigma = matrix(1, 1, 1),
    B_nu = matrix(1, 1, 1),
    prior = p
  )

  expect_equal(noise$prior_mu$type, "normal")
  expect_equal(noise$prior_mu$target, "coef")
  expect_equal(noise$prior_mu$param[2], 1 / 4)

  expect_equal(noise$prior_sigma$type, "pc.sd")
  expect_equal(noise$prior_sigma$target, "field")
  expect_equal(noise$prior_sigma$param[1], -log(0.1) / 1)

  expect_equal(noise$prior_nu$type, "inv.exponential")
  expect_equal(noise$prior_nu$target, "coef")
  expect_equal(noise$prior_nu$param[1], 2)
  expect_equal(noise$prior_nu$param[2], 0.5)
})

test_that("f compiles operator priors from new API", {
  fit <- f(
    map = 1:10,
    model = ar1(),
    noise = noise_normal(),
    prior = priors(
      theta = prior_normal(0, 2),
      rho = prior_half_cauchy(1)
    )
  )

  expect_true(inherits(fit, "ngme_model"))
  expect_equal(length(fit$prior_theta_K), length(fit$operator$theta_K))
  expect_equal(fit$prior_theta_K[[1]]$type, "half.cauchy")
  expect_equal(fit$prior_theta_K[[1]]$target, "coef")

  expect_error(
    f(
      map = 1:10,
      model = ar1(),
      noise = noise_normal(),
      prior = priors(bad_name = prior_normal(0, 1))
    ),
    "Unknown operator prior names"
  )
})

test_that("f sets default nu prior to inv.exponential using median(h)", {
  fit <- f(
    map = 1:20,
    model = ar1(),
    noise = noise_nig()
  )

  expect_equal(fit$noise$prior_nu$type, "inv.exponential")
  expect_equal(fit$noise$prior_nu$target, "coef")
  expect_equal(fit$noise$prior_nu$param[1], log(2), tolerance = 1e-12)
  expect_equal(fit$noise$prior_nu$param[2], 0)
})

test_that("f keeps explicit nu prior and does not override it", {
  fit <- f(
    map = 1:20,
    model = ar1(),
    noise = noise_nig(
      prior = priors(
        nu = prior_normal(0, 2)
      )
    )
  )

  expect_equal(fit$noise$prior_nu$type, "normal")
  expect_equal(fit$noise$prior_nu$param[2], 1 / 4)
})

test_that("f keeps normal default prior for non-stationary nu", {
  B_nu <- cbind(1, seq(0, 1, length.out = 20))
  fit <- f(
    map = 1:20,
    model = ar1(),
    noise = noise_nig(
      B_nu = B_nu,
      theta_nu = c(0, 0)
    )
  )

  expect_equal(fit$noise$prior_nu$type, "normal")
  expect_equal(fit$noise$prior_nu$param[1], 0)
  expect_equal(fit$noise$prior_nu$param[2], 0.01)
})

test_that("beta prior compilation supports global and named forms", {
  b <- compile_beta_priors(
    priors(
      beta = prior_normal(0, 2),
      x = prior_half_cauchy(1),
      feff_3 = prior_normal(1, 0.5)
    ),
    c("(Intercept)", "x", "z")
  )

  expect_equal(length(b), 3)
  expect_equal(b[[1]]$type, "normal")
  expect_equal(b[[1]]$param[2], 1 / 4)
  expect_equal(b[[2]]$type, "half.cauchy")
  expect_equal(b[[3]]$type, "normal")
  expect_equal(b[[3]]$param[1], 1)

  expect_error(
    compile_beta_priors(priors(unknown = prior_normal(0, 1)), c("a", "b")),
    "Unknown beta prior names"
  )
  expect_error(
    compile_beta_priors(priors(beta = prior_normal(0, 1, target = "field")), c("a")),
    "target='coef' only"
  )
})

test_that("default beta prior leaves intercept unconstrained", {
  b <- compile_beta_priors(NULL, c("(Intercept)", "x", "(Intercept)_u_wind", "z"))

  expect_equal(length(b), 4)
  expect_equal(b[[1]]$type, "none")
  expect_equal(b[[1]]$target, "coef")
  expect_equal(length(b[[1]]$param), 0)
  expect_equal(b[[2]]$type, "normal")
  expect_equal(b[[3]]$type, "none")
  expect_equal(b[[4]]$type, "normal")
})

test_that("ngme accepts prior_beta and stores compiled priors per replicate", {
  dat <- data.frame(y = rnorm(20), x = rnorm(20), id = 1:20)

  fit <- ngme(
    y ~ x + f(id, model = ar1(), noise = noise_normal()),
    data = dat,
    family = noise_normal(),
    prior_beta = priors(
      beta = prior_normal(0, 2),
      x = prior_half_cauchy(1)
    ),
    control_opt = control_opt(estimation = FALSE, standardize_fixed = FALSE)
  )

  pb <- fit$replicates[[1]]$prior_beta
  expect_equal(length(pb), ncol(fit$replicates[[1]]$X))
  expect_equal(pb[[2]]$type, "half.cauchy")
  expect_equal(pb[[1]]$type, "normal")
})

#' Non-stationary precision matrix operator with custom matrix combinations
#'
#' Creates a flexible non-stationary precision matrix K by allowing both linear combinations
#' and matrix multiplications with parameter-dependent diagonal matrices. This enables modeling
#' spatially varying parameters through basis expansions.
#'
#' The key distinction from `generic` is that `generic_ns`:
#' 1. Requires explicit position specification for matrix combinations
#' 2. Converts parameters to diagonal matrices for multiplication with fixed matrices
#' 3. Allows for basis expansions via B_theta_K
#'
#' @param theta_K List of parameter vectors (in real scale)
#' @param matrices List of fixed matrices to be used in the model
#' @param position List of vectors specifying matrix combinations (required)
#' @param h The integration weights vector
#' @param model The model type name stored on the returned operator
#' @param B_theta_K List of basis matrices for parameters, defaults to matrices of 1s if not provided
#' @param trans List of transformations for each parameter (one transformation per parameter)
#' @param mesh The mesh object
#' @param zero_trace Whether the trace of K should be zero
#' @param param_name Optional parameter names stored on the returned operator for diagnostics
#' @param param_trans Optional parameter transformations stored on the returned operator for diagnostics
#' @param ... Additional arguments (ignored)
#'
#' @details
#' The `generic_ns` operator constructs K through the following steps:
#'
#' 1. Create diagonal matrices for each parameter using basis expansions: D_param = diag(B_param * theta_param)
#' 2. Combine parameter matrices with fixed matrices according to position specifications
#' 3. Sum all the resulting combinations
#'
#' The `position` parameter defines how matrices are combined:
#' - Each element of the list represents a term in the sum
#' - Each vector contains indices of matrices to multiply (parameter matrices first, then fixed matrices)
#' - For example: `position = list(c(1, 3), c(2, 4))` with one parameter means:
#'   Multiply the parameter diagonal matrix (index 1) with fixed matrix 1 (index 3),
#'   then add parameter diagonal matrix 2 (index 2) multiplied by fixed matrix 2 (index 4)
#'
#' Spatially-varying parameters can be created using basis matrices in B_theta_K.
#'
#' @return An object of class \code{ngme_operator}. The object stores the sparse
#'   non-stationary precision matrix \code{K}, integration weights \code{h}, the
#'   flattened parameter vector \code{theta_K}, basis matrices
#'   \code{B_theta_K}, parameter grouping metadata \code{param_map}, matrix
#'   combination rules \code{position}, and an update function for rebuilding
#'   \code{K}. It is intended for use as the latent \code{model} inside
#'   \code{f()}.
#'
#' @examples
#' # Simple example with one parameter
#' n <- 5
#' A <- matrix(1, n, n)
#' B <- matrix(2, n, n)
#' alpha <- 0.4
#'
#' # Create a simple generic_ns model: D_alpha * A + B
#' model <- generic_ns(
#'   theta_K = list(alpha = c(alpha)),
#'   matrices = list(A, B),
#'   position = list(c(1, 2), c(3)), # D_alpha * A + B
#'   h = rep(1, n)
#' )
#'
#' # AR1 model representation: rho * C + G
#' ar1_obj <- ar1(1:10, rho = 0.5)
#' g <- name2fun("tanh", inv = TRUE)
#' generic_ar1 <- generic_ns(
#'   theta_K = list(x = c(g(0.5))), # Transform from real scale
#'   trans = list(x = "tanh"), # Apply tanh transformation
#'   matrices = list(ar1_obj$C, ar1_obj$G),
#'   position = list(c(1, 2), c(3)), # D_rho * C + G
#'   h = ar1_obj$h,
#'   mesh = 1:10
#' )
#'
#' # Spatially varying Matern model
#' # Create a simple 1D mesh
#' mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = 10))
#'
#' # Create basis for spatially-varying kappa
#' n <- 10
#' B_kappa <- matrix(0, n, 2)
#' B_kappa[1:(n / 2), 1] <- 1 # First half of the domain
#' B_kappa[(n / 2 + 1):n, 2] <- 1 # Second half of the domain
#'
#' # Create standard Matern components
#' matern_model <- matern(mesh)
#'
#' # Create model with space-varying kappa
#' ns_model <- generic_ns(
#'   theta_K = list(kappa = c(log(1), log(2))), # Different values for each region
#'   matrices = list(matern_model$C, matern_model$G),
#'   B_theta_K = list(kappa = B_kappa),
#'   trans = list(kappa = "exp2"), # kappa^2 transformation for C
#'   h = matern_model$h,
#'   position = list(c(1, 2), c(3)), # D_kappa * C + G
#'   mesh = mesh
#' )
#'
#' @export
generic_ns <- function(
    theta_K,
    matrices,
    position,
    h,
    model = "generic_ns",
    B_theta_K = NULL,
    trans = NULL,
    mesh = NULL,
    zero_trace = FALSE,
    param_name = NULL,
    param_trans = NULL,
    ...) {
  if (!is.null(mesh)) mesh <- ngme_build_mesh(mesh)

  # Validate inputs
  stopifnot(
    "theta_K should be a list" = is.list(theta_K),
    "matrices should be a list" = is.list(matrices),
    "matrices should not be NULL" = !is.null(matrices),
    "matrices should not be a list of NULL" = all(!sapply(matrices, is.null)),
    "all matrices should have the same dimensions" = all(sapply(matrices, function(m) nrow(m) == nrow(matrices[[1]])) &
      sapply(matrices, function(m) ncol(m) == ncol(matrices[[1]]))),
    "h should be of the same dimension as the matrices" = length(h) == nrow(matrices[[1]]),
    "position is required" = !is.null(position) && is.list(position)
  )

  # Initialize transformations if not provided
  if (is.null(trans)) {
    trans <- list()
    for (theta_K_name in names(theta_K)) {
      trans[[theta_K_name]] <- "identity"
    }
  }

  # Validate transformation list
  stopifnot(
    "trans should be a list" = is.list(trans),
    "trans should have names matching theta_K" = all(names(trans) %in% names(theta_K))
  )

  # Initialize B_theta_K with matrices of 1s if not provided
  if (is.null(B_theta_K)) {
    B_theta_K <- list()
    n <- nrow(matrices[[1]])
    for (theta_K_name in names(theta_K)) {
      # Create a matrix of 1s with appropriate dimensions
      # Use a single column matrix since most parameters are scalar
      B_theta_K[[theta_K_name]] <- matrix(1, nrow = n, ncol = 1)
    }
  }

  # Validate B_theta_K
  stopifnot(
    "B_theta_K should be a list" = is.list(B_theta_K),
    "B_theta_K should have names matching theta_K" = all(names(B_theta_K) %in% names(theta_K))
  )

  # Process basis matrices for parameters
  param_matrices <- list()
  for (theta_K_name in names(B_theta_K)) {
    B <- B_theta_K[[theta_K_name]]
    theta <- theta_K[[theta_K_name]]

    # Validate dimensions
    if (length(theta) > 1 && ncol(B) != length(theta)) {
      stop(sprintf("B_theta_K[[%s]] should have ncol matching length of theta_K[[%s]]", theta_K_name, theta_K_name))
    }

    # Apply transformation to create parameter matrix
    trans_type <- trans[[theta_K_name]]
    transformed_value <- name2fun(trans_type)(B %*% theta)

    if (trans_type %in% c("identity", "exp", "exp2", "exp4", "sqrt", "square", "log", "tanh")) {
      # For element-wise transformations, create diagonal matrix
      param_matrices[[theta_K_name]] <- Matrix::Diagonal(x = as.numeric(transformed_value))
    } else {
      # For other transformations (like tanh for correlations), create the matrix directly
      param_matrices[[theta_K_name]] <- transformed_value
    }
  }

  # Combine parameters first, then fixed matrices
  all_matrices <- c(param_matrices, matrices)

  # Create flat parameter vector and keep track of mappings
  flat_theta_K <- numeric(0)
  param_map <- list()
  idx <- 1

  for (theta_K_name in names(theta_K)) {
    theta_K_len <- length(theta_K[[theta_K_name]])
    param_map[[theta_K_name]] <- idx:(idx + theta_K_len - 1)
    flat_theta_K <- c(flat_theta_K, theta_K[[theta_K_name]])
    idx <- idx + theta_K_len
  }
  names(flat_theta_K) <- rep(names(theta_K), sapply(theta_K, length))

  flat_names <- names(flat_theta_K)

  # Build parameter names for each flattened coefficient
  expand_group_labels <- function(labels) {
    expanded <- character(0)
    for (group_idx in seq_along(param_map)) {
      group_name <- trimws(labels[group_idx])
      if (is.na(group_name) || !nzchar(group_name)) {
        group_name <- paste0("theta_K_group", group_idx)
      }
      indices <- param_map[[group_idx]]
      if (length(indices) == 1) {
        expanded <- c(expanded, group_name)
      } else {
        tokens <- strsplit(group_name, "\\s+")[[1]]
        last_token <- tail(tokens, 1)
        if (length(tokens) > 1 && grepl("^[0-9]+$", last_token)) {
          base <- trimws(paste(tokens[-length(tokens)], collapse = " "))
          start_idx <- as.integer(last_token)
        } else {
          base <- group_name
          start_idx <- 1
        }
        new_labels <- paste(base, start_idx + seq_along(indices) - 1)
        expanded <- c(expanded, trimws(new_labels))
      }
    }
    expanded
  }

  if (is.null(param_name)) {
    if (!is.null(flat_names) && length(flat_names) == length(flat_theta_K) && all(nzchar(flat_names))) {
      param_name <- make.unique(flat_names)
    } else {
      param_name <- paste0("theta_K", seq_along(flat_theta_K))
    }
  } else if (length(param_name) == length(param_map) && length(flat_theta_K) != length(param_name)) {
    param_name <- expand_group_labels(param_name)
  }

  # Extract parameter transformation functions for each coefficient in flat_theta_K
  if (is.null(param_trans)) {
    param_trans <- vector("list", length(flat_theta_K))
    if (length(flat_theta_K) > 0) {
      for (theta_K_name in names(param_map)) {
        indices <- param_map[[theta_K_name]]
        # Basis coefficients live on the optimizer scale; only scalar parameters
        # get the model-space transform in ngme_result().
        trans_fun <- if (length(indices) == 1) {
          name2fun(trans[[theta_K_name]])
        } else {
          identity
        }
        param_trans[indices] <- replicate(length(indices), trans_fun, simplify = FALSE)
      }
    }
  } else if (length(param_trans) == length(param_map) && length(flat_theta_K) != length(param_trans)) {
    expanded <- vector("list", length(flat_theta_K))
    for (group_idx in seq_along(param_map)) {
      indices <- param_map[[group_idx]]
      trans_fun <- param_trans[[group_idx]]
      expanded[indices] <- replicate(length(indices), trans_fun, simplify = FALSE)
    }
    param_trans <- expanded
  } else if (length(param_trans) != length(flat_theta_K)) {
    stop("param_trans should have the same length as the flattened theta_K vector or the number of parameter groups")
  }

  if (length(param_trans) > 0 && is.null(names(param_trans))) {
    names(param_trans) <- param_name
  }

  # Helper function to compute K (theta_K is a flat vector)
  update_K <- function(theta_K) {
    # Initialize K as zero matrix
    n <- nrow(matrices[[1]])
    K <- Matrix::Matrix(0, n, n)

    # Create parameter matrices using basis expansions
    param_matrices <- list()
    all_matrices <- list()

    # First, create parameter matrices
    for (theta_K_name in names(param_map)) {
      indices <- param_map[[theta_K_name]]

      # Get basis matrix
      B <- if (!is.null(B_theta_K[[theta_K_name]])) {
        B_theta_K[[theta_K_name]]
      } else {
        # Default to matrix of ones (1 column)
        Matrix::Matrix(1, nrow = n, ncol = 1)
      }

      # Extract parameter values from theta_K
      param_values <- theta_K[indices]

      # Apply basis expansion
      expanded_values <- as.vector(B %*% param_values)

      # Apply transformation
      trans_type <- if (!is.null(trans[[theta_K_name]])) trans[[theta_K_name]] else "identity"
      transformed_values <- name2fun(trans_type)(expanded_values)

      # Create diagonal matrix from expanded values
      param_matrices[[theta_K_name]] <- Matrix::Diagonal(x = as.numeric(transformed_values))
    }

    # Create combined list of matrices (parameters first, then fixed matrices)
    all_matrices <- c(param_matrices, matrices)

    # Apply matrix combinations according to position
    for (pos in position) {
      # Start with identity matrix
      term <- Matrix::Diagonal(n)

      # Multiply matrices in the combination
      for (idx in pos) {
        if (idx > 0 && idx <= length(all_matrices)) {
          term <- term %*% all_matrices[[idx]]
        } else {
          stop(sprintf(
            "Position index %d is out of bounds (max: %d)",
            idx, length(all_matrices)
          ))
        }
      }

      # Add to K
      K <- K + term
    }

    return(K)
  }

  # Create operator
  ngme_operator(
    model = model,
    mesh = mesh,
    K = ngme_as_sparse(update_K(theta_K = flat_theta_K)),
    generic_type = "generic_ns",
    h = h,
    theta_K = flat_theta_K,
    trans = trans, # Include the original transforms
    B_theta_K = B_theta_K,
    param_map = param_map,
    position = position,
    matrices = sapply(matrices, function(m) ngme_as_sparse(m)), # Store original matrices, not all_matrices
    update_K_func = update_K,
    symmetric = all(sapply(all_matrices, Matrix::isSymmetric)),
    zero_trace = zero_trace,
    param_name = param_name,
    param_trans = param_trans,
    ...
  )
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

estimate_cpp <- function(R_ngme, control_opt) {
    .Call(`_ngme2_estimate_cpp`, R_ngme, control_opt)
}

sampling_cpp <- function(ngme_replicate, n, n_burnin, posterior, seed) {
    .Call(`_ngme2_sampling_cpp`, ngme_replicate, n, n_burnin, posterior, seed)
}

get_openmp_threads <- function() {
    .Call(`_ngme2_get_openmp_threads`)
}

has_pardiso <- function() {
    .Call(`_ngme2_has_pardiso`)
}

compute_log_like_cpp <- function(R_ngme) {
    .Call(`_ngme2_compute_log_like_cpp`, R_ngme)
}

rGIG_cpp <- function(p, a, b, seed) {
    .Call(`_ngme2_rGIG_cpp`, p, a, b, seed)
}

#' Generic precision matrix operator
#'
#' Creates a flexible precision matrix K by allowing linear combinations of base matrices
#' with parameter-dependent coefficients. The model follows the form:
#' K = sum_i f_i(theta) * matrices_i, where f_i depends on the parameter transformation.
#'
#' This function can be used to represent various models like AR1, Matern, and RW1 in a unified framework.
#'
#' @param theta_K The parameter vector (in real scale). If missing, initialized as double(0)
#' @param matrices A list of matrices to be combined with parameter-dependent coefficients
#' @param h The integration weights vector
#' @param trans Transformation for each parameter. Must be a list where each element
#'   is a character vector defining transformations for each matrix.
#'   For example, `trans=list(kappa=c("exp2", "identity", "null"))` means parameter 'kappa'
#'   affects matrix 1 with exp2 transformation, matrix 2 with identity, and doesn't affect matrix 3.
#'   Available transformations: "identity", "exp", "exp2", "exp4", "sqrt", "square", "log", "tanh"
#' @param mesh The mesh object
#' @param model The model type name
#' @param zero_trace Whether the trace of K should be zero
#' @param param_name Optional parameter names stored on the returned operator for diagnostics
#' @param param_trans Optional parameter transformations stored on the returned operator for diagnostics
#' @param ... Additional arguments (ignored)
#' 
#' @details 
#' The `generic` function provides a flexible way to construct precision matrices by combining 
#' existing matrices with transformed parameters. Here's how it works:
#' 
#' 1. Each parameter in `theta_K` can affect one or more matrices in the `matrices` list
#' 2. The `trans` list defines how each parameter transforms before multiplying each matrix
#' 3. For each matrix, the coefficients from all parameters are multiplied together
#' 4. The final K is the sum of all transformed matrices
#' 
#' Common use cases:
#' 
#' - **AR1 model**: K = rho * C + G (where rho is between -1 and 1)
#' - **Matern (alpha=2)**: K = kappa^2 * C + G (where kappa > 0)
#' - **Matern (alpha=4)**: K = kappa^4 * C + 2*kappa^2 * G + G * Cinv * G
#'
#' @return An object of class \code{ngme_operator}. The object stores the sparse
#'   precision matrix \code{K}, integration weights \code{h}, parameter vector
#'   \code{theta_K}, base \code{matrices}, transformation specification
#'   \code{trans}, and an \code{update_K} function that rebuilds \code{K} for new
#'   parameter values. It is intended for use as the latent \code{model} inside
#'   \code{f()}.
#' 
#' @examples
#' # AR1 model with rho = 0.5
#' ar1_obj <- ar1(1:10, rho = 0.5)
#' g <- name2fun("tanh", inv = TRUE)
#' generic_ar1 <- generic(
#'   theta_K = c(rho = g(0.5)),  # Transform from real scale
#'   trans = list(rho = c("tanh", "null")),  # Apply tanh to first matrix
#'   matrices = list(ar1_obj$C, ar1_obj$G),
#'   h = ar1_obj$h
#' )
#' 
#' # Matern model with kappa = 2
#' mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = 10))
#' matern_obj <- matern(mesh)
#' log_kappa <- log(2)
#' generic_matern <- generic(
#'   theta_K = c(kappa = log_kappa),
#'   trans = list(kappa = c("exp2", "null")),  # exp(2*kappa) for C, nothing for G
#'   matrices = list(matern_obj$C, matern_obj$G),
#'   h = matern_obj$h
#' )
#' 
#' # Matern model with alpha = 4 and kappa = 2
#' set.seed(1)
#' mesh <- fmesher::fm_mesh_2d(cbind(x = runif(20), y = runif(20)))
#' matern_obj <- matern(mesh, alpha = 4)
#' C <- matern_obj$C
#' G <- matern_obj$G
#' Cinv <- C; diag(Cinv) <- 1 / Matrix::diag(C)
#' 
#' generic_matern4 <- generic(
#'   theta_K = c(kappa = log(2)),
#'   trans = list(kappa = c("exp4", "exp2", "null")),
#'   matrices = list(C, 2*G, G %*% Cinv %*% G),
#'   h = matern_obj$h
#' )
#'
#' @export
generic <- function(
    theta_K,
    matrices,
    h,
    trans = NULL,
    mesh = NULL,
    zero_trace = FALSE,
    model = "generic",
    param_name = NULL,
    param_trans = NULL,
    ...
) {
  if (!is.null(mesh)) mesh <- ngme_build_mesh(mesh)

  # If theta_K is not provided, initialize it as double(0)
  if (missing(theta_K)) {
    theta_K <- double(0)
  }

  # Validate matrices
  stopifnot(
    "matrices should not be NULL" = !is.null(matrices),
    "matrices should be a list" = is.list(matrices),
    "matrices should not be a list of NULL" = all(!sapply(matrices, is.null)),
    "all matrices should have the same dimensions" = all(sapply(matrices, function(m) nrow(m) == nrow(matrices[[1]])) &
      sapply(matrices, function(m) ncol(m) == ncol(matrices[[1]]))),
    "h should be of the same dimension as the matrices" = length(h) == nrow(matrices[[1]])
  )

  # Convert old format to new format if needed
  if (is.null(trans) && length(theta_K) > 0) {
    # If trans is NULL, create default trans with only 1s for the identity matrices
    trans <- list()
    for (i in seq_along(theta_K)) {
      trans[[i]] <- rep("null", length(matrices))
      trans[[i]][i] <- "identity"
    }
    # If no parameters provided, just use default behavior (all matrices with coefficient 1)
  } else if (is.character(trans)) {
    # Convert old format to new format
    old_trans <- trans
    trans <- list()
    for (i in seq_along(theta_K)) {
      theta_K_name <- names(theta_K)[i]
      trans[[theta_K_name]] <- rep("null", length(matrices))
      trans[[theta_K_name]][i] <- old_trans[i]
    }
  }

  # Validate the structure of trans
  if (length(theta_K) > 0) {
    # Only validate trans if we have parameters
    for (theta_K_name in names(trans)) {
      if (length(trans[[theta_K_name]]) != length(matrices)) {
        stop(sprintf(
          "For parameter '%s', the transformation vector length (%d) must match the number of matrices (%d)",
          theta_K_name, length(trans[[theta_K_name]]), length(matrices)
        ))
      }
    }
  }

  # Helper function to compute the coefficient of each matrix
  update_K <- function(theta_K) {
    coef <- compute_coef_new(theta_K, trans, length(matrices))

    K <- 0
    for (i in seq_along(matrices)) {
      K <- K + coef[i] * matrices[[i]]
    }
    return(K)
  }

  if (is.null(param_trans)) {
    param_trans <- get_param_trans(trans)
  }
  if (is.null(param_name)) {
    param_name <- names(theta_K)
  }

  ngme_operator(
    model = model,
    mesh = mesh,
    K = ngme_as_sparse(update_K(theta_K)),
    generic_type = "generic",
    h = h,
    theta_K = theta_K,
    trans = trans,
    update_K = update_K,
    matrices = sapply(matrices, ngme_as_sparse),
    symmetric = all(sapply(matrices, Matrix::isSymmetric)),
    zero_trace = zero_trace,
    param_name = param_name,
    param_trans = param_trans,
    ...
  )
}

# Helper function to get param_trans from trans list
# For each parameter, if only one transformation type is used (excluding null),
# use that transformation function, otherwise use identity
get_param_trans <- function(trans) {
  if (is.null(trans) || length(trans) == 0) {
    return(NULL)
  }
  
  result <- list()
  
  for (param_name in names(trans)) {
    # Get all transformations for this parameter (excluding "null")
    trans_types <- unique(trans[[param_name]][trans[[param_name]] != "null"])
    
    if (length(trans_types) == 1) {
      # If only one transformation type is used, use that
      result[[param_name]] <- name2fun(trans_types)
    } else {
      # If multiple or no transformation types are used, use identity
      result[[param_name]] <- name2fun("identity")
    }
  }
  
  return(result)
}

# Calculate coefficients for each matrix based on parameters and transformations
compute_coef_new <- function(theta_K, trans, n_matrices) {
  coef <- rep(1, n_matrices)

  # If no parameters, return all ones (default behavior)
  if (length(theta_K) == 0) {
    return(coef)
  }

  for (param_name in names(theta_K)) {
    if (param_name %in% names(trans)) {
      for (i in seq_along(trans[[param_name]])) {
        trans_type <- trans[[param_name]][i]
        if (trans_type != "null") {
          # Apply the transformation function to the parameter value
          transformed_value <- name2fun(trans_type)(theta_K[param_name])
          # Multiply the coefficient by the transformed parameter value
          coef[i] <- coef[i] * transformed_value
        }
      }
    }
  }

  return(coef)
}

#' Convert transformation name to function
#'
#' This function converts a transformation name to the corresponding function.
#' Convertion is from original scale to real scale by default.
#' Available transformations:
#' \itemize{
#'   \item \code{exp4}: \eqn{exp(4x)}, inverse is \eqn{log(x)/4}
#'   \item \code{exp2}: \eqn{exp(2x)}, inverse is \eqn{log(x)/2}
#'   \item \code{tanh}: Hyperbolic tangent transformation used for AR1 parameter, uses ar1_th2a and ar1_a2th
#'   \item \code{identity}: Identity function, no transformation
#'   \item \code{exp}: \eqn{exp(x)}, inverse is \eqn{log(x)}
#'   \item \code{sqrt}: \eqn{sqrt(x)}, inverse is \eqn{x^2}
#'   \item \code{square}: \eqn{x^2}, inverse is \eqn{sqrt(x)}
#'   \item \code{log}: \eqn{log(x)}, inverse is \eqn{exp(x)}
#' }
#'
#' @param trans Character string specifying the transformation type from original scale to real scale
#' @param inv Logical, if TRUE returns the transformation function from real scale to original scale
#'
#' @return A function that applies the specified transformation
#' @export
name2fun <- function(trans, inv = FALSE) {
  # Define transformation functions in a list
  transformations <- list(
    exp4 = list(
      forward = function(x) exp(4 * x),
      inverse = function(x) log(x) / 4
    ),
    exp2 = list(
      forward = function(x) exp(2 * x),
      inverse = function(x) log(x) / 2
    ),
    tanh = list(
      forward = ar1_th2a,  # restrict the range of rho to (-1, 1)
      inverse = ar1_a2th
    ),
    identity = list(
      forward = function(x) x,
      inverse = function(x) x
    ),
    exp = list(
      forward = function(x) exp(x),
      inverse = function(x) log(x)
    ),
    sqrt = list(
      forward = function(x) sqrt(x),
      inverse = function(x) x^2
    ),
    square = list(
      forward = function(x) x^2,
      inverse = function(x) sqrt(x)
    ),
    log = list(
      forward = function(x) log(x),
      inverse = function(x) exp(x)
    )
  )
  
  # Check if transformation exists
  if (!trans %in% names(transformations)) {
    stop("Unknown transformation: ", trans)
  }
  
  # Return the appropriate function
  if (!inv) {
    return(transformations[[trans]]$forward)
  } else {
    return(transformations[[trans]]$inverse)
  }
}

test_that("generic model operator K works", {
  A <- matrix(1, 2, 2)
  B <- matrix(2, 2, 2)
  C <- matrix(3, 2, 2)
  D <- matrix(4, 2, 2)
  generic_model <- generic(
    theta_K = c(alpha = 1, beta = 2, gamma = 3),
    trans = c(alpha = "identity", beta = "exp", gamma = "exp2"),
    matrices = list(A, B, C, D),
    h = c(1, 1)
  )
  K <- generic_model$K
  expect_equal(as.matrix(K), 1 * A + exp(2) * B + exp(2 * 3) * C + D)
})

test_that("generic model == AR1 model", {
  # Test AR1 representation
  seed <- 10
  n_obs <- 10
  Y <- rnorm(n_obs)
  ar1 <- ar1(1:n_obs, rho = 0.5)
  g <- name2fun("tanh", inv = TRUE)
  ar1$param_name
  ar1$param_trans

  generic_ar1 <- generic(
    theta_K = c(x = g(0.5)), # trans(X) = rho
    trans = c(x = "tanh"),
    matrices = list(ar1$C, ar1$G),
    h = ar1$h,
    mesh = 1:n_obs
  )
  generic_ar1
  generic_ar1$param_name
  generic_ar1$param_trans

  expect_equal(ar1$K, ar1$C * 0.5 + ar1$G)
  expect_equal(generic_ar1$K, ar1$K)
  expect_equal(generic_ar1$matrices, list(ar1$C, ar1$G))

  control <- control_opt(
    seed = seed,
    iterations = 100,
    n_parallel_chain = 4,
    n_batch = 1
  )

  fit_ar1 <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      name = "my_ar",
      model = ar1(rho = 0.5),
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_ar1

  est_rho_ar1 <- ar1_th2a(ngme_result(fit_ar1, "my_ar")$rho)
  print(est_rho_ar1)
  traceplot(fit_ar1, "my_ar")

  fit_generic <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      name = "generic",
      model = generic(
        mesh = 1:n_obs,
        theta_K = c(rho = g(0.5)),
        trans = c(rho = "tanh"),
        matrices = list(ar1$C, ar1$G),
        h = ar1$h
      ),
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_generic

  est_rho_generic <- ar1_th2a(ngme_result(fit_generic, "generic")$rho)
  print(est_rho_generic)

  expect_equal(est_rho_generic[[1]][1], est_rho_ar1[[1]][1])
})

test_that("generic model == Matern model (alpha == 2)", {
  n_obs <- 10
  mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = 10))
  matern <- matern(mesh, kappa = exp(0.7))
  expect_equal(matern$K, matern$C * exp(2 * 0.7) + matern$G)

  generic_model_matern <- generic(
    mesh = mesh,
    theta_K = c(x = 0.7),
    trans = c(x = "exp2"),
    matrices = list(matern$C, matern$G),
    h = matern$h
  )

  expect_equal(generic_model_matern$K, matern$K)

  # Fitting the matern model
  control <- control_opt(
    seed = 10,
    iterations = 10,
    n_parallel_chain = 4,
    n_batch = 1,
  )

  Y <- rnorm(n_obs)
  fit_matern <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      model = matern(
        mesh = mesh,
        alpha = 2,
      ),
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_matern
  ngme_result(fit_matern, "field1")$kappa

  est_kappa_matern <- ngme_result(fit_matern, "field1")$kappa
  est_kappa_matern[[1]]

  fit_generic <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      model = generic(
        mesh = mesh,
        theta_K = c(theta = 0),
        trans = c(theta = "exp2"),
        matrices = list(matern$C, matern$G),
        h = matern$h
      ),
      name = "generic",
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )

  fit_generic
  theta <- ngme_result(fit_generic, "generic")$theta
})

test_that("generic model == Matern model (alpha == 4)", {
  n_obs <- 10
  x <- seq(0, 1, length.out = 10)
  y <- seq(0, 1, length.out = 10)
  mesh <- fmesher::fm_mesh_2d(cbind(x, y))

  kappa <- exp(0.7)
  matern <- matern(mesh, alpha = 4, theta_kappa = log(kappa))
  matern_2 <- matern(mesh, alpha = 4, kappa = kappa)

  # Build Cinv
  C <- matern$C
  G <- matern$G
  Cinv <- C
  diag(Cinv) <- 1 / Matrix::diag(C)

  expect_equal(matern$K, kappa^4 * C + 2 * kappa^2 * G + G %*% Cinv %*% G)

  generic_model_matern <- generic(
    theta_K = c(theta = log(kappa)),
    trans = list(theta = c("exp4", "exp2", "null")),
    matrices = list(C, 2 * G, G %*% Cinv %*% G),
    h = matern$h
  )
  expect_equal(generic_model_matern$symmetric, matern$symmetric)
  expect_equal(generic_model_matern$zero_trace, matern$zero_trace)
  expect_equal(generic_model_matern$h, matern$h)
  expect_equal(generic_model_matern$K, matern$K)

  # Fitting the matern model
  control <- control_opt(
    seed = 10,
    iterations = 100,
    n_parallel_chain = 4,
    n_batch = 1,
  )

  Y <- rnorm(n_obs)
  fit_matern <- ngme(
    Y ~ 0 + f(
      cbind(x, y),
      model = matern(mesh = mesh, alpha = 4)
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_matern
  est_theta_matern <- ngme_result(fit_matern, "field1")$kappa
  est_theta_matern[[1]]

  fit_generic <- ngme(
    Y ~ 0 + f(
      cbind(x, y),
      name = "generic",
      model = generic(
        mesh = mesh,
        theta_K = c(theta = 0),
        trans = list(theta = c("exp4", "exp2", "null")),
        matrices = list(C, 2 * G, G %*% Cinv %*% G),
        h = matern$h
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_generic
  est_theta_generic <- ngme_result(fit_generic, "generic")$theta
})

test_that("generic model == Matern model", {
  seed <- 10
  n_obs <- 5
  Y <- rnorm(n_obs)
  mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = 6))

  matern <- matern(
    mesh,
    theta_kappa = 0.7
  )
  g <- name2fun("exp2", inv = TRUE)
  matern$symmetric
  matern$zero_trace

  generic_matern <- generic(
    theta_K = c(x = 0.7),
    trans = c(x = "exp2"),
    matrices = list(matern$C, matern$G),
    h = matern$h,
    mesh = mesh
  )
  expect_equal(generic_matern$symmetric, matern$symmetric)
  expect_equal(generic_matern$zero_trace, matern$zero_trace)
  expect_equal(generic_matern$mesh, matern$mesh)
  expect_equal(generic_matern$h, matern$h)

  expect_equal(matern$K, matern$C * exp(2 * 0.7) + matern$G)
  expect_equal(generic_matern$K, matern$K)
  expect_equal(generic_matern$matrices, list(matern$C, matern$G))

  control <- control_opt(
    seed = seed,
    iterations = 100,
    n_parallel_chain = 4,
    n_batch = 1,
    # verbose = TRUE
  )

  fit_matern <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      model = matern(mesh = mesh, theta_kappa = 0.7)
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_matern
  est_kappa_matern <- ngme_result(fit_matern, "field1")$kappa
  traceplot(fit_matern, "field1")

  fit_generic <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      name = "generic",
      model = generic(
        theta_K = c(x = 0.7),
        trans = c(x = "exp2"),
        matrices = list(matern$C, matern$G),
        h = matern$h,
        mesh = mesh
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_generic
  est_x_generic <- ngme_result(fit_generic, "generic")$x
  expect_equal(est_x_generic[[1]], est_kappa_matern[[1]]^2)
})

test_that("generic model == RW1 model", {
  # Test AR1 representation
  seed <- 10
  n_obs <- 10
  Y <- rnorm(n_obs)
  rw1 <- rw1(1:n_obs)

  generic_rw1 <- generic(
    matrices = list(rw1$K),
    h = rw1$h,
    mesh = 1:n_obs
  )
  generic_rw1$K

  expect_equal(generic_rw1$K, rw1$K)
  # expect_equal(generic_rw1$matrices, list(rw1$K))

  control <- control_opt(
    seed = seed,
    iterations = 100,
    n_parallel_chain = 4,
    n_batch = 1
  )

  fit_rw1 <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      name = "my_rw1",
      model = rw1(),
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_rw1
  est_sigma_rw1 <- ngme_result(fit_rw1, "my_rw1")$noise$theta_sigma

  fit_generic <- ngme(
    Y ~ 0 + f(
      1:n_obs,
      name = "generic",
      model = generic(
        matrices = list(rw1$K),
        h = rw1$h
      )
    ),
    data = data.frame(Y = Y),
    control_opt = control
  )
  fit_generic
  est_sigma_generic <- ngme_result(fit_generic, "generic")$noise$theta_sigma

  expect_equal(est_sigma_generic[[1]], est_sigma_rw1[[1]])
})

test_that("test compute coef", {
  # Create 3 random matrices for testing
  set.seed(123)
  n <- 5

  # Create random matrices with different patterns
  mat1 <- matrix(rnorm(n * n), nrow = n)
  mat2 <- matrix(runif(n * n), nrow = n)
  mat3 <- matrix(rpois(n * n, lambda = 2), nrow = n)

  # Test with the new compute_coef_new function
  theta_K <- c(param1 = 2, param2 = 3, param3 = 4)

  # Create trans list in the new format
  trans <- list(
    param1 = c("identity", "null", "null"),
    param2 = c("null", "identity", "null"),
    param3 = c("null", "null", "identity")
  )

  coef <- compute_coef_new(theta_K, trans, 3)
  expect_equal(coef, c(2, 3, 4))

  # Test with mixed transformations
  trans2 <- list(
    param1 = c("exp", "null", "null"),
    param2 = c("null", "identity", "null"),
    param3 = c("null", "null", "exp2")
  )

  coef2 <- compute_coef_new(theta_K, trans2, 3)
  expect_equal(coef2, c(exp(2), 3, exp(2 * 4)))

  # Test with a more complex case where parameters affect multiple matrices
  trans3 <- list(
    param1 = c("identity", "identity", "null"),
    param2 = c("null", "identity", "identity")
  )

  theta_K2 <- c(param1 = 2, param2 = 3)

  coef3 <- compute_coef_new(theta_K2, trans3, 3)
  expect_equal(coef3, c(2, 2 * 3, 3))
})

test_that("generic model with new trans format works", {
  # Create test matrices
  A <- matrix(1, 3, 3)
  B <- matrix(2, 3, 3)
  C <- matrix(3, 3, 3)
  D <- matrix(4, 3, 3)

  # Set parameters
  theta_K <- c(a = 1, b = 2, c = 3)

  # Define transformations for each parameter across matrices
  trans <- list(
    a = c("identity", "exp", "null", "null"),
    b = c("null", "identity", "null", "null"),
    c = c("null", "null", "identity", "null")
  )

  # Create the generic model with the new format
  generic_model <- generic(
    theta_K = theta_K,
    trans = trans,
    matrices = list(A, B, C, D),
    h = c(1, 1, 1)
  )

  # Calculate expected K manually
  # K = a*A + b*exp(a)*B + c*C + D
  expected_K <- 1 * A + 2 * exp(1) * B + 3 * C + D

  expect_equal(as.matrix(generic_model$K), expected_K)

  # Test with different transformations
  trans2 <- list(
    a = c("identity", "exp2", "null", "null"),
    b = c("null", "null", "tanh", "null"),
    c = c("null", "null", "null", "exp")
  )

  generic_model2 <- generic(
    theta_K = theta_K,
    trans = trans2,
    matrices = list(A, B, C, D),
    h = c(1, 1, 1)
  )

  # Calculate expected K manually with new transformations
  # K = a*A + exp(2*a)*B + tanh(b)*C + exp(c)*D
  expected_K2 <- 1 * A + exp(2 * 1) * B + ar1_th2a(2) * C + exp(3) * D

  expect_equal(as.matrix(generic_model2$K), expected_K2)

  # Test default behavior with no trans provided
  generic_model3 <- generic(
    matrices = list(A, B, C, D),
    h = c(1, 1, 1)
  )

  # By default, should be the sum of all matrices with coefficient 1
  expected_K3 <- A + B + C + D

  expect_equal(as.matrix(generic_model3$K), expected_K3)
})

test_that("get_param_trans correctly extracts transformation functions", {
  # Case 1: Each parameter has exactly one transformation
  trans1 <- list(
    param1 = c("exp", "null", "null"),
    param2 = c("null", "identity", "null"),
    param3 = c("null", "null", "exp2")
  )

  result1 <- get_param_trans(trans1)
  expect_equal(result1$param1(2), exp(2))
  expect_equal(result1$param2(3), 3)
  expect_equal(result1$param3(4), exp(2 * 4))

  # Case 2: Some parameters have multiple transformations
  trans2 <- list(
    a = c("identity", "exp", "null", "null"),
    b = c("null", "identity", "null", "null"),
    c = c("null", "null", "identity", "null")
  )

  result2 <- get_param_trans(trans2)
  expect_equal(result2$a(5), 5) # Should use identity
  expect_equal(result2$b(6), 6)
  expect_equal(result2$c(7), 7)

  # Case 3: Mixed cases including no transformations
  trans3 <- list(
    x = c("null", "null", "null", "null"),
    y = c("exp2", "null", "null", "null"),
    z = c("identity", "tanh", "null", "null")
  )

  result3 <- get_param_trans(trans3)
  expect_equal(result3$x(8), 8) # Should use identity
  expect_equal(result3$y(2), exp(2 * 2))
  expect_equal(result3$z(9), 9) # Should use identity
})

#############################################
#' Argo float dataset
#' @description Argo floats measurements.
#' @docType data
#' @usage data("argo_float")
#' @format Data frame containing 274 observations on 4 variables.
#' \describe{
#'   \item{lat}{Latitude.}
#'   \item{lon}{Longitude.}
#'   \item{sal}{Salinity.}
#'   \item{temp}{Temperature.}
#' }
#'
#' @details
#' The floats have a pressure case made of aluminium that is about 1.3m long and about 20cm diameter. They weigh about 40kg. On the top is an antenna to communicate with the satellites that fix the float's position and receive the data. Also on the top are the temperature, salinity and pressure sensors.
#'
#' @source Data can be obtained from \href{https://argo.ucsd.edu/}{Argo float website}.
"argo_float"


#' The swamp of Cienaga Grande in Santa Marta, Colombia
#'
#'There is a total of 114 locations where some properties of the
#' swamp were measured. Those measurements were taken twice, however there is no information available about their chronological order so this data cannot
#' be treated as spatiotemporal, despite that, the multiple measurements can be treated as replicates.
#'
#' @format A data frame with 218 rows and 6 columns.
#' \describe{
#'   \item{East, North}{location}
#'   \item{depth}{depth of the swamp}
#'   \item{temp}{temperature}
#'   \item{oxyg}{oxygen}
#'   \item{measurement}{1 means the first measurement, 2 the second}
#' }
#' @source ..
"cienaga"

#' The x y location of the border of the swamp of Cienaga Grande in Santa Marta, Colombia
#'
#' The data is of dimension 472 * 2. It contains the x and y coordinates of the border of the swamp.
#'
#' @format A data frame with 472 rows and 2 columns.
#' \describe{
#'   \item{East, North}{location}
#' }
#' @source ..
"cienaga.border"

# Internal helpers for chain-wise post-processing

strip_ngme_chain_attributes <- function(object) {
  attr(object, "chain_outputs") <- NULL
  attr(object, "chain_fits") <- NULL
  object
}


build_ngme_from_chain_output <- function(object, chain_output) {
  stopifnot(
    inherits(object, "ngme"),
    is.list(chain_output)
  )

  chain_fit <- object
  stopifnot(length(chain_fit$replicates) == length(chain_output))

  for (i in seq_along(chain_fit$replicates)) {
    chain_fit$replicates[[i]] <- update_ngme_est(
      chain_fit$replicates[[i]],
      chain_output[[i]]
    )
  }

  strip_ngme_chain_attributes(chain_fit)
}


get_ngme_chain_fits <- function(object) {
  stopifnot(inherits(object, "ngme"))

  chain_fits <- attr(object, "chain_fits")
  if (is.list(chain_fits) && length(chain_fits) > 0) {
    return(chain_fits)
  }

  chain_outputs <- attr(object, "chain_outputs")
  if (!is.list(chain_outputs) || length(chain_outputs) == 0) {
    return(list(object))
  }

  fits <- lapply(chain_outputs, function(chain_output) {
    build_ngme_from_chain_output(object, chain_output)
  })

  fits
}

# ------- CV -------
#' Compute the cross-validation for the ngme model
#' Perform cross-validation for ngme model
#' first into sub_groups (a list of target, and train data)
#'
#' @param ngme a ngme object, or a list of ngme object (if comparing multiple models)
#' @param type character, in c("k-fold", "loo", "lpo", "custom")
#' k-fold is k-fold cross-validation, provide \code{k}
#' loo is leave-one-out,
#' lpo is leave-percent-out, provide \code{percent} from 1 to 100
#' custom is user-defined group, provide \code{target} and \code{data}
#' @param seed random seed
#' @param N_sim integer, number of simulations (e.g., estimate MAE, MSE, .. N times)
#' @param k integer (only for k-fold type)
#' @param print print information during computation
#' @param percent how many percent for testing? from 0 to 1 (for lpo type)
#' @param times how many test cases (only for lpo type)
#' @param metric Optional function or list of functions (one per model) that maps
#'   the group-wise observations/predictions for a single location to the quantity
#'   that should be scored. The function receives a list containing at least
#'   \code{y} (a named numeric vector of group values) and may optionally use
#'   \code{samples1} and \code{samples2} (matrices with rows named by group and
#'   columns indexing posterior draws). The function must return either a numeric
#'   scalar (optionally named), or a list with component \code{y} and optional
#'   components \code{samples1}, \code{samples2}, and \code{label}. When
#'   \code{NULL}, the original per-group scores are computed. Example:
#'   \code{metric = function(data) 2 * data$y["A"] + data$y["B"]}. To sum all
#'   groups, return \code{sum(data$y)}.
#' @param n_gibbs_samples number of gibbs samples of latent process, used for computing CRPS, sCRPS
#' @param n_burnin number of burnin
#' @param test_idx a list of indices of the data (which data points to be predicted) (only for custom type)
#' @param train_idx  a list of indices of the data (which data points to be used for re-sampling (not re-estimation)) (only for custom type)
#' @param keep_pred logical, keep test information (pred_1, pred_2) in the return (as attributes), pred_1 and pred_2 are the prediction of the two chains
#' @param thining_gap integer, the gap between samples for thinning, if 0, then no thinning, if 1, then keep 50\% of the samples for CRPS, sCRPS, etc.
#' @param parallel logical, run in parallel mode
#' @param cores_layer1 integer, number of cores for the first layer (over testing samples)
#' @param cores_layer2 integer, number of cores for the second layer (over computing scores for N_sim simulations)
#' @param merge_groups logical, if TRUE, merge groups as vector components (e.g., for vector-valued wind data with north_wind, east_wind).
#' MAE becomes Euclidean distance, MSE becomes squared Euclidean distance, etc.
#' @param merged_group_name character, name for the merged group when merge_groups=TRUE.
#' If NULL, uses "group1_group2" format (default: NULL)
#' @param data optional data.frame used to replace the original fitting data before running CV.
#' If `NULL`, the data stored in `ngme` is used. If provided, the model is rebuilt on `data`
#' while reusing fitted parameters from `ngme`.
#' @param chain_combine how to combine multiple optimization chains:
#'   `"param_mean"` uses the fitted object directly (default), while
#'   `"predictive_average"` computes predictions from each optimization chain
#'   and averages at the predictive level.
#' @return A list with components:
#'   \describe{
#'     \item{mean.scores}{Mean of the \code{N_sim} estimates for MSE, MAE, CRPS,
#'       and sCRPS.}
#'     \item{sd.scores}{Standard deviation of the \code{N_sim} estimates for MSE,
#'       MAE, CRPS, and sCRPS.}
#'   }
#' @export
cross_validation <- function(
    ngme,
    type = "k-fold",
    seed = NULL,
    print = TRUE,
    N_sim = 5,
    n_gibbs_samples = 500,
    n_burnin = 100,
    k = 5,
    percent = 0.2,
    times = 10,
    metric = NULL,
    test_idx = NULL,
    train_idx = NULL,
    keep_pred = FALSE,
    parallel = FALSE,
    thining_gap = 1, # Used for computing CRPS, sCRPS, the gap between samples for thinning, if 0, then no thinning, if 1, then keep 50% of the samples for CRPS, sCRPS, etc.
    # merge_replicates = FALSE, # remove this option
    cores_layer1 = if (parallel) min(parallel::detectCores(), 2) else 1, # Limit to 2 cores for safety
    cores_layer2 = if (parallel) min(parallel::detectCores(), 2) else 1, # Limit to 2 cores for safety
    merge_groups = FALSE, # New parameter for merging bivariate groups
    merged_group_name = NULL, # New parameter for custom merged group name
    data = NULL,
    chain_combine = c("param_mean", "predictive_average")
    ) {
  chain_combine <- match.arg(chain_combine)
  merge_replicates <- FALSE

  if (!requireNamespace("parallel", quietly = TRUE)) {
    message("Parallel package not available. Running in serial mode. You can install `parallel` package to speed up the computation.")
    parallel <- FALSE
  }

  if (inherits(ngme, "ngme")) ngme <- list(ngme)
  if (is.null(names(ngme))) names(ngme) <- paste("model", seq_along(ngme), sep = "_")

  if (!is.null(data)) {
    stopifnot("`data` must be a data.frame." = is.data.frame(data))
  }

  ngme_chain_sets <- NULL
  if (chain_combine == "predictive_average") {
    ngme_chain_sets <- lapply(ngme, function(model_fit) {
      chain_models <- resolve_ngme_chain_fits(model_fit)
      if (!is.null(data)) {
        chain_models <- lapply(chain_models, rebuild_cv_model_with_data, data = data)
      }
      chain_models
    })
    ngme <- lapply(ngme_chain_sets, function(chain_models) chain_models[[1]])
  } else if (!is.null(data)) {
    ngme <- lapply(ngme, rebuild_cv_model_with_data, data = data)
  }

  # Handle metric argument: allow NULL, function, or list of functions (one per model)
  if (is.null(metric)) {
    metric <- vector("list", length(ngme))
  } else if (is.list(metric)) {
    if (length(metric) != length(ngme)) {
      stop("If metric is a list, its length must equal number of models (length(ngme)).")
    }
  } else if (is.function(metric)) {
    metric <- rep(list(metric), length(ngme))
  } else {
    stop("metric must be NULL, a function, or a list of functions")
  }

  metric_supplied <- any(vapply(metric, function(fun) !is.null(fun), logical(1)))
  if (metric_supplied && merge_groups) {
    stop("merge_groups cannot be used together with a custom metric; please encode the aggregation inside the metric function.")
  }

  n_data <- attr(ngme[[1]], "fit")$n_data
  if (is.null(n_data)) stop("Please provide ngme object or a list of ngme object")

  if (!is.null(seed)) set.seed(seed)
  stopifnot(
    "type should be in c('k-fold', 'loo', 'lpo', 'custom')" = type %in% c("k-fold", "loo", "lpo", "custom")
  )

  # 1. compute indices of tartget and train if not custom type
  if (type == "k-fold") {
    # Check if we should use paired CV splits for grouped data
    if (merge_groups && !is.null(ngme[[1]]$group)) {
      stop("Paired CV splits are not supported for grouped data. Please use custom type. Check ?create_paired_cv_splits for more details.")
    } else {
      # Regular k-fold split
      idx <- seq_len(n_data)
      folds <- cut(sample(idx), breaks = k, label = FALSE)
      test_idx <- lapply(1:k, function(x) {
        which(folds == x, arr.ind = TRUE)
      })
      train_idx <- lapply(1:k, function(x) {
        which(folds != x, arr.ind = TRUE)
      })
    }
  } else if (type == "loo") {
    idx <- seq_len(n_data)
    test_idx <- lapply(idx, function(i) i)
    train_idx <- lapply(idx, function(i) idx[idx != i])
  } else if (type == "lpo") {
    stopifnot(
      "percent should be between 0 and 1" = percent > 0 && percent < 1,
      "times should be a positive integer" = times > 0
    )
    for (i in 1:times) {
      test_idx[[i]] <- sample(1:n_data, size = percent * n_data)
      train_idx[[i]] <- setdiff(1:n_data, test_idx[[i]])
    }
  } else {
    # check if test_idx and train_idx is provided and of same length
    stopifnot(
      "test_idx and train_idx should be provided" = !is.null(test_idx) && !is.null(train_idx),
      "test_idx and train_idx should be a list" = is.list(test_idx) && is.list(train_idx),
      "test_idx and train_idx should be of same length" = length(test_idx) == length(train_idx)
    )
  }
  # Alternative. do not distinguish between replicates?
  # But the internal mesh may not be the same for each replicate....

  # 2. loop over each test_idx and train_idx, and compute the criterion
  final_mean <- final_sd <- list()
  pred_1 <- list()
  pred_2 <- list()
  Y_1 <- list()
  Y_2 <- list()

  compute_err <- if (merge_replicates) compute_err_merged_reps else compute_err_reps
  cv_message <- function(...) {
    if (isTRUE(print)) {
      message(...)
    }
  }
  cv_message_table <- function(x) {
    if (isTRUE(print)) {
      message(paste(capture.output(base::print(as.data.frame(x))), collapse = "\n"))
    }
  }

  for (idx in seq_along(ngme)) {
    cv_message("Model ", names(ngme)[[idx]], ":")
    scores <- sd_scores <- NULL
    chain_models <- if (is.null(ngme_chain_sets)) NULL else ngme_chain_sets[[idx]]

    # loop over each test_idx and train_idx
    if (parallel && requireNamespace("parallel", quietly = TRUE)) {
      cv_message("Running in parallel mode.")

      ngme_list <- list()
      for (i in 1:length(test_idx)) {
        ngme_list[[i]] <- ngme[[idx]]
      }

      # Add error handling for parallel execution
      tryCatch(
        {
          ret <- parallel::mclapply(
            seq_along(test_idx),
            function(i) {
              tryCatch(
                {
                  result <- compute_err(
                    ngme_list[[i]],
                    test_idx[[i]],
                    train_idx[[i]],
                    N_sim = N_sim,
                    n_gibbs_samples = n_gibbs_samples,
                    seed = seed,
                    keep_pred = keep_pred,
                    parallel = TRUE,
                    metric = metric[[idx]],
                    num_cores = cores_layer2,
                    thining_gap = thining_gap,
                    merge_groups = merge_groups,
                    merged_group_name = merged_group_name,
                    chain_models = chain_models,
                    chain_combine = chain_combine
                  )
                  cv_message("In test batch ", i, ":")
                  cv_message_table(result$scores)
                  result
                },
                error = function(e) {
                  warning(paste("Error in test batch", i, ":", e$message))
                  NULL
                }
              )
            },
            mc.cores = cores_layer1
          )

          # Filter out NULL results
          ret <- ret[!sapply(ret, is.null)]

          if (length(ret) == 0) {
            warning("All parallel attempts failed, falling back to sequential computation")
            parallel <- FALSE
          } else {
            scores <- lapply(ret, function(x) x$scores)
            sd_scores <- lapply(ret, function(x) x$sd_scores)
          }
        },
        error = function(e) {
          warning(paste("Error in parallel execution:", e$message))
          parallel <- FALSE
        }
      )
    }

    # Fall back to sequential if parallel failed or was not requested
    if (!parallel || length(scores) == 0) {
      for (i in seq_along(test_idx)) {
        tryCatch(
          {
            result <- compute_err(
              ngme[[idx]],
              test_idx[[i]],
              train_idx[[i]],
              N_sim = N_sim,
              n_gibbs_samples = n_gibbs_samples,
              seed = seed,
              keep_pred = keep_pred,
              parallel = FALSE,
              metric = metric[[idx]],
              thining_gap = thining_gap,
              merge_groups = merge_groups,
              merged_group_name = merged_group_name,
              chain_models = chain_models,
              chain_combine = chain_combine
            )
            scores[[i]] <- result$scores
            sd_scores[[i]] <- result$sd_scores

            cv_message("In test batch ", i, ":")
            cv_message_table(scores[[i]])
          },
          error = function(e) {
            warning(paste("Error in sequential computation for test batch", i, ":", e$message))
          }
        )
      }
    }

    # Check if we have any valid scores
    if (length(scores) == 0) {
      warning(paste("Failed to compute any valid scores for model", names(ngme)[[idx]]))
      next
    }

    final_mean[[idx]] <- as.data.frame(mean_list(scores))
    final_sd[[idx]] <- as.data.frame(mean_list(sd_scores))
  }

  # Check if we have any results
  if (length(final_mean) == 0) {
    stop("Failed to compute any valid scores for any model")
  }

  # convert to data.frame
  final_mean <- do.call(rbind, final_mean)
  final_sd <- do.call(rbind, final_sd)

  # Check if merging actually succeeded by looking at the number of rows
  # If merge_groups=TRUE but merging failed, we'll have 2 rows per model instead of 1
  n_models <- length(ngme)
  merging_succeeded <- merge_groups && (nrow(final_mean) == n_models)

  if (merging_succeeded) {
    # For successfully merged groups, use simplified naming
    rownames(final_mean) <- names(ngme)
    rownames(final_sd) <- names(ngme)
  } else if (nrow(final_mean) == n_models) {
    # univariate model (1 row per model)
    rownames(final_mean) <- names(ngme)
    rownames(final_sd) <- names(ngme)
  } else {
    # bivariate model (2 rows per model)
    if (nrow(final_mean) > 0) {
      n_groups_per_model <- nrow(final_mean) / n_models
      if (n_groups_per_model == 2) {
        # Extract group names from the first set of results
        group_names <- rownames(final_mean)[1:2]
        names_list <- lapply(names(ngme), function(x) paste(x, group_names, sep = "_"))
        rownames(final_mean) <- do.call(c, names_list)
        rownames(final_sd) <- rownames(final_mean)
      }
    }
  }

  ret <- list(
    mean.scores = final_mean,
    sd.scores = final_sd
  )

  if (!keep_pred) {
    invisible(ret)
  } else {
    structure(
      invisible(ret),
      train_idx = train_idx,
      test_idx = test_idx,
      pred_1 = pred_1,
      pred_2 = pred_2,
      Y_1 = Y_1,
      Y_2 = Y_2
    )
  }
}


resolve_ngme_chain_fits <- function(object) {
  stopifnot(inherits(object, "ngme"))

  if (exists("get_ngme_chain_fits", mode = "function", inherits = TRUE)) {
    return(get_ngme_chain_fits(object))
  }

  chain_outputs <- attr(object, "chain_outputs")
  if (!is.list(chain_outputs) || length(chain_outputs) == 0) {
    return(list(object))
  }

  lapply(chain_outputs, function(chain_output) {
    chain_fit <- object
    if (!is.list(chain_output) || length(chain_fit$replicates) != length(chain_output)) {
      return(object)
    }

    for (i in seq_along(chain_fit$replicates)) {
      chain_fit$replicates[[i]] <- update_ngme_est(
        chain_fit$replicates[[i]],
        chain_output[[i]]
      )
    }

    attr(chain_fit, "chain_outputs") <- NULL
    attr(chain_fit, "chain_fits") <- NULL
    chain_fit
  })
}


rebuild_cv_model_with_data <- function(ngme_obj, data) {
  stopifnot("`ngme_obj` must be of class 'ngme'." = inherits(ngme_obj, "ngme"))
  fit <- attr(ngme_obj, "fit")
  stopifnot(
    "Missing fit metadata in `ngme_obj`." = !is.null(fit),
    "Missing formula in fit metadata." = !is.null(fit$formula)
  )

  replicate_new <- resolve_cv_partition_for_new_data(
    old_data = fit$data,
    old_values = fit$replicate,
    new_data = data,
    arg_name = "replicate"
  )

  group_new <- resolve_cv_partition_for_new_data(
    old_data = fit$data,
    old_values = fit$group,
    new_data = data,
    arg_name = "group"
  )

  standardize_fixed <- if (is.null(ngme_obj$replicates[[1]]$standardize)) {
    TRUE
  } else {
    isTRUE(ngme_obj$replicates[[1]]$standardize)
  }

  build_args <- list(
    formula = fit$formula,
    data = data,
    family = fit$family,
    control_opt = control_opt(
      estimation = FALSE,
      standardize_fixed = standardize_fixed
    ),
    control_ngme = ngme_obj$replicates[[1]]$control_ngme,
    group = group_new,
    replicate = replicate_new
  )
  rebuild_env <- build_cv_rebuild_env(
    formula = fit$formula,
    ngme_obj = ngme_obj,
    data = data
  )

  tryCatch(
    {
      do.call("ngme", c(build_args, list(start = ngme_obj)), envir = rebuild_env)
    },
    error = function(e) {
      if (grepl("length of [WV] should be the same", e$message)) {
        # Fallback for new data with different latent/noise state dimension:
        # rebuild without start and transplant only compatible hyperparameters.
        rebuilt <- do.call("ngme", build_args, envir = rebuild_env)
        rebuilt <- transplant_cv_hyperparameters(rebuilt, ngme_obj)
        return(rebuilt)
      }
      stop(
        "Failed to rebuild ngme object with supplied `data`: ",
        e$message,
        call. = FALSE
      )
    }
  )
}


build_cv_rebuild_env <- function(formula, ngme_obj, data) {
  env <- new.env(parent = parent.frame())
  env$ngme <- ngme

  formula_env <- environment(formula)
  symbols <- setdiff(unique(all.vars(formula)), names(data))
  if (length(symbols) == 0) return(env)
  if (length(ngme_obj$replicates) == 0) return(env)
  if (length(ngme_obj$replicates[[1]]$models) == 0) return(env)

  rep1 <- ngme_obj$replicates[[1]]
  model1 <- rep1$models[[1]]
  operator <- model1$operator
  model_noise <- model1$noise

  for (sym in symbols) {
    if (exists(sym, envir = env, inherits = TRUE)) next
    val <- resolve_cv_symbol_binding(
      sym = sym,
      formula_env = formula_env,
      operator = operator,
      model_noise = model_noise
    )
    if (!is.null(val)) assign(sym, val, envir = env)
  }

  env
}


resolve_cv_symbol_binding <- function(sym, formula_env, operator, model_noise) {
  if (is.environment(formula_env) && exists(sym, envir = formula_env, inherits = TRUE)) {
    return(get(sym, envir = formula_env, inherits = TRUE))
  }

  if (sym == "mesh" && !is.null(operator$mesh)) return(operator$mesh)
  if (sym %in% c("B", "B_kappa", "B_K") && !is.null(operator$B_K)) return(operator$B_K)
  if (sym %in% c("B_sigma", "B_mu", "B_nu") && !is.null(model_noise[[sym]])) return(model_noise[[sym]])
  if (grepl("^n_?basis$", sym) && !is.null(operator$B_K)) return(ncol(operator$B_K))

  NULL
}


transplant_cv_hyperparameters <- function(new_obj, old_obj) {
  stopifnot(
    inherits(new_obj, "ngme"),
    inherits(old_obj, "ngme")
  )

  n_repl <- min(length(new_obj$replicates), length(old_obj$replicates))
  if (n_repl == 0) return(new_obj)

  for (i in seq_len(n_repl)) {
    rep_new <- new_obj$replicates[[i]]
    rep_old <- old_obj$replicates[[i]]

    if (length(rep_new$feff) == length(rep_old$feff)) {
      rep_new$feff <- as.numeric(rep_old$feff)
      names(rep_new$feff) <- names(new_obj$replicates[[i]]$feff)
    }

    rep_new$noise <- transplant_noise_hyperparameters(rep_new$noise, rep_old$noise)

    n_models <- min(length(rep_new$models), length(rep_old$models))
    if (n_models > 0) {
      for (j in seq_len(n_models)) {
        model_new <- rep_new$models[[j]]
        model_old <- rep_old$models[[j]]

        if (!identical(model_new$model, model_old$model)) next

        theta_new <- model_new$operator$theta_K
        theta_old <- model_old$operator$theta_K

        if (!is.null(names(theta_new)) && !is.null(names(theta_old))) {
          common <- intersect(names(theta_new), names(theta_old))
          if (length(common) > 0) theta_new[common] <- theta_old[common]
        } else if (length(theta_new) == length(theta_old)) {
          theta_new <- theta_old
        }

        model_new$operator$theta_K <- theta_new
        model_new$theta_K <- theta_new
        model_new$operator$K <- model_new$operator$update_K(theta_new)

        model_new$noise <- transplant_noise_hyperparameters(model_new$noise, model_old$noise)
        rep_new$models[[j]] <- model_new
      }
    }

    new_obj$replicates[[i]] <- rep_new
  }

  new_obj
}


transplant_noise_hyperparameters <- function(noise_new, noise_old) {
  old_no_state <- noise_old
  old_no_state$V <- NULL
  update_noise(noise_new, new_noise = old_no_state)
}


resolve_cv_partition_for_new_data <- function(old_data, old_values, new_data, arg_name) {
  if (is.null(old_values)) return(NULL)

  if (length(old_values) == nrow(new_data)) {
    return(old_values)
  }

  source_col <- infer_cv_partition_column(old_data, old_values, arg_name)
  if (!is.null(source_col) && source_col %in% names(new_data)) {
    values <- new_data[[source_col]]
    if (is.factor(old_values)) {
      old_levels <- levels(old_values)
      new_levels <- unique(as.character(values[!is.na(values)]))
      ordered_levels <- c(intersect(old_levels, new_levels), setdiff(new_levels, old_levels))
      values <- factor(values, levels = ordered_levels)
    }
    if (anyNA(values)) {
      stop(
        "Detected NA while preparing `", arg_name, "` from `data$",
        source_col, "`. Please ensure there are no missing values."
      )
    }
    return(values)
  }

  if (length(unique(as.character(old_values))) == 1) {
    return(rep(old_values[1], nrow(new_data)))
  }

  stop(
    "Unable to infer `", arg_name, "` for supplied `data`. ",
    "Please include the same `", arg_name, "` column used during fitting."
  )
}


infer_cv_partition_column <- function(old_data, old_values, arg_name) {
  if (!is.data.frame(old_data) || nrow(old_data) != length(old_values)) {
    return(NULL)
  }

  target <- as.character(old_values)

  if (arg_name %in% names(old_data) &&
      identical(as.character(old_data[[arg_name]]), target)) {
    return(arg_name)
  }

  matched_cols <- names(old_data)[vapply(old_data, function(col) {
    length(col) == length(target) && identical(as.character(col), target)
  }, logical(1))]

  if (length(matched_cols) == 0) return(NULL)
  matched_cols[[1]]
}

# Compute error with merged 1 replicate using helper function merge_replicates
compute_err_merged_reps <- function(
    ngme,
    test_idx,
    train_idx,
    N_sim,
    n_gibbs_samples = 100,
    n_burnin = 100,
    seed = NULL,
    keep_pred = FALSE,
    parallel = TRUE,
    metric = NULL,
    num_cores = 1,
    thining_gap = 0,
    merge_groups = FALSE,
    merged_group_name = NULL,
    chain_models = NULL,
    chain_combine = "param_mean") {
  if (is.null(seed)) seed <- Sys.time()
  stopifnot("Not a ngme object." = inherits(ngme, "ngme"))

  test_idx <- sort(test_idx)
  repls <- attr(ngme, "fit")$replicate
  uni_repl <- unique(repls)

  scores <- sd_scores <- weight <- NULL
  n_scores <- 0
  pred_1 <- double(length = length(test_idx))
  pred_2 <- double(length = length(test_idx))
  Y_1 <- double(length = length(test_idx))
  Y_2 <- double(length = length(test_idx))

  ret <- merge_replicates(ngme, train_idx, test_idx)
  merged_rep <- ret$merged_rep

  scores <- list()
  for (sim_n in 1:N_sim) {
    scores[[sim_n]] <- compute_scores(
      merged_rep,
      n_gibbs_samples,
      n_burnin,
      seed + sim_n,
      ret$test_A_block,
      ret$test_noise,
      ret$test_Y,
      ret$test_group,
      ret$test_X,
      metric,
      thining_gap = thining_gap,
      merge_groups = merge_groups,
      merged_group_name = merged_group_name,
      chain_combine = chain_combine
    )
  }

  # compute mean and sd of scores
  n_result_rows <- nrow(scores[[1]])
  n_cols <- ncol(scores[[1]])
  array_3d <- array(unlist(scores),
    dim = c(n_result_rows, n_cols, length(scores))
  )

  mean_array <- apply(array_3d, c(1, 2), mean)
  sd_array <- apply(array_3d, c(1, 2), sd)

  colnames(mean_array) <- colnames(sd_array) <- names(scores[[1]])
  rownames(mean_array) <- rownames(sd_array) <- rownames(scores[[1]])

  list(
    scores = mean_array,
    sd_scores = sd_array
  )
}



# Compute error with multiple replicates using helper function compute_err_1rep
compute_err_reps <- function(
    ngme,
    test_idx,
    train_idx,
    N_sim,
    n_gibbs_samples = 100,
    n_burnin = 100,
    seed = NULL,
    keep_pred = FALSE,
    parallel = TRUE,
    metric = NULL,
    num_cores = 1,
    thining_gap = 1,
    merge_groups = FALSE,
    merged_group_name = NULL,
    chain_models = NULL,
    chain_combine = "param_mean") {
  test_idx <- sort(test_idx)
  stopifnot("Not a ngme object." = inherits(ngme, "ngme"))
  repls <- attr(ngme, "fit")$replicate
  uni_repl <- unique(repls)

  scores <- sd_scores <- weight <- NULL
  n_scores <- 0
  pred_1 <- double(length = length(test_idx))
  pred_2 <- double(length = length(test_idx))
  Y_1 <- double(length = length(test_idx))
  Y_2 <- double(length = length(test_idx))

  for (i in seq_along(uni_repl)) {
    data_idx_rep <- ngme$replicates[[i]]$data_idx
    bool_train_idx <- data_idx_rep %in% train_idx # current rep has train
    bool_test_idx <- data_idx_rep %in% test_idx # current rep has test

    # skip this replicate if no target or train data
    if (sum(bool_train_idx) == 0 || sum(bool_test_idx) == 0) next
    n_scores <- n_scores + 1

    ngme_1rep <- ngme$replicates[[i]]
    ngme_chain_reps <- NULL
    if (chain_combine == "predictive_average" &&
        !is.null(chain_models) &&
        length(chain_models) > 1) {
      ngme_chain_reps <- lapply(chain_models, function(chain_fit) {
        chain_fit$replicates[[i]]
      })
    }
    result_1rep <- compute_err_1rep(
      ngme_1rep,
      bool_train_idx = bool_train_idx,
      bool_test_idx = bool_test_idx,
      N_sim = N_sim,
      n_gibbs_samples = n_gibbs_samples,
      seed = seed,
      keep_pred = keep_pred,
      parallel = parallel,
      metric = metric,
      num_cores = num_cores,
      thining_gap = thining_gap,
      merge_groups = merge_groups,
      merged_group_name = merged_group_name,
      ngme_chain_reps = ngme_chain_reps,
      chain_combine = chain_combine
    )
    scores[[n_scores]] <- result_1rep$mean_scores
    sd_scores[[n_scores]] <- result_1rep$sd_scores

    # Assume which_idx_pred ordered (order test idx)
    which_idx_pred <- data_idx_rep[bool_test_idx]

    if (keep_pred) {
      pred_1[test_idx %in% which_idx_pred] <- result_1rep$pred_1
      pred_2[test_idx %in% which_idx_pred] <- result_1rep$pred_2
      Y_1[test_idx %in% which_idx_pred] <- result_1rep$Y_1
      Y_2[test_idx %in% which_idx_pred] <- result_1rep$Y_2
    }

    which_repl <- which(repls == uni_repl[i])
    weight <- c(weight, length(which_repl))
  }

  # take weighted average over replicates
  list(
    scores = mean_list(scores, weight),
    sd_scores = mean_list(sd_scores, weight),
    pred_1 = pred_1,
    pred_2 = pred_2,
    Y_1 = Y_1,
    Y_2 = Y_2
  )
}


# helper function to compute MSE, MAE, ... for each subset of target / data
# assume test_idx and train_idx belongs to same replicate
##
# A custom metric function can be supplied to combine group-wise observations and predictions into a single quantity before scoring.
compute_err_1rep <- function(
    ngme_1rep,
    bool_test_idx,
    bool_train_idx,
    N_sim,
    n_gibbs_samples = 50,
    n_burnin = 10,
    seed = NULL,
    keep_pred = FALSE,
    parallel = TRUE,
    metric = NULL,
    num_cores = 1, # Default to 1 core to avoid potential issues
    thining_gap = 1,
    merge_groups = FALSE,
    merged_group_name = NULL,
    ngme_chain_reps = NULL,
    chain_combine = "param_mean") {
  stopifnot(
    "bool_<..>_idx should be a logical vector" =
      is.logical(bool_test_idx) && is.logical(bool_train_idx)
  )
  # if test_idx and train_idx overlap, warning message
  if (sum(bool_test_idx & bool_train_idx) > 0) {
    warning("Notice that test_idx and train_idx overlap!")
  }

  # Since we revert the order of Y, now we need to
  # revert the train and test idx to match
  # NOT REALLY, I DID IT in the outside function!!!

  # Subset noise[test_idx, ] for test location
  y_data <- ngme_1rep$Y[bool_test_idx]
  group_data <- ngme_1rep$group[bool_test_idx]
  X_pred <- ngme_1rep$X[bool_test_idx, , drop = FALSE]
  noise_test_idx <- subset_noise(
    ngme_1rep$noise,
    sub_idx = bool_test_idx, compute_corr = TRUE
  )

  # Subset noise, X, Y in train location
  ngme_1rep$X <- ngme_1rep$X[bool_train_idx, , drop = FALSE]
  ngme_1rep$Y <- ngme_1rep$Y[bool_train_idx]

  ngme_1rep$noise <- subset_noise(
    ngme_1rep$noise,
    sub_idx = bool_train_idx, compute_corr = TRUE
  )

  # Subset A for test and train location
  A_preds <- list()
  for (i in seq_along(ngme_1rep$models)) {
    A_preds[[i]] <- ngme_1rep$models[[i]]$A[bool_test_idx, , drop = FALSE]
    ngme_1rep$models[[i]]$A <- ngme_1rep$models[[i]]$A[bool_train_idx, , drop = FALSE]
  }

  if (!is.null(ngme_chain_reps) && length(ngme_chain_reps) > 0) {
    ngme_chain_reps <- lapply(ngme_chain_reps, function(chain_rep) {
      chain_rep$X <- chain_rep$X[bool_train_idx, , drop = FALSE]
      chain_rep$Y <- chain_rep$Y[bool_train_idx]
      chain_rep$noise <- subset_noise(
        chain_rep$noise,
        sub_idx = bool_train_idx, compute_corr = TRUE
      )
      for (ii in seq_along(chain_rep$models)) {
        chain_rep$models[[ii]]$A <- chain_rep$models[[ii]]$A[bool_train_idx, , drop = FALSE]
      }
      chain_rep
    })
  }

  # A_pred_blcok <- [A1_pred .. An_pred]
  # extract A and cbind!
  A_pred_block <- Reduce(cbind, x = A_preds)

  if (is.null(seed)) seed <- as.integer(Sys.time())

  scores <- pred_1 <- pred_2 <- Y_1 <- Y_2 <- list()

  if (parallel && requireNamespace("parallel", quietly = TRUE)) {
    scores <- parallel::mclapply(1:N_sim, function(nn) {
      s <- compute_scores(
        ngme_1rep, n_gibbs_samples, n_burnin, seed + nn, A_pred_block, noise_test_idx, y_data, group_data, X_pred, metric, thining_gap, merge_groups, merged_group_name, ngme_chain_reps, chain_combine
      )
      s
    }, mc.cores = num_cores)
  } else {
    for (nn in 1:N_sim) {
      tryCatch(
        {
          scores[[nn]] <- compute_scores(
            ngme_1rep, n_gibbs_samples, n_burnin, seed + nn, A_pred_block,
            noise_test_idx, y_data, group_data, X_pred, metric, thining_gap, merge_groups, merged_group_name, ngme_chain_reps, chain_combine
          )
        },
        error = function(e) {
          warning(paste("Error in sequential computation:", e$message))
        }
      )
    }
  }

  # Check if we have any valid scores
  if (length(scores) == 0) {
    stop("Failed to compute any valid scores")
  }

  # compute extra pred_N?
  if (keep_pred) {
    tmp <- compute_pred_N(
      ngme_1rep, n_gibbs_samples, n_burnin, seed, A_pred_block, noise_test_idx, y_data, group_data, X_pred
    )
    pred_1 <- tmp$pred_N_1
    pred_2 <- tmp$pred_N_2
    Y_1 <- tmp$Y_N_1
    Y_2 <- tmp$Y_N_2
  }

  # Use the actual number of rows returned by scores instead of original group count
  # This handles the case where merge_groups=TRUE reduces 2 groups to 1 result
  n_result_rows <- nrow(scores[[1]])
  n_cols <- ncol(scores[[1]])

  # compute mean and sd
  array_3d <- array(unlist(scores),
    dim = c(n_result_rows, n_cols, length(scores))
  )

  mean_array <- apply(array_3d, c(1, 2), mean)
  sd_array <- apply(array_3d, c(1, 2), sd)

  colnames(mean_array) <- colnames(sd_array) <- names(scores[[1]])
  rownames(mean_array) <- rownames(sd_array) <- rownames(scores[[1]])

  # scores results and 2 predictions
  list(
    mean_scores = mean_array,
    sd_scores = sd_array,
    Y_1 = Y_1,
    Y_2 = Y_2,
    pred_1 = pred_1,
    pred_2 = pred_2
  )
}



# questions:
# 1. loop over replicates, and do CV for each replicate, and average
# 2. partition first, then do CV for each partition (with many replicates), and average


build_cv_aw_draws_matrix <- function(A_pred_block, Ws_block, n_obs, has_models) {
  if (!has_models) {
    return(matrix(0, nrow = n_obs, ncol = length(Ws_block)))
  }
  if (length(Ws_block) == 0) {
    return(matrix(0, nrow = n_obs, ncol = 0))
  }
  do.call(cbind, lapply(Ws_block, function(W) as.numeric(A_pred_block %*% W)))
}


collect_cv_chain_samples <- function(
    ngme_chain_reps,
    n_gibbs_samples,
    n_burnin,
    seed_int,
    X_pred) {
  Ws_block <- list()
  W2s_block <- list()
  chain_idx_1 <- integer()
  chain_idx_2 <- integer()
  fe_by_chain <- vector("list", length(ngme_chain_reps))

  for (ci in seq_along(ngme_chain_reps)) {
    chain_rep <- ngme_chain_reps[[ci]]
    fe_by_chain[[ci]] <- as.numeric(X_pred %*% chain_rep$feff)

    Ws_result <- sampling_cpp(
      chain_rep,
      n = n_gibbs_samples * 2,
      n_burnin = n_burnin,
      posterior = TRUE,
      seed = seed_int + ci - 1L
    )
    Ws <- Ws_result[["W"]]

    if (is.null(Ws) || length(Ws) == 0) {
      stop("sampling_cpp returned NULL or empty result for chain ", ci)
    }

    Ws_1 <- head(Ws, n_gibbs_samples)
    Ws_2 <- tail(Ws, n_gibbs_samples)
    Ws_block <- c(Ws_block, Ws_1)
    W2s_block <- c(W2s_block, Ws_2)
    chain_idx_1 <- c(chain_idx_1, rep(ci, length(Ws_1)))
    chain_idx_2 <- c(chain_idx_2, rep(ci, length(Ws_2)))
  }

  list(
    Ws_block = Ws_block,
    W2s_block = W2s_block,
    chain_idx_1 = chain_idx_1,
    chain_idx_2 = chain_idx_2,
    fe_by_chain = fe_by_chain
  )
}


# helper function to compute the scores
compute_scores <- function(
    ngme_1rep,
    n_gibbs_samples,
    n_burnin,
    seed,
    A_pred_block,
    test_noise,
    y_data,
    group_data,
    X_pred,
    metric,
    thining_gap,
    merge_groups = FALSE,
    merged_group_name = NULL,
    ngme_chain_reps = NULL,
    chain_combine = "param_mean") {
  tryCatch(
    {
      seed_int <- as.integer(abs(seed) %% 2147483647)
      n_obs <- nrow(X_pred)
      has_models <- length(ngme_1rep$models) > 0
      use_chain_average <- chain_combine == "predictive_average" &&
        !is.null(ngme_chain_reps) &&
        length(ngme_chain_reps) > 1

      if (use_chain_average) {
        sampled <- collect_cv_chain_samples(
          ngme_chain_reps = ngme_chain_reps,
          n_gibbs_samples = n_gibbs_samples,
          n_burnin = n_burnin,
          seed_int = seed_int,
          X_pred = X_pred
        )

        AW_N_1 <- build_cv_aw_draws_matrix(
          A_pred_block,
          sampled$Ws_block,
          n_obs = n_obs,
          has_models = has_models
        )
        AW_N_2 <- build_cv_aw_draws_matrix(
          A_pred_block,
          sampled$W2s_block,
          n_obs = n_obs,
          has_models = has_models
        )

        fe_N_1 <- do.call(cbind, lapply(sampled$chain_idx_1, function(ci) sampled$fe_by_chain[[ci]]))
        fe_N_2 <- do.call(cbind, lapply(sampled$chain_idx_2, function(ci) sampled$fe_by_chain[[ci]]))

        n_draws <- ncol(fe_N_1)
        thin_idx <- seq(1, n_draws, by = thining_gap + 1)

        pred_N_1_thin <- fe_N_1[, thin_idx, drop = FALSE] + AW_N_1[, thin_idx, drop = FALSE]
        pred_N_2_thin <- fe_N_2[, thin_idx, drop = FALSE] + AW_N_2[, thin_idx, drop = FALSE]
      } else {
        Ws_result <- sampling_cpp(
          ngme_1rep,
          n = n_gibbs_samples * 2,
          n_burnin = n_burnin,
          posterior = TRUE,
          seed = seed_int
        )
        Ws <- Ws_result[["W"]]

        if (is.null(Ws) || length(Ws) == 0) {
          stop("sampling_cpp returned NULL or empty result")
        }

        Ws_block <- head(Ws, n_gibbs_samples)
        W2s_block <- tail(Ws, n_gibbs_samples)

        AW_N_1 <- build_cv_aw_draws_matrix(
          A_pred_block,
          Ws_block,
          n_obs = n_obs,
          has_models = has_models
        )
        AW_N_2 <- build_cv_aw_draws_matrix(
          A_pred_block,
          W2s_block,
          n_obs = n_obs,
          has_models = has_models
        )

        fe <- as.numeric(X_pred %*% ngme_1rep$feff)
        fe_N <- matrix(rep(fe, n_gibbs_samples), ncol = n_gibbs_samples, byrow = FALSE)
        thin_idx <- seq(1, n_gibbs_samples, by = thining_gap + 1)

        pred_N_1_thin <- fe_N[, thin_idx, drop = FALSE] + AW_N_1[, thin_idx, drop = FALSE]
        pred_N_2_thin <- fe_N[, thin_idx, drop = FALSE] + AW_N_2[, thin_idx, drop = FALSE]
      }

      n_thin <- ncol(pred_N_1_thin)
      seed_group_1 <- seed_int + seq_len(n_thin)
      seed_group_2 <- seed_int + seq_len(n_thin) + n_thin
      mn_N_1 <- sapply(seq_len(n_thin), function(x) simulate(test_noise, seed = seed_group_1[x])[[1]])
      mn_N_2 <- sapply(seq_len(n_thin), function(x) simulate(test_noise, seed = seed_group_2[x])[[1]])

      Y_N_1_thin <- pred_N_1_thin + mn_N_1
      Y_N_2_thin <- pred_N_2_thin + mn_N_2

      compute_score_given_pred(
        Y_N_1_thin, Y_N_2_thin,
        y_data,
        group_data,
        merge_groups = merge_groups,
        merged_group_name = merged_group_name,
        metric = metric
      )
    },
    error = function(e) {
      if (inherits(e, "invalid_metric_error")) {
        stop(e)
      }
      warning(paste("Error in compute_scores:", e$message))
      n_group <- length(levels(group_data))
      if (n_group == 0) n_group <- 1

      scores <- matrix(NA, nrow = n_group, ncol = 4)
      colnames(scores) <- c("MAE", "MSE", "neg.CRPS", "neg.sCRPS")
      if (n_group == 1) {
        rownames(scores) <- "all"
      } else {
        rownames(scores) <- levels(group_data)
      }
      scores
    }
  )
}



# helper function to generate predictions and simulate Y
compute_pred_N <- function(
    ngme_1rep, n_gibbs_samples, n_burnin, seed, A_pred_block, noise_test_idx, y_data, group_data, X_pred) {
  seed_int <- as.integer(abs(seed) %% 2147483647)
  test_noise <- noise_test_idx

  Ws <- sampling_cpp(
    ngme_1rep,
    n = n_gibbs_samples * 2,
    n_burnin = n_burnin,
    posterior = TRUE,
    seed = seed_int
  )[["W"]]
  # )[["cond_W"]]

  Ws_block <- head(Ws, n_gibbs_samples)
  W2s_block <- tail(Ws, n_gibbs_samples)

  # Note: Ws_block is a list of N realizations of W of current replicate
  # Note: AW_N_1 is a matrix of n_test * N
  AW_N_1 <- if (length(ngme_1rep$models) == 0) {
    0
  } else {
    Reduce(cbind, sapply(Ws_block, function(W) A_pred_block %*% W))
  }
  AW_N_2 <- if (length(ngme_1rep$models) == 0) {
    0
  } else {
    Reduce(cbind, sapply(W2s_block, function(W) A_pred_block %*% W))
  }

  # sampling Y by, Y = X feff + (block_A %*% block_W) + eps
  # AW_N_1[[1]] is concat(A1 W1, A2 W2, ..)

  # generate fixed effect
  fe <- with(ngme_1rep, as.numeric(X_pred %*% feff))
  fe_N <- matrix(
    rep(fe, n_gibbs_samples),
    ncol = n_gibbs_samples,
    byrow = F
  )

  pred_N_1 <- fe_N + AW_N_1
  pred_N_2 <- fe_N + AW_N_2

  # simulate measurement noise
  n_thin <- ncol(pred_N_1)
  seed_group_1 <- seed_int + 1:n_thin
  seed_group_2 <- seed_int + 1:n_thin + n_thin
  mn_N_1 <- sapply(1:n_thin, function(x) simulate(test_noise, seed = seed_group_1[x])[[1]])
  mn_N_2 <- sapply(1:n_thin, function(x) simulate(test_noise, seed = seed_group_2[x])[[1]])

  # simulate y
  Y_N_1 <- pred_N_1 + mn_N_1
  Y_N_2 <- pred_N_2 + mn_N_2

  list(
    pred_N_1 = rowMeans(as.matrix(pred_N_1)),
    pred_N_2 = rowMeans(as.matrix(pred_N_2)),
    Y_N_1 = rowMeans(as.matrix(Y_N_1)),
    Y_N_2 = rowMeans(as.matrix(Y_N_2))
  )
}



#' Compute the scores given the prediction
#'
#' @param Y_N_1_thin posterior predictive draws (rows = observations, columns = samples)
#' @param Y_N_2_thin posterior predictive draws (rows = observations, columns = samples)
#' @param y_data a vector of length n_obs
#' @param group_data a vector of length n_obs
#' @param merge_groups logical, if TRUE and there are exactly two groups,
#'   combine them into a vector-valued score
#' @param merged_group_name optional label for the merged group
#' @param metric optional custom metric function used to combine group-wise values before scoring
compute_score_given_pred <- function(
    Y_N_1_thin, Y_N_2_thin,
    y_data, group_data,
    merge_groups = FALSE,
    merged_group_name = NULL,
    metric = NULL) {
  metric_data <- prepare_metric_data(metric, y_data, Y_N_1_thin, Y_N_2_thin, group_data)

  y_data <- metric_data$y
  group_data <- metric_data$group
  Y_N_1_thin <- metric_data$samples1
  Y_N_2_thin <- metric_data$samples2
  pred <- metric_data$pred

  # Now Y is of dim n_obs * N
  n_obs <- length(y_data)

  if (merge_groups && length(levels(group_data)) == 2) {
    # Handle merged groups for bivariate data (e.g., wind components)
    groups <- levels(group_data)
    group1_idx <- which(group_data == groups[1])
    group2_idx <- which(group_data == groups[2])

    # Check if we have matching observations for both groups
    if (length(group1_idx) != length(group2_idx)) {
      warning("Unequal number of observations for the two groups. Cannot merge.")
      # Fall through to original computation instead of setting merge_groups = FALSE
      # This ensures consistent return structure
    } else {
      # Organize data by groups
      pred_group1 <- pred[group1_idx]
      pred_group2 <- pred[group2_idx]
      y_data_group1 <- y_data[group1_idx]
      y_data_group2 <- y_data[group2_idx]

      # Compute vector-based metrics
      # MAE as Euclidean distance, MSE as squared Euclidean distance
      vec_mae <- sqrt((pred_group1 - y_data_group1)^2 + (pred_group2 - y_data_group2)^2)
      vec_mse <- (pred_group1 - y_data_group1)^2 + (pred_group2 - y_data_group2)^2

      # For CRPS and sCRPS, we need to work with the simulation arrays
      n_thin <- ncol(Y_N_1_thin)
      E_sim_data_thin <- E_sim_sim_thin <- double(length(group1_idx))

      for (i in 1:length(group1_idx)) {
        # Get simulation data for both components
        y_sim1_thin_comp1 <- as.numeric(Y_N_1_thin[group1_idx[i], ])
        y_sim2_thin_comp1 <- as.numeric(Y_N_2_thin[group1_idx[i], ])
        y_sim1_thin_comp2 <- as.numeric(Y_N_1_thin[group2_idx[i], ])
        y_sim2_thin_comp2 <- as.numeric(Y_N_2_thin[group2_idx[i], ])

        # Compute vector distances for CRPS
        obs_vec <- c(y_data_group1[i], y_data_group2[i])

        # E(||X_i - y_data||) where X_i ~ predictive distribution
        vec_dist_sim_data <- sqrt((y_sim1_thin_comp1 - obs_vec[1])^2 + (y_sim1_thin_comp2 - obs_vec[2])^2)
        E_sim_data_thin[i] <- mean(vec_dist_sim_data)

        # E(||X_i - Y_i||) where X_i, Y_i ~ predictive distribution
        vec_dist_sim_sim <- sqrt((y_sim1_thin_comp1 - y_sim2_thin_comp1)^2 + (y_sim1_thin_comp2 - y_sim2_thin_comp2)^2)
        E_sim_sim_thin[i] <- mean(vec_dist_sim_sim)
      }

      denom <- pmax(E_sim_sim_thin, .Machine$double.eps)
      CRPS_thin <- 0.5 * denom - E_sim_data_thin
      sCRPS_thin <- -E_sim_data_thin / denom - 0.5 * log(denom)

      # Create merged results
      scores <- data.frame(
        MAE = mean(vec_mae),
        MSE = mean(vec_mse),
        neg.CRPS = -mean(CRPS_thin),
        neg.sCRPS = -mean(sCRPS_thin)
      )

      # Use custom name if provided, otherwise use default "group1_group2" format
      if (!is.null(merged_group_name)) {
        rownames(scores) <- merged_group_name
      } else {
        rownames(scores) <- paste(groups, collapse = "_")
      }

      return(scores)
    }
  }

  # Original computation for non-merged groups
  E_sim_data_thin <- E_sim_sim_thin <- double(length(y_data))

  for (i in 1:n_obs) {
    # turn row of df into numeric vector.
    y_sim1_thin <- as.numeric(Y_N_1_thin[i, ])
    y_sim2_thin <- as.numeric(Y_N_2_thin[i, ])

    # estimate E(| X_i - y_data |). y_data is observation, X_i ~ predictive distribution at i
    E_sim_data_thin[[i]] <- mean(abs(y_sim1_thin - y_data[i]))

    # estimate E(| X_i - Y_i |) , X_i, Y_i ~ predictive distribution at i
    E_sim_sim_thin[[i]] <- mean(abs(y_sim1_thin - y_sim2_thin))
  }

  # compute MSE, MAE, CRPS, sCRPS within each group
  pred_each_group <- split(pred, group_data)
  y_data_each_group <- split(y_data, group_data)

  denom <- pmax(E_sim_sim_thin, .Machine$double.eps)
  CRPS_thin <- split(0.5 * denom - E_sim_data_thin, group_data)
  sCRPS_thin <- split(
    -E_sim_data_thin / denom - 0.5 * log(denom),
    group_data
  )

  # Compute MAE and MSE within each group
  MAE <- MSE <- double(length(pred_each_group))
  for (j in seq_along(pred_each_group)) {
    MAE[[j]] <- mean(abs(pred_each_group[[j]] - y_data_each_group[[j]]))
    MSE[[j]] <- mean((pred_each_group[[j]] - y_data_each_group[[j]])^2)
  }

  scores <- data.frame(
    MAE = MAE,
    MSE = MSE,
    neg.CRPS = -sapply(CRPS_thin, mean), # mean over 1:n_obs_test within each group
    neg.sCRPS = -sapply(sCRPS_thin, mean) # same
  )

  scores
}


prepare_metric_data <- function(metric, y_data, Y_N_1_thin, Y_N_2_thin, group_data) {
  samples1 <- as.matrix(Y_N_1_thin)
  samples2 <- as.matrix(Y_N_2_thin)
  pred_default <- 0.5 * (rowMeans(samples1) + rowMeans(samples2))

  group_factor <- if (is.factor(group_data)) droplevels(group_data) else factor(group_data)

  if (is.null(metric)) {
    return(list(
      y = as.numeric(y_data),
      pred = pred_default,
      samples1 = samples1,
      samples2 = samples2,
      group = group_factor
    ))
  }

  if (!is.function(metric)) {
    stop("Custom metric must be provided as a function.")
  }

  groups <- levels(group_factor)
  idx_list <- split(seq_along(group_factor), group_factor)
  counts <- lengths(idx_list)
  if (length(unique(counts)) != 1) {
    abort_invalid_metric(
      paste0(
        "Custom metric requires the same number of observations for each group within a fold. ",
        "Consider supplying paired splits via `type = \"custom\"`. Available groups: ",
        paste(groups, collapse = ", "),
        "."
      )
    )
  }

  n_locations <- counts[[1]]
  n_thin <- ncol(samples1)

  metric_y <- numeric(n_locations)
  metric_samples1 <- matrix(NA_real_, n_locations, n_thin)
  metric_samples2 <- matrix(NA_real_, n_locations, n_thin)
  label <- NULL

  for (i in seq_len(n_locations)) {
    obs_vec <- setNames(vapply(groups, function(g) as.numeric(y_data[idx_list[[g]][i]]), numeric(1)), groups)
    sample1_mat <- do.call(rbind, lapply(groups, function(g) samples1[idx_list[[g]][i], , drop = FALSE]))
    rownames(sample1_mat) <- groups
    sample2_mat <- do.call(rbind, lapply(groups, function(g) samples2[idx_list[[g]][i], , drop = FALSE]))
    rownames(sample2_mat) <- groups

    metric_input <- list(
      y = obs_vec,
      samples1 = sample1_mat,
      samples2 = sample2_mat,
      group = groups
    )

    metric_output <- metric(metric_input)
    parsed <- parse_metric_output(
      metric_output,
      expected_length = n_thin,
      available_groups = groups
    )

    metric_y[i] <- parsed$value
    if (is.null(label) && !is.null(parsed$label) && parsed$label != "") {
      label <- parsed$label
    }

    samples1_values <- parsed$samples1
    if (is.null(samples1_values)) {
      samples1_values <- compute_metric_samples(metric, sample1_mat, groups)
    }
    if (length(samples1_values) != n_thin) {
      abort_invalid_metric(
        "`samples1` returned by the custom metric must match the number of posterior draws."
      )
    }

    samples2_values <- parsed$samples2
    if (is.null(samples2_values)) {
      samples2_values <- compute_metric_samples(metric, sample2_mat, groups)
    }
    if (length(samples2_values) != n_thin) {
      abort_invalid_metric(
        "`samples2` returned by the custom metric must match the number of posterior draws."
      )
    }

    metric_samples1[i, ] <- samples1_values
    metric_samples2[i, ] <- samples2_values
  }

  metric_pred <- 0.5 * (rowMeans(metric_samples1) + rowMeans(metric_samples2))
  if (is.null(label) || label == "") {
    label <- "metric"
  }

  list(
    y = metric_y,
    pred = metric_pred,
    samples1 = metric_samples1,
    samples2 = metric_samples2,
    group = factor(rep(label, n_locations), levels = label)
  )
}


parse_metric_output <- function(result, expected_length = NULL, available_groups = NULL) {
  label <- NULL
  samples1 <- NULL
  samples2 <- NULL
  value <- NULL

  if (is.list(result)) {
    if (!is.null(result$label)) {
      label <- as.character(result$label)[1]
    }

    if (!is.null(result$y)) {
      value <- result$y
    } else if (!is.null(result$value)) {
      value <- result$value
    } else if (!is.null(result$result)) {
      value <- result$result
    } else if (length(result) == 1 && is.numeric(result[[1]])) {
      value <- result[[1]]
    }

    if (is.null(value)) {
      abort_invalid_metric(
        build_metric_message(
          "Custom metric must return a numeric value via `y` (or `value`).",
          available_groups
        )
      )
    }

    if (!is.null(names(value)) && length(value) == 1 && is.null(label)) {
      label <- names(value)[1]
    }

    value <- as.numeric(value)
    if (length(value) != 1 || !is.finite(value)) {
      abort_invalid_metric(
        build_metric_message(
          "Custom metric must provide a finite numeric scalar for `y`.",
          available_groups
        )
      )
    }

    if (!is.null(result$samples1)) {
      samples1 <- as.numeric(result$samples1)
    }
    if (!is.null(result$samples2)) {
      samples2 <- as.numeric(result$samples2)
    }
  } else if (is.numeric(result)) {
    value <- as.numeric(result)
    if (length(value) != 1 || !is.finite(value)) {
      abort_invalid_metric(
        build_metric_message(
          "Custom metric must return a scalar numeric value.",
          available_groups
        )
      )
    }
    if (!is.null(names(result))) {
      label <- names(result)[1]
    }
  } else {
    abort_invalid_metric(
      build_metric_message(
        "Custom metric must return either a numeric scalar or a list.",
        available_groups
      )
    )
  }

  if (!is.null(expected_length)) {
    if (!is.null(samples1) && length(samples1) > 0 && length(samples1) != expected_length) {
      abort_invalid_metric(
        "`samples1` returned by the custom metric must match the number of posterior draws."
      )
    }
    if (!is.null(samples2) && length(samples2) > 0 && length(samples2) != expected_length) {
      abort_invalid_metric(
        "`samples2` returned by the custom metric must match the number of posterior draws."
      )
    }
  }

  if (!is.null(samples1) && any(!is.finite(samples1))) {
    abort_invalid_metric("`samples1` returned by the custom metric must be finite.")
  }
  if (!is.null(samples2) && any(!is.finite(samples2))) {
    abort_invalid_metric("`samples2` returned by the custom metric must be finite.")
  }

  list(value = value, label = label, samples1 = samples1, samples2 = samples2)
}


compute_metric_samples <- function(metric, sample_mat, groups) {
  if (ncol(sample_mat) == 0) {
    return(numeric(0))
  }

  apply(sample_mat, 2, function(col) {
    sample_input <- list(y = setNames(as.numeric(col), groups), group = groups)
    parsed <- parse_metric_output(
      metric(sample_input),
      available_groups = groups
    )
    parsed$value
  })
}


abort_invalid_metric <- function(message) {
  stop(
    structure(
      list(message = message, call = sys.call(-1)),
      class = c("invalid_metric_error", "error", "condition")
    )
  )
}


build_metric_message <- function(message, available_groups = NULL) {
  if (is.null(available_groups) || length(available_groups) == 0) {
    return(message)
  }
  paste0(
    message,
    " Available groups: ",
    paste(available_groups, collapse = ", "),
    "."
  )
}


#' Merge model of replicates into model of 1 replicate given train_idx and
#' test_idx, the merged model contains all the information of train_idx from
#' different replicates.
#'
#' @param ngme a ngme object
#' @param train_idx a vector of indices of train data
#' @param test_idx a vector of indices of test data
merge_replicates <- function(
    ngme,
    train_idx,
    test_idx) {
  repls <- attr(ngme, "fit")$replicate
  uni_repl <- unique(repls)
  merged_rep <- ngme$replicates[[1]]
  n_latent <- length(merged_rep$models)


  train_X <- train_Y <- train_noise <- list()
  test_X <- test_Y <- test_noise <- test_group <- list()
  A_preds_block <- list()
  A_train <- vector("list", n_latent)
  # Loop over each replicate
  for (i in seq_along(uni_repl)) {
    data_idx_rep <- ngme$replicates[[i]]$data_idx
    bool_train_idx <- data_idx_rep %in% train_idx # current rep has train
    bool_test_idx <- data_idx_rep %in% test_idx # current rep has test

    ngme_1rep <- ngme$replicates[[i]]
    train_X[[i]] <- ngme_1rep$X[bool_train_idx, , drop = FALSE]
    train_Y[[i]] <- ngme_1rep$Y[bool_train_idx]
    train_noise[[i]] <- subset_noise(
      ngme_1rep$noise,
      sub_idx = bool_train_idx, compute_corr = TRUE
    )

    test_X[[i]] <- ngme_1rep$X[bool_test_idx, , drop = FALSE]
    test_Y[[i]] <- ngme_1rep$Y[bool_test_idx]
    test_group[[i]] <- ngme_1rep$group[bool_test_idx]
    test_noise[[i]] <- subset_noise(
      ngme_1rep$noise,
      sub_idx = bool_test_idx, compute_corr = TRUE
    )

    # Subset A for test and train location
    A_preds <- list()
    for (j in seq_len(n_latent)) {
      A_train[[j]][[i]] <- ngme_1rep$models[[j]]$A[bool_train_idx, , drop = FALSE]
      A_preds[[j]] <- ngme_1rep$models[[j]]$A[bool_test_idx, , drop = FALSE]
    }
    A_preds_block[[i]] <- Reduce(cbind, x = A_preds)
  }

  # Merge train data
  merged_rep$X <- Reduce(rbind, train_X)
  merged_rep$Y <- Reduce(c, train_Y)
  merged_rep$noise <- Reduce(merge_noise, train_noise)
  for (j in seq_len(n_latent)) {
    merged_rep$models[[j]]$A <- Reduce(rbind, A_train[[j]])
  }

  # organize test data
  test_Y <- Reduce(c, test_Y)
  test_group <- Reduce(c, test_group)
  test_X <- Reduce(rbind, test_X)
  test_noise <- Reduce(merge_noise, test_noise)
  test_A_block <- Reduce(rbind, x = A_preds_block)

  return(list(
    merged_rep = merged_rep,
    test_Y = test_Y,
    test_X = test_X,
    test_noise = test_noise,
    test_A_block = test_A_block,
    test_group = test_group
  ))
}

#' Create paired indices for bivariate cross-validation
#' Ensures that paired observations (e.g., u_wind and v_wind at same location)
#' are kept together in the same fold
#'
#' @param data data frame containing the observations
#' @param loc_col character vector of column names defining locations (e.g., c("lon", "lat") or c("x", "y"))
#' @param group character, name of the group column (e.g., "direction" for wind data)
#' @param k number of folds
#' @param seed random seed
#' @return list with test_idx and train_idx, each containing k lists of indices
#' @export
#' @examples
#' # Example with wind data
#' data_long <- data.frame(
#'   lon = rep(c(1, 2, 3, 4), 2),
#'   lat = rep(c(1, 1, 2, 2), 2),
#'   direction = rep(c("u_wind", "v_wind"), each = 4),
#'   wind = rnorm(8)
#' )
#' splits <- create_paired_cv_splits(data_long, c("lon", "lat"), "direction", k = 2)
create_paired_cv_splits <- function(data, loc_col, group, k, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)

  # Validate inputs
  stopifnot(
    "data must be a data frame" = is.data.frame(data),
    "loc_col must be character vector" = is.character(loc_col),
    "group must be character" = is.character(group) && length(group) == 1,
    "k must be positive integer" = k > 0,
    "location columns must exist in data" = all(loc_col %in% names(data)),
    "group column must exist in data" = group %in% names(data)
  )

  # Create location identifier by pasting location columns together
  data$location_id <- do.call(paste, c(data[loc_col], sep = "_"))

  # Get unique locations
  unique_locations <- unique(data$location_id)
  n_locations <- length(unique_locations)

  # Check that each location has observations for all groups
  groups_per_location <- aggregate(data[[group]],
    by = list(location = data$location_id),
    FUN = function(x) length(unique(x))
  )
  unique_groups <- unique(data[[group]])
  n_groups <- length(unique_groups)

  # Verify that all locations have the same number of groups
  if (!all(groups_per_location$x == n_groups)) {
    incomplete_locs <- groups_per_location$location[groups_per_location$x != n_groups]
    warning(paste(
      "Some locations don't have observations for all groups:",
      paste(head(incomplete_locs), collapse = ", ")
    ))
  }

  # Create folds based on unique locations
  if (n_locations < k) {
    warning(paste(
      "Number of unique locations (", n_locations,
      ") is less than k (", k, "). Some folds will be empty."
    ))
    k <- min(k, n_locations)
  }

  # Assign locations to folds
  if (k == 1) {
    # Special case: all locations go to fold 1
    location_folds <- rep(1, n_locations)
    names(location_folds) <- unique_locations
  } else {
    location_folds <- cut(sample(1:n_locations), breaks = k, label = FALSE)
    names(location_folds) <- unique_locations
  }

  # Create test and train indices
  test_idx <- list()
  train_idx <- list()

  for (i in 1:k) {
    # Get locations assigned to this fold
    test_locations <- unique_locations[location_folds == i]
    train_locations <- unique_locations[location_folds != i]

    # Get all observation indices for these locations (across all groups)
    test_idx[[i]] <- which(data$location_id %in% test_locations)
    train_idx[[i]] <- which(data$location_id %in% train_locations)

    # Sort indices to maintain order
    test_idx[[i]] <- sort(test_idx[[i]])
    train_idx[[i]] <- sort(train_idx[[i]])
  }

  # Add some diagnostic information
  attr(test_idx, "n_locations") <- n_locations
  attr(test_idx, "n_groups") <- n_groups
  attr(test_idx, "unique_groups") <- unique_groups
  attr(test_idx, "loc_col") <- loc_col
  attr(test_idx, "group_col") <- group

  return(list(test_idx = test_idx, train_idx = train_idx))
}

# #' prediction of ngme
# #'
# #' @param ngme ngme object
# #' @return a list of outputs contains estimation of operator paramters, noise parameters
# #' @export
# #'
# predict.ngme <- function(
#   ngme,
#   ...
# ) {
#   linear_predictor <- double(ngme$W_sizes)

#   AW_pred <- 0; AW_data <- 0
#   for (i in seq_along(models_in)) {
#     W <- ngme$models[[i]]$W
#     AW_pred <- AW_pred + drop(models_in[[i]]$A_pred %*% W)
#     AW_data <- AW_data + drop(models_in[[i]]$A %*% W)
#   }

#   # fixed effects. watch out! Xb could be double(0)
#   X_pred <- split_data$X_NA;
#   Xb_pred <- drop(X_pred %*% ngme$feff)
#   Xb_data <- drop(X_data %*% ngme$feff)

#   # ngme_response[split_data$index_NA] <- if (length(Xb_pred) == 0) AW_pred else AW_pred + Xb_pred
#   #
#   linear_predictor[split_data$index_NA]   <- if (length(Xb_pred) == 0) AW_pred else AW_pred + Xb_pred
#   linear_predictor[split_data$index_data] <- if (length(Xb_data) == 0) AW_data else AW_data + Xb_data

#   linear_predictor
# }


# #' prediction of ngme given X_pred and A_pred
# #'
# #' @param ngme ngme object
# #' @param X_pred a matrix for covariate at unknown location
# #' @param A_pred a list of observation matrix for each process
# #'
# #' @return linear prediction at unknown locations
# #' @export
# #'
# predict.ngme <- function(
#   ngme,
#   X_pred,
#   A_pred,
#   sampling_iteration = 300
# ) {
#   # posterior sampling
#   ngme <- post_sampling(ngme, sampling_iteration)

#   linear_predictor <- X_pred %*% ngme$feff
#     for (latent in ngme$models)
#       linear_predictor <- linear_predictor + A_pred[i] %*% latent$W

#   linear_predictor
# }

#' Compute Gaussian log-likelihood
#'
#' Creates the underlying C++ block models and evaluates their Gaussian log-likelihood when applicable.
#'
#' @param ngme_model An object created by `ngme()`.
#'
#' @return A numeric scalar. Returns `0` if any replicate is non-Gaussian.
#' @export
compute_log_like <- function(ngme_model) {
  stopifnot(inherits(ngme_model, "ngme"))
  estimation_enabled <- attr(ngme_model, "estimation_enabled")
  if (!is.null(estimation_enabled) && !isTRUE(estimation_enabled)) {
    warning("Model was created with control_opt$estimation = FALSE; fit the model before computing log-likelihood.")
    return(NA_real_)
  }

  tryCatch(
    compute_log_like_cpp(ngme_model),
    error = function(err) {
      warning("Failed to compute log-likelihood: ", conditionMessage(err))
      NA_real_
    }
  )
}

#' @title Calibrate Inverse-Exponential Prior for NIG Driven Noise
#' @description
#' Calibrate \eqn{\lambda} in the prior
#' \eqn{\kappa = 1/\nu \sim \mathrm{Exp}(\lambda)}
#' using a tail-inflation target for driven NIG noise.
#'
#' The calibration target is
#' \deqn{\Pr(R_c(\nu) > r_\text{target}) = \alpha,}
#' where
#' \deqn{R_c(\nu) = \frac{\Pr(|U| > c \mid \nu)}{\Pr(|Z| > c)},\quad Z\sim N(0,1)}
#' and
#' \deqn{U = \frac{\mu(V-h) + \sigma\sqrt{V}Z}{\sigma\sqrt{h}},\quad V\sim\mathrm{GIG}(-1/2,\nu,\nu h^2).}
#'
#' The function solves \eqn{R_c(\nu_r) = r_\text{target}} using log-scale
#' bisection, then returns
#' \deqn{\lambda = -\nu_r \log(\alpha).}
#'
#' @param r_target target tail inflation level \eqn{r_\text{target} > 1}
#' @param alpha target prior probability in \eqn{(0,1)}
#' @param c tail threshold for \eqn{|U|>c}, typically `2.5` or `3`
#' @param mu NIG drift parameter in the driven noise term
#' @param sigma NIG scale parameter (> 0)
#' @param h positive increment scaling (> 0)
#' @param n_samples Monte Carlo sample size used per evaluation of \eqn{R_c(\nu)}
#' @param nu_lower initial lower bracket for \eqn{\nu}
#' @param nu_upper initial upper bracket for \eqn{\nu}
#' @param tol relative tolerance for solving \eqn{R_c(\nu_r)=r_\text{target}}
#' @param max_iter maximum number of bisection iterations
#' @param max_expand maximum bracket expansion steps on each side
#' @param seed optional integer seed for reproducible calibration
#' @param ... arguments forwarded to \code{calibrate_inv_exp_lambda_driven_nig()}
#'
#' @return A list with calibrated `lambda`, solved `nu_r`, achieved `rc_nu_r`,
#'   and diagnostics.
#' @export
calibrate_inv_exp_lambda_driven_nig <- function(
    r_target = 2,
    alpha = 0.1,
    c = 3,
    mu = 0,
    sigma = 1,
    h = 1,
    n_samples = 1e5,
    nu_lower = 0.1,
    nu_upper = 100,
    tol = 0.05,
    max_iter = 30,
    max_expand = 30,
    seed = NULL) {
  stopifnot(
    is.numeric(r_target), length(r_target) == 1, is.finite(r_target), r_target > 1,
    is.numeric(alpha), length(alpha) == 1, is.finite(alpha), alpha > 0, alpha < 1,
    is.numeric(c), length(c) == 1, is.finite(c), c > 0,
    is.numeric(mu), length(mu) == 1, is.finite(mu),
    is.numeric(sigma), length(sigma) == 1, is.finite(sigma), sigma > 0,
    is.numeric(h), length(h) == 1, is.finite(h), h > 0,
    is.numeric(n_samples), length(n_samples) == 1, is.finite(n_samples), n_samples > 0,
    is.numeric(nu_lower), length(nu_lower) == 1, is.finite(nu_lower), nu_lower > 0,
    is.numeric(nu_upper), length(nu_upper) == 1, is.finite(nu_upper), nu_upper > nu_lower,
    is.numeric(tol), length(tol) == 1, is.finite(tol), tol > 0,
    is.numeric(max_iter), length(max_iter) == 1, is.finite(max_iter), max_iter >= 1,
    is.numeric(max_expand), length(max_expand) == 1, is.finite(max_expand), max_expand >= 1
  )

  n_samples <- as.integer(n_samples)
  max_iter <- as.integer(max_iter)
  max_expand <- as.integer(max_expand)

  gaussian_tail <- 2 * stats::pnorm(-c)
  if (gaussian_tail <= 0) {
    stop("Gaussian tail probability underflows for this c; use a smaller c.")
  }

  normalize_seed <- function(x, offset = 0L) {
    as.integer((abs(as.double(x)) + as.double(offset)) %%
      (.Machine$integer.max - 1L) + 1L)
  }

  rc_cache <- new.env(parent = emptyenv())

  eval_rc <- function(nu_value) {
    key <- formatC(nu_value, digits = 17, format = "fg", flag = "#")
    if (exists(key, envir = rc_cache, inherits = FALSE)) {
      return(get(key, envir = rc_cache, inherits = FALSE))
    }

    if (!is.null(seed)) {
      seed_v <- normalize_seed(seed, 101L)
      seed_z <- normalize_seed(seed, 202L)
    } else {
      seed_v <- 0L
      seed_z <- NULL
    }

    V <- rgig(n = n_samples, p = -0.5, a = nu_value, b = nu_value * h * h, seed = seed_v)
    Z <- if (!is.null(seed_z)) {
      withr::with_seed(seed_z, stats::rnorm(n_samples))
    } else {
      stats::rnorm(n_samples)
    }

    U <- (mu * (V - h) + sigma * sqrt(V) * Z) / (sigma * sqrt(h))
    rc <- mean(abs(U) > c) / gaussian_tail
    if (!is.finite(rc)) rc <- NA_real_

    assign(key, rc, envir = rc_cache)
    rc
  }

  nu_lo <- nu_lower
  nu_hi <- nu_upper
  bracket_found <- FALSE
  rc_lo <- NA_real_
  rc_hi <- NA_real_
  rc_min <- NA_real_
  rc_max <- NA_real_

  for (k in seq_len(max_expand + 1L)) {
    nu_grid <- exp(seq(log(nu_lo), log(nu_hi), length.out = 41))
    rc_grid <- vapply(nu_grid, eval_rc, numeric(1))
    ok <- is.finite(rc_grid)
    if (!any(ok)) {
      stop("Numerical failure while evaluating R_c(nu); no finite values were obtained.")
    }

    nu_grid <- nu_grid[ok]
    rc_grid <- rc_grid[ok]
    f_grid <- rc_grid - r_target
    rc_min <- min(rc_grid)
    rc_max <- max(rc_grid)

    crossings <- which(f_grid[-length(f_grid)] * f_grid[-1] <= 0)
    if (length(crossings) > 0) {
      # Use the rightmost crossing (closest to Gaussian regime).
      i <- crossings[length(crossings)]
      nu_lo <- nu_grid[i]
      nu_hi <- nu_grid[i + 1]
      rc_lo <- rc_grid[i]
      rc_hi <- rc_grid[i + 1]
      bracket_found <- TRUE
      break
    }

    if (k > max_expand) break

    if (rc_max < r_target) {
      nu_lo <- nu_lo / 2
    } else if (rc_min > r_target) {
      nu_hi <- nu_hi * 2
    } else {
      # No clean crossing due non-monotonic/noisy curve; expand both sides.
      nu_lo <- nu_lo / 2
      nu_hi <- nu_hi * 2
    }
  }

  if (!bracket_found) {
    stop(
      "Could not bracket nu_r for R_c(nu)=r_target. ",
      "Observed scanned range was R_c in [", signif(rc_min, 4), ", ", signif(rc_max, 4), "]. ",
      "Target may be unattainable for the chosen (c, mu, sigma, h). ",
      "Try larger c (e.g. 2.5 or 3), smaller r_target, or larger n_samples."
    )
  }

  converged <- FALSE
  f_lo <- rc_lo - r_target
  f_hi <- rc_hi - r_target
  nu_mid <- exp((log(nu_lo) + log(nu_hi)) / 2)
  rc_mid <- eval_rc(nu_mid)

  for (iter in seq_len(max_iter)) {
    nu_mid <- exp((log(nu_lo) + log(nu_hi)) / 2)
    rc_mid <- eval_rc(nu_mid)
    f_mid <- rc_mid - r_target

    rel_err <- abs(rc_mid - r_target) / max(1, r_target)
    if (rel_err <= tol) {
      converged <- TRUE
      break
    }

    if (f_lo * f_mid <= 0) {
      nu_hi <- nu_mid
      rc_hi <- rc_mid
      f_hi <- f_mid
    } else {
      nu_lo <- nu_mid
      rc_lo <- rc_mid
      f_lo <- f_mid
    }
  }

  nu_r <- nu_mid
  lambda <- -nu_r * log(alpha)

  structure(
    list(
      lambda = lambda,
      nu_r = nu_r,
      rc_nu_r = rc_mid,
      r_target = r_target,
      alpha = alpha,
      alpha_implied = exp(-lambda / nu_r),
      c = c,
      mu = mu,
      sigma = sigma,
      h = h,
      n_samples = n_samples,
      bracket_nu = c(lower = nu_lo, upper = nu_hi),
      bracket_rc = c(lower = rc_lo, upper = rc_hi),
      rc_scan_range = c(min = rc_min, max = rc_max),
      converged = converged,
      max_iter = max_iter,
      tol = tol,
      seed = seed,
      root_signs = c(lower = f_lo, upper = f_hi)
    ),
    class = "ngme_inv_exp_calibration"
  )
}

#' @rdname calibrate_inv_exp_lambda_driven_nig
#' @export
calibrate_inv_exp_lambda <- function(...) {
  calibrate_inv_exp_lambda_driven_nig(...)
}

test_that("calibrate_inv_exp_lambda_driven_nig returns coherent calibration", {
  out <- calibrate_inv_exp_lambda_driven_nig(
    r_target = 1.3,
    alpha = 0.1,
    c = 2.5,
    mu = 0,
    sigma = 1,
    h = 1,
    n_samples = 20000,
    nu_lower = 0.05,
    nu_upper = 50,
    tol = 0.25,
    max_iter = 12,
    seed = 123
  )

  expect_true(is.list(out))
  expect_true(inherits(out, "ngme_inv_exp_calibration"))
  expect_gt(out$lambda, 0)
  expect_gt(out$nu_r, 0)
  expect_equal(out$lambda, -out$nu_r * log(out$alpha), tolerance = 1e-12)
  expect_equal(out$alpha_implied, out$alpha, tolerance = 1e-12)
  expect_lt(abs(out$rc_nu_r - out$r_target), 0.6)
})

test_that("calibrate_inv_exp_lambda is a direct alias", {
  out1 <- calibrate_inv_exp_lambda_driven_nig(
    r_target = 1.3,
    alpha = 0.2,
    c = 2.5,
    n_samples = 20000,
    nu_lower = 0.05,
    nu_upper = 50,
    tol = 0.25,
    max_iter = 12,
    seed = 99
  )

  out2 <- calibrate_inv_exp_lambda(
    r_target = 1.3,
    alpha = 0.2,
    c = 2.5,
    n_samples = 20000,
    nu_lower = 0.05,
    nu_upper = 50,
    tol = 0.25,
    max_iter = 12,
    seed = 99
  )

  expect_equal(out2$lambda, out1$lambda, tolerance = 1e-12)
  expect_equal(out2$nu_r, out1$nu_r, tolerance = 1e-12)
})

test_that("calibration reports unattainable tail target clearly", {
  expect_error(
    calibrate_inv_exp_lambda_driven_nig(
      r_target = 2,
      alpha = 0.1,
      c = 2,
      mu = 0,
      sigma = 1,
      h = 2,
      n_samples = 20000,
      seed = 1
    ),
    "Target may be unattainable"
  )
})

test_that("re model works with f", {
  out <- f(~ 1 + x,
    model = re(),
    noise = noise_nig(),
    data = data.frame(x = 1:7)
  )
  expect_true(inherits(out, "ngme_model"))
  expect_equal(out$operator$model, "re")
})

---
title: "ngme2 Documentation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{ngme2 Documentation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The full documentation for **ngme2**, including worked examples and rendered
output, is available on the package website:

<https://davidbolin.github.io/ngme2/>

## Getting Started

- [Overview of ngme2](https://davidbolin.github.io/ngme2/articles/ngme2.html):
  Introduction to the package and its main features
- [Installation and configuration (OpenMP)](https://davidbolin.github.io/ngme2/articles/Installation.html):
  How to install ngme2 and enable parallel computation

## Latent Models

- [AR(1) model](https://davidbolin.github.io/ngme2/articles/AR1-model.html):
  Autoregressive order-1 models with non-Gaussian noise
- [Random walk model](https://davidbolin.github.io/ngme2/articles/RW-model.html):
  RW1 and RW2 models
- [Ornstein-Uhlenbeck process](https://davidbolin.github.io/ngme2/articles/ou-process.html):
  Continuous-time OU models
- [SPDE Matérn model](https://davidbolin.github.io/ngme2/articles/SPDE-model.html):
  Spatial models based on the SPDE approach
- [Separable space-time (tensor product) model](https://davidbolin.github.io/ngme2/articles/tensor-product.html):
  Space-time models via tensor products
- [Non-separable space-time model (advection-diffusion)](https://davidbolin.github.io/ngme2/articles/non-separable-model.html):
  Anisotropic and advective space-time models
- [Matérn SPDE model on metric graph](https://davidbolin.github.io/ngme2/articles/matern-on-graph.html):
  Spatial models on networks and metric graphs
- [Bivariate type-G models](https://davidbolin.github.io/ngme2/articles/bivariate.html):
  Joint models for two response variables
- [Using the generic model](https://davidbolin.github.io/ngme2/articles/generic.html):
  How to define custom latent models

## Noise, Estimation, and Prediction

- [Combined Gaussian and NIG noise models](https://davidbolin.github.io/ngme2/articles/normal-nig-noise.html):
  Mixed noise specifications
- [Model estimation and prediction](https://davidbolin.github.io/ngme2/articles/pred-and-est.html):
  How estimation and prediction work in ngme2
- [Models with replicates](https://davidbolin.github.io/ngme2/articles/replicate.html):
  Handling replicated observations
- [Cross-validation](https://davidbolin.github.io/ngme2/articles/cross-validation.html):
  Time series and spatial cross-validation tools
- [Prior templates](https://davidbolin.github.io/ngme2/articles/prior-templates.html):
  Stationary and non-stationary prior specifications
- [SGLD posterior-like sampling](https://davidbolin.github.io/ngme2/articles/sgld-sampling.html):
  Uncertainty quantification via stochastic gradient Langevin dynamics

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{stepsize_schedule}
\alias{stepsize_schedule}
\title{Stepsize schedule}
\usage{
stepsize_schedule(
  method = c("constant", "poly"),
  alpha = 0.501,
  t0 = 1,
  burnin_iter = 0
)
}
\arguments{
\item{method}{schedule type. "constant" keeps \eqn{\eta_t = \eta_0} and
"poly" uses polynomial decay.}

\item{alpha}{polynomial exponent used when \code{method = "poly"}.
Must satisfy \eqn{1/2 < \alpha < 1}.}

\item{t0}{non-negative offset in iteration index.}

\item{burnin_iter}{non-negative integer. During these initial iterations,
schedule scaling is fixed to 1. Afterward, polynomial decay is applied
with reset local time index.}
}
\value{
a list of control variables for stepsize schedule
(used in \code{control_opt}).
}
\description{
Stepsize schedule
}
\details{
Define an iteration-dependent stepsize schedule for use in \code{control_opt()}.
This schedule is independent of \code{stepsize_decay()} and applies a direct
multiplicative scaling by iteration index:
\deqn{\eta_t = \eta_0 (t + t_0)^{-\alpha}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/noise.R
\name{noise_moments}
\alias{noise_moments}
\alias{noise_moments.ngme_noise}
\alias{noise_nig_moments}
\alias{noise_gal_moments}
\title{Moments of stationary NIG and GAL noise}
\usage{
noise_moments(noise, ...)

\method{noise_moments}{ngme_noise}(noise, ...)

noise_nig_moments(delta = -mu, mu = 0, sigma = 1, nu = 1)

noise_gal_moments(delta = -mu, mu = 0, sigma = 1, nu = 1)
}
\arguments{
\item{noise}{An \code{ngme_noise} object created by \code{noise_nig()} or
\code{noise_gal()}.}

\item{...}{Currently unused.}

\item{delta}{Location parameter. Defaults to \code{-mu}, matching the
stationary ngme noise convention.}

\item{mu}{Shift/skewness parameter in the mixture mean.}

\item{sigma}{Positive Gaussian scale parameter.}

\item{nu}{Positive mixing-distribution shape parameter.}
}
\value{
A named numeric vector with entries \code{skewness},
  \code{kurtosis}, \code{excess_kurtosis}, \code{variance}, \code{sd},
  \code{central_moment3}, and \code{central_moment4}.
}
\description{
Compute the mean, variance, skewness, kurtosis, and excess kurtosis for
stationary NIG and GAL noise. The distribution is evaluated through the
normal mean-variance mixture
\deqn{X = \delta + \mu V + \sigma \sqrt{V} Z,}
where \eqn{Z \sim N(0, 1)}. For the package's stationary noise
parameterization, \code{delta = -mu}, so the mean is zero.
}
\examples{
noise_nig_moments(mu = 1, sigma = 2, nu = 3)
noise_gal_moments(mu = 1, sigma = 2, nu = 3)
noise_moments(noise_nig(mu = 1, sigma = 2, nu = 3))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{get_parameter_distance}
\alias{get_parameter_distance}
\title{Calculate parameter distance from true values}
\usage{
get_parameter_distance(
  trajectories,
  true_values,
  norm_type = "euclidean",
  chain_summary = "mean"
)
}
\arguments{
\item{trajectories}{ngme_trajectories object from get_trace_trajectories()}

\item{true_values}{named list or vector of true parameter values}

\item{norm_type}{character, type of norm to use: "euclidean" (default), "manhattan", "max"}

\item{chain_summary}{character, how to summarize across chains: "mean" (default), "median", "chain1"}
}
\value{
numeric vector of distances across iterations
}
\description{
Compute ||theta - theta_hat|| distance for all parameters across iterations.
This provides a single measure of convergence combining all parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control_opt_batch_ci}
\alias{control_opt_batch_ci}
\title{Generate CI-focused control settings for batch-means inference}
\usage{
control_opt_batch_ci(
  optimizer = sgd(stepsize = 0.03),
  burnin = 100,
  iterations = 2000,
  n_batch = 20,
  n_parallel_chain = 4,
  alpha = 0.501,
  t0 = 1,
  schedule_burnin_iter = 0,
  ...
)
}
\arguments{
\item{optimizer}{optimizer object, default \code{sgd(stepsize = 0.03)}.}

\item{burnin}{burn-in iterations before optimization.}

\item{iterations}{optimization iterations.}

\item{n_batch}{number of checkpoints.}

\item{n_parallel_chain}{number of parallel chains.}

\item{alpha}{polynomial stepsize exponent for \code{poly_decay(alpha, t0)}.}

\item{t0}{non-negative schedule offset.}

\item{schedule_burnin_iter}{non-negative integer. Initial optimization
iterations where polynomial schedule scaling is disabled.}

\item{...}{additional arguments forwarded to \code{control_opt()} to override
defaults.}
}
\value{
object of class \code{control_opt}.
}
\description{
Convenience wrapper for \code{control_opt()} with defaults tailored for
Xi-style batch-means confidence intervals.
}
\details{
This helper enforces trajectory-friendly defaults:
\itemize{
  \item \code{store_traj = TRUE}
  \item \code{trend_std_conv_check = FALSE}
  \item \code{R_hat_conv_check = FALSE}
  \item \code{pflug_conv_check = FALSE}
  \item \code{stepsize_control = poly_decay(alpha, t0, schedule_burnin_iter)}
}
Any of these can still be overridden through \code{...}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{adagrad}
\alias{adagrad}
\title{AdaGrad SGD optimization}
\usage{
adagrad(stepsize = 0.05, epsilon = 1e-08)
}
\arguments{
\item{stepsize}{stepsize for SGD}

\item{epsilon}{epsilon for numerical stability}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
AdaGrad SGD optimization
}
\details{
The update rule for AdaGrad is:
\deqn{v_t = v_{t-1} + g_t^2}
\deqn{x_{t+1} = x_t - \text{stepsize} * \frac{g_t}{\sqrt{v_t} + \epsilon}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.R
\name{ngme_result}
\alias{ngme_result}
\title{Access the result of a ngme fitted model}
\usage{
ngme_result(ngme_object, model = NULL, transformed = TRUE)
}
\arguments{
\item{ngme_object}{a ngme fitted model object}

\item{model}{latent model name to filter by (e.g., "my_ar1", "my_matern", "my_rw1", etc.), "data" for fixed effects and measurement noise, if NULL, return all models}

\item{transformed}{logical, if TRUE (default) return transformed parameters, if FALSE return raw parameters}
}
\value{
a list of parameters for the specified model, or all models if model is NULL
}
\description{
Access the result of a ngme fitted model
}
\examples{
# Fit a simple AR(1) model
set.seed(1)
Y <- 1:10
n_obs <- length(Y)
x1 <- rnorm(n_obs)
x2 <- runif(n_obs)

ngme_out <- ngme(
  Y ~ x1 + x2 + f(
    1:n_obs,
    name = "my_ar",
    model = ar1(rho = 0.5),
    noise = noise_nig(mu = 2, sigma = 3, nu = 1)
  ),
  data = data.frame(x1 = x1, x2 = x2, Y = Y),
  control_opt = control_opt(estimation = FALSE)
)

# Get all model parameters (transformed)
all_params <- ngme_result(ngme_out)
# Returns: list(my_ar = list(rho = 0.5, mu = 2, sigma = 3, nu = 1),
#               data = list(fixed_effects = c(...), sigma = 0.5))

# Get parameters for specific latent model
ar_params <- ngme_result(ngme_out, model = "my_ar")
# Returns: list(rho = 0.5, mu = 2, sigma = 3, nu = 1)

# Get raw (untransformed) parameters
ar_raw <- ngme_result(ngme_out, model = "my_ar", transformed = FALSE)
# Returns: list(theta_rho = 1.099, mu = 2, sigma = 3, nu = 1)

# Get fixed effects and measurement noise
data_params <- ngme_result(ngme_out, model = "data")
# Returns: list(fixed_effects = c(...), sigma = 0.5, ...)

# For models with multiple latent processes
ngme_out2 <- ngme(
  Y ~ x1 + x2 + f(
    1:n_obs,
    name = "my_ar",
    model = ar1(rho = 0.5),
    noise = noise_nig(mu = 2, sigma = 3, nu = 1)
  ) + f(
    1:n_obs,
    name = "my_ou",
    model = ou(),
    noise = noise_normal(sigma = 1)
  ),
  data = data.frame(x1 = x1, x2 = x2, Y = Y),
  control_opt = control_opt(estimation = FALSE)
)

# Get all models
all_models <- ngme_result(ngme_out2)
# Returns: list(my_ar = list(...), my_ou = list(...), data = list(...))

# Get specific model
ou_params <- ngme_result(ngme_out2, model = "my_ou")
# Returns: list(theta_K1 = 0.5, sigma = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{summary.ngme_batch_ci}
\alias{summary.ngme_batch_ci}
\title{Summary for Batch-Means CI Results}
\usage{
\method{summary}{ngme_batch_ci}(object, digits = 4, ...)
}
\arguments{
\item{object}{an object of class `ngme_batch_ci`.}

\item{digits}{number of digits used for printing.}

\item{...}{currently unused.}
}
\value{
A list with summary table and covariance matrix.
}
\description{
Summarize an object of class `ngme_batch_ci`, including parameter means,
standard errors, confidence intervals, and covariance matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{matern}
\alias{matern}
\title{ngme Matern SPDE model specification}
\usage{
matern(
  mesh = NULL,
  alpha = 2,
  fix_alpha = TRUE,
  kappa = NULL,
  theta_kappa = NULL,
  rational_order = 2,
  B_kappa = NULL
)
}
\arguments{
\item{mesh}{an fmesher::fm_mesh_2d object, mesh for build the SPDE model}

\item{alpha}{SPDE smoothness parameter (alpha = 2beta) 2 or 4 for integer case, otherwise for fractional case}

\item{fix_alpha}{if FALSE, enable fractional modeling}

\item{kappa}{the range parameter, for the stationary model, it is the range parameter}

\item{theta_kappa}{kappa = exp(B_kappa * theta_kappa), for the non-stationary model, it is the range parameter}

\item{rational_order}{order of rational approximation for fractional operators (default: 2)}

\item{B_kappa}{bases for theta_kappa, ignore if use the stationary model}
}
\value{
ngme_operator object
}
\description{
ngme Matern SPDE model specification
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{sgd}
\alias{sgd}
\title{Vanilla SGD optimization}
\usage{
sgd(stepsize = 0.001)
}
\arguments{
\item{stepsize}{stepsize for SGD}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
Simple stochastic gradient descent without momentum or preconditioning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prior.R
\name{priors}
\alias{priors}
\title{Prior Container}
\usage{
priors(...)
}
\arguments{
\item{...}{named prior specifications}
}
\value{
named prior container
}
\description{
Prior Container
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{compute_score_given_pred}
\alias{compute_score_given_pred}
\title{Compute the scores given the prediction}
\usage{
compute_score_given_pred(
  Y_N_1_thin,
  Y_N_2_thin,
  y_data,
  group_data,
  merge_groups = FALSE,
  merged_group_name = NULL,
  metric = NULL
)
}
\arguments{
\item{Y_N_1_thin}{posterior predictive draws (rows = observations, columns = samples)}

\item{Y_N_2_thin}{posterior predictive draws (rows = observations, columns = samples)}

\item{y_data}{a vector of length n_obs}

\item{group_data}{a vector of length n_obs}

\item{merge_groups}{logical, if TRUE and there are exactly two groups,
combine them into a vector-valued score}

\item{merged_group_name}{optional label for the merged group}

\item{metric}{optional custom metric function used to combine group-wise values before scoring}
}
\description{
Compute the scores given the prediction
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{rw1}
\alias{rw1}
\title{Random Walk Model of Order 1 (RW1)}
\usage{
rw1(mesh = NULL, cyclic = FALSE, constr = TRUE)
}
\arguments{
\item{mesh}{numerical vector or inla.mesh.1d object, locations to build the mesh.
For numerical vectors, assumes equally spaced locations.}

\item{cyclic}{logical, whether the mesh is circular. If TRUE, the first and last
locations are treated as neighbors. Cannot be FALSE when constr = TRUE.}

\item{constr}{logical, whether to enforce the sum-to-zero constraint
\eqn{\sum_{i=1}^n h_i W_i = 0}. If FALSE, fixes the first element \eqn{W_1 = 0}.
Must be TRUE for cyclic models.}
}
\value{
An `ngme_operator` object containing the precision matrix and related
  components for the RW1 model.
}
\description{
Constructs a first-order random walk model for spatial or temporal processes.
The RW1 model assumes that first-order differences \eqn{\Delta W_i = W_i - W_{i-1}}
are independent and identically distributed Gaussian variables.
}
\details{
The RW1 model is defined by the precision matrix \eqn{K} that penalizes
first-order differences. The model has different structures depending on the
constraints:

**Non-cyclic, constrained (default)**: The first row of \eqn{K} enforces the
constraint \eqn{\sum_{i=1}^n h_i W_i = 0}, where \eqn{h_i} are the mesh weights.

**Non-cyclic, unconstrained**: The first element is fixed at 0 (\eqn{W_1 = 0}).

**Cyclic**: Treats the domain as circular, connecting the first and last locations
as neighbors. The constraint \eqn{\sum_{i=1}^n h_i W_i = 0} is always enforced.
The precision matrix is expanded to size \eqn{(n+1) \times (n+1)} to handle
the constraint properly.
}
\examples{
# Non-cyclic constrained RW1 (default)
rw1_default <- rw1(1:5)
print(rw1_default$K)

# Non-cyclic unconstrained RW1 (fixes first element to 0)
rw1_fixed <- rw1(1:5, constr = FALSE)
print(rw1_fixed$K)

# Cyclic RW1 (connects first and last locations)
rw1_cyclic <- rw1(1:5, cyclic = TRUE)
print(rw1_cyclic$K)

# Using with unequally spaced locations
locations <- c(0, 1, 3, 6, 10)
rw1_unequal <- rw1(locations)
print(rw1_unequal$K)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{arma}
\alias{arma}
\title{ngme ARMA(p, q) model specification}
\usage{
arma(
  mesh = NULL,
  ar_order = 1,
  ma_order = 1,
  ar = NULL,
  ma = NULL,
  fix_ar = FALSE,
  fix_ma = FALSE
)
}
\arguments{
\item{mesh}{Integer vector or `inla.mesh.1d` object.}

\item{ar_order}{Integer p (0, 1, or 2).}

\item{ma_order}{Integer q (0, 1, or 2).}

\item{ar}{Numeric length-p vector of AR coefficients in (-1, 1). Defaults to zeros.}

\item{ma}{Numeric length-q vector of MA coefficients in (-1, 1). Defaults to zeros.}

\item{fix_ar}{Logical or length-p logical vector to fix AR coefficients.}

\item{fix_ma}{Logical or length-q logical vector to fix MA coefficients.}
}
\value{
An `ngme_operator` describing the ARMA(p, q) latent model.
}
\description{
General ARMA(p, q) latent operator under AZW: K W = Z eps,
where K encodes the AR part and Z encodes the MA part.
}
\details{
- Supports p, q <= 2. The user provides standard AR/MA coefficients (ar, ma).
  For estimation, these are transformed in R into an unbounded parameter vector
  `theta_K` via PACF-style mappings so optimizers (e.g., Adam/SGD) operate on
  unconstrained variables while K and Z remain valid:
  - AR(1): phi1 = tanh(raw1/2). AR(2): pi1=tanh(raw1/2), pi2=tanh(raw2/2),
    phi2 = pi2, phi1 = pi1 * (1 - pi2).
  - MA(1): theta1 = tanh(raw1/2). MA(2): psi1=tanh(raw1/2), psi2=tanh(raw2/2),
    theta2 = psi2, theta1 = psi1 * (1 - psi2).
- The returned operator includes both K (AR part) and Z (MA part). During
  estimation the backend updates Z synchronously every time `theta_K` changes.
}
\examples{
mesh <- 1:10
op <- arma(mesh, ar_order = 2, ma_order = 2, ar = c(0.5, 0.3), ma = c(0.6, 0.4))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{get_data_from_formula}
\alias{get_data_from_formula}
\title{Extracts design matrix from a formula and data.}
\usage{
get_data_from_formula(form, data)
}
\arguments{
\item{form}{A formula object, e.g., `~ x1 + x2` or `~.`.}

\item{data}{A data frame or matrix from which to extract the design matrix.}
}
\value{
A numeric matrix representing the design matrix.
}
\description{
This function takes a formula and a data frame, and returns a design matrix
based on the right-hand side of the formula. If the formula is `~.`, it treats
all columns in the data as the design matrix. Otherwise, it constructs a
design matrix without an intercept based on the specified formula.
}
\examples{
data(mtcars)
# Extract a design matrix for 'mpg' and 'cyl'
X1 <- get_data_from_formula(~ mpg + cyl, mtcars)
# Extract a design matrix for all columns
X2 <- get_data_from_formula(~., mtcars)
# Extract a design matrix for 'hp'
X3 <- get_data_from_formula(~hp, mtcars)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{print.ngme_trajectories}
\alias{print.ngme_trajectories}
\title{Print method for ngme_trajectories}
\usage{
\method{print}{ngme_trajectories}(x, ...)
}
\arguments{
\item{x}{ngme_trajectories object}

\item{...}{additional arguments}
}
\value{
Invisibly returns \code{x}, an \code{ngme_trajectories} object. The
  method is called for its side effect of printing a summary of stored
  parameter trajectory matrices, including parameter names, chain count, and
  iteration count.
}
\description{
Print method for ngme_trajectories
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{print.parameter_distance}
\alias{print.parameter_distance}
\title{Print method for parameter_distance}
\usage{
\method{print}{parameter_distance}(x, ...)
}
\arguments{
\item{x}{parameter_distance object}

\item{...}{additional arguments}
}
\value{
Invisibly returns \code{x}, a numeric vector with class
  \code{parameter_distance}. The method is called for its side effect of
  printing the norm type, chain summary, number of iterations, final distance,
  and true parameter values.
}
\description{
Print method for parameter_distance
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict.ngme}
\alias{predict.ngme}
\title{Predict function of ngme2
predict using ngme after estimation}
\usage{
\method{predict}{ngme}(
  object,
  map,
  data = NULL,
  type = "lp",
  group = NULL,
  estimator = c("mean", "sd", "0.05q", "0.95q", "median", "mode"),
  sampling_size = 500,
  burnin_size = 100,
  seed = Sys.time(),
  train_idx = NULL,
  chain_combine = c("param_mean", "predictive_average"),
  return_samples = FALSE,
  ...
)
}
\arguments{
\item{object}{a ngme object}

\item{map}{a named list (or dataframe) of the locations to make the prediction}

\item{data}{a data.frame or matrix of covariates (used for fixed effects)
names(loc) corresponding to the name each latent model
vector or matrix (n * 2) for spatial coords}

\item{type}{what type of prediction, c("fe", "lp", <model_name>)
"fe" is fixed effect prediction
<model_name> is prediction of a specific model
"lp" is linear predictor (including fixed effect and all sub-models)
"response" is the linear predictor plus a fresh measurement-noise draw}

\item{group}{which filed to predict
(used for bivariate model, should be of same length as map)}

\item{estimator}{what type of estimator. Options include:
- "mean", "median", "mode", "sd": standard estimators
- "0.XXXq": any quantile specified as probability (e.g., "0.025q", "0.5q", "0.975q")}

\item{sampling_size}{size of posterior sampling}

\item{burnin_size}{size of posterior burnin}

\item{seed}{random seed}

\item{train_idx}{optional vector of training indices to use for posterior sampling.
If provided, only these indices from the original data will be used for training,
similar to cross-validation. If NULL, uses all original training data.}

\item{chain_combine}{how to combine multiple optimization chains:
\itemize{
  \item \code{"param_mean"}: default behavior using the fitted object parameters.
  \item \code{"predictive_average"}: run prediction for each optimization chain and
    average predictions across chains.
}}

\item{return_samples}{logical; when `TRUE`, attach sample draws for the
requested output in `attr(ret, "samples")`. For `type = "response"`, the
attached samples are response predictive draws.}

\item{...}{additional arguments (currently unused)}
}
\value{
a list of outputs contains estimation of operator paramters, noise parameters
}
\description{
Predict function of ngme2
predict using ngme after estimation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{create_paired_cv_splits}
\alias{create_paired_cv_splits}
\title{Create paired indices for bivariate cross-validation
Ensures that paired observations (e.g., u_wind and v_wind at same location)
are kept together in the same fold}
\usage{
create_paired_cv_splits(data, loc_col, group, k, seed = NULL)
}
\arguments{
\item{data}{data frame containing the observations}

\item{loc_col}{character vector of column names defining locations (e.g., c("lon", "lat") or c("x", "y"))}

\item{group}{character, name of the group column (e.g., "direction" for wind data)}

\item{k}{number of folds}

\item{seed}{random seed}
}
\value{
list with test_idx and train_idx, each containing k lists of indices
}
\description{
Create paired indices for bivariate cross-validation
Ensures that paired observations (e.g., u_wind and v_wind at same location)
are kept together in the same fold
}
\examples{
# Example with wind data
data_long <- data.frame(
  lon = rep(c(1, 2, 3, 4), 2),
  lat = rep(c(1, 1, 2, 2), 2),
  direction = rep(c("u_wind", "v_wind"), each = 4),
  wind = rnorm(8)
)
splits <- create_paired_cv_splits(data_long, c("lon", "lat"), "direction", k = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.constant.R
\name{ngme_model_types}
\alias{ngme_model_types}
\title{Show ngme model types}
\usage{
ngme_model_types()
}
\value{
available types for models
}
\description{
Show ngme model types
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{batch_means_estimator}
\alias{batch_means_estimator}
\title{Batch-Means Covariance Estimator for SGD Trajectories}
\usage{
batch_means_estimator(
  trajectory,
  alpha = 0.501,
  M = NULL,
  N = NULL,
  drop_burnin = TRUE,
  burnin_iter = 0
)
}
\arguments{
\item{trajectory}{numeric matrix with rows = iterations and columns = parameters.}

\item{alpha}{stepsize decay exponent in
\eqn{\eta_i = \eta i^{-\alpha}}. Must satisfy \eqn{1/2 < \alpha < 1}.}

\item{M}{number of retained batches (excluding burn-in batch 0). If `NULL`,
use \eqn{\lfloor n^{(1-\alpha)/2} \rfloor}.}

\item{N}{decorrelation constant in
\eqn{e_k = \lfloor ((k+1)N)^{1/(1-\alpha)} \rfloor}. If `NULL`, use
\eqn{N = n^{1-\alpha}/(M+1)}.}

\item{drop_burnin}{logical; if `TRUE`, discard batch 0.}

\item{burnin_iter}{non-negative integer. Explicitly discard the first
`burnin_iter` iterations before building Xi-style batches. After trimming,
batch boundaries are rebuilt from iteration 1 of the retained trajectory.}
}
\value{
A list containing the covariance estimate, pooled mean, batch sizes,
batch boundaries, and chain-level metadata.
}
\description{
Estimate the asymptotic covariance from a single SGD trajectory using the
increasing-batch construction from Xi et al. (2020).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complex-models.R
\name{bv}
\alias{bv}
\title{Ngme bivariate model specification}
\usage{
bv(
  mesh,
  sub_models,
  theta = 0,
  rho = 0,
  c1 = 1,
  c2 = 1,
  share_param = FALSE,
  fix_theta = FALSE,
  use_c_param = FALSE
)
}
\arguments{
\item{mesh}{the mesh where model is defined}

\item{sub_models}{a list of sub_models (total 2 sub_models)}

\item{theta}{the rotation parameter (starting point should be from -pi/4 to pi/4)}

\item{rho}{the parameter related to correlation}

\item{c1}{scaling factor for the first sub-model (default: 1)}

\item{c2}{scaling factor for the second sub-model (default: 1)}

\item{share_param}{TRUE if share the same parameter for 2 sub_models (of same type)}

\item{fix_theta}{TRUE if fix the theta of bv model}

\item{use_c_param}{TRUE to estimate c1 and c2 as parameters, FALSE to fix them at specified values (default: FALSE)}
}
\value{
a ngme_operator of bivariate model
}
\description{
Giving 2 sub_models, build a correlated bivariate operator based on K = D(theta, rho) %*% diag(c1*K_1, c2*K_2)
\deqn{D(\theta, \rho) = \begin{pmatrix}
  \cos(\theta) + \rho \sin(\theta) & -\sin(\theta) \sqrt{1+\rho^2} \\
  \sin(\theta) - \rho \cos(\theta) & \cos(\theta) \sqrt{1+\rho^2}
\end{pmatrix}}
}
\details{
The bivariate model combines two sub-models with a rotation matrix D and optional scaling factors c1 and c2.
When \code{use_c_param = TRUE}, c1 and c2 are estimated as parameters (on log scale).
When \code{use_c_param = FALSE} (default), c1 and c2 are fixed at their specified values (both default to 1).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/noise.R
\name{print.ngme_noise}
\alias{print.ngme_noise}
\title{Print ngme noise}
\usage{
\method{print}{ngme_noise}(x, padding = 0, prefix = "Noise type", model_type = NULL, ...)
}
\arguments{
\item{x}{noise object}

\item{padding}{number of white space padding in front}

\item{prefix}{prefix}

\item{model_type}{model type}

\item{...}{...}
}
\value{
a list (noise specifications)
}
\description{
Print ngme noise
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{ARcov}
\alias{ARcov}
\title{Helper function to compute AR(p) autocovariance}
\usage{
ARcov(rho, p)
}
\arguments{
\item{rho}{vector of AR coefficients}

\item{p}{AR order}
}
\value{
vector of autocovariances from lag 0 to lag p-1
}
\description{
Helper function to compute AR(p) autocovariance
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{stepsize_control}
\alias{stepsize_control}
\title{Unified stepsize control}
\usage{
stepsize_control(
  schedule = stepsize_schedule(method = "constant"),
  decay = stepsize_decay(method = "none")
)
}
\arguments{
\item{schedule}{schedule component. Either an object from
\code{stepsize_schedule()} or a method string accepted by
\code{stepsize_schedule()}.}

\item{decay}{decay component. Either an object from
\code{stepsize_decay()} or a method string accepted by
\code{stepsize_decay()}.}
}
\value{
a bundled stepsize-control object for \code{control_opt()}.
}
\description{
Unified stepsize control
}
\details{
Bundle an iteration schedule (\code{stepsize_schedule}) and an optional
checkpoint-based decay rule (\code{stepsize_decay}) into one object for
\code{control_opt()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{ngme_optimizers}
\alias{ngme_optimizers}
\title{List supported optimizers}
\usage{
ngme_optimizers()
}
\value{
a character vector of supported optimizers
}
\description{
This function returns a list of supported optimizers in the \code{ngme} package.
The optimizers are categorized into three groups:
\itemize{
\item \strong{Gradient descent}: \code{"sgd"}, \code{"sgld"}, \code{"momentum"}, \code{"adaptive_gd"}
\item \strong{Adaptive learning rate}: \code{"adagrad"}, \code{"rmsprop"}, \code{"adam"}, \code{"adamW"}
\item \strong{Preconditioner}: \code{"precond_sgd"}, \code{"bfgs"}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme2-package.R
\docType{package}
\name{ngme2}
\alias{ngme2-package}
\alias{ngme2}
\title{Inference and prediction for mixed effects models with flexible non-Gaussian and Gaussian distributions.}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Fits and analyzes linear latent non-Gaussian models for temporal, spatial, and space-time data. The package provides model components for autoregressive and Ornstein-Uhlenbeck processes, random walks, Matern fields based on stochastic partial differential equations, separable and non-separable space-time models, graph-based Matern models, bivariate type-G fields, and user-defined sparse operators. Latent fields and observation models can use Gaussian and non-Gaussian noise distributions, including normal inverse Gaussian, generalized asymmetric Laplace, and skew-t distributions. Functions are included for simulation, likelihood-based estimation, prediction, cross-validation, convergence diagnostics, stochastic gradient optimization, batch-means confidence intervals, and posterior-like sampling. The modeling framework is described in Bolin, Jin, Simas and Wallin (2026) "A Unified and Computationally Efficient Non-Gaussian Statistical Modeling Framework" \doi{10.48550/arXiv.2602.23987}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://davidbolin.github.io/ngme2/}
  \item \url{https://github.com/davidbolin/ngme2}
  \item Report bugs at \url{https://github.com/davidbolin/ngme2/issues}
}

}
\author{
\strong{Maintainer}: Xiaotian Jin \email{xiaotian.jin@kaust.edu.sa}

Authors:
\itemize{
  \item David Bolin \email{davidbolin@gmail.com} [copyright holder]
  \item Alexandre Simas \email{alexandre.impa@gmail.com}
  \item Jonas Wallin \email{jonas.wallin81@gmail.com}
}

Other contributors:
\itemize{
  \item Andrea V. Rocha \email{dearocha@yahoo.com} (contributed to many vignettes and examples) [contributor]
  \item Timothy A. Davis (SuiteSparse components) [contributor, copyright holder]
  \item Patrick R. Amestoy (SuiteSparse AMD and CAMD components) [contributor, copyright holder]
  \item Iain S. Duff (SuiteSparse AMD and CAMD components) [contributor, copyright holder]
  \item John K. Reid (SuiteSparse AMD-derived code) [contributor, copyright holder]
  \item Yanqing Chen (SuiteSparse CAMD components) [contributor, copyright holder]
  \item Sivasankaran Rajamanickam (SuiteSparse CCOLAMD components) [contributor, copyright holder]
  \item Stefan Larimore (SuiteSparse CCOLAMD and COLAMD components) [contributor, copyright holder]
  \item William W. Hager (SuiteSparse CHOLMOD Modify components) [contributor, copyright holder]
  \item University of Florida (SuiteSparse CHOLMOD and CCOLAMD components) [copyright holder]
  \item Regents of the University of Minnesota (METIS/GKlib components bundled with SuiteSparse) [copyright holder]
  \item Free Software Foundation, Inc. (GKlib regex/getopt code bundled with SuiteSparse) [copyright holder]
  \item Makoto Matsumoto (Mersenne Twister code bundled with SuiteSparse) [contributor, copyright holder]
  \item Takuji Nishimura (Mersenne Twister code bundled with SuiteSparse) [contributor, copyright holder]
  \item Alexander Chemeris (MS integer headers bundled with SuiteSparse) [contributor, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{re}
\alias{re}
\title{ngme random effect model}
\usage{
re(map, theta_K = NULL, ...)
}
\arguments{
\item{map}{numerical vector, covariates to build index for the process (can be formula, provided data)}

\item{theta_K}{initial value for theta_K (build covariance matrix)}

\item{...}{currently ignored}
}
\value{
ngme_operator object
}
\description{
ngme random effect model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prior.R
\name{prior_inv_exponential}
\alias{prior_inv_exponential}
\alias{prior_inv_exp}
\title{Prior Inverse-Exponential}
\usage{
prior_inv_exponential(lambda = 1, lower = 0, target = "coef")

prior_inv_exp(lambda = 1, lower = 0, target = "coef")
}
\arguments{
\item{lambda}{exponential rate on \eqn{\kappa = 1/\nu}}

\item{lower}{lower shift used in \eqn{\nu = \mathrm{lower} + \exp(\theta)}}

\item{target}{apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)}
}
\value{
prior specification
}
\description{
Prior induced by \eqn{\kappa = 1 / \nu \sim \mathrm{Exp}(\lambda)},
giving \eqn{p(\nu) = \lambda \exp(-\lambda / \nu)\nu^{-2}} for \eqn{\nu > 0}.
Internally this prior is applied to \eqn{\nu = \mathrm{lower} + \exp(\theta)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info_datasets.R
\docType{data}
\name{argo_float}
\alias{argo_float}
\title{Argo float dataset}
\format{
Data frame containing 274 observations on 4 variables.
\describe{
  \item{lat}{Latitude.}
  \item{lon}{Longitude.}
  \item{sal}{Salinity.}
  \item{temp}{Temperature.}
}
}
\source{
Data can be obtained from \href{https://argo.ucsd.edu/}{Argo float website}.
}
\usage{
data("argo_float")
}
\description{
Argo floats measurements.
}
\details{
The floats have a pressure case made of aluminium that is about 1.3m long and about 20cm diameter. They weigh about 40kg. On the top is an antenna to communicate with the satellites that fix the float's position and receive the data. Also on the top are the temperature, salinity and pressure sensors.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{momentum}
\alias{momentum}
\title{Momentum SGD optimization}
\usage{
momentum(stepsize = 0.05, beta1 = 0.9, beta2 = 1 - beta1)
}
\arguments{
\item{stepsize}{stepsize for SGD}

\item{beta1}{beta1 for momentum}

\item{beta2}{beta2 for momentum}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
Momentum SGD optimization
}
\details{
The update rule for momentum is:
\deqn{v_t = \beta_1 v_{t-1} + \beta_2 g_t}
\deqn{x_{t+1} = x_t - \text{stepsize} * v_t}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{ar}
\alias{ar}
\title{ngme AR(p) model specification}
\usage{
ar(mesh = NULL, rho = NULL, order = NULL)
}
\arguments{
\item{mesh}{integer vector or inla.mesh.1d object, index to build the mesh}

\item{rho}{vector of AR coefficients of length p (should satisfy stationarity conditions)}

\item{order}{integer, the AR order p. If provided and rho is not specified, rho will be initialized as rep(0, p)}
}
\value{
ngme_operator object
}
\description{
ngme AR(p) model specification
}
\examples{
# AR(2) model with specified coefficients
ar(c(1:10), rho = c(0.5, -0.3))
# AR(3) model with specified coefficients
ar(c(1:10), rho = c(0.4, 0.2, -0.1))
# AR(2) model with default coefficients (zeros)
ar(c(1:10), order = 2)
# AR(3) model with specified order and coefficients
ar(c(1:10), order = 3, rho = c(0.4, 0.2, -0.1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{simulate.ngme}
\alias{simulate.ngme}
\title{Simulate from a ngme object (possibly with replicates)}
\usage{
\method{simulate}{ngme}(object, nsim = 1, seed = NULL, ...)
}
\arguments{
\item{object}{ngme object}

\item{nsim}{number of simulations}

\item{seed}{seed}

\item{...}{optional arguments. Supported names are \code{posterior}
(whether to simulate from posterior sampling of latent fields) and
\code{m_noise} (whether to add the measurement noise).}
}
\value{
a realization of ngme object
}
\description{
Simulate from a ngme object (possibly with replicates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{sgld}
\alias{sgld}
\title{Stochastic Gradient Langevin Dynamics (SGLD) optimization}
\usage{
sgld(stepsize = 0.001, temperature = 1)
}
\arguments{
\item{stepsize}{base stepsize}

\item{temperature}{non-negative Langevin temperature}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
Stochastic Gradient Langevin Dynamics (SGLD) optimization
}
\details{
SGLD adds Gaussian noise to vanilla SGD updates:
\deqn{x_{t+1} = x_t - \eta_t \nabla U(x_t) + \sqrt{2 T \eta_t}\,\xi_t, \quad \xi_t \sim \mathcal{N}(0, I)}
where \eqn{T} is \code{temperature}. The implementation applies this
component-wise using the current effective stepsize.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic_ns.R
\name{generic_ns}
\alias{generic_ns}
\title{Non-stationary precision matrix operator with custom matrix combinations}
\usage{
generic_ns(
  theta_K,
  matrices,
  position,
  h,
  model = "generic_ns",
  B_theta_K = NULL,
  trans = NULL,
  mesh = NULL,
  zero_trace = FALSE,
  param_name = NULL,
  param_trans = NULL,
  ...
)
}
\arguments{
\item{theta_K}{List of parameter vectors (in real scale)}

\item{matrices}{List of fixed matrices to be used in the model}

\item{position}{List of vectors specifying matrix combinations (required)}

\item{h}{The integration weights vector}

\item{model}{The model type name stored on the returned operator}

\item{B_theta_K}{List of basis matrices for parameters, defaults to matrices of 1s if not provided}

\item{trans}{List of transformations for each parameter (one transformation per parameter)}

\item{mesh}{The mesh object}

\item{zero_trace}{Whether the trace of K should be zero}

\item{param_name}{Optional parameter names stored on the returned operator for diagnostics}

\item{param_trans}{Optional parameter transformations stored on the returned operator for diagnostics}

\item{...}{Additional arguments (ignored)}
}
\value{
An object of class \code{ngme_operator}. The object stores the sparse
  non-stationary precision matrix \code{K}, integration weights \code{h}, the
  flattened parameter vector \code{theta_K}, basis matrices
  \code{B_theta_K}, parameter grouping metadata \code{param_map}, matrix
  combination rules \code{position}, and an update function for rebuilding
  \code{K}. It is intended for use as the latent \code{model} inside
  \code{f()}.
}
\description{
Creates a flexible non-stationary precision matrix K by allowing both linear combinations
and matrix multiplications with parameter-dependent diagonal matrices. This enables modeling
spatially varying parameters through basis expansions.
}
\details{
The key distinction from `generic` is that `generic_ns`:
1. Requires explicit position specification for matrix combinations
2. Converts parameters to diagonal matrices for multiplication with fixed matrices
3. Allows for basis expansions via B_theta_K


The `generic_ns` operator constructs K through the following steps:

1. Create diagonal matrices for each parameter using basis expansions: D_param = diag(B_param * theta_param)
2. Combine parameter matrices with fixed matrices according to position specifications
3. Sum all the resulting combinations

The `position` parameter defines how matrices are combined:
- Each element of the list represents a term in the sum
- Each vector contains indices of matrices to multiply (parameter matrices first, then fixed matrices)
- For example: `position = list(c(1, 3), c(2, 4))` with one parameter means:
  Multiply the parameter diagonal matrix (index 1) with fixed matrix 1 (index 3),
  then add parameter diagonal matrix 2 (index 2) multiplied by fixed matrix 2 (index 4)

Spatially-varying parameters can be created using basis matrices in B_theta_K.
}
\examples{
# Simple example with one parameter
n <- 5
A <- matrix(1, n, n)
B <- matrix(2, n, n)
alpha <- 0.4

# Create a simple generic_ns model: D_alpha * A + B
model <- generic_ns(
  theta_K = list(alpha = c(alpha)),
  matrices = list(A, B),
  position = list(c(1, 2), c(3)), # D_alpha * A + B
  h = rep(1, n)
)

# AR1 model representation: rho * C + G
ar1_obj <- ar1(1:10, rho = 0.5)
g <- name2fun("tanh", inv = TRUE)
generic_ar1 <- generic_ns(
  theta_K = list(x = c(g(0.5))), # Transform from real scale
  trans = list(x = "tanh"), # Apply tanh transformation
  matrices = list(ar1_obj$C, ar1_obj$G),
  position = list(c(1, 2), c(3)), # D_rho * C + G
  h = ar1_obj$h,
  mesh = 1:10
)

# Spatially varying Matern model
# Create a simple 1D mesh
mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = 10))

# Create basis for spatially-varying kappa
n <- 10
B_kappa <- matrix(0, n, 2)
B_kappa[1:(n / 2), 1] <- 1 # First half of the domain
B_kappa[(n / 2 + 1):n, 2] <- 1 # Second half of the domain

# Create standard Matern components
matern_model <- matern(mesh)

# Create model with space-varying kappa
ns_model <- generic_ns(
  theta_K = list(kappa = c(log(1), log(2))), # Different values for each region
  matrices = list(matern_model$C, matern_model$G),
  B_theta_K = list(kappa = B_kappa),
  trans = list(kappa = "exp2"), # kappa^2 transformation for C
  h = matern_model$h,
  position = list(c(1, 2), c(3)), # D_kappa * C + G
  mesh = mesh
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info_datasets.R
\docType{data}
\name{cienaga}
\alias{cienaga}
\title{The swamp of Cienaga Grande in Santa Marta, Colombia}
\format{
A data frame with 218 rows and 6 columns.
\describe{
  \item{East, North}{location}
  \item{depth}{depth of the swamp}
  \item{temp}{temperature}
  \item{oxyg}{oxygen}
  \item{measurement}{1 means the first measurement, 2 the second}
}
}
\source{
..
}
\usage{
cienaga
}
\description{
There is a total of 114 locations where some properties of the
swamp were measured. Those measurements were taken twice, however there is no information available about their chronological order so this data cannot
be treated as spatiotemporal, despite that, the multiple measurements can be treated as replicates.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prior_calibration.R
\name{calibrate_inv_exp_lambda_driven_nig}
\alias{calibrate_inv_exp_lambda_driven_nig}
\alias{calibrate_inv_exp_lambda}
\title{Calibrate Inverse-Exponential Prior for NIG Driven Noise}
\usage{
calibrate_inv_exp_lambda_driven_nig(
  r_target = 2,
  alpha = 0.1,
  c = 3,
  mu = 0,
  sigma = 1,
  h = 1,
  n_samples = 1e+05,
  nu_lower = 0.1,
  nu_upper = 100,
  tol = 0.05,
  max_iter = 30,
  max_expand = 30,
  seed = NULL
)

calibrate_inv_exp_lambda(...)
}
\arguments{
\item{r_target}{target tail inflation level \eqn{r_\text{target} > 1}}

\item{alpha}{target prior probability in \eqn{(0,1)}}

\item{c}{tail threshold for \eqn{|U|>c}, typically `2.5` or `3`}

\item{mu}{NIG drift parameter in the driven noise term}

\item{sigma}{NIG scale parameter (> 0)}

\item{h}{positive increment scaling (> 0)}

\item{n_samples}{Monte Carlo sample size used per evaluation of \eqn{R_c(\nu)}}

\item{nu_lower}{initial lower bracket for \eqn{\nu}}

\item{nu_upper}{initial upper bracket for \eqn{\nu}}

\item{tol}{relative tolerance for solving \eqn{R_c(\nu_r)=r_\text{target}}}

\item{max_iter}{maximum number of bisection iterations}

\item{max_expand}{maximum bracket expansion steps on each side}

\item{seed}{optional integer seed for reproducible calibration}

\item{...}{arguments forwarded to \code{calibrate_inv_exp_lambda_driven_nig()}}
}
\value{
A list with calibrated `lambda`, solved `nu_r`, achieved `rc_nu_r`,
  and diagnostics.
}
\description{
Calibrate \eqn{\lambda} in the prior
\eqn{\kappa = 1/\nu \sim \mathrm{Exp}(\lambda)}
using a tail-inflation target for driven NIG noise.

The calibration target is
\deqn{\Pr(R_c(\nu) > r_\text{target}) = \alpha,}
where
\deqn{R_c(\nu) = \frac{\Pr(|U| > c \mid \nu)}{\Pr(|Z| > c)},\quad Z\sim N(0,1)}
and
\deqn{U = \frac{\mu(V-h) + \sigma\sqrt{V}Z}{\sigma\sqrt{h}},\quad V\sim\mathrm{GIG}(-1/2,\nu,\nu h^2).}

The function solves \eqn{R_c(\nu_r) = r_\text{target}} using log-scale
bisection, then returns
\deqn{\lambda = -\nu_r \log(\alpha).}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prior.R
\name{prior_none}
\alias{prior_none}
\title{Prior None}
\usage{
prior_none(target = "coef")
}
\arguments{
\item{target}{apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)}
}
\value{
prior specification
}
\description{
Prior None
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{merge_replicates}
\alias{merge_replicates}
\title{Merge model of replicates into model of 1 replicate given train_idx and
test_idx, the merged model contains all the information of train_idx from
different replicates.}
\usage{
merge_replicates(ngme, train_idx, test_idx)
}
\arguments{
\item{ngme}{a ngme object}

\item{train_idx}{a vector of indices of train data}

\item{test_idx}{a vector of indices of test data}
}
\description{
Merge model of replicates into model of 1 replicate given train_idx and
test_idx, the merged model contains all the information of train_idx from
different replicates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{openmp_test}
\alias{openmp_test}
\title{Test OpenMP availability and report the number of threads.}
\usage{
openmp_test()
}
\value{
Invisibly returns the detected number of OpenMP threads. A value of
  zero means OpenMP is not available.
}
\description{
This function checks if OpenMP is available in the current R environment
and, if so, reports the number of OpenMP threads detected.
If OpenMP is not available, it reports a corresponding message.
It relies on an internal or external function `get_openmp_threads()`
which is assumed to return the number of threads or 0 if unavailable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{adamW}
\alias{adamW}
\title{AdamW SGD optimization}
\usage{
adamW(
  stepsize = 0.05,
  beta1 = 0.9,
  beta2 = 0.999,
  lambda = 0.01,
  epsilon = 1e-08
)
}
\arguments{
\item{stepsize}{stepsize for SGD}

\item{beta1}{beta1 for AdamW}

\item{beta2}{beta2 for AdamW}

\item{lambda}{lambda (weight decay) for AdamW}

\item{epsilon}{epsilon for numerical stability}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
AdamW SGD optimization
}
\details{
The update rule for AdamW is:
\deqn{m_t = \beta_1 m_{t-1} + (1 - \beta_1) g_t}
\deqn{v_t = \beta_2 v_{t-1} + (1 - \beta_2) g_t^2}
\deqn{\hat{m_t} = m_t / (1 - \beta_1^t)}
\deqn{\hat{v_t} = v_t / (1 - \beta_2^t)}
\deqn{x_{t+1} = x_t - \text{stepsize} * \left( \lambda x_t + \frac{\hat{m_t}}{\sqrt{\hat{v_t}} + \epsilon} \right)}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{ngme_sgld_ci}
\alias{ngme_sgld_ci}
\title{Quantile Confidence Intervals from SGLD Samples}
\usage{
ngme_sgld_ci(samples, lower = 0.025, upper = 0.975)
}
\arguments{
\item{samples}{data.frame (or list of data.frames) returned by
[ngme_sgld_samples()].}

\item{lower}{lower quantile probability (e.g. 0.025).}

\item{upper}{upper quantile probability (e.g. 0.975).}
}
\value{
A list with posterior-like point estimates (`estimates`), quantile
intervals (`ci`), and sample covariance matrix (`covariance`).
Class: `ngme_sgld_ci`.
}
\description{
Compute parameter-wise quantile confidence intervals from posterior-like
SGLD samples returned by [ngme_sgld_samples()].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.constant.R
\name{ngme_prior_types}
\alias{ngme_prior_types}
\title{Show ngme priors}
\usage{
ngme_prior_types()
}
\value{
available types of priors
}
\description{
Show ngme priors
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{stepsize_decay}
\alias{stepsize_decay}
\title{Stepsize decay schedule}
\usage{
stepsize_decay(
  method = c("none", "grad_norm_plateau"),
  patience = 3,
  gamma = 0.5,
  min_delta = 0,
  warmup = 0,
  min_stepsize = 0
)
}
\arguments{
\item{method}{decay strategy. "none" disables decay; "grad_norm_plateau"
decays when mean grad.norm() across chains fails to decrease for a number
of epochs.}

\item{patience}{number of consecutive epochs without grad.norm() improvement
before decaying stepsize.}

\item{gamma}{decay factor applied when triggered (0 < gamma < 1).}

\item{min_delta}{minimum required decrease in grad.norm() to be counted as improvement.}

\item{warmup}{number of initial epochs to skip decay checks.}

\item{min_stepsize}{lower bound for stepsize after decay (absolute value).}
}
\value{
a list of control variables for stepsize decay (used in \code{control_opt}).
}
\description{
Stepsize decay schedule
}
\details{
Define a stepsize decay strategy for use in \code{control_opt()}.
Currently supports the plateau-based schedule using grad.norm().
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions.R
\name{ig}
\alias{ig}
\alias{dig}
\alias{rig}
\alias{pig}
\alias{qig}
\title{The Inverse-Gaussian (IG) Distribution}
\usage{
dig(x, a, b, log = FALSE)

rig(n, a, b, seed = 0)

pig(q, a, b, lower.tail = TRUE, log.p = FALSE)

qig(p, a, b, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x, q}{vector of quantiles.}

\item{a, b}{parameters \code{a} and \code{b}. Must be positive.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities/densities \eqn{p} are
returned as \eqn{log(p)}.}

\item{n, }{number of observations.}

\item{seed}{Seed for the random generation.}

\item{lower.tail}{logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
otherwise, \eqn{P[X>x]}.}

\item{p}{vector of probabilities.}
}
\value{
dig gives the density, pig gives the distribution function,
qig gives the quantile function, and rig generates random deviates.

Invalid arguments will result in return value NaN, with a warning.

The length of the result is determined by \code{n} for rig.
}
\description{
Density, distribution function, quantile function and
random generation for the inverse-Gaussian distribution
 with parameters \code{a} and \code{b}.
}
\details{
The inverse-Gaussian distribution has density given
by
 \deqn{f(x; a, b) = \frac{\sqrt{b}}{\sqrt{2\pi x^3}}\exp(
 -\frac{a}{2}x -\frac{b}{2x} + \sqrt{ab}),}
 where \eqn{x>0} and \eqn{a,b>0}. In this parameterization,
 \eqn{E(X) = \sqrt{b}/\sqrt{a}}. See Tweedie (1957a, 1957b) for
 further details.
}
\examples{
rig(100, a = 1, b = 1)
pig(0.4, a = 1, b = 1)
qig(0.8, a = 1, b = 1)
plot(function(x){dig(x, a = 1, b = 1)}, main =
"Inverse-Gaussian density", ylab = "Probability density",
xlim = c(0,10))
}
\references{
Tweedie, M. C. K. (1957a). "Statistical Properties of Inverse Gaussian Distributions I". Annals of Mathematical Statistics. 28 (2): 362–377. \doi{10.1214/aoms/1177706964}

Tweedie, M. C. K. (1957b). "Statistical Properties of Inverse Gaussian Distributions II". Annals of Mathematical Statistics. 28 (3): 696–705. \doi{10.1214/aoms/1177706881}
}
\seealso{
\code{\link{dnig}}, \code{\link{dgig}}, \code{\link{digam}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{arma11}
\alias{arma11}
\title{Convenience wrapper for ARMA(1,1)}
\usage{
arma11(mesh = NULL, ar = 0, ma = 0, fix_ar = FALSE, fix_ma = FALSE)
}
\arguments{
\item{mesh}{Integer vector or `inla.mesh.1d` object.}

\item{ar}{Numeric length-p vector of AR coefficients in (-1, 1). Defaults to zeros.}

\item{ma}{Numeric length-q vector of MA coefficients in (-1, 1). Defaults to zeros.}

\item{fix_ar}{Logical or length-p logical vector to fix AR coefficients.}

\item{fix_ma}{Logical or length-q logical vector to fix MA coefficients.}
}
\value{
An \code{ngme_operator} object describing an ARMA(1,1) latent
  operator. The object contains the AR precision component \code{K}, the MA
  filter component \code{Z}, integration weights \code{h}, and the
  unconstrained AR and MA parameters used by the optimizer. If \code{mesh} is
  \code{NULL}, returns an \code{ngme_operator_def} specification for delayed
  construction inside \code{f()}.
}
\description{
Convenience wrapper for ARMA(1,1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{bfgs}
\alias{bfgs}
\title{BFGS optimization}
\usage{
bfgs(line_search = "wolfe")
}
\arguments{
\item{line_search}{line search method, can be c("backtracking", "wofle")
"bfgs" means use BFGS preconditioner}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
BFGS optimization
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.R
\name{ngme_parse_formula}
\alias{ngme_parse_formula}
\title{Parse the formula for ngme function}
\usage{
ngme_parse_formula(
  fm,
  data,
  control_ngme,
  noise,
  group,
  replicate,
  prior_beta,
  standardize
)
}
\arguments{
\item{fm}{formula}

\item{data}{data.frame}

\item{control_ngme}{control_ngme}

\item{noise}{noise}

\item{group}{group factor}

\item{replicate}{replicate vector}

\item{prior_beta}{prior specification for fixed effects}

\item{standardize}{logical, whether fixed effects are standardized}
}
\value{
a list (replicate) of ngme_replicate models
}
\description{
Parse the formula for ngme function
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{make_time_series_cv_index}
\alias{make_time_series_cv_index}
\title{Create Time Series Cross-Validation Indices}
\usage{
make_time_series_cv_index(
  time_idx,
  train_length = NULL,
  test_length = 1,
  replicate = time_idx,
  gap = 0
)
}
\arguments{
\item{time_idx}{A numeric vector of time indices in ascending order}

\item{train_length}{An integer specifying the fixed length of training sets.
If NULL (default), an expanding window approach is used.}

\item{test_length}{An integer specifying the number of observations to include in each test set.
Default is 1 (single-step forecasting).}

\item{replicate}{An optional vector of the same length as time_idx, indicating which observations
belong to the same replicate group. When provided, ensures that all observations with the
same replicate value are either entirely in the training set or entirely in the test set.}

\item{gap}{An integer specifying the gap between the training set and test set.
Default is 0 (no gap, test set starts immediately after training set).
For example, gap=1 means skip one time point between training and test (useful for 2-step ahead forecasting).}
}
\value{
A list with two components:
\describe{
  \item{train}{A list of numeric vectors, where each vector contains the time indices for training in that fold}
  \item{test}{A list of numeric vectors, where each vector contains the time indices for testing in that fold}
}
}
\description{
Creates indices for time series cross-validation with options for expanding window
or sliding window approaches. Supports both single-step and multi-step forecasting, and can
handle replicated observations.
}
\details{
Time series cross-validation requires respecting the temporal order of observations.
This function implements two common approaches:

1. Expanding window (when train_length = NULL): The training set grows with each fold,
   starting with a minimal set and expanding to include all but the test data.

2. Sliding window (when train_length is specified): Uses a fixed-length window
   that slides through the time series, maintaining the same training size across folds.

The test_length parameter allows for multi-step forecasting evaluation.

When replicate is provided, the function ensures that all observations with the same
replicate value are kept together, either all in the training set or all in the test set.
This is useful for scenarios where multiple observations at the same time point should
be treated as a group.

The gap parameter creates a separation between training and test sets, which is useful
for multi-step ahead forecasting validation.
}
\examples{
# Expanding window approach with single-step forecasting
cv_expanding <- make_time_series_cv_index(1:10)

# Sliding window approach with window size 3 and single-step forecasting
cv_sliding <- make_time_series_cv_index(1:10, train_length = 3)

# Sliding window with multi-step forecasting (predict 2 steps ahead)
cv_multistep <- make_time_series_cv_index(1:10, train_length = 3, test_length = 2)

# Working with replicates
time_idx <- c(1, 1, 1, 2, 2, 3, 3)
replicates <- c(1, 1, 1, 2, 2, 3, 3)
cv_with_replicates <- make_time_series_cv_index(time_idx, replicate = replicates)

# 2-step ahead forecasting with a gap
cv_with_gap <- make_time_series_cv_index(1:10, gap = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operator.R
\name{print.ngme_model}
\alias{print.ngme_model}
\title{Print ngme model}
\usage{
\method{print}{ngme_model}(x, padding = 0, ...)
}
\arguments{
\item{x}{ngme model object}

\item{padding}{number of white space padding in front}

\item{...}{...}
}
\value{
a list (model specifications)
}
\description{
Print ngme model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prior.R
\name{prior_normal}
\alias{prior_normal}
\title{Prior Normal}
\usage{
prior_normal(mean = 0, sd = 1, target = "coef")
}
\arguments{
\item{mean}{prior mean}

\item{sd}{prior standard deviation}

\item{target}{apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)}
}
\value{
prior specification
}
\description{
Prior Normal
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info_datasets.R
\docType{data}
\name{pm25_quarterly_2022}
\alias{pm25_quarterly_2022}
\title{Quarterly PM2.5 monitor averages over CONUS in 2022}
\format{
A data frame with 2640 rows and 9 columns.
\describe{
  \item{site_id}{EPA monitor site identifier.}
  \item{state}{State name.}
  \item{county}{County name.}
  \item{lat}{Monitor latitude in decimal degrees.}
  \item{lon}{Monitor longitude in decimal degrees.}
  \item{quarter}{Calendar quarter, coded as \code{Q1}, \code{Q2}, \code{Q3}, or \code{Q4}.}
  \item{pm25_q}{Quarterly average PM2.5 concentration in micrograms per cubic meter.}
  \item{n_days}{Number of daily monitor summaries contributing to the quarterly average.}
  \item{replicate}{Integer replicate index corresponding to quarter.}
}
}
\source{
US EPA Air Quality System Data Mart and EPA AirData pre-generated
files for PM2.5 monitor summaries.
\url{https://www.epa.gov/aqs}
\url{https://aqs.epa.gov/aqsweb/airdata/download_files.html}
}
\usage{
data("pm25_quarterly_2022")
}
\description{
A processed example dataset containing quarterly fine particulate matter
(PM2.5) averages for monitoring sites in the contiguous United States.
The data are derived from the US EPA Air Quality System and AirData
pre-generated monitor files and aggregated to site-quarter means.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{batch_means_ci}
\alias{batch_means_ci}
\title{Pooled Batch-Means Confidence Intervals from Multiple Chains}
\usage{
batch_means_ci(
  chain_trajectories,
  level = 0.95,
  alpha = 0.501,
  M = NULL,
  N = NULL,
  drop_burnin = TRUE,
  burnin_iter = 0
)
}
\arguments{
\item{chain_trajectories}{list of numeric matrices, each with rows = iterations
and columns = parameters.}

\item{level}{confidence level.}

\item{alpha}{stepsize decay exponent in \eqn{(1/2, 1)}.}

\item{M}{number of retained batches per chain (excluding burn-in batch 0).}

\item{N}{decorrelation constant used in the batch boundary formula.}

\item{drop_burnin}{logical; if `TRUE`, discard batch 0.}

\item{burnin_iter}{non-negative integer. Explicitly discard the first
`burnin_iter` iterations of each chain before Xi-style batching.}
}
\value{
A list with pooled estimates, covariance, standard errors, and
confidence intervals.
}
\description{
Compute pooled point estimates, covariance, and Wald confidence intervals from
multiple chain trajectories.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{traceplot}
\alias{traceplot}
\title{Trace plot of ngme fitting}
\usage{
traceplot(
  ngme,
  name = "all",
  moving_window = 1,
  hline = NULL,
  combine = TRUE,
  ncol = NULL
)
}
\arguments{
\item{ngme}{ngme object}

\item{name}{name of latent models, otherwise plot fixed effects and measurement noise.
  Use \code{"all"} to plot all latent models plus data parameters in one figure.
should be in names(ngme$models) or other}

\item{moving_window}{moving window for the traceplot}

\item{hline}{vector, add hline to each plot}

\item{combine}{bool, if TRUE, return a single faceted ggplot; otherwise return
a list of ggplot objects and print them one by one.}

\item{ncol}{number of facet columns when \code{combine = TRUE}. If \code{NULL}
(default), use 2 columns for multi-parameter plots and 1 for single-parameter plots.}
}
\value{
A ggplot object when \code{combine = TRUE}; otherwise a list of ggplot
  objects. The mean parameter trajectories are stored in the \code{avg_lines}
  attribute of the returned object.
}
\description{
Trace plot of ngme fitting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{plot.parameter_distance}
\alias{plot.parameter_distance}
\title{Plot method for parameter_distance}
\usage{
\method{plot}{parameter_distance}(x, ...)
}
\arguments{
\item{x}{parameter_distance object}

\item{...}{additional arguments for ggplot}
}
\value{
A \code{ggplot} object showing the parameter-distance trajectory over
  optimization iterations. The y-axis is the stored distance
  \eqn{||theta - \hat{theta}||}; plot annotations report the norm and
  across-chain summary used when \code{x} was computed.
}
\description{
Plot method for parameter_distance
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme_replicate.R
\name{print.ngme_replicate}
\alias{print.ngme_replicate}
\title{Print ngme object}
\usage{
\method{print}{ngme_replicate}(x, ...)
}
\arguments{
\item{x}{ngme object}

\item{...}{ignored}
}
\value{
Invisibly returns \code{x}, an \code{ngme_replicate} object. The
  method is called for its side effect of printing a replicate summary.
}
\description{
Print ngme object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.R
\name{compute_index_corr_from_map}
\alias{compute_index_corr_from_map}
\title{Helper function to compute the index_corr vector}
\usage{
compute_index_corr_from_map(map, eps = 0.1)
}
\arguments{
\item{map}{used as location to compute distance, can be 1d (numerical) or 2d (data.frame)}

\item{eps}{threshold to determine if two points are close (if close, we consider them as the same point)}
}
\value{
the index_corr vector for ngme correlated measurement noise
}
\description{
Helper function to compute the index_corr vector
}
\examples{
x_coord <- c(1.11, 1.12, 2, 1.3, 1.3)
y_coord <- c(2.11, 2.11, 2, 3.3, 3.3)
coord <- data.frame(x_coord, y_coord)
compute_index_corr_from_map(map = coord, 0.1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{adaptive_gd}
\alias{adaptive_gd}
\title{Adaptive gradient descent}
\usage{
adaptive_gd(stepsize = 0.01)
}
\arguments{
\item{stepsize}{initial stepsize for SGD}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
Adaptive gradient descent optimizer.
}
\details{
Based on the method described in
\url{https://arxiv.org/pdf/1910.09529}.
The update rule for adaptive gradient descent is:
\deqn{\lambda_k = \min(\sqrt{1 + \theta_{k-1}} \lambda_{k-1}, \frac{||x_k - x_{k-1}||}{2 ||\nabla f(x_k) - \nabla f(x_{k-1})||} )}
\deqn{x_{k+1} = x_k - \lambda_k \nabla f(x_k)}
\deqn{\theta_k = \lambda_k / \lambda_{k-1}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{simulate.ngme_model}
\alias{simulate.ngme_model}
\title{Simulate latent process with noise}
\usage{
\method{simulate}{ngme_model}(object, nsim = 1, seed = NULL, ...)
}
\arguments{
\item{object}{ngme model specified by f() function}

\item{nsim}{number of simulations}

\item{seed}{seed}

\item{...}{ignored}
}
\value{
a realization of latent model
}
\description{
Simulate latent process with noise
}
\examples{
simulate(f(1:10, model = ar1(rho = 0.4), noise = noise_nig()))
simulate(f(rnorm(10), model = rw1(), noise = noise_normal()))
simulate(f(1:10, model = ar1(rho = 0.4), noise = noise_t(nu = 5)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions.R
\name{igam}
\alias{igam}
\alias{digam}
\alias{rigam}
\alias{pigam}
\alias{qigam}
\title{The Inverse-Gamma (IGam) Distribution}
\usage{
digam(x, a, b, log = FALSE)

rigam(n, a, b)

pigam(q, a, b, lower.tail = TRUE, log.p = FALSE)

qigam(p, a, b, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x, q}{vector of quantiles.}

\item{a, b}{parameters \code{a} and \code{b}. Must be positive.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities/densities \eqn{p} are
returned as \eqn{log(p)}.}

\item{n, }{number of observations.}

\item{lower.tail}{logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
otherwise, \eqn{P[X>x]}.}

\item{p}{vector of probabilities.}
}
\value{
digam gives the density, pigam gives the distribution function,
qigam gives the quantile function, and rigam generates random deviates.

Invalid arguments will result in return value NaN, with a warning.

The length of the result is determined by \code{n} for rig.
}
\description{
Density, distribution function, quantile function and
random generation for the inverse-Gamma distribution
 with parameters \code{a} and \code{b}.
}
\details{
The inverse-Gamma distribution has density given
by
 \deqn{f(x; a, b) = \frac{b^a}{\Gamma(a)}x^{-a-1}\exp(
 -\frac{b}{x}),}
 where \eqn{x>0} and \eqn{a,b>0}.
}
\examples{
rigam(100, a = 1, b = 1)
pigam(0.4, a = 1, b = 1)
qigam(0.8, a = 1, b = 1)
plot(function(x){digam(x, a = 1, b = 1)}, main =
"Inverse-Gamma density", ylab = "Probability density",
xlim = c(0,10))
}
\seealso{
\code{\link{dnig}}, \code{\link{dgig}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complex-models.R
\name{tp}
\alias{tp}
\title{ngme tensor-product model specification}
\usage{
tp(first, second)
}
\arguments{
\item{first}{ngme_operator, left side of kronecker model (usually a temporal or iid model)}

\item{second}{ngme_operator, right side of kronecker model (ususally a temporal or spatial model)}
}
\value{
a list of specification of model
}
\description{
Given 2 operator (first and second), build a tensor-product operator based on K = K_first x K_second (here x is Kronecker product)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/f.R
\name{f}
\alias{f}
\title{Specifying a latent process model (wrapper function for each model)}
\usage{
f(
  map,
  model,
  noise = noise_normal(),
  name = "field",
  data = NULL,
  group = NULL,
  which_group = NULL,
  replicate = NULL,
  A = NULL,
  W = NULL,
  fix_W = FALSE,
  fix_K = FALSE,
  prior = NULL,
  subset = rep(TRUE, length_map(map)),
  debug = FALSE
)
}
\arguments{
\item{map}{symbol or numerical value: index or covariates to build index}

\item{model}{an \code{ngme_operator} or \code{ngme_operator_def} created by
model constructors such as \code{ar1()}, \code{matern()}, \code{bv()}, etc.}

\item{noise}{ngme_noise object, noise_nig() or noise_gal()}

\item{name}{name of the field, for later use, if not provided, will be "field1" etc.}

\item{data}{specifed or inherit from ngme() function}

\item{group}{group factor indicate resposne variable, can be inherited from ngme() function, (used for bivariate model)}

\item{which_group}{belong to which group}

\item{replicate}{factor indicating replicate structure for INLA-style replicates.
When provided, operators and A matrices will be block-diagonalized across replicates.}

\item{A}{observation matrix, automatically computed given map and model}

\item{W}{starting value of the process}

\item{fix_W}{stop sampling for W}

\item{fix_K}{fix the estimation of parameter of K}

\item{prior}{prior specification created by \code{prior_*()} or
\code{priors(...)} for operator parameters (\code{theta_K}).}

\item{subset}{subset of the model}

\item{debug}{debug mode}
}
\value{
a list for constructing latent model, e.g. A, h, C, G,
which also has
1. Information about K matrix
2. Information about noise
3. Control variables
}
\description{
Function used for defining of smooth and spatial terms
within ngme model formulae.
The function is a wrapper function for specific submodels.
(see ngme_models_types() for available models).
}
\details{
When using different meshes for different replicates, provide the mesh parameter
as a list of mesh objects. The number of meshes should match the number of replicates.
For example: \code{mesh = list(mesh1, mesh2, mesh3)} for 3 replicates.

When using the `replicate` argument, the operator matrices will be block-diagonalized.
For a generic operator K = param1 * Matrix1 + param2 * Matrix2, with 2 replicates,
the resulting operator becomes:
K = param1 * bdiag(Matrix1_rep1, Matrix1_rep2) + param2 * bdiag(Matrix2_rep1, Matrix2_rep2)
Similarly, the A matrix will be block-diagonalized as bdiag(A_rep1, A_rep2, ...).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{get_noise_param}
\alias{get_noise_param}
\title{get the parameters of the noise}
\usage{
get_noise_param(noise, prefix)
}
\arguments{
\item{noise}{noise object}

\item{prefix}{prefix of the parameter}
}
\value{
a list with parameter name as key and parameter value as value
}
\description{
Get the parameters of the noise
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control_ngme}
\alias{control_ngme}
\title{Generate control specifications for the ngme model}
\usage{
control_ngme(
  n_gibbs_samples = 5,
  fix_beta = FALSE,
  n_post_samples = 100,
  beta_init = NULL,
  debug = FALSE,
  ...
)
}
\arguments{
\item{n_gibbs_samples}{number of gibbs samples at each iteration}

\item{fix_beta}{logical, fix fixed effect estimation}

\item{n_post_samples}{number of posterior samples, see ?ngme_post_samples()}

\item{beta_init}{fixed effect initial value on the original design
scale. If \code{control_opt(standardize_fixed = TRUE)} (the default), the
vector is internally rotated/scaled to match the SVD-standardized design
matrix before being passed to the optimizer. Provide values on the raw
design scale; no manual rescaling is required.}

\item{debug}{debug mode}

\item{...}{additional arguments. Legacy aliases \code{feff} and
\code{fix_feff} are still recognized and mapped to \code{beta_init} and
\code{fix_beta}.}
}
\value{
a list of control variables for ngme
}
\description{
Generate control specifications for the ngme model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complex-models.R
\name{spacetime}
\alias{spacetime}
\title{Ngme space-time non-separable model specification}
\usage{
spacetime(
  mesh,
  lambda = 1,
  alpha = 2,
  cc = 1,
  kappa = 1,
  fix_gamma = FALSE,
  theta_gamma_x = 0,
  theta_gamma_y = 0,
  shared_theta_gamma = FALSE,
  B_gamma_x = matrix(1, nrow = mesh[[2]]$n, ncol = 1),
  B_gamma_y = matrix(1, nrow = mesh[[2]]$n, ncol = 1),
  B_gamma_x_list = NULL,
  B_gamma_y_list = NULL,
  stabilization = TRUE
)
}
\arguments{
\item{mesh}{A list of two objects:
\itemize{
  \item mesh_t - The temporal mesh
  \item mesh_s - The spatial mesh
}}

\item{lambda}{The spatial damping parameter.}

\item{alpha}{2 or 4, SPDE smoothness parameter.}

\item{cc}{Parameter c in the SPDE.}

\item{kappa}{Kappa parameter from Matern SPDE.}

\item{fix_gamma}{TRUE if fix gamma (advection term), FALSE if estimate gamma.}

\item{theta_gamma_x}{The x component of the advection term: \code{gamma_x = B_gamma_x \%*\% theta_gamma_x}.}

\item{theta_gamma_y}{The y component of the advection term: \code{gamma_y = B_gamma_y \%*\% theta_gamma_y}.}

\item{shared_theta_gamma}{TRUE if share the same theta_gamma for all time nodes. (theta_gamma_x and theta_gamma_y will be the same)}

\item{B_gamma_x}{The design matrix for the x component of the advection term.}

\item{B_gamma_y}{The design matrix for the y component of the advection term.}

\item{B_gamma_x_list}{A list of design matrices for the x component of the advection term on every time node, length(B_gamma_x_list) == nt-1.}

\item{B_gamma_y_list}{A list of design matrices for the y component of the advection term on every time node, length(B_gamma_y_list) == nt-1.}

\item{stabilization}{TRUE if using a stabilization term (for implicit Euler).}
}
\value{
ngme_operator object.
}
\description{
Implements a non-separable space-time model based on the advection-diffusion SPDE with first-order derivative in time.
The model combines temporal and spatial components through a finite difference method (implicit Euler) for temporal discretization
and finite element method (continuous Galerkin) for spatial discretization. When advection dominates diffusion,
the "Streamline Diffusion" stabilization technique is applied.
}
\details{
The model is particularly useful for large space-time datasets in environmental science, offering computationally
efficient methods for parameter estimation, kriging prediction, and conditional simulations.

For details, see \doi{10.1016/j.spasta.2024.100847}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulation.R
\name{simulate.ngme_noise}
\alias{simulate.ngme_noise}
\title{Simulate ngme noise object}
\usage{
\method{simulate}{ngme_noise}(object, nsim = 1, seed = NULL, h = NULL, ...)
}
\arguments{
\item{object}{ngme noise object}

\item{nsim}{number of simulations}

\item{seed}{seed}

\item{h}{should be of same length as nsim}

\item{...}{ignored}
}
\value{
data.frame (each col is a realization)
}
\description{
Simulate ngme noise object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R
\name{generic}
\alias{generic}
\title{Generic precision matrix operator}
\usage{
generic(
  theta_K,
  matrices,
  h,
  trans = NULL,
  mesh = NULL,
  zero_trace = FALSE,
  model = "generic",
  param_name = NULL,
  param_trans = NULL,
  ...
)
}
\arguments{
\item{theta_K}{The parameter vector (in real scale). If missing, initialized as double(0)}

\item{matrices}{A list of matrices to be combined with parameter-dependent coefficients}

\item{h}{The integration weights vector}

\item{trans}{Transformation for each parameter. Must be a list where each element
is a character vector defining transformations for each matrix.
For example, `trans=list(kappa=c("exp2", "identity", "null"))` means parameter 'kappa'
affects matrix 1 with exp2 transformation, matrix 2 with identity, and doesn't affect matrix 3.
Available transformations: "identity", "exp", "exp2", "exp4", "sqrt", "square", "log", "tanh"}

\item{mesh}{The mesh object}

\item{zero_trace}{Whether the trace of K should be zero}

\item{model}{The model type name}

\item{param_name}{Optional parameter names stored on the returned operator for diagnostics}

\item{param_trans}{Optional parameter transformations stored on the returned operator for diagnostics}

\item{...}{Additional arguments (ignored)}
}
\value{
An object of class \code{ngme_operator}. The object stores the sparse
  precision matrix \code{K}, integration weights \code{h}, parameter vector
  \code{theta_K}, base \code{matrices}, transformation specification
  \code{trans}, and an \code{update_K} function that rebuilds \code{K} for new
  parameter values. It is intended for use as the latent \code{model} inside
  \code{f()}.
}
\description{
Creates a flexible precision matrix K by allowing linear combinations of base matrices
with parameter-dependent coefficients. The model follows the form:
K = sum_i f_i(theta) * matrices_i, where f_i depends on the parameter transformation.
}
\details{
This function can be used to represent various models like AR1, Matern, and RW1 in a unified framework.


The `generic` function provides a flexible way to construct precision matrices by combining 
existing matrices with transformed parameters. Here's how it works:

1. Each parameter in `theta_K` can affect one or more matrices in the `matrices` list
2. The `trans` list defines how each parameter transforms before multiplying each matrix
3. For each matrix, the coefficients from all parameters are multiplied together
4. The final K is the sum of all transformed matrices

Common use cases:

- **AR1 model**: K = rho * C + G (where rho is between -1 and 1)
- **Matern (alpha=2)**: K = kappa^2 * C + G (where kappa > 0)
- **Matern (alpha=4)**: K = kappa^4 * C + 2*kappa^2 * G + G * Cinv * G
}
\examples{
# AR1 model with rho = 0.5
ar1_obj <- ar1(1:10, rho = 0.5)
g <- name2fun("tanh", inv = TRUE)
generic_ar1 <- generic(
  theta_K = c(rho = g(0.5)),  # Transform from real scale
  trans = list(rho = c("tanh", "null")),  # Apply tanh to first matrix
  matrices = list(ar1_obj$C, ar1_obj$G),
  h = ar1_obj$h
)

# Matern model with kappa = 2
mesh <- fmesher::fm_mesh_1d(seq(0, 1, length.out = 10))
matern_obj <- matern(mesh)
log_kappa <- log(2)
generic_matern <- generic(
  theta_K = c(kappa = log_kappa),
  trans = list(kappa = c("exp2", "null")),  # exp(2*kappa) for C, nothing for G
  matrices = list(matern_obj$C, matern_obj$G),
  h = matern_obj$h
)

# Matern model with alpha = 4 and kappa = 2
set.seed(1)
mesh <- fmesher::fm_mesh_2d(cbind(x = runif(20), y = runif(20)))
matern_obj <- matern(mesh, alpha = 4)
C <- matern_obj$C
G <- matern_obj$G
Cinv <- C; diag(Cinv) <- 1 / Matrix::diag(C)

generic_matern4 <- generic(
  theta_K = c(kappa = log(2)),
  trans = list(kappa = c("exp4", "exp2", "null")),
  matrices = list(C, 2*G, G \%*\% Cinv \%*\% G),
  h = matern_obj$h
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{ngme_batch_ci}
\alias{ngme_batch_ci}
\title{Batch-Means Confidence Intervals from an ngme Fit}
\usage{
ngme_batch_ci(
  ngme,
  name = "all",
  level = 0.95,
  alpha = 0.501,
  M = NULL,
  N = NULL,
  apply_transform = TRUE,
  drop_burnin = TRUE,
  burnin_iter = 0
)
}
\arguments{
\item{ngme}{fitted `ngme` object with `store_traj = TRUE` during optimization.}

\item{name}{either `"all"` (default) to use all parameters jointly, a latent
model name, or `"general"` for measurement-noise/fixed-effect block.}

\item{level}{confidence level.}

\item{alpha}{stepsize decay exponent in \eqn{(1/2, 1)}.}

\item{M}{number of retained batches (excluding burn-in batch 0).}

\item{N}{decorrelation constant in the batch boundary formula.}

\item{apply_transform}{logical; if `TRUE`, inference is first computed in the
raw optimization space and then mapped to user scale via post-hoc Delta
method (Post-delta).}

\item{drop_burnin}{logical; if `TRUE`, discard batch 0.}

\item{burnin_iter}{non-negative integer. Explicitly discard the first
`burnin_iter` optimization iterations before Xi-style batching.}
}
\value{
Same structure as [batch_means_ci()] with extra metadata fields
`name` and `apply_transform`.
}
\description{
Compute Xi-style batch-means confidence intervals directly from trajectories
stored in an `ngme` object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{ngme_post_samples}
\alias{ngme_post_samples}
\title{posterior samples of different latent models}
\usage{
ngme_post_samples(ngme_object, model_name = 1, type = "W", replicate = 1)
}
\arguments{
\item{ngme_object}{ngme object}

\item{model_name}{name of the model, or index of the model}

\item{type}{type of samples, "W" or "V"}

\item{replicate}{which replicate}
}
\value{
a data.frame of posterior samples (mesh_size * n_post_samples)
}
\description{
Extract the posterior samples of different latent models
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.R
\name{summary.ngme}
\alias{summary.ngme}
\title{Summary of ngme fit result}
\usage{
\method{summary}{ngme}(object, name = NULL, ...)
}
\arguments{
\item{object}{an object of class \code{ngme}}

\item{name}{name of the latent model to be summarized (if NULL, will print all)}

\item{...}{other arguments}
}
\value{
a list of summary
}
\description{
Summary of ngme fit result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/control.R
\name{control_opt}
\alias{control_opt}
\title{Generate control specifications for \code{ngme()} function.}
\usage{
control_opt(
  seed = Sys.time(),
  burnin = 100,
  iterations = 500,
  estimation = TRUE,
  standardize_fixed = TRUE,
  n_batch = 10,
  iters_per_check = iterations/n_batch,
  optimizer = adam(),
  start = NULL,
  start_sd = 0.5,
  n_parallel_chain = 4,
  max_num_threads = n_parallel_chain,
  print_check_info = FALSE,
  max_relative_step = 0.5,
  max_absolute_step = 0.5,
  rao_blackwellization = FALSE,
  n_trace_iter = 10,
  sampling_strategy = "all",
  solver_backend = "cholmod",
  solver_type = "llt",
  verbose = FALSE,
  store_traj = TRUE,
  robust = FALSE,
  stepsize_control = NULL,
  n_min_batch = min(n_batch, 3),
  n_slope_check = min(n_batch, 3),
  trend_std_conv_check = TRUE,
  std_lim = 0.01,
  trend_lim = 0.01,
  R_hat_conv_check = TRUE,
  max_R_hat = 1.1,
  pflug_conv_check = TRUE,
  pflug_alpha = 0.9
)
}
\arguments{
\item{seed}{set the seed for pesudo random number generator}

\item{burnin}{interations for burn-in periods (before optimization)}

\item{iterations}{optimization iterations}

\item{estimation}{run the estimation process (call C++ in backend)}

\item{standardize_fixed}{whether or not standardize the fixed effect. When
\code{TRUE} (default), the design matrix is SVD-standardized internally and
any user-supplied \code{control_ngme(beta_init = ...)} is automatically
mapped from the original design scale to that standardized basis. Set to
\code{FALSE} to keep both the design matrix and any provided
\code{beta_init} on
their original scale.}

\item{n_batch}{number of checkpoints; optimization is split into \code{n_batch} equal batches}

\item{iters_per_check}{run how many iterations between each check point (or specify \code{n_batch})}

\item{optimizer}{choose different sgd optimization method,
currently support "sgd", "sgld", "precond_sgd", "momentum", "adagrad", "rmsprop", "adam", "adamW"
see ?sgd, ?precond_sgd, ?momentum, ?adagrad, ?rmsprop, ?adam, ?adamW}

\item{start}{deprecated guard argument. Do not pass model starts through
\code{control_opt()}; use \code{ngme(..., start = previous_fit)} instead.}

\item{start_sd}{standard deviation of the initial parameter (1st chain fixed, other chains random), set 0 to be fixed for all chains}

\item{n_parallel_chain}{number of parallel chains}

\item{max_num_threads}{maximum number of threads used for parallel computing, by default will be set same as n_parallel_chain.
If it is more than n_parallel_chain, the rest will be used to parallel different replicates of the model.}

\item{print_check_info}{print the convergence information}

\item{max_relative_step}{max relative step allowed in 1 iteration}

\item{max_absolute_step}{max absolute step allowed in 1 iteration}

\item{rao_blackwellization}{use rao_blackwellization}

\item{n_trace_iter}{use how many iterations to approximate the trace (Hutchinson’s trick)}

\item{sampling_strategy}{subsampling method of replicates of model, c("all", "ws")
"all" means using all replicates in each iteration,
"ws" means weighted sampling (each iteration use 1 replicate to compute the gradient, the sample probability is proption to its number of observations)}

\item{solver_backend}{backend in ("eigen", "cholmod", "pardiso")}

\item{solver_type}{factorization type: "llt" or "ldlt"}

\item{verbose}{print estimation}

\item{store_traj}{store the optimizer trajectory for diagnostics (set FALSE to reduce memory)}

\item{robust}{use robust mode in the backend optimizer/model updates}

\item{stepsize_control}{unified stepsize configuration created by
\code{stepsize_control()}, \code{poly_decay()}, or \code{batch_decay()}.
For polynomial schedule, \code{poly_decay(..., burnin_iter = B)} keeps
schedule scale at 1 for the first \code{B} iterations, then starts decay
with reset local time index.}

\item{n_min_batch}{minimum number of checkpoints before any convergence diagnostic is attempted}

\item{n_slope_check}{number of checkpoints used as the regression window for the trend test}

\item{trend_std_conv_check}{enable the trend/std diagnostic (uses \code{std_lim}, \code{trend_lim}, \code{n_slope_check})}

\item{std_lim}{maximum allowed standard deviation}

\item{trend_lim}{maximum allowed slope}

\item{R_hat_conv_check}{use the R-hat diagnostic for convergence checking}

\item{max_R_hat}{maximum allowed R_hat}

\item{pflug_conv_check}{use Pflug diagnostic for convergence check}

\item{pflug_alpha}{scaling factor (0-1] for Pflug criterion: require \code{pflug_sum < pflug_alpha * max_pflug_sum}}
}
\value{
list of control variables
}
\description{
These are configurations for \code{ngme}
optimization process.
}
\details{
Convergence diagnostics (multi-chain):
* R-hat: per-parameter Gelman–Rubin statistic; passes if \code{R_hat <= max_R_hat}.
* Trend/Std: uses the last \code{n_slope_check} checkpoints after at least \code{n_min_batch} batches.
  Passes when both the relative std (\code{sqrt(var)/|mean| <= std_lim}) and linear trend
  of the means (\code{|slope| <= trend_lim}) satisfy their thresholds.
* Pflug: per-chain criterion \code{pflug_sum < pflug_alpha * max_pflug_sum} in the latest batch;
  if all chains satisfy it, overall convergence is declared.
Checks are evaluated every \code{iters_per_check = iterations / n_batch}. A parameter is marked
converged if any enabled parameter-level diagnostic (R-hat or Trend/Std) passes; the run stops
when all parameters converge or when the Pflug diagnostic triggers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/test.ngme.R
\name{test_ngme}
\alias{test_ngme}
\title{Test ngme function}
\usage{
test_ngme(
  model,
  n_obs_per_rep,
  n_replicate = 1,
  control_opt = NULL,
  numer_grad = TRUE,
  f_noise = noise_nig(mu = -2, sigma = 1.5, nu = 0.5),
  n_gibbs_samples = 5,
  family = "normal",
  max.n = 1000,
  debug = FALSE,
  debug_f = FALSE,
  start = NULL,
  fix_theta_K = FALSE,
  fix_theta_mu = FALSE,
  fix_theta_sigma = FALSE,
  fix_theta_nu = FALSE,
  seed = Sys.time()
)
}
\arguments{
\item{model}{model name}

\item{n_obs_per_rep}{number of observation per replicate}

\item{n_replicate}{number of replicate}

\item{control_opt}{control options for optimization, see \code{\link{control_opt}}}

\item{numer_grad}{numerical gradient}

\item{f_noise}{noise function}

\item{n_gibbs_samples}{number of gibbs samples}

\item{family}{family of noise}

\item{max.n}{maximum number for building mesh}

\item{debug}{debug mode}

\item{debug_f}{debug mode for latent process}

\item{start}{start value for optimization}

\item{fix_theta_K}{fix theta_K}

\item{fix_theta_mu}{fix theta_mu}

\item{fix_theta_sigma}{fix theta_sigma}

\item{fix_theta_nu}{fix nu}

\item{seed}{seed}
}
\description{
Test ngme function for different models
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generic.R
\name{name2fun}
\alias{name2fun}
\title{Convert transformation name to function}
\usage{
name2fun(trans, inv = FALSE)
}
\arguments{
\item{trans}{Character string specifying the transformation type from original scale to real scale}

\item{inv}{Logical, if TRUE returns the transformation function from real scale to original scale}
}
\value{
A function that applies the specified transformation
}
\description{
This function converts a transformation name to the corresponding function.
Convertion is from original scale to real scale by default.
Available transformations:
\itemize{
  \item \code{exp4}: \eqn{exp(4x)}, inverse is \eqn{log(x)/4}
  \item \code{exp2}: \eqn{exp(2x)}, inverse is \eqn{log(x)/2}
  \item \code{tanh}: Hyperbolic tangent transformation used for AR1 parameter, uses ar1_th2a and ar1_a2th
  \item \code{identity}: Identity function, no transformation
  \item \code{exp}: \eqn{exp(x)}, inverse is \eqn{log(x)}
  \item \code{sqrt}: \eqn{sqrt(x)}, inverse is \eqn{x^2}
  \item \code{square}: \eqn{x^2}, inverse is \eqn{sqrt(x)}
  \item \code{log}: \eqn{log(x)}, inverse is \eqn{exp(x)}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prior.R
\name{prior_half_cauchy}
\alias{prior_half_cauchy}
\title{Prior Half-Cauchy}
\usage{
prior_half_cauchy(scale = 1, target = "coef")
}
\arguments{
\item{scale}{half-Cauchy scale}

\item{target}{apply prior on coefficient scale (`"coef"`) or field scale (`"field"`)}
}
\value{
prior specification
}
\description{
Prior Half-Cauchy
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complex-models.R
\name{bv_matern}
\alias{bv_matern}
\title{Ngme bivariate model with Matern sub_models}
\usage{
bv_matern(
  mesh,
  sub_models,
  theta = 0,
  rho = 0,
  sd1 = 1,
  sd2 = 1,
  group = NULL,
  share_param = FALSE,
  fix_theta = FALSE,
  ...
)
}
\arguments{
\item{mesh}{the mesh where model is defined}

\item{sub_models}{a list of sub_models (should be two matern models)}

\item{theta}{the parameter related to rotation}

\item{rho}{the parameter related to correlation}

\item{sd1}{scaling parameter for the first sub-model}

\item{sd2}{scaling parameter for the second sub-model}

\item{group}{group vector, can be inherited from ngme() function}

\item{share_param}{TRUE if share the same parameter for 2 sub_models (of same type)}

\item{fix_theta}{TRUE if the rotation parameter is fixed}

\item{...}{ignore}
}
\value{
a list of specification of model
}
\description{
Giving 2 sub_models, build a correlated bivaraite operator based on K = D(theta, eta) %*% diag(K_1, K_2)
\deqn{D(\theta, \rho) = \begin{pmatrix}
  \cos(\theta) + \rho \sin(\theta) & -\sin(\theta) \sqrt{1+\rho^2} \\
  \sin(\theta) - \rho \cos(\theta) & \cos(\theta) \sqrt{1+\rho^2}
\end{pmatrix}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{ngme_cov_matrix}
\alias{ngme_cov_matrix}
\title{variance of the data or the latent field}
\usage{
ngme_cov_matrix(ngme_object, model_name = "data", replicate = 1)
}
\arguments{
\item{ngme_object}{ngme_model}

\item{model_name}{if model_name = "data", then return the covariance matrix of the data (without measurement noise)
if the model_name is the name or index of the latent, then return the covariance matrix of the latent field}

\item{replicate}{which replicate (default = 1)}
}
\value{
a data.frame of posterior samples (mesh_size * n_post_samples)
}
\description{
Compute the variance of the data or the latent field
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.constant.R
\name{ngme_noise_types}
\alias{ngme_noise_types}
\title{Show ngme noise types}
\usage{
ngme_noise_types()
}
\value{
available types for noise
}
\description{
Show ngme noise types
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validation.R
\name{cross_validation}
\alias{cross_validation}
\title{Compute the cross-validation for the ngme model
Perform cross-validation for ngme model
first into sub_groups (a list of target, and train data)}
\usage{
cross_validation(
  ngme,
  type = "k-fold",
  seed = NULL,
  print = TRUE,
  N_sim = 5,
  n_gibbs_samples = 500,
  n_burnin = 100,
  k = 5,
  percent = 0.2,
  times = 10,
  metric = NULL,
  test_idx = NULL,
  train_idx = NULL,
  keep_pred = FALSE,
  parallel = FALSE,
  thining_gap = 1,
  cores_layer1 = if (parallel) min(parallel::detectCores(), 2) else 1,
  cores_layer2 = if (parallel) min(parallel::detectCores(), 2) else 1,
  merge_groups = FALSE,
  merged_group_name = NULL,
  data = NULL,
  chain_combine = c("param_mean", "predictive_average")
)
}
\arguments{
\item{ngme}{a ngme object, or a list of ngme object (if comparing multiple models)}

\item{type}{character, in c("k-fold", "loo", "lpo", "custom")
k-fold is k-fold cross-validation, provide \code{k}
loo is leave-one-out,
lpo is leave-percent-out, provide \code{percent} from 1 to 100
custom is user-defined group, provide \code{target} and \code{data}}

\item{seed}{random seed}

\item{print}{print information during computation}

\item{N_sim}{integer, number of simulations (e.g., estimate MAE, MSE, .. N times)}

\item{n_gibbs_samples}{number of gibbs samples of latent process, used for computing CRPS, sCRPS}

\item{n_burnin}{number of burnin}

\item{k}{integer (only for k-fold type)}

\item{percent}{how many percent for testing? from 0 to 1 (for lpo type)}

\item{times}{how many test cases (only for lpo type)}

\item{metric}{Optional function or list of functions (one per model) that maps
the group-wise observations/predictions for a single location to the quantity
that should be scored. The function receives a list containing at least
\code{y} (a named numeric vector of group values) and may optionally use
\code{samples1} and \code{samples2} (matrices with rows named by group and
columns indexing posterior draws). The function must return either a numeric
scalar (optionally named), or a list with component \code{y} and optional
components \code{samples1}, \code{samples2}, and \code{label}. When
\code{NULL}, the original per-group scores are computed. Example:
\code{metric = function(data) 2 * data$y["A"] + data$y["B"]}. To sum all
groups, return \code{sum(data$y)}.}

\item{test_idx}{a list of indices of the data (which data points to be predicted) (only for custom type)}

\item{train_idx}{a list of indices of the data (which data points to be used for re-sampling (not re-estimation)) (only for custom type)}

\item{keep_pred}{logical, keep test information (pred_1, pred_2) in the return (as attributes), pred_1 and pred_2 are the prediction of the two chains}

\item{parallel}{logical, run in parallel mode}

\item{thining_gap}{integer, the gap between samples for thinning, if 0, then no thinning, if 1, then keep 50\% of the samples for CRPS, sCRPS, etc.}

\item{cores_layer1}{integer, number of cores for the first layer (over testing samples)}

\item{cores_layer2}{integer, number of cores for the second layer (over computing scores for N_sim simulations)}

\item{merge_groups}{logical, if TRUE, merge groups as vector components (e.g., for vector-valued wind data with north_wind, east_wind).
MAE becomes Euclidean distance, MSE becomes squared Euclidean distance, etc.}

\item{merged_group_name}{character, name for the merged group when merge_groups=TRUE.
If NULL, uses "group1_group2" format (default: NULL)}

\item{data}{optional data.frame used to replace the original fitting data before running CV.
If `NULL`, the data stored in `ngme` is used. If provided, the model is rebuilt on `data`
while reusing fitted parameters from `ngme`.}

\item{chain_combine}{how to combine multiple optimization chains:
`"param_mean"` uses the fitted object directly (default), while
`"predictive_average"` computes predictions from each optimization chain
and averages at the predictive level.}
}
\value{
A list with components:
  \describe{
    \item{mean.scores}{Mean of the \code{N_sim} estimates for MSE, MAE, CRPS,
      and sCRPS.}
    \item{sd.scores}{Standard deviation of the \code{N_sim} estimates for MSE,
      MAE, CRPS, and sCRPS.}
  }
}
\description{
Compute the cross-validation for the ngme model
Perform cross-validation for ngme model
first into sub_groups (a list of target, and train data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/noise.R
\name{ngme_noise}
\alias{ngme_noise}
\alias{noise_nig}
\alias{noise_normal}
\alias{noise_t}
\alias{noise_skew_t}
\alias{noise_gal}
\alias{noise_normal_nig}
\title{ngme noise specification}
\usage{
ngme_noise(
  noise_type,
  mu = 0,
  sigma = 1,
  nu = 1,
  B_mu = NULL,
  theta_mu = NULL,
  B_sigma = NULL,
  theta_sigma = NULL,
  B_nu = NULL,
  theta_nu = NULL,
  theta_sigma_normal = NULL,
  B_sigma_normal = NULL,
  fix_theta_mu = FALSE,
  fix_theta_sigma = FALSE,
  fix_rho = FALSE,
  fix_theta_sigma_normal = FALSE,
  fix_theta_nu = FALSE,
  V = NULL,
  fix_V = FALSE,
  single_V = FALSE,
  share_V = FALSE,
  corr_measurement = FALSE,
  index_corr = NULL,
  map_corr = NULL,
  nu_lower_bound = 0,
  rho = double(0),
  prior = NULL,
  ...
)

noise_normal(
  sigma = NULL,
  theta_sigma = NULL,
  B_sigma = matrix(1),
  corr_measurement = FALSE,
  index_corr = NULL,
  ...
)

noise_nig(
  mu = NULL,
  sigma = NULL,
  nu = NULL,
  V = NULL,
  theta_mu = NULL,
  theta_sigma = NULL,
  theta_nu = NULL,
  nu_lower_bound = 0,
  B_mu = matrix(1),
  B_sigma = matrix(1),
  B_nu = matrix(1),
  corr_measurement = FALSE,
  index_corr = NULL,
  ...
)

noise_gal(
  mu = NULL,
  sigma = NULL,
  nu = NULL,
  V = NULL,
  theta_mu = NULL,
  theta_sigma = NULL,
  theta_nu = NULL,
  nu_lower_bound = 0.01,
  B_mu = matrix(1),
  B_sigma = matrix(1),
  B_nu = matrix(1),
  corr_measurement = FALSE,
  index_corr = NULL,
  ...
)

noise_skew_t(
  mu = NULL,
  sigma = NULL,
  nu = NULL,
  theta_mu = NULL,
  theta_sigma = NULL,
  theta_nu = NULL,
  nu_lower_bound = 0.01,
  B_mu = matrix(1),
  B_sigma = matrix(1),
  B_nu = matrix(1),
  corr_measurement = FALSE,
  index_corr = NULL,
  ...
)

noise_t(
  nu = NULL,
  theta_nu = NULL,
  nu_lower_bound = 0,
  B_nu = matrix(1),
  corr_measurement = FALSE,
  index_corr = NULL,
  ...
)

noise_normal_nig(
  sigma_normal = NULL,
  mu = NULL,
  sigma_nig = NULL,
  nu = NULL,
  V = NULL,
  theta_mu = NULL,
  theta_sigma_nig = NULL,
  theta_sigma_normal = NULL,
  theta_nu = NULL,
  B_mu = matrix(1),
  B_sigma_nig = matrix(1),
  B_sigma_normal = matrix(1),
  B_nu = matrix(1),
  corr_measurement = FALSE,
  index_corr = NULL,
  ...
)
}
\arguments{
\item{noise_type}{type of noise, "normal", "nig", "gal", "t", "skew_t"}

\item{mu}{specify the NIG noise parameter mu, see \code{?nig}}

\item{sigma}{specify the noise parameter sigma, see \code{?nig}}

\item{nu}{specify the noise parameter nu, see \code{?nig} and \code{?gal}}

\item{B_mu}{Basis matrix for mu (if non-stationary)}

\item{theta_mu}{specify a non-stationary noise using theta_mu}

\item{B_sigma}{Basis matrix for sigma (if non-stationary)}

\item{theta_sigma}{specify a non-stationary noise using theta_sigma}

\item{B_nu}{Basis matrix for nu (if non-stationary)}

\item{theta_nu}{specify a non-stationary noise using theta_nu}

\item{theta_sigma_normal}{for normal nosie with nig noise sharing same parameter}

\item{B_sigma_normal}{for normal nosie with nig noise sharing same parameter}

\item{fix_theta_mu}{fix the parameter of theta_mu}

\item{fix_theta_sigma}{fix the parameter of theta_sigma, can be a single
logical value or a vector of logical values with length equal to
length(theta_sigma)}

\item{fix_rho}{fix the parameter of rho}

\item{fix_theta_sigma_normal}{fix the parameter of sigma_normal, used in noise_normal_nig()}

\item{fix_theta_nu}{fix the parameter of nu}

\item{V}{start value for V}

\item{fix_V}{fix the sampling of V
gives Y|W ~ N(mean * sigma, sigma^2)}

\item{single_V}{TRUE if V is a single number}

\item{share_V}{used only for bivariate model}

\item{corr_measurement}{TRUE if we use correlated measurement noise}

\item{index_corr}{used when corr_measurement=TRUE, indicate which observation has correlation}

\item{map_corr}{1d, 2d, or formula, used when corr_measurement=TRUE, specify use which covariate to infer the index_corr.}

\item{nu_lower_bound}{specify the lower bound of parameter nu; effective
parametrization is \eqn{nu = nu_lower_bound + \exp(B_\nu \theta_\nu)} so
\code{theta_nu} remains unconstrained. Default 0.}

\item{rho}{used when corr_measurement=TRUE, starting point for correlation}

\item{prior}{prior specification created by \code{priors(...)}. Supported
keys are \code{mu}, \code{sigma}, and \code{nu}.}

\item{...}{additional arguments}

\item{sigma_normal}{for normal nosie with nig noise sharing same parameter}

\item{sigma_nig}{similar to sigma_normal}

\item{theta_sigma_nig}{similar to theta_sigma_normal}

\item{B_sigma_nig}{similar to B_sigma_nig}
}
\value{
a list of specification of noise
}
\description{
Function for specifying ngme noise.
Please use \code{noise_nig} and \code{noise_normal} for simpler usage.
Use \code{ngme_noise_types()} to check all the available types.
}
\details{
The parameterization is given in \code{?nig} and \code{?gal}. Moreover,
for specifying non-stationary mu and sigma, nu
\deqn{\mu = B_{\mu} \theta_{\mu},} and
\deqn{\sigma = \exp (B_{\sigma} \theta_{\sigma}),}
\deqn{\nu = \nu_{\mathrm{lower}} + \exp (B_{\nu} \theta_{\nu}).}
}
\examples{
noise_normal(sigma = 2)
noise_nig(mu = 1, sigma = 2, nu = 1)
noise_gal(mu = 1, sigma = 2, nu = 1)
noise_skew_t(mu = 0, sigma = 1, nu = 5)
noise_t(nu = 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gal.R
\name{gal}
\alias{gal}
\alias{dgal}
\alias{rgal}
\alias{pgal}
\alias{qgal}
\title{The Generalized Asymmetric Laplace (GAL) Distribution}
\usage{
dgal(x, delta, mu, nu, sigma, log = FALSE)

rgal(n, delta, mu, nu, sigma, seed = 0)

pgal(q, delta, mu, nu, sigma, lower.tail = TRUE, log.p = FALSE)

qgal(p, delta, mu, nu, sigma, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x, q}{vector of quantiles.}

\item{delta}{A numeric value for the location parameter.}

\item{mu}{A numeric value for the shift parameter.}

\item{nu}{A numeric value for the shape parameter.}

\item{sigma}{A numeric value for the scaling parameter.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities/densities \eqn{p} are
returned as \eqn{log(p)}.}

\item{n, }{number of observations.}

\item{seed}{Seed for the random generation.}

\item{lower.tail}{logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
otherwise, \eqn{P[X>x]}.}

\item{p}{vector of probabilities.}
}
\value{
dgal gives the density, pgal gives the distribution function,
qgal gives the quantile function, and rgal generates random deviates.

Invalid arguments will result in return value NaN, with a warning.

The length of the result is determined by \code{n} for rgal.
}
\description{
Density, distribution function, quantile function and
random generation for the generalized asymmetric Laplace distribution
 with parameters \code{mu}, \code{sigma} and \code{nu}, \code{delta}.
}
\details{
The generalized asymmetric Laplace distribution has density given
by
 \deqn{f(x; p, a, b) =
 \frac{e^{\nu+\mu(x-\delta)/\sigma^2}\sqrt{\nu\mu^2/\sigma^2+\nu^2}}{\pi\sqrt{\nu\sigma^2+(x-\delta)^2}}
 K_1(\sqrt{(\nu\sigma^2+(x-\delta)^2)(\mu^2/\sigma^4+\nu/\sigma^2)}),}
 where \eqn{K_p} is modified Bessel function of the second kind of order \eqn{p},
 \eqn{x>0}, \eqn{\nu>0} and \eqn{\mu,\delta, \sigma\in\mathbb{R}}.
 See Barndorff-Nielsen (1977, 1978 and 1997) for further details.

If the mixing variable \eqn{V} follows a Gamma distribution (same parameterization in R):
\deqn{V \sim \Gamma(h \nu, \nu),} then the poserior follows the GAL distribution (a special case of GIG distribution):
\deqn{
-\mu +\mu V + \sigma \sqrt{V} Z \sim GIG(h \nu - 0.5, 2 \nu + (\frac{\mu}{\sigma})^{2}, 0)
}
}
\examples{
rgal(100, delta = 0, mu = 5, sigma = 1, nu = 1)
pgal(0.4, delta = 0, mu = 5, sigma = 1, nu = 1)
qgal(0.8, delta = 0, mu = 5, sigma = 1, nu = 1)
plot(function(x){dgal(x, delta = 0, mu = 5, sigma = 1, nu = 1)}, main =
"generalized asymmetric Laplace density", ylab = "Probability density",
xlim = c(0,10))
}
\references{
Barndorff-Nielsen, O. (1977) Exponentially decreasing distributions for the logarithm of particle size. Proceedings of the Royal Society of London.

 Series A, Mathematical and Physical Sciences. The Royal Society. 353, 401–409. \doi{10.1098/rspa.1977.0041}

 Barndorff-Nielsen, O. (1978) Hyperbolic Distributions and Distributions on Hyperbolae, Scandinavian Journal of Statistics. 5, 151–157.
}
\seealso{
\code{\link{dgig}}, \code{\link{dig}}, \code{\link{digam}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info_datasets.R
\docType{data}
\name{cienaga.border}
\alias{cienaga.border}
\title{The x y location of the border of the swamp of Cienaga Grande in Santa Marta, Colombia}
\format{
A data frame with 472 rows and 2 columns.
\describe{
  \item{East, North}{location}
}
}
\source{
..
}
\usage{
cienaga.border
}
\description{
The data is of dimension 472 * 2. It contains the x and y coordinates of the border of the swamp.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{get_trace_trajectories}
\alias{get_trace_trajectories}
\title{Get trace trajectories from ngme fitting}
\usage{
get_trace_trajectories(ngme, name = "general", apply_transform = TRUE)
}
\arguments{
\item{ngme}{ngme object}

\item{name}{name of latent models, otherwise get fixed effects and measurement noise
should be in names(ngme$models) or other}

\item{apply_transform}{logical, whether to apply parameter transformations (default: TRUE)}
}
\value{
list containing:
\describe{
  \item{trajectories}{named list of trajectory matrices for each parameter}
  \item{parameter_names}{character vector of parameter names}
  \item{transformations}{list of transformation functions used}
  \item{n_chains}{number of chains}
  \item{n_iterations}{number of iterations}
}
}
\description{
Extract numerical trajectories of parameters during ngme optimization
without creating plots. This function provides the underlying data
used by traceplot().
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{precision_matrix_multivariate_spde}
\alias{precision_matrix_multivariate_spde}
\title{Compute the precision matrix for multivariate spde Matern model}
\usage{
precision_matrix_multivariate_spde(
  p,
  mesh,
  rho,
  alpha_list = NULL,
  theta_K_list = NULL,
  variance_list = NULL,
  B_K_list = NULL,
  theta = NULL,
  Q = NULL
)
}
\arguments{
\item{p}{dimension, should be integer and greater than 1}

\item{mesh}{an fmesher::fm_mesh_2d object, mesh for build the SPDE model}

\item{rho}{vector with the p(p-1)/2 correlation parameters rho_11, rho_21, rho_22, ... rho_p1, rho_p2, ... rho_p(p-1)}

\item{alpha_list}{a list of SPDE smoothness parameter}

\item{theta_K_list}{a list (length is p) of theta_K}

\item{variance_list}{If provided, it should be a vector of length p, where the
kth element corresponds to a desired variance of the kth field. The kth operator
is then scaled by a constant c so that this variance is achieved in the stationary case
(default no scaling)}

\item{B_K_list}{a list (length is p) of B_K (non-stationary case)}

\item{theta}{parameter for Q matrix (length of 1 when p=2, length of 3 when p=3)}

\item{Q}{orthogonal matrix of dim p*p (provide when p > 3)}
}
\value{
the precision matrix of the multivariate model
}
\description{
Compute the precision matrix for multivariate spde Matern model
}
\details{
The general model is defined as $D diag(L_1, ..., L_p) x = M$. D is the dependence matrix, it is paramterized by $D = Q(theta) * D_l(cor_mat)$, where $Q$ is the orthogonal matrix, and $D_l$ is matrix controls the cross-correlation.
See the section 2.2 of Bolin and Wallin (2020) for exact parameterization of Dependence matrix.
}
\examples{
library(fmesher)

# Use a small mesh so the example stays lightweight.
x <- seq(from = 0, to = 1, length.out = 6)
mesh <- fm_rcdt_2d_inla(lattice = fm_lattice_2d(x, x), extend = FALSE)

Q <- precision_matrix_multivariate_spde(
  p = 2,
  mesh = mesh,
  rho = 0.25,
  alpha_list = list(2, 2),
  theta_K_list = list(0, 0),
  variance_list = list(1, 1)
)
dim(Q)
}
\references{
Bolin, D. and Wallin, J. (2020), Multivariate type G Matérn stochastic partial differential equation random fields. J. R. Stat. Soc. B, 82: 215-239. https://doi.org/10.1111/rssb.12351
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{rw2}
\alias{rw2}
\title{Random Walk Model of Order 2 (RW2)}
\usage{
rw2(mesh = NULL, cyclic = FALSE, constr = TRUE)
}
\arguments{
\item{mesh}{numerical vector or inla.mesh.1d object, locations to build the mesh.
For numerical vectors, assumes equally spaced locations. Must have at least 3 locations.}

\item{cyclic}{logical, whether the mesh is circular. If TRUE, the first and last
locations are treated as neighbors with second-order differences computed
across the boundary.}

\item{constr}{logical, whether to enforce the sum-to-zero constraint \eqn{\sum_{i=1}^n h_i W_i = 0}
and the linear trend constraint \eqn{\sum_{i=1}^n (\sum_{j=1}^i h_j - h_1) \cdot W_i = 0}.
If FALSE, fixes the first and second elements \eqn{W_1 = 0} and \eqn{W_2 = 0}.}
}
\value{
An `ngme_operator` object containing the precision matrix and related
  components for the RW2 model.
}
\description{
Constructs a second-order random walk model for spatial or temporal processes.
The RW2 model assumes that second-order differences
\eqn{\Delta^2 W_i = W_i - 2W_{i-1} + W_{i-2}} are independent and identically
distributed Gaussian variables.
}
\details{
The RW2 model is defined by the precision matrix \eqn{K} that penalizes
second-order differences, making it smoother than RW1. The model enforces
constraints to ensure identifiability:

**Non-cyclic (default)**: The first two rows of \eqn{K} enforce constraints:
the first row implements \eqn{\sum_{i=1}^n h_i W_i = 0} (sum-to-zero),
and the second row implements \eqn{\sum_{i=1}^n h_i \cdot i \cdot W_i = 0}
(removes linear trend). The remaining rows penalize second-order differences.

**Cyclic**: Treats the domain as circular, connecting the first and last locations
as neighbors. The constraint \eqn{\sum_{i=1}^n h_i W_i = 0} is enforced.
The precision matrix is expanded to size \eqn{(n+1) \times (n+1)} to handle
the constraint and circular structure properly.
}
\examples{
# Non-cyclic RW2 with constraints (default)
rw2_default <- rw2(1:6)
print(rw2_default$K)

# Cyclic RW2 (connects first and last locations)
rw2_cyclic <- rw2(1:6, cyclic = TRUE)
print(rw2_cyclic$K)

# Using with unequally spaced locations
locations <- c(0, 1, 3, 6, 10, 15)
rw2_unequal <- rw2(locations)
print(rw2_unequal$K)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{mean_list}
\alias{mean_list}
\title{taking mean over a list of nested lists}
\usage{
mean_list(lls, weights = NULL)
}
\arguments{
\item{lls}{a list}

\item{weights}{weights of each list}
}
\value{
a list of nested lists
}
\description{
taking mean over a list of nested lists
}
\examples{
ls <- list(
  list(a = 1, b = 2, t = "nig", ll = list(a = 1, b = 2, w = "ab")),
  list(a = 3, b = 5, t = "nig", ll = list(a = 1, b = 6, w = "ab")),
  list(a = 5, b = 5, t = "nig", ll = list(a = 4, b = 2, w = "ab"))
)
mean_list(ls)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loglik.R
\name{compute_log_like}
\alias{compute_log_like}
\title{Compute Gaussian log-likelihood}
\usage{
compute_log_like(ngme_model)
}
\arguments{
\item{ngme_model}{An object created by `ngme()`.}
}
\value{
A numeric scalar. Returns `0` if any replicate is non-Gaussian.
}
\description{
Creates the underlying C++ block models and evaluates their Gaussian log-likelihood when applicable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{poly_decay}
\alias{poly_decay}
\title{Polynomial schedule helper}
\usage{
poly_decay(alpha = 0.501, t0 = 1, burnin_iter = 0)
}
\arguments{
\item{alpha}{polynomial exponent in \eqn{(1/2,1)}.}

\item{t0}{non-negative schedule offset.}

\item{burnin_iter}{non-negative integer. Initial iterations without
polynomial schedule scaling.}
}
\value{
a \code{stepsize_control()} object.
}
\description{
Polynomial schedule helper
}
\details{
Convenience helper that enables polynomial schedule and disables
checkpoint-based decay.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{compute_ngme_sgld_samples}
\alias{compute_ngme_sgld_samples}
\title{Refit an Existing ngme Object with SGLD and Extract Samples}
\usage{
compute_ngme_sgld_samples(
  fit,
  iterations = 4000,
  optimizer = sgld(stepsize = 0.01, temperature = 1),
  burnin = 100,
  n_batch = 20,
  n_parallel_chain = 4,
  alpha = 0.6,
  t0 = 10,
  start_sd = 0.2,
  seed = Sys.time(),
  verbose = FALSE,
  name = "all",
  burnin_iter = 0,
  thinning = 1,
  n_gibbs_samples = NULL,
  apply_transform = TRUE,
  combine_chains = TRUE,
  control_opt = NULL,
  ...
)
}
\arguments{
\item{fit}{existing fitted `ngme` object (can be obtained with any optimizer).}

\item{iterations}{optimization iterations for the SGLD stage.}

\item{optimizer}{optimizer for the sampling stage; must be `sgld(...)`.}

\item{burnin}{burn-in iterations before optimization.}

\item{n_batch}{number of optimization checkpoints.}

\item{n_parallel_chain}{number of parallel chains.}

\item{alpha}{polynomial schedule exponent used by `poly_decay(alpha, t0)`.}

\item{t0}{non-negative schedule offset.}

\item{start_sd}{standard deviation for randomized chain initialization.}

\item{seed}{random seed for SGLD stage.}

\item{verbose}{logical; print optimization progress.}

\item{name}{parameter block to extract: `"all"` (default), latent model name,
or `"general"`.}

\item{burnin_iter}{non-negative integer used both as optimizer schedule
warmup (`stepsize_schedule_burnin_iter`) and as explicit trajectory
trimming before sampling.}

\item{thinning}{positive integer thinning interval.}

\item{n_gibbs_samples}{optional positive integer. If supplied, override
`control_ngme$n_gibbs_samples` for the SGLD refit stage only; otherwise
inherit the value from `fit`.}

\item{apply_transform}{logical; apply parameter transforms to user scale.}

\item{combine_chains}{logical; if `TRUE`, return one combined data.frame,
otherwise return one data.frame per chain.}

\item{control_opt}{optional pre-built `control_opt` object for the SGLD
stage. If supplied, it is used directly (with `store_traj` forced to
`TRUE`).}

\item{...}{additional arguments forwarded to [control_opt()] when
`control_opt` is `NULL`.}
}
\value{
A data.frame (or list of data.frames) from [ngme_sgld_samples()]
with extra attributes `refit`, `refit_control_opt`, and `refit_source`.
}
\description{
Run one additional SGLD stage (warm-started from an existing `ngme` fit),
then extract posterior-like samples from stored trajectories.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{precision_matrix_multivariate}
\alias{precision_matrix_multivariate}
\title{Compute the precision matrix for multivariate model}
\usage{
precision_matrix_multivariate(
  p,
  operator_list,
  rho,
  theta = NULL,
  Q = NULL,
  scale = NULL
)
}
\arguments{
\item{p}{dimension, should be integer and greater than 1}

\item{operator_list}{a list of ngme_operator object (length should be p)}

\item{rho}{vector with the p(p-1)/2 correlation parameters rho_11, rho_21, rho_22, ... rho_p1, rho_p2, ... rho_p(p-1)}

\item{theta}{parameter for Q matrix (length of 1 when p=2, length of 3 when p=3)}

\item{Q}{orthogonal matrix of dim p*p (provide when p > 3)}

\item{scale}{A vector of length p with constants to multiply each operator matrix with}
}
\value{
the precision matrix of the multivariate model
}
\description{
Compute the precision matrix for multivariate model
}
\details{
The general model is defined as $D diag(L_1, ..., L_p) x = M$. D is the dependence matrix, it is paramterized by $D = Q(theta) * D_l(cor_mat)$, where $Q$ is the orthogonal matrix, and $D_l$ is matrix controls the cross-correlation.
See the section 2.2 of Bolin and Wallin (2020) for exact parameterization of Dependence matrix.
}
\examples{
rho <- c(-0.5, 0.5, -0.25) # correlation parameters
operator_list <- list(ar1(1:5, rho = 0.4), ar1(1:5, rho = 0.5), ar1(1:5, rho = 0.6))
precision_matrix_multivariate(3, operator_list, rho, theta = c(1, 2, 3))
}
\references{
Bolin, D. and Wallin, J. (2020), Multivariate type G Matérn stochastic partial differential equation random fields. J. R. Stat. Soc. B, 82: 215-239. https://doi.org/10.1111/rssb.12351
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{ngme_ts_make_A}
\alias{ngme_ts_make_A}
\title{Make observation matrix for time series}
\usage{
ngme_ts_make_A(loc, replicate = NULL, range = c(min(loc), max(loc)))
}
\arguments{
\item{loc}{integers (after sorting, no gaps > 1)}

\item{replicate}{indicating replicate measure at same location}

\item{range}{range for the mesh
by default range=(min(loc), max(loc))}
}
\value{
A matrix (length(loc) * length(unique(loc)))
}
\description{
Make observation matrix for time series
}
\examples{
ngme_ts_make_A(c(1, 2, 2), replicate = c(1, 1, 2))
ngme_ts_make_A(c(1, 2, 2), range = c(1, 5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{ngme_as_sparse}
\alias{ngme_as_sparse}
\title{Convert sparse matrix into sparse dgCMatrix}
\usage{
ngme_as_sparse(G)
}
\arguments{
\item{G}{matrix}
}
\value{
sparse dgCMatrix
}
\description{
Convert sparse matrix into sparse dgCMatrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{ngme_update}
\alias{ngme_update}
\title{Check whether a newer stable version of ngme2 is available}
\usage{
ngme_update()
}
\value{
Invisibly returns a list with the local version, remote version,
  repository URL, and a logical \code{update_available} flag.
}
\description{
This function checks the package repository for the latest available ngme2
version. It does not install or update packages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/operator.R
\name{print.ngme_operator}
\alias{print.ngme_operator}
\title{Print ngme operator}
\usage{
\method{print}{ngme_operator}(x, padding = 0, prefix = "Model type", ...)
}
\arguments{
\item{x}{ngme operator object}

\item{padding}{number of white space padding in front}

\item{prefix}{prefix string}

\item{...}{...}
}
\value{
a list (operator specifications)
}
\description{
Print ngme operator
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions.R
\name{gig}
\alias{gig}
\alias{dgig}
\alias{rgig}
\alias{pgig}
\alias{qgig}
\title{The Generalised Inverse-Gaussian (GIG) Distribution}
\usage{
dgig(x, p, a, b, log = FALSE)

rgig(n, p, a, b, seed = 0)

pgig(q, p, a, b, lower.tail = TRUE, log.p = FALSE)

qgig(prob, p, a, b, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x, q}{vector of quantiles.}

\item{p}{parameter \code{p}.}

\item{a, b}{parameters \code{a} and \code{b}. Must be positive.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities/densities \eqn{p} are
returned as \eqn{log(p)}.}

\item{n, }{number of observations.}

\item{seed}{Seed for the random generation.}

\item{lower.tail}{logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
otherwise, \eqn{P[X>x]}.}

\item{prob}{vector of probabilities.}
}
\value{
dgig gives the density, pgig gives the distribution function,
qgig gives the quantile function, and rgig generates random deviates.

Invalid arguments will result in return value NaN, with a warning.

The length of the result is determined by \code{n} for rgig.
}
\description{
Density, distribution function, quantile function and
random generation for the generalised inverse-Gaussian distribution
 with parameters \code{p}, \code{a} and \code{b}.
}
\details{
The generalised inverse-Gaussian distribution has density given
by
 \deqn{f(x; p, a, b) = ((a/b)^{p/2})/(2K_p(\sqrt{ab})) x^{p-1} \exp\{-(a/2)x - (b/2)/x\},}
 where \eqn{K_p} is modified Bessel function of the second kind of order \eqn{p},
 \eqn{x>0}, \eqn{a,b>0} and \eqn{p\in\mathbb{R}}.
 See Jørgensen (1982) for further details.
}
\examples{
rgig(20, p = 1, a = 1, b = 1)
pgig(0.4, p = 1, a = 1, b = 1)
qgig(0.8, p = 1, a = 1, b = 1)
plot(function(x){dgig(x, p = 1, a = 1, b = 1)}, main =
"Generalised inverse-Gaussian density", ylab = "Probability density",
xlim = c(0,10))
}
\references{
Jørgensen, Bent (1982). Statistical Properties of the Generalized
Inverse Gaussian Distribution. Lecture Notes in Statistics. 9.
New York–Berlin: Springer-Verlag. \doi{10.1007/978-1-4612-5698-4}
}
\seealso{
\code{\link{dnig}}, \code{\link{dig}}, \code{\link{digam}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{precond_sgd}
\alias{precond_sgd}
\title{Preconditioner SGD optimization}
\usage{
precond_sgd(stepsize = 1, numerical_eps = 1e-05)
}
\arguments{
\item{stepsize}{stepsize for SGD}

\item{numerical_eps}{numerical, the gap used for estimate preconditioner, default is 1e-5}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
Preconditioner SGD optimization
}
\details{
Preconditioner SGD is using fisher information matrix as preconditioner (natural gradient descent).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{compute_ngme_ci}
\alias{compute_ngme_ci}
\alias{compute_ngme_CI}
\title{Refit an Existing ngme Object with SGD and Compute Batch-Means CI}
\usage{
compute_ngme_ci(
  fit,
  iterations = 4000,
  alpha = 0.501,
  optimizer = sgd(stepsize = 0.03),
  burnin = 100,
  n_batch = 20,
  n_parallel_chain = 4,
  t0 = 1,
  start_sd = 0.2,
  seed = Sys.time(),
  verbose = FALSE,
  level = 0.95,
  name = "all",
  M = NULL,
  N = NULL,
  apply_transform = TRUE,
  drop_burnin = TRUE,
  burnin_iter = 0,
  control_opt = NULL,
  ...
)

compute_ngme_CI(
  fit,
  iterations = 4000,
  alpha = 0.501,
  optimizer = sgd(stepsize = 0.03),
  burnin = 100,
  n_batch = 20,
  n_parallel_chain = 4,
  t0 = 1,
  start_sd = 0.2,
  seed = Sys.time(),
  verbose = FALSE,
  level = 0.95,
  name = "all",
  M = NULL,
  N = NULL,
  apply_transform = TRUE,
  drop_burnin = TRUE,
  burnin_iter = 0,
  control_opt = NULL,
  ...
)
}
\arguments{
\item{fit}{existing fitted `ngme` object (can be obtained with any optimizer).}

\item{iterations}{optimization iterations for the SGD CI stage.}

\item{alpha}{Xi-style exponent used for both polynomial schedule and
batch-means inference.}

\item{optimizer}{optimizer for the CI stage; must be `sgd(...)` for Xi theory.}

\item{burnin}{burn-in iterations before SGD optimization.}

\item{n_batch}{number of optimization checkpoints.}

\item{n_parallel_chain}{number of parallel chains for SGD.}

\item{t0}{non-negative schedule offset in
\eqn{\eta_t = \eta_0 (t + t0)^{-\alpha}}.}

\item{start_sd}{standard deviation for randomized chain initialization.}

\item{seed}{random seed for CI-stage optimization.}

\item{verbose}{logical; print optimization progress.}

\item{level}{confidence level for CI.}

\item{name}{parameter block for CI (`"all"`, latent name, or `"general"`).}

\item{M}{number of retained Xi batches (excluding batch 0).}

\item{N}{decorrelation constant in Xi batch boundaries.}

\item{apply_transform}{logical; if `TRUE`, apply Post-delta transform after
raw-space inference.}

\item{drop_burnin}{logical; if `TRUE`, discard Xi batch 0.}

\item{burnin_iter}{non-negative integer used both as optimizer schedule
warmup (`stepsize_schedule_burnin_iter`) and as explicit CI trajectory
trimming before Xi-style batching.}

\item{control_opt}{optional pre-built `control_opt` object for the CI stage.
If supplied, it is used directly (with `store_traj` forced to `TRUE`).}

\item{...}{additional arguments forwarded to [control_opt_batch_ci()] when
`control_opt` is `NULL`.}
}
\value{
An object of class `ngme_batch_ci` with extra fields:
  `refit` (the SGD-refit model), `refit_control_opt`, and `refit_source`.
}
\description{
Run one additional SGD stage (warm-started from an existing `ngme` fit),
then compute Xi-style batch-means confidence intervals from the newly stored
trajectories.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{ar1}
\alias{ar1}
\title{ngme AR(1) model specification}
\usage{
ar1(mesh = NULL, rho = 0)
}
\arguments{
\item{mesh}{integer vector or inla.mesh.1d object, index to build the mesh}

\item{rho}{the correlation parameter (between -1 and 1)}
}
\value{
ngme_operator object
}
\description{
ngme AR(1) model specification
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{batch_decay}
\alias{batch_decay}
\title{Batch/checkpoint decay helper}
\usage{
batch_decay(
  patience = 3,
  gamma = 0.5,
  min_delta = 0,
  warmup = 0,
  min_stepsize = 0
)
}
\arguments{
\item{patience}{number of consecutive checkpoints without improvement before decay.}

\item{gamma}{decay factor in \eqn{(0,1)}.}

\item{min_delta}{minimum required decrease to count as improvement.}

\item{warmup}{number of initial checkpoints to skip.}

\item{min_stepsize}{lower bound for absolute stepsize.}
}
\value{
a \code{stepsize_control()} object.
}
\description{
Batch/checkpoint decay helper
}
\details{
Convenience helper that enables grad-norm plateau decay and keeps
iteration schedule constant.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{plot.ngme_noise}
\alias{plot.ngme_noise}
\title{Plot the density of one or more stationary noise objects}
\usage{
\method{plot}{ngme_noise}(x = NULL, ...)
}
\arguments{
\item{x}{An ngme_noise object (required).}

\item{...}{Additional ngme_noise objects to plot, or plotting parameters such as \code{xlim}.
Named arguments will be used as legend labels.}
}
\value{
A ggplot object showing the density curves for the provided noise objects.
}
\description{
This function plots the probability density function for one or more stationary noise objects
(e.g., NIG, GAL, or normal noise). Multiple noise objects can be compared on the same plot.
}
\examples{
plot(noise_nig(mu = 1, sigma = 2, nu = 1))
plot(n1 = noise_nig(mu = 0, sigma = 1, nu = 1), n2 = noise_nig(mu = 1, sigma = 1.5, nu = 0.5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{compare_noise_kld}
\alias{compare_noise_kld}
\title{Compare noise objects using Kullback-Leibler divergence}
\usage{
compare_noise_kld(x = NULL, ..., xlim = c(-10, 10), n_points = 1000)
}
\arguments{
\item{x}{first noise object (reference)}

\item{...}{additional noise objects to compare, and optional parameters}

\item{xlim}{x-axis range for evaluation (default: c(-10, 10))}

\item{n_points}{number of evaluation points (default: 1000)}
}
\value{
named vector of KLD values
}
\description{
This function compares multiple noise objects by calculating the KLD
of each noise against the first noise object (reference).
}
\examples{
n1 <- noise_nig(mu = 0, sigma = 1, nu = 1)
n2 <- noise_nig(mu = 0.5, sigma = 1.2, nu = 0.8)
compare_noise_kld(n1, method2 = n2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{get_trajectories}
\alias{get_trajectories}
\title{get the trajectories of parameters of the model}
\usage{
get_trajectories(ngme_object, model_name)
}
\arguments{
\item{ngme_object}{ngme object}

\item{model_name}{name of the model, if NULL, return the trajectories of the parameters of the measurement noise and fixed effects}
}
\value{
a list of trajectories, each element is a matrix of trajectories (n_samples * n_trajectories)
}
\description{
Get the trajectories of the parameters of the model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{iid}
\alias{iid}
\title{ngme iid model specification}
\usage{
iid(mesh = NULL)
}
\arguments{
\item{mesh}{integer or factor, index to build the mesh}
}
\value{
ngme_operator object
}
\description{
ngme iid model specification
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.R
\name{ngme}
\alias{ngme}
\title{Fit an additive linear mixed effect model over replicates}
\usage{
ngme(
  formula,
  data,
  family = "normal",
  control_opt = NULL,
  control_ngme = NULL,
  group = NULL,
  replicate = NULL,
  start = NULL,
  moving_window = 1,
  prior_beta = NULL,
  debug = FALSE
)
}
\arguments{
\item{formula}{formula}

\item{data}{a dataframe or a list providing data
(Only response variable can contain \code{NA} value,
 \code{NA} value in other columns will cause problem)}

\item{family}{likelihood type, same as measurement noise specification. It can be provided as:
\enumerate{
  \item a string, e.g., \code{"normal"}, \code{"nig"}, \code{"t"}.
  \item an ngme noise object, e.g., \code{noise_normal()}, \code{noise_nig(mu = 0, sigma = 1, nu = 1)}, \code{noise_t(nu=5)}.
}}

\item{control_opt}{control for optimizer. by default it is \code{control_opt()}. See \code{?control_opt} for details.}

\item{control_ngme}{control for ngme model. by default it is \code{control_ngme()}. See \code{?control_ngme} for details.}

\item{group}{factor, used for bivariate model, indicating which group the observation belongs to}

\item{replicate}{factor, used for divide data into different replicates}

\item{start}{starting ngme object (usually object from last fit)}

\item{moving_window}{number of iterations to average the estimation}

\item{prior_beta}{prior specification for fixed effects (`beta`), created by
\code{prior_*()} or \code{priors(...)}.}

\item{debug}{toggle debug mode}
}
\value{
random effects (for different replicate) + models(fixed effects, measuremnt noise, and latent process)
}
\description{
\code{ngme} function performs an analysis of non-gaussian additive models.
 It does the maximum likelihood estimation via stochastic gradient descent.
 The prediction of unknown location can be performed by leaving the response
 variable to be \code{NA}. The likelihood is specified by \code{family}.
The model estimation control can be setted in \code{control} using
 \code{control_opt()} function, see \code{?control_opt} for details.
See \code{ngme_model_types()} for available models.
}
\examples{
ngme(
  formula = Y ~ x1 + f(
    x2,
    model = ar1(rho = 0.5),
    noise = noise_nig()
  ) + f(x1,
    model = rw1(),
    noise = noise_normal(),
  ),
  family = noise_normal(sd = 0.5),
  data = data.frame(Y = 1:5, x1 = 2:6, x2 = 3:7),
  control_opt = control_opt(
    estimation = FALSE
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/noise.R
\name{merge_noise}
\alias{merge_noise}
\title{Merge 2 noise into 1 noise}
\usage{
merge_noise(noise1, noise2)
}
\arguments{
\item{noise1}{noise 1}

\item{noise2}{noise 2}
}
\value{
merged noise
}
\description{
Merge 2 noise into 1 noise
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{ou}
\alias{ou}
\title{Ornstein-Uhlenbeck Process Model}
\usage{
ou(mesh = NULL, theta = 1)
}
\arguments{
\item{mesh}{numerical vector or `inla.mesh.1d` object giving the ordered
time locations. Must be strictly increasing.}

\item{theta}{positive mean-reversion rate (stiffness parameter). Internally
stored on the log scale so optimization remains unconstrained.}
}
\value{
An `ngme_operator` object containing:
\itemize{
  \item `K`: the precision matrix
  \item `h`: integration weights
  \item `theta_K`: log-transformed theta for unconstrained optimization
  \item `mesh`: the mesh object
}
}
\description{
Implements the exact discrete-time representation of the continuous
Ornstein-Uhlenbeck (OU) process using a matrix formulation
\eqn{\mathbf{K}\mathbf{X} = \boldsymbol{\delta}}.
}
\details{
The OU process is defined by the stochastic differential equation (SDE):

\deqn{dX_t = -\theta X_t dt + \sigma dW_t}

where \eqn{\theta > 0} is the mean-reversion rate (stiffness) and
\eqn{\sigma} is the volatility (diffusion coefficient).

## Exact Discretization

Unlike Euler-Maruyama approximations, the OU process has an exact solution
between time points \eqn{t_i} and \eqn{t_{i+1}} with step size \eqn{\Delta t_i}:

\deqn{X_{i+1} = X_i e^{-\theta \Delta t_i} + \epsilon_i}

This is an AR(1) form with autoregressive coefficient:

\deqn{\rho_i = e^{-\theta \Delta t_i}}

and Gaussian noise:

\deqn{\epsilon_i \sim \mathcal{N}\left(0, \frac{\sigma^2}{2\theta}(1 - \rho_i^2)\right)}

## Matrix Representation

The precision matrix \eqn{\mathbf{K}} for the linear system
\eqn{\mathbf{K}\mathbf{X} = \boldsymbol{\delta}} is constructed as:

\deqn{\mathbf{K} = \begin{bmatrix}
\sqrt{1-\rho_1^2} & 0 & 0 & \cdots & 0 \\
-\rho_1 & 1 & 0 & \cdots & 0 \\
0 & -\rho_2 & 1 & \cdots & 0 \\
\vdots & \vdots & \ddots & \ddots & \vdots \\
0 & 0 & \cdots & -\rho_{n-1} & 1
\end{bmatrix}}

where \eqn{\rho_i = \exp(-\theta \Delta t_i)} with \eqn{\Delta t_i = t_{i+1} - t_i}.

**Why the \eqn{\sqrt{1-\rho_1^2}} term?** The first row ensures that \eqn{X_1}
is drawn from the stationary distribution \eqn{\mathcal{N}(0, \sigma^2/(2\theta))}.
This scaling factor matches the variance of the first component of
\eqn{\boldsymbol{\delta}} with the transition noise variance in subsequent steps.

## Non-uniform Mesh

For non-uniform meshes, each \eqn{\rho_i} is computed locally based on the
varying time step \eqn{\Delta t_i}, allowing the model to naturally handle
irregularly spaced observations.

## The \eqn{h} vector

The \eqn{h} vector represents integration weights for the mass matrix.
For the OU model, it is set as:

\deqn{h = [\Delta t_1, \Delta t_2, \ldots, \Delta t_{n-1}, \Delta t_{n-1}]}

where the last element is duplicated. This ensures proper weighting in the
finite element formulation.

## The \eqn{\boldsymbol{\delta}} vector

In the equation \eqn{\mathbf{K}\mathbf{X} = \boldsymbol{\delta}}, the vector
\eqn{\boldsymbol{\delta}} contains independent identically distributed (i.i.d.)
random variables with unit variance. The precision matrix \eqn{\mathbf{K}}
is constructed such that the resulting process \eqn{\mathbf{X}} has the
correct OU covariance structure.
}
\examples{
# Uniform mesh
mesh_uniform <- seq(0, 10, by = 0.5)
ou_uniform <- ou(mesh = mesh_uniform, theta = 0.5)
print(ou_uniform)

# Non-uniform mesh
mesh_nonuniform <- c(0, 1, 2.5, 3.5, 5, 8, 11)
ou_nonunif <- ou(mesh = mesh_nonuniform, theta = 0.8)
print(ou_nonunif)
}
\references{
Uhlenbeck, G. E., & Ornstein, L. S. (1930). On the theory of the Brownian motion.
Physical review, 36(5), 823.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/util.R
\name{ngme_make_mesh_repls}
\alias{ngme_make_mesh_repls}
\title{ngme make mesh for different replicates}
\usage{
ngme_make_mesh_repls(data, map, replicate, mesh_type = "regular")
}
\arguments{
\item{data}{provide the data.frame}

\item{map}{provide the map to make mesh, i.g. ~x+y, x and y will be extracted from data to make a 2d mesh}

\item{replicate}{provide the replicate information, i.g. ~id}

\item{mesh_type}{type of mesh, "regular" means use all the point from same replicate to make a mesh}
}
\value{
a list of mesh of length of different replicates
}
\description{
Make different mesh for different replicates
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{posterior_plot}
\alias{posterior_plot}
\alias{plot.ngme_sgld_ci}
\title{Plot Posterior Distributions from SGLD Samples}
\usage{
posterior_plot(
  x,
  parameters = NULL,
  type = c("density", "histogram"),
  by_chain = FALSE,
  show_estimate = TRUE,
  show_interval = TRUE,
  lower = NULL,
  upper = NULL,
  bins = 30,
  ncol = NULL,
  ...
)

\method{plot}{ngme_sgld_ci}(x, ...)
}
\arguments{
\item{x}{a data.frame (or list of data.frames) returned by
[ngme_sgld_samples()], or an object of class `ngme_sgld_ci`.}

\item{parameters}{optional character vector of parameter names to plot.
Defaults to all parameter columns.}

\item{type}{character; `"density"` (default) or `"histogram"`.}

\item{by_chain}{logical; if `TRUE`, color distributions by chain when a
`.chain` column is available.}

\item{show_estimate}{logical; add a vertical line at the posterior mean.}

\item{show_interval}{logical; add vertical lines for the equal-tail interval.}

\item{lower}{lower quantile probability used for intervals. If `x` is an
`ngme_sgld_ci` object, the stored value is used by default.}

\item{upper}{upper quantile probability used for intervals. If `x` is an
`ngme_sgld_ci` object, the stored value is used by default.}

\item{bins}{number of bins for histograms.}

\item{ncol}{number of facet columns. If `NULL`, use 2 columns for multiple
parameters and 1 for a single parameter.}

\item{...}{unused.}
}
\value{
A faceted ggplot object.
}
\description{
Visualize marginal posterior distributions of parameters extracted from
[ngme_sgld_samples()] or summarized by [ngme_sgld_ci()].
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complex-models.R
\name{var1}
\alias{var1}
\title{ngme VAR(1) bivariate model specification (Cayley re-parameterization)}
\usage{
var1(mesh = NULL, p1 = 0, p2 = 1, p3 = 0, p4 = 1)
}
\arguments{
\item{mesh}{Integer vector or \code{inla.mesh.1d} object for the time mesh
(length \eqn{T}).  Pass \code{NULL} for deferred construction.}

\item{p1}{Unconstrained skew-symmetric parameter (\eqn{\in \mathbb{R}});
controls the rotation part of the Cayley mapping.  Default \code{0}.}

\item{p2, p3, p4}{Unconstrained lower-triangular Cholesky parameters
(\eqn{\in \mathbb{R}}); control the contraction (positive-definite) part.
Defaults \code{p2 = 1}, \code{p3 = 0}, \code{p4 = 1}.}
}
\value{
An \code{ngme_operator} (or \code{ngme_operator_def} when
  \code{mesh = NULL}) suitable for use inside \code{f()}.
}
\description{
Builds a Vector Autoregressive order-1 (VAR(1)) operator for bivariate
time-series data.  The model follows:
\deqn{Y_t = A\,Y_{t-1} + \varepsilon_t, \quad
  A = \begin{pmatrix} a_{11} & a_{12} \\ a_{21} & a_{22} \end{pmatrix}}
}
\section{Cayley re-parameterization}{

To guarantee stationarity (\eqn{\rho(A) < 1}) at every SGD step, the
\eqn{2 \times 2} coefficient matrix \eqn{A} is obtained via the
\emph{Cayley transform}:
\deqn{A = (I + S)(I - S)^{-1}}
where \eqn{S} is constructed from four unconstrained real parameters
\eqn{(p_1, p_2, p_3, p_4) \in \mathbb{R}^4} as \eqn{S = J - R}:
\describe{
  \item{Skew-symmetric part (rotation):}{
    \eqn{J = \begin{pmatrix} 0 & p_1 \\ -p_1 & 0 \end{pmatrix}}}
  \item{Positive-definite part (contraction):}{
    \eqn{R = L L^{\top} + \varepsilon I}, where
    \eqn{L = \begin{pmatrix} p_2 & 0 \\ p_3 & p_4 \end{pmatrix}}
    and \eqn{\varepsilon = 10^{-5}}.}
}
Because \eqn{S + S^{\top} = -2R} is negative definite, all eigenvalues of
\eqn{S} have strictly negative real parts, and the Cayley transform then
guarantees \eqn{|\lambda_i(A)| < 1} for every \eqn{i}.

The default values \eqn{(p_1, p_2, p_3, p_4) = (0, 1, 0, 1)} give
\eqn{A \approx 0} (no auto-regression at initialization).
}

\section{Precision operator}{

The 2T x 2T precision operator is:
\deqn{K = M_0 + a_{11}\,M_{11} + a_{22}\,M_{22}
         + a_{12}\,M_{12} + a_{21}\,M_{21}}
where the \eqn{M_{ij}} are fixed sparse block matrices constructed from the
\eqn{T \times T} first-order difference matrix \eqn{C_T}.
}

\examples{
set.seed(1)
n_obs <- 10
dat <- data.frame(
  y     = rnorm(2 * n_obs),
  idx   = rep(seq_len(n_obs), 2),
  group = factor(rep(c("y1", "y2"), each = n_obs))
)

fit <- ngme(
  y ~ 0 + f(idx, model = var1(mesh = 1:n_obs), group = group,
            noise = noise_nig()),
  data    = dat,
  family  = noise_normal(sigma = 0.01, fix_sigma = TRUE),
  control_opt = control_opt(estimation = FALSE)
)
print(fit)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributions.R
\name{nig}
\alias{nig}
\alias{dnig}
\alias{rnig}
\alias{pnig}
\alias{qnig}
\title{The Normal Inverse-Gaussian (NIG) Distribution}
\usage{
dnig(x, delta, mu, nu, sigma, h = NULL, log = FALSE)

rnig(n, delta, mu, nu, sigma, h = NULL, seed = 0)

pnig(q, delta, mu, nu, sigma, h = NULL, lower.tail = TRUE, log.p = FALSE)

qnig(p, delta, mu, nu, sigma, h = NULL, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x, q}{vector of quantiles.}

\item{delta}{A numeric value for the location parameter.}

\item{mu}{A numeric value for the shift parameter.}

\item{nu}{A numeric value for the shape parameter.}

\item{sigma}{A numeric value for the scaling parameter.}

\item{h}{A numeric value for the additional parameter, see details.}

\item{log, log.p}{logical; if \code{TRUE}, probabilities/densities \eqn{p} are
returned as \eqn{log(p)}.}

\item{n, }{number of observations.}

\item{seed}{Seed for the random generation.}

\item{lower.tail}{logical; if \code{TRUE}, probabilities are \eqn{P[X\leq x]},
otherwise, \eqn{P[X>x]}.}

\item{p}{vector of probabilities.}
}
\value{
dnig gives the density, pnig gives the distribution function,
qnig gives the quantile function, and rnig generates random deviates.

Invalid arguments will result in return value NaN, with a warning.

The length of the result is determined by \code{n} for rnig.
}
\description{
Density, distribution function, quantile function and
random generation for the normal inverse-Gaussian distribution
 with parameters \code{p}, \code{a} and \code{b}.
}
\details{
The normal inverse-Gaussian distribution has density given
by
 \deqn{f(x; \delta, \mu, \sigma, \nu) =
 \frac{e^{\nu+\mu(x-\delta)/\sigma^2}\sqrt{\nu\mu^2/\sigma^2+\nu^2}}{\pi\sqrt{\nu\sigma^2+(x-\delta)^2}}
 K_1(\sqrt{(\nu\sigma^2+(x-\delta)^2)(\mu^2/\sigma^4+\nu/\sigma^2)}),}
 where \eqn{K_p} is modified Bessel function of the second kind of order \eqn{p},
 \eqn{x>0}, \eqn{\nu>0} and \eqn{\mu,\delta, \sigma\in\mathbb{R}}.
 See Barndorff-Nielsen (1977, 1978 and 1997) for further details.

The additional parameter h is used when \deqn{V\sim IG(\nu,\nu h^{2})}.
By the infinite divisibility, \deqn{\frac{1}{h} V \sim IG(\nu h, \nu h)}.
Then \deqn{\delta+\mu V + \sigma \sqrt{V} Z} has the distribution of \deqn{NIG(\delta=-\mu h,\mu= \mu h, \sigma=\sigma \sqrt{h}, \nu=\nu h).}
}
\examples{
rnig(100, delta = 0, mu = 5, sigma = 1, nu = 1)
pnig(0.4, delta = 0, mu = 5, sigma = 1, nu = 1)
qnig(0.8, delta = 0, mu = 5, sigma = 1, nu = 1)
plot(function(x){dnig(x, delta = 0, mu = 5, sigma = 1, nu = 1)}, main =
"Normal inverse-Gaussian density", ylab = "Probability density",
xlim = c(0,10))
}
\references{
Barndorff-Nielsen, O. (1977) Exponentially decreasing distributions for the logarithm of particle size. Proceedings of the Royal Society of London.

 Series A, Mathematical and Physical Sciences. The Royal Society. 353, 401–409. \doi{10.1098/rspa.1977.0041}

 Barndorff-Nielsen, O. (1978) Hyperbolic Distributions and Distributions on Hyperbolae, Scandinavian Journal of Statistics. 5, 151–157.

 Barndorff-Nielsen, O. (1997) Normal Inverse Gaussian Distributions and Stochastic Volatility Modelling, Scandinavian Journal of Statistics. 24, 1-13. \doi{10.1111/1467-9469.00045}
}
\seealso{
\code{\link{dgig}}, \code{\link{dig}}, \code{\link{digam}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ngme.R
\name{print.ngme}
\alias{print.ngme}
\title{Print an ngme model}
\usage{
\method{print}{ngme}(x, ...)
}
\arguments{
\item{x}{ngme model object}

\item{...}{...}
}
\value{
Invisibly returns \code{x}, an \code{ngme} object. The method is
  called for its side effect of printing a model summary.
}
\description{
Print an ngme model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch-means.R
\name{ngme_sgld_samples}
\alias{ngme_sgld_samples}
\title{Extract Posterior-like Samples from Stored SGLD Trajectories}
\usage{
ngme_sgld_samples(
  ngme,
  name = "all",
  burnin_iter = 0,
  thinning = 1,
  apply_transform = TRUE,
  combine_chains = TRUE
)
}
\arguments{
\item{ngme}{fitted `ngme` object with `store_traj = TRUE`.}

\item{name}{parameter block to extract: `"all"` (default), latent model name,
or `"general"`.}

\item{burnin_iter}{non-negative integer. Number of initial iterations to
discard before sampling.}

\item{thinning}{positive integer thinning interval.}

\item{apply_transform}{logical; apply parameter transforms to user scale.}

\item{combine_chains}{logical; if `TRUE`, return one combined data.frame,
otherwise return one data.frame per chain.}
}
\value{
A data.frame (or list of data.frames when `combine_chains = FALSE`)
with columns `.chain`, `.draw`, `.iter`, and one column per parameter.
}
\description{
Build posterior-like samples from optimizer trajectories by dropping an
initial burn-in segment and applying thinning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.ngme.R
\name{print.noise_kld_comparison}
\alias{print.noise_kld_comparison}
\title{Print method for noise_kld_comparison}
\usage{
\method{print}{noise_kld_comparison}(x, ...)
}
\arguments{
\item{x}{noise_kld_comparison object}

\item{...}{additional arguments}
}
\value{
Invisibly returns \code{x}, a \code{noise_kld_comparison} object. The
  method is called for its side effect of printing the reference noise object,
  the Kullback-Leibler divergence values for each comparison noise, and the
  closest comparison.
}
\description{
Print method for noise_kld_comparison
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimizer.R
\name{rmsprop}
\alias{rmsprop}
\title{Root Mean Square Propagation (RMSProp) SGD optimization}
\usage{
rmsprop(stepsize = 0.05, beta1 = 0.9, epsilon = 1e-08)
}
\arguments{
\item{stepsize}{stepsize for SGD}

\item{beta1}{beta1 for momentum}

\item{epsilon}{epsilon for numerical stability}
}
\value{
a list of control variables for optimization
(used in \code{control_opt} function)
}
\description{
Root Mean Square Propagation (RMSProp) SGD optimization
}
\details{
The update rule for RMSProp is:
\deqn{v_t = \beta_1 v_{t-1} + (1 - \beta_1) g_t^2}
\deqn{x_{t+1} = x_t - \text{stepsize} * \frac{g_t}{\sqrt{v_t} + \epsilon}}
}
